{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_802dot11_radiotap;

interface

uses base_structures,
     class_parser_802dot11_radiotap,
     class_parser_tzsp,
     class_assembly_baseclass;

type t802dot11RadioTapAssembler = class (tFrameAssembler)

                                        private
                                               priv_Version               : byte;
                                               priv_HeaderFlags           : t802dot11HeaderFlagSet;

                                               priv_optTSFT               : uint64;
                                               priv_optFlags              : t802dot11OptionalFlagSet;
                                               priv_optDataRateMbps       : double;
                                               priv_optChannelFrequency   : word;
                                               priv_optChannelFlags       : t802dot11OptionalChannelTypeFlagSet;
                                               priv_FHSS                  : byte;
                                               priv_optdbmAntennaSignal   : shortint;
                                               priv_optdbmAntennaNoise    : shortint;
                                               priv_optSignalQuality      : word;
                                               priv_optAntennaIndex       : byte;
                                               priv_optdbAntennaSignal    : shortint;
                                               priv_optdbAntennaNoise     : shortint;

                                               procedure initialize;

                                               function OptionalFlagsToByte : byte;

                                               function getHeaderFlagsValue : longword;
                                               function getChannelFlagsValue : word;

                                               procedure writeRadioTapOptions (var Offset : integer);

                                               function ChannelTo24GhzFrequency (ChannelID : word) : word;

                                        protected
                                                 // called from base class from GetAssembly

                                                 procedure DetermineAssemblySize; override;
                                                 procedure Assemble; override;

                                        public
                                              property Version : byte read priv_version write priv_version;
                                              property HeaderLength : integer read prot_length;
                                              property HeaderFlags : t802dot11HeaderFlagSet read priv_headerflags write priv_headerflags;

                                              property TSFT : uint64 read priv_optTSFT write priv_optTSFT;
                                              property OptionalFlags : t802dot11OptionalFlagSet read priv_optflags write priv_optflags;
                                              property DataRateMBps : double read priv_optdatarateMbps write priv_optdatarateMbps;
                                              property ChannelFrequency : word read priv_optChannelFrequency write priv_optChannelFrequency;
                                              property ChannelFlags : t802dot11OptionalChannelTypeFlagSet read priv_optChannelFlags write priv_optChannelFlags;
                                              property FHSS : byte read priv_FHSS write priv_FHSS;
                                              property dbmAntennaSignal : shortint read priv_optdbmAntennaSignal write priv_optdbmAntennaSignal;
                                              property dbmAntennaNoise : shortint read priv_optdbmAntennaNoise write priv_optdbmAntennaNoise;
                                              property SignalQuality : word read priv_optSignalQuality write priv_optSignalQuality;
                                              property Antenna : byte read priv_optAntennaIndex write priv_optAntennaIndex;
                                              property dbAntennaSignal : shortint read priv_optdbAntennaSignal write priv_optdbAntennaSignal;
                                              property dbAntennaNoise : shortint read priv_optdbAntennaNoise write priv_optdbAntennaNoise;

                                              procedure CopyFromParser (Parser : t802dot11RadioTapParser); overload;
                                              procedure CopyFromParser (Parser : tTZSPParser); overload;
                                   end;


implementation

uses toolkit,
     common_parse_helpers;

(* ------------------------------------------- *)
{ t802dot11RadioTapAssembler }
(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.DetermineAssemblySize;

var flag : t802dot11HeaderFlag;

begin
     prot_length := sizeOf (t802dot11RadioTapHeader);

     for flag in priv_HeaderFlags do
         begin
              case flag of
               rtTSFT             : inc (prot_length, 8);
               rtFlags            : inc (prot_length);
               rtRate             : inc (prot_length);
               rtChannel          : inc (prot_length, 4);
               rtFHSS             : inc (prot_length);
               rtdBmAntennaSignal : inc (prot_length);
               rtdBmAntennaNoise  : inc (prot_length);
               rtLockQuality      : inc (prot_length, 2);
               rtTXAttenuation    : inc (prot_length, 2);
               rtdbTXAttenuation  : inc (prot_length, 2);
               rtdBmTXPower       : inc (prot_length);
               rtAntenna          : inc (prot_length);
               rtdbAntennaSignal  : inc (prot_length);
               rtdbAntennaNoise   : inc (prot_length);
               rtRXflags          : inc (prot_length, 2);
               rtChannelPlus      : ;
               rtHTinformation    : ;
               rtAMPDUStatus      : ;
               rtVHTinformation   : ;
               rtRadioNSTapNext   : ;
               rtVendorNSNext     : ;
               rtExt              : ;
              end;
         end;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapAssembler.getHeaderFlagsValue : longword;

var value : longword;
    flag  : t802dot11HeaderFlag;

begin
     value := 0;

     // this determines what the radiotap header contains, so
     // length adjustments are required depending on the flag set

     for flag in priv_HeaderFlags do
         begin
              case flag of
               rtTSFT             : value := value or (1 shl 0);
               rtFlags            : value := value or (1 shl 1);
               rtRate             : value := value or (1 shl 2);
               rtChannel          : value := value or (1 shl 3);
               rtFHSS             : value := value or (1 shl 4);
               rtdBmAntennaSignal : value := value or (1 shl 5);
               rtdBmAntennaNoise  : value := value or (1 shl 6);
               rtLockQuality      : value := value or (1 shl 7);
               rtTXAttenuation    : value := value or (1 shl 8);
               rtdbTXAttenuation  : value := value or (1 shl 9);
               rtdBmTXPower       : value := value or (1 shl 10);
               rtAntenna          : value := value or (1 shl 11);
               rtdbAntennaSignal  : value := value or (1 shl 12);
               rtdbAntennaNoise   : value := value or (1 shl 13);
               rtRXflags          : value := value or (1 shl 14);
               rtChannelPlus      : value := value or (1 shl 18);
               rtHTinformation    : value := value or (1 shl 19);
               rtAMPDUStatus      : value := value or (1 shl 20);
               rtVHTinformation   : value := value or (1 shl 21);
               rtRadioNSTapNext   : value := value or (1 shl 29);
               rtVendorNSNext     : value := value or (1 shl 30);
               rtExt              : value := value or (1 shl 31);
              end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapAssembler.getChannelFlagsValue : word;

var value : longword;
    flag  : t802dot11OptionalChannelTypeFlag;

begin
     value := 0;

     // this determines what the radiotap header contains, so
     // length adjustments are required depending on the flag set

     for flag in priv_optChannelFlags do
         begin
              case flag of
               rtcfTurbo       : value := value or (1 shl 4);
               rtcfCCK         : value := value or (1 shl 5);
               rtcfOFDM        : value := value or (1 shl 6);
               rtcf2GHZ        : value := value or (1 shl 7);
               rtcf5GHz        : value := value or (1 shl 8);
               rtcfPassive     : value := value or (1 shl 9);
               rtcfDynCCKOFDM  : value := value or (1 shl 10);
               rtcfGFSK        : value := value or (1 shl 11);
               rtcfGSM         : value := value or (1 shl 12);
               rtcfStaticTurbo : value := value or (1 shl 13);
               rtcfHalfRate    : value := value or (1 shl 14);
               rtcfQuarterRate : value := value or (1 shl 15);
              end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapAssembler.OptionalFlagsToByte : byte;

var flag  : t802dot11OptionalFlag;
    value : byte;

begin
     value := 0;

     for flag in priv_optFlags do
         begin
              case flag of
               rtofCFP           : value := value or (1 shl 0);
               rtofPreamble      : value := value or (1 shl 1);
               rtofWEP           : value := value or (1 shl 2);
               rtofFragmentation : value := value or (1 shl 3);
               rtofFCSAtEnd      : value := value or (1 shl 4);
               rtofDataPad       : value := value or (1 shl 5);
               rtofBadFCS        : value := value or (1 shl 6);
               rtofShortGI       : value := value or (1 shl 7);
              end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function t802dot11RadioTapAssembler.ChannelTo24GhzFrequency (ChannelID: word): word;

begin
     case ChannelID of
        1 : result := 2412;
        2 : result := 2417;
        3 : result := 2422;
        4 : result := 2427;
        5 : result := 2432;
        6 : result := 2437;
        7 : result := 2442;
        8 : result := 2447;
        9 : result := 2452;
       10 : result := 2457;
       11 : result := 2462;
       12 : result := 2467;
       13 : result := 2472;
       14 : result := 2484;
     end;
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.writeRadioTapOptions (var Offset: integer);

var bHelp : pByte;
    sHelp : pShortInt;
    wHelp : pWord;
    lHelp : pLongword;
    uHelp : pUInt64;

begin
     if rtTSFT in priv_HeaderFlags then
        begin
             // Value in microseconds of the MACs 64-bit 802.11 Time Synchronization Function timer

             uHelp := pUInt64 (prot_data + offset);
             uHelp^ := priv_optTSFT;

             inc (Offset, 8);
        end;

     if rtFlags in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_FLAGS, length 1 Byte

             bHelp := pByte (prot_data + offset);
             bHelp^ := OptionalFlagsToByte;

             inc (Offset);
        end;

     if rtRate in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_RATE, Length 1 Byte

             bHelp := pByte (prot_data + offset);
             bHelp^ := trunc ((priv_optdatarateMbps * 1000) / 500);

             inc (Offset);
        end;

     if rtChannel in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_CHANNEL and Flags, Length 2 Words

              wHelp := pWord (prot_data + offset);
              wHelp^ := priv_optChannelFrequency;

              inc (Offset, 2);

              wHelp := pWord (prot_data + offset);
              wHelp^ := getChannelFlagsValue;

              inc (Offset, 2);
        end;

     if rtFHSS in priv_HeaderFlags then
        begin
             bHelp := pByte (prot_data + offset);
             bHelp^ := priv_FHSS;

             inc (Offset);
        end;

     if rtdBmAntennaSignal in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_DBM_ANTSIGNAL (decibels from one milliwatt (dBm)), Length 1 Shortint

             sHelp := pShortInt (prot_data + offset);
             sHelp^ := priv_optdbmAntennaSignal;

             inc (Offset);
        end;

     if rtdBmAntennaNoise in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_ANTENNA_NOISE, Length 1 Shortint

             sHelp := pShortInt (prot_data + offset);
             sHelp^ := priv_optdbmAntennaNoise;

             inc (Offset);
        end;

     if rtLockQuality in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_LOCK_QUALITY, Length 1 Word

              wHelp := pWord (prot_data + offset);
              wHelp^ := priv_optSignalQuality;

              inc (Offset, 2);
        end;

     if rtdbTXAttenuation in priv_HeaderFlags then
        begin


        end;

     if rtTXAttenuation in priv_HeaderFlags then
        begin


        end;

     if rtdBmTXPower in priv_HeaderFlags then
        begin


        end;

     if rtAntenna in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_ANTENNA, Length 1 Byte

             bHelp := pByte (prot_data + offset);
             bHelp^ := priv_optAntennaIndex;

             inc (Offset);
        end;

     if rtdbAntennaSignal in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_DB_ANTSIGNAL (decibels from one milliwatt (dBm)), Length 1 Shortint

             sHelp := pShortInt (prot_data + offset);
             sHelp^ := priv_optdbAntennaSignal;

             inc (Offset);
        end;

     if rtdbAntennaNoise in priv_HeaderFlags then
        begin
             // IEEE80211_RADIOTAP_DB_ANTNOISE (decibels from one milliwatt (dBm)), Length 1 Shortint

             sHelp := pShortInt (prot_data + offset);
             sHelp^ := priv_optdbAntennaNoise;

             inc (Offset);
        end;

     if rtRXflags in priv_HeaderFlags then
        begin



        end;

     if rtChannelPlus in priv_HeaderFlags then
        begin


        end;

     if rtHTinformation in priv_HeaderFlags then
        begin


        end;

     if rtAMPDUStatus in priv_HeaderFlags then
        begin


        end;

     if rtVHTinformation in priv_HeaderFlags then
        begin


        end;

     if rtRadioNSTapNext in priv_HeaderFlags then
        begin


        end;

     if rtVendorNSNext in priv_HeaderFlags then
        begin


        end;

     if rtExt in priv_HeaderFlags then
        begin


        end;

     assert (Offset <= prot_length, 'Memory offset check mismatch (t802dot11ManagementAssembler.writeRadioTapOptions)');
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.initialize;

begin
     priv_Version := 0;
     priv_HeaderFlags := [];

     priv_optTSFT := 0;
     priv_optFlags := [];
     priv_optDataRateMbps := 0;
     priv_optChannelFrequency := 0;
     priv_optChannelFlags := [];
     priv_FHSS := 0;
     priv_optdbmAntennaSignal := 0;
     priv_optdbmAntennaNoise := 0;
     priv_optSignalQuality := 0;
     priv_optAntennaIndex := 0;
     priv_optdbAntennaSignal := 0;
     priv_optdbAntennaNoise := 0;
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.Assemble;

var pData  : p802dot11RadioTapHeader;
    offset : integer;

begin
     pData := p802dot11RadioTapHeader (prot_data);

     // Write static stuff

     pData^.HeaderVersion := priv_version;
     pData^.HeaderPadding := 0;
     pData^.HeaderLength := prot_length;
     pData^.HeaderFlags := getHeaderFlagsValue;

     offset := sizeOf (t802dot11RadioTapHeader);

     // Write Optional data

     writeRadioTapOptions (offset);
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.CopyFromParser (Parser: t802dot11RadioTapParser);

begin
     priv_Version := Parser.Version;
//   priv_HeaderLen := Parser.HeaderLength;         nope. determined by actual content
     priv_HeaderFlags := (Parser.HeaderFlags);

     // Optional data

     priv_optFlags := Parser.OptionalFlags;
     priv_optDataRateMbps := Parser.DataRateMBps;
     priv_optChannelFrequency := Parser.ChannelFrequency;
     priv_optChannelFlags := Parser.ChannelFlags;
     priv_optdbmAntennaSignal := Parser.dbmAntennaSignal;
     priv_optdbmAntennaNoise := Parser.dbmAntennaNoise;
     priv_optSignalQuality := Parser.SignalQuality;
     priv_optAntennaIndex := Parser.Antenna;
     priv_optdbAntennaSignal := Parser.dbAntennaSignal;
     priv_optdbAntennaNoise := Parser.dbAntennaNoise;
end;

(* ------------------------------------------- *)

procedure t802dot11RadioTapAssembler.CopyFromParser (Parser: tTZSPParser);

begin
     assert (Parser.EncapsulatedProtocol = NEXT_HEADER_802dot11Hdr, 'TZSP is carrying a WiFi frame (t802dot11RadioTapAssembler.CopyFromParser)');

     initialize;
     priv_Version := 0;

     // tzfRawRSSI, tzfSNR, tzfDataRate, tzfTimestamp, tzfContentionFree, tzfDecrypted, tzfFCSError, tzfRXChannel, tzfPacketCount, tzfRXFrameLength

     if tzfRawRSSI in Parser.TaggedFields then
        begin
             priv_HeaderFlags := priv_HeaderFlags + [rtdbAntennaSignal];

             priv_optdbAntennaSignal := Parser.RawRSSI;
        end;

     if tzfSNR in Parser.TaggedFields then
        begin
             priv_HeaderFlags := priv_HeaderFlags + [rtdbAntennaNoise];

             priv_optdbAntennaSignal := Parser.SNR;
        end;

     if tzfDataRate in Parser.TaggedFields then
        begin
             priv_HeaderFlags := priv_HeaderFlags + [rtRate];

             priv_optDataRateMbps := Parser.DataRateMBps;
        end;

     // TODO: determine what to do with tzfFCSError

     if tzfRXChannel in Parser.TaggedFields then
        begin
             priv_HeaderFlags := priv_HeaderFlags + [rtChannel];

             priv_optChannelFrequency := ChannelTo24GhzFrequency (Parser.RXChannel);
        end;


end;

(* ------------------------------------------- *)

end.
