{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_802dot1q;

interface

uses base_structures,
     class_parser_802dot1q,
     class_assembly_baseclass;


type t802dot1qAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_priority    : byte;
                                       priv_cfi         : boolean;
                                       priv_vlanid      : integer;
                                       priv_ethertype   : word;

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property Priority : byte read priv_priority write priv_priority;
                                      property CFI : boolean read priv_cfi write priv_cfi;
                                      property VLANID : integer read priv_vlanid write priv_vlanid;
                                      property EtherType : word read priv_ethertype write priv_ethertype;

                                      procedure CopyFromParser (Parser : t802dot1qParser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ t802dot1qAssembler }
(* ------------------------------------------- *)

constructor t802dot1qAssembler.create;

begin
     inherited;

     priv_priority := 0;
     priv_cfi := false;
     priv_vlanid := 0;
     priv_ethertype := 0;
end;

(* ------------------------------------------- *)

procedure t802dot1qAssembler.DetermineAssemblySize;

begin
     prot_length := 4;
end;

(* ------------------------------------------- *)

procedure t802dot1qAssembler.Assemble;

var pData : pWord;
    help  : word;

begin
     help := priv_priority;
     help := help shl 13;

     if priv_cfi then
        help := (help or $1000);

     help := help or priv_vlanid;

     pData := pWord (prot_data);
     pData^ := swap16 (help);

     pData := pWord (prot_data + 2);
     pData^ := swap16 (priv_ethertype);
end;

(* ------------------------------------------- *)

procedure t802dot1qAssembler.CopyFromParser (Parser: t802dot1qParser);

begin
     priv_priority := Parser.Priority;
     priv_cfi := Parser.CFI;
     priv_vlanid := Parser.VLANID;
     priv_ethertype := Parser.EtherType;
end;

(* ------------------------------------------- *)

end.
