{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_ayiya;

interface

uses base_structures,
     class_parser_ayiya,
     class_assembly_baseclass;


type tAyiyaAssembler = class (tFrameAssembler)

                             constructor create;

                             private
                                    priv_IDLen     : byte;
                                    priv_IDType    : tAYIYAIDType;
                                    priv_SigLen    : word;
                                    priv_HshMeth   : tAYIYAHashMethod;
                                    priv_AuthMeth  : tAYIYAAuthenticationMethod;
                                    priv_OpCode    : tAYIYAOperationCode;
                                    priv_NextHdr   : byte;
                                    priv_EpochTime : longword;
                                    priv_Identity  : tBufferBytes;
                                    priv_signature : tBufferBytes;

                                    procedure SetIdentity (Value : tBufferBytes);
                                    procedure SetSignature (Value : tBufferBytes);

                             protected
                                      // called from base class from GetAssembly

                                      procedure DetermineAssemblySize; override;
                                      procedure Assemble; override;

                             public
                                   property IDLength : byte read priv_IDLen;
                                   property IDType : tAYIYAIDType read priv_IDType write priv_IDType;
                                   property SignatureLength : word read priv_SigLen;
                                   property HashMethod : tAYIYAHashMethod read priv_HshMeth write priv_HshMeth;
                                   property AuthMethod : tAYIYAAuthenticationMethod read priv_AuthMeth write priv_AuthMeth;
                                   property OpCode : tAYIYAOperationCode read priv_OpCode write priv_OpCode;
                                   property NextHeader : byte read priv_NextHdr write priv_NextHdr;
                                   property EpochTime : longword read priv_EpochTime write priv_EpochTime;
                                   property Identity : tBufferBytes read priv_Identity write SetIdentity;
                                   property Signature : tBufferBytes read priv_signature write SetSignature;

                                   procedure CopyFromParser (Parser : tAYIYAParser);
                        end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tAyiyaAssembler }
(* ------------------------------------------- *)

constructor tAyiyaAssembler.create;

begin
     inherited;

     priv_IDLen := 0;
     priv_SigLen := 0;
end;

(* ------------------------------------------- *)

procedure tAyiyaAssembler.DetermineAssemblySize;

begin
     prot_length := sizeOf (tAYIYAStruct) + priv_SigLen + priv_IDLen;
end;

(* ------------------------------------------- *)

procedure tAyiyaAssembler.SetIdentity (Value: tBufferBytes);

begin
     priv_IDLen := length (Value);
     priv_Identity := copy (Value);
end;

(* ------------------------------------------- *)

procedure tAyiyaAssembler.SetSignature (Value: tBufferBytes);

begin
     priv_SigLen := length (Value);
     priv_signature := Copy (Value);
end;

(* ------------------------------------------- *)

procedure tAyiyaAssembler.Assemble;

var pData  : pAYIYAStruct;
    offset : integer;
    pHelp  : pByte;
    hByte  : byte;
    i      : integer;

begin
     pData := pAYIYAStruct (prot_data);

     // Len and Type (each a nibble)

     hByte := ((priv_IDLen div 4) shl 4);

     case priv_IDType of
      aitInteger : hByte := hByte or 1;
      aitASCII   : hByte := hByte or 2;
     end;

     pData^.IDLenType := hByte;

     // Sig Len and Hash Method

     hByte := ((priv_SigLen div 4) shl 4);

     case priv_HshMeth of
      ahmMD5  : hByte := hByte or 1;
      ahmSHA1 : hByte := hByte or 2;
     end;

     pData^.SigLenHshMeth := hByte;

     // Auth Method and OpCode

     case priv_AuthMeth of
      aamSharedSecret : hByte := 16;
      aamPubKey       : hByte := 32;
     end;

     case priv_opcode of
      aocForward       : hByte := hByte or 1;
      aocEchoReq       : hByte := hByte or 2;
      aocEchoReqAndFwd : hByte := hByte or 3;
      aocEchoResp      : hByte := hByte or 4;
      aocMOTD          : hByte := hByte or 5;
      aocQueryReq      : hByte := hByte or 6;
      aocQueryResp     : hByte := hByte or 7;
     end;

     pData^.AuthMethOpCode := hByte;

     // Next Header

     pData^.NextHeader := priv_nexthdr;

     // Epoch Time

     pData^.EpochTime := Swap32 (priv_EpochTime);

     // Identity

     offset := sizeOf (tAYIYAStruct);

     pHelp := pByte (prot_data + offset);

     for i := 0 to (length (priv_Identity) - 1) do
         begin
              pHelp^ := priv_Identity[i];
              inc (pHelp);
              inc (offset);
         end;

     // Signature

     for i := 0 to (length (priv_signature) - 1) do
         begin
              pHelp^ := priv_Signature[i];
              inc (pHelp);
              inc (offset);
         end;

     assert (offset = prot_length, 'Memory offset check mismatch (tAyiyaAssembler.Assemble)');
end;


(* ------------------------------------------- *)

procedure tAyiyaAssembler.CopyFromParser(Parser: tAYIYAParser);

begin
     IDType := Parser.IDType;
     HashMethod := Parser.HashMethod;
     AuthMethod := Parser.AuthMethod;
     OpCode := Parser.OpCode;
     NextHeader := Parser.NextAnchor.HeaderType;
     EpochTime := Parser.EpochTime;
     Identity := Parser.Identity;
     Signature := Parser.Signature;
end;

(* ------------------------------------------- *)

end.
