{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_dhcpv4;

interface

uses base_structures,
     definition_ip_addresses,
     class_parser_dhcpv4,
     class_assembly_baseclass;

type tDHCPv4Assembler  = class (tFrameAssembler)

                               constructor create;

                                      private
                                             priv_Messagetype         : tDHCPv4MessageOpCode;
                                             priv_HardwareAddrType    : byte;
                                             priv_HardwareAddrLength  : byte;
                                             priv_Hops                : byte;
                                             priv_TransactionID       : longword;
                                             priv_Seconds             : word;
                                             priv_Flags               : tDHCPv4Flag;
                                             priv_ClientIPAddress     : tIPv4Address;
                                             priv_YourClientIPAddress : tIPv4Address;
                                             priv_ServerIPAddress     : tIPv4Address;
                                             priv_RelayIPAddress      : tIPv4Address;
                                             priv_ClientHWAddress     : tMACAddress;
                                             priv_ClientHWPadding     : tDHCPv4MACPAdding;
                                             priv_ServerHostName      : string;
                                             priv_BootFileName        : string;
                                             priv_MagicCookie         : longword;

                                             priv_dhcpoptions         : tDHCPv4OptionList;
                                             priv_optionlen           : integer;
                                             priv_padsize             : integer;

                                             procedure writeDHCPOptions (Offset : integer);

                                      protected
                                               // called from base class from GetAssembly

                                               procedure Assemble; override;
                                               procedure DetermineAssemblySize; override;

                                      public
                                            property MessageOpCode : tDHCPv4MessageOpCode read priv_Messagetype write priv_Messagetype;
                                            property HardwareAddressType : byte read priv_HardwareAddrType write priv_HardwareAddrType;
                                            property HardwareAddressLength : byte read priv_HardwareAddrLength write priv_HardwareAddrLength;
                                            property Hops : byte read priv_Hops write priv_Hops;
                                            property TransactionID : longword read priv_TransactionID write priv_TransactionID;
                                            property Seconds : word read priv_Seconds write priv_Seconds;
                                            property Flags : tDHCPv4Flag read priv_Flags write priv_Flags;
                                            property ClientIPAddress : tIPv4Address read priv_ClientIPAddress write priv_ClientIPAddress;
                                            property YourClientIPAddress : tIPv4Address read priv_YourClientIPAddress write priv_YourClientIPAddress;
                                            property ServerIPAddress : tIPv4Address read priv_ServerIPAddress write priv_ServerIPAddress;
                                            property RelayIPAddress : tIPv4Address read priv_RelayIPAddress write priv_RelayIPAddress;
                                            property ClientHWAddress : tMACAddress read priv_ClientHWAddress write priv_ClientHWAddress;
                                            property ClientHWPadding : tDHCPv4MACPAdding read priv_ClientHWPadding write priv_ClientHWPadding;
                                            property ServerHostName : string read priv_ServerHostName write priv_ServerHostName;
                                            property BootFileName : string read priv_BootFileName write priv_BootFileName;
                                            property MagicCookie : longword read priv_MagicCookie write priv_MagicCookie;

                                            property OptionList : tDHCPv4OptionList read priv_dhcpoptions write priv_dhcpoptions;
                                            property PaddingSize : integer read priv_padsize write priv_padsize;

                                            procedure AddDHCPOption (Option : tDHCPv4OptionHdr);
                                            procedure CopyFromParser (Parser : tDHCPv4Parser);
                                end;


implementation

uses Toolkit,
     common_assembly_helpers;

type tDHCPOptHdr = packed record
                                code : byte;
                                len  : byte;
                   end;
     pDHCPOptHdr = ^tDHCPOptHdr;

(* ------------------------------------------- *)
{ tDHCPv4Assembler }
(* ------------------------------------------- *)

constructor tDHCPv4Assembler.create;

begin
     inherited create;

     setLength (priv_dhcpoptions, 0);

     priv_padsize := 0;
end;

(* ------------------------------------------- *)

procedure tDHCPv4Assembler.DetermineAssemblySize;

var Option : tDHCPv4OptionHdr;
    i      : integer;

begin
     prot_length := sizeOf (tStruct_DHCPv4) + priv_padsize;

     for i := 1 to length (priv_dhcpoptions) do
         begin
              option := priv_dhcpoptions[i-1];

              if (Option.Code > DHCP_OPT_Pad) and (Option.Code < DHCP_OPT_End) then
                 begin
                      prot_length := prot_length + 2 + Option.Length;                             // 2 for the code/len bytes
                      priv_optionlen := priv_optionlen + 2 + Option.Length;
                 end
              else
                  begin
                       inc (prot_length);
                       inc (priv_optionlen);
                  end;
         end;
end;

(* ------------------------------------------- *)

procedure tDHCPv4Assembler.writeDHCPOptions (Offset : integer);

var LocalOffset : integer;
    opt         : tDHCPv4OptionHdr;
    hdr         : pDHCPOptHdr;
    help8       : pByte;
    help16      : pWord;
    help32      : pLongword;
    helpMAC     : pMACBytes;
    helpStr     : pANSIString;
    helpstring  : string;
    helparray   : tByteArray;
    i,j, k      : integer;
    modifier    : integer;

begin
     LocalOffset := Offset;

     for i := 1 to length (priv_dhcpoptions) do
         begin
              opt := priv_dhcpoptions[i-1];

              // write Header

              if (opt.Code > DHCP_OPT_Pad) and (opt.Code < DHCP_OPT_End) then
                 begin
                      hdr := pDHCPOptHdr (prot_data + LocalOffset);

                      hdr^.code := opt.Code;
                      hdr^.len := opt.Length;

                      inc (LocalOffset, 2);
                 end
              else
                  begin
                       help8 := pByte (prot_data + LocalOffset);

                       help8^ := opt.Code;
                  end;

              // Write option value

              case opt.Code of
               DHCP_OPT_Subnet_Mask                                 : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := swap32 (opt.addresslist[1].ValueAsLong);
                                                                      end;
               DHCP_OPT_Time_Offset                                 : begin
                                                                      end;
               DHCP_OPT_Router                                      : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := swap32 (opt.addresslist[1].ValueAsLong);
                                                                      end;
               DHCP_OPT_Time_Server                                 : begin
                                                                           for j := 1 to (opt.Length DIV 4) do
                                                                               begin
                                                                                    help32 := pLongword (prot_data + LocalOffset + ((j - 1) * 4));
                                                                                    help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                               end;
                                                                      end;
               DHCP_OPT_Name_Server                                 : begin
                                                                           for j := 1 to (opt.Length DIV 4) do
                                                                               begin
                                                                                    help32 := pLongword (prot_data + LocalOffset + ((j - 1) * 4));
                                                                                    help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                               end;
                                                                      end;
               DHCP_OPT_Domain_Name_Server                          : begin
                                                                           for j := 1 to (opt.Length DIV 4) do
                                                                               begin
                                                                                    help32 := pLongword (prot_data + LocalOffset + ((j - 1) * 4));
                                                                                    help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                               end;
                                                                      end;
               DHCP_OPT_Host_Name                                   : begin
                                                                           WriteStringToOffset (prot_data, opt.FQDNList[1].FullFQDN, LocalOffset);
{
                                                                           helpstring := opt.FQDNList[1].FullFQDN;

                                                                            for j := 0 to length (helpstring) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j);

                                                                                    help8^ := ord (helpstring[j+1]);
                                                                               end;
}
                                                                      end;
               DHCP_OPT_Domain_Name                                 : begin
                                                                           WriteStringToOffset (prot_data, opt.FQDNList[1].FullFQDN, LocalOffset);
{
                                                                           helpstring := opt.FQDNList[1].FullFQDN;

                                                                           for j := 0 to length (helpstring) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j);

                                                                                    help8^ := ord (helpstring[j+1]);
                                                                               end;
}
                                                                      end;
               DHCP_OPT_Broadcast_Address                           : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := swap32 (opt.addresslist[1].ValueAsLong);
                                                                      end;
               DHCP_OPT_TCP_keepalive_interval                      : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := Swap32 (opt.numberlongword);
                                                                      end;
               DHCP_OPT_NTP_servers                                 : begin
                                                                           for j := 1 to (opt.Length DIV 4) do
                                                                               begin
                                                                                    help32 := pLongword (prot_data + LocalOffset + ((j - 1) * 4));
                                                                                    help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                               end;
                                                                      end;
               DHCP_OPT_NetBIOS_over_TCPIP_nameserver               : begin
                                                                           for j := 1 to (opt.Length DIV 4) do
                                                                               begin
                                                                                    help32 := pLongword (prot_data + LocalOffset + ((j - 1) * 4));
                                                                                    help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                               end;
                                                                      end;
               DHCP_OPT_NetBIOS_over_TCPIP_Node_Type                : begin
                                                                           help8 := pByte (prot_data + LocalOffset);

                                                                           help8^ := opt.numberbyte;
                                                                      end;
               DHCP_OPT_Requested_IP_Address                        : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := swap32 (opt.addresslist[1].ValueAsLong);
                                                                      end;
               DHCP_OPT_IP_address_lease_time                       : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := Swap32 (opt.numberlongword);
                                                                      end;
               DHCP_OPT_Option_overload                             : begin

                                                                      end;
               DHCP_OPT_Message                                     : begin
                                                                           for j := 0 to length (opt.text) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j);

                                                                                    help8^ := ord (opt.text[j+1]);
                                                                               end;
                                                                      end;
               DHCP_OPT_Maximum_DHCP_message_size                   : begin
                                                                           help16 := pWord (prot_data + LocalOffset);
                                                                           help16^ := Swap16 (opt.numberword);
                                                                      end;
               DHCP_OPT_DHCP_message_type                           : begin
                                                                           help8 := pByte (prot_data + LocalOffset);

                                                                           help8^ := opt.numberbyte;
                                                                      end;
               DHCP_OPT_Server_identifier                           : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := swap32 (opt.addresslist[1].ValueAsLong);
                                                                      end;
               DHCP_OPT_Parameter_request_list                      : begin
                                                                           for j := 0 to length (opt.ParamReqList) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j);
                                                                                    help8^ := opt.ParamReqList[j];
                                                                               end;
                                                                      end;
               DHCP_OPT_Renew_time_value                            : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := Swap32 (opt.numberlongword);
                                                                      end;
               DHCP_OPT_Rebinding_time_value                        : begin
                                                                           help32 := pLongword (prot_data + LocalOffset);
                                                                           help32^ := Swap32 (opt.numberlongword);
                                                                      end;
               DHCP_OPT_ClassIdentifier                             : begin
                                                                           for j := 0 to length (opt.text) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j);

                                                                                    help8^ := ord (opt.text[j+1]);
                                                                               end;
                                                                      end;
               DHCP_OPT_ClientIdentifier                            : begin
                                                                           help8 := pByte (prot_data + LocalOffset);
                                                                           help8^ := byte (opt.numberbyte);

                                                                           helpMAC := pMACBytes (prot_data + LocalOffset + 1);
                                                                           helpMAC^ := opt.MACAddress.RangeBytes;
                                                                      end;
               DHCP_OPT_FQDN_Fully_Qualified_Domain_Name            : begin
                                                                           // Write the 3 bytes containing flags, A-RR, PTR-RR from

                                                                           help8 := pByte (prot_data + LocalOffset);
                                                                           help8^ := opt.Value[0];

                                                                           help8 := pByte (prot_data + LocalOffset + 1);
                                                                           help8^ := opt.Value[1];

                                                                           help8 := pByte (prot_data + LocalOffset + 2);
                                                                           help8^ := opt.Value[2];

                                                                           for j := 0 to length (opt.text) - 1 do
                                                                               begin
                                                                                    help8 := pByte (prot_data + LocalOffset + j + 3);

                                                                                    help8^ := ord (opt.text[j+1]);
                                                                               end;
                                                                      end;
               DHCP_OPT_AutoConfigure                               : begin
                                                                           help8 := pByte (prot_data + LocalOffset);

                                                                           help8^ := opt.numberbyte;
                                                                      end;

               DHCP_OPT_SIP_Servers_DHCP_Option                     : begin
                                                                           case opt.SIPSrvEncoding of
                                                                            dssFQDNList : begin
                                                                                               help8 := pByte (prot_data + LocalOffset);
                                                                                               help8^ := 0;

                                                                                               modifier := 1;

                                                                                               for j := 1 to opt.FQDNCount do
                                                                                                   begin
                                                                                                        helparray := opt.FQDNList[j].getAsFrameBufferArray;

                                                                                                        for k := 0 to length (helparray) - 1 do
                                                                                                            begin
                                                                                                                help8 := pByte (prot_data + LocalOffset + modifier + k);
                                                                                                                help8^ := helparray[k];
                                                                                                            end;

                                                                                                        // Move the modifier to continue writing after the current FQDN

                                                                                                        inc (modifier, length (helparray));
                                                                                                   end;
                                                                                          end;
                                                                            dssIPList   : begin
                                                                                               help8 := pByte (prot_data + LocalOffset);
                                                                                               help8^ := 1;

                                                                                               for j := 1 to opt.addresscount do
                                                                                                   begin
                                                                                                        help32 := pLongword (prot_data + LocalOffset + 1 + ((j - 1) * 4));
                                                                                                        help32^ := swap32 (opt.addresslist[j].ValueAsLong);
                                                                                                   end;
                                                                                          end;
                                                                           end;
                                                                      end
              else
                  begin
                       // TODO: diag message when we cannot assemble the option

                       ;
                  end;
              end;

              inc (LocalOffset, opt.Length);
         end;
end;

(* ------------------------------------------- *)

procedure tDHCPv4Assembler.AddDHCPOption (Option: tDHCPv4OptionHdr);

var count : integer;

begin
     count := length (priv_dhcpoptions);
     inc (count);
     setLength (priv_dhcpoptions, count);

     priv_dhcpoptions[count-1] := Option;
end;

(* ------------------------------------------- *)

procedure tDHCPv4Assembler.Assemble;

var pData  : pStruct_DHCPv4;
    Help8  : pByte;
    Offset : integer;
    i      : integer;

begin
     pData := pStruct_DHCPv4 (prot_data);

     case priv_Messagetype of
      docBootRequest : pData^.MessageType := 1;
      docBootReply   : pData^.MessageType := 2;
     end;

     pData^.HardwareAddrType := priv_HardwareAddrType;
     pData^.HardwareAddrLength := priv_HardwareAddrLength;
     pData^.Hops := priv_Hops;
     pData^.TransactionID := priv_TransactionID;
     pData^.Seconds := priv_Seconds;

     case priv_flags of
      dfBroadcast : pData^.Flags := $8000;
      dfUnicast   : pData^.Flags := 0;
     end;

     pData^.ClientIPAddress := swap32 (priv_ClientIPAddress.ValueAsLong);
     pData^.YourClientIPAddress := swap32 (priv_YourClientIPAddress.ValueAsLong);
     pData^.ServerIPAddress := Swap32 (priv_ServerIPAddress.ValueAsLong);
     pData^.RelayIPAddress := Swap32 (priv_RelayIPAddress.ValueAsLong);
     pData^.ClientHWAddress := priv_ClientHWAddress.RangeBytes;
     pData^.ServerHostName := priv_ServerHostName;
     pData^.BootFileName := priv_BootFileName;
     pData^.MagicCookie := Swap32 (priv_MagicCookie);

     if length (priv_dhcpoptions) > 0 then
        writeDHCPOptions (sizeOf (tStruct_DHCPv4));

     // Handle padding

     Offset := sizeOf (tStruct_DHCPv4) + priv_optionlen;

     for i := 1 to priv_padsize do
         begin
              Help8 := pByte (prot_data + Offset);
              Help8^ := 0;

              inc (Offset);
         end;

     assert (offset = prot_length, 'Memory offset check mismatch (tDHCPv4Assembler.Assemble)');
end;

(* ------------------------------------------- *)

procedure tDHCPv4Assembler.CopyFromParser (Parser: tDHCPv4Parser);

var opt : tDHCPv4OptionHdr;
    i   : integer;

begin
     priv_Messagetype := Parser.MessageOpCode;
     priv_HardwareAddrType := Parser.HardwareAddressType;
     priv_HardwareAddrLength := Parser.HardwareAddressLength;
     priv_Hops := Parser.Hops;
     priv_TransactionID := Parser.TransactionID;
     priv_Seconds := Parser.Seconds;
     priv_Flags := Parser.Flags;
     priv_ClientIPAddress := Parser.ClientIPAddress;
     priv_YourClientIPAddress := Parser.YourClientIPAddress;
     priv_ServerIPAddress := Parser.ServerIPAddress;
     priv_RelayIPAddress := Parser.RelayIPAddress;
     priv_ClientHWAddress := Parser.ClientHWAddress;
     priv_ClientHWPadding := Parser.ClientHWPadding;
     priv_ServerHostName := Parser.ServerHostName;
     priv_BootFileName := Parser.BootFileName;
     priv_MagicCookie := Parser.MagicCookie;
     priv_padsize :=  Parser.PaddingSize;

     for i := 1 to length (Parser.OptionList) do
         begin
              opt := Parser.OptionList[i-1];

              AddDHCPOption (opt);
         end;
end;

(* ------------------------------------------- *)

end.


