{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_icmpv6;

interface

uses base_structures,
     definition_ip_addresses,
     base_checksums,
     class_parser_icmpv6,
     class_assembly_baseclass;


type tICMPv6Assembler = class (tFrameAssembler)

                                constructor create (IPv6PseudoHeader : tIPv6PseudoHeader); overload;
                                constructor create (IPv6PseudoHeader : tIPv6PseudoHeader; QuotedPayload: tFrameChunk); overload;

                                private
                                       priv_type                    : byte;
                                       priv_code                    : byte;
                                       priv_checksum                : word;

                                       priv_payload                 : tRawBuffer;
                                       priv_payloadlength           : integer;

                                       priv_echo_identifier         : word;
                                       priv_echo_sequence           : word;
                                       priv_next_hop_mtu            : longword;
                                       priv_problem_offset          : longword;
                                       priv_source_linklayer_addr   : tMACAddress;
                                       priv_target_linklayer_addr   : tMACAddress;
                                       priv_nd_target_address       : tIPv6Address;
                                       priv_redirect_target_address : tIPv6Address;
                                       priv_redirect_dest_address   : tIPv6Address;
                                       priv_router_hoplimit         : byte;
                                       priv_routeradv_flags         : tICMPv6RtrAdvtFlagSet;
                                       priv_router_preference       : tICMPv6RtrPreference;
                                       priv_router_lifetime         : word;
                                       priv_router_reachtime        : longword;
                                       priv_router_retranstime      : longword;
                                       priv_router_mtu              : longword;
                                       priv_neighboradv_flags       : tICMPv6NDAdvFlagSet;
                                       priv_keepcrc                 : boolean;
                                       priv_force_badcrc            : boolean;

                                       priv_optionlist              : tICMPv6NDOptionList;

                                       priv_IPv6PseudoHeader        : tIPv6PseudoHeader;

                                       procedure initialize;

//                                     procedure setICMPType (value : byte);
                                       procedure setCheckSum (Value : word);

                                       procedure WriteOptionList (var Offset : integer);

                                       {$REGION 'Documentation'}
                                       /// <summary>
                                       ///   Returns the first option that has
                                       ///   a matching OptionType. Returns -1
                                       ///   if nothing was found
                                       /// </summary>
                                       /// <param name="OptionType">
                                       ///   Option Type to look for
                                       /// </param>
                                       /// <returns>
                                       ///   index of the option record in the
                                       ///   option list
                                       /// </returns>
                                       {$ENDREGION}
                                       function FindOptionListIndexByOptionType (OptionType : integer) : integer;

                                       procedure SetSourceLinkLayerAddress (Address : tMACAddress);
                                       procedure SetTargetLinkLayerAddress (Address : tMACAddress);

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property ICMPType : byte read priv_type write priv_type;
                                      property ICMPCode : byte read priv_code write priv_code;
                                      property Checksum : word read priv_checksum write setChecksum;
                                      property EchoIdentifier : word read priv_echo_identifier write priv_echo_identifier;
                                      property EchoSequence : word read priv_echo_sequence write priv_echo_sequence;
                                      property NextHopMTU : longword read priv_next_hop_mtu write priv_next_hop_mtu;
                                      property ParameterProblemOffset : longword read priv_problem_offset write priv_problem_offset;
                                      property SourceLinkLayerAddress : tMACAddress read priv_source_linklayer_addr write SetSourceLinkLayerAddress;
                                      property TargetLinkLayerAddress : tMACAddress read priv_target_linklayer_addr write SetTargetLinkLayerAddress;
                                      property NDTargetAddress : tIPv6Address read priv_nd_target_address write priv_nd_target_address;
                                      property RedirectTargetAddress : tIPv6Address read priv_redirect_target_address write priv_redirect_target_address;
                                      property RedirectDestinationAddress : tIPv6Address read priv_redirect_dest_address write priv_redirect_dest_address;
                                      property RAHopLimit : byte read priv_router_hoplimit write priv_router_hoplimit;
                                      property RAFlags : tICMPv6RtrAdvtFlagSet read priv_routeradv_flags write priv_routeradv_flags;
                                      property RAPreference : tICMPv6RtrPreference read priv_router_preference write priv_router_preference;
                                      property RALifeTime : word read priv_router_lifetime write priv_router_lifetime;
                                      property RAReachableTime : longword read priv_router_reachtime write priv_router_reachtime;
                                      property RARetransTimer : longword read priv_router_retranstime write priv_router_retranstime;
                                      property RAMTU : longword read priv_router_mtu write priv_router_mtu;
                                      property NeighborAdvertisementFlags : tICMPv6NDAdvFlagSet read priv_neighboradv_flags write priv_neighboradv_flags;
                                      property ForceBadCRC : boolean read priv_force_badcrc write priv_force_badcrc;
                                      property OptionList : tICMPv6NDOptionList read priv_optionlist write priv_optionlist;
                                      property Payload : tRawBuffer read priv_payload;
                                      property PayloadLength : integer read priv_payloadlength;

                                      procedure setPayloadFromRawbuffer (buffer : tRawBuffer; Len : integer);
                                      procedure setPayloadFromFrameChunk (Chunk : tFrameChunk);
                                      procedure clearPayload;

                                      procedure AddOption (Option : tICMPv6NDOption);
                                      procedure CopyFromParser (Parser : tICMPv6Parser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tICMPv6Assembler }
(* ------------------------------------------- *)

constructor tICMPv6Assembler.create (IPv6PseudoHeader : tIPv6PseudoHeader);

begin
     inherited create;

     priv_IPv6PseudoHeader := IPv6PseudoHeader;

     priv_force_badcrc := false;
     prot_length := 0;

     initialize;

     priv_payloadlength := 0;
end;

(* ------------------------------------------- *)

constructor tICMPv6Assembler.create (IPv6PseudoHeader: tIPv6PseudoHeader; QuotedPayload: tFrameChunk);

begin
     inherited create;

     priv_IPv6PseudoHeader := IPv6PseudoHeader;

     priv_force_badcrc := false;
     prot_length := 0;

     initialize;

     priv_payload := QuotedPayload.getAsRawBuffer;
     priv_payloadlength := QuotedPayload.Len;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.DetermineAssemblySize;

var i : integer;

begin
     case priv_type of
      128,                             // Echo Request/Reply
      129 : prot_length := 8 + priv_payloadlength;
      134 : prot_length := 16;         // Router Advertisement
      135 : prot_length := 24;         // Neighbor Solicitation
      136 : prot_length := 24;         // Neighbor Advertisement
      137 : prot_length := 40;         // Redirect (4 + 4 + 16 + 16)
     else
         prot_length := sizeOf (tICMPv6Struct) + 4;
     end;

     // Adjust the length of the protocol block to make sure we allocate enough bytes

     for i := 1 to length (OptionList) do
         begin
              case OptionList[i-1].option_type of
               4 : inc (prot_length, 8);           // Redirect option is 8 bytes (plus quote, which we ignore as it is handled in other assemblers)
              else
                  inc (prot_length, OptionList[i-1].option_len);
              end;
         end;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.initialize;

begin
     priv_echo_identifier := 0;
     priv_echo_sequence := 0;
     priv_next_hop_mtu := 0;
     priv_nd_target_address.initialize;
     priv_redirect_target_address.initialize;
     priv_redirect_dest_address.initialize;
     priv_router_hoplimit := 0;
     priv_routeradv_flags := [];
     priv_router_preference := rpNone;
     priv_router_lifetime := 0;
     priv_router_reachtime := 0;
     priv_router_retranstime := 0;
     priv_router_mtu := 0;
     setlength (priv_optionlist, 0);
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.setCheckSum(Value: word);

begin
     priv_checksum := value;
     priv_keepcrc := true;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.clearPayload;

begin
     InitializeRawBuffer (priv_payload);
     priv_payloadlength := 0;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.setPayloadFromFrameChunk (Chunk: tFrameChunk);

begin
     move (Chunk.data^, priv_payload, Chunk.Len);
     priv_payloadlength := Chunk.Len;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.setPayloadFromRawbuffer (buffer: tRawBuffer; Len: integer);

begin
     priv_payload := buffer;
     priv_payloadlength := Len;
end;

(* ------------------------------------------- *)

function tICMPv6Assembler.FindOptionListIndexByOptionType (OptionType: integer): integer;

var len : integer;
    i   : integer;

begin
     i := 0;
     len := length (priv_optionlist);

     // Simple search loop

     while (i < len) and (priv_optionlist[i].option_type <> OptionType) do
           inc (i);

     if (i < len) and (priv_optionlist[i].option_type = OptionType) then
        result := i
     else
         result := -1;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.AddOption (Option: tICMPv6NDOption);

var i : integer;

begin
     i := length (priv_optionlist);
     inc (i);
     setlength (priv_optionlist, i);

     priv_optionlist[i-1] := Option;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.SetSourceLinkLayerAddress (Address: tMACAddress);

var index : integer;
    opt   : tICMPv6NDOption;

begin
     priv_source_linklayer_addr := Address;

     index := FindOptionListIndexByOptionType (1);

     if index > -1 then
        priv_optionlist[index].MACAddress := priv_source_linklayer_addr
     else
         begin
              opt.initialize;
              opt.option_type := 1;
              opt.option_len := 1;        // = 8 Bytes
              opt.MACAddress := priv_source_linklayer_addr;

              AddOption (opt);
         end;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.SetTargetLinkLayerAddress (Address: tMACAddress);

var index : integer;
    opt   : tICMPv6NDOption;

begin
     priv_target_linklayer_addr := Address;

     index := FindOptionListIndexByOptionType (2);

     if index > -1 then
        priv_optionlist[index].MACAddress := priv_target_linklayer_addr
     else
         begin
              opt.initialize;
              opt.option_type := 2;
              opt.option_len := 1;        // = 8 Bytes
              opt.MACAddress := priv_target_linklayer_addr;

              AddOption (opt);
         end;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.WriteOptionList (var Offset: integer);

var option   : tICMPv6NDOption;
    bHelp    : byte;
    pbHelp   : pByte;
    pwHelp   : pWord;
    plHelp   : pLongword;
    pMACAddr : pMACBytes;
    pIPAddr  : pIPv6AddressByteStruct;
    FQDNbuf  : tByteArray;
    i,j,k    : integer;
    count    : integer;
    len      : integer;

begin
     for i := 1 to length (priv_optionlist) do
         begin
              option := priv_optionlist[i-1];

              // Write type / len

              pbHelp := pByte (prot_data + Offset);
              pbHelp^ := option.option_type;

              pbHelp := pByte (prot_data + Offset + 1);

              if option.option_type = 4 then
                 begin
                      // for redirects, the option length includes the quote following the header

                      len := 8 + priv_payloadlength;
                      pbHelp^ := (len div 8);
                 end
              else
                  pbHelp^ := (option.option_len div 8);

              // Write actual data

              case option.option_type of
               1,2 : begin
                          // Source/Target Link Layer address

                          pMACAddr := pMACBytes (prot_data + Offset + 2);
                          pMACAddr^ := option.MACAddress.RangeBytes;
                     end;
                 3 : begin
                          // Prefix Information

                          // Length

                          pbHelp := pByte (prot_data + Offset + 2);
                          pbHelp^ := option.PrefixLength;

                          // Flags

                          if poOnLink in option.PrefixFlags then
                             bHelp := $80
                          else
                              bHelp := 0;

                          if poAutonomous in option.PrefixFlags then
                             bHelp := bHelp or $40;

                          if poRouterAddress in option.PrefixFlags then
                             bHelp := bHelp or $20;

                          bHelp := bHelp or option.PrefixReserved;

                          // Lifetimes

                          pbHelp := pByte (prot_data + Offset + 3);
                          pbHelp^ := bHelp;

                          plHelp := pLongword (prot_data + Offset + 4);
                          plHelp^ := Swap32 (option.PrefixValidLifetime);

                          plHelp := pLongword (prot_data + Offset + 8);
                          plHelp^ := Swap32 (option.PrefixPrefLifetime);

                          // Reserved

                          plHelp := pLongword (prot_data + Offset + 12);
                          plHelp^ := 0;

                          // Prefix Address

                          pIPAddr := pIPv6AddressByteStruct (prot_data + Offset + 16);
                          pIPAddr^ := option.PrefixAddress.ValueAsByteStruct;
                     end;
                 4 : begin
                          // Redirect, clear 6 reserved bytes

                          plHelp := pLongword (prot_data + Offset + 2);
                          plHelp^ := 0;

                          pwHelp := pword (prot_data + Offset + 6);
                          pwHelp^ := 0;
                     end;
                 5 : begin
                          // MTU

                          pwHelp := pWord (prot_data + Offset + 2);
                          pwHelp^ := 0;

                          plHelp := pLongword (prot_data + Offset + 4);
                          plHelp^ := Swap32 (option.MTU);
                     end;
                25 : begin
                          // Recursive DNS option

                          pwHelp := pWord (prot_data + Offset + 2);
                          pwHelp^ := 0;

                          plHelp := pLongword (prot_data + Offset + 4);
                          plHelp^ := swap32 (option.DNSLifeTime);

                          for j := 0 to length (option.DNSServerList) - 1 do
                              begin
                                   pIPAddr := pIPv6AddressByteStruct (prot_data + Offset + 8 + (j * 16));
                                   pIPAddr^ := option.DNSServerList[j].ValueAsByteStruct;
                              end;
                     end;
                31 : begin
                          // DNS Search List Option

                          pwHelp := pWord (prot_data + Offset + 2);
                          pwHelp^ := 0;

                          plHelp := pLongWord (prot_data + Offset + 4);
                          plHelp^ := swap32 (option.DNSLifeTime);

                          count := 0;

                          for j := 0 to length (option.DNSDomainNameList) - 1 do
                              begin
                                   FQDNbuf := option.DNSDomainNameList[j].getAsFrameBufferArray;

                                   for k := 0 to length (FQDNbuf) - 1 do
                                       begin
                                            pbHelp := pByte (prot_data + Offset + 8 + count);
                                            pbHelp^ := FQDNbuf[k];

                                            inc (count);
                                       end;

                                   // Write terminating 0 byte

                                   pbHelp := pByte (prot_data + Offset + 8 + count);
                                   pbHelp^ := 0;

                                   inc (count);
                              end;

                          // Pad if not multiple of 8

                          for j := 1 to (count mod 8) do
                              begin
                                   pbHelp := pByte (prot_data + Offset + 8 +count);
                                   pbHelp^ := 0;

                                   inc (count);
                              end;
                     end;
              end;

              // for redirects, the option length includes the quote following the header
              // which we can't add because we add the payload from assemblers, not as blob

              if option.option_type = 4 then
                 inc (Offset, 8)
              else
                  inc (Offset, option.option_len);
         end;
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.Assemble;

var pData      : pICMPv6Struct;
    pbHelp     : pByte;
    bHelp      : Byte;
    pwHelp     : pWord;
    wHelp      : word;
    plHelp     : pLongword;
    lHelp      : longword;
    help2      : byte;
    pAddr      : pIPv6AddressByteStruct;
    hdrBlock   : tFrameChunk;
    crcbuf     : tBufferWords;
    payloadbuf : tBufferWords;
    offset     : integer;
    i          : integer;

begin
     pData := pICMPv6Struct (prot_data);

     pData^.icmp_type := priv_type;
     pData^.icmp_code := priv_code;
     pData^.checksum := swap16 (priv_checksum);

     offset := sizeOf (tICMPv6Struct);

     case priv_type of
      1   : begin
                 // Destination Unreachable

                 plHelp := pLongword (prot_data + offset);
                 plHelp^ := 0;

                 inc (offset, 4);
            end;
      2   : begin
                 // Packet too big

                 plHelp := pLongword (prot_data + offset);
                 plHelp^ := Swap32 (priv_next_hop_mtu);

                 inc (offset, 4);
            end;
      3   : begin
                 // Time exceeded

                 plHelp := pLongword (prot_data + offset);
                 plHelp^ := Swap32 (priv_next_hop_mtu);

                 inc (offset, 4);
            end;
      4   : begin
                 // Parameter Problem

                 plHelp := pLongword (prot_data + offset);
                 plHelp^ := Swap32 (priv_problem_offset);
            end;
      128,
      129 : begin
                 // Echo Request/Reply

                 pwHelp := pWord (prot_data + offset);
                 pwHelp^ := Swap16 (priv_echo_identifier);

                 pwHelp := pWord (prot_data + offset + 2);
                 pwHelp^ := Swap16 (priv_echo_sequence);

                 inc (offset, 4);

                 // Copy Echo Payload

                 pbHelp := pByte (prot_data + offset);

                 for i := 1 to priv_payloadlength do
                     begin
                          pbHelp^ := priv_payload[i];

                          inc (pbHelp);
                     end;

                 inc (offset, priv_payloadlength);
            end;
      133 : begin
                 // Router Solicitation

                 inc (offset, 4);
            end;
      134 : begin
                 // Router Advertisement

                 pbHelp := pByte (prot_data + offset);
                 pbHelp^ := priv_router_hoplimit;

                 // Flags

                 if rfManaged in priv_routeradv_flags then
                    bHelp := $80
                 else
                     bHelp := 0;

                 if rfOther in priv_routeradv_flags then
                    bHelp := bHelp or $40;

                 if rfHomeAgent in priv_routeradv_flags then
                    bHelp := bHelp or $20;

                 if rfProxy in priv_routeradv_flags then
                    bHelp := bHelp or $4;

                 if rfReserved in priv_routeradv_flags then
                    bHelp := bHelp or $2;

                 // Router Preference

                 case priv_router_preference of
                  rpLow      : help2 := 3;
                  rpMedium   : help2 := 0;
                  rpHigh     : help2 := 1;
                  rpReserved : help2 := 2;
                 end;

                 bHelp := bHelp or (help2 shl 3);

                 pbHelp := pByte (prot_data + offset + 1);
                 pbHelp^ := bHelp;

                 inc (offset, 2);

                 // Timers

                 pwHelp := pWord (prot_data + offset);
                 pwHelp^ := Swap16 (priv_router_lifetime);

                 plHelp := pLongword (prot_data + offset + 2);
                 plHelp^ := Swap32 (priv_router_reachtime);

                 plHelp := pLongword (prot_data + offset + 6);
                 plHelp^ := Swap32 (priv_router_retranstime);

                 inc (offset, 10);
            end;
      135 : begin
                 // Neighbor Solicitation

                 pAddr := pIPv6AddressByteStruct (prot_data + offset + 4);
                 pAddr^ := priv_nd_target_address.ValueAsByteStruct;

                 inc (offset, 20);
            end;
      136 : begin
                 // Neighbor Advertisement

                 if afRouter in priv_neighboradv_flags then
                    lHelp := $4
                 else
                     lHelp := 0;

                 if afSolicited in priv_neighboradv_flags then
                    lHelp := lHelp or $2;

                 if afOverride in priv_neighboradv_flags then
                    lHelp := lHelp or $1;

                 lHelp := lHelp shl 29;

                 plHelp := pLongword (prot_data + offset);
                 plHelp^ := swap32 (lHelp);

                 pAddr := pIPv6AddressByteStruct (prot_data + offset + 4);
                 paddr^ := priv_nd_target_address.ValueAsByteStruct;

                 inc (offset, 20);
            end;
      137 : begin
                 // Redirect

                 pAddr := pIPv6AddressByteStruct (prot_data + offset + 4);
                 pAddr^ := priv_redirect_target_address.ValueAsByteStruct;

                 pAddr := pIPv6AddressByteStruct (prot_data + offset + 20);
                 pAddr^ := priv_redirect_dest_address.ValueAsByteStruct;

                 // Offset increment by 2*16 + 4 bytes reserved

                 inc (offset, 36);
            end;
     end;

     if length (priv_optionlist) > 0 then
        WriteOptionList (offset);

     assert (offset = prot_length, 'Memory offset check mismatch (tICMPv6Assembler.Assemble)');

     // Calculate the checksum, which for IPv6 needs to include the PseudoHeader

     if not priv_keepcrc then
        begin
             hdrBlock.Data := prot_data;
             hdrBlock.Len := prot_length;

             crcbuf := hdrBlock.getAsBufferWords;

             // For Echo Request/Reply, the payload is part of ICMP and contained in hdrBlock, otherwise its a quote

             if (priv_type = 128) or (priv_type = 129) then
                setLength (payloadbuf, 0)
             else
                 payloadbuf := GetRawBufferAsWordBuffer (priv_payload, priv_payloadlength);

             priv_checksum := getICMPv6Checksum (priv_IPv6PseudoHeader, crcbuf, payloadbuf);

             if priv_force_badcrc then
                priv_checksum := getBadCRC (priv_checksum);
        end;

     pData^.Checksum := swap16 (priv_checksum);
end;

(* ------------------------------------------- *)

procedure tICMPv6Assembler.CopyFromParser (Parser: tICMPv6Parser);

var option : tICMPv6NDOption;
    i      : integer;

begin
     priv_type := Parser.ICMPType;
     priv_code := Parser.ICMPCode;
     priv_checksum := Parser.Checksum;

     case priv_type of
      128,
      129 : begin
                 if Parser.PayloadLength > 0 then
                    setPayloadFromRawbuffer (Parser.Payload, Parser.PayloadLength)
                 else
                     priv_payloadlength := 0;
            end;
     end;

     priv_echo_identifier := Parser.EchoIdentifier;
     priv_echo_sequence := Parser.EchoSequence;
     priv_next_hop_mtu := Parser.NextHopMTU;
     priv_problem_offset := Parser.ParameterProblemOffset;
     priv_source_linklayer_addr := Parser.SourceLinkLayerAddress;
     priv_target_linklayer_addr := Parser.TargetLinkLayerAddress;
     priv_nd_target_address := Parser.NDTargetAddress;
     priv_redirect_target_address := Parser.RedirectTargetAddress;
     priv_redirect_dest_address := Parser.RedirectDestinationAddress;
     priv_router_hoplimit := Parser.RAHopLimit;
     priv_routeradv_flags := Parser.RAFlags;
     priv_router_preference := Parser.RAPreference;
     priv_router_lifetime := Parser.RALifeTime;
     priv_router_reachtime := Parser.RAReachableTime;
     priv_router_retranstime := Parser.RARetransTimer;
     priv_router_mtu := Parser.RAMTU;
     priv_neighboradv_flags := Parser.NeighborAdvertisementFlags;

     for i := 1 to length (Parser.OptionList) do
         begin
              option := Parser.OptionList[i-1];

              AddOption (option);
         end;
end;

(* ------------------------------------------- *)

end.
