{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_ipv6;

interface

uses base_structures,
     definition_ip_addresses,
     class_parser_ipv6,
     class_assembly_baseclass;

type tIPv6Assembler  = class (tFrameAssembler)

                             {$REGION 'Documentation'}
                             ///	<param name="ForcePayloadLength">
                             ///	  In case of a truncated packet this optional
                             ///	  parameter holds the length that will be
                             ///	  written as IPpayload length to the IP header
                             ///	</param>
                             {$ENDREGION}
                             constructor create (const ForcePayloadLength : word = 0);

                             private
                                    priv_version       : byte;
                                    priv_trafficclass  : byte;
                                    priv_flowlabel     : longword;
                                    priv_payloadlength : word;
                                    priv_nextheader    : byte;
                                    priv_hoplimit      : byte;
                                    priv_source        : tIPv6Address;
                                    priv_destination   : tIPv6Address;

                                    priv_forcedpayloadlen : word;

                             protected
                                      // called from base class from GetAssembly

                                      procedure DetermineAssemblySize; override;
                                      procedure Assemble; override;

                             public
                                   property Version : byte read priv_version write priv_version;
                                   property TrafficClass : byte read priv_trafficclass write priv_trafficclass;
                                   property FlowLabel : longword read priv_flowlabel write priv_flowlabel;
                                   property PayloadLength : word read priv_payloadlength write priv_payloadlength;
                                   property NextHeader : byte read priv_nextheader write priv_nextheader;
                                   property HopLimit : byte read priv_hoplimit write priv_hoplimit;
                                   property Source : tIPv6Address read priv_source write priv_source;
                                   property Destination : tIPv6Address read priv_destination write priv_destination;

                                   {$REGION 'Documentation'}
                                   ///	<summary>
                                   ///	  copies all values from a parser object
                                   ///	  to the assembler
                                   ///	</summary>
                                   {$ENDREGION}
                                   procedure CopyFromParser (Parser : tIPv6Parser);
                       end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ tIPv6Assembler }
(* ------------------------------------------- *)

constructor tIPv6Assembler.create (const ForcePayloadLength : word = 0);

begin
     inherited create;

     priv_version := 0;
     priv_trafficclass := 0;
     priv_flowlabel := 0;
     priv_payloadlength := 0;
     priv_nextheader := 0;
     priv_hoplimit := 0;
     priv_source.initialize;
     priv_destination.initialize;

     priv_forcedpayloadlen := ForcePayloadLength;
end;

(* ------------------------------------------- *)

procedure tIPv6Assembler.DetermineAssemblySize;

begin
     prot_length := 40;
end;

(* ------------------------------------------- *)

procedure tIPv6Assembler.Assemble;

var pData  : pIPv6Struct;
    lHelp  : longword;

begin
     pData := pIPv6Struct (prot_data);

     // Combine Version, Class and Flow Label into a longword

     lHelp := 0;
     lHelp := (lHelp or (priv_version shl 28));
     lHelp := (lHelp or (priv_trafficclass shl 20));
     lHelp := (lHelp or priv_flowlabel);

     pData^.VersionClassLabel := swap32 (lHelp);

     if priv_forcedpayloadlen = 0 then
        pData^.PayloadLength := swap16 (priv_payloadlength)
     else
         pData^.PayloadLength := swap16 (priv_forcedpayloadlen);

     pData^.NextHeader := priv_nextheader;
     pData^.HopLimit := priv_hoplimit;
     pData^.SourceAddress := priv_source.ValueAsByteStruct;
     pData^.DestinationAddress := priv_destination.ValueAsByteStruct;
end;

(* ------------------------------------------- *)

procedure tIPv6Assembler.CopyFromParser (Parser: tIPv6Parser);

begin
     priv_version := Parser.Version;
     priv_trafficclass := Parser.TrafficClass;
     priv_flowlabel := Parser.FlowLabel;
     priv_forcedpayloadlen := Parser.PayloadLength;
     priv_nextheader := Parser.NextHeader;
     priv_hoplimit := Parser.HopLimit;
     priv_source := Parser.Source;
     priv_destination := Parser.Destination;
end;

(* ------------------------------------------- *)

end.


