{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_linux_sll;

interface

uses base_structures,
     definition_ip_addresses,
     class_parser_linux_sll,
     class_assembly_baseclass;


type tLinuxSLLAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_packettype         : tSLLPacketType;
                                       priv_ARPHdrType         : word;
                                       priv_LinkLayerAddrLen   : word;
                                       priv_MACAddress         : tMACAddress;
                                       priv_OtherLinkLayerAddr : uint64;
                                       priv_ProtocolType       : word;

                                       procedure setMACAddress (Value : tMACAddress);

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property PacketType : tSLLPacketType read priv_packettype write priv_packettype;
                                      property ARPHeaderType : word read priv_ARPHdrType write priv_ARPHdrType;
                                      property LinkLayerAddressLength : word read priv_LinkLayerAddrLen write priv_LinkLayerAddrLen;
                                      property MACAddress : tMACAddress read priv_MACAddress write setMACAddress;
                                      property LinkLayerAddress : uint64 read priv_OtherLinkLayerAddr write priv_OtherLinkLayerAddr;
                                      property ProtocolType : word read priv_ProtocolType write priv_ProtocolType;

                                      procedure CopyFromParser (Parser : tLinuxSLLParser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tLinuxSLLAssembler }
(* ------------------------------------------- *)

constructor tLinuxSLLAssembler.create;

begin
     inherited;

     priv_packettype := sptUnicastToUs;
     priv_ARPHdrType := 0;
     priv_LinkLayerAddrLen := 0;
     priv_MACAddress.initialize;
     priv_OtherLinkLayerAddr := 0;
     priv_ProtocolType := 0;
end;

(* ------------------------------------------- *)

procedure tLinuxSLLAssembler.DetermineAssemblySize;

begin
     prot_length := sizeOf (tSLLStruct);
end;

(* ------------------------------------------- *)

procedure tLinuxSLLAssembler.setMACAddress (Value: tMACAddress);

var i    : integer;
    help : uint64;

begin
     // We need to the the address and fix the length/general address block

     priv_MACAddress := Value;
     priv_LinkLayerAddrLen := 6;

     help := Value.RangeBytes[1];

     for i := 2 to 6 do
         begin
              help := help shl 8;
              help := help + Value.RangeBytes[i];
         end;

     // set the last two bytes to zero

     help := help shl 16;

     priv_OtherLinkLayerAddr := help;
end;

(* ------------------------------------------- *)

procedure tLinuxSLLAssembler.Assemble;

var pData : pSLLStruct;
    help  : word;

begin
     pData := pSLLStruct (prot_data);

     case priv_packettype of
      sptUnicastToUs        : help := 0;
      sptBroadcast          : help := 1;
      sptMulticast          : help := 2;
      sptUnicastToOtherHost : help := 3;
      sptSentByUs           : help := 4;
     end;

     pData^.PacketType := Swap16 (help);
     pData^.ARPHdrType := swap16 (priv_ARPHdrType);
     pData^.LinkLayerAddrLen := Swap16 (priv_LinkLayerAddrLen);
     pData^.LinkLayerAddr := swap64 (priv_OtherLinkLayerAddr);
     pData^.ProtocolType := Swap16 (priv_ProtocolType);
end;

(* ------------------------------------------- *)

procedure tLinuxSLLAssembler.CopyFromParser(Parser: tLinuxSLLParser);

begin
     priv_packettype := Parser.PacketType;
     priv_ARPHdrType := Parser.ARPHeaderType;
     priv_LinkLayerAddrLen := Parser.LinkLayerAddressLength;
     priv_MACAddress := Parser.MACAddress;                           // necessary, because we might check against it
     priv_OtherLinkLayerAddr := Parser.LinkLayerAddress;
     priv_ProtocolType := Parser.ProtocolType;
end;

(* ------------------------------------------- *)

end.
