{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_llc;

interface

uses base_structures,
     class_parser_llc,
     class_assembly_baseclass;


type tLLCAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       // 802.2 (not extracting IG Bit and CR Bit and other detail stuff here, maybe later)

                                       priv_dsap            : byte;
                                       priv_ssap            : byte;
                                       priv_control         : byte;

                                       // SNAP

                                       priv_vendorcode      : tSNAPVendorCode;
                                       priv_protocoltype    : word;

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property DSAP : byte read priv_dsap write priv_dsap;
                                      property SSAP : byte read priv_ssap write priv_ssap;
                                      property Control : byte read priv_control write priv_control;
                                      property VendorCode : tSNAPVendorCode read priv_vendorcode write priv_vendorcode;
                                      property ProtocolType : word read priv_protocoltype write priv_protocoltype;

                                public
                                      procedure CopyFromParser (Parser : tLLCParser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tLLCAssembler }
(* ------------------------------------------- *)

constructor tLLCAssembler.create;

begin
     inherited;

     priv_dsap := 0;
     priv_ssap := 0;
     priv_control := 0;

     priv_vendorcode[0] := 0;
     priv_vendorcode[1] := 0;
     priv_vendorcode[2] := 0;

     priv_protocoltype := 0;
end;

(* ------------------------------------------- *)

procedure tLLCAssembler.DetermineAssemblySize;

begin
     if priv_dsap = $AA then
        prot_length := sizeOf (tStruct_802dot3) + sizeOf (tStruct_SNAP)
     else
         prot_length := sizeOf (tStruct_802dot3);
end;

(* ------------------------------------------- *)

procedure tLLCAssembler.Assemble;

var pData     : pStruct_802dot3;
    pSNAPData : pStruct_SNAP;

begin
     pData := pStruct_802dot3 (prot_data);

     pData^.DSAP := priv_dsap;
     pData^.SSAP := priv_ssap;
     pData^.Control := priv_control;

     if priv_dsap = $AA then
        begin
             pSNAPData := pStruct_SNAP (prot_data + sizeOf (tStruct_802dot3));

             pSNAPData^.VendorCode := priv_Vendorcode;
             pSNAPData^.ProtocolType := Swap16 (priv_ProtocolType);
        end;
end;

(* ------------------------------------------- *)

procedure tLLCAssembler.CopyFromParser (Parser: tLLCParser);

begin
     priv_dsap := Parser.DSAP;
     priv_ssap := Parser.SSAP;
     priv_control := Parser.Control;

     // SNAP

     priv_vendorcode := Parser.VendorCode;
     priv_protocoltype := Parser.ProtocolType;
end;

(* ------------------------------------------- *)

end.
