{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_mpls;

interface

uses base_structures,
     class_parser_mpls,
     class_assembly_baseclass;


type tMPLSAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_label       : longword;
                                       priv_tos         : byte;
                                       priv_bottom      : boolean;
                                       priv_ttl         : byte;

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property MPLSLabel : longword read priv_label write priv_label;
                                      property ToSExperimental : byte read priv_tos write priv_tos;
                                      property BottomOfStack : boolean read priv_bottom write priv_bottom;
                                      property TTL : byte read priv_ttl write priv_ttl;

                                      procedure CopyFromParser (Parser : tMPLSParser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tMPLSAssembler }
(* ------------------------------------------- *)

constructor tMPLSAssembler.create;

begin
     inherited;

     priv_label := 0;
     priv_tos := 0;
     priv_bottom := false;
     priv_ttl := 0;
end;

(* ------------------------------------------- *)

procedure tMPLSAssembler.DetermineAssemblySize;

begin
     prot_length := sizeOf (longword);
end;

(* ------------------------------------------- *)

procedure tMPLSAssembler.Assemble;

var pData : pLongword;
    help  : longword;
    help2 : longword;

begin
     pData := pLongword (prot_data);

     help := priv_label shl 12;

     help2 := priv_tos shl 9;
     help := help or help2;

     if priv_bottom then
        help := help or $100;

     help := help or priv_ttl;

     pData^:= Swap32 (help);
end;

(* ------------------------------------------- *)

procedure tMPLSAssembler.CopyFromParser (Parser: tMPLSParser);

begin
     priv_label := Parser.MPLSLabel;
     priv_tos := Parser.ToSExperimental;
     priv_bottom := Parser.BottomOfStack;
     priv_ttl := Parser.TTL;
end;

(* ------------------------------------------- *)

end.
