{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parse_snortalertlog;

interface

{$i 'jedi.inc')}

uses DBXJSON,
{$ifdef DELPHI20_UP}
     System.JSON,
{$endif}
     System.Generics.Collections,
     base_structures,
     definition_ip_addresses,
     definition_filter_structures,
     definition_tcp_structures;

type tSnortAlertFileFormat = (sfUnknownFormat, sfTextAlerts, sfUnified2);
     tSnortExtraDataType = (edtOriginalClientIPv4, edtOriginalClientIPv6, edtUnused, edtGZIPCompressedData, edtSMTPFilename, edtSMTPMailFrom, edtSMTPRCPTTo, edtSMTPEmailHeaders, edtHTTPURI, edtHTTPHostname, edtIPv6SourceAddress, edtIPv6DestinationAddress, edtNormalizedJavascriptData);

     tOnAdvanceAlertPercentProcessed = procedure (Sender : tObject; const FileSize : uint64; const AlertsAdded : integer; const PercentageParsed : integer) of object;

     tSnortRule = record
                        SignatureID            : longword;
                        RuleName               : string;
                        Definition             : string;

                        function ParseFromLine (Line : string) : boolean;
                  end;

     tSnortAlert = record
                         gid            : integer;
                         sid            : integer;
                         revision       : integer;
                         rulename       : string;
                         classification : string;
                         priority       : integer;
                         timestamp      : tFrameTimestamp;
                         IPv4SocketPair : tIPv4SocketPair;
                         IPv6SocketPair : tIPv6SocketPair;
                         Layer4Protocol : tLayer4Protocol;
                         TTL            : byte;
                         IPVersion      : byte;
                         IPID           : word;
                         IPHdrLen       : integer;
                         IPTotalLen     : integer;
                         TCPSequenceNo  : longword;
                         TCPAcknowledge : longword;
                         TCPWindow      : longword;
                         TCPFlags       : tTCPFlagSet;
                         ICMPType       : byte;
                         ICMPCode       : byte;
                         ICMPSequence   : word;
                         Reference      : string;

                         procedure initialize;
                         function DictionaryKey : string;
                         function getFilterLabel (const Template : string) : string;
                   end;

     tSnortAlertList = array of tSnortAlert;

     tMetaRuleDetailsEntry = record
                                   Classification           : string;
                                   Priority                 : integer;
                                   sid                      : integer;
                                   Comment                  : string;

                                   procedure initialize;
                             end;

     tMetaDataEntry = record
                            IPVersion                : byte;
                            IPv4SocketPair           : tIPv4SocketPair;
                            IPv6SocketPair           : tIPv6SocketPair;
                            IPID                     : word;
                            Layer4Protocol           : tLayer4Protocol;
                            TCPSequenceNumber        : longword;
                            TCPFlags                 : tTCPFlagSet;
                            Timestamp                : tFrameTimestamp;
                            RuleDetails              : tMetaRuleDetailsEntry;

                            class operator equal (a, b : tMetaDataEntry) : boolean;
                            class operator notequal (a, b : tMetaDataEntry) : boolean;
                            class operator greaterthan (a, b : tMetaDataEntry) : boolean;
                            class operator lessthan (a, b : tMetaDataEntry) : boolean;
                            class operator greaterthanorequal (a, b : tMetaDataEntry) : boolean;
                            class operator lessthanorequal (a, b : tMetaDataEntry) : boolean;

                            procedure initialize; overload;
                            procedure initialize (const SnortAlert : tSnortAlert); overload;
                            procedure initialize (Settings : tJSONObject); overload;
                            procedure initialize (Settings : string); overload;
                            {$REGION 'Documentation'}
                            /// <summary>
                            ///   Creates a DictionaryKey for the current entry.
                            ///    OverrideSID can be used to set a specific
                            ///   SID, which is required if enhancing frames,
                            ///   because then we don't know the SID of course
                            /// </summary>
                            /// <param name="OverrideSID">
                            ///   0 if the SID from the entry should used
                            /// </param>
                            {$ENDREGION}
                            function DictionaryKey (const OverrideSID : integer = 0) : string;
                            /// <summary>
                            ///   Returns a JSON object containing the
                            ///   Metadata. The calling routine needs to handle
                            ///   the freeing of the returned object.
                            /// </summary>
                            function toJSON : tJSONObject;
                      end;

     tMetaDataSet = array of tMetaDataEntry;

     tUnifiedRecordType = (urtUnknown, urtPacket, urtIDSEvent, urtIDSEventIPv6, urtIDSEvent2, urtIDSEventIPv62, urtExtraData);

     tUnifiedPacketInfo = packed record
                                       SensorID             : longword;
                                       EventID              : longword;
                                       EventSeconds         : longword;
                                       PacketSeconds        : longword;
                                       PacketMicroseconds   : longword;
                                       LinkType             : longword;
                                       PacketLength         : longword;

                                       procedure SwapValues;
                          end;

     tUnifiedIDSIPv4EventIfnfo = packed record
                                              SensorID             : longword;
                                              EventID              : longword;
                                              EventSeconds         : longword;
                                              EvenMicroseconds     : longword;
                                              SignatureID          : longword;
                                              GeneratorID          : longword;
                                              SignatureRevision    : longword;
                                              ClassificationID     : longword;
                                              PriorityID           : longword;
                                              IPSource             : longword;
                                              IPDestination        : longword;
                                              SourcePort           : word;
                                              DestinationPort      : word;
                                              Protocol             : byte;
                                              ImpactFlag           : byte;
                                              Impact               : byte;
                                              Blocked              : byte;

                                              procedure SwapValues;
                                 end;

     tUnifiedIDSIPv6EventIfnfo = packed record
                                              SensorID             : longword;
                                              EventID              : longword;
                                              EventSeconds         : longword;
                                              EvenMicroseconds     : longword;
                                              SignatureID          : longword;
                                              GeneratorID          : longword;
                                              SignatureRevision    : longword;
                                              ClassificationID     : longword;
                                              PriorityID           : longword;
                                              IPSource             : tIPv6AddressByteStruct;
                                              IPDestination        : tIPv6AddressByteStruct;
                                              SourcePort           : word;
                                              DestinationPort      : word;
                                              Protocol             : byte;
                                              ImpactFlag           : byte;
                                              Impact               : byte;
                                              Blocked              : byte;

                                              procedure SwapValues;
                                 end;

     tUnifiedExtraData = packed record
                                      EventType            : longword;
                                      EventLength          : longword;
                                      SensorID             : longword;
                                      EventID              : longword;
                                      EventSeconds         : longword;
                                      ExtraType            : longword;
                                      DataType             : longword;
                                      DataLength           : longword;

                                      procedure SwapValues;
                         end;

     tUnifiedRecord  = record
                             RecordType           : tUnifiedRecordType;
                             RecordLength         : longword;
                             SensorID             : longword;
                             EventID              : longword;
                             EventSeconds         : longword;
                             EvenMicroseconds     : longword;

                             LinkType             : longword;
                             PacketLength         : longword;
                             PacketData           : tRawBuffer;

                             SignatureID          : longword;
                             GeneratorID          : longword;
                             SignatureRevision    : longword;
                             ClassificationID     : longword;
                             PriorityID           : longword;
                             IPSource             : tIPAddress;
                             IPDestination        : tIPAddress;
                             SourcePort           : word;
                             DestinationPort      : word;
                             Protocol             : tLayer4Protocol;
                             TTL                  : byte;
                             IPVersion            : byte;
                             IPID                 : word;
                             IPHdrLen             : integer;
                             IPTotalLen           : integer;
                             TCPSequenceNo        : longword;
                             TCPAcknowledge       : longword;
                             TCPWindow            : longword;
                             TCPFlags             : tTCPFlagSet;
                             ICMPType             : byte;
                             ICMPCode             : byte;
                             ICMPSequence         : word;
                             ImpactFlag           : byte;
                             Impact               : byte;
                             Blocked              : boolean;

                             procedure initialize;
                             procedure assign (PacketInfo : tUnifiedPacketInfo; PacketData : tRawBuffer); overload;
                             procedure assign (IPv4Event : tUnifiedIDSIPv4EventIfnfo); overload;
                             procedure assign (IPv6Event : tUnifiedIDSIPv6EventIfnfo); overload;
                       end;


     tSnortAlertParser = class (tObject)

                               constructor create (const Filename : string);

                               private
                                      EventOnAlertPercentProcessed : tOnAdvanceAlertPercentProcessed;

                                      priv_filename                : string;
                                      priv_fileformat              : tSnortAlertFileFormat;
                                      priv_labeltemplate           : string;
                                      priv_alerts                  : tSnortAlertList;
                                      priv_rules                   : tDictionary<integer, tSnortRule>;

                                      priv_metadatalist            : tMetaDataSet;

                                      priv_ignoreSYN               : boolean;
                                      priv_ignoreICMP              : boolean;

                                      priv_UnifiedReadFail         : boolean;
                                      priv_percentageprocessed     : integer;

                                      procedure ReportProgress (const FileSize, Offset : uint64);

                                      function getFilterForAlert (const Alert : tSnortAlert) : tFrameFilter;
                                      function SnortDateTimeToFrameTimeStamp (value : string) : tFrameTimestamp;

                                      procedure ParseFirstLine (sLine : string; var alert : tSnortAlert);
                                      procedure ParseSecondLine (sLine : string; var alert : tSnortAlert);
                                      procedure ParseThirdLine (sLine : string; var alert : tSnortAlert);
                                      procedure ParseFourthLine (sLine : string; var alert : tSnortAlert);
                                      procedure ParseFifthLine (sLine : string; var alert : tSnortAlert);
                                      procedure ParseSixtLine (sLine : string; var alert : tSnortAlert);

                                      /// <summary>
                                      ///   Adds a snort alert based on a
                                      ///   unified2 record
                                      /// </summary>
                                      procedure AddSnortAlert (UnifiedRec : tUnifiedRecord);

                                      procedure ReadTextAlerts;
                                      procedure ReadUnified2;

                               public
                                     property FileFormat : tSnortAlertFileFormat read priv_fileformat;
                                     property AlertList : tSnortAlertList read priv_alerts;
                                     property MetaDataList : tMetaDataSet read priv_metadatalist;
                                     property IgnoreSYN : boolean read priv_ignoreSYN write priv_ignoreSYN;
                                     property IgnoreICMP : boolean read priv_ignoreICMP write priv_ignoreICMP;
                                     property LabelTemplate : string read priv_labeltemplate write priv_labeltemplate;
                                     property UnifiedFileReadFailure : boolean read priv_UnifiedReadFail;

                                     property OnAdvanceAlertPercentProcessed : tOnAdvanceAlertPercentProcessed read EventOnAlertPercentProcessed write EventOnAlertPercentProcessed;

                                     destructor Destroy; override;

                                     function isValidFile : boolean;
                                     /// <summary>
                                     ///   Creates filters from the snort data
                                     ///   already parsed and returns a new
                                     ///   filterset object. Destruction of
                                     ///   that object must be handled by the
                                     ///   calling routine.
                                     /// </summary>
                                     /// <returns>
                                     ///   New tFilterSet Object
                                     /// </returns>
                                     function GetFilterList : tFilterSet;
                                     procedure AddAlertToList (const alert : tSnortAlert);
                                     procedure LoadRuleDefinitions (const Filename : string; const IncludeDisabledRules : boolean = false);
                                     procedure clear;
                                     procedure execute;
                         end;

const DefaultSnortLabelTemplate = '<rulename>';
      UNKNOWN_SID = 1;

     function getDictionaryKeyForMetaDataEntry (const SocketPair : tIPv4SocketPair; const IPID : integer; const L4Protocol : tLayer4Protocol; const TCPSequenceNo : longword; const sid : integer) : string; overload;
     function getDictionaryKeyForMetaDataEntry (const SocketPair : tIPv6SocketPair; const IPID : integer; const L4Protocol : tLayer4Protocol; const TCPSequenceNo : longword; const sid : integer) : string; overload;

     procedure mergeMetaDataToSet (const entry : tMetaDataEntry; var MetaDataSet : tMetaDataSet);
     procedure mergeMetaSetDataToSet (const NewEntries : tMetaDataSet; var MetaDataSet : tMetaDataSet);
     procedure deleteMetaDataFromSet (const Index : integer; var DataSet : tMetaDataSet);

implementation

uses SiAuto,
     SysUtils,
     DateUtils,
     Math,
     Classes,
     Toolkit,
     class_rawframe,
     class_parsemaster,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_tcp,
     class_parser_udp,
     class_parser_ICMPv4;


(* ----------------------------------- *)
{ Helper Functions}
(* ----------------------------------- *)

function getDictionaryKeyForMetaDataEntry (const SocketPair : tIPv4SocketPair; const IPID : integer; const L4Protocol : tLayer4Protocol; const TCPSequenceNo : longword; const sid : integer) : string;

var value : string;

begin
     // For TCP we ignore IP ID, because the sequence number is much better. Plus in Snort alert logs,
     // the IP ID is often from a different packet when reassembly occurred, so we'd never hit the entry

     case L4Protocol of
      l4Any,
      l4UDP,
      l4ICMP : value := SocketPair.DictionaryKey + '-' + IntToStr (IPID) + '-' + Layer4ProtocolToStr (L4Protocol) + '-0' + IntToStr (sid);
      l4TCP  : value := SocketPair.DictionaryKey + '-0-' + Layer4ProtocolToStr (L4Protocol) + '-' + IntToStr (TCPSequenceNo) + IntToStr (sid);
     end;

     // Remove the separators again (doesn't seem to make sense after having added them just now, but they make it easier to debug)

     value := stringReplace (value, ':', '', [rfReplaceAll]);
     value := stringReplace (value, '-', '', [rfReplaceAll]);
     value := stringReplace (value, '.', '', [rfReplaceAll]);
     value := stringReplace (value, '/', '', [rfReplaceAll]);

     result := value;
end;

(* ----------------------------------- *)

function getDictionaryKeyForMetaDataEntry (const SocketPair : tIPv6SocketPair; const IPID : integer; const L4Protocol : tLayer4Protocol; const TCPSequenceNo : longword; const sid : integer) : string;

var value : string;

begin
     // For TCP we ignore IP ID, because the sequence number is much better. Plus in Snort alert logs,
     // the IP ID is often from a different packet when reassembly occurred, so we'd never hit the entry

     case L4Protocol of
      l4Any,
      l4UDP,
      l4ICMP : value := SocketPair.DictionaryKey + '-' + IntToStr (IPID) + '-' + Layer4ProtocolToStr (L4Protocol) + '-0' + '-' + IntToStr (sid);
      l4TCP  : value := SocketPair.DictionaryKey + '-0-' + Layer4ProtocolToStr (L4Protocol) + '-' + IntToStr (TCPSequenceNo) + '-' + IntToStr (sid);
     end;

     // Remove the separators again (doesn't seem to make sense after having added them just now, but they make it easier to debug)

     value := stringReplace (value, ':', '', [rfReplaceAll]);
     value := stringReplace (value, '-', '', [rfReplaceAll]);
     value := stringReplace (value, '.', '', [rfReplaceAll]);
     value := stringReplace (value, '/', '', [rfReplaceAll]);

     result := value;
end;

(* ----------------------------------- *)

procedure mergeMetaDataToSet (const entry : tMetaDataEntry; var MetaDataSet : tMetaDataSet);

var found : boolean;
    len   : integer;
    i     : integer;

begin
     found := false;
     len := length (MetaDataSet);

     if len > 0 then
        begin
             i := 0;

             repeat
                   if entry = MetaDataSet[i] then
                      begin
                           found := true;
                      end;

                   inc (i);

             until found or (i >= len);
        end;

     if not found then
        begin
             setLength (MetaDataSet, len + 1);
             MetaDataSet[len] := entry;
        end;
end;

(* ----------------------------------- *)

procedure mergeMetaSetDataToSet (const NewEntries : tMetaDataSet; var MetaDataSet : tMetaDataSet);

var entry  : tMetaDataEntry;
    len    : integer;
    i      : integer;

begin
     len := length (NewEntries);

     for i := 1 to len do
         begin
              entry := NewEntries[i-1];

              mergeMetaDataToSet (entry, MetadataSet);
         end;
end;

(* ----------------------------------- *)

procedure deleteMetaDataFromSet (const Index : integer; var DataSet : tMetaDataSet);

var len : integer;
    i   : integer;

begin
     len := length (DataSet);

     Assert (len > 0);
     Assert (Index < len);

     for i := Index + 1 to len - 1 do
         DataSet[i-1] := DataSet[i];

     SetLength (DataSet, len - 1);
end;

(* ----------------------------------- *)
{ tSnortRule }
(* ----------------------------------- *)

function tSnortRule.ParseFromLine (Line: string) : boolean;

var help  : string;
    dummy : integer;

begin
     Definition := Line;

     help := rightstring (Line, 'sid:');
     help := leftstring (help, ';');
     val (help, SignatureID, dummy);

     help := rightstring (Line, 'msg:"');
     RuleName := leftstring (help, '"');

     result := (SignatureID <> 0);
end;

(* ----------------------------------- *)
{ tMetaRuleDetailsEntry }
(* ----------------------------------- *)

procedure tMetaRuleDetailsEntry.initialize;

begin
     Classification := '';
     Priority := 0;
     sid := 0;
     Comment := '';
end;

(* ----------------------------------- *)
{ tSnortAlert }
(* ----------------------------------- *)

procedure tSnortAlert.initialize;

begin
     gid := 0;
     sid := 0;
     revision := 0;
     rulename := '';
     classification := '';
     priority := 0;
     timestamp.initialize;
     IPv4SocketPair.initialize;
     IPv6SocketPair.initialize;
     Layer4Protocol := l4Any;
     TTL := 0;
     IPVersion := 4;
     IPID := 0;
     IPHdrLen := 0;
     IPTotalLen := 0;
     TCPSequenceNo := 0;
     TCPFlags := [];
     ICMPType := 0;
     ICMPCode := 0;
     ICMPSequence := 0;
     Reference := '';
end;

(* ----------------------------------- *)

function tSnortAlert.DictionaryKey: string;

begin
     case IPVersion of
      4 : result := getDictionaryKeyForMetaDataEntry (IPv4SocketPair, IPID, Layer4Protocol, TCPSequenceNo, sid);
      6 : result := getDictionaryKeyForMetaDataEntry (IPv6SocketPair, IPID, Layer4Protocol, TCPSequenceNo, sid);
     end;
end;

(* ----------------------------------- *)

function tSnortAlert.getFilterLabel (const Template: string): string;

var value : string;

begin
     value := trim (Template);

     value := stringReplace (value, '<sid>', IntToStr (sid), [rfReplaceAll, rfIgnoreCase]);
     value := stringReplace (value, '<rulename>', rulename, [rfReplaceAll, rfIgnoreCase]);
     value := stringReplace (value, '<classification>', classification, [rfReplaceAll, rfIgnoreCase]);

     result := value;
end;

(* ----------------------------------- *)
{ tMetaDataEntry }
(* ----------------------------------- *)

procedure tMetaDataEntry.initialize;

begin
     IPVersion := 4;
     IPv4SocketPair.initialize;
     IPv6SocketPair.initialize;
     IPID := 0;
     Layer4Protocol := l4Any;
     TCPSequenceNumber := 0;
     RuleDetails.initialize;
end;

(* ----------------------------------- *)

procedure tMetaDataEntry.initialize (const SnortAlert: tSnortAlert);

begin
     IPVersion := SnortAlert.IPVersion;
     IPv4SocketPair := SnortAlert.IPv4SocketPair;
     IPv6SocketPair := SnortAlert.IPv6SocketPair;
     IPID := SnortAlert.IPID;
     Layer4Protocol := SnortAlert.Layer4Protocol;
     TCPSequenceNumber := SnortAlert.TCPSequenceNo;
     TCPFlags := SnortAlert.TCPFlags;
     Timestamp := SnortAlert.timestamp;

     RuleDetails.sid := SnortAlert.sid;
     RuleDetails.Comment := IntToStr (SnortAlert.gid) + ':' + IntToStr (SnortAlert.sid) + ':' + IntToStr (SnortAlert.revision) + ' - ' + SnortAlert.rulename;
     RuleDetails.Priority := SnortAlert.priority;
     RuleDetails.Classification := SnortAlert.classification;
end;

(* ----------------------------------- *)

procedure tMetaDataEntry.initialize (Settings : tJSONObject);

var dummy : integer;
    help  : string;
    JVal  : tJSONValue;

begin
     initialize;

     JVal := Settings.GetValue ('IPVersion');
     IPVersion := StrToInt (StripQuotes (JVal.Value));


     case IPVersion of
      4 : begin
               JVal := Settings.GetValue ('SourceAddress');
               help := stripQuotes (JVal.ToString);
               IPv4SocketPair.Source.IPAddress.setValueFromString (help);

               JVal := Settings.GetValue ('SourcePort');
               help := stripQuotes (JVal.ToString);
               IPv4SocketPair.Source.Port := StrToInt (help);

               JVal := Settings.GetValue ('DestinationAddress');
               help := stripQuotes (JVal.ToString);
               IPv4SocketPair.Destination.IPAddress.setValueFromString (help);

               JVal := Settings.GetValue ('DestinationPort');
               help := stripQuotes (JVal.ToString);
               IPv4SocketPair.Destination.Port := StrToInt (help);
          end;
      6 : begin
               JVal := Settings.GetValue ('SourceAddress');
               help := stripQuotes (JVal.ToString);
               IPv6SocketPair.Source.IPAddress.setValueFromString (help);

               JVal := Settings.GetValue ('SourcePort');
               help := stripQuotes (JVal.ToString);
               IPv6SocketPair.Source.Port := StrToInt (help);

               JVal := Settings.GetValue ('DestinationAddress');
               help := stripQuotes (JVal.ToString);
               IPv6SocketPair.Destination.IPAddress.setValueFromString (help);

               JVal := Settings.GetValue ('DestinationPort');
               help := stripQuotes (JVal.ToString);
               IPv6SocketPair.Destination.Port := StrToInt (help);
          end;
     end;

     JVal := Settings.GetValue ('IPID');
     IPID := StrToInt (StripQuotes (JVal.Value));

     JVal := Settings.GetValue ('L4Protocol');
     Layer4Protocol := StrToLayer4Protocol (stripQuotes (JVal.Value));

     JVal := Settings.GetValue ('TCPSequenceNumber');
     val (stripQuotes (JVal.Value), TCPSequenceNumber, dummy);

     JVal := Settings.GetValue ('TCPFlags');
     help := stripQuotes (JVal.Value);
     TCPFlags := TCPDumpToTCPFlagSet (help);

     JVal := Settings.GetValue ('Timestamp');
     help := stripQuotes (JVal.Value);
     timestamp.setFromString (help);

     JVal := Settings.GetValue ('sid');
     RuleDetails.sid := StrToInt (StripQuotes (JVal.Value));

     JVal := Settings.GetValue ('Comment');
     RuleDetails.Comment := stripQuotes (JVal.Value);

     JVal := Settings.GetValue ('Priority');
     RuleDetails.Priority := StrToInt (StripQuotes (JVal.Value));

     JVal := Settings.GetValue ('Classification');
     RuleDetails.Classification := stripQuotes (JVal.Value);
end;

(* ----------------------------------- *)

procedure tMetaDataEntry.initialize (Settings : string);

var json : tJSONObject;

begin
     json := TJSONObject.ParseJSONValue (TEncoding.ASCII.GetBytes (Settings),0) as TJSONObject;

     initialize (json);

     json.Free;
end;

(* ----------------------------------- *)

function tMetaDataEntry.toJSON : tJSONObject;

var value : tJSONObject;

begin
     value := tJSONObject.Create;

     value.AddPair (TJSONPair.Create ('IPVersion', tJSONString.Create (IntToStr (IPVersion))));

     case IPVersion of
      4 : begin
               value.AddPair (TJSONPair.Create ('SourceAddress', tJSONString.Create (IPv4SocketPair.IPAddressPair.Source.ValueAsString)));
               value.AddPair (TJSONPair.Create ('SourcePort', tJSONString.Create (IntToStr (IPv4SocketPair.Source.Port))));
               value.AddPair (TJSONPair.Create ('DestinationAddress', tJSONString.Create (IPv4SocketPair.IPAddressPair.Destination.ValueAsString)));
               value.AddPair (TJSONPair.Create ('DestinationPort', tJSONString.Create (IntToStr (IPv4SocketPair.Destination.Port))));
          end;
      6 : begin
               value.AddPair (TJSONPair.Create ('SourceAddress', tJSONString.Create (IPv6SocketPair.IPAddressPair.Source.ValueAsString)));
               value.AddPair (TJSONPair.Create ('SourcePort', tJSONString.Create (IntToStr (IPv6SocketPair.Source.Port))));
               value.AddPair (TJSONPair.Create ('DestinationAddress', tJSONString.Create (IPv6SocketPair.IPAddressPair.Destination.ValueAsString)));
               value.AddPair (TJSONPair.Create ('DestinationPort', tJSONString.Create (IntToStr (IPv6SocketPair.Destination.Port))));
          end;
     end;

     value.AddPair (TJSONPair.Create ('IPID', tJSONString.Create (IntToStr (IPID))));
     value.AddPair (TJSONPair.Create ('L4Protocol', tJSONString.Create (Layer4ProtocolToStr (Layer4Protocol))));
     value.AddPair (TJSONPair.Create ('TCPSequenceNumber', tJSONString.Create (IntToStr (TCPSequenceNumber))));
     value.AddPair (TJSONPair.Create ('TCPFlags', tJSONString.Create (TCPFlagSetToTCPDump (TCPFlags))));
     value.AddPair (TJSONPair.Create ('Timestamp', tJSONString.Create (Timestamp.asFullString)));
     value.AddPair (TJSONPair.Create ('sid', tJSONString.Create (IntToStr (RuleDetails.sid))));
     value.AddPair (TJSONPair.Create ('Comment', tJSONString.Create (RuleDetails.Comment)));
     value.AddPair (TJSONPair.Create ('Priority', tJSONString.Create (IntToStr (RuleDetails.Priority))));
     value.AddPair (TJSONPair.Create ('Classification', tJSONString.Create (RuleDetails.Classification)));

     result := value;
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.equal(a, b: tMetaDataEntry): boolean;

begin
     result := (a.IPVersion = b.IPVersion) and
               (a.IPv4SocketPair = b.IPv4SocketPair) and
               (a.IPv6SocketPair = b.IPv6SocketPair) and
               (a.IPID = b.IPID) and
               (a.Layer4Protocol = b.Layer4Protocol) and
               (a.TCPSequenceNumber = b.TCPSequenceNumber) and
               (a.TCPFlags = b.TCPFlags) and
               (a.Timestamp = b.Timestamp) and
               (a.RuleDetails.sid = b.RuleDetails.sid) and
               (a.RuleDetails.Comment = b.RuleDetails.Comment);
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.notequal(a, b: tMetaDataEntry): boolean;

begin
     result := not (a=b);
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.greaterthan (a, b: tMetaDataEntry): boolean;

var help : boolean;

begin
     if (a.IPVersion > b.IPVersion) then
        result := true
     else
         if (b.IPVersion > a.IPVersion) then
            result := false
         else
             begin
                  // IP versions equal

                  help := false;

                  case a.IPVersion of
                   4 : begin
                            if a.IPv4SocketPair = b.IPv4SocketPair then
                               help := true
                            else
                                result := a.IPv4SocketPair > b.IPv4SocketPair;
                       end;
                   6 : begin
                            if a.IPv6SocketPair = b.IPv6SocketPair then
                               help := true
                            else
                                result := a.IPv6SocketPair > b.IPv6SocketPair;
                       end;
                  end;

                  if help then
                     begin
                          // Sockets were equal, too, so test the other stuff. Which looks awkward, but what can you do...

                          if a.IPID = b.IPID then
                             begin
                                  if a.Layer4Protocol = b.Layer4Protocol then
                                     begin
                                          if a.TCPSequenceNumber = b.TCPSequenceNumber then
                                             begin
                                                  if a.RuleDetails.sid = b.RuleDetails.sid then
                                                     begin
                                                          if a.Timestamp = b.Timestamp then
                                                             result := a.RuleDetails.Comment > b.RuleDetails.Comment
                                                          else
                                                              result := a.Timestamp > b.Timestamp;
                                                     end
                                                  else
                                                      result := a.RuleDetails.sid > b.RuleDetails.sid;
                                             end
                                          else
                                              result := a.TCPSequenceNumber > b.TCPSequenceNumber;
                                     end
                                  else
                                      result := a.Layer4Protocol > b.Layer4Protocol;
                             end
                          else
                              result := a.IPID > b.IPID;
                     end;
             end;
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.greaterthanorequal (a, b: tMetaDataEntry): boolean;

begin
     result := (a = b) or (a > b);
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.lessthan (a, b: tMetaDataEntry): boolean;

begin
     result := not (a >= b);
end;

(* ----------------------------------- *)

class operator tMetaDataEntry.lessthanorequal (a, b: tMetaDataEntry): boolean;

begin
     result := not (a > b);
end;

(* ----------------------------------- *)

function tMetaDataEntry.DictionaryKey (const OverrideSID : integer = 0) : string;

var value : string;

begin
     if OverrideSID = 0 then
        begin
             case IPVersion of
              4 : value := getDictionaryKeyForMetaDataEntry (IPv4SocketPair, IPID, Layer4Protocol, TCPSequenceNumber, RuleDetails.sid);
              6 : value := getDictionaryKeyForMetaDataEntry (IPv6SocketPair, IPID, Layer4Protocol, TCPSequenceNumber, RuleDetails.sid);
             end;
        end
     else
         begin
              case IPVersion of
               4 : value := getDictionaryKeyForMetaDataEntry (IPv4SocketPair, IPID, Layer4Protocol, TCPSequenceNumber, OverrideSID);
               6 : value := getDictionaryKeyForMetaDataEntry (IPv6SocketPair, IPID, Layer4Protocol, TCPSequenceNumber, OverrideSID);
              end;
         end;

     result := value;
end;

(* ----------------------------------- *)
{ tUnifiedPacketInfo }
(* ----------------------------------- *)

procedure tUnifiedPacketInfo.SwapValues;

begin
     SensorID := swap32 (SensorID);
     EventID := swap32 (EventID);
     EventSeconds := swap32 (EventSeconds);
     PacketSeconds := swap32 (PacketSeconds);
     PacketMicroseconds := swap32 (PacketMicroseconds);
     LinkType := swap32 (LinkType);
     PacketLength := swap32 (PacketLength);
end;

(* ----------------------------------- *)
{ tUnifiedIDSIPv4EventIfnfo }
(* ----------------------------------- *)

procedure tUnifiedIDSIPv4EventIfnfo.SwapValues;

begin
     SensorID := swap32 (SensorID);
     EventID := swap32 (EventID);
     EventSeconds := swap32 (EventSeconds);
     EvenMicroseconds := swap32 (EvenMicroseconds);
     SignatureID := swap32 (SignatureID);
     GeneratorID := swap32 (GeneratorID);
     SignatureRevision := swap32 (SignatureRevision);
     ClassificationID := swap32 (ClassificationID);
     PriorityID := swap32 (PriorityID);
     IPSource := swap32 (IPSource);
     IPDestination := swap32 (IPDestination);
     SourcePort := swap16 (SourcePort);
     DestinationPort := swap16 (DestinationPort);
end;

(* ----------------------------------- *)
{ tUnifiedIDSIPv6EventIfnfo }
(* ----------------------------------- *)

procedure tUnifiedIDSIPv6EventIfnfo.SwapValues;

begin
     SensorID := swap32 (SensorID);
     EventID := swap32 (EventID);
     EventSeconds := swap32 (EventSeconds);
     EvenMicroseconds := swap32 (EvenMicroseconds);
     SignatureID := swap32 (SignatureID);
     GeneratorID := swap32 (GeneratorID);
     SignatureRevision := swap32 (SignatureRevision);
     ClassificationID := swap32 (ClassificationID);
     PriorityID := swap32 (PriorityID);
     SourcePort := swap16 (SourcePort);
     DestinationPort := swap16 (DestinationPort);
end;

(* ----------------------------------- *)
{ tUnifiedExtraData }
(* ----------------------------------- *)

procedure tUnifiedExtraData.SwapValues;

begin
     EventType := swap32 (EventType);
     EventLength := swap32 (EventLength);
     SensorID := swap32 (SensorID);
     EventID := swap32 (EventID);
     EventSeconds := swap32 (EventSeconds);
     ExtraType := swap32 (ExtraType);
     DataType := swap32 (DataType);
     DataLength := swap32 (DataLength);
end;

(* ----------------------------------- *)
{ tUnifiedRecord }
(* ----------------------------------- *)

procedure tUnifiedRecord.initialize;

begin
     RecordType := urtUnknown;
     RecordLength := 0;
     SensorID := 0;
     EventID := 0;
     EventSeconds := 0;
     EvenMicroseconds := 0;

     LinkType := 0;
     PacketLength := 0;
     FillChar (PacketData, SizeOf (PacketData), 0);
//   setLength (PacketData, 0);

     SignatureID := 0;
     GeneratorID := 0;
     SignatureRevision := 0;
     ClassificationID := 0;
     PriorityID := 0;
     IPSource.initialize;
     IPDestination.initialize;
     SourcePort := 0;
     DestinationPort := 0;
     ICMPType := 0;
     ICMPCode := 0;
     Protocol := l4Unknown;
     TTL := 0;
     IPVersion := 0;
     IPID := 0;
     IPHdrLen := 0;
     IPTotalLen := 0;
     TCPSequenceNo := 0;
     TCPAcknowledge := 0;
     TCPWindow := 0;
     TCPFlags := [];
     ImpactFlag := 0;
     Impact := 0;
     Blocked := false;
end;

(* ----------------------------------- *)

procedure tUnifiedRecord.assign (PacketInfo: tUnifiedPacketInfo; PacketData : tRawBuffer);

var TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    Frame        : tRawFrame;
    ParseMaster  : tParseMaster;
    IPv4Parser   : tIPv4Parser;
    IPv6Parser   : tIPv6Parser;
    TCPParser    : tTCPParser;
    UDPParser    : tUDPParser;
    ICMPv4Parser : tICMPv4Parser;

begin
     SensorID := PacketInfo.SensorID;
     EventID := PacketInfo.EventID;
     EventSeconds := PacketInfo.EventSeconds;
     LinkType := PacketInfo.LinkType;

     // Build a RawFrame we can parse for additional IP/TCP Info we want

     TimeDetails.initialize;
     TimeDetails.AbsoluteTime := (PacketInfo.PacketSeconds / 86400) + 25569;

     WireDetails.initialize;
     WireDetails.LinkType := PacketInfo.LinkType;
     WireDetails.CaptureSize := PacketInfo.PacketLength;
     WireDetails.WireSize := PacketInfo.PacketLength;

     Fileposition.initialize;

     Frame := tRawFrame.create (TimeDetails, WireDetails, Fileposition, PacketData);

     try
        ParseMaster := tParseMaster.create (Frame, ppmUpToLayer4);

        try
           // IPv4

           if ppIPv4 in ParseMaster.ParsedProtocols then
              begin
                   IPv4Parser := ParseMaster.getParserByProtocol (ppIPv4) as tIPV4Parser;

                   IPSource.initialize (IPv4Parser.Source);
                   IPDestination.initialize (IPv4Parser.Destination);

                   TTL := IPv4Parser.TTL;
                   IPVersion := 4;
                   IPID := IPv4Parser.Identification;
                   IPHdrLen := IPv4Parser.IHL;
                   IPTotalLen := IPv4Parser.TotalLength;
              end;

           // IPv6

           if ppIPv6 in ParseMaster.ParsedProtocols then
              begin
                   IPv6Parser := ParseMaster.getParserByProtocol (ppIPv6) as tIPV6Parser;

                   IPSource.initialize (IPv6Parser.Source);
                   IPDestination.initialize (IPv6Parser.Destination);

                   TTL := IPv6Parser.HopLimit;
                   IPVersion := 6;
                   IPID := 0;
                   IPHdrLen := 40;
                   IPTotalLen := 40 + IPv6Parser.PayloadLength;
              end;

           // TCP

           if ppTCP in ParseMaster.ParsedProtocols then
              begin
                   TCPParser := ParseMaster.getParserByProtocol (ppTCP) as tTCPParser;

                   TCPSequenceNo := TCPParser.SequenceNumber.Value;
                   TCPAcknowledge := TCPParser.AcknowledgeNumber.Value;
                   TCPWindow := TCPParser.WindowSize;
                   TCPFlags := TCPParser.Flags;

                   SourcePort := TCPParser.SourcePort;
                   DestinationPort := TCPParser.DestinationPort;
              end;

           // UDP

           if ppUDP in ParseMaster.ParsedProtocols then
              begin
                   UDPParser := ParseMaster.getParserByProtocol (ppUDP) as tUDPParser;

                   SourcePort := UDPParser.SourcePort;
                   DestinationPort := UDPParser.DestinationPort;
              end;

           // ICMP

           if ppICMPv4 in ParseMaster.ParsedProtocols then
              begin
                   ICMPv4Parser := ParseMaster.getParserByProtocol (ppICMPv4) as tICMPv4Parser;

                   ICMPType := ICMPv4Parser.ICMPType;
                   ICMPCode := ICMPv4Parser.ICMPCode;
                   ICMPSequence := ICMPv4Parser.EchoSequence;
              end;

        finally
               ParseMaster.free;
        end;

     finally
            Frame.Free;
     end;
end;

(* ----------------------------------- *)

procedure tUnifiedRecord.assign (IPv4Event : tUnifiedIDSIPv4EventIfnfo);

begin
     SensorID := IPv4Event.SensorID;
     EventID := IPv4Event.EventID;
     EventSeconds := IPv4Event.EventSeconds;
     EvenMicroseconds := IPv4Event.EvenMicroseconds;
     SignatureID := IPv4Event.SignatureID;
     GeneratorID := IPv4Event.GeneratorID;
     SignatureRevision := IPv4Event.SignatureRevision;
     ClassificationID := IPv4Event.ClassificationID;
     PriorityID := IPv4Event.PriorityID;
     IPSource.initialize (IPv4Event.IPSource);
     IPDestination.initialize (IPv4Event.IPDestination);

     if IPv4Event.Protocol = 1 then
        begin
             SourcePort := 0;
             DestinationPort := 0;
             ICMPType := IPv4Event.SourcePort;
             ICMPCode := IPv4Event.DestinationPort;
        end
     else
         begin
              SourcePort := IPv4Event.SourcePort;
              DestinationPort := IPv4Event.DestinationPort;
              ICMPType := 0;
              ICMPCode := 0;
         end;

     IPVersion := 4;

     case IPv4Event.Protocol of
       1 : Protocol := l4ICMP;
       6 : Protocol := l4TCP;
      17 : Protocol := l4UDP;
     end;

     ImpactFlag := IPv4Event.ImpactFlag;
     Impact := IPv4Event.Impact;
     Blocked := (IPv4Event.Blocked = 1);
end;

(* ----------------------------------- *)

procedure tUnifiedRecord.assign (IPv6Event : tUnifiedIDSIPv6EventIfnfo);

begin
     SensorID := IPv6Event.SensorID;
     EventID := IPv6Event.EventID;
     EventSeconds := IPv6Event.EventSeconds;
     EvenMicroseconds := IPv6Event.EvenMicroseconds;
     SignatureID := IPv6Event.SignatureID;
     GeneratorID := IPv6Event.GeneratorID;
     SignatureRevision := IPv6Event.SignatureRevision;
     ClassificationID := IPv6Event.ClassificationID;
     PriorityID := IPv6Event.PriorityID;
     IPSource.initialize (IPv6Event.IPSource);
     IPDestination.initialize (IPv6Event.IPDestination);

     if IPv6Event.Protocol = 1 then
        begin
             SourcePort := 0;
             DestinationPort := 0;
             ICMPType := IPv6Event.SourcePort;
             ICMPCode := IPv6Event.DestinationPort;
        end
     else
         begin
              SourcePort := IPv6Event.SourcePort;
              DestinationPort := IPv6Event.DestinationPort;
              ICMPType := 0;
              ICMPCode := 0;
         end;

     IPVersion := 6;

     case IPv6Event.Protocol of
       1 : Protocol := l4ICMP;
       6 : Protocol := l4TCP;
      17 : Protocol := l4UDP;
     end;

     ImpactFlag := IPv6Event.ImpactFlag;
     Impact := IPv6Event.Impact;
     Blocked := (IPv6Event.Blocked = 1);
end;

(* ----------------------------------- *)
{ tSnortAlertParser }
(* ----------------------------------- *)

constructor tSnortAlertParser.create (const Filename: string);

var stream : tFileStream;
    buf    : array[0..3] of byte;

begin
     inherited create;

     priv_labeltemplate := DefaultSnortLabelTemplate;

     priv_filename := Filename;

     // Read the first four bytes to determine the file type

     try
        stream := tFileStream.Create (priv_Filename, fmOpenRead, fmShareDenyNone);
        stream.read (buf, 4);

        if (buf[0] = $5b) and (buf[1] = $2a) and (buf[2] = $2a) and (buf[3] = $5d) then
           priv_fileformat := sfTextAlerts
        else
            if (buf[0] = $0) and (buf[1] = $0) and (buf[2] = $0) and ((buf[3] = $2) or (buf[3] = $7) or (buf[3] = $72) or(buf[3] = $104) or(buf[3] = $105) or(buf[3] = $110)) then
               priv_fileformat := sfUnified2
            else
                priv_fileformat := sfUnknownFormat;
     finally
            stream.Free;
     end;

     setlength (priv_alerts, 0);
     setLength (priv_metadatalist, 0);

     priv_rules := tDictionary<integer, tSnortRule>.create;
end;

(* ----------------------------------- *)

destructor tSnortAlertParser.destroy;

begin
     priv_rules.Free;

     inherited;
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.clear;

begin
     setlength (priv_alerts, 0);
     setLength (priv_metadatalist, 0);
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.AddAlertToList (const alert: tSnortAlert);

var len : integer;

begin
     len := length (priv_alerts);
     setLength (priv_alerts, len + 1);
     priv_alerts[len] := alert;
end;

(* ----------------------------------- *)

function tSnortAlertParser.SnortDateTimeToFrameTimeStamp (value: string): tFrametimestamp;

var datestr : string;
    timestr : string;
    sIndex  : string;
    day     : integer;
    month   : integer;
    hour    : integer;
    minute  : integer;
    second  : integer;
    micro   : integer;
    tstamp  : tFrametimeStamp;

begin
     datestr := getEntry (value, '-', 1);
     timestr := getEntry (value, '-', 2);

     // Month & Day

     sIndex := getEntry (datestr, '/', 1);
     tryStrToInt (sIndex, month);

     sIndex := getEntry (datestr, '/', 2);
     tryStrToInt (sIndex, day);

     // Time

     sIndex := getEntry (timestr, ':', 1);
     tryStrToInt (sIndex, hour);

     sIndex := getEntry (timestr, ':', 2);
     tryStrToInt (sIndex, minute);

     sIndex := getEntry (timestr, ':', 3);
     tryStrToInt (leftstring (sIndex, '.'), second);
     tryStrToInt (rightstring (sIndex, '.'), micro);

     tstamp.initialize;
     tstamp.datetime := encodeDateTime (1970, month, day, hour, minute, second, micro div 1000);
     tstamp.nanoseconds := 1000 * micro;

     result := tstamp;
end;

(* ----------------------------------- *)

// Indexes and rule name

procedure tSnortAlertParser.ParseFirstLine(sLine: string; var alert: tSnortAlert);

var block  : string;
    sIndex : string;

begin
     // Get indexes

     block := getEntry (sLine, ']', 2);
     block := rightstring (block, '[');

     sIndex := getEntry (block, ':', 1);
     tryStrToInt (sIndex, alert.gid);

     sIndex := getEntry (block, ':', 2);
     tryStrToInt (sIndex, alert.sid);

     sIndex := getEntry (block, ':', 3);
     tryStrToInt (sIndex, alert.revision);

     // Get rule name

     block := getEntry (sLine, ']', 3);
     alert.rulename := trim (leftstring (block, '['));
end;

(* ----------------------------------- *)

// Classification and priority

procedure tSnortAlertParser.ParseSecondLine (sLine: string; var alert: tSnortAlert);

var block  : string;
    sIndex : string;

begin
     block := getEntry (sLine, ']', 1);
     alert.classification := rightstring (block, '[');

     block := getEntry (sLine, ']', 2);
     sIndex := rightstring (block, '[Priority: ');
     tryStrToInt (sIndex, alert.priority);
end;

(* ----------------------------------- *)

// Timestamp and Socket Pair

procedure tSnortAlertParser.ParseThirdLine (sLine: string; var alert: tSnortAlert);

var block   : string;

begin
     block := getEntry (sLine, ' ', 1);
     alert.timestamp := SnortDateTimeToFrameTimeStamp (block);           // year is always 1970 because snort alert log doesn't store the year

     block := getEntry (sLine, ' ', 2);

     if isValidIPv4Socket (block) then
        begin
             alert.IPVersion := 4;

             alert.IPv4SocketPair.Source.initialize (block);
        end
     else
         if isValidIPv4String (block) then
            begin
                 alert.IPVersion := 4;

                 alert.IPv4SocketPair.Source.IPAddress.setValueFromString (block);
                 alert.IPv4SocketPair.Source.Port := 0;
            end
        else
            if isValidIPv6Socket (block, true) then
               begin
                    alert.IPVersion := 6;

                    alert.IPv6SocketPair.Source.initialize (block);
               end
            else
                if isValidIPv6String (block) then
                   begin
                        alert.IPVersion := 6;

                        alert.IPv6SocketPair.Source.IPAddress.setValueFromString (block);
                        alert.IPv6SocketPair.Source.Port := 0;
                   end;

     block := getEntry (sLine, ' -> ', 2);

     if isValidIPv4Socket (block) then
        begin
             alert.IPVersion := 4;

             alert.IPv4SocketPair.Destination.initialize (block);
        end
     else
         if isValidIPv4String (block) then
            begin
                 alert.IPVersion := 4;

                 alert.IPv4SocketPair.Destination.IPAddress.setValueFromString (block);
                 alert.IPv4SocketPair.Destination.Port := 0;
            end
        else
            if isValidIPv6Socket (block, true) then
               begin
                    alert.IPVersion := 6;

                    alert.IPv6SocketPair.Destination.initialize (block);
               end
            else
                if isValidIPv6String (block) then
                   begin
                        alert.IPVersion := 6;

                        alert.IPv6SocketPair.Destination.IPAddress.setValueFromString (block);
                        alert.IPv6SocketPair.Destination.Port := 0;
                   end;
end;

(* ----------------------------------- *)

// Layer 4 protocol, TTL etc.

procedure tSnortAlertParser.ParseFourthLine (sLine: string; var alert: tSnortAlert);

var block   : string;
    sIndex  : string;
    help    : integer;

begin
     block := uppercase (getEntry (sLine, ' ', 1));

     // Layer 4 protocol

     if block = 'UDP' then
        alert.Layer4Protocol := l4UDP
     else
     if block = 'TCP' then
        alert.Layer4Protocol := l4TCP
     else
     if (block = 'ICMP') or (block = 'IPV6-ICMP') then
        alert.Layer4Protocol := l4ICMP;

     // TTL

     block := uppercase (getEntry (sLine, ' ', 2));
     sIndex := getEntry (block, ':', 2);

     tryStrToInt (sIndex, help);
     alert.TTL := byte (help);

     // IPID

     block := uppercase (getEntry (sLine, ' ', 4));
     sIndex := getEntry (block, ':', 2);

     tryStrToInt (sIndex, help);
     alert.IPID := word (help);

     // IPHdrLen

     block := uppercase (getEntry (sLine, ' ', 5));
     sIndex := getEntry (block, ':', 2);

     tryStrToInt (sIndex, alert.IPHdrLen);

     // IPTotalLen

     block := uppercase (getEntry (sLine, ' ', 6));
     sIndex := getEntry (block, ':', 2);

     tryStrToInt (sIndex, alert.IPTotalLen);

end;

(* ----------------------------------- *)

procedure tSnortAlertParser.ParseFifthLine (sLine: string; var alert: tSnortAlert);

var block   : string;
    sIndex  : string;
    dummy   : integer;

begin
     // Reduce that STUPID multi space delimiters to a single space

     while pos ('  ', sLine) > 0 do
           sLine := stringReplace (sLine, '  ', ' ', [rfReplaceAll]);

     case alert.Layer4Protocol of
      l4TCP  : begin
                    // Flags

                    block := getEntry (sLine, ' ', 1);
                    alert.TCPFlags := TCPDumpToTCPFlagSet (block);

                    // Sequence

                    block := getEntry (sLine, ' ', 3);
                    val (block, alert.TCPSequenceNo, dummy);

                    // Acknowledge

                    block := getEntry (sLine, ' ', 5);
                    val (block, alert.TCPAcknowledge, dummy);

                    // Window Size

                    block := getEntry (sLine, ' ', 7);
                    val (block, alert.TCPWindow, dummy);
               end;
      l4ICMP : begin
                    // Type

                    block := getEntry (sLine, ' ', 1);
                    sIndex := getEntry (block, ':', 2);

                    val (sIndex, alert.ICMPType, dummy);

                    // Code

                    block := getEntry (sLine, ' ', 2);
                    sIndex := getEntry (block, ':', 2);

                    val (sIndex, alert.ICMPCode, dummy);

                    if (alert.ICMPType = 0) or (alert.ICMPType = 8) then
                       begin
                            // Sequence

                            block := getEntry (sLine, ' ', 4);
                            sIndex := getEntry (block, ':', 2);
                            val (sIndex, alert.ICMPSequence, dummy);
                       end;
               end;
     end;
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.ParseSixtLine(sLine: string; var alert: tSnortAlert);

var block : string;

begin
     block := getEntry (sLine, ']', 1);

     alert.Reference := rightstring (block, '=> ');;
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.ReadTextAlerts;

var infile  : tStringlist;
    sLine   : string;
    alert   : tSnortAlert;
    doParse : boolean;
    row     : integer;
    i       : Integer;

begin
     // TODO: handle IPv6 hits

//   assignfile (fHandle, priv_filename);
//   reset (fHandle);

     infile := tStringlist.Create;
     infile.LoadFromFile (priv_filename);

     doParse := false;
     row := 0;
     i := 0;

//   while not eof (fHandle) do
     for i := 1 to infile.count do
           begin
//              inc (i);
//              readln (fHandle, sLine);
//              SiMain.LogVerbose ('Parsing line ' + IntToStr (i) + ': "' + sLine + '"');

                sLine := infile[i-1];

                // Check for a alert block start

                if pos ('[**] [', sLine) = 1 then
                   begin
                        doParse := true;

                        alert.initialize;
                        row := 1;

                        // parse the ids and rule name

                        parseFirstLine (sLine, alert);
                   end
                else
                    if doParse then
                       begin
                            // Check for block terminator (empty line)

                            if trim (sLine) = '' then
                               begin
                                    doParse := false;

                                    AddAlertToList (alert);
                                    row := 0;
                               end
                            else
                                inc (row);

                            // Parse other stuff

                            case row of
                             2 : ParseSecondLine (sLine, alert);
                             3 : ParseThirdLine (sLine, alert);
                             4 : ParseFourthLine (sLine, alert);
                             5 : ParseFifthLine (sLine, alert);
                             6 : ParseSixtLine (sLine, alert);
                            end;
                       end;

                ReportProgress (infile.Count, i);
           end;

//   closefile (fHandle);
     infile.Free;

     // Add the last alert if it wasn't yet

     if row > 0 then
        AddAlertToList (alert);
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.AddSnortAlert (UnifiedRec: tUnifiedRecord);

var alert : tSnortAlert;
    rule  : tSnortRule;
    len   : integer;

begin
     alert.initialize;

     alert.sid := UnifiedRec.SignatureID;

     // Find the rule if possible, to get the rule msg. This requires loading a snort ruleset first

     if priv_rules.ContainsKey (alert.sid) then
        begin
             rule := priv_rules.Items[alert.sid];

             alert.rulename := rule.RuleName;
             alert.classification := '';
        end
     else
         begin
              alert.rulename := '';
              alert.classification := '';
         end;

     alert.timestamp.datetime := (UnifiedRec.EventSeconds / 86400) + 25569;
     alert.timestamp.nanoseconds := UnifiedRec.EvenMicroseconds * 1000;

     if UnifiedRec.IPVersion = 4 then
        begin
             alert.IPv4SocketPair.Source.IPAddress := UnifiedRec.IPSource.IPv4Address;
             alert.IPv4SocketPair.Source.Port := UnifiedRec.SourcePort;
             alert.IPv4SocketPair.Destination.IPAddress := UnifiedRec.IPDestination.IPv4Address;
             alert.IPv4SocketPair.Destination.Port := UnifiedRec.DestinationPort;
        end
     else
         begin
              alert.IPv6SocketPair.Source.IPAddress := UnifiedRec.IPSource.IPv6Address;
              alert.IPv6SocketPair.Source.Port := UnifiedRec.SourcePort;
              alert.IPv6SocketPair.Destination.IPAddress := UnifiedRec.IPDestination.IPv6Address;
              alert.IPv6SocketPair.Destination.Port := UnifiedRec.DestinationPort;
         end;

     alert.Layer4Protocol := UnifiedRec.Protocol;
     alert.TTL := UnifiedRec.TTL;
     alert.IPVersion := UnifiedRec.IPVersion;
     alert.IPID := UnifiedRec.IPID;
     alert.IPHdrLen := UnifiedRec.IPHdrLen;
     alert.IPTotalLen := UnifiedRec.IPTotalLen;
     alert.TCPSequenceNo := UnifiedRec.TCPSequenceNo;
     alert.TCPAcknowledge := UnifiedRec.TCPAcknowledge;
     alert.TCPWindow := UnifiedRec.TCPWindow;
     alert.TCPFlags := UnifiedRec.TCPFlags;
     alert.ICMPType := UnifiedRec.ICMPType;
     alert.ICMPCode := UnifiedRec.ICMPCode;
     alert.ICMPSequence := UnifiedRec.ICMPSequence;

     len := length (priv_alerts);
     setLength (priv_alerts, len + 1);
     priv_alerts[len] := alert;
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.ReadUnified2;

var stream     : tFileStream;
    lHelp      : longword;
    Rectype    : tUnifiedRecordType;
    Rec        : tUnifiedRecord;
    PacketHdr  : tUnifiedPacketInfo;
    PacketData : tRawBuffer;
    IPv4Event  : tUnifiedIDSIPv4EventIfnfo;
    IPv6Event  : tUnifiedIDSIPv6EventIfnfo;
    ExtraHdr   : tUnifiedExtraData;
    ExtraData  : tRawbuffer;
    alert      : tSnortAlert;
    pos        : int64;
    count      : integer;

begin
     SiMain.EnterMethod ('tSnortAlertParser.ReadUnified2');

     try
        stream := tFileStream.Create (priv_Filename, fmOpenRead, fmShareDenyNone);

        priv_UnifiedReadFail := false;
        count := 0;

        while not (stream.Position >= stream.Size) and not priv_UnifiedReadFail do
              begin
                   rec.initialize;
                   FillChar (PacketData, SizeOf(PacketData), 0);

                   stream.Read (lHelp, 4);
                   lHelp := swap32 (lHelp);

                   case lHelp of
                      2 : rec.RecordType := urtPacket;
                      7 : rec.RecordType := urtIDSEvent;
                     72 : rec.RecordType := urtIDSEventIPv6;
                    104 : rec.RecordType := urtIDSEvent2;
                    105 : rec.RecordType := urtIDSEventIPv62;
                    110 : rec.RecordType := urtExtraData;
                   else
                       priv_UnifiedReadFail := true;
                       SiMain.LogWarning ('Failed parsing Snort Unified2 alert file after alert #' + count.ToString);
                   end;

                   inc (count);

                   stream.Read (lHelp, 4);
                   rec.RecordLength := swap32 (lHelp);

                   case rec.RecordType of
                    urtPacket        : begin
                                            stream.Read (PacketHdr, sizeOf (PacketHdr));
                                            PacketHdr.SwapValues;
                                            stream.Read (PacketData, min (PacketHdr.PacketLength, maxFrameSize));

                                            Rec.assign (PacketHdr, PacketData);
                                       end;
                    urtIDSEvent      : begin
                                            stream.Read (IPv4Event, sizeOf (tUnifiedIDSIPv4EventIfnfo));
                                            IPv4Event.SwapValues;

                                            Rec.Assign (IPv4Event);

                                            // Check if this is followed by a packet record we can merge

                                            pos := stream.Position;

                                            stream.Read (lHelp, 4);
                                            lHelp := swap32 (lHelp);

                                            if lHelp = 2 then
                                               begin
                                                    // dummy-read the length

                                                    stream.Read (lHelp, 4);

                                                    // read the packet record

                                                    stream.Read (PacketHdr, sizeOf (PacketHdr));
                                                    PacketHdr.SwapValues;

                                                    // Check if the packet record has the same event ID as the last record we processed
                                                    // if not, this is a completely new record, otherwise we merge

                                                    if (PacketHdr.SensorID = rec.SensorID) and (PacketHdr.EventID = rec.EventID) then
                                                       begin
                                                            stream.Read (PacketData, min (PacketHdr.PacketLength, maxFrameSize));
                                                            Rec.assign (PacketHdr, PacketData);
                                                       end
                                                    else
                                                        stream.Seek (pos, soFromBeginning);
                                               end
                                            else
                                                stream.Seek (pos, soFromBeginning);
                                       end;
                    urtIDSEventIPv6  : begin
                                            stream.Read (IPv6Event, sizeOf (tUnifiedIDSIPv6EventIfnfo));
                                            IPv6Event.SwapValues;

                                            rec.initialize;
                                            Rec.assign (IPv6Event);

                                            // Check if this is followed by a packet record we can merge

                                            pos := stream.Position;

                                            stream.Read (lHelp, 4);
                                            lHelp := swap32 (lHelp);

                                            if lHelp = 2 then
                                               begin
                                                    stream.Read (PacketHdr, sizeOf (PacketHdr));
                                                    PacketHdr.SwapValues;

                                                    // Check if the packet record has the same event ID as the last record we processed
                                                    // if not, this is a completely new record, otherwise we merge

                                                    if (PacketHdr.SensorID = rec.SensorID) and (PacketHdr.EventID = rec.EventID) then
                                                       begin
                                                            stream.Read (PacketData, min (PacketHdr.PacketLength, maxFrameSize));
                                                            Rec.assign (PacketHdr, PacketData);
                                                       end
                                                    else
                                                        stream.Seek (pos, soFromBeginning);
                                               end
                                            else
                                                stream.Seek (pos, soFromBeginning);
                                       end;
                    urtIDSEvent2     : begin

                                       end;
                    urtIDSEventIPv62 : begin

                                       end;
                    urtExtraData     : begin
                                            stream.Read (ExtraHdr, sizeOf (tUnifiedExtraData));
                                            ExtraHdr.SwapValues;

                                            // DataLength seems to be broken, at least in some cases it's too big

                                            if ExtraHdr.DataLength > (rec.RecordLength - sizeOf (ExtraHdr)) then
                                               ExtraHdr.DataLength := (rec.RecordLength - sizeOf (ExtraHdr));

                                            stream.Read (ExtraData, min (ExtraHdr.DataLength, maxFrameSize));

                                            rec.initialize;
//                                          rec.assign (ExtraData);    not needed at
                                       end;
                   end;

                   // Normalize into Alert info

                   if not priv_UnifiedReadFail then
                      AddSnortAlert (rec);

                   ReportProgress (stream.Size, stream.Position);
              end;

     finally
            stream.Free;
     end;

     SiMain.LeaveMethod ('tSnortAlertParser.ReadUnified2');
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.execute;

var i : integer;

begin
     SiMain.EnterMethod ('tSnortAlertParser.execute');

     priv_percentageprocessed := 0;

     case priv_FileFormat of
      sfTextAlerts : readTextAlerts;
      sfUnified2   : readUnified2;
     end;

     // Transfer alerts to metadata list

     setLength (priv_metadatalist, length (priv_alerts));

     for i := 1 to length (priv_alerts) do
         priv_metadatalist[i-1].initialize (priv_alerts[i-1]);

     SiMain.LeaveMethod ('tSnortAlertParser.execute');
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.ReportProgress (const FileSize, Offset : uint64);

var percentagescanned : integer;
    iHelp             : integer;
    dHelp             : double;

begin
     if (FileSize > 0) and assigned (EventOnAlertPercentProcessed) then
        begin
             dHelp := 100 * Offset / FileSize;
             iHelp := trunc (dHelp);

             if iHelp > priv_percentageprocessed then
                begin
                     priv_percentageprocessed := iHelp;

                     if assigned (EventOnAlertPercentProcessed) then
                        EventOnAlertPercentProcessed (self, FileSize, length (priv_alerts), priv_percentageprocessed);
                end;
        end;
end;

(* ----------------------------------- *)

function tSnortAlertParser.getFilterForAlert (const Alert: tSnortAlert): tFrameFilter;

var filter      : tFrameFilter;
    socketpair4 : tIPv4SocketPair;
    socketpair6 : tIPv6SocketPair;
    goAhead     : boolean;

begin
     goAhead := true;

     filter.initialize;

     if priv_ignoreSYN and (tcpSYN in Alert.TCPFlags) then
        goAhead := false;

     if priv_ignoreICMP and (Alert.Layer4Protocol = l4ICMP) then
        goAhead := false;

     if goAhead then
        begin
             // Set the filter label

             filter.FilterLabel := Alert.getFilterLabel (priv_labeltemplate);

             case Alert.IPVersion of
              4 : begin
                       socketpair4 := Alert.IPv4SocketPair;
                       socketpair4.normalize;

                       if Alert.Layer4Protocol = l4UDP then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4UDP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair4.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair4.Destination.Port;
                          end;

                       if Alert.Layer4Protocol = l4TCP then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4TCP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair4.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair4.Destination.Port;
                          end;

                       if Alert.Layer4Protocol = l4ICMP then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4ICMP;
                               filter.PortFilter.PortScope := fsAny;
                               filter.PortFilter.PortFilterType := ftSingle;
                          end;
                  end;
              6 : begin
                       socketpair6 := Alert.IPv6SocketPair;
                       socketpair6.normalize;

                       if Alert.Layer4Protocol = l4UDP then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4UDP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair6.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair6.Destination.Port;
                          end;

                       if Alert.Layer4Protocol = l4TCP then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4TCP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair6.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair6.Destination.Port;
                          end;

                       if Alert.Layer4Protocol = l4ICMP then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4ICMP;
                               filter.PortFilter.PortScope := fsAny;
                               filter.PortFilter.PortFilterType := ftSingle;
                          end;
                  end;
             end;
     end;

     result := filter;
end;

(* ----------------------------------- *)

function tSnortAlertParser.isValidFile: boolean;

begin
     result := not (priv_fileformat = sfUnknownFormat);

{
var fHandle : textfile;
    sLine   : string;
    valid   : boolean;

begin
     valid := false;

     assignfile (fHandle, priv_filename);

     try
        reset (fHandle);

        if not eof (fHandle) then
           begin
                sLine := '';

                repeat
                      readln (fHandle, sLine);
                until (trim (sLine) <> '') or eof (fHandle);

                // Check for a alert block start

                if pos ('[**] [', sLine) = 1 then
                   valid :=  true;
           end;
     finally
            closefile (fHandle);
     end;

     result := valid;
}
end;

(* ----------------------------------- *)

function tSnortAlertParser.GetFilterList : tFilterSet;

var len    : integer;
    alert  : tSnortAlert;
    filter : tFrameFilter;
    list   : tFilterSet;
    i      : integer;

begin
     len := length (priv_alerts);
     list := tFilterSet.create;

     for i := 1 to len do
         begin
              alert := priv_alerts[i-1];
              filter := getFilterForAlert (alert);

              if not filter.isUnspecified then
                 list.AddFilter (Filter);
         end;

     result := list;
end;

(* ----------------------------------- *)

procedure tSnortAlertParser.LoadRuleDefinitions (const Filename: string; const IncludeDisabledRules : boolean = false);

var rulefile : tStringlist;
    line     : string;
    rule     : tSnortRule;
    doparse  : boolean;
    i,p      : integer;

begin
     SiMain.EnterMethod ('tSnortAlertParser.LoadRuleDefinitions');

     assert (fileexists (Filename), 'Rule file does not exist');

     rulefile := tStringlist.Create;
     rulefile.LoadFromFile (Filename);

     SiMain.LogVerbose ('Alert log file has a total of ' + rulefile.count.ToString + ' lines, including headers and comments');

     for i := 1 to rulefile.Count do
         begin
              line := rulefile[i-1];

              p := pos ('alert ', lowercase (line));
              doParse := (p > 0) and (p < 4);

              if (pos ('#', line) = 1) and not IncludeDisabledRules then
                 doParse := false;

              if doParse then
                 begin
//                    SiMain.LogDebug (line);

                      if rule.ParseFromLine (line) then
                         begin
                              if not priv_rules.ContainsKey (rule.SignatureID) then
                                 priv_rules.Add (rule.SignatureID, rule);
                         end;
                 end;
         end;

     rulefile.Free;

     SiMain.LeaveMethod ('tSnortAlertParser.LoadRuleDefinitions');
end;

(* ----------------------------------- *)

end.
