{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit commander_tools;

interface
         function getTotalCommanderDescription (Filename: string): string;
         procedure setTotalCommanderDescription (Filename, Description : string);

implementation

uses Classes,
     SysUtils,
     Toolkit;


type tDescriptionEntry = record
                               Filename    : string;
                               Description : string;
                               ID          : byte;
                         end;

(* --------------------------------------- *)

function getDescription (DescriptionLine : string) : tDescriptionEntry;

var entry     : tDescriptionEntry;

begin
     // Maybe the Filename is encapsulated in quotation marks

     if pos ('"', DescriptionLine) = 0 then
        begin
             entry.Filename := getEntry (DescriptionLine, ' ', 1);
             entry.Description := rightstring (DescriptionLine, ' ');
        end
     else
         begin
              entry.Filename := getEntry (DescriptionLine, '"', 2);
              entry.Description := trim (getEntry (DescriptionLine, '"', 3));
         end;

     if entry.Description <> '' then
        begin
             entry.Description := stringReplace (entry.Description, '\n', CRLF, [rfReplaceAll]);

             if pos (chr(4), entry.Description) > 0 then
                entry.Description := leftstring (entry.Description, chr(4));
        end;

     result := entry;
end;


(* --------------------------------------- *)

function getTotalCommanderDescription (Filename: string): string;

var list   : tStringList;
    sPath  : string;
    sFile  : string;
    entry  : tDescriptionEntry;
    i      : integer;

begin
     sPath := extractFilePath (Filename);
     sFile := extractFilename (Filename);

     // See if there is a description file

     if FileExists (sPath + 'descript.ion') then
        begin
             list := tStringlist.Create;
             list.LoadFromFile(sPath + 'descript.ion');

             for i := 1 to list.Count do
                 begin
                      entry := getDescription (list[i-1]);

                      if lowercase (entry.Filename) = lowercase (filename) then
                         result := entry.Description;
                 end;

             list.Free;
        end
     else
         result := '';
end;

(* --------------------------------------- *)

procedure setTotalCommanderDescription (Filename, Description : string);

var list   : tStringList;
    sPath  : string;
    sFile  : string;
    entry  : tDescriptionEntry;
    i      : integer;

begin
     list := tStringlist.Create;

     sPath := extractFilePath (Filename);
     sFile := extractFilename (Filename);

     Description := stringreplace (Description, CRLF, '\n', [rfReplaceAll]);

     if FileExists (sPath + 'descript.ion') then
        begin
             // Load the existing description file

             list.LoadFromFile(sPath + 'descript.ion');

             for i := 1 to list.Count do
                 begin
                      entry := getDescription (list[i-1]);

                      // If there is an entry for the file, remove it

                      if lowercase (entry.Filename) = lowercase (filename) then
                         list.Delete(i-1);
                 end;
        end;

     // Add the description to the list

     list.Add ('"' + sFile + '" ' + description + chr(4) + chr ($c2));

     list.SaveToFile(sPath + 'descript.ion');
     list.Free;

     FileSetAttr (sPath + 'descript.ion', faHidden);
end;

(* --------------------------------------- *)

end.
