unit definition_symptoms;

interface

const LowTTLThreshold = 8;

type tAnalysisFlag = (afClientAllZeroCRC, afServerAllZeroCRC);
     tAnalysisFlagSet = set of tAnalysisFlag;

     tTCPSymptom = (tsNone, tsFrameDuplicate, tsIPAllowsFragmentation, tsIPLowTTL, tsIPStrangeSource, tsRoutingDuplicate, tsFrameOutOfOrder, tsHandshakeOutOfOrder, tsSYNRetransmit, tsSYNACKRetransmit, tsSYNSYNACKSameSource,
                    tsSYNACKNotMatchingSYN, tsHalfOpenReset, tsHalfOpenFIN, tsFINSeqNoWrong, tsFINwithoutACKFlag, tsClientAbort, tsServerAbort, tsSocketRefused, tsClosedInactive, tsClientReceivedNoData, tsClientSocketTest,
                    tsSYNFlagWithPayload, tsSpuriousPacket, tcpSYNwithSequenceChg, tsNoClientMSSOption, tsNoServerMSSOption, tsNoServerScaleOption, tsNoClientScaleOption, tsSYNwithNonZeroACKNo, tsZeroClientWindow,
                    tsZeroServerWindow, tsKeepAlive, tsInvalidFlagSet, tsUnseenPacket, tsSocketsReused, tsPortReuseConfusion, tsForgedReset, tsRSTOutOfWindow, tsRSTAfterTeardown, tsPacketsAfterReset, tsClientFINRetransmit,
                    tsServerFINRetransmit, tsOneSidedClientFIN, tsUrgentPointerWithoutFlag, tsOneSidedServerFIN, tsServerPacketsAfterRST, tsClientPacketsAfterRST, tsMultiClientTTL, tsMultiServerTTL, tsClientIPIDIncrements,
                    tsServerIPIDIncrements, tsEstablishAfterRefuse, tsPrematureNewSession, tsIdenticalReverseSession, tsDuplicateInterface, tsInvalidConnectionState, tsZeroWindowOnSYNACK, tsThirdAckWithPayload, tstcpSequenceWrapped,
                    tsSACKEdgesSeen, tsClientMSSTooLow, tsServerMSSTooLow, tsHandshakeCompleteDetermined, tsZeroISN, tsPortAndISNReusedAfterReset, tsMultiTCPClientSYNOptionSets, tsMultiTCPServerSYNOptionSets, tsTimestampValueMismatch,
                    tsHandshakeFailByTimestamp, tsCumbersomeHandshake, tsClientUsesMinMSS, tsServerUsesMinMSS, tsSegmentExceedsMSS, tsTrueCRCErrors, tsSYNACKWithNewISN);

     tTCPSymptomSet = set of tTCPSymptom;

     tSymptomGroup = (tsgNone, tsgCapture, tsgIPLayer, tsgHandshake, tsgTeardown, tsgWindow, tsgOptions, tsgConversation, tsgSecurity, tsgMisc);
     tSymptomGroups = set of tSymptomGroup;

     tSymptomCriticality = (tscNone, tscInfo, tscMinor, tscMajor, tscCritical);
     tTCPSymptomCriticalities = set of tSymptomCriticality;

     tTCPSymptomMeta = record
                             Symptom        : tTCPSymptom;
                             Name           : string;
                             Description    : string;
                             Group          : tSymptomGroup;
                             Criticality    : tSymptomCriticality;
                             Score          : integer;

                             procedure initialize (const TCPSymptom : tTCPSymptom);
                             procedure setRemainingTCPSymptomMeta (const TCPSymptom : tTCPSymptom);
                       end;

     tSocketSymptom = (ssNone, ssLegitConnectionsRefused);
     tSocketSymptomSet = set of tSocketSymptom;

     tSocketSymptomMeta = record
                                Symptom        : tSocketSymptom;
                                Name           : string;
                                Description    : string;
                                Group          : tSymptomGroup;
                                Criticality    : tSymptomCriticality;
                                Score          : integer;

                                procedure initialize (const SocketSymptom : tSocketSymptom);
                                procedure setRemainingSocketSymptomMeta (const SocketSymptom : tSocketSymptom);
                          end;

     function getTCPSymptomDescription (const TCPSymptom : tTCPSymptom) : string;
     function getTCPSymptomMeta (const TCPSymptom : tTCPSymptom) : tTCPSymptomMeta;

     function getSocketSymptomDescription (const SocketSymptom : tSocketSymptom) : string;

     function GetSwappedSymptoms (Symptoms : tTCPSymptomSet) : tTCPSymptomSet;
     function SymptomToStr (const Symptom : tTCPSymptom) : string;
     function SymptomSetToStr (const SymptomSet : tTCPSymptomSet; const Separator : string = '|') : string;

     function TCPSymptomSetToString (const symptoms : tTCPSymptomSet) : string;

implementation

uses SysUtils;

(* ------------------------------------------- *)

function getTCPSymptomDescription (const TCPSymptom : tTCPSymptom) : string;

var value : string;

begin
     value := '';

     case TCPSymptom of
      tsNone                        : value := 'No Symptom';
      tsFrameDuplicate              : value := 'The TCP flow seems to contain layer 2 duplicates';
      tsIPAllowsFragmentation       : value := 'At least one frame in the TCP flow allows fragmentation via the IP flags';
      tsIPLowTTL                    : value := 'Low IP TTL values observed (less than ' + IntToStr (LowTTLThreshold) + ' hops)';
//    tsIPStrangeSource             : value := '';
      tsRoutingDuplicate            : value := 'The TCP flow seems to contain routed duplicates';
      tsFrameOutOfOrder             : value := 'The TCP flow contains packets that arrived out of order (capture problem)';
      tsHandshakeOutOfOrder         : value := 'The TCP handshake was captured out of order';
      tsSYNRetransmit               : value := 'SYN packets were retransmitted, usually because no SYN/ACK arrived in time';
      tsSYNACKRetransmit            : value := 'SYN/ACK packets were retransmitted, usually because they got lost on the way to the client';
      tsSYNSYNACKSameSource         : value := 'SYN and SYN/ACK packets were seen coming from the same source IP';
      tsSYNACKNotMatchingSYN        : value := 'A SYN/ACK packet arrived with an acknowledge number that does not meet the expectation';
      tsHalfOpenReset               : value := 'The connection was already half open (SYN - SYN/ACK) before it was killed with an RST packet';
      tsHalfOpenFIN                 : value := 'The connection was already half open (SYN - SYN/ACK) before it was shut down with a FIN packet';
      tsFINSeqNoWrong               : value := 'The sequence number of a FIN packet does not match the expectation';
      tsFINwithoutACKFlag           : value := 'A FIN packet arrived without the ACK flag set';
      tsClientAbort                 : value := 'The client reset the connection without any data having been transferred';
      tsServerAbort                 : value := 'The server reset the connection without any data having been transferred';
      tsSocketRefused               : value := 'The server shut down the connection immediately after accepting it (TCP socket refused by application)';
      tsClosedInactive              : value := 'The server shut down the connection due to inactivity without any data having been transferred';
      tsClientReceivedNoData        : value := 'The client shut down the connection without having received any data';
      tsClientSocketTest            : value := 'The client shut down the connection immediately after establishing it. This often indicates a TCP socket test.';
      tsSYNFlagWithPayload          : value := 'The SYN packet contains TCP payload data';
      tsSpuriousPacket              : value := 'A packet was retransmitted even though the original had been received and acknowledged';
//    tcpSYNwithSequenceChg         : value := '';
      tsNoClientMSSOption           : value := 'The client sent no MSS option in the SYN packet';
      tsNoServerMSSOption           : value := 'The server sent no MSS option in the SYN/ACK packet';
      tsNoServerScaleOption         : value := 'The server offered no window scaling option in the SYN/ACK packet';
      tsNoClientScaleOption         : value := 'The client offered no window scaling option in the SYN packet';
      tsSYNwithNonZeroACKNo         : value := 'The acknowledge number of a SYN packet was greater than zero';
      tsZeroClientWindow            : value := 'The client sent a zero window value';
      tsZeroServerWindow            : value := 'The server sent a zero window value';
//    tsKeepAlive                   : value := '';
      tsInvalidFlagSet              : value := 'An invalid TCP flag set was detected';
      tsUnseenPacket                : value := 'The TCP handshake was complete without having seen the SYN/ACK packet';
      tsSocketsReused               : value := 'The connection reused a socket pair (both IPs and ports identical to a previous connection)';
      tsPortReuseConfusion          : value := 'A connection reused a socket pair of a previous connection too soon. Packets arrived for the new connection with sequence or acknowledge values for the old.';
      tsForgedReset                 : value := 'A reset packet arrived with a TTL that does not match the expected TTL';
      tsRSTOutOfWindow              : value := 'A reset packet arrived with a sequence number outside the expected TCP window';
      tsRSTAfterTeardown            : value := 'A reset packet arrived after at least one side used a FIN packet to shut down the connection';
      tsPacketsAfterReset           : value := 'Packets arrived after the connection was terminated by the same sender using a reset packet';
      tsClientFINRetransmit         : value := 'The client retransmitted a FIN packet';
      tsServerFINRetransmit         : value := 'The server retransmitted a FIN packet';
      tsOneSidedClientFIN           : value := 'The connection was shut down by the client using a FIN packet, but none seen from the server';
      tsUrgentPointerWithoutFlag    : value := 'A packet contained an urgent pointer, but the urgent flag was not set';
      tsOneSidedServerFIN           : value := 'The connection was shut down by the server using a FIN packet, but none seen from the client';
      tsServerPacketsAfterRST       : value := 'Packets arrived from the server after the connection was known to be already reset';
      tsClientPacketsAfterRST       : value := 'Packets arrived from the client after the connection was known to be already reset';
      tsMultiClientTTL              : value := 'Observed multiple TTL values coming from the client';
      tsMultiServerTTL              : value := 'Observed multiple TTL values coming from the server';
      tsClientIPIDIncrements        : value := 'The IPID values coming from the clients are incremented for subsequent packets';
      tsServerIPIDIncrements        : value := 'The IPID values coming from the server are incremented for subsequent packets';
      tsEstablishAfterRefuse        : value := 'The TCP connection was established succesfully after being refused at first';
      tsPrematureNewSession         : value := 'The connection reused a socket pair without waiting enough time';
      tsIdenticalReverseSession     : value := 'An identical connection using the same socket pair exists with client and server role reversed';
      tsDuplicateInterface          : value := 'Packets coming from the same source were captured on more than one physical interface';
      tsInvalidConnectionState      : value := 'The connection state is invalid, e.g. when receiving a SYN/ACK from the same origin as the SYN';
      tsZeroWindowOnSYNACK          : value := 'The SYN packet contains a zero TCP window value';
      tsThirdAckWithPayload         : value := 'The ACK packet of the handshake contains TCP payload';
      tstcpSequenceWrapped          : value := 'The TCP sequence number was wrapped when hitting the 32 bit boundary';
      tsSACKEdgesSeen               : value := 'The connection has packets containing TCP SACK edges';
      tsClientMSSTooLow             : value := 'The client sent a MSS value that is invalid (< 536 for IPv4, < 1280 for IPv6)';
      tsServerMSSTooLow             : value := 'The server sent a MSS value that is invalid (< 536 for IPv4, < 1280 for IPv6)';
      tsHandshakeCompleteDetermined : value := 'Not all handshake packets were seen in the capture, but later packets indicate it must have been completed';
      tsZeroISN                     : value := 'The initial sequence number of the SYN packet is zero';
      tsPortAndISNReusedAfterReset  : value := 'A socket pair was reused with an identical ISN as the previous connection';
      tsMultiTCPClientSYNOptionSets : value := 'Multiple SYN packets coming from the client carry different TCP option sets';
      tsMultiTCPServerSYNOptionSets : value := 'Multiple SYN/ACK packets coming from the server carry different TCP option sets';
      tsTimestampValueMismatch      : value := 'The echo value of a TCP timestamp did not match the original value';
      tsHandshakeFailByTimestamp    : value := 'The TCP handshake failed because the server echoed an invalid TCP timestamp option value';
      tsCumbersomeHandshake         : value := 'The TCP handshake had some trouble to complete sucessfully';
      tsClientUsesMinMSS            : value := 'The client uses the minimum MSS of 536, which is highly inefficient';
      tsServerUsesMinMSS            : value := 'The server uses the minimum MSS of 536, which is highly inefficient';
      tsSegmentExceedsMSS           : value := 'At least one TCP segment is larger than the MSS of either client or server';
      tsTrueCRCErrors               : value := 'At least one TCP packet had a truly broken CRC, probably leading to it being ignored';
      tsSYNACKWithNewISN            : value := 'Another SYN/ACK was seen with a different sequence number than before';
     else
         raise Exception.Create('No description available for TCP symptom');
     end;

     result := value;
end;

(* ------------------------------------------- *)
{ tTCPSymptomMeta }
(* ------------------------------------------- *)

procedure tTCPSymptomMeta.initialize (const TCPSymptom : tTCPSymptom);

begin
     Symptom := TCPSymptom;
     Description := getTCPSymptomDescription (Symptom);

     setRemainingTCPSymptomMeta (Symptom);
end;

(* ------------------------------------------- *)

function getTCPSymptomMeta (const TCPSymptom : tTCPSymptom) : tTCPSymptomMeta;

var value : tTCPSymptomMeta;

begin
     value.initialize (TCPSymptom);

     result := value;
end;

(* ------------------------------------------- *)

procedure tTCPSymptomMeta.setRemainingTCPSymptomMeta (const TCPSymptom : tTCPSymptom);

begin
     case TCPSymptom of
      tsNone                        : begin
                                           Name := 'tsNone';
                                           Group := tsgNone;
                                           Criticality := tscNone;
                                           Score := 0;
                                      end;
      tsFrameDuplicate              : begin
                                           Name := 'tsFrameDuplicate';
                                           Group := tsgCapture;
                                           Criticality := tscMinor;
                                           Score := 1;
                                      end;
      tsIPAllowsFragmentation       : begin
                                           Name := 'tsIPAllowsFragmentation';
                                           Group := tsgConversation;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsIPLowTTL                    : begin
                                           Name := 'tsIPLowTTL';
                                           Group := tsgIPLayer;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
//    tsIPStrangeSource             : begin

//                                    end;
      tsRoutingDuplicate            : begin
                                           Name := 'tsRoutingDuplicate';
                                           Group := tsgCapture;
                                           Criticality := tscMinor;
                                           Score := 1;
                                      end;
      tsFrameOutOfOrder             : begin
                                           Name := 'tsFrameOutOfOrder';
                                           Group := tsgCapture;
                                           Criticality := tscMinor;
                                           Score := 1;
                                      end;
      tsHandshakeOutOfOrder         : begin
                                           Name := 'tsHandshakeOutOfOrder';
                                           Group := tsgCapture;
                                           Criticality := tscMinor;
                                           Score := 1;
                                      end;
      tsSYNRetransmit               : begin
                                           Name := 'tsSYNRetransmit';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 10;
                                      end;
      tsSYNACKRetransmit            : begin
                                           Name := 'tsSYNACKRetransmit';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 10;
                                      end;
      tsSYNSYNACKSameSource         : begin
                                           Name := 'tsSYNSYNACKSameSource';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsSYNACKNotMatchingSYN        : begin
                                           Name := 'tsSYNACKNotMatchingSYN';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsHalfOpenReset               : begin
                                           Name := 'tsHalfOpenReset';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsHalfOpenFIN                 : begin
                                           Name := 'tsHalfOpenFIN';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsFINSeqNoWrong               : begin
                                           Name := 'tsFINSeqNoWrong';
                                           Group := tsgTeardown;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsFINwithoutACKFlag           : begin
                                           Name := 'tsFINwithoutACKFlag';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsClientAbort                 : begin
                                           Name := 'tsClientAbort';
                                           Group := tsgTeardown;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsServerAbort                 : begin
                                           Name := 'tsServerAbort';
                                           Group := tsgTeardown;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsSocketRefused               : begin
                                           Name := 'tsSocketRefused';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsClosedInactive              : begin
                                           Name := 'tsClosedInactive';
                                           Group := tsgTeardown;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsClientReceivedNoData        : begin
                                           Name := 'tsClientReceivedNoData';
                                           Group := tsgConversation;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsClientSocketTest            : begin
                                           Name := 'tsClientSocketTest';
                                           Group := tsgSecurity;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsSYNFlagWithPayload          : begin
                                           Name := 'tsSYNFlagWithPayload';
                                           Group := tsgHandshake;
                                           Criticality := tscInfo;
                                           Score := 1;
                                      end;
      tsSpuriousPacket              : begin
                                           Name := 'tsSpuriousPacket';
                                           Group := tsgConversation;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
//    tcpSYNwithSequenceChg         : begin

//                                    end;
      tsNoClientMSSOption           : begin
                                           Name := 'tsNoClientMSSOption';
                                           Group := tsgOptions;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsNoServerMSSOption           : begin
                                           Name := 'tsNoServerMSSOption';
                                           Group := tsgOptions;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsNoServerScaleOption         : begin
                                           Name := 'tsNoServerScaleOption';
                                           Group := tsgOptions;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsNoClientScaleOption         : begin
                                           Name := 'tsNoClientScaleOption';
                                           Group := tsgOptions;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsSYNwithNonZeroACKNo         : begin
                                           Name := 'tsSYNwithNonZeroACKNo';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsZeroClientWindow            : begin
                                           Name := 'tsZeroClientWindow';
                                           Group := tsgWindow;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsZeroServerWindow            : begin
                                           Name := 'tsZeroServerWindow';
                                           Group := tsgWindow;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
//    tsKeepAlive                   : begin

//                                    end;
      tsInvalidFlagSet              : begin
                                           Name := 'tsInvalidFlagSet';
                                           Group := tsgSecurity;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsUnseenPacket                : begin
                                           Name := 'tsUnseenPacket';
                                           Group := tsgCapture;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsSocketsReused               : begin
                                           Name := 'tsSocketsReused';
                                           Group := tsgHandshake;
                                           Criticality := tscInfo;
                                           Score := 10;
                                      end;
      tsPortReuseConfusion          : begin
                                           Name := 'tsPortReuseConfusion';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 50;
                                      end;
      tsForgedReset                 : begin
                                           Name := 'tsForgedReset';
                                           Group := tsgSecurity;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsRSTOutOfWindow              : begin
                                           Name := 'tsRSTOutOfWindow';
                                           Group := tsgSecurity;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsRSTAfterTeardown            : begin
                                           Name := 'tsRSTAfterTeardown';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsPacketsAfterReset           : begin
                                           Name := 'tsPacketsAfterReset';
                                           Group := tsgTeardown;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsClientFINRetransmit         : begin
                                           Name := 'tsClientFINRetransmit';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsServerFINRetransmit         : begin
                                           Name := 'tsServerFINRetransmit';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsOneSidedClientFIN           : begin
                                           Name := 'tsOneSidedClientFIN';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsUrgentPointerWithoutFlag    : begin
                                           Name := 'tsUrgentPointerWithoutFlag';
                                           Group := tsgConversation;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsOneSidedServerFIN           : begin
                                           Name := 'tsOneSidedServerFIN';
                                           Group := tsgTeardown;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsServerPacketsAfterRST       : begin
                                           Name := 'tsServerPacketsAfterRST';
                                           Group := tsgTeardown;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsClientPacketsAfterRST       : begin
                                           Name := 'tsClientPacketsAfterRST';
                                           Group := tsgTeardown;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsMultiClientTTL              : begin
                                           Name := 'tsMultiClientTTL';
                                           Group := tsgIPLayer;
                                           Criticality := tscInfo;
                                           Score := 10;
                                      end;
      tsMultiServerTTL              : begin
                                           Name := 'tsMultiServerTTL';
                                           Group := tsgIPLayer;
                                           Criticality := tscInfo;
                                           Score := 10;
                                      end;
      tsClientIPIDIncrements        : begin
                                           Name := 'tsClientIPIDIncrements';
                                           Group := tsgIPLayer;
                                           Criticality := tscInfo;
                                           Score := 0;
                                      end;
      tsServerIPIDIncrements        : begin
                                           Name := 'tsServerIPIDIncrements';
                                           Group := tsgIPLayer;
                                           Criticality := tscInfo;
                                           Score := 0;
                                      end;
      tsEstablishAfterRefuse        : begin
                                           Name := 'tsEstablishAfterRefuse';
                                           Group := tsgHandshake;
                                           Criticality := tscInfo;
                                           Score := 50;
                                      end;
      tsPrematureNewSession         : begin
                                           Name := 'tsPrematureNewSession';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsIdenticalReverseSession     : begin
                                           Name := 'tsIdenticalReverseSession';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsDuplicateInterface          : begin
                                           Name := 'tsDuplicateInterface';
                                           Group := tsgCapture;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsInvalidConnectionState      : begin
                                           Name := 'tsInvalidConnectionState';
                                           Group := tsgConversation;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsZeroWindowOnSYNACK          : begin
                                           Name := 'tsZeroWindowOnSYNACK';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsThirdAckWithPayload         : begin
                                           Name := 'tsThirdAckWithPayload';
                                           Group := tsgHandshake;
                                           Criticality := tscInfo;
                                           Score := 0;
                                      end;
      tstcpSequenceWrapped          : begin
                                           Name := 'tstcpSequenceWrapped';
                                           Group := tsgConversation;
                                           Criticality := tscInfo;
                                           Score := 5;
                                      end;
      tsSACKEdgesSeen               : begin
                                           Name := 'tsSACKEdgesSeen';
                                           Group := tsgConversation;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsClientMSSTooLow             : begin
                                           Name := 'tsClientMSSTooLow';
                                           Group := tsgOptions;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsServerMSSTooLow             : begin
                                           Name := 'tsServerMSSTooLow';
                                           Group := tsgOptions;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsHandshakeCompleteDetermined : begin
                                           Name := 'tsHandshakeCompleteDetermined';
                                           Group := tsgHandshake;
                                           Criticality := tscInfo;
                                           Score := 0;
                                      end;
      tsZeroISN                     : begin
                                           Name := 'tsZeroISN';
                                           Group := tsgHandshake;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsPortAndISNReusedAfterReset  : begin
                                           Name := 'tsPortAndISNReusedAfterReset';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsMultiTCPClientSYNOptionSets : begin
                                           Name := 'tsMultiTCPClientSYNOptionSets';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 25;
                                      end;
      tsMultiTCPServerSYNOptionSets : begin
                                           Name := 'tsMultiTCPServerSYNOptionSets';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 25;
                                      end;
      tsTimestampValueMismatch      : begin
                                           Name := 'tsTimestampValueMismatch';
                                           Group := tsgConversation;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsHandshakeFailByTimestamp    : begin
                                           Name := 'tsHandshakeFailByTimestamp';
                                           Group := tsgHandshake;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsCumbersomeHandshake         : begin
                                           Name := 'tsCumbersomeHandshake';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsClientUsesMinMSS            : begin
                                           Name := 'tsClientUsesMinMSS';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsServerUsesMinMSS            : begin
                                           Name := 'tsServerUsesMinMSS';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 10;
                                      end;
      tsSegmentExceedsMSS           : begin
                                           Name := 'tsSegmentExceedsMSS';
                                           Group := tsgConversation;
                                           Criticality := tscMajor;
                                           Score := 25;
                                      end;
      tsTrueCRCErrors               : begin
                                           Name := 'tsTrueCRCErrors';
                                           Group := tsgConversation;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
      tsSYNACKWithNewISN            : begin
                                           Name := 'tsSYNACKWithNewISN';
                                           Group := tsgHandshake;
                                           Criticality := tscMinor;
                                           Score := 5;
                                      end
     else
         raise Exception.Create('No meta available for TCP symptom');
     end;
end;

(* ------------------------------------------- *)

function TCPSymptomSetToString (const symptoms : tTCPSymptomSet) : string;

var symptom : tTCPSymptom;

begin
     result := '';

     for symptom in symptoms do
         result := result + ' ' + getTCPSymptomMeta (symptom).Name;

     result := trim (result);
     result := stringreplace (result, ' ', ',', [rfReplaceAll]);
end;

(* ------------------------------------------- *)

function getSocketSymptomDescription (const SocketSymptom : tSocketSymptom) : string;

var value : string;

begin
     value := '';

     case SocketSymptom of
      ssNone                        : value := 'No Symptom';
      ssLegitConnectionsRefused     : value := 'Socket accepted some connections and refused others';
     end;

     result := value;
end;

(* ------------------------------------------- *)

function GetSwappedSymptoms (Symptoms : tTCPSymptomSet) : tTCPSymptomSet;

var value : tTCPSymptomSet;

begin
     value := Symptoms;

     if (tsClientFINRetransmit in Symptoms) then
        value := value - [tsClientFINRetransmit] + [tsServerFINRetransmit];

     if (tsMultiClientTTL in Symptoms) then
        value := value - [tsMultiClientTTL] + [tsMultiServerTTL];

     if (tsClientIPIDIncrements in Symptoms) then
        value := value - [tsClientIPIDIncrements] + [tsServerIPIDIncrements];

     if (tsZeroClientWindow in Symptoms) then
        value := value - [tsZeroClientWindow] + [tsZeroServerWindow];

     if (tsClientAbort in Symptoms) then
        value := value - [tsClientAbort] + [tsServerAbort];

     if (tsNoClientMSSOption in Symptoms) then
        value := value - [tsNoClientMSSOption] + [tsNoServerMSSOption];

     if (tsNoClientScaleOption in Symptoms) then
        value := value - [tsNoClientScaleOption] + [tsNoServerScaleOption];


     if (tsServerFINRetransmit in Symptoms) then
        value := value - [tsServerFINRetransmit] + [tsClientFINRetransmit];

     if (tsMultiServerTTL in Symptoms) then
        value := value - [tsMultiServerTTL] + [tsMultiClientTTL];

     if (tsServerIPIDIncrements in Symptoms) then
        value := value - [tsServerIPIDIncrements] + [tsClientIPIDIncrements];

     if (tsZeroServerWindow in Symptoms) then
        value := value - [tsZeroServerWindow] + [tsZeroClientWindow];

     if (tsServerAbort in Symptoms) then
        value := value - [tsServerAbort] + [tsClientAbort];

     if (tsNoServerMSSOption in Symptoms) then
        value := value - [tsNoServerMSSOption] + [tsNoClientMSSOption];

     if (tsNoServerScaleOption in Symptoms) then
        value := value - [tsNoServerScaleOption] + [tsNoClientScaleOption];

     if (tsClientMSSTooLow in Symptoms) then
        value := value - [tsClientMSSTooLow] + [tsServerMSSTooLow];

     if (tsServerMSSTooLow in Symptoms) then
        value := value - [tsServerMSSTooLow] + [tsClientMSSTooLow];

     result := value;
end;

(* ------------------------------------------- *)

function SymptomToStr (const Symptom : tTCPSymptom) : string;

begin
     case Symptom of
      tsNone                        : result := '';
      tsFrameDuplicate              : result := 'Duplicate Frame';
      tsIPAllowsFragmentation       : result := 'IP allows Fragmentation';
      tsIPLowTTL                    : result := 'Low IP TTL';
      tsIPStrangeSource             : result := 'Strange IP Source';
      tsRoutingDuplicate            : result := 'Routing Duplicate';
      tsFrameOutOfOrder             : result := 'Frames Out of Order';
      tsHandshakeOutOfOrder         : result := 'TCP Handshake Out of Order';
      tsSYNRetransmit               : result := 'SYN retranmitted';
      tsSYNACKRetransmit            : result := 'SYN/ACK retransmitted';
      tsSYNSYNACKSameSource         : result := 'SYN and SYN/ACK from same source IP';
      tsSYNACKNotMatchingSYN        : result := 'SYN/ACK without matching SYN';
      tsHalfOpenReset               : result := 'Session HalfOpen and RST';
      tsHalfOpenFIN                 : result := 'Session HalfOpen and FIN';
      tsFINSeqNoWrong               : result := 'FIN Sequence Wrong';
      tsFINwithoutACKFlag           : result := 'FIN without ACK flag set';
      tsClientAbort                 : result := 'Client aborted connection';
      tsServerAbort                 : result := 'Server aborted connection';
      tsSocketRefused               : result := 'Socket refused';
      tsClosedInactive              : result := 'Socket closed for Inactivity';
      tsClientReceivedNoData        : result := 'Client received no data';
      tsClientSocketTest            : result := 'Socket tested by client';
      tsSYNFlagWithPayload          : result := 'SYN flag with payload';
      tsSpuriousPacket              : result := 'Spurious packet(s)';
      tcpSYNwithSequenceChg         : result := 'Multiple SYN with different ISNs';
      tsNoClientMSSOption           : result := 'Client sent no MSS option';
      tsNoServerMSSOption           : result := 'Server sent no MSS option';
      tsNoServerScaleOption         : result := 'Server sent no window scale option';
      tsNoClientScaleOption         : result := 'Client sent no window scale option';
      tsSYNwithNonZeroACKNo         : result := 'SYN with non-zero ACK number';
      tsZeroClientWindow            : result := 'Zero client Window';
      tsZeroServerWindow            : result := 'Zero server Window';
      tsKeepAlive                   : result := 'TCP Keep Alive';
      tsInvalidFlagSet              : result := 'Invalid TCP flag set';
      tsUnseenPacket                : result := 'Some packets not captured';
      tsSocketsReused               : result := 'Socket reused';
      tsPortReuseConfusion          : result := 'Port reuse confusion';
      tsForgedReset                 : result := 'Forged RST packet';
      tsRSTOutOfWindow              : result := 'RST out of TCP window';
      tsRSTAfterTeardown            : result := 'RST after session teardown';
      tsPacketsAfterReset           : result := 'Packets after RST';
      tsClientFINRetransmit         : result := 'Client FIN retransmitted';
      tsServerFINRetransmit         : result := 'Server FIN retransmitted';
      tsOneSidedClientFIN           : result := 'One-sided client FIN';
      tsUrgentPointerWithoutFlag    : result := 'Urgent pointer with not URG flag';
      tsOneSidedServerFIN           : result := 'One-sided server FIN';
      tsServerPacketsAfterRST       : result := 'Server packets after RST';
      tsClientPacketsAfterRST       : result := 'Client packets after RST';
      tsMultiClientTTL              : result := 'Multiple TTLs seen from client';
      tsMultiServerTTL              : result := 'Multiple TTLs seen from server';
      tsClientIPIDIncrements        : result := 'Client IPID increments';
      tsServerIPIDIncrements        : result := 'Server IPID increments';
      tsEstablishAfterRefuse        : result := 'Session established after initial refuse';
      tsPrematureNewSession         : result := 'Premature new session';
      tsIdenticalReverseSession     : result := 'Identical reverse session';
      tsDuplicateInterface          : result := 'Duplicate interface';
      tsInvalidConnectionState      : result := 'Invalid connection state';
      tsZeroWindowOnSYNACK          : result := 'Zero Window on SYN/ACK';
      tsThirdAckWithPayload         : result := 'Third ACK with payload';
      tstcpSequenceWrapped          : result := 'TCP sequence number wrapped';
      tsSACKEdgesSeen               : result := 'SACK edges seen';
      tsClientMSSTooLow             : result := 'Client MSS too low';
      tsServerMSSTooLow             : result := 'Server MSS too low';
      tsClientUsesMinMSS            : result := 'Client uses minimum MSS';
      tsServerUsesMinMSS            : result := 'Server uses minimum MSS';
      tsMultiTCPClientSYNOptionSets : result := 'Client has multiple SYN option sets';
      tsMultiTCPServerSYNOptionSets : result := 'Client has multiple SYN option sets';
      tsZeroISN                     : result := 'Zero ISN';
      tsCumbersomeHandshake         : result := 'Handshake had some trouble';
      tsSegmentExceedsMSS           : result := 'TCP segment was larger than MSS';
      tsTrueCRCErrors               : result := 'TCP packet had a truly broken CRC';
      tsSYNACKWithNewISN            : result := 'SYN/ACK was seen with a new ISN';

     else
         result := 'Unknown Symptom';
     end;
end;

(* ------------------------------------------- *)

function SymptomSetToStr (const SymptomSet : tTCPSymptomSet; const Separator : string = '|') : string;

var symptom : tTCPSymptom;
    meta    : tTCPSymptomMeta;
    i       : integer;

begin
     result := '';
     i := 0;

     for symptom in SymptomSet do
         begin
              inc (i);
{
              if i = 1 then
                 result := SymptomToStr (symptom)
              else
                  result := result + Separator + SymptomToStr (symptom);
}
              meta := getTCPSymptomMeta (symptom);

              if i = 1 then
                 result := meta.Name
              else
                  result := result + Separator + meta.Name;
         end;
end;

(* ------------------------------------------- *)
{ tSocketSymptomMeta }
(* ------------------------------------------- *)

procedure tSocketSymptomMeta.initialize (const SocketSymptom: tSocketSymptom);

begin
     Symptom := SocketSymptom;
     Description := getSocketSymptomDescription (Symptom);

     setRemainingSocketSymptomMeta (Symptom);
end;

(* ------------------------------------------- *)

procedure tSocketSymptomMeta.setRemainingSocketSymptomMeta (const SocketSymptom: tSocketSymptom);

begin
     case SocketSymptom of
      ssNone                        : begin
                                           Name := 'ssNone';
                                           Group := tsgConversation;
                                           Criticality := tscNone;
                                           Score := 0;
                                      end;
      ssLegitConnectionsRefused     : begin
                                           Name := 'ssLegitConnectionsRefused';
                                           Group := tsgConversation;
                                           Criticality := tscCritical;
                                           Score := 50;
                                      end;
     end;
end;

(* ------------------------------------------- *)

end.
