{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit definition_tcp_structures;

interface

uses System.Types,
     base_structures,
     definition_symptoms,
     definition_ip_addresses;

const TCPWindowUnknown = -1;
      TCPNoWindowScalingUsed = -2;

      TCPMaximumSegmentLifetime = 120;      // 2 Minutes in seconds, see RFC 793, Page 80, MSL

      TCPFlagSetPlaceHolder = '';
      TCPHistorySize = 32;

type tTCPFlags = (tcpSYN, tcpACK, tcpFIN, tcpRST, tcpPSH, tcpURG, tcpECN, tcpCWR, tcpNON, tcpRS1, tcpRS2);
     tTCPFlagSet = set of tTCPFlags;

     tTCPMPathSubType = (mtcpNone, mtcpCapable, mtcpDataSequenceSignal, mtcpJoinConnection, mtcpAddAddress, mtcpRemovAddress, mtcpFastClose, mtcpFallback, mtcpSubflowPrio);
     tStateLocation = (slUnknown, slClientLocal, slServerLocal, slSPAN, slSPANClient, slSPANServer, slSPANClientServer, slSPANClientLocation, slSPANServerLocation, slFDXTAP, slFDXTAPClient, slFDXTAPServer, slFDXTAPClientServer, slFDXTAPClientLocation, slFDXTAPServerLocation);
     tTCPHandshakeState = (hsNone, hsSYNSent, hsSYNACKSent, hsSYNTwoWay, hsSYNACKTwoWay, hsSYNSYNACKSameSource, hsHalfOpen, hsSYNReset, hsHalfOpenReset, hsComplete);
     tTCPTeardownState = (tdNone, tdFINSeen, tdFINACked);
     tTCPResetState = (rsNone, rsClient, rsServer, rsBoth);
{
     tTCPSymptom = (tsNone, tsFrameDuplicate, tsIPFlagsInvalid, tsIPLowTTL, tsIPStrangeSource, tsRoutingDuplicate, tsFrameOutOfOrder, tsHandshakeOutOfOrder, tsSYNRetransmit, tsSYNACKRetransmit, tsSYNSYNACKSameSource,
                    tsSYNACKNotMatchingSYN, tsHalfOpenReset, tsHalfOpenFIN, tsFINSeqNoWrong, tsFINwithoutACKFlag, tsClientAbort, tsServerAbort, tsSocketRefused, tsClosedInactive, tsClientReceivedNoData, tsClientSocketTest,
                    tsSYNFlagWithPayload, tsSpuriousPacket, tcpSYNwithSequenceChg, tsNoClientMSSOption, tsNoServerMSSOption, tsNoServerScaleOption, tsNoClientScaleOption, tsSYNwithNonZeroACKNo, tsZeroClientWindow,
                    tsZeroServerWindow, tsKeepAlive, tsInvalidFlagSet, tsUnseenPacket, tsSocketsReused, tsPortReuseConfusion, tsForgedReset, tsRSTOutOfWindow, tsRSTAfterTeardown, tsPacketsAfterReset, tsClientFINRetransmit,
                    tsServerFINRetransmit, tsOneSidedClientFIN, tsUrgentPointerWithoutFlag, tsOneSidedServerFIN, tsServerPacketsAfterRST, tsClientPacketsAfterRST, tsMultiClientTTL, tsMultiServerTTL, tsClientIPIDIncrements,
                    tsServerIPIDIncrements, tsEstablishAfterRefuse, tsPrematureNewSession, tsIdenticalReverseSession, tsDuplicateInterface, tsInvalidConnectionState, tsZeroWindowOnSYNACK, tsThirdAckWithPayload, tstcpSequenceWrapped,
                    tsSACKEdgesSeen, tsClientMSSTooLow, tsServerMSSTooLow, tsHandshakeCompleteDetermined, tsZeroISN, tsPortAndISNReusedAfterReset, tsMultiTCPClientSYNOptionSets, tsMultiTCPServerSYNOptionSets, tsTimestampValueMismatch);

     tTCPSymptomSet = set of tTCPSymptom;
}
     tTCPMPathFlags = (tmpChecksumReq, tmpExtensibility, tmpCryptoUseHMAC, tmpBackup, tmpDataFin, tmpDataSeq8Octets, tmpDNSetcPresent, tmpDataACK8Octets, tmpDataACKpresent);
     tTCPMPathFlagSet = set of tTCPMPathFlags;

     tTCPOption = (tcpNOP, tcpMSS, tcpScalefactor, tcpSACKPermitted, tcpSACK, tcpTimestamp, tcpEOL, tcpCC, tcpCCNEW, tcpCCECHO, tcpMPath, tcpFastOpen, tcpRiverbed, tcpOptionsTruncated);
     tTCPOptionSet = set of tTCPOption;
     tTCPOptionList = array of tTCPOption;         // exact order of options, including multiples (e.g. NOP)

     pTCPSackOption = ^tTCPSackOption;
     tTCPSACKOption = packed record
                             leftEdge  : longword;
                             rightEdge : longword;
                      end;

     tTCPScaleFactorOption = record
                                   ShiftCount : integer;
                                   Multiplier : integer;

                                   procedure initialize; overload;
                                   procedure initialize (const Factor : integer); overload;
                             end;
     tTCPSequence = record
                          WrapCount   : integer;
                          Value       : longword;

                          class operator equal (a, b : tTCPSequence) : boolean;
                          class operator notequal (a, b : tTCPSequence) : boolean;
                          class operator greaterthan (a, b : tTCPSequence) : boolean;
                          class operator lessthan (a, b : tTCPSequence) : boolean;
                          class operator greaterthanorequal (a, b : tTCPSequence) : boolean;
                          class operator lessthanorequal (a, b : tTCPSequence) : boolean;

                          class operator equal (a : tTCPSequence; b : longword) : boolean;
                          class operator notequal (a : tTCPSequence; b : longword) : boolean;
                          class operator greaterthan (a : tTCPSequence; b : longword) : boolean;
                          class operator lessthan (a : tTCPSequence; b : longword) : boolean;
                          class operator greaterthanorequal (a : tTCPSequence; b : longword) : boolean;
                          class operator lessthanorequal (a : tTCPSequence; b : longword) : boolean;

                          class operator add (a : tTCPSequence; b : longword) : tTCPSequence;
                          class operator subtract (a : tTCPSequence; b : longword) : tTCPSequence;

                          procedure initialize; overload;
                          procedure initialize (const SequenceNumber : longword; const WrappedCount : integer = 0); overload;
                          function DictionaryKey : string;
                          function toInt64 : int64;
                    end;

     tTCPSACKEdges = record
                           leftEdge : tTCPSequence;
                           rightEdge : tTCPSequence;

                           function SequenceIsInside (const Sequence : tTCPSequence) : boolean;
                     end;

     tTCPSACKRanges = array of tTCPSACKEdges;

     tTCPWindow = record
                        ScaleFactor    : tTCPScaleFactorOption;
                        BaseSequence   : tTCPSequence;
                        TCPWindowValue : word;
                        CurrentSize    : integer;

                        procedure initialize; overload;
                        procedure initialize (const Factor : integer); overload;
                        procedure initialize (const TCPWindowScaling : tTCPScaleFactorOption; const Sequence : tTCPSequence; const TCPWindow: word); overload;

                        procedure UpdateWindow (const Sequence : tTCPSequence; const TCPWindow : word);
                        function FirstSequence : tTCPSequence;
                        function LastSequence : tTCPSequence;
                  end;

     pTCPTimeStampOption = ^tTCPTimeStampOption;
     tTCPTimeStampOption = packed record
                                 tsval    : longword;
                                 tsecho   : longword;
                           end;

     tTCPHistoryEntry = record
                              ConversationFrameIndex : integer;
                              FrameTimestamp         : tFrametimestamp;
                              TCPSequenceNumber      : tTCPSequence;
                              TCPACKNumber           : tTCPSequence;
                              TCPWindow              : tTCPWindow;

                              procedure initialize; overload;
                              procedure initialize (const FrameIndex : integer; const Timestamp : tFrameTimestamp; const Sequence, ACK : tTCPSequence; const Window : tTCPWindow); overload;
                        end;

     tTCPHistoryList  = array [1..TCPHistorySize] of tTCPHistoryEntry;

     tTCPSequenceHistory = record
                                 HistoryList     : tTCPHistoryList;
                                 FirstEntryIndex : integer;
                                 LastEntryIndex  : integer;

                                 procedure initialize;
                                 procedure Add (const Entry : tTCPHistoryEntry); overload;
                                 procedure Add (const FrameIndex : integer; const Timestamp : tFrameTimestamp; const Sequence, ACK : tTCPSequence; const Window : tTCPWindow); overload;
                           end;

     tTCPBasicConversationState = record
                                        Handshake                : tTCPHandshakeState;     // TCP Handshake state
                                        TeardownClient           : tTCPTeardownState;      // FIN state for the client
                                        TeardownServer           : tTCPTeardownState;      // FIN state for the server
                                        Reset                    : tTCPResetState;         // Reset state
                                        Symptoms                 : tTCPSymptomSet;         // "Expert" symptoms detected
                                        Location                 : tStateLocation;         // Guess of where and how this flow was captured
                                        TotalPackets             : integer;                // All packets seen for this connection

                                        ClientIPAddress          : tIPAddress;             // Used to track which of the two nodes is the client
                                        FirstSourceIPAddress     : tIPAddress;             // The first source IP address we ever saw for this conversation (needed for display purposes)
                                        ClientConfirmed          : boolean;                // Only true if SYN or SYN/ACK flag was seen, otherwise client IP is an assumption based on ports
                                        FirstFrameTimestamp      : tFrameTimeStamp;        // Timestamp of the first frame in the conversation
                                        LastFrameTimestamp       : tFrameTimeStamp;        // Timestamp of the last frame in the conversation
                                        FirstFrameFileLocation   : tFileFrameLocation;     // Location of the first frame in the file
                                        LastFrameFileLocation    : tFileFrameLocation;     // Location of the last frame in the file
                                        CommonMSS                : word;                   // Min MSS of client and server
                                        GapRatio                 : double;                 // A ratio of how often sequence numbers matched the expected sequence number
                                        MostRecentFromClient     : boolean;                // true if the most recent added frame was a client frame, else false
                                        SymptomGroups            : tSymptomGroups;         // contains all groups touched by symptoms
                                        SymptomCriticalities     : tTCPSymptomCriticalities; // contains all criticallities raised by symptoms
                                        SymptomScore             : integer;                  // Total Score of all symptoms

                                        // Handshake Tracking

                                        SynTimestamp             : tFrameTimeStamp;        // Timestamp of the SYN packet, needed for iRTT calculation
                                        SynIPID                  : word;                   // IP ID of the last SYN seen, needed for duplicate/retransmission detection
                                        SynAckTimestamp          : tFrameTimeStamp;        // Timestamp of the SYN/ACK packet, needed for iRTT calculation
                                        SynAckIPID               : word;                   // IP ID of the last SYN/ACK seen, needed for duplicate/retransmission detection
                                        InitialSequenceNo        : tTCPSequence;           // Set when encountering a SYN packet
                                        SYNACKSequenceNo         : tTCPSequence;           // Set when encountering a SYN/ACK packet
                                        HandshakeThird           : tTCPSequence;           // sequence number of the expected third handshake packet
                                        HandshakeThirdTime       : tFrameTimeStamp;        // Timestamp of the third handshake packet
                                        AckTimestamp             : tFrameTimeStamp;        // Timestamp of the ACK handshake packet, needed for iRTT calculation
                                        AckIPID                  : word;                   // IP ID of the ACK handshake packet seen, needed for duplicate/retransmission detection
                                        InitialRTTns             : int64;                  // Calculated as soon as the handshake is complete
                                        SynPacketsSeen           : integer;                // Initial SYN counter (SYN flag set only, not SYN/ACK)
                                        SYNResetCount            : integer;                // Counts how often SYNs are answered with a direct RST
                                        ClientTCPSYNOptions      : tTCPOptionSet;          // Lists the options found in the latest client SYN
                                        ServerTCPSYNOptions      : tTCPOptionSet;          // Lists the options found in the latest server SYN/ACK
                                        CaptureToClientns        : int64;                  // time in MS from the capture location to the client
                                        CaptureToServerns        : int64;                  // time in MS from the capture location to the server

                                        // Teardown Tracking

                                        ClientFINIPID            : word;                   // IP ID of the client FIN packet seen, needed for duplicate/retransmission detection
                                        ServerFINIPID            : word;                   // IP ID of the server FIN packet seen, needed for duplicate/retransmission detection
                                        ClientFINTimestamp       : tFrameTimeStamp;        // Timestamp of the client FIN packet
                                        ServerFINTimestamp       : tFrameTimeStamp;        // Timestamp of the server FIN packet
                                        ClientRSTIPID            : word;                   // IP ID of the client FIN packet seen, needed for duplicate/retransmission detection
                                        ServerRSTIPID            : word;                   // IP ID of the server FIN packet seen, needed for duplicate/retransmission detection
                                        ClientRSTTimestamp       : tFrameTimeStamp;        // Timestamp of the client RST packet
                                        ServerRSTTimestamp       : tFrameTimeStamp;        // Timestamp of the server RST packet
                                        ACKforClientFIN          : tTCPSequence;           // sequence number of the expected FIN ACK client packet
                                        ACKforServerFIN          : tTCPSequence;           // sequence number of the expected FIN ACK server packet
                                        ClientFINCount           : integer;                // how often the client has the FIN flag set
                                        ServerFINCount           : integer;                // how often the server has the FIN flag set
                                        ClientRSTCount           : integer;                // how often the client has the RST flag set
                                        ServerRSTCount           : integer;                // how often the server has the RST flag set
                                        TeardownComplete         : tFrameTimeStamp;        // Timestamp of the last teardown packet of the connection
                                        ClientFINSequenceNo      : tTCPSequence;           // Set when encountering a Client FIN packet
                                        ServerFINSequenceNo      : tTCPSequence;           // Set when encountering a Server FIN packet
                                        ClientRSTSequenceNo      : tTCPSequence;           // Set when encountering a Client RST packet
                                        ServerRSTSequenceNo      : tTCPSequence;           // Set when encountering a Server RST packet

                                        // General Client Tracking

                                        ClientPackets            : integer;                // Client packets seen
                                        ClientDiscardedPackets   : integer;                // Packets from the client discarded as duplicates
                                        LastSeenClientSeq        : tTCPSequence;           // Last sequence number seen from the client (or what we think the client is)
                                        NextExpectedClientSeq    : tTCPSequence;           // Calculated sequence we expect to see next from the client
                                        LastSeenClientAck        : tTCPSequence;           // last acknowledge seen from the client
                                        ClientCRCErrors          : integer;                // CRC errors seen - used to diagnose local capture situations and problems
                                        ClientCRCUnknown         : integer;                // CRC errors unknown - used to diagnose local capture situations
                                        LastClientTimestamp      : tFrameTimeStamp;        // Timestamp of the last frame coming from the client seen in the conversation so far
                                        LastClientIPID           : word;                   // Last IP ID seen from the client
                                        ClientTTL                : byte;                   // TTL seen for the client
                                        ClientWindow             : tTCPWindow;             // current client window
                                        ClientMSS                : word;                   // Client MSS from handshake

                                        ClientPayloadSegments    : integer;                // Number of segments that had payload bytes coming from the client
                                        ClientPayloadBytes       : int64;                  // Counts the accumulated TCP payload bytes, e.g. to be able to see if a connection is inactive

                                        ClientSequenceMatch      : integer;                // Counter how often the expected sequence is correctly received from the client
                                        ClientSequenceGap        : integer;                // Counter how often the expected sequence has a gap coming from the client
                                        ClientIPIDIncrements     : integer;                // how often the IP ID of the client is higher than the previous
                                        ClientIPIDZeroCount      : integer;                // how often the IP ID of the client is zero, used to diff against increments to find random ID usage
                                        ClientZeroWindowCount    : integer;                // how often the client window was down to zero
                                        ClientTTLIdentical       : integer;                // how often the Client TTL was a match to the TTL already observed
                                        ClientPushCount          : integer;                // how often the client has the PSH flag set
                                        LastClientInterfaceID    : integer;                // The ID of the last interface seen for a client packet (only relevant for PCAPng probably)
                                        LastClientTimestampValue : tTCPTimeStampOption;    // Tracks the most recent timestamp option set by the client

 //                                     ClientSequenceHistory    : tTCPSequenceHistory;    // array containing previous client sequences

                                        // General Server Tracking

                                        ServerPackets            : integer;                // Server packets seen
                                        ServerDiscardedPackets   : integer;                // Packets from the server discarded as duplicates
                                        ServerCRCErrors          : integer;                // CRC errors seen - used to diagnose local capture situations
                                        LastSeenServerSeq        : tTCPSequence;           // The last sequence we saw coming from the server
                                        NextExpectedServerSeq    : tTCPSequence;           // The next sequence we should see coming from the server
                                        LastSeenServerAck        : tTCPSequence;           // last acknowledge seen from the server
                                        ServerCRCUnknown         : integer;                // CRC errors unknown - used to diagnose local capture situations
                                        LastServerTimestamp      : tFrameTimeStamp;        // Timestamp of the last frame coming from the server seen in the conversation so far
                                        LastServerIPID           : word;                   // Last IP ID seen from the server
                                        ServerTTL                : byte;                   // TTL seen for the server
                                        ServerWindow             : tTCPWindow;             // current server window
                                        ServerMSS                : word;                   // Server MSS from handshake

                                        ServerPayloadSegments    : integer;                // Number of segments that had payload bytes coming from the server
                                        ServerPayloadBytes       : int64;                  // Counts the accumulated TCP payload bytes, e.g. to be able to see if a connection is inactive

                                        ServerSequenceMatch      : integer;                // Counter how often the expected sequence is correctly received from the server
                                        ServerSequenceGap        : integer;                // Counter how often the expected sequence has a gap coming from the server
                                        ServerIPIDIncrements     : integer;                // how often the IP ID of the server is higher than the previous
                                        ServerIPIDZeroCount      : integer;                // how often the IP ID of the server is zero, used to diff against increments to find random ID usage
                                        ServerZeroWindowCount    : integer;                // how often the server window was down to zero
                                        ServerTTLIdentical       : integer;                // how often the server TTL was a match to the TTL already observed
                                        ServerPushCount          : integer;                // how often the server has the PSH flag set
                                        LastServerInterfaceID    : integer;                // The ID of the last interface seen for a server packet (only relevant for PCAPng probably)
                                        LastServerTimestampValue : tTCPTimeStampOption;    // Tracks the most recent timestamp option set by the server

                                        // Internal stuff

                                        WasFinalized             : boolean;                // True if final calculations were performed (should happen when no more packets are added)

//                                      ServerSequenceHistory    : tTCPSequenceHistory;    // array containing previous client sequences

                                        procedure initialize;
                                        procedure SwapClientServer (const TrueClientIPAddress : tIPAddress);
                                        function toCSV (const Separator : string = ',') : string;
                                  end;

     tTCPFlowState = record
                           InitialSeq        : tTCPSequence;
                           NextExpectedSeq   : tTCPSequence;
                           LowestSeq         : tTCPSequence;
                           HighestSeq        : tTCPSequence;
                           LastSeq           : tTCPSequence;
                           MaxSegmentSize    : word;
                           LastWindow        : longword;
                           WindowMultiplier  : integer;
                           LastScaledWindow  : longword;
                           TotalSegmentBytes : int64;
                           PacketCount       : longword;
                           AcknowledgedTo    : longword;
                           SackRanges        : tTCPSackRanges;
                           BytesInFlight     : integer;
                           Timestamp         : tFrametimeStamp;
                           Location          : tStateLocation;

                           procedure initialize; overload;
                           procedure initialize (const Sequence : tTCPSequence; const SegmentLen : integer; const WinSize : longword;Flags : tTCPFlagSet; const PartnerACK : longword = 0); overload;
                           procedure initialize (const Sequence : longword; const SegmentLen : integer; const WinSize : longword;Flags : tTCPFlagSet; const PartnerACK : longword = 0); overload;

//                         procedure AddSackRange (const NewRange : tTCPSackOption);
//                         procedure AddSackRanges (const NewRanges : tTCPSACKRanges);

//                         procedure AddState (const NewState : tTCPFlowState);
                     end;

     tTCPConversationState = record
                                   SourceState      : tTCPFlowState;
                                   DestinationState : tTCPFlowState;


                                   procedure initialize;
                             end;

{
     tTCPSessionState = record
                              IPVersion             : byte;
                              ClientIPv4Address     : tIPv4Address;
                              ClientIPv6Address     : tIPv6Address;
                              ClientSynSeen         : boolean;
                              ClientRstSeen         : boolean;
                              ClientFinSeen         : boolean;
                              ClientFlowState       : tTCPFlowState;

                              ServerIPv4Address     : tIPv4Address;
                              ServerIPv6Address     : tIPv6Address;
                              ServerSynSeen         : boolean;
                              ServerRstSeen         : boolean;
                              ServerFinSeen         : boolean;
                              ServerFlowState       : tTCPFlowState;

                              FirstPacketTimestamp  : tFrameTimestamp;
                              LastPacketTimestamp   : tFrameTimestamp;

                              procedure initialize;
                        end;
}

     function getTCPFlagSetFromWord (Value : word) : tTCPFlagSet;
     function getWordFromTCPFlagSet (FlagSet : tTCPFlagSet) : word;
     function TCPFlagToString (Flag : tTCPFlags) : string;
     function StringToTCPFlag (Flag : string) : tTCPFlags;
     function TCPFlagToFullString (Flag : tTCPFlags) : string;
     function TCPDumpToTCPFlagSet (value : string) : tTCPFlagSet;
     function TCPFlagSetToTCPDump (FlagSet : tTCPFlagSet; const placeHolder : string = TCPFlagSetPlaceHolder) : string;
     function CompareFlags (const a, b : tTCPFlagSet) : tValueRelationship;

     function CalculateNextExpextedTCPSequence (const Sequence : tTCPSequence; const SegmentLen : longword; const TCPFlags : tTCPFlagSet) : tTCPSequence;
     function SequenceInWindow (const Sequence : tTCPSequence; const Window : tTCPWindow) : boolean;

     function getStateCSVHeader (const Separator: string)  : string;

implementation

uses SysUtils,
     Math;

(* ------------------------------------------- *)

function getTCPFlagSetFromWord (Value : word) : tTCPFlagSet;

var flagset : tTCPFlagSet;

begin
     flagset := [];

     if (Value and $0001) = $0001 then
        flagset := [tcpFIN];

     if (Value and $0002) = $0002 then
        flagset := flagset + [tcpSYN];

     if (Value and $0004) = $0004 then
        flagset := flagset + [tcpRST];

     if (Value and $0008) = $0008 then
        flagset := flagset + [tcpPSH];

     if (Value and $0010) = $0010 then
        flagset := flagset + [tcpACK];

     if (Value and $0020) = $0020 then
        flagset := flagset + [tcpURG];

     if (Value and $0040) = $0040 then
        flagset := flagset + [tcpECN];

     if (Value and $0080) = $0080 then
        flagset := flagset + [tcpCWR];

     if (Value and $0100) = $0100 then
        flagset := flagset + [tcpNON];

     if (Value and $0200) = $0200 then
        flagset := flagset + [tcpRS1];

     if (Value and $0400) = $0400 then
        flagset := flagset + [tcpRS2];

     result := flagset;
end;

(* ------------------------------------------- *)

function getWordFromTCPFlagSet (FlagSet : tTCPFlagSet) : word;

var value : word;
    help  : word;

begin
     // Set bits according to flags

     if tcpFIN in FlagSet then
        value := $0001
     else
         value := 0;

     if tcpSYN in FlagSet then
        value := value or $0002;

     if tcpRST in FlagSet then
        value := value or $0004;

     if tcpPSH in FlagSet then
        value := value or $0008;

     if tcpACK in FlagSet then
        value := value or $0010;

     if tcpURG in FlagSet then
        value := value or $0020;

     if tcpECN in FlagSet then
        value := value or $0040;

     if tcpCWR in FlagSet then
        value := value or $0080;

     if tcpNON in FlagSet then
        value := value or $0100;

     if tcpRS1 in FlagSet then
        value := value or $0200;

     if tcpRS2 in FlagSet then
        value := value or $0400;

     result := value;
end;

(* ------------------------------------------- *)

function TCPFlagToString (Flag : tTCPFlags) : string;

begin
     case flag of
      tcpSYN: result := 'SYN';
      tcpACK: result := 'ACK';
      tcpFIN: result := 'FIN';
      tcpRST: result := 'RST';
      tcpPSH: result := 'PSH';
      tcpURG: result := 'URG';
      tcpECN: result := 'ENC';
      tcpCWR: result := 'CWR';
      tcpNON: result := 'NON';
      tcpRS1: result := 'RS1';
      tcpRS2: result := 'RS2';
     end;
end;

(* ------------------------------------------- *)

function TCPFlagToFullString (Flag : tTCPFlags) : string;

begin
     case flag of
      tcpSYN: result := 'Syncronize';
      tcpACK: result := 'Acknowledge';
      tcpFIN: result := 'Finish';
      tcpRST: result := 'Reset';
      tcpPSH: result := 'Push';
      tcpURG: result := 'Urgent';
      tcpECN: result := 'Explicit Congestion Notification';
      tcpCWR: result := 'Congestion Window Reduced';
      tcpNON: result := 'Nonce';
      tcpRS1: result := 'Reserved1';
      tcpRS2: result := 'Reserved2';
     end;
end;

(* ------------------------------------------- *)

function StringToTCPFlag (Flag : string) : tTCPFlags;

begin
     if Flag = 'SYN' then
        result := tcpSYN
     else
     if Flag = 'ACK' then
        result := tcpACK
     else
     if Flag = 'FIN' then
        result := tcpFIN
     else
     if Flag = 'RST' then
        result := tcpRST
     else
     if Flag = 'PSH' then
        result := tcpPSH
     else
     if Flag = 'URG' then
        result := tcpURG
     else
     if Flag = 'ECN' then
        result := tcpECN
     else
     if Flag = 'CWR' then
        result := tcpCWR
     else
     if Flag = 'NON' then
        result := tcpNON
     else
     if Flag = 'RS1' then
        result := tcpRS1
     else
     if Flag = 'RS2' then
        result := tcpRS2
end;

(* ------------------------------------------- *)

function CalculateNextExpextedTCPSequence (const Sequence : tTCPSequence; const SegmentLen : longword; const TCPFlags : tTCPFlagSet) : tTCPSequence; // inline;

var size : longword;

begin
     // SYN and FIN flag count as one phantom byte

     if (tcpSYN in TCPFlags) or (tcpFIN in TCPFlags) then
        size := SegmentLen + 1
     else
         size := SegmentLen;

     result := Sequence + size;
end;

(* ------------------------------------------- *)

function SequenceInWindow (const Sequence : tTCPSequence; const Window : tTCPWindow) : boolean;

var First : tTCPSequence;
    Last  : tTCPSequence;

begin
     First := Window.FirstSequence;
     Last := Window.LastSequence;

     result := (Sequence >= First) and
               (Sequence <= Last);
end;

(* ------------------------------------------- *)

function TCPDumpToTCPFlagSet (value : string) : tTCPFlagSet;

var flagset : tTCPFlagSet;

begin
     flagset := [];

     if pos ('N', value) > 0 then
        flagset := flagset + [tcpNON];

     if pos ('C', value) > 0 then
        flagset := flagset + [tcpCWR];

     if pos ('E', value) > 0 then
        flagset := flagset + [tcpECN];

     if pos ('U', value) > 0 then
        flagset := flagset + [tcpURG];

     if pos ('A', value) > 0 then
        flagset := flagset + [tcpACK];

     if pos ('P', value) > 0 then
        flagset := flagset + [tcpPSH];

     if pos ('R', value) > 0 then
        flagset := flagset + [tcpRST];

     if pos ('S', value) > 0 then
        flagset := flagset + [tcpSYN];

     if pos ('F', value) > 0 then
        flagset := flagset + [tcpFIN];

     result := flagset;
end;

(* ------------------------------------------- *)

function TCPFlagSetToTCPDump (FlagSet : tTCPFlagSet; const placeHolder : string = TCPFlagSetPlaceHolder) : string;

var value : string;

begin
     if tcpNON in FlagSet then
        value := 'N'
     else
         value := placeholder;

     if tcpCWR in FlagSet then
        value := 'C'
     else
         value := value + placeholder;

     if tcpECN in FlagSet then
        value := value + 'E'
     else
         value := value + placeholder;

     if tcpURG in FlagSet then
        value := value + 'U'
     else
         value := value + placeholder;

     if tcpACK in FlagSet then
        value := value + 'A'
     else
         value := value + placeholder;

     if tcpPSH in FlagSet then
        value := value + 'P'
     else
         value := value + placeholder;

     if tcpRST in FlagSet then
        value := value + 'R'
     else
         value := value + placeholder;

     if tcpSYN in FlagSet then
        value := value + 'S'
     else
         value := value + placeholder;

     if tcpFIN in FlagSet then
        value := value + 'F'
     else
         value := value + placeholder;

     result := value;
end;

(* ------------------------------------------- *)

function CompareFlags (const a, b : tTCPFlagSet) : tValueRelationship;

begin
     if A = B then
       Result := EqualsValue
     else
         begin
              // Now lets check the flag set

              if a <= b then
                 Result := LessThanValue
              else
                  Result := GreaterThanValue;
         end;
end;

(* ------------------------------------------- *)

function HandshakeStateToStr (State : tTCPHandshakeState) : string;

begin
     case State of
      hsNone                : result := 'None';
      hsSYNSent             : result := 'SYNSent';
      hsSYNACKSent          : result := 'SYNACKSent';
      hsSYNTwoWay           : result := 'SYNTwoWay';
      hsSYNACKTwoWay        : result := 'SYNACKTwoWay';
      hsSYNSYNACKSameSource : result := 'SYNSYNACKSameSource';
      hsHalfOpen            : result := 'HalfOpen';
      hsSYNReset            : result := 'SYNReset';
      hsHalfOpenReset       : result := 'HalfOpenReset';
      hsComplete            : result := 'Complete';
     end;
end;

(* ------------------------------------------- *)

function TeardownStateToStr (State : tTCPTeardownState) : string;

begin
     case State of
      tdNone     : result := 'None';
      tdFINSeen  : result := 'FINSeen';
      tdFINACked : result := 'FINACKed';
     end;
end;

(* ------------------------------------------- *)

function ResetStateToStr (State : tTCPResetState) : string;

begin
     case state of
      rsNone   : result := 'None';
      rsClient : result := 'Client';
      rsServer : result := 'Server';
      rsBoth   : result := 'Both';
     end;
end;

(* ------------------------------------------- *)

function StateLocationToStr (State : tStateLocation) : string;

begin
     case State of
      slUnknown              : result := 'Unknown';
      slClientLocal          : result := 'ClientLocal';
      slServerLocal          : result := 'ServerLocal';
      slSPAN                 : result := 'SPAN';
      slSPANClient           : result := 'SPANClient';
      slSPANServer           : result := 'SPANServer';
      slSPANClientServer     : result := 'SPANClientServer';
      slSPANClientLocation   : result := 'SPANClientLocation';
      slSPANServerLocation   : result := 'SPANServerLocation';
      slFDXTAP               : result := 'FDXTap';
      slFDXTAPClient         : result := 'FDXTAPClient';
      slFDXTAPServer         : result := 'FDXTAPServer';
      slFDXTAPClientServer   : result := 'FDXTAPClientServer';
      slFDXTAPClientLocation : result := 'FDXTAPClientLocation';
      slFDXTAPServerLocation : result := 'FDXTAPServerLocation';
     end;
end;
(* ------------------------------------------- *)
{ tTCPScaleFactorOption }
(* ------------------------------------------- *)

procedure tTCPScaleFactorOption.initialize;

begin
     ShiftCount := TCPWindowUnknown;
     Multiplier := 1;
end;

(* ------------------------------------------- *)

procedure tTCPScaleFactorOption.initialize (const Factor: integer);

begin
     ShiftCount := Factor;
     Multiplier := trunc (Power (2, Factor));
end;

(* ------------------------------------------- *)
{ tTCPSequence }
(* ------------------------------------------- *)

class operator tTCPSequence.equal (a, b: tTCPSequence): boolean;

begin
     result := (a.WrapCount = b.WrapCount) and (a.Value = b.Value);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.notequal (a, b: tTCPSequence): boolean;

begin
     result := not (a=b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.greaterthan (a, b: tTCPSequence): boolean;

begin
     result := (a.WrapCount > b.WrapCount) or ((a.WrapCount = b.WrapCount) and (a.Value > b.Value));
end;

(* ------------------------------------------- *)

class operator tTCPSequence.greaterthanorequal (a, b: tTCPSequence): boolean;

begin
     result := (a=b) or (a>b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.lessthan (a, b: tTCPSequence): boolean;

begin
     result := not (a >= b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.lessthanorequal (a, b: tTCPSequence): boolean;

begin
     result := not (a > b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.equal (a: tTCPSequence; b: longword): boolean;

begin
     result := (a.Value = b) and (a.WrapCount = 0);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.notequal (a: tTCPSequence; b: longword): boolean;

begin
     result := not (a=b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.greaterthan (a: tTCPSequence; b: longword): boolean;

begin
     result := (a.WrapCount > 0) or (a.Value > b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.greaterthanorequal (a: tTCPSequence; b: longword): boolean;

begin
     result := (a = b) or (a > b)
end;

(* ------------------------------------------- *)

class operator tTCPSequence.lessthan (a: tTCPSequence; b: longword): boolean;

begin
     result := not (a >= b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.lessthanorequal (a: tTCPSequence; b: longword): boolean;

begin
     result := not (a > b);
end;

(* ------------------------------------------- *)

class operator tTCPSequence.add (a: tTCPSequence; b: longword): tTCPSequence;

var help : uint64;

begin
     // Check if the sequence would wrap when adding a Segment of 65535 bytes
     // if not, we can just add the two values and return the result

     help := a.Value;
     help := help + b;

     if help > $ffffffff then
        begin
             // we got a wrapped sequence here

              help := help mod $ffffffff;
              result.Value := help - 1;   // for the 0 we start at, not 1

              result.WrapCount := a.WrapCount + 1;
        end
     else
         begin
              result.Value := help;
              result.WrapCount := a.WrapCount;
         end;
end;

(* ------------------------------------------- *)

class operator tTCPSequence.subtract (a: tTCPSequence; b: longword): tTCPSequence;

var help : int64;

begin
     help := a.Value;
     help := help - b;

     if help < 0 then
        begin
             result.Value := $ffffffff + (help + 1);    // for the 0 we start at, not 1
             result.WrapCount := Min (0, a.WrapCount - 1);
        end
     else
         begin
              result.Value := help;
              result.WrapCount := a.WrapCount;
         end;
end;

(* ------------------------------------------- *)

procedure tTCPSequence.initialize;

begin
     WrapCount := 0;
     Value := 0;
end;

(* ------------------------------------------- *)

procedure tTCPSequence.initialize (const SequenceNumber: longword; const WrappedCount : integer = 0);

begin
     WrapCount := WrappedCount;
     Value := SequenceNumber;
end;

(* ------------------------------------------- *)

function tTCPSequence.DictionaryKey: string;

begin
     result := IntToStr (WrapCount) + ':' + IntToStr (Value);
end;

(* ------------------------------------------- *)

function tTCPSequence.toInt64: int64;

begin
     result := (WrapCount * cardinal.MaxValue) + Value;
end;

(* ------------------------------------------- *)
{ tTCPSACKEdges }
(* ------------------------------------------- *)

function tTCPSACKEdges.SequenceIsInside (const Sequence: tTCPSequence): boolean;

begin
     result := (Sequence >= leftEdge) and (Sequence <= rightEdge);
end;

(* ------------------------------------------- *)
{ tTCPWindow }
(* ------------------------------------------- *)

procedure tTCPWindow.initialize;

begin
     ScaleFactor.initialize;
     BaseSequence.initialize;
     TCPWindowValue := 0;
     CurrentSize := 0
end;

(* ------------------------------------------- *)

procedure tTCPWindow.initialize (const Factor : integer);

begin
     initialize;

     assert ((Factor >= 0) and (Factor < 15));

     ScaleFactor.initialize (Factor);
end;

(* ------------------------------------------- *)

procedure tTCPWindow.initialize (const TCPWindowScaling: tTCPScaleFactorOption; const Sequence : tTCPSequence; const TCPWindow: word);

begin
     ScaleFactor := TCPWindowScaling;

     UpdateWindow (Sequence, TCPWindow);
end;

(* ------------------------------------------- *)

procedure tTCPWindow.UpdateWindow (const Sequence : tTCPSequence; const TCPWindow: word);

begin
     TCPWindowValue := TCPWindow;
     BaseSequence := Sequence;

     CurrentSize := ScaleFactor.Multiplier * TCPWindow;
end;

(* ------------------------------------------- *)

function tTCPWindow.FirstSequence: tTCPSequence;

begin
     result := BaseSequence;
end;

(* ------------------------------------------- *)

function tTCPWindow.LastSequence: tTCPSequence;

begin
     result := BaseSequence + CurrentSize;     // TODO: test this, may be off by one
end;

(* ------------------------------------------- *)
{ tTCPHistoryEntry }
(* ------------------------------------------- *)

procedure tTCPHistoryEntry.initialize;

begin
     ConversationFrameIndex := 0;
     Frametimestamp.initialize;
     TCPSequenceNumber.initialize;
     TCPACKNumber.initialize;
     TCPWindow.initialize;
end;

(* ------------------------------------------- *)

procedure tTCPHistoryEntry.initialize(const FrameIndex: integer; const Timestamp : tFrameTimestamp;const Sequence, ACK: tTCPSequence; const Window: tTCPWindow);

begin
     ConversationFrameIndex := FrameIndex;
     Frametimestamp := Timestamp;
     TCPSequenceNumber := Sequence;
     TCPACKNumber := ACK;
     TCPWindow := Window;
end;

(* ------------------------------------------- *)
{ tTCPSequenceHistory }
(* ------------------------------------------- *)

procedure tTCPSequenceHistory.initialize;

begin
     FirstEntryIndex := 0;
     LastEntryIndex := 0;
end;

(* ------------------------------------------- *)

procedure tTCPSequenceHistory.Add (const Entry: tTCPHistoryEntry);

begin
     if FirstEntryIndex = 0 then
        begin
             // First entry ever

             FirstEntryIndex := 1;
             LastEntryIndex := 1;
        end
     else
         begin
              inc (LastEntryIndex);

              // Check if the ring is full, and advance indexes if required

              if (LastEntryIndex > TCPHistorySize) then
                 LastEntryIndex := 1;

              // Check if we caught up to the first entry, and move it if true

              if (LastEntryIndex = FirstEntryIndex) then
                 begin
                      inc (FirstEntryIndex);

                      // Wrap the first entry index, if required

                      if FirstEntryIndex > TCPHistorySize then
                         FirstEntryIndex := 1;
                 end;
         end;

     HistoryList[LastEntryIndex] := Entry;
end;

(* ------------------------------------------- *)

procedure tTCPSequenceHistory.Add(const FrameIndex: integer; const Timestamp: tFrameTimestamp; const Sequence, ACK: tTCPSequence; const Window: tTCPWindow);

begin

end;

(* ------------------------------------------- *)
{ tTCPBasicConversationState }
(* ------------------------------------------- *)

procedure tTCPBasicConversationState.initialize;

begin
     Handshake := hsNone;
     TeardownClient := tdNone;
     TeardownServer := tdNone;
     Reset := rsNone;
     Symptoms := [];
     Location := slUnknown;
     CommonMSS := 0;
     SymptomGroups := [];
     SymptomCriticalities := [];
     SymptomScore := 0;

     FirstFrameTimestamp.initialize;
     LastFrameTimestamp.initialize;
     FirstFrameFileLocation.initialize;
     LastFrameFileLocation.initialize;

     ClientIPAddress.initialize;
     ClientConfirmed := False;
     ClientPayloadSegments := 0;
     ServerPayloadSegments := 0;
     ClientPayloadBytes := 0;
     ServerPayloadBytes := 0;
     ClientTCPSYNOptions := [];
     ServerTCPSYNOptions := [];
     InitialRTTns := 0;
     InitialSequenceNo.initialize;
     SYNACKSequenceNo.initialize;
     ClientFINSequenceNo.initialize;
     ServerFINSequenceNo.initialize;
     ClientRSTSequenceNo.initialize;
     ServerRSTSequenceNo.initialize;
     LastSeenClientSeq.initialize;
     LastSeenClientAck.initialize;
     LastSeenServerSeq.initialize;
     LastSeenServerAck.initialize;
     NextExpectedClientSeq.initialize;
     NextExpectedServerSeq.initialize;
     LastClientTimestamp.initialize;
     LastServerTimestamp.initialize;
     SynTimestamp.initialize;
     CaptureToClientns := 0;
     CaptureToServerns := 0;
     SynIPID := 0;
     ClientFINIPID := 0;
     ServerFINIPID := 0;
     ClientRSTIPID := 0;
     ServerRSTIPID := 0;
     ClientTTL := 0;
     ClientMSS := 0;
     SynAckTimestamp.initialize;
     SynAckIPID := 0;
     AckTimestamp.initialize;
     AckIPID := 0;
     ServerTTL := 0;
     ServerMSS := 0;
     LastClientIPID := 0;
     LastServerIPID := 0;
     ClientFINTimestamp.initialize;
     ServerFINTimestamp.initialize;
     ClientRSTTimestamp.initialize;
     ServerRSTTimestamp.initialize;
     TeardownComplete.initialize;
     HandshakeThird.initialize;
     HandshakeThirdTime.initialize;
     ACKforClientFIN.initialize;
     ACKforServerFIN.initialize;
     SynPacketsSeen := 0;
//   FinPacketsSeen := 0;
     TotalPackets := 0;
     ClientPackets := 0;
     ServerPackets := 0;
     ClientDiscardedPackets := 0;
     ServerDiscardedPackets := 0;
     ClientCRCErrors := 0;
     ServerCRCErrors := 0;
     ClientCRCUnknown := 0;
     ServerCRCUnknown := 0;
     ClientSequenceMatch := 0;
     ServerSequenceMatch := 0;
     ClientSequenceGap := 0;
     ServerSequenceGap := 0;
     ClientWindow.initialize;
     ServerWindow.initialize;
     ClientIPIDIncrements := 0;
     ServerIPIDIncrements := 0;
     ClientIPIDZeroCount := 0;
     ServerIPIDZeroCount := 0;
     ClientZeroWindowCount := 0;
     ServerZeroWindowCount := 0;
     ClientPushCount := 0;
     ServerPushCount := 0;
     ClientFINCount := 0;
     ServerFINCount := 0;
     ClientRSTCount := 0;
     ServerRSTCount := 0;
     ClientTTLIdentical := 0;
     ServerTTLIdentical := 0;

     SYNResetCount := 0;

     LastClientInterfaceID := -1;   // can't be 0 because that would be the first interface
     LastServerInterfaceID := -1;   // can't be 0 because that would be the first interface

     GapRatio := 100.0;

//   ClientSequenceHistory.initialize;
//   ServerSequenceHistory.initialize;

     MostRecentFromClient := true;
     WasFinalized := false;
end;

(* ------------------------------------------- *)
{ tTCPFlowState }
(* ------------------------------------------- *)

procedure tTCPFlowState.initialize;

begin
     InitialSeq.initialize;
     NextExpectedSeq.initialize;
     LowestSeq.initialize;
     HighestSeq.initialize;
     LastSeq.initialize;
     MaxSegmentSize := 0;
     LastWindow := 0;
     WindowMultiplier := TCPWindowUnknown;
     LastScaledWindow := 0;
     TotalSegmentBytes := 0;
     PacketCount := 0;
     AcknowledgedTo := 0;
     BytesInFlight := 0;
     Timestamp.initialize;
     Location := slUnknown;

     setLength (SackRanges, 0);
end;

(* ------------------------------------------- *)

procedure tTCPFlowState.initialize (const Sequence: tTCPSequence; const SegmentLen: integer; const WinSize : longword; Flags: tTCPFlagSet; const PartnerACK : longword = 0);

begin
     // Determine next expected sequence number. This also deals with
     // packets where the SYN carries a payload

     NextExpectedSeq := CalculateNextExpextedTCPSequence (Sequence, SegmentLen, Flags);

     // If we have a SYN flag the initial sequence is obvious
     // Otherwise assume SYN was just the packet before this one

     if tcpSYN in Flags then
        InitialSeq := Sequence
     else
         begin
              if Sequence.Value > 0 then
                 InitialSeq := Sequence - 1
              else
                  InitialSeq.Value := $ffffffff;
         end;

     LowestSeq := Sequence;
     HighestSeq := Sequence;
     LastSeq := Sequence;
     LastWindow := WinSize;
     WindowMultiplier := TCPWindowUnknown;
     MaxSegmentSize := 0;
     TotalSegmentBytes := SegmentLen;
     PacketCount := 1;
     AcknowledgedTo := 0;
     BytesInFlight := SegmentLen;
     Timestamp.initialize;
     Location := slUnknown;
end;

(* ------------------------------------------- *)

procedure tTCPFlowState.initialize (const Sequence: longword; const SegmentLen: integer; const WinSize : longword; Flags: tTCPFlagSet; const PartnerACK : longword = 0);

var help : tTCPSequence;

begin
     help.Value := Sequence;
     help.WrapCount := 0;

     initialize (help, SegmentLen, WinSize, Flags, PartnerACK);
end;

(* ------------------------------------------- *)

procedure tTCPBasicConversationState.SwapClientServer (const TrueClientIPAddress : tIPAddress);

var HelpState : tTCPBasicConversationState;

begin
     HelpState := self;

     case Location of
      slClientLocal : Location := slServerLocal;
      slServerLocal : Location := slClientLocal;
     end;

     TeardownClient := HelpState.TeardownServer;
     TeardownServer := HelpState.TeardownClient;

     ClientIPAddress := TrueClientIPAddress;
     Symptoms := GetSwappedSymptoms (Symptoms);
     ClientConfirmed := True;

     ClientTCPSYNOptions := HelpState.ServerTCPSYNOptions;
     ServerTCPSYNOptions := HelpState.ClientTCPSYNOptions;

     ClientPayloadSegments := HelpState.ServerPayloadSegments;
     ServerPayloadSegments := HelpState.ClientPayloadSegments;

     ClientPayloadBytes := HelpState.ServerPayloadBytes;
     ServerPayloadBytes := HelpState.ClientPayloadBytes;

     ClientFINSequenceNo := HelpState.ServerFINSequenceNo;
     ServerFINSequenceNo := HelpState.ClientFINSequenceNo;

     LastSeenClientSeq := HelpState.LastSeenServerSeq;
     LastSeenServerSeq := HelpState.LastSeenClientSeq;

     LastSeenClientAck := HelpState.LastSeenServerAck;
     LastSeenServerAck := HelpState.LastSeenClientAck;

     NextExpectedClientSeq := HelpState.NextExpectedServerSeq;
     NextExpectedServerSeq := HelpState.NextExpectedClientSeq;

     LastClientTimestamp := HelpState.LastServerTimestamp;
     LastServerTimestamp := HelpState.LastClientTimestamp;

     ClientFINIPID := HelpState.ServerFINIPID;
     ServerFINIPID := HelpState.ClientFINIPID;

     ClientRSTIPID := HelpState.ServerRSTIPID;
     ServerRSTIPID := HelpState.ClientRSTIPID;

     ClientTTL := HelpState.ServerTTL;
     ServerTTL := HelpState.ClientTTL;

     ClientMSS := HelpState.ServerMSS;
     ServerMSS := HelpState.ClientMSS;

     LastClientIPID := HelpState.LastServerIPID;
     LastServerIPID := HelpState.LastClientIPID;

     ClientFINTimestamp := HelpState.ServerFINTimestamp;
     ServerFINTimestamp := HelpState.ClientFINTimestamp;

     ClientRSTTimestamp := HelpState.ServerRSTTimestamp;
     ServerRSTTimestamp := HelpState.ClientRSTTimestamp;

     ACKforClientFIN := HelpState.ACKforServerFIN;
     ACKforServerFIN := HelpState.ACKforClientFIN;

     ClientPackets := HelpState.ServerPackets;
     ServerPackets := HelpState.ClientPackets;

     ClientPackets := HelpState.ServerDiscardedPackets;
     ServerPackets := HelpState.ClientDiscardedPackets;

     ClientCRCErrors := HelpState.ServerCRCErrors;
     ServerCRCErrors := HelpState.ClientCRCErrors;

     ClientSequenceMatch := HelpState.ServerSequenceMatch;
     ServerSequenceMatch := HelpState.ClientSequenceMatch;

     ClientSequenceGap := HelpState.ServerSequenceGap;
     ServerSequenceGap := HelpState.ClientSequenceGap;

     ClientWindow := HelpState.ServerWindow;
     ServerWindow := HelpState.ClientWindow;

     ClientIPIDIncrements := HelpState.ServerIPIDIncrements;
     ServerIPIDIncrements := HelpState.ClientIPIDIncrements;
     ClientIPIDZeroCount := HelpState.ServerIPIDZeroCount;
     ServerIPIDZeroCount := HelpState.ClientIPIDZeroCount;

     ClientZeroWindowCount := ServerZeroWindowCount;
     ServerZeroWindowCount := ClientZeroWindowCount;

     ClientPushCount := HelpState.ServerPushCount;
     ServerPushCount := HelpState.ClientPushCount;
     ClientFINCount := HelpState.ServerFINCount;
     ServerFINCount := HelpState.ClientFINCount;
     ClientRSTCount := HelpState.ServerRSTCount;
     ServerRSTCount := HelpState.ClientRSTCount;

     ClientTTLIdentical := HelpState.ServerTTLIdentical;
     ServerTTLIdentical := HelpState.ClientTTLIdentical;

     LastClientInterfaceID := HelpState.LastServerInterfaceID;
     LastServerInterfaceID := HelpState.LastClientInterfaceID;

     LastClientTimestampValue := HelpState.LastServerTimestampValue;
     LastServerTimestampValue := HelpState.LastClientTimestampValue;

     CaptureToClientns := HelpState.CaptureToServerns;
     CaptureToServerns := HelpState.CaptureToClientns;

     MostRecentFromClient := not MostRecentFromClient;
end;

(* ------------------------------------------- *)

function getStateCSVHeader (const Separator: string) : string;

begin
     result := 'Handshake State' + Separator +
               'Teardown Client' + Separator +
               'Teardown Server' + Separator +
               'Reset State' + Separator +
               'Symptoms' + Separator +
               'State Location' + Separator +
               'Client IP' + Separator +
               'First IP seen' + Separator +
               'Client determined' + Separator +
               'Common MSS' + Separator +
               'GapRatio' + Separator +
               'Total Symptom Score' + Separator +
               'Initial RTT (ns)' + Separator +
               'SYN Packets Seen' + Separator +
               'SYN Reset count' + Separator +
               'Client FIN count' + Separator +
               'Server FIN count' + Separator +
               'Client RST count' + Separator +
               'Server RST count' + Separator +
               'Client packet count' + Separator +
               'Server packet count' + Separator +
               'Client discarded packets' + Separator +
               'Server discarded packets' + Separator +
               'Client CRC errors' + Separator +
               'Server CRC errors' + Separator +
               'Client TTL' + Separator +
               'Server TTL' + Separator +
               'Client MSS' + Separator +
               'Server MSS';
end;

(* ------------------------------------------- *)

function tTCPBasicConversationState.toCSV (const Separator: string) : string;

var line         : string;
    subseparator : string;


begin
     if Separator = '|' then
        subSeparator := ','
     else
         subSeparator := '|';

     line := HandshakeStateToStr (Handshake) + Separator +
             TeardownStateToStr (TeardownClient) + Separator +
             TeardownStateToStr (TeardownServer) + Separator +
             ResetStateToStr (Reset) + Separator +
             SymptomSetToStr (Symptoms, subSeparator) + Separator +
             StateLocationToStr (Location) + Separator +
             ClientIPAddress.ValueAsString + Separator +
             FirstSourceIPAddress.ValueAsString + Separator;

     if ClientConfirmed then
        line := line + 'true' + Separator
     else
         line := line + 'false' + Separator;

     line := line + CommonMSS.ToString + Separator +
                    GapRatio.ToString + Separator +
                    SymptomScore.ToString + Separator +
                    InitialRTTns.ToString + Separator +
                    SYNPacketsSeen.ToString + Separator +
                    SYNResetCount.ToString + Separator +
                    ClientFINCount.ToString + Separator +
                    ServerFINCount.ToString + Separator +
                    ClientRSTCount.ToString + Separator +
                    ServerRSTCount.ToString + Separator +
                    ClientPackets.ToString + Separator +
                    ServerPackets.ToString + Separator +
                    ClientDiscardedPackets.ToString + Separator +
                    ServerDiscardedPackets.ToString + Separator +
                    ClientCRCErrors.ToString + Separator +
                    ServerCRCErrors.ToString + Separator +
                    ClientTTL.ToString + Separator +
                    ServerTTL.ToString + Separator +
                    ClientMSS.ToString + Separator +
                    ServerMSS.ToString;


     result := line;
end;

(* ------------------------------------------- *)
{ tTCPConversationState }
(* ------------------------------------------- *)

procedure tTCPConversationState.initialize;

begin
     SourceState.initialize;
     DestinationState.initialize;
end;

(* ------------------------------------------- *)
{
procedure tTCPFlowState.AddSackRange (const NewRange : tTCPSACKOption);

var range : tTCPSACKOption;
    found : boolean;
    len   : integer;
    i     : integer;

begin
     len := length (SackRanges);

     // If we have no SACK ranges yet, it's easy, otherwise we need to merge ranges

     if len = 0 then
        begin
             setlength (SackRanges, 1);
             SackRanges[len] := NewRange;
        end
     else
         begin
              for i := 1 to len do
                  begin
                       // Check if the new range replaces an existing one

                       range := SackRanges[i-1];

                       if (NewRange.leftEdge <= range.leftEdge) and (NewRange.rightEdge >= range.rightEdge) then
                          SackRanges[i-1] := NewRange
                       else


                  end;
         end;
end;

(* ------------------------------------------- *)

procedure tTCPFlowState.AddSackRanges (const NewRanges: tTCPSACKRanges);

var len : integer;
    i   : integer;

begin
     len := length (SackRanges);

     // If we have no SACK ranges yet, it's easy, otherwise we need to merge ranges

     if len = 0 then
        begin
             setLength (SackRanges, length (NewRanges));

             for i := 1 to length (NewRanges) do
                 SackRanges[i-1] := NewRanges[i-1];
        end
     else
         begin




         end;
end;

(* ------------------------------------------- *)

procedure tTCPFlowState.AddState (const NewState: tTCPFlowState);

begin
     // Under normal circumstances, the new state should be following directly
     // after what we already have. Or maybe the other way around.
     // If not, things get a little more complicated

     if NextExpectedSeq = NewState.LowestSeq then
        begin
             // Looks good, merge

             NextExpectedSeq := NewState.NextExpectedSeq;
             HighestSeq := NewState.HighestSeq;
             LastSeq := NewState.LastSeq;
             LastWindow := NewState.LastWindow;
             SeqWrapCount := SeqWrapCount + NewState.SeqWrapCount;
             TotalSegmentBytes := TotalSegmentBytes + NewState.TotalSegmentBytes;
             PacketCount := PacketCount + NewState.PacketCount;
             AcknowledgedTo := max (AcknowledgedTo, NewState.AcknowledgedTo);
             BytesInFlight := TotalSegmentBytes - AcknowledgedTo;
        end
     else
         if NewState.NextExpectedSeq = LowestSeq then
            begin
                 // The new state is actually "earlier"

                 LowestSeq := NewState.LowestSeq;
                 InitialSeq := NewState.InitialSeq;
                 SeqWrapCount := SeqWrapCount + NewState.SeqWrapCount;
                 TotalSegmentBytes := TotalSegmentBytes + NewState.TotalSegmentBytes;
                 PacketCount := PacketCount + NewState.PacketCount;
                 AcknowledgedTo := max (AcknowledgedTo, NewState.AcknowledgedTo);
                 BytesInFlight := TotalSegmentBytes - AcknowledgedTo;
            end
         else
             begin
                  // Okay, complicated it is.






             end;
end;
}
(* ------------------------------------------- *)
{ tTCPSessionState }
(* ------------------------------------------- *)
{
procedure tTCPSessionState.initialize;

begin
     ClientIPv4Address.initialize;
     ClientIPv6Address.initialize;
     ClientSynSeen := false;
     ClientRstSeen := false;
     ClientFinSeen := false;
     ClientFlowState.initialize;

     ServerIPv4Address.initialize;
     ServerIPv6Address.initialize;
     ServerSynSeen := false;
     ServerRstSeen := false;
     ServerFinSeen := false;
     ServerFlowState.initialize;

     FirstPacketTimestamp.initialize;
     LastPacketTimestamp.initialize;
end;
}

(* ------------------------------------------- *)

end.
