{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit dialog_anon_replacementdetails;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,

  base_structures,
  class_conversationstorage;

type tEventAddAndNew = procedure (const ReplacementEditType : tReplacementEditType; Original, Replacement : string) of object;

     TFormReplacementDetails = class(TForm)
       EditOriginal: TEdit;
       EditReplacement: TEdit;
       LabelOriginal: TLabel;
       LabelReplacement: TLabel;
       ButtonOkay: TButton;
       ButtonCancel: TButton;
       ButtonAddAndNew: TButton;
       MemoActionHint: TMemo;
       LabelMessage: TLabel;

       procedure onCreate(Sender: TObject);
       procedure onShow(Sender: TObject);
       procedure onKeyPressEdit(Sender: TObject; var Key: Char);
       procedure ButtonAddAndNewClick(Sender: TObject);
       procedure onKeyUpEdit(Sender: TObject; var Key: Word; Shift: TShiftState);
       procedure onKeyDownEdit(Sender: TObject; var Key: Word; Shift: TShiftState);
       procedure onEnterOriginal(Sender: TObject);
       procedure onEnterReplacement(Sender: TObject);
       procedure onExitEdit(Sender: TObject);

       private
              { Private declarations }

              EventOnAddAndNew       : tEventAddAndNew;

              priv_editmode          : tReplacementEditMode;
              priv_edittype          : tReplacementEditType;

              priv_conversationstore : tConversationStorage;

              procedure DoInputSanityChecks;
       public
             { Public declarations }

             property onAddAndNew : tEventAddAndNew read EventOnAddAndNew write EventOnAddAndNew;
             property ReplacementEditMode : tReplacementEditMode read priv_editmode write priv_editmode;
             property ReplacementEditType : tReplacementEditType read priv_edittype write priv_edittype;
             property ConversationStore : tConversationStorage read priv_conversationstore write priv_conversationstore;
     end;

var FormReplacementDetails : TFormReplacementDetails;

implementation

{$R *.dfm}

uses Toolkit,
     definition_ip_addresses,
     definition_conversation_types;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.ButtonAddAndNewClick(Sender: TObject);

var Original    : string;
    Replacement : string;

begin
     Original := EditOriginal.Text;
     Replacement := EditReplacement.Text;

     if assigned (EventOnAddAndNew) then
        EventOnAddAndNew (priv_edittype, Original, Replacement);

     EditOriginal.SelectAll;
     EditOriginal.SetFocus;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onCreate(Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     priv_edittype := tETGeneric;
     priv_editmode := tEMAdd;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onEnterOriginal (Sender: TObject);

begin
     case priv_edittype of
      tETMACAddress    : MemoActionHint.Text := 'Enter original MAC addresses with the octets seprarated by colons.' + CRLF + CRLF + 'Example: 01:02:03:04:05:06';
      tETVLANID        : MemoActionHint.Text := 'Enter original VLAN ID between 0 and 4095.' + CRLF + CRLF + 'Example: 1234';
      tETIPv4Address   : MemoActionHint.Text := 'Enter original IPv4 addresses.' + CRLF + CRLF + 'Example: 192.168.0.1';
      tETIPv6Address   : MemoActionHint.Text := 'Enter original IPv6 addresses in long or shortened notation.' + CRLF + CRLF + 'Example: 2001:db8:1234::1';
      tETIPv4Subnet    : MemoActionHint.Text := 'Enter original IPv4 subnets using slashed notation.' + CRLF + CRLF + 'Example: 192.168.0.0/24';
      tETIPv6Subnet    : MemoActionHint.Text := 'Enter original IPv6 subnets using slashed notation.' + CRLF + 'Example: 2001:db8:1234:5678::/64';
      tETTCPPort,
      tETUDPPort       : MemoActionHint.Text := 'Enter original Port number.' + CRLF + CRLF + 'Example: 49152';
      tETTCPSocketPort,
      tETUDPSocketPort : MemoActionHint.Text := 'Enter original IPv4 or IPv6 socket. Use zero address as wildcard.' + CRLF + CRLF + 'Examples: 192.168.0.1:80' + CRLF + '          [2001:db8:1234::1]:80';
     end;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onEnterReplacement (Sender: TObject);

begin
     case priv_edittype of
      tETMACAddress    : MemoActionHint.Text := 'Enter replacement MAC addresses with the octets seprarated by colons.' + CRLF + CRLF + 'Example: 01:02:03:04:05:06';
      tETVLANID        : MemoActionHint.Text := 'Enter replacement VLAN ID between 0 and 4095.' + CRLF + CRLF + 'Example: 1234';
      tETIPv4Address   : MemoActionHint.Text := 'Enter replacement IPv4 addresses.' + CRLF + CRLF + 'Example: 192.168.0.1';
      tETIPv6Address   : MemoActionHint.Text := 'Enter replacement IPv6 addresses in long or shortened notation.' + CRLF + CRLF + 'Example: 2001:db8:1234::1';
      tETIPv4Subnet    : MemoActionHint.Text := 'Enter replacement IPv4 subnets using slashed notation. Original and replacement must have the same mask bit count.' + CRLF + CRLF + 'Example: 192.168.0.0/24';
      tETIPv6Subnet    : MemoActionHint.Text := 'Enter replacement IPv6 subnets using slashed notation. Original and replacement must have the same mask bit count.' + CRLF + 'Example: 2001:db8:1234:5678::/64';
      tETTCPPort,
      tETUDPPort       : MemoActionHint.Text := 'Enter replacement Port number.' + CRLF + CRLF + 'Example: 49152';
      tETTCPSocketPort,
      tETUDPSocketPort : MemoActionHint.Text := 'Enter replacement IPv4 or IPv6 socket.' + CRLF + CRLF + 'Examples: 192.168.0.1:80' + CRLF + '          [2001:db8:1234::1]:80';
     end;
end;


(* ------------------------------------------- *)

procedure TFormReplacementDetails.onExitEdit (Sender: TObject);

begin
     DoInputSanityChecks;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onKeyDownEdit (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if Shift = [ssCtrl] then
        begin
             Case Key of
               Ord ('C') : (Sender as tEdit).CopyToClipboard;
               Ord ('V') : (Sender as tEdit).PasteFromClipboard;
               Ord ('X') : (Sender as tEdit).CutToClipboard;
             end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onKeyPressEdit (Sender: TObject; var Key: Char);

begin
     case priv_edittype of
      tETMACAddress   : begin
                             if not (key in [#8, #9, '0'..'9','a'..'f','A'..'F', ':']) then
                                key:=#0;
                        end;
      tETVLANID        : begin
                              if not (key in [#8, #9, '0'..'9']) then
                                 key:=#0;
                         end;
      tETIPv4Address   : begin
                              if not (key in [#8, #9, '0'..'9','.']) then
                                 key:=#0;
                         end;
      tETIPv6Address   : begin
                              if not (key in [#8, #9, '0'..'9','a'..'f','A'..'F', ':']) then
                                 key:=#0;
                         end;
      tETIPv4Subnet    : begin
                              if not (key in [#8, #9, '0'..'9','.','/']) then
                                 key:=#0;
                         end;
      tETIPv6Subnet    : begin
                              if not (key in [#8, #9, '0'..'9','a'..'f','A'..'F', ':','/']) then
                                 key:=#0;
                         end;
      tETTCPPort,
      tETUDPPort       : begin
                              if not (key in [#8, #9, '0'..'9']) then
                                 key:=#0;
                         end;
      tETTCPSocketPort,
      tETUDPSocketPort : begin
                              if (Sender as tEdit).Name = 'EditOriginal' then
                                 begin
                                      if not (key in [#8, #9, '0'..'9','.','a'..'f','A'..'F', ':', '[',']']) then
                                         key:=#0;
                                 end
                              else
                                      if not (key in [#8, #9, '0'..'9','.','a'..'f','A'..'F', ':', '[',']']) then
                                         key:=#0;
                         end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onKeyUpEdit (Sender: TObject; var Key: Word; Shift: TShiftState);

var value   : string;
    number  : integer;
    valid   : boolean;
    help    : string;
    mask1   : integer;
    mask2   : integer;
    AllGood : boolean;

begin
     Value := (Sender as TEdit).Text;
     TryStrToInt (value, number);

     case priv_edittype of
      tetGeneric       : valid := true;
      tETMACAddress    : valid := isValidMAC (value);
      tETVLANID        : valid := (number <= 4096);
      tETIPv4Address   : valid := isValidIPv4String (value);
      tETIPv6Address   : valid := isValidIPv6String (value);
      tETIPv4Subnet    : valid := isValidIPv4Subnet (value);
      tETIPv6Subnet    : valid := isValidIPv6Subnet (value);
      tETTCPPort       : valid := (number > 0) and (number <= 65535);
      tETUDPPort       : valid := (number > 0) and (number <= 65535);
      tETTCPSocketPort,
      tETUDPSocketPort : begin
                              if (Sender as tEdit).Name = 'EditOriginal' then
                                 valid := isValidIPv4Socket (value) or isValidIPv6Socket (value)
                              else
                                  valid := isValidIPv4Socket (value) or isValidIPv6Socket (value);
                         end;
     end;

     // mark the input red or black

     if valid then
        (Sender as TEdit).Font.Color := clWindowText
     else
         (Sender as TEdit).Font.Color := clRed;

     // enable Okay button if everything is okay

     case priv_edittype of
      tETIPv4Subnet,
      tETIPv6Subnet : begin
                           help := GetEntry (EditOriginal.Text, '/', 2);
                           TryStrToInt (help, mask1);

                           help := GetEntry (EditReplacement.Text, '/', 2);
                           TryStrToInt (help, mask2);

                           AllGood := valid and (mask1 = mask2);
                      end;
     else
         AllGood := valid;
     end;

     ButtonOkay.Enabled := (EditOriginal.Text <> '') and (EditReplacement.Text <> '') and AllGood;
     ButtonAddAndNew.Enabled := ButtonOkay.Enabled;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.onShow (Sender: TObject);

begin
     if priv_editmode = tEMEdit then
        begin
             ButtonOkay.Left := 44;
             ButtonOkay.Caption := '&Okay';
             ButtonCancel.Left := 125;

             ButtonAddAndNew.Visible := false;

             ButtonOkay.Default := true;
             ButtonOkay.Enabled := true;
        end
     else
         begin
             ButtonOkay.Left := 8;
             ButtonOkay.Caption := '&Add';

             ButtonCancel.Left := 166;

             ButtonAddAndNew.Visible := true;
             ButtonAddAndNew.Default := true;
         end;

     EditOriginal.Font.Color := clWindowText;
     LabelMessage.Caption := '';

     EditOriginal.SetFocus;
end;

(* ------------------------------------------- *)

procedure TFormReplacementDetails.DoInputSanityChecks;

var boOkay      : boolean;
    value1      : integer;
    value2      : integer;
    count       : integer;
    Address     : tIPAddress;
    socket      : tIPSocket;
    sockdetails : tSocketDetails;
    boExists    : boolean;

begin
     boOkay := false;

     if (EditOriginal.Text <> '') and (EditReplacement.Text <> '') then
        begin
             TryStrToInt (EditOriginal.Text, value1);
             TryStrToInt (EditReplacement.Text, value2);

             case priv_edittype of
              tETMACAddress    : begin
                                      boOkay := isValidMAC (EditOriginal.Text) and isValidMAC (EditReplacement.Text);
                                 end;
              tETVLANID        : begin
                                      boOkay := (value1 < 4096) and (value2 < 4096);
                                 end;
              tETIPv4Address   : begin
                                      if priv_conversationstore.HasRecords then
                                         begin
                                              if isValidIPString (EditOriginal.Text) then
                                                 begin
                                                      address.setValueFromString (EditOriginal.Text);

                                                      // Check if we have an entry for the given IP address

                                                      boExists := not (priv_conversationstore.GetIPv4EndpointEntry (address.IPv4Address).IPv4Address.isUndefined);

                                                      if boExists then
                                                         begin
                                                              EditOriginal.Font.Color := clWindowText;
                                                              LabelMessage.Caption := '';
                                                         end
                                                      else
                                                          begin
                                                               EditOriginal.Font.Color := clGray;

                                                               LabelMessage.Caption := '** Original IP address not found in current fileset **';
                                                               LabelMessage.Font.Color := clRed;
                                                          end;
                                                 end;
                                         end;

                                      boOkay := (isValidIPv4String (EditOriginal.Text) and isValidIPv4String (EditReplacement.Text));
                                 end;
              tETIPv6Address   : begin
                                      if priv_conversationstore.HasRecords then
                                         begin
                                              if isValidIPString (EditOriginal.Text) then
                                                 begin
                                                      address.setValueFromString (EditOriginal.Text);

                                                      // Check if we have an entry for the given IP address

                                                      boExists := not (priv_conversationstore.GetIPv6EndpointEntry (address.IPv6Address).IPv6Address.isUndefined);

                                                      if boExists then
                                                         begin
                                                              EditOriginal.Font.Color := clWindowText;
                                                              LabelMessage.Caption := '';
                                                         end
                                                      else
                                                          begin
                                                               EditOriginal.Font.Color := clGray;

                                                               LabelMessage.Caption := '** Original IP address not found in current fileset **';
                                                               LabelMessage.Font.Color := clRed;
                                                          end;
                                                 end;
                                         end;

                                      boOkay := (isValidIPv6String (EditOriginal.Text) and isValidIPv6String (EditReplacement.Text));
                                 end;
              tETIPv4Subnet    : begin
                                      boOkay := (isValidIPv4Subnet (EditOriginal.Text) and isValidIPv4Subnet (EditReplacement.Text));
                                 end;
              tETIPv6Subnet    : begin
                                      boOkay := (isValidIPv6Subnet (EditOriginal.Text) and isValidIPv6Subnet (EditReplacement.Text));
                                 end;
              tETTCPPort,
              tETUDPPort       : begin
                                      boOkay := (value1 > 0) and (value1 <= 65535) and (value2 > 0) and (value2 <= 65535);
                                 end;
              tETTCPSocketPort,
              tETUDPSocketPort : begin
                                      if priv_conversationstore.HasRecords then
                                         begin
                                              if isValidSocketString (EditOriginal.Text) then
                                                 begin
                                                      socket.initialize (EditOriginal.Text);

                                                      // Get the details for TCP or UDP

                                                      if priv_edittype = tETTCPSocketPort then
                                                         sockdetails := priv_conversationstore.GetSocketEntry (protTCP, socket)
                                                      else
                                                          sockdetails := priv_conversationstore.GetSocketEntry (protUDP, socket);

                                                      if sockdetails.UseCount = 0 then
                                                         begin
                                                              EditOriginal.Font.Color := clGray;

                                                              LabelMessage.Caption := '** Original Socket not found in current fileset **';
                                                              LabelMessage.Font.Color := clRed;
                                                         end
                                                      else
                                                          begin
                                                               EditOriginal.Font.Color := clWindowText;
                                                               LabelMessage.Caption := '';
                                                          end;
                                                 end;
                                         end;

                                      boOkay := (isValidIPv4Socket (EditOriginal.Text) and isValidIPv4Socket (EditReplacement.Text)) or
                                                (isValidIPv6Socket (EditOriginal.Text) and isValidIPv6Socket (EditReplacement.Text));
                                 end;
             end;
        end;

     ButtonOkay.Enabled := boOkay;
     ButtonAddAndNew.Enabled := boOkay;
end;

(* ------------------------------------------- *)

end.
