{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit dialog_anon_selectaddresses;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ComCtrls, Vcl.Grids,
  Vcl.ValEdit, VirtualTrees, Vcl.StdCtrls,

  base_structures,
  definition_conversation_types, Vcl.ExtCtrls;

type TFormSelectAddresses = class(TForm)
                              VSTAddressList: TVirtualStringTree;
                              ValueListEditorReplacementList: TValueListEditor;
                              StatusBar: TStatusBar;
                              ButtonAdd: TButton;
                              ButtonRemove: TButton;
                              ButtonOkay: TButton;
                              ButtonCancel: TButton;
                              TimerClearSearchStatus: TTimer;
                              procedure onCreate(Sender: TObject);
                              procedure onDestroy(Sender: TObject);
                              procedure onShow(Sender: TObject);
                              procedure onVSTAddressListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                              procedure onChangeVSTAddressList(Sender: TBaseVirtualTree; Node: PVirtualNode);
                              procedure ButtonAddClick(Sender: TObject);
                              procedure onVLEReplacementListValidate(Sender: TObject; ACol, ARow: Integer; const KeyName, KeyValue: string);
                              procedure onVLEReplacementListDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
                              procedure onVLEReplacementListKeyPress(Sender: TObject; var Key: Char);
                              procedure onVSTAddressListIncrementalSearch(Sender: TBaseVirtualTree; Node: PVirtualNode; const SearchText: string; var Result: Integer);
                              procedure onTimeClearSearchStatus(Sender: TObject);
                              procedure onDblClickVSTAddressList(Sender: TObject);
                              procedure onVSTAddressListCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                              procedure ButtonRemoveClick(Sender: TObject);
                              procedure onVLEReplacementSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
                            private
                                   priv_ReplacementEditType  : tReplacementEditType;
//                                 priv_addresslist          : tStringlist;
                                   priv_MACEndpointList      : tMACEndpointList;
                                   priv_IPv4EndpointList     : tIPv4EndpointList;
                                   priv_IPv6EndpointList     : tIPv6EndpointList;
                                   priv_IPv4ConversationList : tIPv4ConversationList;
                                   priv_IPv6ConversationList : tIPv6ConversationList;

                                   function GetAddressFromNode (Node : pVirtualNode) : string;

                            public
                                  property ReplacementEditType : tReplacementEditType read priv_ReplacementEditType write priv_ReplacementEditType;

                                  procedure PrepareValues (AddressList : tMACEndpointList); overload;
                                  procedure PrepareValues (AddressList : tIPv4EndpointList); overload;
                                  procedure PrepareValues (AddressList : tIPv6EndpointList); overload;
                                  procedure PrepareValues (AddressList : tIPv4ConversationList); overload;
                                  procedure PrepareValues (AddressList : tIPv6ConversationList); overload;

                                  procedure getReplacementPairs (var Values : tStringList);

                            end;

var
  FormSelectAddresses: TFormSelectAddresses;

implementation

uses Math,
     definition_ip_addresses;

{$R *.dfm}

type tVSTSourceAddress = record
                               EditType           : tReplacementEditType;
                               Added              : boolean;
                               Index              : integer;
                         end;
     pVSTSourceAddress = ^tVSTSourceAddress;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onCreate(Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     VSTAddressList.NodeDataSize := sizeOf (tVSTSourceAddress);

//   priv_addresslist := tStringlist.Create;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onDblClickVSTAddressList(Sender: TObject);

var node    : pVirtualNode;
    data    : pVSTSourceAddress;
    Address : string;

begin
     node := VSTAddressList.GetFirstSelected;

     if node <> nil then
        begin
             Address := GetAddressFromNode (Node);
             ValueListEditorReplacementList.InsertRow (Address, '', true);

//           ButtonRemove.Enabled := true;
        end;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onDestroy(Sender: TObject);

begin
//   priv_addresslist.Free;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onShow(Sender: TObject);

var node  : pVirtualNode;
    data  : pVSTSourceAddress;
    count : integer;
    i     : integer;

begin
//   priv_addresslist.Sort;

     ValueListEditorReplacementList.Strings.Clear;

     VSTAddressList.Clear;
     VSTAddressList.BeginUpdate;

     case priv_ReplacementEditType of
      tETMACAddress       : count := length (priv_MACEndpointList);
      tETIPv4Address      : count := length (priv_IPv4EndpointList);
      tETIPv6Address      : count := length (priv_IPv6EndpointList);
      tETIPv4Conversation : count := length (priv_IPv4ConversationList);
      tETIPv6Conversation : count := length (priv_IPv6ConversationList);
     end;

     for i := 1 to count do
         begin
              node := VSTAddressList.AddChild(nil);
              data := VSTAddressList.GetNodeData(node);

              data^.EditType := priv_ReplacementEditType;
              data^.Index := i-1;
              data^.Added := false;
         end;

     VSTAddressList.SortTree (0, sdAscending);
     VSTAddressList.EndUpdate;

     TimerClearSearchStatus.Interval := VSTAddressList.IncrementalSearchTimeout;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onTimeClearSearchStatus(Sender: TObject);

begin
     TimerClearSearchStatus.Enabled := false;

     if pos ('Searching', Statusbar.Panels[1].Text) > 0 then
        Statusbar.Panels[1].Text := '';
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVSTAddressListCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var addr1 : string;
    addr2 : string;

begin
     addr1 := GetAddressFromNode (Node1);
     addr2 := GetAddressFromNode (Node2);

     Result := compareStr (addr1, addr2);
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVSTAddressListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

begin
     CellText := GetAddressFromNode (Node);
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onChangeVSTAddressList (Sender: TBaseVirtualTree; Node: PVirtualNode);

begin
     ButtonAdd.Enabled := VSTAddressList.SelectedCount > 0;
end;

(* ------------------------------------------- *)


procedure TFormSelectAddresses.onVSTAddressListIncrementalSearch (Sender: TBaseVirtualTree; Node: PVirtualNode; const SearchText: string; var Result: Integer);

var address : string;

begin
     if assigned (node) then
        begin
             TimerClearSearchStatus.Enabled := false;

             Statusbar.Panels[1].Text := 'Searching for: ' + SearchText;

             address := GetAddressFromNode (node);

             Result := StrLIComp(PChar(SearchText), PChar(address), Min(Length(SearchText), Length(address)));

             TimerClearSearchStatus.Enabled := true;
        end;
end;

(* ------------------------------------------- *)

function TFormSelectAddresses.GetAddressFromNode (Node: pVirtualNode): string;

var data : pVSTSourceAddress;
    addr : string;

begin
     data := VSTAddressList.GetNodeData(node);

     case data^.EditType of
      tETMACAddress       : addr := priv_MACEndpointList[data^.Index].MACAddress.AsString;
      tETIPv4Address      : addr := priv_IPv4EndpointList[data^.Index].IPv4Address.ValueAsString;
      tETIPv6Address      : addr := priv_IPv6EndpointList[data^.Index].IPv6Address.ValueAsString;
      tETIPv4Conversation : addr := priv_IPv4ConversationList[data^.Index].IPv4AddressPair.ValueAsString (false);
      tETIPv6Conversation : addr := priv_IPv6ConversationList[data^.Index].IPv6AddressPair.ValueAsString (false);
     end;

     result := addr;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.getReplacementPairs (var Values: tStringList);

var i : integer;

begin
     for i := 1 to ValueListEditorReplacementList.Strings.Count do
         Values.Add (ValueListEditorReplacementList.Strings[i-1]);
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVLEReplacementListDrawCell (Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);

var value : string;

begin
     if (ARow > 0) and (ACol = 1) then
        begin
             value := ValueListEditorReplacementList.Cells[ACol, ARow];

             case priv_ReplacementEditType of
              tETMACAddress  : begin
                                    if isValidMAC (value) then
                                       ValueListEditorReplacementList.Canvas.Font.Color := clWindowText
                                    else
                                        ValueListEditorReplacementList.Canvas.Font.Color := clRed;

                                    ValueListEditorReplacementList.Canvas.TextRect (Rect, Rect.Left + 4, Rect.Top + 2, value);
                               end;
              tETIPv4Address : begin
                                    if isValidIPv4String (value) then
                                       ValueListEditorReplacementList.Canvas.Font.Color := clWindowText
                                    else
                                        ValueListEditorReplacementList.Canvas.Font.Color := clRed;

                                    ValueListEditorReplacementList.Canvas.TextRect (Rect, Rect.Left + 4, Rect.Top + 2, value);
                               end;
              tETIPv6Address : begin
                                    if isValidIPv6String (value) then
                                       ValueListEditorReplacementList.Canvas.Font.Color := clWindowText
                                    else
                                        ValueListEditorReplacementList.Canvas.Font.Color := clRed;

                                    ValueListEditorReplacementList.Canvas.TextRect (Rect, Rect.Left + 4, Rect.Top + 2, value);
                               end;
             end;
        end
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVLEReplacementListKeyPress (Sender: TObject; var Key: Char);

begin
     if not (key in [#8, #9, '0'..'9','a'..'f','A'..'F', '.', ':']) then
        key:=#0;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVLEReplacementListValidate (Sender: TObject; ACol, ARow: Integer; const KeyName, KeyValue: string);

begin
     case priv_ReplacementEditType of
      tETMACAddress  : begin
                            ButtonOkay.Enabled := isValidMAC (KeyValue);
                       end;
      tETIPv4Address : begin
                            ButtonOkay.Enabled := isValidIPv4String (KeyValue);
                       end;
      tETIPv6Address : begin
                            ButtonOkay.Enabled := isValidIPv6String (KeyValue);
                       end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.onVLEReplacementSelectCell (Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);

begin

end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.PrepareValues (AddressList: tMACEndpointList);

var count : integer;
    i     : integer;

begin
     priv_ReplacementEditType := tETMACAddress;

     setLength (priv_MACEndpointList, 0);
     count := length (AddressList);
     setLength (priv_MACEndpointList, count);

     for i := 1 to count do
         priv_MACEndpointList[i-1] := AddressList[i-1];
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.PrepareValues (AddressList: tIPv4EndpointList);

var count : integer;
    i     : integer;

begin
     priv_ReplacementEditType := tETIPv4Address;

     setLength (priv_IPv4EndpointList, 0);
     count := length (AddressList);
     setLength (priv_IPv4EndpointList, count);

     for i := 1 to count do
         priv_IPv4EndpointList[i-1] := AddressList[i-1];
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.PrepareValues (AddressList: tIPv6EndpointList);

var count : integer;
    i     : integer;

begin
     priv_ReplacementEditType := tETIPv6Address;

     setLength (priv_IPv6EndpointList, 0);
     count := length (AddressList);
     setLength (priv_IPv6EndpointList, count);

     for i := 1 to count do
         priv_IPv6EndpointList[i-1] := AddressList[i-1];
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.PrepareValues (AddressList: tIPv4ConversationList);

var count : integer;
    i     : integer;

begin
     priv_ReplacementEditType := tETIPv4Conversation;

     setLength (priv_IPv4ConversationList, 0);
     count := length (AddressList);
     setLength (priv_IPv4ConversationList, count);

     for i := 1 to count do
         priv_IPv4ConversationList[i-1] := AddressList[i-1];
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.PrepareValues (AddressList: tIPv6ConversationList);

var count : integer;
    i     : integer;

begin
     priv_ReplacementEditType := tETIPv6Conversation;

     setLength (priv_IPv6ConversationList, 0);
     count := length (AddressList);
     setLength (priv_IPv6ConversationList, count);

     for i := 1 to count do
         priv_IPv6ConversationList[i-1] := AddressList[i-1];
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.ButtonAddClick(Sender: TObject);

var node : pVirtualNode;
    data : pVSTSourceAddress;
    i    : integer;

begin
     case priv_ReplacementEditType of
      tETMACAddress,
      tETIPv4Address,
      tETIPv6Address      : begin
                                 i := 0;

                                 node := VSTAddressList.GetFirstSelected;

                                 repeat
                                       data := VSTAddressList.GetNodeData (node);

                                       if not data^.Added then
                                          ValueListEditorReplacementList.InsertRow (GetAddressFromNode (node), '', true);

                                       node := VSTAddressList.GetNextSelected (node);

                                 until node = nil;
                            end;
     end;


//   ButtonRemove.Enabled := true;
end;

(* ------------------------------------------- *)

procedure TFormSelectAddresses.ButtonRemoveClick(Sender: TObject);

begin

end;

(* ------------------------------------------- *)

end.

