{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit dialog_conversation_summary;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ComCtrls, VirtualTrees, Vcl.Menus, Vcl.StdCtrls,
  Vcl.ExtCtrls, Vcl.ExtDlgs, System.Generics.Collections,

  base_structures,
  class_tracefilelist,
  definition_ip_addresses,
  definition_filter_structures,
  definition_interface_mapping_types,
  definition_conversation_types,
{$ifdef fqdnstats}
  class_fqdnmanager,
{$endif}
  class_conversationstorage;


const tabcount = 9;

type tConversationMenuMode = (cmmMainForm, cmmMergeForm, cmmExtractForm);
     tNodeProcessAction = (paMarkRows);
     tMarkJumpMode = (jmNext, jmPrevious);
     tHighlightMode = (hlmNone, hlmTimeEmbedded);

     tOnAddFilterFromList = procedure (const Filter : tFrameFilter) of object;

     tSelectedStats = record
                            TotalFrames     : int64;
                            TotalBytes      : int64;

                            procedure initialize;
                      end;

     tHighlightTimeEdges = record
                                 FirstFrameTime : tFrameTimestamp;
                                 LastFrameTime : tFrameTimestamp;

                                 procedure initialize;
                                 function isSet : boolean;
                           end;

     tTabValues     = record
                            CurrentFilterStr     : string;
                            CurrentSocketFilter  : tIPSocket;
                            currentmarkedrow     : integer;
                            RowIndexList         : tDictionary<integer, PVirtualNode>;
                            MarkedRowList        : tDictionary<integer, PVirtualNode>;
                            CurrentRow           : integer;
                            HighlightMode        : tHighlightMode;
                            HighlightTimeEdges   : tHighlightTimeEdges;
                            AggregatedMode       : boolean;
                            MainTree             : TBaseVirtualTree;

                            procedure initialize;
                            procedure free;
                      end;

     TFormConversationSummary = class (TForm)
                                      PageControlConversations: TPageControl;
                                      StatusBar: TStatusBar;
                                      TabSheetMACEndpoints: TTabSheet;
                                      TabSheetIPv4Addresses: TTabSheet;
                                      TabSheetIPv6Addresses: TTabSheet;
                                      TabSheetMACConversations: TTabSheet;
                                      TabSheetIPv4Conversations: TTabSheet;
                                      TabSheetIPv6Conversations: TTabSheet;
                                      TabSheetTCPConversations: TTabSheet;
                                      TabSheetUDPConversations: TTabSheet;
                                      TabSheetNameResolution: TTabSheet;
                                      VSTMACEndpoints: TVirtualStringTree;
                                      VSTIPv4Endpoints: TVirtualStringTree;
                                      VSTIPv6Endpoints: TVirtualStringTree;
                                      VSTIPv4Conversations: TVirtualStringTree;
                                      VSTIPv6Conversations: TVirtualStringTree;
                                      VSTTCPConversations: TVirtualStringTree;
                                      VSTUDPConversations: TVirtualStringTree;
                                      VSTMACConversations: TVirtualStringTree;
                                      VSTNameResolution: TVirtualStringTree;

                                      PopupMenuEndpointTreeView: TPopupMenu;
                                      MenuCreateEndpointSharkFilter: TMenuItem;
                                      PopupMenuConversationTreeView: TPopupMenu;
                                      MenuCreateConversationSharkFilter: TMenuItem;
                                      MenuCreateConversationSharkFilterAandB: TMenuItem;
                                      MenuCreateConversationSharkFilterAtoB: TMenuItem;
                                      MenuCreateConversationSharkFilterBtoA: TMenuItem;
                                      MenuCreateConversationSharkFilterAtoAny: TMenuItem;
                                      MenuCreateConversationSharkFilterBtoAny: TMenuItem;

                                      PopupMenuMergeFilter: TPopupMenu;
                                      MenuAddAsMergeFilter: TMenuItem;
                                      MenuConversationConfigureDisplay: TMenuItem;
                                      LabeledEditFilter: TLabeledEdit;
                                      MenuEndpointConfigureDisplay: TMenuItem;
                                      MenuConversationExtract: TMenuItem;
                                      MenuConversationExtractToSingleFile: TMenuItem;
                                      MenuConversationExtractToMultipleFiles: TMenuItem;
                                      MenuEndpointExtract: TMenuItem;
                                      MenuEndpointExtractToSingleFile: TMenuItem;
                                      MenuEndpointExtractToMultipleFiles: TMenuItem;
                                      SaveDialogExtract: TSaveDialog;
                                      TimerApplyFilter: TTimer;
                                      TimerDisplayAll: TTimer;
                                      MenuConversationExportToCSV: TMenuItem;
                                      SaveTextFileExportCSV: TSaveTextFileDialog;
                                      MenuConversationOpen: TMenuItem;
                                      MenuEndpointOpen: TMenuItem;
                                      MenuEndpointExportToCSV: TMenuItem;
                                      PopupMenuHeaderColumnSelector: TPopupMenu;
                                      MenuColumnIndex: TMenuItem;
                                      MenuColumnAddressA: TMenuItem;
                                      MenuColumnPortA: TMenuItem;
                                      MenuConversationCopyCellToClipboard: TMenuItem;
                                      MenuConversationSeparator1: TMenuItem;
                                      MenuConversationView: TMenuItem;
                                      MenuConversationViewTCPHighlightMode: TMenuItem;
                                      MenuConversationViewTCPAggregatedMode: TMenuItem;
                                      MenuConversationMarkRow: TMenuItem;
                                      MenuConversationGotoRow: TMenuItem;
                                      MenuConversationGoto: TMenuItem;
                                      MenuConversationPreviousMark: TMenuItem;
                                      MenuConversationNextMark: TMenuItem;
                                      MenuConversationClearAllMarks: TMenuItem;
                                      MenuConversationLookup: TMenuItem;
                                      MenuConversationLookupIPAPI: TMenuItem;

                                      procedure onCreate(Sender: TObject);
                                      procedure onDestroy(Sender: TObject);
                                      procedure onShow (Sender: TObject);

                                      procedure onVSTMACEndPointsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTIPv4EndpointsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTIPv6EndpointsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTMACConversationsGetText(Sender: TBaseVirtualTree;Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onIPv4ConversationsGetText(Sender: TBaseVirtualTree;Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTIPv6ConversationsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onTCPConversationsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTUDPConversationsGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTNameResolutionGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
                                      procedure onVSTHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);
                                      procedure onVSTMACEndpointsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTIPv4EndpointsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTIPv6EndpointsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTMACConversationsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTIPv4ConversationsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTIPv6ConversationsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTTCPConversationsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTUDPConversationsCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure onVSTNameResolutionCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
                                      procedure MenuCreateEndpointSharkFilterClick(Sender: TObject);
                                      procedure MenuCreateConversationSharkFilterClick(Sender: TObject);
                                      procedure onVSTTCPConversationsBeforeCellPaint (Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; CellPaintMode: TVTCellPaintMode; CellRect: TRect; var ContentRect: TRect);
                                      procedure onVSTTCPConnectionsPaintText(Sender: TBaseVirtualTree; const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType);
                                      procedure MenuAddAsMergeFilterClick(Sender: TObject);
                                      procedure onStatusbarDrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel; const Rect: TRect);
                                      procedure onKeyUpFilter(Sender: TObject; var Key: Word; Shift: TShiftState);
                                      procedure onKeyPressFilter(Sender: TObject; var Key: Char);
                                      procedure onChangePage(Sender: TObject);
                                      procedure MenuConversationOpenClick(Sender: TObject);
                                      procedure MenuConversationExtractToSingleFileClick(Sender: TObject);
                                      procedure MenuConversationExtractToMultipleFilesClick(Sender: TObject);
                                      procedure MenuExportToCSVClick(Sender: TObject);
                                      procedure MenuConversationConfigureDisplayClick(Sender: TObject);
                                      procedure onTimerApplyFilter(Sender: TObject);
                                      procedure onTimerDisplayAll(Sender: TObject);
                                      procedure onVSTChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
                                      procedure onVSTGetHint(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; var LineBreakStyle: TVTTooltipLineBreakStyle; var HintText: string);
                                      procedure onVSTDblClick(Sender: TObject);
                                      procedure MenuEndpointOpenClick(Sender: TObject);
                                      procedure onPopupHeaderColumnSelector(Sender: TObject);
                                      procedure MenuColumnIndexClick(Sender: TObject);
                                      procedure onVSTKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
                                      procedure MenuConversationCopyCellToClipboardClick(Sender: TObject);
                                      procedure onVSTGetPopupMenu(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; const P: TPoint; var AskParent: Boolean; var PopupMenu: TPopupMenu);
                                      procedure MenuConversationViewTCPHighlightModeClick(Sender: TObject);
                                      procedure MenuConversationViewTCPAggregatedModeClick(Sender: TObject);
                                      procedure MenuConversationMarkRowClick(Sender: TObject);
                                      procedure MenuConversationGotoRowClick(Sender: TObject);
                                      procedure MenuConversationNextMarkClick(Sender: TObject);
                                      procedure MenuConversationPreviousMarkClick(Sender: TObject);
                                      procedure MenuConversationClearAllMarksClick(Sender: TObject);
                                      procedure MenuConversationLookupIPAPIClick(Sender: TObject);
                                      procedure MenuNameResolutionAggregatedModeClick(Sender: TObject);

                                    private
                                           EventOnAddFilterToList  : tOnAddFilterFromList;

                                           priv_filelist           : tTraceFileList;
                                           priv_conversationstore  : tConversationStorage;
{$ifdef fqdnstats}
                                           priv_FQDNManager        : tFQDNManager;
{$endif}
                                           priv_filterlist         : tFilterSet;
                                           priv_interfacemaplist   : tInterfaceFileIDList;

                                           priv_menumode           : tConversationMenuMode;

                                           priv_popuptree          : tBaseVirtualTree;
                                           priv_extractwritemode   : tMultiWriteMode;

                                           priv_settings           : tConversationDialogSettings;
                                           priv_loadcolorsfromdb   : boolean;
                                           priv_firstshow          : boolean;

                                           priv_tabvalues          : array [0..tabcount] of tTabValues;
                                           priv_selectedstats      : tSelectedStats;

                                           priv_temppath           : string;

                                           priv_lastsourcefilename : string;
                                           priv_CellTextCache      : string;

                                           // row lists, with all entries in the conversation store sorted by frame index,
                                           // so we don't have to pull them each and every time. Also avoiding calculating
                                           // the filter text this way. Right now only doing this for TCP

                                           priv_TCPList            : tTCPConversationList;
                                           priv_NRList             : tNameResolutionList;

//                                         priv_nanosecondprecision : boolean; do this later

                                           function RTTtoStr (InitialRTT : uint64) : string;

                                           procedure UpdateWindowCaption;
                                           procedure OpenWebBrowser (const URL : string);

                                           procedure DisplayMACEndpoints;
                                           procedure DisplayIPv4Endpoints;
                                           procedure DisplayIPv6Endpoints;
                                           procedure DisplayMACConversations;
                                           procedure DisplayIPv4Conversations;
                                           procedure DisplayIPv6Conversations;
                                           procedure DisplayTCPConversations;
                                           procedure DisplayUDPConversations;
{$ifdef fqdnstats}
                                           procedure DisplayNameResolutions;
{$endif}

                                           function CreateMACConversationFilter (Conversation : tMACConversationTreeEntry; MenuTag : integer) : string;
                                           function CreateIPv4ConversationFilter (Conversation : tIPv4ConversationTreeEntry; MenuTag : integer) : string;
                                           function CreateIPv6ConversationFilter (Conversation : tIPv6ConversationTreeEntry; MenuTag : integer) : string;
                                           function CreateTCPConversationFilter (Conversation : tTCPConversationTreeEntry; MenuTag : integer) : string;
                                           function CreateUDPConversationFilter (Conversation : tUDPConversationTreeEntry; MenuTag : integer) : string;
                                           function CopySharkFilterToClipboard (Tree : TBaseVirtualTree; const FilterType : integer) : string;

                                           function GetMACEndpointCSVExportLine (Node : pVirtualNode) : string;
                                           function GetIPv4EndpointCSVExportLine (Node : pVirtualNode) : string;
                                           function GetIPv6EndpointCSVExportLine (Node : pVirtualNode) : string;
                                           function GetMACConversationCSVExportLine (Node : pVirtualNode) : string;
                                           function GetIPv4ConversationCSVExportLine (Node : pVirtualNode) : string;
                                           function GetIPv6ConversationCSVExportLine (Node : pVirtualNode) : string;
                                           function GetTCPConversationCSVExportLine (Node : pVirtualNode) : string;
                                           function GetUDPConversationCSVExportLine (Node : pVirtualNode) : string;

                                           function getIPv4ConversationCellText (const column : integer; const Data : tIPv4ConversationTreeEntry) : string;
                                           function getIPv6ConversationCellText (const column : integer; const Data : tIPv6ConversationTreeEntry) : string;
                                           function getTCPConversationCellText (const column : integer; const Data : tTCPConversationTreeEntry) : string;
{$ifdef fqdnstats}
                                           function GetNameResolutionCellText (Const column : integer; const Data : tFQDNTreeNode) : string;
{$endif}

                                           procedure ProcessTCPNodes (ProcessAction : tNodeProcessAction);
                                           procedure ToggleHighlightMode;
                                           procedure ToggleAggregatedMode;
                                           procedure GotoRow;

                                           procedure ClearAllTCPMarks;
                                           procedure JumpToTCPMark  (const mode : tMarkJumpMode);

                                           function FilterBoxIsAMatch (FilterValue : string; const CompareTo : string) : boolean;
                                           /// <summary>
                                           ///   Extracts packets to a single
                                           ///   file. It returns a string
                                           ///   containing the path of the
                                           ///   last file that is used to
                                           ///   extract, to be able to call
                                           ///   ShellExecute with that path to
                                           ///   avoid jumping to temp.
                                           /// </summary>
                                           /// <returns>
                                           ///   Path of the last source file
                                           /// </returns>
                                           function ExtractToSingleFile (const Filename : string) : string;
                                           procedure ExtractToMultipleFiles (const Directory, TargetFile : string; const TruncateAfterLayer : byte);

                                           procedure DisplayConversationStore;
                                           procedure ExtractPacketsAndOpen (Tree : tBaseVirtualTree);

                                           procedure ShowColumnStatus (Tree : TBaseVirtualTree);

                                           procedure onTaskExtractNextFile (Sender : tObject; const FileIndex : integer; const FileEntry : tTraceFileEntry; var SkipFile : boolean);
                                           procedure onTaskExtractAdvancePercentProcessed (Sender : tObject; FileEntry : tTracefileEntry; const PercentageParsed : integer);

                                           {$REGION 'Documentation'}
                                           /// <summary>
                                           ///   Scans throught the currently
                                           ///   visible list and creates
                                           ///   filters for all selected rows.
                                           ///   At the same time, an interface
                                           ///   mapping list is created to
                                           ///   note which capture interfaces
                                           ///   in which files are being used
                                           ///   by each row.
                                           /// </summary>
                                           {$ENDREGION}
                                           procedure getFilterAndInterfaceMappingsFromList (Tree : tBaseVirtualTree);
                                           /// <summary>
                                           ///   Determines what extraction
                                           ///   write mode is required, e.g.
                                           ///   mwmL4Conversation for TCP
                                           ///   conversations. This helps
                                           ///   speed up building data
                                           ///   structures and can improve
                                           ///   filter speed by not looking at
                                           ///   irrelevant things <br />
                                           /// </summary>
                                           function determineWriteMode : tMultiWriteMode;

                                           /// <summary>
                                           ///   This routine is called by
                                           ///   assigning a
                                           ///   ConversationStorage object to
                                           ///   the form, which is required
                                           ///   before showing it. It will
                                           ///   also prepare the FQDNManager.
                                           /// </summary>
                                           procedure AssignNewConversationStore (ConversationStore : tConversationStorage);

                                    public
                                          property onAddFilterToList : tOnAddFilterFromList read EventOnAddFilterToList write EventOnAddFilterToList;

                                          property TracefileList : tTraceFileList read priv_filelist write priv_filelist;
                                          property ConversationStore : tConversationStorage read priv_conversationstore write AssignNewConversationStore;
                                          property FilterList : tFilterSet read priv_filterlist;

                                          property TempPath : string read priv_temppath write priv_temppath;

                                          procedure SetMode (MenuMode : tConversationMenuMode);
                                end;

var FormConversationSummary: TFormConversationSummary;

implementation

{$R *.dfm}

uses SiAuto,
     Math,
     Types,
     Clipbrd,
     ShellAPI,
     ShlObj,
     toolkit,
     definition_tcp_structures,
     dialog_configure_conversation_summary,
     class_interfacemapper,
     class_taskprocessor_extract,
     form_preferences,
     dialog_conversation_multifilesettings, dialog_goto;

const csvseparator = ',';
      TabMACEndpoints      = 0;
      TabIPv4Endpoints     = 1;
      TabIPv6Endpoints     = 2;
      TabMACConversations  = 3;
      TabIPv4Conversations = 4;
      TabIPv6Conversations = 5;
      TabTCPConversations  = 6;
      TabUDPConversations  = 7;
      TabNameResolutions   = 8;

      SharkFilterTypeAandB  = 1;
      SharkFilterTypeAtoB   = 2;
      SharkFilterTypeBtoA   = 3;
      SharkFilterTypeAtoAny = 4;
      SharkFilterTypeBtoAny = 5;

(* ------------------------------------------ *)

function BrowseDialog (const Title: string; const Flag: integer): string;

var lpItemID    : PItemIDList;
    BrowseInfo  : TBrowseInfo;
    DisplayName : array[0..MAX_PATH] of char;
    TempPath    : array[0..MAX_PATH] of char;

begin
     Result:='';
     FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);

     with BrowseInfo do
          begin
               hwndOwner := Application.Handle;
               pszDisplayName := @DisplayName;
               lpszTitle := PChar(Title);
               ulFlags := Flag;
          end;

  lpItemID := SHBrowseForFolder(BrowseInfo);

  if lpItemId <> nil then
     begin
          SHGetPathFromIDList(lpItemID, TempPath);
          Result := TempPath;
          GlobalFreePtr(lpItemID);
     end;
end;

(* ------------------------------------------- *)
{ tSelectedStats }
(* ------------------------------------------- *)

procedure tSelectedStats.initialize;

begin
     TotalFrames := 0;
     TotalBytes := 0;
end;

(* ------------------------------------------- *)
{ tHighlightTimeEdges }
(* ------------------------------------------- *)

procedure tHighlightTimeEdges.initialize;

begin
     FirstFrametime.initialize;
     LastFrametime.initialize;
end;

(* ------------------------------------------- *)

function tHighlightTimeEdges.isSet: boolean;

begin
     result := not ((FirstFrametime.datetime = 0) or (LastFrametime.datetime = 0));
end;

(* ------------------------------------------- *)
{ tTabValues }
(* ------------------------------------------- *)

procedure tTabValues.initialize;

begin
     currentfilterstr := '';
     currentSocketFilter.initialize;
     currentmarkedrow := 0;
     currentrow := 0;
     HighlightMode := hlmNone;
     AggregatedMode := false;
     MainTree := nil;

     rowindexlist := tDictionary<integer, PVirtualNode>.create;
     markedrowlist := tDictionary<integer, PVirtualNode>.create;
end;

(* ------------------------------------------- *)

procedure tTabValues.free;

begin
     rowindexlist.Free;
     markedrowlist.Free;
end;

(* ------------------------------------------- *)
{ TFormConversationSummary }
(* ------------------------------------------- *)

procedure TFormConversationSummary.onCreate (Sender: TObject);

var i: Integer;

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     PageControlConversations.ActivePageIndex := 0;

     VSTMACEndpoints.NodeDataSize := sizeOf (tMACEndpointTreeEntry);
     VSTIPv4Endpoints.NodeDataSize := sizeOf (tIPv4EndpointTreeEntry);
     VSTIPv6Endpoints.NodeDataSize := sizeOf (tIPv6EndpointTreeEntry);
     VSTMACConversations.NodeDataSize := sizeOf (tMACConversationTreeEntry);
     VSTIPv4Conversations.NodeDataSize := sizeOf (tIPv4ConversationTreeEntry);
     VSTIPv6Conversations.NodeDataSize := sizeOf (tIPv6ConversationTreeEntry);
     VSTTCPConversations.NodeDataSize := sizeOf (tTCPConversationTreeEntry);
     VSTUDPConversations.NodeDataSize := sizeOf (tUDPConversationTreeEntry);
{$ifdef fqdnstats}
     VSTNameResolution.NodeDataSize := sizeOf (tFQDNTreeNode);
{$endif}

     for i := 1 to tabcount do
         priv_tabvalues[i-1].initialize;

     priv_TabValues[TabMACEndpoints].MainTree := VSTMACEndpoints;
     priv_TabValues[TabIPv4Endpoints].MainTree := VSTIPv4Endpoints;
     priv_TabValues[TabIPv6Endpoints].MainTree := VSTIPv6Endpoints;
     priv_TabValues[TabMACConversations].MainTree := VSTMACConversations;
     priv_TabValues[TabIPv4Conversations].MainTree := VSTIPv4Conversations;
     priv_TabValues[TabIPv6Conversations].MainTree := VSTIPv6Conversations;
     priv_TabValues[TabTCPConversations].MainTree := VSTTCPConversations;
     priv_TabValues[TabUDPConversations].MainTree := VSTUDPConversations;
     priv_TabValues[TabNameResolutions].MainTree := VSTNameResolution;

     if Screen.DesktopWidth <= self.Width then
        begin
             // Projector fix - if the screen is small, do not autoresize, otherwise the last column may disappear

             for i := 1 to PageControlConversations.PageCount do
                 begin
                      priv_tabvalues[i-1].MainTree.Header.Options := priv_tabvalues[i-1].MainTree.Header.Options  - [hoAutoResize];
                      priv_tabvalues[i-1].MainTree.Header.Columns.Items[priv_tabvalues[i-1].MainTree.Header.AutosizeIndex].Width := 100;
                 end;
{
             VSTMACEndpoints.Header.Options := VSTMACEndpoints.Header.Options - [hoAutoResize];
             VSTMACEndpoints.Header.Columns.Items[VSTMACEndpoints.Header.AutoSizeIndex].Width := 100;
             VSTIPv4Endpoints.Header.Options := VSTIPv4Endpoints.Header.Options - [hoAutoResize];
             VSTIPv4Endpoints.Header.Columns.Items[VSTIPv4Endpoints.Header.AutoSizeIndex].Width := 100;
             VSTIPv6Endpoints.Header.Options := VSTIPv6Endpoints.Header.Options - [hoAutoResize];
             VSTIPv6Endpoints.Header.Columns.Items[VSTIPv6Endpoints.Header.AutoSizeIndex].Width := 100;
             VSTMACConversations.Header.Options := VSTMACConversations.Header.Options - [hoAutoResize];
             VSTMACConversations.Header.Columns.Items[VSTMACConversations.Header.AutoSizeIndex].Width := 100;
             VSTIPv4Conversations.Header.Options := VSTIPv4Conversations.Header.Options - [hoAutoResize];
             VSTIPv4Conversations.Header.Columns.Items[VSTIPv4Conversations.Header.AutoSizeIndex].Width := 100;
             VSTIPv6Conversations.Header.Options := VSTIPv6Conversations.Header.Options - [hoAutoResize];
             VSTIPv6Conversations.Header.Columns.Items[VSTIPv6Conversations.Header.AutoSizeIndex].Width := 100;
             VSTTCPConversations.Header.Options := VSTTCPConversations.Header.Options - [hoAutoResize];
             VSTTCPConversations.Header.Columns.Items[VSTTCPConversations.Header.AutoSizeIndex].Width := 100;
             VSTUDPConversations.Header.Options := VSTUDPConversations.Header.Options - [hoAutoResize];
             VSTUDPConversations.Header.Columns.Items[VSTUDPConversations.Header.AutoSizeIndex].Width := 100;
             VSTNameResolution.Header.Options := VSTNameResolution.Header.Options - [hoAutoResize];
             VSTNameResolution.Header.Columns.Items[VSTNameResolution.Header.AutoSizeIndex].Width := 100;
}
        end;

     priv_popupTree := nil;

     // create FQDNManager (initialized when assigning the ConversationStore property)

     priv_filterlist := tFilterSet.create;
{$ifdef fqdnstats}
     priv_FQDNManager := tFQDNManager.create (false);
{$endif}

     priv_loadcolorsfromdb := true;
     priv_firstshow := true;

     // Move LabelEdit to Statusbar

     Statusbar.Panels[3].Style := psOwnerDraw;
     LabeledEditFilter.Parent := Statusbar;

     setLength (priv_TCPList, 0);
     setLength (priv_NRList, 0);

     // Hide NameResolution tab for now except during debug

{$ifndef fqdnstats}
     TabsheetNameResolution.TabVisible := false;
{$endif}
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onDestroy (Sender: TObject);

var i : integer;

begin
     for i := 1 to tabcount do
         priv_tabvalues[i-1].free;

     priv_filterlist.Free;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.AssignNewConversationStore (ConversationStore: tConversationStorage);

begin
     SiMain.EnterMethod ('TFormConversationSummary.AssignNewConversationStore');

     priv_conversationstore := ConversationStore;

     // Empty the FQDN Manager and
{$ifdef fqdnstats}
     priv_conversationstore.InitFQDNManager (priv_FQDNManager);
{$endif}

     SiMain.LeaveMethod ('TFormConversationSummary.AssignNewConversationStore');
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.UpdateWindowCaption;

begin
     FormConversationSummary.Caption := 'Endpoint and Conversation Summary';

     if trim (TWWindowLabel) <> '' then
        FormConversationSummary.Caption := FormConversationSummary.Caption + ' [' + TWWindowLabel + ']';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onShow (Sender: TObject);

var i : integer;

begin
     SiMain.EnterMethod ('TFormConversationSummary.onShow');

     // Clear the filter list

//   setlength (priv_filterlist, 0);
     priv_filterlist.Clear;

     setLength (priv_interfacemaplist, 0);

     if priv_loadcolorsfromdb then
        begin
             if FormPreferences.HasTCPColorTable then
                priv_settings := FormPreferences.RetrieveConversationListSettings
             else
                 begin
                      priv_settings.TCPColors := FormConfigureConversationDisplay.ColorTable;
                      priv_settings.InitialPage := FormConfigureConversationDisplay.ComboBoxInitialPage.ItemIndex;
                      priv_settings.RememberPage := FormConfigureConversationDisplay.CheckBoxGoToPageEveryTime.Checked;
                      priv_settings.ClearFilterOnChange := FormConfigureConversationDisplay.CheckboxClearFilterOnPageChange.checked;
                 end;
        end;

     // remember that we tried, no matter if it worked or not

     priv_loadcolorsfromdb := false;

     // Apply display settings

     if priv_firstshow or priv_settings.RememberPage then
        begin
             PageControlConversations.ActivePageIndex := priv_settings.InitialPage;
             priv_firstshow := false;
        end;

     LabeledEditFilter.Clear;
     priv_CellTextCache := '';

     // Prepare the TCP row list, in standard mode (meaning, not aggregated by IP pair)

     priv_TCPList := priv_conversationstore.GetTCPConversationList (csmStandard);

{$ifdef fqdnstats}

     // Do the same for NameResolutions

     priv_NRList := priv_conversationstore.GetNameResolutionList;
{$endif}

     // Display the stuff

     TimerDisplayAll.Enabled := true;

     UpdateWindowCaption;

     SiMain.LeaveMethod ('TFormConversationSummary.onShow');
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onChangePage(Sender: TObject);

begin
     // Check if the current filter is different than what the new page is filtered for

     if (LabeledEditFilter.Text <> priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr) then
        begin
             // Clear the filter if the setting requires it, otherwise make sure the current filter is applied

             if priv_settings.ClearFilterOnChange then
                begin
                     priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr := '';
                     LabeledEditFilter.Clear;
                end
             else
                 priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr := LabeledEditFilter.Text;

             // Only update the list that is shown

             case PageControlConversations.ActivePageIndex of
              0 : DisplayMACEndpoints;
              1 : DisplayIPv4Endpoints;
              2 : DisplayIPv6Endpoints;
              3 : DisplayMACConversations;
              4 : DisplayIPv4Conversations;
              5 : DisplayIPv6Conversations;
              6 : DisplayTCPConversations;
              7 : DisplayUDPConversations;
             end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTGetHint (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; var LineBreakStyle: TVTTooltipLineBreakStyle; var HintText: string);

var tree     : tBaseVirtualTree;
    hintlist : tStringlist;
    scannode : pVirtualNode;
    macendp  : pMACEndpointTreeEntry;
    ipv4endp : pIPv4EndpointTreeEntry;
    ipv6endp : pIPv6EndpointTreeEntry;
    macdata  : pMACConversationTreeEntry;
    ipv4data : pIPv4ConversationTreeEntry;
    ipv6data : pIPv6ConversationTreeEntry;
    tcpdata  : pTCPConversationTreeEntry;
    udpdata  : pUDPConversationTreeEntry;
    maplist  : tInterfaceFileIDList;
    udpconv  : tUDPConversation;
//  udpconv4 : tUDPIPv4Conversation;
//  udpconv6 : tUDPIPv6Conversation;
    tcpconv  : tTCPConversation;
//  tcpconv4 : tTCPIPv4Conversation;
//  tcpconv6 : tTCPIPv6Conversation;
    entry    : tTraceFileEntry;
    i        : integer;

begin
     tree := (Sender as tBaseVirtualTree);

     hintlist := tStringlist.Create;

     if tree.Name = 'VSTMACEndpoints' then
        begin
             macendp := tree.GetNodeData (Node);

//           HintText := '';

             for i := macendp.HelpFirstSeen.FileIndex to macendp.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                      hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTIPv4Endpoints' then
        begin
             ipv4endp := tree.GetNodeData (Node);

//           HintText := '';

             for i := ipv4endp.HelpFirstSeen.FileIndex to ipv4endp.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                      hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTIPv6Endpoints' then
        begin
             ipv6endp := tree.GetNodeData (Node);

//           HintText := '';

             for i := ipv6endp.HelpFirstSeen.FileIndex to ipv6endp.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                      hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTMACConversations' then
        begin
             macdata := tree.GetNodeData (Node);

//           HintText := '';

             for i := macdata.HelpFirstSeen.FileIndex to macdata.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                      hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTIPv4Conversations' then
        begin
             ipv4data := tree.GetNodeData (Node);

//           HintText := '';

             for i := ipv4data.HelpFirstSeen.FileIndex to ipv4data.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                      hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTIPv6Conversations' then
        begin
             ipv6data := tree.GetNodeData (Node);

//           HintText := '';

             for i := ipv6data.HelpFirstSeen.FileIndex to ipv6data.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                       hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTUDPConversations' then
        begin
             udpdata := tree.GetNodeData (Node);

//           HintText := '';

             if udpdata.AddressA.IPVersion = 4 then
                begin
                     udpconv := priv_conversationstore.GetUDPConversationEntry (udpdata.GetIPv4SocketPair);

                     assert (udpconv.SocketPair.Source.IPAddress.IPVersion = 4, 'failed to get IPv4 based UDP conversation (TFormConversationSummary.onVSTGetHint)');

                     maplist := udpdata.InterfaceList;
                end
             else
                 begin
                      udpconv := priv_conversationstore.GetUDPConversationEntry (udpdata.GetIPv6SocketPair);

                      assert (udpconv.SocketPair.Source.IPAddress.IPVersion = 6, 'failed to get IPv6 based TCP conversation (TFormConversationSummary.onVSTGetHint)');

                      maplist := udpdata.InterfaceList;
                 end;

             for i := udpdata.HelpFirstSeen.FileIndex to udpdata.HelpLastSeen.FileIndex do
                 begin
                      entry := priv_filelist.getFile(i);
                       hintlist.Add (entry.FilePath + entry.FileName);
{
                      if i < tcpdata.HelpLastSeen.FileIndex then
                         HintText := HintText + entry.FilePath + entry.FileName + crlf
                      else
                          HintText := HintText + entry.FilePath + entry.FileName;
}
                 end;
        end;

     if tree.Name = 'VSTTCPConversations' then
        begin
             tcpdata := tree.GetNodeData (Node);

             // Determine the interface mapping list. With that we can tell in which file the conversation is
             // present. Sometimes it is not in all files between first and last.

             if not tcpdata^.IsParent then
                begin
                     if tcpdata.AddressA.IPVersion = 4 then
                        begin
                             tcpconv := priv_conversationstore.GetTCPConversationEntry (tcpdata.GetIPv4SocketPair);

                             assert (tcpconv.SocketPair.Source.IPAddress.IPVersion = 4, 'failed to get IPv4 based TCP conversation (TFormConversationSummary.onVSTGetHint)');

                             maplist := tcpconv.InterfaceList;
                        end
                     else
                         begin
                              tcpconv := priv_conversationstore.GetTCPConversationEntry (tcpdata.GetIPv6SocketPair);

                              assert (tcpconv.SocketPair.Source.IPAddress.IPVersion = 6, 'failed to get IPv6 based TCP conversation (TFormConversationSummary.onVSTGetHint)');

                              maplist := tcpconv.InterfaceList;
                         end;

                     for i := tcpdata.HelpFirstSeen.FileIndex to tcpdata.HelpLastSeen.FileIndex do
                         begin
                              // Check if the current file index is relevant, and if it is, put its name in the hint

                              if FileIndexExistsInList (maplist, i) then
                                 begin
                                      entry := priv_filelist.getFile(i);
                                      hintlist.Add (entry.FilePath + entry.FileName);
                                 end;
                         end;
                end;
        end;

     // Create the hint

     HintText := '';

     if hintlist.Count > 0 then
        begin
             for i := 1 to hintlist.Count - 1 do
                 HintText := HintText + hintlist[i-1] + crlf;

             HintText := HintText + hintlist[hintlist.Count - 1];
        end;

     hintlist.Free;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTGetPopupMenu (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; const P: TPoint; var AskParent: Boolean; var PopupMenu: TPopupMenu);

var macendp  : pMACEndpointTreeEntry;
    ipv4endp : pIPv4EndpointTreeEntry;
    ipv6endp : pIPv6EndpointTreeEntry;
    macdata  : pMACConversationTreeEntry;
    ipv4data : pIPv4ConversationTreeEntry;
    ipv6data : pIPv6ConversationTreeEntry;
    tcpdata  : pTCPConversationTreeEntry;
    udpdata  : pUDPConversationTreeEntry;
{$ifdef fqdnstats}
    nrdata   : pFQDNTreeNode;
{$endif}
    CellText : string;

begin
     priv_popuptree := Sender;
     priv_extractwritemode := mwmNone;

     // Set defaults

     MenuConversationCopyCellToClipboard.Visible := false;
     MenuConversationLookup.Visible := false;
     MenuConversationView.Visible := false;
     MenuConversationGoto.Visible := false;
     MenuConversationMarkRow.Visible := false;
//   MenuNameResolutionCopyCellToClipboard.Visible := false;

     case priv_MenuMode of
      cmmMainForm    : begin
                            if priv_tabvalues[TabTCPConversations].MarkedRowList.Count > 0 then
                               MenuConversationExportToCSV.Caption := 'Export ' + priv_tabvalues[TabTCPConversations].MarkedRowList.Count.ToString + ' marked lines to &CSV'
                            else
                                MenuConversationExportToCSV.Caption := 'Export all to &CSV';

                            // TCP is most important, so deal with it first

                            if Sender.Name = 'VSTTCPConversations' then
                               begin
                                    PopupMenu := PopupMenuConversationTreeView;

                                    if assigned (node) then
                                       begin
                                            tcpdata := Sender.GetNodeData (Node);

                                            CellText := GetTCPConversatioNCellText (Column, tcpdata^);
                                            MenuConversationCopyCellToClipboard.Caption := 'Copy "' + CellText + '" to clipboard';
                                            priv_CellTextCache := CellText;

                                            MenuConversationCopyCellToClipboard.Visible := true;

                                            if isValidIPv4String (CellText) or isValidIPv6String (CellText) then
                                               MenuConversationLookup.Visible := true;
                                       end;

                                    // Enable menus only available for TCP

                                    MenuConversationView.Visible := true;
                                    MenuConversationGoto.Visible := true;
                                    MenuConversationMarkRow.Visible := true;

//                                  priv_extractwritemode := mwmTCPConversation;
                               end
                            else
                            if Sender.Name = 'VSTIPv4Conversations' then
                               begin
                                    PopupMenu := PopupMenuConversationTreeView;

                                    if assigned (node) then
                                       begin
                                            ipv4data := Sender.GetNodeData (Node);

                                            CellText := GetIPv4ConversationCellText (Column, ipv4data^);
                                            MenuConversationCopyCellToClipboard.Caption := 'Copy "' + CellText + '" to clipboard';
                                            priv_CellTextCache := CellText;
                                       end;

                                    MenuConversationCopyCellToClipboard.Visible := true;

//                                  priv_extractwritemode := mwmIPv4Conversation;
                               end
                            else
                            if Sender.Name = 'VSTIPv6Conversations' then
                               begin
                                    PopupMenu := PopupMenuConversationTreeView;

                                    if assigned (node) then
                                       begin
                                            ipv6data := Sender.GetNodeData (Node);

                                            CellText := getIPv6ConversationCellText (Column, ipv6data^);
                                            MenuConversationCopyCellToClipboard.Caption := 'Copy "' + CellText + '" to clipboard';
                                            priv_CellTextCache := CellText;
                                       end;

                                    MenuConversationCopyCellToClipboard.Visible := true;

//                                  priv_extractwritemode := mwmIPv6Conversation;
                               end
                            else
                                begin
                                     // Check everything else

                                     if Sender.Name = 'VSTMACEndpoints' then
                                        PopupMenu := PopupMenuEndpointTreeView;

                                     if Sender.Name = 'VSTIPv4Endpoints' then
                                        PopupMenu := PopupMenuEndpointTreeView;

                                     if Sender.Name = 'VSTIPv6Endpoints' then
                                        PopupMenu := PopupMenuEndpointTreeView;

                                     if Sender.Name = 'VSTMACConversations' then
                                        begin
                                             PopupMenu := PopupMenuConversationTreeView;

                                             priv_extractwritemode := mwmEtherConversation;
                                        end;

                                     if Sender.Name = 'VSTUDPConversations' then
                                        PopupMenu := PopupMenuConversationTreeView;

                                     if Sender.Name = 'VSTNameResolution' then
                                        begin
{
                                              PopupMenu := PopupMenuNameResolution;

                                             if assigned (node) then
                                                begin
                                                     nrdata := Sender.GetNodeData (Node);

                                                     CellText := GetNameResolutionCellText (Column, nrdata^);
                                                     MenuNameResolutionCopyCellToClipboard.Caption := 'Copy "' + CellText + '" to clipboard';
                                                     priv_CellTextCache := CellText;

                                                     MenuNameResolutionCopyCellToClipboard.Visible := true;

                                                     if isValidIPv4String (CellText) or isValidIPv6String (CellText) then
                                                        MenuConversationLookup.Visible := true;
                                                end;
}
                                             PopupMenu := nil;
                                        end;
                                end;
                       end;
      cmmMergeForm   : begin
                            PopupMenu := PopupMenuMergeFilter;
                       end;
      cmmExtractForm : begin
                            // for now, this is the same as getting filters for a merge

                            PopupMenu := PopupMenuMergeFilter;
                       end;
     end;

     // Enable or disable menu items

     if assigned (PopupMenu) then
        begin
             if PopupMenu.Name = 'PopupMenuConversationTreeView' then
                begin
                     MenuConversationOpen.Enabled := (Sender.SelectedCount > 0);

                     MenuConversationConfigureDisplay.Enabled := (priv_menumode = cmmMainForm);
                     MenuConversationConfigureDisplay.Visible := MenuConversationConfigureDisplay.Enabled;

                     MenuCreateConversationSharkFilter.Enabled := (Sender.SelectedCount > 0);
                     MenuCreateConversationSharkFilterAandB.Enabled := MenuCreateConversationSharkFilter.Enabled;
                     MenuCreateConversationSharkFilterAtoB.Enabled := MenuCreateConversationSharkFilter.Enabled;
                     MenuCreateConversationSharkFilterBtoA.Enabled := MenuCreateConversationSharkFilter.Enabled;
                     MenuCreateConversationSharkFilterAtoAny.Enabled := MenuCreateConversationSharkFilter.Enabled;
                     MenuCreateConversationSharkFilterBtoAny.Enabled := MenuCreateConversationSharkFilter.Enabled;

                     MenuConversationMarkRow.Enabled := (Sender.SelectedCount > 0);
                     MenuConversationPreviousMark.Enabled := (priv_tabvalues[PageControlConversations.ActivePageIndex].MarkedRowList.Count > 0);
                     MenuConversationNextMark.Enabled := MenuConversationPreviousMark.Enabled;
                     MenuConversationClearAllMarks.Enabled := MenuConversationPreviousMark.Enabled;

                     MenuConversationExtract.Enabled := (Sender.SelectedCount > 0);
                     MenuConversationExtractToSingleFile.Enabled := MenuConversationExtract.Enabled;

                     MenuConversationExtractToMultipleFiles.Enabled := (Sender.SelectedCount > 1);

                     MenuConversationExportToCSV.Enabled := (Sender.TotalCount > 0);
                end;

             if PopupMenu.Name =  'PopupMenuEndpointTreeView' then
                begin
                     MenuEndpointConfigureDisplay.Enabled := (priv_menumode = cmmMainForm);
                     MenuEndpointConfigureDisplay.Visible := MenuEndpointConfigureDisplay.Enabled;

                     MenuCreateEndpointSharkFilter.Enabled := (Sender.SelectedCount > 0);

                     MenuEndpointExtract.Enabled := (Sender.SelectedCount > 0);
                     MenuEndpointExtractToSingleFile.Enabled := MenuEndpointExtract.Enabled;
                     MenuEndpointExtractToMultipleFiles.Enabled := MenuEndpointExtract.Enabled;
                end;
        end;
end;


(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTChange (Sender: TBaseVirtualTree; Node: PVirtualNode);

var tree     : tBaseVirtualTree;
    count    : integer;
    scannode : pVirtualNode;
    macendp  : pMACEndpointTreeEntry;
    ipv4endp : pIPv4EndpointTreeEntry;
    ipv6endp : pIPv6EndpointTreeEntry;
    macdata  : pMACConversationTreeEntry;
    ipv4data : pIPv4ConversationTreeEntry;
    ipv6data : pIPv6ConversationTreeEntry;
    tcpdata  : pTCPConversationTreeEntry;
    udpdata  : pUDPConversationTreeEntry;

begin
     tree := (Sender as tBaseVirtualTree);
     count := tree.SelectedCount;

     if count > 0 then
        begin
             priv_selectedstats.initialize;
             scannode := tree.GetFirstSelected (false);

             if tree.Name = 'VSTMACEndpoints' then
                begin
                     while scannode <> nil do
                           begin
                                macendp := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + macendp^.Frames;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + macendp^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := macendp^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTIPv4Endpoints' then
                begin
                     while scannode <> nil do
                           begin
                                ipv4endp := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + ipv4endp^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + ipv4endp^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := ipv4endp^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTIPv6Endpoints' then
                begin
                     while scannode <> nil do
                           begin
                                ipv6endp := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + ipv6endp^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + ipv6endp^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := ipv6endp^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTMACConversations' then
                begin
                     while scannode <> nil do
                           begin
                                macdata := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + macdata^.Frames;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + macdata^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := macdata^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTIPv4Conversations' then
                begin
                     while scannode <> nil do
                           begin
                                ipv4data := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + ipv4data^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + ipv4data^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := ipv4data^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTIPv6Conversations' then
                begin
                     while scannode <> nil do
                           begin
                                ipv6data := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + ipv6data^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + ipv6data^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := ipv6data^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTUDPConversations' then
                begin
                     while scannode <> nil do
                           begin
                                udpdata := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + udpdata^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + udpdata^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := udpdata^.index;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             if tree.Name = 'VSTTCPConversations' then
                begin
                     while scannode <> nil do
                           begin
                                tcpdata := tree.GetNodeData (scannode);

                                priv_selectedstats.TotalFrames := priv_selectedstats.TotalFrames + tcpdata^.Packets;
                                priv_selectedstats.TotalBytes := priv_selectedstats.TotalBytes + tcpdata^.Bytes;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].currentrow := tcpdata^.index;

                                priv_tabvalues[PageControlConversations.ActivePageIndex].HighlightTimeEdges.FirstFrameTime := tcpdata^.FirstTimestamp;
                                priv_tabvalues[PageControlConversations.ActivePageIndex].HighlightTimeEdges.LastFrameTime := tcpdata^.LastTimestamp;

                                // Repaint the tree in case the highlight pivot row changed

                                if priv_tabvalues[PageControlConversations.ActivePageIndex].HighlightMode <> hlmNone then
                                   tree.Repaint;

                                scannode := tree.GetNextSelected (scannode);
                           end;
                end;

             Statusbar.Panels[1].Text := 'Selected Rows: ' + IntToStr (count) + ' -  Total Frames: ' + PointText (priv_selectedstats.TotalFrames) + ' - Total Bytes: ' + PointText (priv_selectedstats.TotalBytes);
        end
     else
         Statusbar.Panels[1].Text := '';

     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ShowColumnStatus (Tree: TBaseVirtualTree);

var count : integer;
    i     : integer;

begin
     count := 0;

     for i := 1 to Tree.Header.Columns.Count do
         if not (coVisible in Tree.Header.Columns.Items[i-1].Options) then
            inc (count);

     if count = 0 then
        StatusBar.Panels[2].Text := ''
     else
         if count = 1 then
            StatusBar.Panels[2].Text := '1 hidden Column'
         else
             StatusBar.Panels[2].Text := IntToStr (count) + ' hidden Columns'
end;

(* ------------------------------------------- *)

function TFormConversationSummary.RTTtoStr (InitialRTT: uint64): string;

var value : string;
    nano  : uint64;
    micro : uint64;
    secs  : uint64;

begin
     if InitialRTT >= 1000000000 then
        begin
             // Woah, more than a second...

             secs := InitialRTT div 1000000000;
             nano := InitialRTT - (secs * 1000000000);
        end
     else
         begin
              secs := 0;
              nano := InitialRTT;
         end;

     // By default, we use microseconds since nobody cares about nanoseconds in RTT... yet.

     micro := nano DIV 1000;
     value := IntToStr (secs) + FormatSettings.DecimalSeparator + Format ('%.6d', [micro]);

     result := value;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.SetMode (MenuMode : tConversationMenuMode);

begin
     priv_menumode := MenuMode;
{
     case MenuMode of
      cmmMainForm    : begin
                            VSTMACEndpoints.PopupMenu := PopupMenuEndpointTreeView;
                            VSTIPv4Endpoints.PopupMenu := PopupMenuEndpointTreeView;
                            VSTIPv6Endpoints.PopupMenu := PopupMenuEndpointTreeView;
                            VSTIPv4Conversations.PopupMenu := PopupMenuConversationTreeView;
                            VSTIPv6Conversations.PopupMenu := PopupMenuConversationTreeView;
//                          VSTTCPConversations.PopupMenu := PopupMenuConversationTreeView;
                            VSTUDPConversations.PopupMenu := PopupMenuConversationTreeView;
                            VSTMACConversations.PopupMenu := PopupMenuConversationTreeView;
                       end;
      cmmMergeForm   : begin
                            VSTMACEndpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv4Endpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv6Endpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv4Conversations.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv6Conversations.PopupMenu := PopupMenuMergeFilter;
                            VSTTCPConversations.PopupMenu := PopupMenuMergeFilter;
                            VSTUDPConversations.PopupMenu := PopupMenuMergeFilter;
                            VSTMACConversations.PopupMenu := PopupMenuMergeFilter;
                       end;
      cmmExtractForm : begin
                            // for now, this is the same as getting filters for a merge

                            VSTMACEndpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv4Endpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv6Endpoints.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv4Conversations.PopupMenu := PopupMenuMergeFilter;
                            VSTIPv6Conversations.PopupMenu := PopupMenuMergeFilter;
                            VSTTCPConversations.PopupMenu := PopupMenuMergeFilter;
                            VSTUDPConversations.PopupMenu := PopupMenuMergeFilter;
                            VSTMACConversations.PopupMenu := PopupMenuMergeFilter;
                       end;
     end;
}
end;

(* ------------------------------------------- *)

function TFormConversationSummary.FilterBoxIsAMatch (FilterValue: string; const CompareTo: string): boolean;

var invert : boolean;
    match  : boolean;
    help   : string;

begin
     // Check if there is a not clause

     invert := false;

     if (pos ('not ', FilterValue) = 1) then
        begin
             invert := true;

             help := rightstring (FilterValue, 5);
        end
     else
         if (pos ('!', FilterValue) = 1) then
            begin
                 invert := true;

                 help := rightstring (FilterValue, 2);
            end
         else
             help := FilterValue;

     // Do the check

     match := pos (lowercase (help), lowercase (CompareTo)) > 0;

     if invert then
        result := not match
     else
         result := match;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTMACEndPointsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pMACEndpointTreeEntry;

begin
     data := VSTMACEndpoints.GetNodeData (node);

     case column of
      0 : CellText := intToStr (data^.index);
      1 : CellText := data^.Address.asString(':', true);
      2 : CellText := pointtext (data^.Frames);
      3 : CellText := pointtext (data^.Bytes);
      4 : CellText := pointtext (data^.FramesSent);
      5 : CellText := pointtext (data^.FramesReceived);
      6 : CellText := pointtext (data^.BytesSent);
      7 : CellText := pointtext (data^.BytesReceived);
      8 : CellText := GetAttributeDescription (data^.Attributes);
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv4EndpointsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pIPv4EndpointTreeEntry;

begin
     data := VSTIPv4Endpoints.GetNodeData (node);

     case column of
       0 : CellText := intToStr (data^.index);
       1 : CellText := data^.Address.ValueAsString;
       2 : CellText := pointtext (data^.Packets);
       3 : CellText := pointtext (data^.Bytes);
       4 : CellText := pointtext (data^.PacketsSent);
       5 : CellText := pointtext (data^.PacketsReceived);
       6 : CellText := pointtext (data^.BytesSent);
       7 : CellText := pointtext (data^.BytesReceived);
       8 : CellText := IntToStr (data^.PartnerCount);
       9 : CellText := IntToStr (data^.Flowcount);
      10 : CellText := GetAttributeDescription (data^.Attributes);
     end;
end;

(* ------------------------------------------- *)


procedure TFormConversationSummary.onVSTIPv6EndpointsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pIPv6EndpointTreeEntry;

begin
     data := VSTIPv6Endpoints.GetNodeData (node);

     case column of
       0 : CellText := intToStr (data^.index);
       1 : CellText := data^.Address.ValueAsString;
       2 : CellText := pointtext (data^.Packets);
       3 : CellText := pointtext (data^.Bytes);
       4 : CellText := pointtext (data^.PacketsSent);
       5 : CellText := pointtext (data^.PacketsReceived);
       6 : CellText := pointtext (data^.BytesSent);
       7 : CellText := pointtext (data^.BytesReceived);
       8 : CellText := IntToStr (data^.PartnerCount);
       9 : CellText := IntToStr (data^.Flowcount);
      10 : CellText := GetAttributeDescription (data^.Attributes);
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTMACConversationsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pMACConversationTreeEntry;

begin
     data := VSTMACConversations.GetNodeData (node);

     case column of
       0 : CellText := intToStr (data^.index);
       1 : CellText := data^.AddressA.asString(':', true);
       2 : CellText := data^.AddressB.asString(':', true);
       3 : CellText := pointtext (data^.Frames);
       4 : CellText := pointtext (data^.Bytes);
       5 : CellText := pointtext (data^.FramesAtoB);
       6 : CellText := pointtext (data^.FramesBtoA);
       7 : CellText := pointtext (data^.BytesAtoB);
       8 : CellText := pointtext (data^.BytesBtoA);
       9 : CellText := data^.DurationString;
      10 : CellText := FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2);
      11 : CellText := FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2);
      12 : CellText := GetAttributeDescription (data^.Attributes);
     end;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.getIPv4ConversationCellText (const column: integer; const Data: tIPv4ConversationTreeEntry): string;

var CellText : string;

begin
     case column of
       0 : CellText := intToStr (data.index);
       1 : CellText := data.AddressA.ValueAsString;
       2 : CellText := data.AddressB.ValueAsString;
       3 : CellText := pointtext (data.Packets);
       4 : CellText := pointtext (data.Bytes);
       5 : CellText := pointtext (data.PacketsAtoB);
       6 : CellText := pointtext (data.PacketsBtoA);
       7 : CellText := pointtext (data.BytesAtoB);
       8 : CellText := pointtext (data.BytesBtoA);
       9 : CellText := data.DurationString;
      10 : CellText := FloatToStrF (data.MBpSAtoB, ffNumber, 15, 2);
      11 : CellText := FloatToStrF (data.MBpSBtoA, ffNumber, 15, 2);
      12 : begin
                if data.AverageRTTns > 0 then
                   CellText := RTTtoStr (data.AverageRTTns)
                else
                    CellText := '';
           end;
      13 : CellText := IntToStr (data.FlowCount);
      14 : CellText := GetAttributeDescription (data.Attributes);
     end;

     result := CellText;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onIPv4ConversationsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pIPv4ConversationTreeEntry;

begin
     data := VSTIPv4Conversations.GetNodeData (node);

     CellText := getIPv4ConversationCellText (Column, data^);
end;

(* ------------------------------------------- *)

function TFormConversationSummary.getIPv6ConversationCellText (const column: integer; const Data: tIPv6ConversationTreeEntry): string;

var CellText : string;

begin
     case column of
       0 : CellText := intToStr (data.index);
       1 : CellText := data.AddressA.ValueAsString;
       2 : CellText := data.AddressB.ValueAsString;
       3 : CellText := pointtext (data.Packets);
       4 : CellText := pointtext (data.Bytes);
       5 : CellText := pointtext (data.PacketsAtoB);
       6 : CellText := pointtext (data.PacketsBtoA);
       7 : CellText := pointtext (data.BytesAtoB);
       8 : CellText := pointtext (data.BytesBtoA);
       9 : CellText := data.DurationString;
      10 : CellText := FloatToStrF (data.MBpSAtoB, ffNumber, 15, 2);
      11 : CellText := FloatToStrF (data.MBpSBtoA, ffNumber, 15, 2);
      12 : begin
                if data.AverageRTTns > 0 then
                   CellText := RTTtoStr (data.AverageRTTns)
                else
                    CellText := '';
           end;
      13 : CellText := IntToStr (data.FlowCount);
      14 : CellText := GetAttributeDescription (data.Attributes);
     end;

     result := CellText;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv6ConversationsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pIPv6ConversationTreeEntry;

begin
     data := VSTIPv6Conversations.GetNodeData (node);

     CellText := getIPv6ConversationCellText (Column, data^);
end;

(* ------------------------------------------- *)
{$ifdef fqdnstats}
function TFormConversationSummary.GetNameResolutionCellText (Const column : integer; const Data : tFQDNTreeNode) : string;

var CellText : string;

begin

     case Column of
       0 : CellText := data.name;
       1 : CellText := data.framecount.ToString;
       2 : CellText := data.querycount.ToString;
       3 : CellText := data.answercount.ToString;
       4 : CellText := data.SuccessfullQueries.ToString;
       5 : CellText := data.FailedQueries.ToString;
       6 : CellText := data.FormatErrors.ToString;
       7 : CellText := data.NXDomain.ToString;
       8 : CellText := data.NotImplemented.ToString;
       9 : CellText := data.QueryRefused.ToString;
      10 : CellText := data.NotAuthorized.ToString;
     end;

     result := CellText;
end;
{$endif}

(* ------------------------------------------- *)
{$ifdef fqdnstats}
procedure TFormConversationSummary.onVSTNameResolutionGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pFQDNTreeNode;

begin
     data := VSTNameResolution.GetNodeData (Node);

     CellText := GetNameResolutionCellText (Column, data^);
end;
{$else}
procedure TFormConversationSummary.onVSTNameResolutionGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

begin

end;
{$endif}
(* ------------------------------------------- *)

procedure TFormConversationSummary.onTaskExtractAdvancePercentProcessed (Sender : tObject; FileEntry: tTracefileEntry; const PercentageParsed: integer);

begin
     // Due to currently inexact calculation of total frames to process the percentage can exceed 100%... so let's just hide that little problem :-)

     if Percentageparsed <= 100 then
        Statusbar.Panels[0].Text := 'Status: Extracting frames (' + IntToStr (PercentageParsed) + '% complete)'
     else
         Statusbar.Panels[0].Text := 'Status: Extracting frames (100% complete)';

     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onTaskExtractNextFile (Sender: tObject; const FileIndex : integer; const FileEntry: tTracefileEntry; var SkipFile : boolean);

begin
     priv_lastsourcefilename := FileEntry.FullName;

     Statusbar.Panels[1].Text := FileEntry.FileName;
     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.getTCPConversationCellText (const column: integer; const Data: tTCPConversationTreeEntry) : string;

var CellText : string;

begin
     if data.IsParent then
        begin
         case column of
           1 : CellText := data.AddressA.ValueAsString;
           3 : CellText := data.AddressB.ValueAsString;
         else
             CellText := '';
         end;
        end
     else
         case column of
           0 : CellText := intToStr (data.index);
           1 : CellText := data.AddressA.ValueAsString;
           2 : CellText := intToStr (data.PortA);
           3 : CellText := data.AddressB.ValueAsString;
           4 : CellText := intToStr (data.PortB);
           5 : CellText := pointtext (data.Packets);
           6 : CellText := pointtext (data.Bytes);
           7 : CellText := pointtext (data.PacketsAtoB);
           8 : CellText := pointtext (data.PacketsBtoA);
           9 : CellText := pointtext (data.BytesAtoB);
          10 : CellText := pointtext (data.BytesBtoA);
          11 : CellText := data.PacketRatioAtoB.ToString + '% / ' + (100 - data.PacketRatioAtoB).ToString + '%';
          12 : begin
                    if data.InitialRTTns > 0 then
                       CellText := RTTtoStr (data.InitialRTTns)
                    else
                        CellText := '';
               end;
          13 : CellText := data.DurationString;
          14 : CellText := FloatToStrF (data.MBpSAtoB, ffNumber, 15, 2);
          15 : CellText := FloatToStrF (data.MBpSBtoA, ffNumber, 15, 2);
          16 : CellText := TCPFlagSetToTCPDump (data.flags);
          17 : CellText := DateTimeToStr (data.FirstTimestamp.datetime);
          18 : CellText := GetTCPStatusDescription (data.SessionStatus, data.Attributes);
         end;

     result := CellText;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onTCPConversationsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pTCPConversationTreeEntry;

begin
     data := VSTTCPConversations.GetNodeData (node);

     CellText := getTCPConversationCellText (Column, data^);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTUDPConversationsGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var data : pUDPConversationTreeEntry;

begin
     data := VSTUDPConversations.GetNodeData (node);

     case column of
       0 : CellText := intToStr (data^.index);
       1 : begin
                if data^.AddressA.IPVersion = 4 then
                   CellText := data^.AddressA.ValueAsString
                else
                    CellText := data^.AddressA.ValueAsString;
           end;
       2 : CellText := intToStr (data^.PortA);
       3 : begin
                if data^.AddressA.IPVersion = 4 then
                   CellText := data^.AddressB.ValueAsString
                else
                    CellText := data^.AddressB.ValueAsString
           end;
       4 : CellText := intToStr (data^.PortB);
       5 : CellText := pointtext (data^.Packets);
       6 : CellText := pointtext (data^.Bytes);
       7 : CellText := pointtext (data^.PacketsAtoB);
       8 : CellText := pointtext (data^.PacketsBtoA);
       9 : CellText := pointtext (data^.BytesAtoB);
      10 : CellText := pointtext (data^.BytesBtoA);
      11 : CellText := pointtext (data^.PacketsPerSecondAtoB);
      12 : CellText := pointtext (data^.PacketsPerSecondAtoB);
      13 : CellText := FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2);
      14 : CellText := FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2);
      15 : begin
                if data^.Packets > 1 then
                   CellText := data^.DurationString
                else
                    CellText := '';
           end;
      16 : begin
                if (data^.PacketsAtoB = 1) and (data^.PacketsBtoA = 1) then
                   CellText := FloatToStrF (data^.DeltaTime.toFloat, ffNumber, 15, 9)
                else
                    CellText := '';
           end;
      17 : CellText := GetAttributeDescription (data^.Attributes);
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ProcessTCPNodes (ProcessAction: tNodeProcessAction);

var Node     : PVirtualNode;
    data     : pTCPConversationTreeEntry;
    helpnode : PVirtualNode;
    helpdata : pTCPConversationTreeEntry;

begin
     Node := VSTTCPConversations.GetFirstSelected;

     while assigned (Node) do
           begin
                data := VSTTCPConversations.GetNodeData (node);

                case ProcessAction of
                 paMarkRows   : begin
                                     if data^.Marked then
                                        begin
                                             // Set the display item to unmarked

                                             data^.Marked := False;
                                             priv_TCPList[data^.index-1].Marked := False;

                                             if priv_tabvalues[TabTCPConversations].markedrowlist.ContainsKey (data^.index) then
                                                priv_tabvalues[TabTCPConversations].markedrowlist.Remove (data^.index);

                                             // clear the last mark to avoid trouble

                                             priv_tabvalues[TabTCPConversations].currentmarkedrow := 0;
                                        end
                                     else
                                         begin
                                              // Set the display item to marked

                                              data^.Marked := true;
                                              priv_TCPList[data^.index-1].Marked := true;

                                              if not priv_tabvalues[TabTCPConversations].markedrowlist.ContainsKey (data^.index) then
                                                priv_tabvalues[TabTCPConversations].markedrowlist.Add (data^.index, Node);

                                              priv_tabvalues[TabTCPConversations].currentmarkedrow := data^.index;
                                         end;
                                end;
                end;

                // Repaint

                VSTTCPConversations.RepaintNode (Node);

                // continue with next selected node

                Node := VSTTCPConversations.GetNextSelected (Node);
           end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.GotoRow;

var node     : pVirtualNode;
    TabIndex : integer;
    row      : integer;

begin
     FormGoToDialog.ShowModal;

     if FormGoToDialog.ModalResult = mrOK then
        begin
             TryStrToInt (FormGoToDialog.LabeledEditGoTo.Text, row);

             if row > 0 then
                begin
                     TabIndex := PageControlConversations.ActivePageIndex;

                     if priv_tabvalues[TabIndex].rowindexlist.ContainsKey (row) then
                        begin
                             node := priv_tabvalues[TabIndex].rowindexlist.Items[row];

                             case TabIndex of
                              TabTCPConversations : VSTTCPConversations.ScrollIntoView (node, true);
                             end;
                        end;
                end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ClearAllTCPMarks;

var Node     : PVirtualNode;
    data     : pTCPConversationTreeEntry;
    index    : integer;

begin
     for index in priv_tabvalues[TabTCPConversations].markedrowlist.Keys do
         begin
              Node := priv_tabvalues[TabTCPConversations].markedrowlist.Items[index];

              data := VSTTCPConversations.GetNodeData (node);
              data^.Marked := False;

              // Repaint

              VSTTCPConversations.RepaintNode (Node);
         end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.JumpToTCPMark  (const mode : tMarkJumpMode);

var HelpArray    : TArray<integer>;
    Node         : PVirtualNode;
    data         : pTCPConversationTreeEntry;
    index        : integer;
    current      : integer;
    count        : integer;
    len          : integer;
    marklist     : array of PVirtualNode;

begin
     len := priv_tabvalues[TabTCPConversations].markedrowlist.Count;
     Node := nil;

     if len > 0 then
        begin
             // Use a TArray<integer> to get sorted access to the dictionary

             HelpArray := priv_tabvalues[TabTCPConversations].markedrowlist.Keys.ToArray;
             TArray.Sort<integer>(HelpArray);

             current := 0;
             count := 0;

             // Find the current marked row

             setLength (marklist, len);

             for index in HelpArray do
                 begin
                      inc (count);

                      marklist[count-1] := priv_tabvalues[TabTCPConversations].markedrowlist.Items[index];

                      if index = priv_tabvalues[TabTCPConversations].currentmarkedrow then
                         current := count;
                 end;

             // If we have a current marked row, go up or down to the next/previous

             if current > 0 then
                begin
                     case mode of
                      jmNext     : begin
                                        inc (current);

                                        if current > len then
                                           current := len;
                                   end;
                      jmPrevious : begin
                                        dec (current);

                                        if current < 1 then
                                           current := 1;
                                   end;
                     end;
                end
             else
                 current := 1;

             Node := marklist[current-1];
             data := VSTTCPConversations.GetNodeData (node);
             priv_tabvalues[TabTCPConversations].currentmarkedrow := Data^.index;

             if assigned (Node) then
                VSTTCPConversations.ScrollIntoView (Node, true);
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ToggleAggregatedMode;

begin
     priv_tabvalues[TabTCPConversations].AggregatedMode := not priv_tabvalues[TabTCPConversations].AggregatedMode;

     if priv_tabvalues[TabTCPConversations].AggregatedMode then
        begin
             setLength (priv_TCPlist, 0);

             priv_TCPList := priv_conversationstore.GetTCPConversationList (csmIPPair);

             DisplayTCPConversations;
        end
     else
         begin
              setLength (priv_TCPlist, 0);

              priv_TCPList := priv_conversationstore.GetTCPConversationList (csmStandard);

              DisplayTCPConversations;
         end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ToggleHighlightMode;

begin
     if (priv_tabvalues[TabTCPConversations].HighlightMode = hlmNone) then
        priv_tabvalues[TabTCPConversations].HighlightMode := hlmTimeEmbedded
     else
         priv_tabvalues[TabTCPConversations].HighlightMode := hlmNone;

     VSTTCPConversations.Repaint;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuNameResolutionAggregatedModeClick (Sender: TObject);

begin
     priv_tabvalues[TabNameResolutions].AggregatedMode := not priv_tabvalues[TabNameResolutions].AggregatedMode;

{$ifdef fqdnstats}
     DisplayNameResolutions;
{$endif}
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTKeyUp (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
{$ifdef debug}
     StatusBar.Panels[0].Text := 'Key: ' + IntToStr (Key);
{$endif}
     case key of
       66 : begin
                 // Ctrl + Shift - B

                 if (Shift = [ssCtrl,ssShift]) and (priv_tabvalues[TabTCPConversations].markedrowlist.count > 0) then
                    begin
                         JumpToTCPMark (jmPrevious);
                         Key := 0;
                    end;
            end;
       71 : begin
                 // STRG-G - Goto Row

                 if (Shift = [ssCtrl]) then
                    begin
                         GotoRow;
                         Key := 0;
                    end;
            end;
       72 : begin
                 // STRG-H - Highlight Mode toggle

                 if (Shift = [ssCtrl]) then
                    begin
                         ToggleHighlightMode;
                         Key := 0;
                    end;
            end;
       77 : begin
                 // STRG-M - Mark Row(s) toggle

                 if (Shift = [ssCtrl]) then
                    begin
                         ProcessTCPNodes (paMarkRows);
                         Key := 0;
                    end;

                 // Clear all marks

                 if (Shift = [ssCtrl, ssAlt]) then
                    begin
                         ClearAllTCPMarks;
                         Key := 0;
                    end;
            end;
       78 : begin
                 // Ctrl + Shift - N

                 if (Shift = [ssCtrl,ssShift]) and (priv_tabvalues[TabTCPConversations].markedrowlist.Count > 0) then
                    begin
                         JumpToTCPMark (jmNext);
                         Key := 0;
                    end;
            end;
       79 : begin
                 // Ctrl + O

                 if (Shift = [ssCtrl]) then
                    begin
                         // Toggle aggregated mode

                         ToggleAggregatedMode;
                         Key := 0;
                    end;
            end;
       83 : begin
                 if (Shift = [ssCtrl]) then
                    begin
                         // Copy filter in Wireshark syntax to clipboard

                         CopySharkFilterToClipboard (priv_tabvalues[PageControlConversations.ActivePageIndex].MainTree, SharkFilterTypeAandB);
                         Key := 0;
                    end;
            end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTTCPConversationsBeforeCellPaint (Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; CellPaintMode: TVTCellPaintMode; CellRect: TRect; var ContentRect: TRect);

var data : pTCPConversationTreeEntry;

begin
     data := VSTTCPConversations.GetNodeData (node);

     if data^.Marked then
        begin
             TargetCanvas.Brush.Color := clBlack;
             TargetCanvas.FillRect(CellRect);
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTTCPConnectionsPaintText (Sender: TBaseVirtualTree; const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType);

var data     : pTCPConversationTreeEntry;
    Embedded : boolean;

begin
     data := VSTTCPConversations.GetNodeData (node);

     case data^.SessionStatus of
      TCPHandshakeComplete,
      TCPHandshakeAndTeardown,
      TCPTeardown             : begin
                                     if (caTCPSYNReset in data^.Attributes) then
                                        TargetCanvas.Font.Color := priv_settings.TCPColors[TCPHandshakeSYNReset]
                                     else
                                         if (caTCPMultiSYNRetry in data^.Attributes) then
                                            TargetCanvas.Font.Color := priv_settings.TCPColors[TCPHandshakeHalfOpen]
                                         else
                                             TargetCanvas.Font.Color := priv_settings.TCPColors[data^.SessionStatus];
                                end
     else
         TargetCanvas.Font.Color := priv_settings.TCPColors[data^.SessionStatus];
     end;

     // Invert black if marked (because of black background)

     if data^.Marked and (TargetCanvas.Font.Color = clWindowText) then
        TargetCanvas.Font.Color := clWhite;

     // Highlight Mode, but only if just one row is selected and times are set

     if (priv_tabvalues[TabTCPConversations].HighlightMode = hlmTimeEmbedded) and (priv_tabvalues[TabTCPConversations].HighlightTimeEdges.isSet) and (VSTTCPConversations.SelectedCount = 1) then
        begin
             Embedded := (data^.FirstTimestamp > priv_tabvalues[TabTCPConversations].HighlightTimeEdges.FirstFrameTime) and (data^.LastTimestamp < priv_tabvalues[TabTCPConversations].HighlightTimeEdges.LastFrameTime);

             // gray out the row if it is not starting after the selected row and ending before it

             if not Embedded then
                TargetCanvas.Font.Color := clSilver;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTMACEndpointsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pMACEndpointTreeEntry;
    data2 : pMACEndpointTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTMACEndpoints.GetNodeData (node1);
     data2 := VSTMACEndpoints.GetNodeData (node2);

     case column of
      0 : Result := compareValue (data1^.index, data2^.index);
      1 : Result := compareValue (data1^.Address, data2^.Address);
      2 : Result := compareValue (data1^.Frames, data2^.Frames);
      3 : Result := compareValue (data1^.Bytes, data2^.Bytes);
      4 : Result := compareValue (data1^.FramesSent, data2^.FramesSent);
      5 : Result := compareValue (data1^.FramesReceived, data2^.FramesReceived);
      6 : Result := compareValue (data1^.BytesSent, data2^.BytesSent);
      7 : Result := compareValue (data1^.BytesReceived, data2^.BytesReceived);
      8 : begin
               help1 := GetAttributeDescription (data1^.Attributes);
               help2 := GetAttributeDescription (data2^.Attributes);

               Result := compareText (help1, help2);
          end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv4EndpointsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pIPv4EndpointTreeEntry;
    data2 : pIPv4EndpointTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTIPv4Endpoints.GetNodeData (node1);
     data2 := VSTIPv4Endpoints.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := compareValue (data1^.Address, data2^.Address);
       2 : Result := compareValue (data1^.Packets, data2^.Packets);
       3 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       4 : Result := compareValue (data1^.PacketsSent, data2^.PacketsSent);
       5 : Result := compareValue (data1^.PacketsReceived, data2^.PacketsReceived);
       6 : Result := compareValue (data1^.BytesSent, data2^.BytesSent);
       7 : Result := compareValue (data1^.BytesReceived, data2^.BytesReceived);
       8 : Result := CompareValue (data1^.PartnerCount, data2^.PartnerCount);
       9 : Result := CompareValue (data1^.Flowcount, data2^.Flowcount);
      10 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv6EndpointsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pIPv6EndpointTreeEntry;
    data2 : pIPv6EndpointTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTIPv6Endpoints.GetNodeData (node1);
     data2 := VSTIPv6Endpoints.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := compareValue (data1^.Address, data2^.Address);
       2 : Result := compareValue (data1^.Packets, data2^.Packets);
       3 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       4 : Result := compareValue (data1^.PacketsSent, data2^.PacketsSent);
       5 : Result := compareValue (data1^.PacketsReceived, data2^.PacketsReceived);
       6 : Result := compareValue (data1^.BytesSent, data2^.BytesSent);
       7 : Result := compareValue (data1^.BytesReceived, data2^.BytesReceived);
       8 : Result := CompareValue (data1^.PartnerCount, data2^.PartnerCount);
       9 : Result := CompareValue (data1^.Flowcount, data2^.Flowcount);
      10 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTMACConversationsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pMACConversationTreeEntry;
    data2 : pMACConversationTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTMACConversations.GetNodeData (node1);
     data2 := VSTMACConversations.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := compareValue (data1^.AddressA, data2^.AddressA);
       2 : Result := compareValue (data1^.AddressB, data2^.AddressB);
       3 : Result := compareValue (data1^.Frames, data2^.Frames);
       4 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       5 : Result := compareValue (data1^.FramesAtoB, data2^.FramesAtoB);
       6 : Result := compareValue (data1^.FramesBtoA, data2^.FramesBtoA);
       7 : Result := compareValue (data1^.BytesAtoB, data2^.BytesAtoB);
       8 : Result := compareValue (data1^.BytesBtoA, data2^.BytesBtoA);
       9 : Result := compareText (data1^.DurationString, data2^.DurationString);
      10 : Result := compareValue (data1^.MBpSAtoB, data2^.MBpSAtoB);
      11 : Result := compareValue (data1^.MBpSBtoA, data2^.MBpSBtoA);
      12 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv4ConversationsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pIPv4ConversationTreeEntry;
    data2 : pIPv4ConversationTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTIPv4Conversations.GetNodeData (node1);
     data2 := VSTIPv4Conversations.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := compareValue (data1^.AddressA, data2^.AddressA);
       2 : Result := compareValue (data1^.AddressB, data2^.AddressB);
       3 : Result := compareValue (data1^.Packets, data2^.Packets);
       4 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       5 : Result := compareValue (data1^.PacketsAtoB, data2^.PacketsAtoB);
       6 : Result := compareValue (data1^.PacketsBtoA, data2^.PacketsBtoA);
       7 : Result := compareValue (data1^.BytesAtoB, data2^.BytesAtoB);
       8 : Result := compareValue (data1^.BytesBtoA, data2^.BytesBtoA);
       9 : Result := compareText (data1^.DurationString, data2^.DurationString);
      10 : Result := compareValue (data1^.MBpSAtoB, data2^.MBpSAtoB);
      11 : Result := compareValue (data1^.MBpSBtoA, data2^.MBpSBtoA);
      12 : Result := CompareValue (data1^.AverageRTTns, data2^.AverageRTTns);
      13 : Result := CompareValue (data1^.FlowCount, data2^.FlowCount);
      14 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTIPv6ConversationsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pIPv6ConversationTreeEntry;
    data2 : pIPv6ConversationTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTIPv6Conversations.GetNodeData (node1);
     data2 := VSTIPv6Conversations.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := compareValue (data1^.AddressA, data2^.AddressA);
       2 : Result := compareValue (data1^.AddressB, data2^.AddressB);
       3 : Result := compareValue (data1^.Packets, data2^.Packets);
       4 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       5 : Result := compareValue (data1^.PacketsAtoB, data2^.PacketsAtoB);
       6 : Result := compareValue (data1^.PacketsBtoA, data2^.PacketsBtoA);
       7 : Result := compareValue (data1^.BytesAtoB, data2^.BytesAtoB);
       8 : Result := compareValue (data1^.BytesBtoA, data2^.BytesBtoA);
       9 : Result := compareText (data1^.DurationString, data2^.DurationString);
      10 : Result := compareValue (data1^.MBpSAtoB, data2^.MBpSAtoB);
      11 : Result := compareValue (data1^.MBpSBtoA, data2^.MBpSBtoA);
      12 : Result := CompareValue (data1^.AverageRTTns, data2^.AverageRTTns);
      13 : Result := CompareValue (data1^.FlowCount, data2^.FlowCount);
      14 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTTCPConversationsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pTCPConversationTreeEntry;
    data2 : pTCPConversationTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTTCPConversations.GetNodeData (node1);
     data2 := VSTTCPConversations.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := CompareValue (data1^.AddressA, data2^.AddressA);
       2 : Result := CompareValue (data1^.PortA, data2^.PortA);
       3 : Result := CompareValue (data1^.AddressB, data2^.AddressB);
       4 : Result := compareValue (data1^.PortB, data2^.PortB);
       5 : Result := compareValue (data1^.Packets, data2^.Packets);
       6 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       7 : Result := compareValue (data1^.PacketsAtoB, data2^.PacketsAtoB);
       8 : Result := compareValue (data1^.PacketsBtoA, data2^.PacketsBtoA);
       9 : Result := compareValue (data1^.BytesAtoB, data2^.BytesAtoB);
      10 : Result := compareValue (data1^.BytesBtoA, data2^.BytesBtoA);
      11 : Result := compareValue (data1^.PacketRatioAtoB, data2^.PacketRatioAtoB);
      12 : Result := compareValue (data1^.InitialRTTns, data2^.InitialRTTns);
      13 : Result := compareText (data1^.DurationString, data2^.DurationString);
      14 : Result := compareValue (data1^.MBpSAtoB, data2^.MBpSAtoB);
      15 : Result := compareValue (data1^.MBpSBtoA, data2^.MBpSBtoA);
      16 : Result := CompareText (TCPFlagSetToTCPDump (data1^.Flags), TCPFlagSetToTCPDump (data2^.Flags));
      17 : Result := compareText (DateTimeToStr (data1^.FirstTimestamp.datetime), DateTimeToStr (data2^.FirstTimestamp.datetime));
      18 : begin
                help1 := GetTCPStatusDescription (data1^.SessionStatus, data1^.Attributes);
                help2 := GetTCPStatusDescription (data2^.SessionStatus, data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTUDPConversationsCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pUDPConversationTreeEntry;
    data2 : pUDPConversationTreeEntry;
    help1 : string;
    help2 : string;

begin
     data1 := VSTUDPConversations.GetNodeData (node1);
     data2 := VSTUDPConversations.GetNodeData (node2);

     case column of
       0 : Result := compareValue (data1^.index, data2^.index);
       1 : Result := CompareValue (data1^.AddressA, data2^.AddressA);
       2 : Result := CompareValue (data1^.PortA, data2^.PortA);
       3 : Result := CompareValue (data1^.AddressB, data2^.AddressB);
       4 : Result := compareValue (data1^.PortB, data2^.PortB);
       5 : Result := compareValue (data1^.Packets, data2^.Packets);
       6 : Result := compareValue (data1^.Bytes, data2^.Bytes);
       7 : Result := compareValue (data1^.PacketsAtoB, data2^.PacketsAtoB);
       8 : Result := compareValue (data1^.PacketsBtoA, data2^.PacketsBtoA);
       9 : Result := compareValue (data1^.BytesAtoB, data2^.BytesAtoB);
      10 : Result := compareValue (data1^.BytesBtoA, data2^.BytesBtoA);
      11 : Result := compareValue (data1^.PacketsPerSecondAtoB, data2^.PacketsPerSecondAtoB);
      12 : Result := compareValue (data1^.PacketsPerSecondBtoA, data2^.PacketsPerSecondBtoA);
      13 : Result := compareValue (data1^.MBpSAtoB, data2^.MBpSAtoB);
      14 : Result := compareValue (data1^.MBpSBtoA, data2^.MBpSBtoA);
      15 : begin
                if data1^.Packets > 1 then
                   begin
                        if data2^.Packets > 1 then
                           Result := compareText (data1^.DurationString, data2^.DurationString)
                        else
                            Result := GreaterThanValue;
                   end
                else
                    begin
                         if data2^.Packets > 1 then
                            Result := LessThanValue
                         else
                             Result := EqualsValue;
                    end;
           end;
      16 : begin
                // Check if both have a delta time. If not, the one with a delta time wins, and if both don't have one, they're equal.

                if (data1^.PacketsAtoB = 1) and (data1^.PacketsBtoA = 1) then
                   begin
                        if (data2^.PacketsAtoB = 1) and (data2^.PacketsBtoA = 1) then
                           Result := compareValue (data1^.DeltaTime.NanoTicks, data2^.DeltaTime.NanoTicks)
                        else
                            result := GreaterThanValue;
                   end
                else
                    begin
                        if (data2^.PacketsAtoB = 1) and (data2^.PacketsBtoA = 1) then
                           Result := LessThanValue
                        else
                            result := EqualsValue;
                    end;
           end;
      17 : begin
                help1 := GetAttributeDescription (data1^.Attributes);
                help2 := GetAttributeDescription (data2^.Attributes);

                Result := compareText (help1, help2);
           end;
     end;
end;

(* ------------------------------------------- *)
{$ifdef fqdnstats}
procedure TFormConversationSummary.onVSTNameResolutionCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pFQDNTreeNode;
    data2 : pFQDNTreeNode;

begin
     data1 := VSTNameResolution.GetNodeData (Node1);
     data2 := VSTNameResolution.GetNodeData (Node2);

     case Column of
       0 : Result := compareText (data1^.name, data2^.name);
       1 : Result := compareValue (data1^.framecount, data2^.framecount);
       2 : Result := compareValue (data1^.querycount, data2^.querycount);
       3 : Result := compareValue (data1^.answercount, data2^.answercount);
       4 : Result := compareValue (data1^.SuccessfullQueries, data2^.SuccessfullQueries);
       5 : Result := compareValue (data1^.FailedQueries, data2^.FailedQueries);
       6 : Result := compareValue (data1^.FormatErrors, data2^.FormatErrors);
       7 : Result := compareValue (data1^.NXDomain, data2^.NXDomain);
       8 : Result := compareValue (data1^.NotImplemented, data2^.NotImplemented);
       9 : Result := compareValue (data1^.QueryRefused, data2^.QueryRefused);
      10 : Result := compareValue (data1^.NotAuthorized, data2^.NotAuthorized);
     end;
end;
{$else}
procedure TFormConversationSummary.onVSTNameResolutionCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

begin
end;
{$endif}
(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTHeaderClick (Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);

var node : pVirtualNode;

begin
     case HitInfo.Button of
      mbLeft  : begin
                     // Sort the tree by the current colum, alternating ascending and descending

                     node := Sender.Treeview.GetFirstSelected(false);

                     Sender.Treeview.SortTree (HitInfo.Column, Sender.SortDirection, true);

                     // Since we alternate between ascending and descending we need to reverse the setting for next time

                     if Sender.sortDirection = sdAscending then
                        Sender.SortDirection := sdDescending
                     else
                         Sender.SortDirection := sdAscending;

                     if node <> nil then
                        Sender.Treeview.ScrollIntoView (node, true);
                end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onVSTDblClick (Sender: TObject);

var tree : tBaseVirtualTree;

begin
     tree := (Sender as tBaseVirtualTree);

     ExtractPacketsAndOpen (tree);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayMACEndpoints;

var list   : tMACEndpointList;
    entry  : tMACEndpoint;
    node   : pVirtualNode;
    data   : pMACEndpointTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doshow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating Ethernet Endpoints';
     Application.ProcessMessages;

     list := priv_conversationstore.GetMACEndpointList;
     len := length (list);

     VSTMACEndpoints.Clear;
     VSTMACEndpoints.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabMACEndpoints] <> '' then
                      if priv_tabvalues[TabMACEndpoints].CurrentFilterStr <> '' then
                         begin
                              help := entry.MACAddress.asString(':', true);

//                            doShow := FilterBoxIsAMatch (priv_currentFilter[TabMACEndpoints], help);
                              doshow := FilterBoxIsAMatch (priv_tabvalues[TabMACEndpoints].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTMACEndpoints.AddChild (nil);
                              data := VSTMACEndpoints.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTMACEndpoints.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTMACEndpoints);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetMACEndpoints.Caption := 'MAC Addresses (' + IntToStr (count) + ')'
     else
         TabSheetMACEndpoints.Caption := 'MAC Addresses (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayIPv4Endpoints;

var list   : tIPv4EndpointList;
    entry  : tIPv4Endpoint;
    node   : pVirtualNode;
    data   : pIPv4EndpointTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating IPv4 Endpoints';
     Application.ProcessMessages;

     list := priv_conversationstore.GetIPv4EndpointsList;
     len := length (list);

     VSTIPv4Endpoints.Clear;
     VSTIPv4Endpoints.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabIPv4Endpoints] <> '' then
                      if priv_tabvalues[TabIPv4Endpoints].CurrentFilterStr <> '' then
                         begin
                              help := entry.IPv4Address.ValueAsString;

//                            doShow := FilterBoxIsAMatch (priv_currentFilter[TabIPv4Endpoints], help);
                              doShow := FilterBoxIsAMatch (priv_tabvalues[TabIPv4Endpoints].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTIPv4Endpoints.AddChild (nil);
                              data := VSTIPv4Endpoints.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTIPv4Endpoints.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTIPv4Endpoints);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetIPv4Addresses.Caption := 'IPv4 Addresses (' + IntToStr (count) + ')'
     else
         TabSheetIPv4Addresses.Caption := 'IPv4 Addresses (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayIPv6Endpoints;

var list   : tIPv6EndpointList;
    entry  : tIPv6Endpoint;
    node   : pVirtualNode;
    data   : pIPv6EndpointTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;


begin
     Statusbar.Panels[1].Text := 'Updating IPv6 Endpoints';
     Application.ProcessMessages;

     list := priv_conversationstore.GetIPv6EndpointsList;
     len := length (list);

     VSTIPv6Endpoints.Clear;
     VSTIPv6Endpoints.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabIPv6Endpoints] <> '' then
                      if priv_tabvalues[TabIPv6Endpoints].CurrentFilterStr <> '' then
                         begin
                              help := entry.IPv6Address.ValueAsString;

//                            doShow := FilterBoxIsAMatch (priv_currentFilter[TabIPv6Endpoints], help);
                              doShow := FilterBoxIsAMatch (priv_tabvalues[TabIPv6Endpoints].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTIPv6Endpoints.AddChild (nil);
                              data := VSTIPv6Endpoints.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTIPv6Endpoints.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTIPv6Endpoints);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetIPv6Addresses.Caption := 'IPv6 Addresses (' + IntToStr (count) + ')'
     else
         TabSheetIPv6Addresses.Caption := 'IPv6 Addresses (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayMACConversations;

var list   : tMACConversationList;
    entry  : tMACConversation;
    node   : pVirtualNode;
    data   : pMACConversationTreeEntry;
    len    : integer;
    total  : integer;
    count  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating Ethernet Conversations';
     Application.ProcessMessages;

     list := priv_conversationstore.GetMACConversationList;
     len := length (list);

     VSTMACConversations.Clear;
     VSTMACConversations.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabMACConversations] <> '' then
                      if priv_tabvalues[TabMACConversations].CurrentFilterStr <> '' then
                         begin
                              help := entry.MACPair.asString (':', true);

//                            doShow := FilterBoxIsAMatch (priv_currentFilter[TabMACConversations], help);
                              doShow := FilterBoxIsAMatch (priv_tabvalues[TabMACConversations].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTMACConversations.AddChild (nil);
                              data := VSTMACConversations.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTMACConversations.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTMACConversations);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetMACConversations.Caption := 'MAC Conversations (' + IntToStr (count) + ')'
     else
         TabSheetMACConversations.Caption := 'MAC Conversations (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayIPv4Conversations;

var list   : tIPv4ConversationList;
    entry  : tIPv4Conversation;
    node   : pVirtualNode;
    data   : pIPv4ConversationTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating IPv4 Conversations';
     Application.ProcessMessages;

     list := priv_conversationstore.GetIPv4ConversationList;
     len := length (list);

     VSTIPv4Conversations.Clear;
     VSTIPv4Conversations.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabIPv4Conversations] <> '' then
                      if priv_tabvalues[TabIPv4Conversations].CurrentFilterStr <> '' then
                         begin
                              help := entry.IPv4AddressPair.ValueAsString (true);

//                              doShow := FilterBoxIsAMatch (priv_currentFilter[TabIPv4Conversations], help);
                                doShow := FilterBoxIsAMatch (priv_tabvalues[TabIPv4Conversations].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTIPv4Conversations.AddChild (nil);
                              data := VSTIPv4Conversations.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTIPv4Conversations.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTIPv4Conversations);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetIPv4Conversations.Caption := 'IPv4 Conversations (' + IntToStr (count) + ')'
     else
         TabSheetIPv4Conversations.Caption := 'IPv4 Conversations (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayIPv6Conversations;

var list   : tIPv6ConversationList;
    entry  : tIPv6Conversation;
    node   : pVirtualNode;
    data   : pIPv6ConversationTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating IPv6 Conversations';
     Application.ProcessMessages;

     list := priv_conversationstore.GetIPv6ConversationList;
     len := length (list);

     VSTIPv6Conversations.Clear;
     VSTIPv6Conversations.BeginUpdate;

     count := 0;
     total := 0;

     for i := 1 to len do
         begin
              entry := list[i-1];

              // We only display it if it was found on the network, not in some DNS packet or ARP frame

              doShow := containsLiveNetworkMarker (entry.Markers);

              if doShow then
                 begin
                      inc (total);

                      // Check if a filter is active

//                    if priv_currentfilter[TabIPv6Conversations] <> '' then
                      if priv_tabvalues[TabIPv6Conversations].CurrentFilterStr <> '' then
                         begin
                              help := entry.IPv6AddressPair.ValueAsString (true);

//                            doShow := FilterBoxIsAMatch (priv_currentFilter[TabIPv6Conversations], help);
                              doShow := FilterBoxIsAMatch (priv_tabvalues[TabIPv6Conversations].CurrentFilterStr, help);
                         end;

                      if doShow then
                         begin
                              inc (count);

                              node := VSTIPv6Conversations.AddChild (nil);
                              data := VSTIPv6Conversations.GetNodeData (node);

                              data^.CopyFrom (entry);
                              data^.index := total;
                         end;
                 end;
         end;

     VSTIPv6Conversations.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTIPv6Conversations);

     Application.ProcessMessages;

     // Update Header Tab

     if count = total then
        TabSheetIPv6Conversations.Caption := 'IPv6 Conversations (' + IntToStr (count) + ')'
     else
         TabSheetIPv6Conversations.Caption := 'IPv6 Conversations (' + IntToStr (count) + '/'+ IntToStr (total) + ')';
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayTCPConversations;

var node      : pVirtualNode;
    data      : pTCPConversationTreeEntry;
    entry     : tTCPConversation;
    len       : integer;
    IPPair    : tIPAddressPair;
    count     : integer;
    total     : integer;
    i         : integer;
    row       : integer;
    help      : string;
    searchfor : string;
    doShow    : boolean;
    parent    : pVirtualNode;

begin
     SiMain.EnterMethod ('TFormConversationSummary.DisplayTCPConversations');

     Statusbar.Panels[1].Text := 'Updating TCP Conversations';
     Application.ProcessMessages;

     node := VSTTCPConversations.GetFirstSelected;

     if assigned (node) then
        begin
             data := VSTTCPConversations.GetNodeData (node);
             priv_tabvalues[TabTCPConversations].CurrentRow := data^.index;
        end
     else
         priv_tabvalues[TabTCPConversations].CurrentRow := 0;

     len := length (priv_TCPList);

     priv_tabvalues[TabTCPConversations].rowindexlist.Clear;

     VSTTCPConversations.Clear;
     VSTTCPConversations.BeginUpdate;

     count := 0;
     total := 0;
     parent := nil;
     IPPair.initialize;

     if priv_tabvalues[TabTCPConversations].AggregatedMode then
        VSTTCPConversations.TreeOptions.PaintOptions := VSTTCPConversations.TreeOptions.PaintOptions + [toShowRoot]
     else
         VSTTCPConversations.TreeOptions.PaintOptions := VSTTCPConversations.TreeOptions.PaintOptions - [toShowRoot];

     // Make sure the conversation store was finalized

     if len > 0 then
        assert (priv_TCPList[0].FilterMatchText <> '');

     // Display the entries
     // TODO: check if using "VirtualTree.IsVisible[Node] := False;" is faster/easier

     searchfor := priv_tabvalues[TabTCPConversations].CurrentFilterStr;

     if searchfor <> ''  then
        begin
             for i := 1 to len do
                 begin
                      entry := priv_TCPList[i-1];

                      inc (total);

                      // check if the entry is marked or matches the filter

                      if (priv_tabvalues[TabTCPConversations].CurrentSocketFilter.Port > 0) and
                         ((priv_tabvalues[TabTCPConversations].CurrentSocketFilter = entry.SocketPair.Source) or (priv_tabvalues[TabTCPConversations].CurrentSocketFilter = entry.SocketPair.Destination)) then
                         doShow := not (pos ('!', searchfor) = 1)
                      else
                          doShow := entry.Marked or FilterBoxIsAMatch (searchfor, entry.FilterMatchText);

                      if doShow then
                         begin
                              inc (count);

                              // If aggregation mode is active, group conversations by IP pair

                              if priv_tabvalues[TabTCPConversations].AggregatedMode then
                                 begin
                                      // Check if we have a new IP pair in aggregated mode

                                      if not (IPPair = entry.SocketPair.IPAddressPair) then
                                         begin
                                              // Create new parent node

                                              parent := VSTTCPConversations.AddChild (nil);
                                              data := VSTTCPConversations.GetNodeData (parent);

                                              data^.IsParent := true;
                                              data^.AddressA := entry.SocketPair.Source.IPAddress;
                                              data^.AddressB := entry.SocketPair.Destination.IPAddress;
                                              data^.Marked := entry.Marked;

                                              IPPair := priv_TCPList[i-1].SocketPair.IPAddressPair
                                         end;
                                 end;

                              node := VSTTCPConversations.AddChild (parent);
                              data := VSTTCPConversations.GetNodeData (node);

                              data^.CopyFrom (entry, total);

                              VSTTCPConversations.FullExpand (parent);

                              priv_tabvalues[TabTCPConversations].rowindexlist.Add (total, node);
                         end;
                 end;
        end
     else
         begin
              if priv_tabvalues[TabTCPConversations].AggregatedMode then
                 begin
                      for i := 1 to len do
                          begin
                               // Check if we have a new IP pair in aggregated mode

                               if not (IPPair = priv_TCPList[i-1].SocketPair.IPAddressPair) then
                                  begin
                                       // Create new parent node

                                       parent := VSTTCPConversations.AddChild (nil);
                                       data := VSTTCPConversations.GetNodeData (parent);

                                       data^.IsParent := true;
                                       data^.AddressA := priv_TCPList[i-1].SocketPair.Source.IPAddress;
                                       data^.AddressB := priv_TCPList[i-1].SocketPair.Destination.IPAddress;

                                       IPPair := priv_TCPList[i-1].SocketPair.IPAddressPair
                                  end;

                               // Add the actual TCP row

                               node := VSTTCPConversations.AddChild (parent);
                               data := VSTTCPConversations.GetNodeData (node);

                               data^.CopyFrom (priv_TCPList[i-1], i);

                               VSTTCPConversations.FullExpand (parent);

                               priv_tabvalues[TabTCPConversations].rowindexlist.Add (i, node);
                          end;
                 end
              else
                  begin
                       for i := 1 to len do
                           begin
                                node := VSTTCPConversations.AddChild (parent);
                                data := VSTTCPConversations.GetNodeData (node);
                                data^.CopyFrom (priv_TCPList[i-1], i);
                                priv_tabvalues[TabTCPConversations].rowindexlist.Add (i, node);
                           end;
                  end;

              total := len;
              count := total;
         end;

{   // Slow, for WHATEVER REASON.
              for i := 1 to len do
                  begin
                       if priv_tabvalues[TabTCPConversations].AggregatedMode then
                          begin
                               // Check if we have a new IP pair in aggregated mode

                               if not (IPPair = priv_TCPList[i-1].SocketPair.IPAddressPair) then
                                  begin
                                       // Create new parent node

                                       parent := VSTTCPConversations.AddChild (nil);
                                       data := VSTTCPConversations.GetNodeData (parent);

                                       data^.IsParent := true;
                                       data^.AddressA := priv_TCPList[i-1].SocketPair.Source.IPAddress;
                                       data^.AddressB := priv_TCPList[i-1].SocketPair.Destination.IPAddress;

                                       IPPair := priv_TCPList[i-1].SocketPair.IPAddressPair
                                  end;
                          end;

                       node := VSTTCPConversations.AddChild (parent);
                       data := VSTTCPConversations.GetNodeData (node);

                       data^.CopyFrom (priv_TCPList[i-1], i);

                       VSTTCPConversations.FullExpand (parent);

                       priv_tabvalues[TabTCPConversations].rowindexlist.Add (i, node);
                  end;
}
{ // fast
              for i := 1 to len do
                  begin
                       node := VSTTCPConversations.AddChild (parent);
                       data := VSTTCPConversations.GetNodeData (node);
                       data^.CopyFrom (priv_TCPList[i-1], i);
                       priv_tabvalues[TabTCPConversations].rowindexlist.Add (i, node);
                  end;
}
{
              total := len;
              count := total;
         end;
}


     VSTTCPConversations.EndUpdate;

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTTCPConversations);

     Application.ProcessMessages;

     // Bring last row into view if there was one before rebuilding the table

     row := priv_tabvalues[TabTCPConversations].CurrentRow;

     if (row > 0) and (priv_tabvalues[TabTCPConversations].rowindexlist.ContainsKey (row)) then
        begin
             node := priv_tabvalues[TabTCPConversations].rowindexlist.Items[row];
             VSTTCPConversations.ScrollIntoView (node, true);
        end;

     // Update Header Tab

     if total = count then
        TabSheetTCPConversations.Caption := 'TCP Conversations (' + IntToStr (total) + ')'
     else
         TabSheetTCPConversations.Caption := 'TCP Conversations (' + IntToStr (count) + '/' + IntToStr (total) + ')';

     SiMain.LeaveMethod ('TFormConversationSummary.DisplayTCPConversations');
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayUDPConversations;

var list   : tUDPConversationList;
    entry  : tUDPConversation;
    node   : pVirtualNode;
    data   : pUDPConversationTreeEntry;
    len    : integer;
    count  : integer;
    total  : integer;
    i      : integer;
    help   : string;
    doShow : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating UDP Conversations';
     Application.ProcessMessages;

     list := priv_conversationstore.GetUDPConversationList;
     len := length (list);

     VSTUDPConversations.Clear;
     VSTUDPConversations.BeginUpdate;

     count := 0;
     total := 0;

     // If we're not obviously filtering on IPv6 we show IPv4 stuff

             for i := 1 to len do
                 begin
                      entry := list[i-1];

                      // We only display it if it was found on the network, not in some DNS packet or ARP frame

                      doShow := containsLiveNetworkMarker (entry.Markers);

                      if doShow then
                         begin
                              inc (total);

                              // Check if a filter is active

                              if priv_tabvalues[TabUDPConversations].CurrentFilterStr <> '' then
                                 begin
                                      help := entry.SocketPair.ValueAsString (true);

                                      doShow := FilterBoxIsAMatch (priv_tabvalues[TabUDPConversations].CurrentFilterStr, help);
                                 end;

                              if doShow then
                                 begin
                                      inc (count);

                                      node := VSTUDPConversations.AddChild (nil);
                                      data := VSTUDPConversations.GetNodeData (node);

                                      data^.CopyFrom (entry);
                                      data^.index := total;
                                 end;
                         end;
                 end;

     VSTUDPConversations.EndUpdate;

     // Update Header Tab

     if total = count then
        TabSheetUDPConversations.Caption := 'UDP Conversations (' + IntToStr (count) + ')'
     else
         TabSheetUDPConversations.Caption := 'UDP Conversations (' + IntToStr (count) + '/' + IntToStr (total) + ')';

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTUDPConversations);

     Application.ProcessMessages;
end;

(* ------------------------------------------- *)
{$ifdef fqdnstats}
procedure TFormConversationSummary.DisplayNameResolutions;

var entry      : tNameResolutionEntry;
    parentnode : pVirtualNode;
    parentdata : pFQDNTreeNode;
    node       : pVirtualNode;
    data       : pFQDNTreeNode;
    key        : string;
    len        : integer;
    count      : integer;
    total      : integer;
    i          : integer;
    help       : string;
    doShow     : boolean;

begin
     Statusbar.Panels[1].Text := 'Updating NameResolutions';
     Application.ProcessMessages;

     len := length (priv_NRList);

     if len > 0 then
        begin
             priv_tabvalues[TabNameResolutions].rowindexlist.Clear;

             priv_FQDNManager.DisplayTree (VSTNameResolution, priv_tabvalues[TabNameResolutions].AggregatedMode);

             VSTNameResolution.BeginUpdate;

             // Add more details

             for i := 1 to len do
                 begin
                      entry := priv_NRList[i-1];

                      assert (entry.FQDN.FullFQDN <> '', 'FQDN is empty (TFormConversationSummary.DisplayNameResolutions)');

                      parentnode := priv_FQDNManager.GetFQDNVirtualTreeNode (entry.FQDN);

                      assert (assigned (parentnode), 'Node not found for FQDN "' + entry.FQDN.FullFQDN + '" (TFormConversationSummary.DisplayNameResolutions)');

                      parentdata := VSTNameResolution.GetNodeData(ParentNode);

                      assert (lowercase (parentdata^.name) = lowercase (entry.FQDN.FullFQDN), 'FQDN node mismatch for "' + entry.FQDN.FullFQDN + '" (TFormConversationSummary.DisplayNameResolutions)');

                      // Aggregate details to parent

                      parentdata^.framecount := parentdata^.framecount + entry.FrameCount;

                      // Add subnode for query or answer

                      node := VSTNameResolution.AddChild (parentnode);
                      data := VSTNameResolution.GetNodeData (node);

                      // Check return codes

                      case entry.ReplyCode of
                       dnsrcNoError  : parentdata^.SuccessfullQueries := parentdata^.SuccessfullQueries + entry.FrameCount;
                       dnsrcFormErr  : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentdata^.FormatErrors := parentdata^.FormatErrors + entry.FrameCount;
                                            data^.FailedQueries := entry.FrameCount;
                                            data^.FormatErrors := entry.FrameCount;
                                       end;
                       dnsrcServFail : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentdata^.ServerFailures := parentdata^.ServerFailures + entry.FrameCount;
                                            data^.FailedQueries := entry.FrameCount;
                                            data^.ServerFailures := entry.FrameCount;
                                       end;
                       dnsrcNXDomain : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentdata^.NXDomain := parentdata^.NXDomain + entry.FrameCount;
                                            data^.FailedQueries := entry.FrameCount;
                                            data^.NXDomain := entry.FrameCount;
                                       end;
                       dnsrcNotImp   : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentdata^.NotImplemented := parentdata^.NotImplemented + entry.FrameCount;
                                            data^.FailedQueries := entry.FrameCount;
                                            data^.NotImplemented := entry.FrameCount;
                                       end;
                       dnsrcRefused  : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentdata^.QueryRefused := parentdata^.QueryRefused + entry.FrameCount;
                                            data^.FailedQueries := entry.FrameCount;
                                            data^.QueryRefused := entry.FrameCount;
                                       end;
                       dnsrcNotAuth  : begin
                                            parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                                            parentData^.NotAuthorized := parentdata^.NotAuthorized + entry.FrameCount;
                                            data^.FormatErrors := entry.FrameCount;
                                            data^.NotAuthorized := entry.FrameCount;
                                       end;
                      else
                          parentdata^.FailedQueries := parentdata^.FailedQueries + entry.FrameCount;
                      end;

                      // Add detail nodes

                      if entry.isAnswer then
                         begin
                              parentdata^.answercount := parentdata^.answercount + entry.FrameCount;

                              data^.nodetype := tntDetailQuery;
                              data^.name := 'Answer: ' + entry.Answer.asString;
                              data^.framecount := entry.FrameCount;
                              data^.querycount := 0;
                              data^.answercount := entry.FrameCount;
                         end
                      else
                          begin
                               parentdata^.querycount := parentdata^.querycount + entry.FrameCount;

                               data^.nodetype := tntDetailQuery;
                               data^.name := 'Query of type: ' + QueryTypeSetToString (entry.QueryTypes);
                               data^.framecount := entry.FrameCount;
                               data^.querycount := entry.FrameCount;
                               data^.answercount := 0;
                          end;
                 end;

             VSTNameResolution.EndUpdate;
        end;

     // Update Header Tab

     count := priv_FQDNManager.count;
     total := count;                     // will change later when filtering is implemented

     if (total = count) then
        TabSheetNameResolution.Caption := 'Name Resolution (' + IntToStr (count) + ')'
     else
         TabSheetNameResolution.Caption := 'Name Resolution (' + IntToStr (count) + '/' + IntToStr (total) + ')';

     Statusbar.Panels[1].Text := '';
     ShowColumnStatus (VSTNameResolution);

     Application.ProcessMessages;
end;
{$endif}

(* ------------------------------------------- *)

function TFormConversationSummary.CreateMACConversationFilter (Conversation: tMACConversationTreeEntry; MenuTag: integer): string;

var help : string;

begin
     case MenuTag of
      1 : begin
               // A <-> B

               help := 'eth.addr==' + Conversation.AddressA.asString (':', true) + ' and eth.addr==' + Conversation.AddressB.asString (':', true);
          end;
      2 : begin
               // A -> B

               help := 'eth.src==' + Conversation.AddressA.asString (':', true) + ' and eth.dest==' + Conversation.AddressB.asString (':', true);
          end;
      3 : begin
               // B -> A

               help := 'eth.src==' + Conversation.AddressB.asString (':', true) + ' and eth.dest==' + Conversation.AddressA.asString (':', true);
          end;
      4 : begin
               // A to Any

               help := 'eth.addr==' + Conversation.AddressA.asString (':', true)
          end;
      5 : begin
               // B to Any

               help := 'eth.addr==' + Conversation.AddressB.asString (':', true)
          end;
     end;

     result := help;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.CreateIPv4ConversationFilter (Conversation: tIPv4ConversationTreeEntry; MenuTag: integer): string;

var help : string;

begin
     case MenuTag of
      1 : begin
               // A <-> B

               help := 'ip.addr==' + Conversation.AddressA.ValueAsString + ' and ip.addr==' + Conversation.AddressB.ValueAsString;
          end;
      2 : begin
               // A -> B

               help := 'ip.src==' + Conversation.AddressA.ValueAsString + ' and ip.dest==' + Conversation.AddressB.ValueAsString;
          end;
      3 : begin
               // B -> A

               help := 'ip.src==' + Conversation.AddressB.ValueAsString + ' and ip.dest==' + Conversation.AddressA.ValueAsString;
          end;
      4 : begin
               // A to Any

               help := 'ip.addr==' + Conversation.AddressA.ValueAsString;
          end;
      5 : begin
               // B to Any

               help := 'ip.addr==' + Conversation.AddressB.ValueAsString
          end;
     end;

     result := help;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.CreateIPv6ConversationFilter (Conversation: tIPv6ConversationTreeEntry; MenuTag: integer): string;

var help : string;

begin
     case MenuTag of
      1 : begin
               // A <-> B

               help := 'ipv6.addr==' + Conversation.AddressA.ValueAsString + ' and ipv6.addr==' + Conversation.AddressB.ValueAsString;
          end;
      2 : begin
               // A -> B

               help := 'ipv6.src==' + Conversation.AddressA.ValueAsString + ' and ipv6.dest==' + Conversation.AddressB.ValueAsString;
          end;
      3 : begin
               // B -> A

               help := 'ipv6.src==' + Conversation.AddressB.ValueAsString + ' and ipv6.dest==' + Conversation.AddressA.ValueAsString;
          end;
      4 : begin
               // A to Any

               help := 'ipv6.addr==' + Conversation.AddressA.ValueAsString;
          end;
      5 : begin
               // B to Any

               help := 'ipv6.addr==' + Conversation.AddressB.ValueAsString
          end;
     end;

     result := help;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.CreateTCPConversationFilter (Conversation: tTCPConversationTreeEntry; MenuTag: integer): string;

var help : string;

begin
     case MenuTag of
      1 : begin
               // A <-> B

               if Conversation.AddressA.IPVersion = 4 then
//                help := 'ip.addr==' + Conversation.AddressIPv4A.ValueAsString + ' and ip.addr==' + Conversation.AddressIPv4B.ValueAsString +
                  help := 'ip.addr==' + Conversation.AddressA.IPv4Address.ValueAsString + ' and ip.addr==' + Conversation.AddressB.IPv4Address.ValueAsString +
                          ' and tcp.port==' + IntTostr (Conversation.PortA) + ' and tcp.port==' + IntTostr (Conversation.PortB)
               else
//                 help := 'ipv6.addr==' + Conversation.AddressIPv6A.ValueAsString + ' and ipv6.addr==' + Conversation.AddressIPv6B.ValueAsString +
                   help := 'ipv6.addr==' + Conversation.AddressA.IPv6Address.ValueAsString + ' and ipv6.addr==' + Conversation.AddressB.IPv6Address.ValueAsString +
                           ' and tcp.port==' + IntTostr (Conversation.PortA) + ' and tcp.port==' + IntTostr (Conversation.PortB);
          end;
      2 : begin
               // A -> B

               if Conversation.AddressA.IPVersion = 4 then
//                help := 'ip.src==' + Conversation.AddressIPv4A.ValueAsString + ' and ip.dest==' + Conversation.AddressIPv4B.ValueAsString +
                  help := 'ip.src==' + Conversation.AddressA.IPv4Address.ValueAsString + ' and ip.dest==' + Conversation.AddressB.IPv4Address.ValueAsString +
                          ' and tcp.srcport==' + IntTostr (Conversation.PortA) + ' and tcp.dstport==' + IntTostr (Conversation.PortB)
               else
//                 help := 'ipv6.src==' + Conversation.AddressIPv6A.ValueAsString + ' and ipv6.dest==' + Conversation.AddressIPv6B.ValueAsString +
                   help := 'ipv6.src==' + Conversation.AddressA.IPv6Address.ValueAsString + ' and ipv6.dest==' + Conversation.AddressB.IPv6Address.ValueAsString +
                           ' and tcp.srcport==' + IntTostr (Conversation.PortA) + ' and tcp.dstport==' + IntTostr (Conversation.PortB)
          end;
      3 : begin
               // B -> A

               if Conversation.AddressA.IPVersion = 4 then
//                help := 'ip.src==' + Conversation.AddressIPv4B.ValueAsString + ' and ip.dest==' + Conversation.AddressIPv4A.ValueAsString +
                  help := 'ip.src==' + Conversation.AddressB.IPv4Address.ValueAsString + ' and ip.dest==' + Conversation.AddressA.IPv4Address.ValueAsString +
                          ' and tcp.srcport==' + IntTostr (Conversation.PortB) + ' and tcp.dstport==' + IntTostr (Conversation.PortA)
               else
//                 help := 'ipv6.src==' + Conversation.AddressIPv6B.ValueAsString + ' and ipv6.dest==' + Conversation.AddressIPv6A.ValueAsString +
                   help := 'ipv6.src==' + Conversation.AddressB.IPv6Address.ValueAsString + ' and ipv6.dest==' + Conversation.AddressA.IPv6Address.ValueAsString +
                           ' and tcp.srcport==' + IntTostr (Conversation.PortB) + ' and tcp.dstport==' + IntTostr (Conversation.PortA)
          end;
      4 : begin
               // A to Any

               if Conversation.AddressA.IPVersion = 4 then
//                help := 'ip.addr==' + Conversation.AddressIPv4A.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA)
                  help := 'ip.addr==' + Conversation.AddressA.IPv4Address.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA)
               else
//                 help := 'ipv6.addr==' + Conversation.AddressIPv6A.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA);
                   help := 'ipv6.addr==' + Conversation.AddressA.IPv6Address.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA);
          end;
      5 : begin
               // B to Any

               if Conversation.AddressB.IPVersion = 4 then
//                help := 'ip.addr==' + Conversation.AddressIPv4B.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB)
                  help := 'ip.addr==' + Conversation.AddressB.IPv4Address.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB)
               else
//                 help := 'ipv6.addr==' + Conversation.AddressIPv6B.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB);
                   help := 'ipv6.addr==' + Conversation.AddressB.IPv6Address.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB);
          end;
     end;

     result := help;
end;
(* ------------------------------------------- *)

function TFormConversationSummary.CreateUDPConversationFilter (Conversation: tUDPConversationTreeEntry; MenuTag: integer): string;

var help : string;

begin
     case MenuTag of
      1 : begin
               // A <-> B

               if Conversation.AddressA.IPVersion = 4 then
                  help := 'ip.addr==' + Conversation.AddressA.ValueAsString + ' and ip.addr==' + Conversation.AddressB.ValueAsString +
                          ' and tcp.port==' + IntTostr (Conversation.PortA) + ' and tcp.port==' + IntTostr (Conversation.PortB)
               else
                   help := 'ipv6.addr==' + Conversation.AddressA.ValueAsString + ' and ipv6.addr==' + Conversation.AddressB.ValueAsString +
                           ' and tcp.port==' + IntTostr (Conversation.PortA) + ' and tcp.port==' + IntTostr (Conversation.PortB);
          end;
      2 : begin
               // A -> B

               if Conversation.AddressA.IPVersion = 4 then
                  help := 'ip.src==' + Conversation.AddressA.ValueAsString + ' and ip.dest==' + Conversation.AddressB.ValueAsString +
                          ' and tcp.srcport==' + IntTostr (Conversation.PortA) + ' and tcp.dstport==' + IntTostr (Conversation.PortB)
               else
                   help := 'ipv6.src==' + Conversation.AddressA.ValueAsString + ' and ipv6.dest==' + Conversation.AddressB.ValueAsString +
                           ' and tcp.srcport==' + IntTostr (Conversation.PortA) + ' and tcp.dstport==' + IntTostr (Conversation.PortB)
          end;
      3 : begin
               // B -> A

               if Conversation.AddressA.IPVersion = 4 then
                  help := 'ip.src==' + Conversation.AddressB.ValueAsString + ' and ip.dest==' + Conversation.AddressA.ValueAsString +
                          ' and tcp.srcport==' + IntTostr (Conversation.PortB) + ' and tcp.dstport==' + IntTostr (Conversation.PortA)
               else
                   help := 'ipv6.src==' + Conversation.AddressB.ValueAsString + ' and ipv6.dest==' + Conversation.AddressA.ValueAsString +
                           ' and tcp.srcport==' + IntTostr (Conversation.PortB) + ' and tcp.dstport==' + IntTostr (Conversation.PortA)
          end;
      4 : begin
               // A to Any

               if Conversation.AddressA.IPVersion = 4 then
                  help := 'ip.addr==' + Conversation.AddressA.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA)
               else
                   help := 'ipv6.addr==' + Conversation.AddressA.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortA);
          end;
      5 : begin
               // B to Any

               if Conversation.AddressA.IPVersion = 4 then
                  help := 'ip.addr==' + Conversation.AddressB.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB)
               else
                   help := 'ipv6.addr==' + Conversation.AddressB.ValueAsString + ' and tcp.port==' + IntToStr (Conversation.PortB);
          end;
     end;

     result := help;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.determineWriteMode: tMultiWriteMode;

begin
     case PageControlConversations.ActivePageIndex of
      0   : begin
               // MAC Endpoints

               result := mwmEtherEndpoint;
            end;
      1,2 : begin
                 // IPv4/IPv6 Endpoints

                 result := mwmIPEndpoint;
            end;
      3   : begin
                 // MAC Conversations

                 result := mwmEtherConversation;
            end;
      4,5 : begin
                 // IPv6 Conversations

                 result := mwmIPConversation;
            end;
      6,7 : begin
                 // TCP Conversations

                 result := mwmL4Conversation;
            end;
      else
          result := mwmNone;
     end;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.CopySharkFilterToClipboard (Tree: TBaseVirtualTree; const FilterType : Integer): string;

var node     : pVirtualNode;
    macdata  : pMACConversationTreeEntry;
    ipv4data : pIPv4ConversationTreeEntry;
    ipv6data : pIPv6ConversationTreeEntry;
    tcpdata  : pTCPConversationTreeEntry;
    udpdata  : pUDPConversationTreeEntry;
    filter   : string;
    help     : string;
    count    : integer;

begin
     filter := '';
     count := 0;

     node := Tree.GetFirstSelected;

     while node <> nil do
           begin
                inc (count);

                if Tree.Name = 'VSTMACConversations' then
                   begin
                        macdata := Tree.GetNodeData (node);

                        help := CreateMACConversationFilter (macdata^, FilterType);
                   end;

                if Tree.Name = 'VSTIPv4Conversations' then
                   begin
                        ipv4data := Tree.GetNodeData (node);

                        help := CreateIPv4ConversationFilter (ipv4data^, FilterType);
                   end;

                if Tree.Name = 'VSTIPv6Conversations' then
                   begin
                        ipv6data := Tree.GetNodeData (node);

                        help := CreateIPv6ConversationFilter (ipv6data^, FilterType);
                   end;

                if Tree.Name = 'VSTTCPConversations' then
                   begin
                        tcpdata := Tree.GetNodeData (node);

                        help := CreateTCPConversationFilter (tcpdata^, FilterType);
                   end;

                if Tree.Name = 'VSTUDPConversations' then
                   begin
                        udpdata := Tree.GetNodeData (node);

                        help := CreateUDPConversationFilter (udpdata^, FilterType);
                   end;

                case count of
                 1 : filter := help;
                 2 : filter := '(' + filter + ') or (' + help + ')';
                else
                    filter := filter + ' or (' + help + ')';
                end;

                node := Tree.GetNextSelected (node);
           end;

     Clipboard.AsText := filter;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuCreateConversationSharkFilterClick (Sender: TObject);

var Menu : tMenuItem;

begin
     Menu := (Sender as tMenuItem);

     if priv_popupTree <> nil then
        CopySharkFilterToClipboard (priv_popupTree, Menu.Tag);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuCreateEndpointSharkFilterClick (Sender: TObject);

var node     : pVirtualNode;
    macdata  : pMACEndpointTreeEntry;
    ipv4data : pIPv4EndpointTreeEntry;
    ipv6data : pIPv6EndpointTreeEntry;
    filter   : string;
    help     : string;
    count    : integer;

begin
     if priv_popupTree <> nil then
        begin
             filter := '';
             count := 0;

             node := priv_popupTree.GetFirstSelected;

             while node <> nil do
                   begin
                        inc (count);

                        if priv_popupTree.Name = 'VSTMACEndpoints' then
                           begin
                                macdata := priv_popupTree.GetNodeData (node);

                                help := 'eth.addr==' + macdata^.Address.asString(':', true);
                           end;

                        if priv_popupTree.Name = 'VSTIPv4Endpoints' then
                           begin
                                ipv4data := priv_popupTree.GetNodeData (node);

                                help := 'ip.addr==' + ipv4data^.Address.ValueAsString;
                           end;

                        if priv_popupTree.Name = 'VSTIPv6Endpoints' then
                           begin
                                ipv6data := priv_popupTree.GetNodeData (node);

                                help := 'ipv6.addr==' + ipv6data^.Address.ValueAsString;
                           end;

                        if count = 1 then
                           filter := help
                        else
                            filter := filter + ' or ' + help;

                        node := priv_popupTree.GetNextSelected (node);
                   end;

             Clipboard.AsText := filter;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuEndpointOpenClick(Sender: TObject);

begin
     ExtractPacketsAndOpen (priv_popuptree);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.getFilterAndInterfaceMappingsFromList (Tree : tBaseVirtualTree);

var node     : pVirtualNode;
    macendp  : pMACEndpointTreeEntry;
    ipv4endp : pIPv4EndpointTreeEntry;
    ipv6endp : pIPv6EndpointTreeEntry;
    macconv  : pMACConversationTreeEntry;
    ipv4conv : pIPv4ConversationTreeEntry;
    ipv6conv : pIPv6ConversationTreeEntry;
    tcpconv  : pTCPConversationTreeEntry;
    udpconv  : pUDPConversationTreeEntry;
    filter   : tFrameFilter;
    help     : string;
    count    : integer;

begin
      if Tree <> nil then
        begin
             setlength (priv_interfacemaplist, 0);

//           count := length (priv_filterlist);
             count := priv_filterlist.count;

             node := Tree.GetFirstSelected;

             while node <> nil do
                   begin
                        inc (count);
                        filter.initialize;

                        if Tree.Name = 'VSTMACEndpoints' then
                           begin
                                macendp := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := macendp.index;

                                filter.MACFilter.AddressFilterType := ftSingle;
                                filter.MACFilter.AddressScope := fsAny;
                                filter.MACFilter.SingleAddress := macendp.Address;

                                filter.HelpFirstSeen := macendp.HelpFirstSeen;
                                filter.HelpLastSeen := macendp.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, macendp.InterfaceList);
                           end;

                        if Tree.Name = 'VSTIPv4Endpoints' then
                           begin
                                ipv4endp := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := ipv4endp.index;

                                filter.IPv4Filter.AddressFilterType := ftSingle;
                                filter.IPv4Filter.AddressScope := fsAny;
                                filter.IPv4Filter.SingleAddress := ipv4endp.Address;

                                filter.HelpFirstSeen := ipv4endp.HelpFirstSeen;
                                filter.HelpLastSeen := ipv4endp.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, ipv4endp.InterfaceList);
                           end;

                        if Tree.Name = 'VSTIPv6Endpoints' then
                           begin
                                ipv6endp := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := ipv6endp.index;

                                filter.IPv6Filter.AddressFilterType := ftSingle;
                                filter.IPv6Filter.AddressScope := fsAny;
                                filter.IPv6Filter.SingleAddress := ipv6endp.Address;

                                filter.HelpFirstSeen := ipv6endp.HelpFirstSeen;
                                filter.HelpLastSeen := ipv6endp.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, ipv6endp.InterfaceList);
                           end;

                        if Tree.Name = 'VSTMACConversations' then
                           begin
                                macconv := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := macconv.index;

                                filter.MACFilter.AddressFilterType := ftSingle;
                                filter.MACFilter.AddressScope := fsBoth;
                                filter.MACFilter.SourceAddress := macconv.AddressA;
                                filter.MACFilter.DestinationAddress := macconv.AddressB;

                                filter.HelpFirstSeen := macconv.HelpFirstSeen;
                                filter.HelpLastSeen := macconv.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, macconv.InterfaceList);
                           end;

                        if Tree.Name = 'VSTIPv4Conversations' then
                           begin
                                ipv4conv := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := ipv4conv.index;

                                filter.IPv4Filter.AddressFilterType := ftSingle;
                                filter.IPv4Filter.AddressScope := fsBoth;
                                filter.IPv4Filter.SourceAddress := ipv4conv.AddressA;
                                filter.IPv4Filter.DestinationAddress := ipv4conv.AddressB;

                                filter.HelpFirstSeen := ipv4conv.HelpFirstSeen;
                                filter.HelpLastSeen := ipv4conv.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, ipv4conv.InterfaceList);
                           end;

                        if Tree.Name = 'VSTIPv6Conversations' then
                           begin
                                ipv6conv := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := ipv6conv.index;

                                filter.IPv6Filter.AddressFilterType := ftSingle;
                                filter.IPv6Filter.AddressScope := fsBoth;
                                filter.IPv6Filter.SourceAddress := ipv6conv.AddressA;
                                filter.IPv6Filter.DestinationAddress := ipv6conv.AddressB;

                                filter.HelpFirstSeen := ipv6conv.HelpFirstSeen;
                                filter.HelpLastSeen := ipv6conv.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, ipv6conv.InterfaceList);
                           end;

                        if Tree.Name = 'VSTTCPConversations' then
                           begin
                                tcpconv := Tree.GetNodeData (node);

                                if not tcpconv.IsParent then
                                   begin
                                        filter.HelpConversationIndex := tcpconv.index;

                                        filter.PortFilter.Layer4Protocol := l4TCP;

                                        if tcpconv.AddressA.IPVersion = 4 then
                                           begin
                                                filter.IPv4Filter.AddressFilterType := ftSingle;
                                                filter.IPv4Filter.AddressScope := fsBoth;
                                                filter.IPv4Filter.SourceAddress := tcpconv.AddressA.IPv4Address;
                                                filter.IPv4Filter.DestinationAddress := tcpconv.AddressB.IPv4Address;
                                           end
                                        else
                                            begin
                                                 filter.IPv6Filter.AddressFilterType := ftSingle;
                                                 filter.IPv6Filter.AddressScope := fsBoth;
                                                 filter.IPv6Filter.SourceAddress := tcpconv.AddressA.IPv6Address;
                                                 filter.IPv6Filter.DestinationAddress := tcpconv.AddressB.IPv6Address;
                                            end;

                                        filter.PortFilter.PortFilterType := ftSingle;
                                        filter.PortFilter.PortScope := fsBoth;
                                        filter.PortFilter.SourcePort := tcpconv.PortA;
                                        filter.PortFilter.DestinationPort := tcpconv.PortB;

                                        filter.HelpFirstSeen := tcpconv.HelpFirstSeen;
                                        filter.HelpLastSeen := tcpconv.HelpLastSeen;

                                        // Add the interfaces in use by the current conversation to the mapping list.

                                        priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, tcpconv.InterfaceList);
                                   end;
                           end;

                        if Tree.Name = 'VSTUDPConversations' then
                           begin
                                udpconv := Tree.GetNodeData (node);

                                filter.HelpConversationIndex := udpconv.index;

                                filter.PortFilter.Layer4Protocol := l4UDP;

                                if udpconv.AddressA.IPVersion = 4 then
                                   begin
                                        filter.IPv4Filter.AddressFilterType := ftSingle;
                                        filter.IPv4Filter.AddressScope := fsBoth;
                                        filter.IPv4Filter.SourceAddress := udpconv.AddressA.IPv4Address;
                                        filter.IPv4Filter.DestinationAddress := udpconv.AddressB.IPv4Address;
                                   end
                                else
                                    begin
                                         filter.IPv6Filter.AddressFilterType := ftSingle;
                                         filter.IPv6Filter.AddressScope := fsBoth;
                                         filter.IPv6Filter.SourceAddress := udpconv.AddressA.IPv6Address;
                                         filter.IPv6Filter.DestinationAddress := udpconv.AddressB.IPv6Address;
                                    end;

                                filter.PortFilter.PortFilterType := ftSingle;
                                filter.PortFilter.PortScope := fsBoth;
                                filter.PortFilter.SourcePort := udpconv.PortA;
                                filter.PortFilter.DestinationPort := udpconv.PortB;

                                filter.HelpFirstSeen := udpconv.HelpFirstSeen;
                                filter.HelpLastSeen := udpconv.HelpLastSeen;

                                priv_interfacemaplist := mergeInterfaceLists (priv_interfacemaplist, udpconv.InterfaceList);
                           end;

//                      setlength (priv_filterlist, count);
//                      priv_filterlist[count-1] := filter;
                        priv_filterlist.AddFilter (filter);

                        if assigned (EventOnAddFilterToList) then
                           EventOnAddFilterToList (filter);

                        node := Tree.GetNextSelected (node);
                   end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuColumnIndexClick (Sender: TObject);

var item : TMenuItem;

begin
     item := Sender as TMenuItem;

     if item.Checked then
        priv_popupTree.Header.Columns.Items[item.Tag].Options := priv_popupTree.Header.Columns.Items[item.Tag].Options - [coVisible]
     else
         priv_popupTree.Header.Columns.Items[item.Tag].Options := priv_popupTree.Header.Columns.Items[item.Tag].Options + [coVisible];

     ShowColumnStatus (priv_popuptree);
end;

procedure TFormConversationSummary.MenuConversationConfigureDisplayClick (Sender: TObject);

begin
     FormConfigureConversationDisplay.ColorTable := priv_settings.TCPColors;
     FormConfigureConversationDisplay.ComboBoxInitialPage.ItemIndex := priv_settings.InitialPage;
     FormConfigureConversationDisplay.CheckBoxGoToPageEveryTime.Checked := priv_settings.RememberPage;
     FormConfigureConversationDisplay.CheckboxClearFilterOnPageChange.Checked := priv_settings.ClearFilterOnChange;


     FormConfigureConversationDisplay.PageControl.ActivePageIndex := 0;
     FormConfigureConversationDisplay.ShowModal;

     if FormConfigureConversationDisplay.ModalResult = mrOk then
        begin
             priv_settings.TCPColors := FormConfigureConversationDisplay.ColorTable;
             priv_settings.InitialPage := FormConfigureConversationDisplay.ComboBoxInitialPage.ItemIndex;
             priv_settings.RememberPage := FormConfigureConversationDisplay.CheckBoxGoToPageEveryTime.Checked;
             priv_settings.ClearFilterOnChange := FormConfigureConversationDisplay.CheckboxClearFilterOnPageChange.Checked;

             FormPreferences.StoreConversationListSettings (priv_settings);
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationCopyCellToClipboardClick (Sender: TObject);

begin
     Clipboard.AsText := priv_CellTextCache;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetMACEndpointCSVExportLine (Node: pVirtualNode): string;

var data  : pMACEndpointTreeEntry;
    sLine : string;

begin
     data := VSTMACEndpoints.GetNodeData (node);

     sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
              '"' + data^.Address.asString (':') +  '"' + csvseparator +
              '"' + IntToStr (data^.Frames) +  '"' + csvseparator +
              '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
              '"' + IntToStr (data^.FramesSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.FramesReceived) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesReceived) +  '"' + csvseparator +
              '"' + GetAttributeDescription (data^.Attributes) + '"';

     result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetIPv4EndpointCSVExportLine (Node: pVirtualNode): string;

var data  : pIPv4EndpointTreeEntry;
    sLine : string;

begin
     data := VSTIPv4Endpoints.GetNodeData (node);

     sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
              '"' + data^.Address.ValueAsString +  '"' + csvseparator +
              '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
              '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
              '"' + IntToStr (data^.PacketsSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.PacketsReceived) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesReceived) +  '"' + csvseparator +
              '"' + IntToStr (data^.PartnerCount) +  '"' + csvseparator +
              '"' + IntToStr (data^.Flowcount) +  '"' + csvseparator +
              '"' + GetAttributeDescription (data^.Attributes) + '"';

     result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetIPv6EndpointCSVExportLine (Node: pVirtualNode): string;

var data  : pIPv6EndpointTreeEntry;
    sLine : string;

begin
     data := VSTIPv6Endpoints.GetNodeData (node);

     sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
              '"' + data^.Address.ValueAsString +  '"' + csvseparator +
              '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
              '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
              '"' + IntToStr (data^.PacketsSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.PacketsReceived) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesSent) +  '"' + csvseparator +
              '"' + IntToStr (data^.BytesReceived) +  '"' + csvseparator +
              '"' + IntToStr (data^.PartnerCount) +  '"' + csvseparator +
              '"' + IntToStr (data^.Flowcount) +  '"' + csvseparator +
              '"' + GetAttributeDescription (data^.Attributes) + '"';

     result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetMACConversationCSVExportLine (Node: pVirtualNode): string;

var data  : pMACConversationTreeEntry;
    sLine : string;

begin
      data := VSTMACConversations.GetNodeData (node);

      sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
               '"' + data^.AddressA.asString (':') +  '"' + csvseparator +
               '"' + data^.AddressB.asString (':') +  '"' + csvseparator +
               '"' + IntToStr (data^.Frames) +  '"' + csvseparator +
               '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
               '"' + IntToStr (data^.FramesAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.FramesBtoA) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesBtoA) +  '"' + csvseparator +
               '"' + data^.DurationString + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2) +  '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2)  +  '"' + csvseparator +
               '"' + GetAttributeDescription (data^.Attributes) + '"';

      result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetIPv4ConversationCSVExportLine (Node: pVirtualNode): string;

var data  : pIPv4ConversationTreeEntry;
    sLine : string;
    rtt   : string;

begin
      data := VSTIPv4Conversations.GetNodeData (node);

      if data^.AverageRTTns > 0 then
         rtt := RTTtoStr (data^.AverageRTTns)
      else
          rtt := '';

      sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
               '"' + data^.AddressA.ValueAsString +  '"' + csvseparator +
               '"' + data^.AddressB.ValueAsString +  '"' + csvseparator +
               '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
               '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
               '"' + IntToStr (data^.PacketsAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.PacketsBtoA) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesBtoA) +  '"' + csvseparator +
               '"' + data^.DurationString + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2) + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2)  + '"' + csvseparator +
               '"' + rtt + '"' + csvseparator +
               '"' + IntToStr (data^.FlowCount) + '"' + csvseparator +
               '"' + GetAttributeDescription (data^.Attributes) + '"';

      result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetIPv6ConversationCSVExportLine (Node: pVirtualNode): string;

var data  : pIPv6ConversationTreeEntry;
    sLine : string;
    rtt   : string;

begin
      data := VSTIPv6Conversations.GetNodeData (node);

      if data^.AverageRTTns > 0 then
         rtt := RTTtoStr (data^.AverageRTTns)
      else
          rtt := '';

      sLine := '"' + intToStr (data^.index) +  '"' + csvseparator +
               '"' + data^.AddressA.ValueAsString +  '"' + csvseparator +
               '"' + data^.AddressB.ValueAsString +  '"' + csvseparator +
               '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
               '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
               '"' + IntToStr (data^.PacketsAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.PacketsBtoA) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesAtoB) +  '"' + csvseparator +
               '"' + IntToStr (data^.BytesBtoA) +  '"' + csvseparator +
               '"' + data^.DurationString + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2) + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2) + '"' + csvseparator +
               '"' + rtt + '"' + csvseparator +
               '"' + IntToStr (data^.FlowCount) + '"' + csvseparator +
               '"' + GetAttributeDescription (data^.Attributes) + '"';

      result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetTCPConversationCSVExportLine (Node: pVirtualNode): string;

var data  : pTCPConversationTreeEntry;
    sLine : string;

begin
      data := VSTTCPConversations.GetNodeData (node);

      sLine := '"' + intToStr (data^.index) + '"' + csvseparator;
{
      if data^.IPVersion = 4 then
         sLine := sLine + '"' + data^.AddressIPv4A.ValueAsString +  '"' + csvseparator +
                          '"' + intToStr (data^.PortA) +  '"' + csvseparator +
                          '"' + data^.AddressIPv4B.ValueAsString +  '"' + csvseparator +
                          '"' + intToStr (data^.PortB) +'"' +  csvseparator
      else
          sLine := sLine + '"' + data^.AddressIPv6A.ValueAsString +  '"' + csvseparator +
                           '"' + intToStr (data^.PortA) +  '"' + csvseparator +
                           '"' + data^.AddressIPv6B.ValueAsString +  '"' + csvseparator +
                           '"' + intToStr (data^.PortB) + '"' + csvseparator;
}
     sLine := sLine + '"' + data^.AddressA.ValueAsString +  '"' + csvseparator +
                       '"' + intToStr (data^.PortA) +  '"' + csvseparator +
                       '"' + data^.AddressB.ValueAsString +  '"' + csvseparator +
                       '"' + intToStr (data^.PortB) +'"' +  csvseparator;

      sLine := sLine + '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
                       '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsAtoB) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsBtoA) +  '"' + csvseparator +
                       '"' + IntToStr (data^.BytesAtoB) +  '"' + csvseparator +
                       '"' + IntToStr (data^.BytesBtoA) + '"' + csvseparator +
                       '"' + data.PacketRatioAtoB.ToString + '% / ' + (100 - data.PacketRatioAtoB).ToString + '%"' + csvseparator;

      if data^.InitialRTTns > 0 then
         sLine := sLine + '"' + RTTtoStr (data^.InitialRTTns) + '"' + csvseparator
      else
          sLine := sLine + '""' + csvseparator;

      sLine := sLine + '"' + data^.DurationString + '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2) +  '"' + csvseparator +
               '"' + FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2)  +  '"' + csvseparator +
               '"' + TCPFlagSetToTCPDump (data^.Flags) + '"' + csvseparator +
               '"' + DateTimeToStr (data^.FirstTimestamp.datetime)  +  '"' + csvseparator +
               '"' + GetTCPStatusDescription (data^.SessionStatus, data^.Attributes) +  '"';

      result := sLine;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.GetUDPConversationCSVExportLine (Node: pVirtualNode): string;

var data  : pUDPConversationTreeEntry;
    sLine : string;

begin
      data := VSTUDPConversations.GetNodeData (node);

      sLine := '"' + intToStr (data^.index) + '"' + csvseparator;

      if data^.AddressA.IPVersion = 4 then
         sLine := sLine + '"' + data^.AddressA.ValueAsString +  '"' + csvseparator +
                          '"' + intToStr (data^.PortA) +  '"' + csvseparator +
                          '"' + data^.AddressB.ValueAsString +  '"' + csvseparator +
                          '"' + intToStr (data^.PortB) + '"' + csvseparator
      else
          sLine := sLine + '"' + data^.AddressA.ValueAsString +  '"' + csvseparator +
                           '"' + intToStr (data^.PortA) +  '"' + csvseparator +
                           '"' + data^.AddressB.ValueAsString +  '"' + csvseparator +
                           '"' + intToStr (data^.PortB) + '"' + csvseparator;

      sLine := sLine + '"' + IntToStr (data^.Packets) +  '"' + csvseparator +
                       '"' + IntToStr (data^.Bytes) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsAtoB) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsBtoA) +  '"' + csvseparator +
                       '"' + IntToStr (data^.BytesAtoB) +  '"' + csvseparator +
                       '"' + IntToStr (data^.BytesBtoA) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsPerSecondAtoB) +  '"' + csvseparator +
                       '"' + IntToStr (data^.PacketsPerSecondBtoA) +  '"' + csvseparator +
                       '"' + FloatToStrF (data^.MBpSAtoB, ffNumber, 15, 2) +  '"' + csvseparator +
                       '"' + FloatToStrF (data^.MBpSBtoA, ffNumber, 15, 2)  +  '"' + csvseparator;

                       if data^.Packets > 1 then
                          sLine := sLine + '"' + data^.DurationString +  '"' + csvseparator
                       else
                           sLine := sLine + '""' + csvseparator;

                       if (data^.PacketsAtoB = 1) and (data^.PacketsBtoA = 1) then
                          sLine := sLine + '"' + FloatToStrF (data^.DeltaTime.toFloat, ffNumber, 15, 9) +  '"' + csvseparator
                       else
                           sLine := sLine + '""' + csvseparator;

                       sLine := sLine + '"' + GetAttributeDescription (data^.Attributes) + '"';

      result := sLine;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuExportToCSVClick (Sender: TObject);

var Filename  : string;
//  fHandle   : textfile;
    HelpArray : TArray<integer>;
    node      : pVirtualNode;
    sLine     : string;
    all       : TStringList;
    count     : integer;
    help      : double;
    iHelp     : integer;
    index     : integer;
    percent   : integer;
    boOkay    : boolean;
    i         : integer;

begin
     boOkay := false;

     if SaveTextFileExportCSV.Execute then
        begin
             boOkay := not FileExists (SaveTextFileExportCSV.FileName);

             if not boOkay then
                begin
                     boOkay := (MessageDlg ('File "' + SaveTextFileExportCSV.FileName + '" exists, overwrite?', mtConfirmation, [mbYes, mbNo], 0) = mrYes);

                     if boOkay then
                        begin
                             try
                                deletefile (SaveTextFileExportCSV.FileName);
                                boOkay := not FileExists (SaveTextFileExportCSV.FileName);

                             except
                                   boOkay := false;
                             end;

                             if not boOkay then
                                MessageDlg('Error: could not overwrite file, aborting', mtError, [mbOK], 0);
                        end;
                end;
        end;

     if boOkay then
        begin
             Filename := SaveTextFileExportCSV.FileName;
             all := TStringList.Create;

             if priv_popuptree.Name = 'VSTMACEndpoints' then
                count := VSTMACEndpoints.RootNodeCount;

             if priv_popuptree.Name = 'VSTIPv4Endpoints' then
                count := VSTIPv4Endpoints.RootNodeCount;

             if priv_popuptree.Name = 'VSTIPv6Endpoints' then
                count := VSTIPv6Endpoints.RootNodeCount;

             if priv_popuptree.Name = 'VSTMACConversations' then
                count := VSTMACConversations.RootNodeCount;

             if priv_popuptree.Name = 'VSTIPv4Conversations' then
                count := VSTIPv4Conversations.RootNodeCount;

             if priv_popuptree.Name = 'VSTIPv6Conversations' then
                count := VSTIPv6Conversations.RootNodeCount;

             if priv_popuptree.Name = 'VSTTCPConversations' then
                count := VSTTCPConversations.RootNodeCount;

             if priv_popuptree.Name = 'VSTUDPConversations' then
                count := VSTUDPConversations.RootNodeCount;

             // create header line

             sLine := '';

             for i := 1 to priv_popuptree.Header.Columns.Count do
                 begin
                      if i < priv_popuptree.Header.Columns.Count then
                         sLine := sLine + '"' + priv_popuptree.Header.Columns.Items[i-1].Text + '",'
                      else
                          sLine := sLine + '"' + priv_popuptree.Header.Columns.Items[i-1].Text + '"';
                 end;

             all.Add (sLine);

             // process rows

             if (priv_popuptree.Name = 'VSTTCPConversations') and (priv_tabvalues[TabTCPConversations].MarkedRowList.Count > 0) then
                begin
                     HelpArray := priv_tabvalues[TabTCPConversations].markedrowlist.Keys.ToArray;
                     TArray.Sort<integer>(HelpArray);

                     i := 0;

                     for index in HelpArray do
                         begin
                              inc (i);

                              // Status update

                              help := 100 * i / priv_tabvalues[TabTCPConversations].MarkedRowList.Count;
                              iHelp := trunc (Help);

                              if iHelp > percent then
                                 begin
                                      percent := iHelp;
                                      Statusbar.Panels[0].Text := 'Status: Storing CSV Lines (' + IntToStr (iHelp) + '% complete)';

                                      Application.ProcessMessages;
                                 end;

                              node := priv_tabvalues[TabTCPConversations].markedrowlist.Items[index];
                              sLine := GetTCPConversationCSVExportLine (node);
                              all.Add (sLine);
                         end;
                end
             else
                 begin
                      node := priv_popuptree.GetFirst;

                      percent := 0;
                      i := 0;

                      while node <> nil do
                            begin
                                 inc (i);

                                 // Status update

                                 help := 100 * i / count;
                                 iHelp := trunc (Help);

                                 if iHelp > percent then
                                    begin
                                         percent := iHelp;
                                         Statusbar.Panels[0].Text := 'Status: Storing CSV Lines (' + IntToStr (iHelp) + '% complete)';

                                         Application.ProcessMessages;
                                    end;

                                 if priv_popuptree.Name = 'VSTMACEndpoints' then
                                    sLine := GetMACEndpointCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTIPv4Endpoints' then
                                    sLine := GetIPv4EndpointCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTIPv6Endpoints' then
                                    sLine := GetIPv6EndpointCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTMACConversations' then
                                    sLine := GetMACConversationCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTIPv4Conversations' then
                                    sLine := GetIPv4ConversationCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTIPv6Conversations' then
                                    sLine := GetIPv6ConversationCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTTCPConversations' then
                                    sLine := GetTCPConversationCSVExportLine (node);

                                 if priv_popuptree.Name = 'VSTUDPConversations' then
                                    sLine := GetUDPConversationCSVExportLine (node);

                                 all.Add (sLine);

                                 node := priv_popuptree.GetNext(node);
                            end;
                 end;

             all.SaveToFile (Filename);
             all.Free;

             Statusbar.Panels[0].Text := '';

             Application.ProcessMessages;
        end;
end;

(* ------------------------------------------- *)

function TFormConversationSummary.ExtractToSingleFile (const Filename : string) : string;

var extractor : tTaskProcessorExtract;
    params    : tExtractionParameters;
{$ifdef debug}
    count     : integer;
{$endif}

begin
     priv_lastsourcefilename := '';

     params.initialize;

     params.TraceFileList := priv_filelist;
     params.FilterList := priv_filterlist;
     params.InterfaceTransferScope := itsKeepActive;
     params.ActiveInterfaceList := priv_interfacemaplist;
     params.ExtractionMode := xmSingleFile;
     params.OutputSchema := Filename;
     params.TotalExpectedFrameCount := priv_selectedstats.TotalFrames;
     params.WriteMode := determineWriteMode;

{$ifdef debug}
     count := priv_filterlist.count;
{$endif}

     Statusbar.Panels[0].Text := 'Status: Extracting frames';
     Application.ProcessMessages;

     extractor := tTaskProcessorExtract.create (priv_conversationstore, params);
     extractor.onNextFile := onTaskExtractNextFile;
     extractor.OnAdvancePercentageProcessed := onTaskExtractAdvancePercentProcessed;

     extractor.execute;

     extractor.free;

     Statusbar.Panels[0].Text := '';
     Statusbar.Panels[1].Text := '';

     if priv_lastsourcefilename <> '' then
        result := ExtractFilePath (priv_lastsourcefilename)
     else
         result := '';

     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ExtractToMultipleFiles (const Directory, TargetFile: string; const TruncateAfterLayer : byte);

var extractor : tTaskProcessorExtract;
    params    : tExtractionParameters;

begin
     priv_lastsourcefilename := '';

     params.initialize;

     params.TraceFileList := priv_filelist;
     params.FilterList := priv_filterlist;
     params.ActiveInterfaceList := priv_interfacemaplist;
     params.ExtractionMode := xmFilePerConversation;
     params.OutputSchema := Directory + TargetFile;
     params.TotalExpectedFrameCount := priv_selectedstats.TotalFrames;
     params.TruncateAfterLayer := TruncateAfterLayer;
//   params.WriteMode := priv_extractwritemode;

     Statusbar.Panels[0].Text := 'Status: Extracting frames';
     Application.ProcessMessages;

     extractor := tTaskProcessorExtract.create (priv_conversationstore, params);
     extractor.onNextFile := onTaskExtractNextFile;
     extractor.OnAdvancePercentageProcessed := onTaskExtractAdvancePercentProcessed;

     extractor.execute;

     extractor.free;

     Statusbar.Panels[0].Text := '';
     Statusbar.Panels[1].Text := '';
     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.ExtractPacketsAndOpen (Tree: tBaseVirtualTree);

var sFilename : string;
    sPath     : string;

begin
     if tree.SelectedCount = 1 then
        begin
             // Check Temp Path

             if not DirectoryExists (priv_temppath) or not IsDirectoryWriteable (priv_temppath) then
                begin
                     MessageDlg ('Please select a temporary path to write the file', mtInformation, [mbOK], 0);

                     sPath := BrowseDialog ('Select temporary path', BIF_RETURNONLYFSDIRS);

                     if not IsDirectoryWriteable (sPath) then
                        begin
                             MessageDlg ('Path does not allow creating files, aborting', mtError, [mbOK], 0);
                             exit;
                        end
                     else
                         priv_temppath := includeTrailingPathDelimiter (sPath);
                end;

             priv_filterlist.Clear;
             setLength (priv_interfacemaplist, 0);

             // Determine the filter for the current selection (stored in a object variable called "priv_filterlist", so no return value)

             getFilterAndInterfaceMappingsFromList (tree);

             if trim (TWWindowLabel) <> '' then
                sFilename := priv_temppath + 'tracewrangler_temp_' + TWWindowLabel + '.pcapng'
             else
                 sFilename := priv_temppath + 'tracewrangler_temp.pcapng';

             // Extract file (using the priv_filterlist variable, among others)

             SiMain.LogVerbose ('Exporting frames to "' + sFilename + '"');

             sPath := ExtractToSingleFile (sFilename);

             if sPath = '' then
                sPath := priv_temppath;

             SiMain.LogVerbose ('Running ShellExecute with filename "' + sFilename + '" at "' + sPath + '"');

             // Run whatever is configured to open PCAPng files (Wireshark at this time ignores the working directory, but
             // hopefull that'll change, see https://bugs.wireshark.org/bugzilla/show_bug.cgi?id=11908)

             ShellExecute (Handle, 'open', pChar ('"' + sFilename + '"'), nil, pChar ('"' + sPath + '"'), SW_SHOWNORMAL);
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationExtractToSingleFileClick (Sender: TObject);

var boOkay : boolean;

begin
//   setLength (priv_filterlist, 0);
     priv_filterlist.Clear;
     setLength (priv_interfacemaplist, 0);

     getFilterAndInterfaceMappingsFromList (priv_popuptree);

//   if length (priv_filterlist) > 0 then
     if priv_filterlist.count > 0 then
        begin
             // Get output filename

             if SaveDialogExtract.Execute then
                begin
                     boOkay := true;

                     if FileExists (SaveDialogExtract.FileName) then
                        boOkay := (MessageDlg('Do you want to overwrite "' + SaveDialogExtract.FileName + '"?', mtConfirmation, [mbYes, mbNo], 0) = mrYes);

                     if boOkay then
                        ExtractToSingleFile (SaveDialogExtract.FileName);
                end;
        end;

     Statusbar.Panels[0].Text := '';
     Statusbar.Panels[1].Text := '';
     Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationExtractToMultipleFilesClick (Sender: TObject);

var fileentry : tTraceFileEntry;
    extractor : tTaskProcessorExtract;
    params    : tExtractionParameters;
    layer     : byte;

begin
     priv_filterlist.Clear;
     getFilterAndInterfaceMappingsFromList (priv_popuptree);

     if priv_filterlist.count > 0 then
        begin
             fileentry := priv_filelist.getFile (1);

             if FormMultiFileExtractSettings.EditTargetDirectory.Text = '' then
                FormMultiFileExtractSettings.EditTargetDirectory.Text := fileentry.FilePath + 'extract\';

             FormMultiFileExtractSettings.showmodal;

             if FormMultiFileExtractSettings.ModalResult = mrOk then
                begin
                     if FormMultiFileExtractSettings.CheckBoxTruncateAfter.Checked then
                        layer := FormMultiFileExtractSettings.ComboBoxTruncateAfterLayer.ItemIndex + 2
                     else
                         layer := 0;

                     ExtractToMultipleFiles (FormMultiFileExtractSettings.EditTargetDirectory.Text, FormMultiFileExtractSettings.EditFileFormat.Text, layer);
                end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationOpenClick (Sender: TObject);

begin
     ExtractPacketsAndOpen (priv_popuptree);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationViewTCPAggregatedModeClick (Sender: TObject);

begin
     ToggleAggregatedMode;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationViewTCPHighlightModeClick (Sender: TObject);

begin
     ToggleHighlightMode;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationMarkRowClick (Sender: TObject);

begin
     ProcessTCPNodes (paMarkRows);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationGotoRowClick (Sender: TObject);

begin
     GotoRow;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationLookupIPAPIClick (Sender: TObject);

begin
     OpenWebBrowser ('http://ip-api.com/#' + priv_CellTextCache);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationPreviousMarkClick (Sender: TObject);

begin
     JumpToTCPMark (jmPrevious);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationNextMarkClick (Sender: TObject);

begin
     JumpToTCPMark (jmNext);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuConversationClearAllMarksClick (Sender: TObject);

var node : pVirtualNode;
    data : pTCPConversationTreeEntry;
    key  : integer;

begin
     for key in priv_tabvalues[TabTCPConversations].markedrowlist.Keys do
         begin
              node := priv_tabvalues[TabTCPConversations].markedrowlist.Items[key];
              data := VSTTCPConversations.GetNodeData (node);

              // Set the display item to unmarked

              data^.Marked := False;
              priv_TCPList[data^.index-1].Marked := False;
         end;

     priv_tabvalues[TabTCPConversations].markedrowlist.Clear;

     // clear the last mark to avoid trouble

     priv_tabvalues[TabTCPConversations].currentmarkedrow := 0;

     VSTTCPConversations.Repaint;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.MenuAddAsMergeFilterClick (Sender: TObject);

begin
     // We probably only need the filter, but the interface list is done at the same time

     getFilterAndInterfaceMappingsFromList (priv_popuptree);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onPopupHeaderColumnSelector (Sender: TObject);

var Item : tMenuItem;
    i    : integer;

begin
     if (Sender as tPopupMenu).PopupComponent <> nil then
        begin
             priv_popupTree := ((Sender as tPopupMenu).PopupComponent) as tBaseVirtualTree;

             // clear popup menu

             for i := 1 to PopupMenuHeaderColumnSelector.Items.Count do
                 PopupMenuHeaderColumnSelector.Items[0].Free;

             // Build the menu

             for i := 1 to priv_popuptree.Header.Columns.Count do
                 begin
                      Item := TMenuItem.Create(PopupMenuHeaderColumnSelector);
                      Item.Caption := priv_popuptree.Header.Columns[i-1].Text;
                      Item.Checked := coVisible in priv_popuptree.Header.Columns[i-1].Options;
                      Item.Tag := i-1;

                      Item.OnClick := MenuColumnIndexClick;

                      PopupMenuHeaderColumnSelector.Items.Add (item);
                 end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onStatusbarDrawPanel (StatusBar: TStatusBar; Panel: TStatusPanel; const Rect: TRect);

begin
     if Panel = StatusBar.Panels[3] then
        with LabeledEditFilter do begin
             Top := Rect.Top;
             Left := Rect.Left + LabeledEditFilter.EditLabel.Width + 5;
             Width := (Rect.Right - Rect.Left - 15) - (LabeledEditFilter.EditLabel.Width + 5);
             Height := Rect.Bottom - Rect.Top;
        end;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onKeyPressFilter(Sender: TObject; var Key: Char);

begin
     // #1 is CTRL-A, #3 is CTRL-C, #8 is tab, #9 is backspace, #27 is Escape, #$16 is CTRL-V, #$1a is CTRL-Z

     if not (key in [#1, #3, #8, #9, #27, #32, #$16, #$1a, '!', '0'..'9','a'..'z','A'..'Z', '.', ':']) then
        key := #0;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onKeyUpFilter(Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if Key = 27 then
        LabeledEditFilter.Clear;

     TimerApplyFilter.Enabled := false;
     TimerApplyFilter.Enabled := true;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onTimerApplyFilter(Sender: TObject);

var filtertext : string;
    DoApply    : boolean;

begin
     filtertext := Trim (LowerCase (LabeledEditFilter.Text));

     if filtertext = '' then
        begin
             // Clear socket filters

             priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentSocketFilter.initialize;
        end;

     // Check if filter is not complete yet or unchanged

     DoApply := (filtertext <> '!') and (filtertext <> '') and (filtertext <> priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr);

     // Apply if filter was cleared

     if not DoApply and ((filtertext = '') and (filtertext <> priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr)) then
        DoApply := True;

     if DoApply then
        begin
             Statusbar.Panels[1].Text := 'Applying filter';
             Application.ProcessMessages;

             // Check if this is a socket filter

             if pos ('!', filtertext) = 1 then
                priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentSocketFilter.initialize (rightstring (filtertext, 2))
             else
                 priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentSocketFilter.initialize (filtertext);
{
             if pos ('!', filtertext) = 1 then
                begin
                     if isValidIPv4Socket (rightstring (filtertext, 2)) then
                        priv_tabvalues[PageControlConversations.ActivePageIndex].currentIPv4Filter.initialize (rightstring (filtertext, 2))
                     else
                         if isValidIPv6Socket (rightstring (filtertext, 2)) then
                            priv_tabvalues[PageControlConversations.ActivePageIndex].currentIPv6Filter.initialize (rightstring (filtertext, 2));
                end
             else
                 begin
                      if isValidIPv4Socket (filtertext) then
                         priv_tabvalues[PageControlConversations.ActivePageIndex].currentIPv4Filter.initialize (filtertext)
                     else
                         if isValidIPv6Socket (filtertext) then
                            priv_tabvalues[PageControlConversations.ActivePageIndex].currentIPv6Filter.initialize (filtertext);
                 end;
}
             priv_tabvalues[PageControlConversations.ActivePageIndex].CurrentFilterStr := filtertext;

             // Only update the list that is shown

             case PageControlConversations.ActivePageIndex of
              0 : DisplayMACEndpoints;
              1 : DisplayIPv4Endpoints;
              2 : DisplayIPv6Endpoints;
              3 : DisplayMACConversations;
              4 : DisplayIPv4Conversations;
              5 : DisplayIPv6Conversations;
              6 : DisplayTCPConversations;
              7 : DisplayUDPConversations;
             end;
        end;

        Statusbar.Panels[1].Text := '';
        Application.ProcessMessages;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.onTimerDisplayAll (Sender: TObject);

begin
     TimerDisplayAll.Enabled := false;

     DisplayConversationStore;
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.OpenWebBrowser(const URL: string);

begin
     ShellExecute (self.WindowHandle, 'open', PChar(URL), nil, nil, SW_SHOWNORMAL);
end;

(* ------------------------------------------- *)

procedure TFormConversationSummary.DisplayConversationStore;

begin
     SiMain.EnterMethod ('TFormConversationSummary.DisplayConversationStore');

     DisplayMACEndpoints;
     DisplayIPv4Endpoints;
     DisplayIPv6Endpoints;
     DisplayMACConversations;
     DisplayIPv4Conversations;
     DisplayIPv6Conversations;
     DisplayTCPConversations;
     DisplayUDPConversations;
{$ifdef fqdnstats}
     DisplayNameResolutions;
{$endif}

     SiMain.LeaveMethod ('TFormConversationSummary.DisplayConversationStore');
end;

(* ------------------------------------------- *)


end.

