{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit dialog_rename_files;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ComCtrls,
  base_structures,
  class_tracefilelist;

type
  TFormRenameFiles = class(TForm)
    StatusBar: TStatusBar;
    GroupBox1: TGroupBox;
    LabelNewNamePattern: TLabel;
    EditNewNamePattern: TEdit;
    MemoPlaceholderHint: TMemo;
    Label1: TLabel;
    EditCounterStart: TEdit;
    UpDownCounterStart: TUpDown;
    LabelCounterIncrement: TLabel;
    EditCounterIncrement: TEdit;
    UpDownCounterIncrement: TUpDown;
    LabelCounterDigits: TLabel;
    ComboBoxCounterDigits: TComboBox;
    LabelTimestamp: TLabel;
    ComboBoxTimestampSource: TComboBox;
    LabelTimestampFormat: TLabel;
    ComboBoxTimestampFormat: TComboBox;
    LabelExamplePreview: TLabel;
    LabelValueExamplePreview: TLabel;
    ButtonOkay: TButton;
    ButtonCancel: TButton;
    procedure onShow(Sender: TObject);
    procedure onResize(Sender: TObject);
    procedure onChangeValues(Sender: TObject);
    procedure onKeyPressPattern(Sender: TObject; var Key: Char);
    procedure onCreate(Sender: TObject);
  private
         priv_filelist : tTraceFileList;
         priv_preview  : string;

         function getCounterSettings : tCounterStruct;
         function getTimestampMode : tFileTimestampMode;
         procedure updatePreview;
  public
        property TraceFileList : tTraceFileList read priv_filelist write priv_filelist;

        function getParams : tRenameParams;
  end;

var
  FormRenameFiles: TFormRenameFiles;

implementation

uses Toolkit;

{$R *.dfm}

(* ------------------------------------------ *)
{ TFormRenameFiles }
(* ------------------------------------------ *)

procedure TFormRenameFiles.onShow(Sender: TObject);

begin
     updatePreview;
end;

(* ------------------------------------------ *)

procedure TFormRenameFiles.onResize (Sender: TObject);

var len : integer;

begin
     len := LabelValueExamplePreview.Width div 6;

     LabelValueExamplePreview.Caption := mince (priv_preview , len);
end;

(* ------------------------------------------ *)

procedure TFormRenameFiles.onChangeValues(Sender: TObject);

begin
     updatePreview;
end;

(* ------------------------------------------ *)

procedure TFormRenameFiles.onCreate(Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');
end;

(* ------------------------------------------ *)

procedure TFormRenameFiles.onKeyPressPattern(Sender: TObject; var Key: Char);

begin
     // #1 is CTRL-A, #3 is CTRL-C, #8 is tab, #9 is backspace, #$16 is CTRL-V, #$1a is CTRL-Z

     if not (key in [#1, #3, #8, #9, #$16, #$1a, ' ', 'a'..'z', 'A'..'Z', '0'..'9', '-', '_', '#', '[', ']', '(', ')', '<', '>']) then
        Key := #0;
end;

(* ------------------------------------------ *)

function TFormRenameFiles.getCounterSettings: tCounterStruct;

var value : tCounterStruct;

begin
     value.initialize;

     value.CountStart := UpDownCounterStart.Position;
     value.CountStep := UpDownCounterIncrement.Position;
     value.Digits := ComboBoxCounterDigits.ItemIndex + 1;
     value.Current := value.CountStart;

     result := value;
end;

(* ------------------------------------------ *)

function TFormRenameFiles.getTimestampMode: tFileTimestampMode;

begin
     result := tsmNone;

     case ComboBoxTimestampSource.ItemIndex of
      0 : result := tsmFirstFrame;
      1 : result := tsmLastFrame;
      2 : result := tsmOriginalFile;
     end;
end;

(* ------------------------------------------ *)

procedure TFormRenameFiles.updatePreview;

var entry   : tTraceFileEntry;
    params  : tRenameParams;

begin
     if priv_filelist.Count > 0 then
        begin
             entry := priv_filelist.getFile(1);

             params.FilePattern := EditNewNamePattern.Text;
             params.Counter := getCounterSettings;
             params.TimestampMode := getTimestampMode;
             params.TimestampPatternIndex := ComboBoxTimestampFormat.ItemIndex;


             priv_preview  := entry.determineRenamedFilename (params);

             // Yeah, I know... bad habit

             onResize (nil);
        end
     else
         LabelValueExamplePreview.Caption := 'n/a';

     // Check if we have a valid filename without brackets

     if (pos ('<', priv_preview) > 0) or (pos ('>', priv_preview) > 0) then
        begin
             EditNewNamePattern.Font.Color := clRed;
             ButtonOkay.Enabled := false
        end
     else
         begin
              EditNewNamePattern.Font.Color := clWindowText;
              ButtonOkay.Enabled := True;
         end;
end;

(* ------------------------------------------ *)

function TFormRenameFiles.getParams: tRenameParams;

var value : tRenameParams;

begin
     value.initialize;

     value.FilePattern := EditNewNamePattern.Text;
     value.Counter := getCounterSettings;
     value.TimestampMode := getTimestampMode;
     value.TimestampPatternIndex := ComboBoxTimestampFormat.ItemIndex;

     result := value;
end;

(* ------------------------------------------ *)

end.
