{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit dialog_selecttask;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls;

type
  TFormSelectTask = class(TForm)
    ComboBoxSelectTask: TComboBox;
    Label1: TLabel;
    MemoTaskInfo: TMemo;
    ButtonOkay: TButton;
    ButtonCancel: TButton;
    EditTaskname: TEdit;
    LabelTaskName: TLabel;
    procedure onChangeFields(Sender: TObject);
    procedure onFormShow(Sender: TObject);
    procedure onKeyPressEditTaskname(Sender: TObject; var Key: Char);
    procedure onCreate(Sender: TObject);
    procedure onDestroy(Sender: TObject);
    procedure onExitTaskname(Sender: TObject);
  private
         priv_NameSetByTaskSelect : boolean;
         priv_ExistingTaskNames   : tStringList;

         procedure ValidateName (var TaskName : string);

  public
        procedure ClearTaskNames;
        procedure AddTaskName (Taskname : string);
  end;

var
  FormSelectTask: TFormSelectTask;

implementation

{$R *.dfm}

uses Toolkit;

(* -------------------------------------- *)

procedure TFormSelectTask.onCreate(Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     priv_ExistingTaskNames := tStringlist.Create;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.onDestroy(Sender: TObject);

begin
     priv_ExistingTaskNames.free;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.onFormShow(Sender: TObject);

begin
     EditTaskname.SetFocus;

     priv_NameSetByTaskSelect := false;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.AddTaskName (Taskname: string);

begin
     priv_ExistingTaskNames.Add (lowercase (TaskName));
end;

(* -------------------------------------- *)

procedure TFormSelectTask.ClearTaskNames;

begin
     priv_ExistingTaskNames.Clear;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.ValidateName (var TaskName: string);

var i : integer;

begin
     if priv_ExistingTaskNames.IndexOf (lowercase (TaskName)) > -1 then
        begin
             i := 1;

             repeat
                   TaskName := leftstring (Taskname, ' (');
                   TaskName := TaskName + ' (' + IntToStr (i) + ')';

                   inc (i);

             until (priv_ExistingTaskNames.IndexOf (lowercase (TaskName)) = -1);
        end;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.onKeyPressEditTaskname(Sender: TObject; var Key: Char);

begin
     if not (key in [#8, #9, '-', ' ', '_', 'a' .. 'z', 'A' .. 'Z', '0'..'9']) then
        key:=#0
     else
         priv_NameSetByTaskSelect := false;
end;

(* -------------------------------------- *)

procedure TFormSelectTask.onChangeFields(Sender: TObject);

var Taskname : string;

begin
     TaskName := '';

     case ComboBoxSelectTask.ItemIndex of
      0 : begin
               MemoTaskInfo.Text := 'Select a task to process the list of trace files.';
          end;
      1 : begin
               MemoTaskInfo.Text := 'Sanitize trace files and remove sensitive information, like IP addresses and other packet details.';

               if (EditTaskname.Text = '') or priv_NameSetByTaskSelect then
                  begin
                       TaskName := 'Anonymize Files';
                       priv_NameSetByTaskSelect := true;
                  end;
          end;
      2 : begin
               MemoTaskInfo.Text := 'Extract specific conversations into new files';

               if (EditTaskname.Text = '') or priv_NameSetByTaskSelect then
                  begin
                       TaskName := 'Extract Files';
                       priv_NameSetByTaskSelect := true;
                  end;
          end;
      3 : begin
               MemoTaskInfo.Text := 'Edit trace files elements, for example packet order, PCAPng headers, packet layer removal or insertion etc.';

               if (EditTaskname.Text = '') or priv_NameSetByTaskSelect then
                  begin
                       TaskName := 'Edit Files';
                       priv_NameSetByTaskSelect := true;
                  end;
          end;
      4 : begin
               MemoTaskInfo.Text := 'Merges trace files.';

               if (EditTaskname.Text = '') or priv_NameSetByTaskSelect then
                  begin
                       TaskName := 'Merge Files';
                       priv_NameSetByTaskSelect := true;
                  end;
          end;
      5 : begin
               MemoTaskInfo.Text := 'Splits a PCAPng File with multiple sections into single PCAPng files again.';

               if (EditTaskname.Text = '') or priv_NameSetByTaskSelect then
                  begin
                       TaskName :=  'Separate Sections';
                       priv_NameSetByTaskSelect := true;
                  end;
          end;
     end;

     if TaskName <> '' then
        begin
             ValidateName (TaskName);
             EditTaskname.Text := TaskName;
        end;

     ButtonOkay.Enabled := (ComboBoxSelectTask.ItemIndex > 0) and (Trim (EditTaskname.Text) <> '');
end;

(* -------------------------------------- *)

procedure TFormSelectTask.onExitTaskname (Sender: TObject);

begin
     if priv_ExistingTaskNames.IndexOf (lowercase (EditTaskName.Text)) > -1 then
        begin
             EditTaskName.Font.Color := clRed;
             ButtonOkay.Enabled := false;
        end
     else
         begin
              EditTaskName.Font.Color := clWindowText;
              ButtonOkay.Enabled := true;
         end;
end;

(* -------------------------------------- *)

end.
