{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_metaupdate;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorMetaUpdate = class (tFrameProcessor)

                                 private
                                        priv_fixframelengthmeta        : boolean;
                                        priv_forcefixframelengthmeta   : boolean;

                                 public
                                       property FixFrameLengthMeta : boolean read priv_fixframelengthmeta write priv_fixframelengthmeta;
                                       property FixFrameLengthWhenNotHardSliced : boolean read priv_forcefixframelengthmeta write priv_forcefixframelengthmeta;

                                       function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;
                          end;

implementation

uses class_parser_baseclass,
     common_parse_helpers,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parsemaster;

(* ------------------------------------------- *)
{ tFrameProcessorMetaUpdate }
(* ------------------------------------------- *)

function tFrameProcessorMetaUpdate.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var ParseMaster  : tParseMaster;
    NewFrame     : tRawFrame;
    FixSize      : integer;
    IPv4Parser   : tIPv4Parser;
    IPv6Parser   : tIPv6Parser;

begin
     NewFrame := nil;

     // We only need to check this if the Wiresize is not already larger than the capture size

    if priv_forcefixframelengthmeta or (Frame.FrameSize = Frame.WireSize) then
        begin
             ParseMaster := tParseMaster.create(Frame, ppmEdit);

             // Check if we have hard slicing we can fix

             if (ppIPv4 in ParseMaster.ParsedProtocols) then
                begin
                     // should be enough to check any of the IPv4 parsers because if the frame has a wiresize that is too short
                     // it'll be too short for all of them

                     IPv4Parser := ParseMaster.getParserByProtocol (ppIPv4) as tIPv4Parser;

                     // Set correct wiresize if required. The delta value is negative if the frame is shorter than it should be.

                     if IPv4Parser.FramePayloadLengthDelta <> 0 then
                        Frame.WireSize := IPv4Parser.FrameOffset + IPv4Parser.TotalLength;
                end;

             // Free up the parsemaster

             ParseMaster.free;
        end;

     if NewFrame = nil then
        result := Frame
     else
         result := NewFrame;
end;

(* ------------------------------------------- *)

end.
