{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_multiwriter;

interface

uses System.Generics.Collections,
     base_structures,
     class_rawframe,
     class_tracewriter,
     class_writer_pcapng,
     class_parsemaster,
     class_interfacelist,
     class_frameprocessor_baseclass,
     class_parse_snortalertlog,
     class_conversationstorage;

type tMultiWriteParams = record
                               FileSchema                  : string;
//                             WriteMode                   : tMultiWriteMode;
                               ConversationStore           : tConversationStorage;
                               InterfaceList               : tCaptureInterfaceList;
                               MaxFrameCountBeforeWrite    : integer;
                               FileTimestampMode           : tFileTimestampMode;
                               AbsoluteTimestamp           : tDateTime;
                               WriteSnortInfoToFileComment : boolean;                  // TODO: Implement this
                               TruncateAfterLayer          : byte;

                               procedure initialize;
                         end;

     tMultiWriteStatistics = record
                                   ActiveWriterCount    : integer;
                                   MaxWriterCount       : integer;
                                   FilesCreated         : integer;
                                   FramesWritten        : int64;
//                                 MetaInfoSetSize      : integer;
//                                 MetaInfoMatchCount   : integer;
//                                 MetaInfoLastMatch    : tMetaDataEntry;

                                   procedure initialize;
                             end;

     tOnMultiWriterChange = procedure (Sender : tObject; const Statistics : tMultiWriteStatistics) of Object;

     tFrameProcessorMultiWriter = class (tFrameProcessor)

                                        constructor create (const Name : string; const Params : tMultiWriteParams);

                                        private
                                               priv_OnMultiWriterChange      : tOnMultiWriterChange;
//                                             priv_OnMetaInfoMatch          : tOnMetaInfoMatch;

                                               priv_tracewriters             : tDictionary<string, tPCAPNGWriter>;
//                                             priv_metadataset              : tDictionary<string, tMetaDataEntry>;
                                               priv_conversationstore        : tConversationStorage;

                                               priv_WriteMode                : tMultiWriteMode;
                                               priv_fileschema               : string;
                                               priv_interfacelist            : tCaptureInterfaceList;
                                               priv_outfileindex             : integer;
                                               priv_writesnortinfotocomment  : boolean;
                                               priv_truncateAfterLayer       : byte;
//                                             priv_currentRowIndex          : integer;           no idea how to get this at this time, so disabled for now

//                                             priv_frame                    : tRawFrame;
//                                             priv_parser                   : tParseMaster;

                                               priv_absolutetimestamp        : tDateTime;
                                               priv_timestampmode            : tFileTimestampMode;

                                               priv_multiwritestats          : tMultiWriteStatistics;

                                               priv_currentInputFilename     : string;

                                               /// <summary>
                                               ///   Sets the new file schema
                                               ///   string, and calculates the
                                               ///   current write mode (which
                                               ///   is required to determine
                                               ///   when a writer object can
                                               ///   be closed)
                                               /// </summary>
                                               /// <param name="FileSchema">
                                               ///   New file naming schema
                                               /// </param>
                                               procedure setNewFileSchema (const FileSchema : string);

                                               /// <summary>
                                               ///   Replaces all the
                                               ///   placeholders to create the
                                               ///   final output filename.
                                               ///   Placeholders not relevant
                                               ///   to the given WriteMode
                                               ///   will be removed by
                                               ///   replacing them with an
                                               ///   empty string
                                               /// </summary>
                                               /// <param name="Frame">
                                               ///   Frame to create the
                                               ///   filename for
                                               /// </param>
                                               function getOutputFileName (const Frame : tRawFrame) : string;
                                               /// <summary>
                                               ///   This routine checks if a
                                               ///   write for the conversation
                                               ///   is already open. Otherwise
                                               ///   a new writer object is
                                               ///   created.
                                               /// </summary>
                                               /// <param name="Filename">
                                               ///   Generated from the frame
                                               ///   using the file name schema
                                               /// </param>
                                               function getWriterForFrame (const Filename : string) : tPCAPNGWriter;
//                                             function getWriterForKey (const Key : string; const CurrentWriteMode : tMultiWriteMode) : tPCAPNGWriter;
                                               procedure FinalizeWriter (Writer : tPCAPNGWriter);
{
                                               procedure EnhanceFrameWithMetaData;

                                               procedure writeEtherconversationFrame (const Frame : tRawFrame);
                                               procedure writeIPv4conversationFrame (const Frame : tRawFrame);
                                               procedure writeIPv6conversationFrame (const Frame : tRawFrame);
                                               procedure writeTCPconversationFrame (const Frame : tRawFrame);
                                               procedure writeUDPconversationFrame (const Frame : tRawFrame);
                                               procedure writeAutoFrame (const Frame : tRawFrame);
}
                                               /// <summary>
                                               ///   This function truncates
                                               ///   the frame if required,
                                               ///   destroying the original
                                               ///   frame and returning a
                                               ///   newly contructed frame
                                               /// </summary>
                                               /// <param name="Frame">
                                               ///   frame to be truncated
                                               /// </param>
                                               function getTruncatedFrame (Frame : tRawFrame) : tRawFrame;

                                               /// <summary>
                                               ///   Determines the correct
                                               ///   writer object based on the
                                               ///   filename scheme, writes
                                               ///   the frame, and checks if
                                               ///   the writer can be closed
                                               ///   depending on conversation
                                               ///   details (matching on "last
                                               ///   frame")
                                               /// </summary>
                                               /// <param name="Frame">
                                               ///   Frame to be written
                                               /// </param>
                                               procedure AddFrameToWriter (const Frame : tRawFrame);
                                        public
                                              destructor Destroy; override;

                                              {$REGION 'Documentation'}
                                              /// <summary>
                                              ///   Schema for the file name
                                              ///   created for each writer.
                                              ///   Can contain placeholders
                                              ///   for values like IP
                                              ///   addresses etc.
                                              /// </summary>
                                              /// <remarks>
                                              ///   Directory and File Format
                                              ///   fields will accept the
                                              ///   following placeholders:
                                              ///   <br /><br />&lt;filename&gt;
                                              ///    name of the source file,
                                              ///   without extension and path
                                              ///   <br />&lt;timestamp&gt;
                                              ///   time of the first packet
                                              ///   <br />&lt;sourcemac&gt;
                                              ///   Source Ethernet address
                                              ///   <br />&lt;destinationmac&gt;
                                              ///    Destination Ethernet
                                              ///   address<br />&lt;sourceip&
                                              ///   gt; Source IPv4 or IPv6
                                              ///   address<br />&lt;
                                              ///   destinationip&gt;
                                              ///   Destination IPv4 or IPv6
                                              ///   address<br />&lt;sourceport&
                                              ///   gt; Source UDP or TCP port
                                              ///   <br />&lt;destinationport&
                                              ///   gt; Destination UDP or TCP
                                              ///   port<br />&lt;index&gt;
                                              ///   index of the conversation
                                              ///   <br />&lt;count&gt; counter
                                              ///   starting at 1, going up for
                                              ///   each new file<br />
                                              /// </remarks>
                                              {$ENDREGION}
                                              property TargetFileSchema : string read priv_fileschema write setNewFileSchema;
                                              {$REGION 'Documentation'}
                                              /// <summary>
                                              ///   Value that holds the
                                              ///   current file the frames are
                                              ///   being read from. The
                                              ///   multiwriter needs to know
                                              ///   this in case the input file
                                              ///   name is used as a
                                              ///   placeholder for the output
                                              ///   file name
                                              /// </summary>
                                              {$ENDREGION}
                                              property CurrentInputFilename : string read priv_currentInputFilename write priv_currentInputFilename;
                                              property TruncateAfterLayer : byte read priv_truncateAfterLayer write priv_truncateAfterLayer;
                                              property Statistics : tMultiWriteStatistics read priv_multiwritestats;

//                                            procedure AddMetaDataSet (DataSet : tMetaDataSet);
                                              procedure AddFrame (Frame : tRawFrame); override;
                                  end;

implementation

uses SiAuto,
     SysUtils,
     DateUtils,
     Toolkit,
     definition_conversation_types,
     definition_ip_addresses,
     definition_tcp_structures;

const UnspecifiedFileMarker = 'UNSPECIFIEDFILENAME';

(* ------------------------------------------- *)
{ tMultiWriteParams }
(* ------------------------------------------- *)

procedure tMultiWriteParams.initialize;

begin
     FileSchema := '';
//   WriteMode := mwmNone;
//   ConversationStore := nil;
     InterfaceList := nil;
     MaxFrameCountBeforeWrite := 10000;

     FileTimestampMode := tsmNone;
     AbsoluteTimestamp := 0;

     WriteSnortInfoToFileComment := true;
end;

(* ------------------------------------------- *)
{ tMultiWriteStatistics }
(* ------------------------------------------- *)

procedure tMultiWriteStatistics.initialize;

begin
     ActiveWriterCount := 0;
     MaxWriterCount := 0;
     FilesCreated := 0;
     FramesWritten := 0;

//   MetaInfoSetSize := 0;
//   MetaInfoMatchCount := 0;
//   MetaInfoLastMatch.initialize;
end;

(* ------------------------------------------- *)
{ tFrameProcessorMultiWriter }
(* ------------------------------------------- *)

constructor tFrameProcessorMultiWriter.create (const Name: string; const Params : tMultiWriteParams);

begin
     inherited create (Name, nil, Params.MaxFrameCountBeforeWrite);

     setNewFileSchema (Params.FileSchema);
     priv_tracewriters := tDictionary<string, tPCAPNGWriter>.create;
//   priv_metadataset := tDictionary<string, tMetaDataEntry>.create;
//   priv_writemode := Params.WriteMode;
//   priv_currentRowIndex := 0;

     priv_conversationstore := Params.ConversationStore;
     priv_interfacelist := Params.InterfaceList;

     priv_absolutetimestamp := Params.AbsoluteTimestamp;
     priv_timestampmode := Params.FileTimestampMode;

     priv_writesnortinfotocomment := Params.WriteSnortInfoToFileComment;
     priv_truncateAfterLayer := Params.TruncateAfterLayer;

     priv_outfileindex := 0;
//   priv_frame := nil;
//   priv_parser := nil;

     priv_multiwritestats.initialize;

     // make it obvious if values that should have been initialized aren't

     priv_currentInputFilename := UnspecifiedFileMarker;
end;

(* ------------------------------------------- *)

destructor tFrameProcessorMultiWriter.destroy;

var writer : tTraceWriter;
    key    : string;

begin
     for key in priv_tracewriters.Keys do
         begin
              writer := priv_tracewriters.Items[key];

              if writer <> nil then
                 begin
                      writer.Flush;
                      writer.Close;
                      writer.free;
                      writer := nil;
                 end;
         end;

     priv_tracewriters.Free;
//   priv_metadataset.Free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.setNewFileSchema (const FileSchema: string);

begin
     priv_fileschema := FileSchema;
     priv_WriteMode := mwmNone;

     // Check if we're using MACs in the file name - if so, it's at least Ether Endpoint

     if pos ('<sourcemac>', priv_fileschema) > 0 then
        priv_WriteMode := mwmEtherEndpoint;

     // If both MACs are used, it's Ether Conversation

     if pos ('<destinationmac>', priv_fileschema) > 0 then
        begin
             if priv_WriteMode = mwmEtherEndpoint then
                priv_WriteMode := mwmEtherEndpoint
             else
                 priv_WriteMode := mwmEtherConversation;
        end;

     // Same for IPs

     if pos ('<sourceip>', priv_fileschema) > 0 then
        priv_WriteMode := mwmIPEndpoint;

     if pos ('<destinationip>', priv_fileschema) > 0 then
        begin
             if priv_WriteMode = mwmIPEndpoint then
                priv_WriteMode := mwmIPConversation
             else
                 priv_WriteMode := mwmIPEndpoint;
        end;

     // If we have ports then it's L4

     if (pos ('<sourceport>', priv_fileschema) > 0) or (pos ('<destinationport>', priv_fileschema) > 0) then
        priv_writemode := mwmL4Conversation;
end;

(* ------------------------------------------- *)
{
procedure tFrameProcessorMultiWriter.EnhanceFrameWithMetaData;

var entry  : tMetaDataEntry;
    l4prot : tLayer4Protocol;
    key    : string;

begin
     l4prot := l4Any;

     if ppTCP in priv_parser.ParsedProtocols then
        l4prot := l4TCP
     else
     if ppUDP in priv_parser.ParsedProtocols then
        l4prot := l4UDP
     else
     if (ppICMPv4 in priv_parser.ParsedProtocols) or (ppICMPv6 in priv_parser.ParsedProtocols) then
        l4prot := l4ICMP;

     // To avoid commenting TCP ACK packets we require a non-zero TCP payload lenght, or no TCP

     if not ((l4prot = l4TCP) and (tcpACK in priv_parser.TCPFlags) and (priv_parser.TCPPayloadLength = 0)) then
        begin
             // We don't know the SID (as we're matching on flows), so using UNKNOWN_SID

             if priv_parser.FirstIPLayerVersion = 4 then
                key := getDictionaryKeyForMetaDataEntry (priv_parser.IPv4SocketPair, priv_parser.Identification, l4prot, priv_parser.TCPSequenceNo.Value, UNKNOWN_SID)
             else
                 key := getDictionaryKeyForMetaDataEntry (priv_parser.IPv6SocketPair, priv_parser.Identification, l4prot, priv_parser.TCPSequenceNo.Value, UNKNOWN_SID);

             // get the metadata entry if available

             if priv_metadataset.ContainsKey (key) then
                begin
                     entry := priv_metadataset.Items[key];

                     priv_frame.Commentlist.Add (entry.RuleDetails.Comment);

                     // Update Statistics

                     priv_multiwritestats.MetaInfoMatchCount := priv_multiwritestats.MetaInfoMatchCount + 1;
                     priv_multiwritestats.MetaInfoLastMatch := entry;

                     if assigned (priv_OnMetaInfoMatch) then
                        priv_OnMetaInfoMatch (self, priv_multiwritestats);
                end;

        end;
end;
}
(* ------------------------------------------- *)
{
procedure tFrameProcessorMultiWriter.AddMetaDataSet (DataSet : tMetaDataSet);

var entry    : tMetaDataEntry;
    oldentry : tMetaDataEntry;
    key      : string;
    i        : integer;

begin
     for i := 1 to length (DataSet) do
         begin
              entry := DataSet[i-1];

              // Generate entry key with a static SID, because we need to be able to find the
              // entry without one

              key := entry.DictionaryKey (UNKNOWN_SID);

              // Add the alert to the alert list

              if not priv_metadataset.ContainsKey (key) then
                 priv_metadataset.Add (key, entry)
              else
                  begin
                       oldentry := priv_metadataset.Items[key];

                       // TODO: Handle collisions, which can happen when a single frame has hits on multiple snort signatures



                  end;
         end;

     priv_multiwritestats.MetaInfoSetSize := priv_metadataset.Count;
end;
}
(* ------------------------------------------- *)

function tFrameProcessorMultiWriter.getOutputFileName (const Frame : tRawFrame) : string;

var filename       : string;
    frametimestamp : string;
    sourcemac      : string;
    destmac        : string;
    sourceip       : string;
    destip         : string;
    sourceport     : string;
    destport       : string;
    vlanids        : string;
    help           : string;
    MACPair        : tMACAddressPair;
    IPv4SockPair   : tIPv4SocketPair;
    IPv6SockPair   : tIPv6SocketPair;
    year           : integer;
    month          : integer;
    day            : integer;
    hour           : integer;
    minute         : integer;
    second         : integer;
    startpos       : integer;
    endpos         : integer;

begin
     assert (priv_currentInputFilename <> UnspecifiedFileMarker, 'Error: input filename unspecified (tFrameProcessorMultiWriter.getOutputFileName)');

     filename := priv_fileschema;

     // Extract date/time values in case we need them for the timestamp placeholder

     year := yearOf (Frame.Timestamp.datetime);
     month := monthOf (Frame.Timestamp.datetime);
     day := dayOf (Frame.Timestamp.datetime);
     hour := hourOf (Frame.Timestamp.datetime);
     minute := minuteOf (Frame.Timestamp.datetime);
     second := dayOf (Frame.Timestamp.datetime);

     vlanids := prot_ParseMaster.VLANIDsAsString;

     if vlanids = '' then
        vlanids := '0';

     frametimestamp := IntToStr (year) + IntToStr (month) + IntToStr (day) + IntToStr (hour) + IntToStr (minute) + IntToStr (second);

     // Determine output file name based on the write mode, because
     // we do not have access to values for all possible placeholders

     filename := stringreplace (filename, '<filename>', priv_currentInputFilename, [rfReplaceAll, rfIgnoreCase]);
     filename := stringreplace (filename, '<count>', IntToStr (priv_outfileindex), [rfReplaceAll, rfIgnoreCase]);
//   filename := stringreplace (filename, '<index>', IntToStr (priv_currentconversationindex), [rfReplaceAll, rfIgnoreCase]);
     filename := stringreplace (filename, '<timestamp>', frametimestamp, [rfReplaceAll, rfIgnoreCase]);
     filename := StringReplace (filename, '<vlanid>', vlanids, [rfReplaceAll, rfIgnoreCase]);

     help := SanitizeFilename (Frame.FilterLabel);
     filename := stringreplace (filename, '<filterlabel>', help, [rfReplaceAll, rfIgnoreCase]);

     if (ppEthernet in prot_ParseMaster.ParsedProtocols) then
        begin
             // Ethernet stuff

             MACPair := prot_ParseMaster.MACAddressPair;
             MACPair.normalize;

             sourcemac := MACPair.Source.asString('');
             filename := stringreplace (filename, '<sourcemac>', sourcemac, [rfReplaceAll, rfIgnoreCase]);

             destmac := MACPair.Destination.asString('');
             filename := stringreplace (filename, '<destinationmac>', destmac, [rfReplaceAll, rfIgnoreCase]);
        end
     else
         begin
              filename := stringreplace (filename, '<sourcemac>', '', [rfReplaceAll, rfIgnoreCase]);
              filename := stringreplace (filename, '<destinationmac>', '', [rfReplaceAll, rfIgnoreCase]);
         end;

     // Determine the highest socket for IP replacements and normalize

     case prot_ParseMaster.LastIPLayerVersion of
      4 : begin
               IPv4SockPair := prot_ParseMaster.IPv4SocketPair;
               IPv4SockPair.normalize;
          end;
      6 : begin
               IPv6SockPair := prot_ParseMaster.IPv6SocketPair;
               IPv6SockPair.normalize;
          end;
      else
          begin
               IPv4SockPair.initialize;
               IPv6SockPair.initialize;
          end;
     end;

     // Replace IP placeholders

     if (ppIPv4 in prot_ParseMaster.ParsedProtocols) or (ppIPv6 in prot_ParseMaster.ParsedProtocols) then
        begin
             case prot_ParseMaster.LastIPLayerVersion of
              4 : begin
                       filename := stringreplace (filename, '<sourceip>', IPv4SockPair.Source.IPAddress.ValueAsString, [rfReplaceAll, rfIgnoreCase]);
                       filename := stringreplace (filename, '<destinationip>', IPv4SockPair.Destination.IPAddress.ValueAsString, [rfReplaceAll, rfIgnoreCase]);
                  end;
              6 : begin
                       // Get IPv6 addresses and replace colons

                       sourceIP := stringReplace (IPv6SockPair.Source.IPAddress.ValueAsString, ':', '-', [rfReplaceAll]);
                       destIP := stringReplace (IPv6SockPair.Destination.IPAddress.ValueAsString, ':', '-', [rfReplaceAll]);

                       filename := stringreplace (filename, '<sourceip>', sourceIP, [rfReplaceAll, rfIgnoreCase]);
                       filename := stringreplace (filename, '<destinationip>', destIP, [rfReplaceAll, rfIgnoreCase]);
                  end;
             end;
        end
     else
         begin
              filename := stringreplace (filename, '<sourceip>', '', [rfReplaceAll, rfIgnoreCase]);
              filename := stringreplace (filename, '<destinationip>', '', [rfReplaceAll, rfIgnoreCase]);
         end;

     // Replace ports

     if (ppTCP in prot_ParseMaster.ParsedProtocols) or (ppUDP in prot_ParseMaster.ParsedProtocols) then
        begin
             case prot_ParseMaster.LastIPLayerVersion of
              4 : begin
                       filename := stringreplace (filename, '<sourceport>', IPv4SockPair.Source.Port.ToString, [rfReplaceAll, rfIgnoreCase]);
                       filename := stringreplace (filename, '<destinationport>', IPv4SockPair.Destination.Port.ToString, [rfReplaceAll, rfIgnoreCase]);
                  end;
              6 : begin
                       filename := stringreplace (filename, '<sourceport>', IPv6SockPair.Source.Port.ToString, [rfReplaceAll, rfIgnoreCase]);
                       filename := stringreplace (filename, '<destinationport>', IPv6SockPair.Destination.Port.ToString, [rfReplaceAll, rfIgnoreCase]);
                  end;
             end;
        end
     else
         begin
              filename := stringreplace (filename, '<sourceport>', '', [rfReplaceAll, rfIgnoreCase]);
              filename := stringreplace (filename, '<destinationport>', '', [rfReplaceAll, rfIgnoreCase]);
         end;

     // remove remaining pointy brackets, which can happen when subdirectory settings were modified

     repeat
           startpos := pos ('<', filename);

           if startpos > 0 then
              begin
                   endpos := pos ('>', filename);

                   if endpos = 0 then
                      filename := leftstring (filename, startpos)
                   else
                       filename := copy (filename, 1, startpos - 1) + copy (filename, endpos + 1, length (filename));
              end;
     until startpos = 0;

     result := filename;
end;

(* ------------------------------------------- *)

function tFrameProcessorMultiWriter.getWriterForFrame (const Filename : string): tPCAPNGWriter;

var writer   : tPCAPNGWriter;
    path     : string;

begin
     if priv_tracewriters.ContainsKey (filename) then
        writer := priv_tracewriters.Items[filename]
     else
         begin
              // Prepare values for a new Writer object

              inc (priv_outfileindex);

              // Track statistics

              priv_multiwritestats.FilesCreated := priv_outfileindex;

              inc (priv_multiwritestats.ActiveWriterCount);

              if priv_multiwritestats.ActiveWriterCount > priv_multiwritestats.MaxWriterCount then
                 priv_multiwritestats.MaxWriterCount := priv_multiwritestats.ActiveWriterCount;

              // Notify of the new writer having been created

              if assigned (priv_OnMultiWriterChange) then
                 priv_OnMultiWriterChange (self, priv_multiwritestats);

              // Check if the output directory exists

              path := extractfilepath (filename);

              if not DirectoryExists (path) then
                 forcedirectories (path);

              // Create the new writer, and store it

              writer := tPCAPngWriter.create (filename, true);
              priv_tracewriters.Add (filename, writer);

              writer.TimeStampMode := priv_timestampmode;
              writer.AbsoluteTimestamp := priv_absolutetimestamp;

              // Open it

              writer.Open;

              // Prepare comment, mimicking Wireshark's mergecap except for the mentioning of TraceWrangler

              writer.HeaderCommentList.Add ('File extracted by TraceWrangler in MultiWriteMode');

              // Write header (yes, all files get all interfaces - reducing that list to the interfaces in use by the conversation is too complicated)

              writer.WriteGenericHeader (priv_interfacelist);
         end;

     result := writer;
end;

(* ------------------------------------------- *)
{
function tFrameProcessorMultiWriter.getWriterForKey (const Key: string; const CurrentWriteMode : tMultiWriteMode): tPCAPNGWriter;

var writer   : tPCAPNGWriter;
    filename : string;
    path     : string;

begin
     if priv_tracewriters.ContainsKey (key) then
        writer := priv_tracewriters.Items[key]
     else
         begin
              // Prepare values for a new Writer object

              inc (priv_outfileindex);

              // Track statistics

              priv_multiwritestats.FilesCreated := priv_outfileindex;

              inc (priv_multiwritestats.ActiveWriterCount);

              if priv_multiwritestats.ActiveWriterCount > priv_multiwritestats.MaxWriterCount then
                 priv_multiwritestats.MaxWriterCount := priv_multiwritestats.ActiveWriterCount;

              if assigned (priv_OnMultiWriterChange) then
                 priv_OnMultiWriterChange (self, priv_multiwritestats);

              // Determine Filename

              filename := getOutputFileName (CurrentWriteMode);

              // Check if the output directory exists

              path := extractfilepath (filename);

              if not DirectoryExists (path) then
                 forcedirectories (path);

              // Create the new writer, and store it

              writer := tPCAPngWriter.create (filename, true);
              priv_tracewriters.Add (key, writer);

              writer.TimeStampMode := priv_timestampmode;
              writer.AbsoluteTimestamp := priv_absolutetimestamp;

              // Open it

              writer.Open;

              // Prepare comment, mimicking Wireshark's mergecap except for the mentioning of TraceWrangler

              writer.HeaderCommentList.Add ('File extracted by TraceWrangler in MultiWriteMode');

              // Write header (yes, all files get all interfaces - reducing that list to the interfaces in use by the conversation is too complicated)

              writer.WriteGenericHeader (priv_interfacelist);
         end;

     result := writer;
end;
}
(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.FinalizeWriter (Writer: tPCAPNGWriter);

begin
     writer.Flush;
     writer.Close;

     writer.free;
     writer := nil;

     // Track how many active writers we have

     dec (priv_multiwritestats.ActiveWriterCount);

     if assigned (priv_OnMultiWriterChange) then
        priv_OnMultiWriterChange (self, priv_multiwritestats);
end;

(* ------------------------------------------- *)
{
procedure tFrameProcessorMultiWriter.writeAutoFrame (const Frame: tRawFrame);

begin
     // Parse the frame to determine which write mode we need

     priv_parser := tParseMaster.create (Frame, ppmUpToLayer4);

     if priv_metadataset.Count > 0 then
        EnhanceFrameWithMetaData;

     if ppTCP in priv_parser.ParsedProtocols then
        writeTCPconversationFrame (Frame)
     else
     if ppUDP in priv_parser.ParsedProtocols then
        writeUDPconversationFrame (Frame)
     else
     if ppIPv4 in priv_parser.ParsedProtocols then
        writeIPv4ConversationFrame (Frame)
     else
     if ppIPv6 in priv_parser.ParsedProtocols then
        writeIPv6ConversationFrame (Frame)
     else
     if ppEthernet in priv_parser.ParsedProtocols then
        writeEtherConversationFrame (Frame);
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.writeEtherconversationFrame (const Frame: tRawFrame);

var writer     : tPCAPNGWriter;
    macconv    : tMACConversation;

begin
     // If the parser is already created, we're on auto mode, otherwise write mode was dictated and we need to parse

     if priv_parser = nil then
        priv_parser := tParseMaster.create (Frame, ppmUpToLayer2);

     macconv := priv_conversationstore.GetMACConversationEntry (priv_parser.MACAddressPair);

     writer := getWriterForKey (macconv.DictionaryKey, mwmEtherConversation);

     // Write the frame

     Writer.addFrameToList (Frame);

     // Check if the Writer can be closed. This is the case when the current frame is known to
     // be the last for this conversation

     if (Frame.TracefileID >= macconv.LastFileID) and (Frame.FileFrameID >= macconv.LastFrame) then
        begin
             FinalizeWriter (writer);

             priv_tracewriters.Remove (macconv.DictionaryKey);
        end;

     priv_parser.free;
     priv_parser := nil;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.writeIPv4conversationFrame (const Frame: tRawFrame);

var writer     : tPCAPNGWriter;
    ipv4conv   : tIPv4Conversation;

begin
     // If the parser is already created, we're on auto mode, otherwise write mode was dictated and we need to parse

     if priv_parser = nil then
        priv_parser := tParseMaster.create (Frame, ppmUpToLayer3);

     ipv4conv := priv_conversationstore.GetIPv4ConversationEntry (priv_parser.IPv4SocketPair.IPAddressPair);

     writer := getWriterForKey (ipv4conv.DictionaryKey, mwmIPv4Conversation);

     // Write the frame

     Writer.addFrameToList (Frame);

     // Check if the Writer can be closed. This is the case when the current frame is known to
     // be the last for this conversation

     if (Frame.TracefileID >= ipv4conv.LastFileID) and (Frame.FileFrameID >= ipv4conv.LastFrame) then
        begin
             FinalizeWriter (writer);

             priv_tracewriters.Remove (ipv4conv.DictionaryKey);
        end;

     priv_parser.free;
     priv_parser := nil;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.writeIPv6conversationFrame (const Frame: tRawFrame);

var writer     : tPCAPNGWriter;
    ipv6conv   : tIPv6Conversation;

begin
     // If the parser is already created, we're on auto mode, otherwise write mode was dictated and we need to parse

     if priv_parser = nil then
        priv_parser := tParseMaster.create (Frame, ppmUpToLayer3);

     ipv6conv := priv_conversationstore.GetIPv6ConversationEntry (priv_parser.IPv6SocketPair.IPAddressPair);

     writer := getWriterForKey (ipv6conv.DictionaryKey, mwmIPv6Conversation);

     // Write the frame

     Writer.addFrameToList (Frame);

     // Check if the Writer can be closed. This is the case when the current frame is known to
     // be the last for this conversation

     if (Frame.TracefileID >= ipv6conv.LastFileID) and (Frame.FileFrameID >= ipv6conv.LastFrame) then
        begin
             FinalizeWriter (writer);

             priv_tracewriters.Remove (ipv6conv.DictionaryKey);
        end;

     priv_parser.free;
     priv_parser := nil;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.writeTCPconversationFrame (const Frame: tRawFrame);

var writer     : tPCAPNGWriter;
    IPv4Socket : tIPv4SocketPair;
    IPv6Socket : tIPv6SocketPair;
    TCPConv    : tTCPConversation;
//  TCPv4conv  : tTCPIPv4Conversation;
//  TCPv6conv  : tTCPIPv6Conversation;

begin
     // If the parser is already created, we're on auto mode, otherwise write mode was dictated and we need to parse

     if priv_parser = nil then
        priv_parser := tParseMaster.create (Frame, ppmUpToLayer4);

     case priv_parser.LastIPLayerVersion of
      4 : begin
               IPv4Socket := priv_parser.IPv4SocketPair;
               IPv4Socket.normalize;

//             TCPv4conv := priv_conversationstore.GetTCPIPv4ConversationEntry (IPv4Socket);
               TCPconv := priv_conversationstore.GetTCPConversationEntry (IPv4Socket);

               writer := getWriterForKey (IPv4Socket.DictionaryKey, mwmTCPConversation);

               // Write the frame

               Writer.addFrameToList (Frame);

               // Check if the Writer can be closed. This is the case when the current frame is known to
               // be the last for this conversation

//             if (Frame.TracefileID >= TCPv4conv.LastFileID) and (Frame.FileFrameID >= TCPv4conv.LastFrame) then
               if (Frame.TracefileID >= TCPconv.LastFileID) and (Frame.FileFrameID >= TCPconv.LastFrame) then
                  begin
                       FinalizeWriter (writer);

//                     priv_tracewriters.Remove (TCPv4conv.DictionaryKey);
                       priv_tracewriters.Remove (TCPconv.DictionaryKey);
                  end;
          end;
      6 : begin
               IPv6Socket := priv_parser.IPv6SocketPair;
               IPv6Socket.normalize;

//             TCPv6conv := priv_conversationstore.GetTCPIPv6ConversationEntry (IPv6Socket);
               TCPconv := priv_conversationstore.GetTCPConversationEntry (IPv6Socket);

               writer := getWriterForKey (IPv6Socket.DictionaryKey, mwmTCPConversation);

               // Write the frame

               Writer.addFrameToList (Frame);

               // Check if the Writer can be closed. This is the case when the current frame is known to
               // be the last for this conversation

//             if (Frame.TracefileID >= TCPv6conv.LastFileID) and (Frame.FileFrameID >= TCPv6conv.LastFrame) then
               if (Frame.TracefileID >= TCPconv.LastFileID) and (Frame.FileFrameID >= TCPconv.LastFrame) then
                  begin
                       FinalizeWriter (writer);

//                     priv_tracewriters.Remove (TCPv6conv.DictionaryKey);
                       priv_tracewriters.Remove (TCPconv.DictionaryKey);
                  end;
          end;
     end;

     priv_parser.free;
     priv_parser := nil;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.writeUDPconversationFrame (const Frame: tRawFrame);

var writer     : tPCAPNGWriter;
    IPv4Socket : tIPv4SocketPair;
    IPv6Socket : tIPv6SocketPair;
    UDPv4conv  : tUDPIPv4Conversation;
    UDPv6conv  : tUDPIPv6Conversation;

begin
     // If the parser is already created, we're on auto mode, otherwise write mode was dictated and we need to parse

     if priv_parser = nil then
        priv_parser := tParseMaster.create (Frame, ppmUpToLayer4);

     case priv_parser.LastIPLayerVersion of
      4 : begin
               IPv4Socket := priv_parser.IPv4SocketPair;
               IPv4Socket.normalize;

               UDPv4conv := priv_conversationstore.GetUDPIPv4ConversationEntry (IPv4Socket);

               writer := getWriterForKey (IPv4Socket.DictionaryKey, mwmUDPConversation);

               // Write the frame

               Writer.addFrameToList (Frame);

               // Check if the Writer can be closed. This is the case when the current frame is known to
               // be the last for this conversation

               if (Frame.TracefileID >= UDPv4conv.LastFileID) and (Frame.FileFrameID >= UDPv4conv.LastFrame) then
                  begin
                       FinalizeWriter (writer);

                       priv_tracewriters.Remove (UDPv4conv.DictionaryKey);
                  end;
          end;
      6 : begin
               IPv6Socket := priv_parser.IPv6SocketPair;
               IPv6Socket.normalize;

               UDPv6conv := priv_conversationstore.GetUDPIPv6ConversationEntry (IPv6Socket);

               writer := getWriterForKey (IPv6Socket.DictionaryKey, mwmUDPConversation);

               // Write the frame

               Writer.addFrameToList (Frame);

               // Check if the Writer can be closed. This is the case when the current frame is known to
               // be the last for this conversation

               if (Frame.TracefileID >= UDPv6conv.LastFileID) and (Frame.FileFrameID >= UDPv6conv.LastFrame) then
                  begin
                       FinalizeWriter (writer);

                       priv_tracewriters.Remove (UDPv6conv.DictionaryKey);
                  end;
          end;
     end;

     priv_parser.free;
     priv_parser := nil;
end;
}
(* ------------------------------------------- *)

function tFrameProcessorMultiWriter.getTruncatedFrame (Frame: tRawFrame): tRawFrame;

var TruncatedFrame : tRawFrame;
    offset         : integer;

begin
     offset := prot_ParseMaster.LayerOffsetList[priv_truncateAfterLayer].NextLayerOffset;

     if (offset > -1) and (offset < frame.FrameSize) then
        begin
             TruncatedFrame := tRawFrame.create (Frame, Frame.WireSize, offset);
             Frame.free;
             Frame := TruncatedFrame;
        end;

     result := frame;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.AddFrameToWriter (const Frame: tRawFrame);

var Filename : string;
    writer   : tPCAPNGWriter;
    boClose  : boolean;
    status   : tConversationStoreRecTypeSet;

begin
     filename := getOutputFileName (Frame);

     writer := getWriterForFrame (Filename);

     // Write the frame

     Writer.addFrameToList (Frame);

     // Check if the Writer can be closed. This is the case when the current frame is known to
     // be the last for this conversation. For that the frame needs to be the last frame of the according
     // conversation type, wich depends on the current write mode (which depends on the file name scheme)

     boClose := false;

     status := priv_conversationstore.LastFrameStatusType (Frame.FilePosition);

     case priv_WriteMode of
      mwmNone              : boClose := false;
      mwmEtherEndpoint     : boClose := (csrMACEndpoint in status);
      mwmIPEndpoint        : boClose := (csrIPv4Endpoint in status) or (csrIPv6Endpoint in status);
      mwmEtherConversation : boClose := (csrMACConversation in status);
      mwmIPConversation    : boClose := (csrIPv4Conversation in status) or (csrIPv6Conversation in status);
      mwmL4Conversation    : boClose := (csrTCPConversation in status) or (csrUDPConversation in status);
     end;

     if boClose then
        begin
             FinalizeWriter (writer);

             priv_tracewriters.Remove (Filename);
        end;
end;

(* ------------------------------------------- *)

procedure tFrameProcessorMultiWriter.addFrame (Frame: tRawFrame);

begin
//   SiMain.LogVerbose ('Processing Frame #' + IntToStr (Frame.FileFrameID) + ' in processor "' + prot_processorname + '"');

     // Do statistics

     DoFrameStatistics (Frame);
     priv_multiwritestats.FramesWritten := priv_multiwritestats.FramesWritten + 1;

     // Create Parsemaster if not handed over

     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create (Frame, ppmUpToLayer4);

     if priv_truncateAfterLayer > 0 then
        Frame := getTruncatedFrame (Frame);

     AddFrameToWriter (Frame);

     // Get rid of the ParseMaster if it exists

//   FreeAndNil (prot_ParseMaster);
     DiscardParseMaster;
//   SiMain.LogDebug ('ParseMaster Object destroyed after adding frame to Writer object (tFrameProcessorMultiWriter.addFrame)');
end;

(* ------------------------------------------- *)


end.
