{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_tracefilesetloader;

interface

uses Contnrs,
     Classes,
     base_structures,
     class_rawframe,
     class_diagmessagehandler,
     class_tracefilelist,
     class_traceloader,
     class_interfacelist,
     class_inteldb_manager,
     class_conversationstorage;

type tOnNextFile = procedure (Sender : tObject; const FileIndex : integer; const FileEntry : tTracefileEntry; var SkipFile : boolean) of object;
     tOnCloseFile = procedure (Sender : tObject; const FileIndex : integer; const FileEntry : tTracefileEntry) of object;
     tOnAddDuplicateFile = procedure (Sender : tObject; const Filename : string; var discardFile : boolean) of object;
     tOnAddUnknownFileType = procedure (Sender : tObject; const Filename : string) of object;

     // Tracefile Entry Forwards

//   tOnAdvancePercentProcessed = procedure (Sender : tObject; const PercentageParsed : integer) of object;    // Primarily for passthrough from tTraceFileEntry

     tOnStartScanOfFile = procedure (Sender : tObject;  const Status : tProcessingStatus) of object;
     tOnEndScanOfFile = procedure (Sender : tObject;  const Status : tProcessingStatus) of object;
     tOnScanMessage = procedure (Sender : tObject;  MessageText : string; out Continue : boolean) of object;
     tOnFullTraceScan = procedure (Sender : tObject;  const Filename : string; out Continue : boolean) of object;
     tOnFileStatusChange = procedure (Sender : tObject; const Status : tProcessingStatus) of object;
     tEventOnFileDiagnosticMessages = procedure (Sender : tObject; FileEntry : tTraceFileEntry; Messagelist : tDiagnosticMsgList) of object;

     tTracefileSetLoader  = class (tObject)

                              constructor create (const DefaultBatchSize : integer = 100000); overload;
                              constructor create (FileList : tTraceFileList; const DefaultBatchSize : integer = 100000); overload;
                              procedure free;

                              private
                                     EventOnNextFile                : tOnNextFile;
                                     EventOnCloseFile               : tOnCloseFile;
                                     EventOnAddDuplicate            : tOnAddDuplicateFile;
                                     EventOnAddUnknownFileType      : tOnAddUnknownFileType;
//                                   EventOnAdvancePercentProcessed : tOnAdvancePercentProcessed;

                                     EventOnScanMessage             : tOnScanMessage;
                                     EventOnFullScan                : tOnFullTraceScan;
                                     EventOnStartScanOfFile         : tOnStartScanOfFile;
                                     EventOnEndScanOfFile           : tOnEndScanOfFile;
                                     EventOnAdvancePercentScanned   : tOnAdvancePercentProcessed;

                                     EventOnFileStatusChange        : tOnFileStatusChange;

                                     EventonDiagnosticsMessages     : tEventOnFileDiagnosticMessages;

                                     priv_namelist                  : tStringlist;
                                     priv_filelist                  : tTracefileList;
                                     priv_framestorage              : tObjectlist;
                                     priv_conversationstore         : tConversationStorage;

                                     priv_currentfileindex          : integer;
                                     priv_stopatfileindex           : integer;
                                     priv_filesToProcess            : integer;
                                     priv_filesProcessed            : integer;
                                     priv_traceloader               : tTraceloader;

                                     priv_interfacelist             : tCaptureInterfaceList;

                                     priv_batchsize                 : integer;
                                     priv_autoscanthreshold         : int64;
                                     priv_currentframe              : integer;

                                     priv_discardduplicates         : boolean;
                                     priv_eof                       : boolean;

                                     priv_ignoredfiles              : integer;

                                     priv_currentfileframecount     : int64;
                                     priv_currentfileframebytes     : int64;
                                     priv_totalframecount           : int64;
                                     priv_totalframebytes           : int64;

                                     priv_totalstatsincomplete      : boolean;

                                     priv_ownsfilelist              : boolean;

                                     priv_IntelDBParams             : tIntelDBParameters;

                                     procedure initialize;

                                     function getFrameCount         : integer;
                                     function getFileCount          : integer;

                                     procedure openTraceFile (FileIndex : integer);
//                                   function FrameIsNotFilteredOut (NewFrame : tRawFrame) : boolean;

                                     function getCurrentFileName : string;
                                     function getCAPFileCount: integer;
                                     function getENCFileCount: integer;
                                     function getPCAPFileCount: integer;
                                     function getPCAPngFileCount: integer;

                                     function getMaximumSectionCount : integer;

                                     function CalculateTotalFrameCount : int64;
                                     function CalculateFrameBytes : int64;

                                     // Event forwarders

//                                   procedure OnFileListEntryAdvancedPercentProcessed (Sender : tObject; const PercentageParsed : integer);
                                     procedure OnFileListEntryStartOfScan (Sender : tObject; const Status : tProcessingStatus);
                                     procedure OnFileListEntryEndOfScan (Sender : tObject; const Status : tProcessingStatus);
                                     procedure OnFileListEntryAdvancePercentageScanned (Sender : tObject; const PercentageScanned, FramesPerSecond : integer; var AbortScan : boolean);
                                     procedure OnFileListEntryScanMessage (Sender : tObject; MessageText : string; out Continue : boolean);
                                     procedure OnFileListEntryStatusChange (Sender : tObject; const Status : tProcessingStatus);

                                     procedure onParserHasDiagnostics (Sender : tObject; Messagelist : tDiagnosticMsgList);

                              protected

                              public
                                    property onNextFile : tOnNextFile read EventOnNextFile write EventOnNextFile;
                                    property onCloseFile : tOnCloseFile read EventOnCloseFile write EventOnCloseFile;
                                    property onAddDuplicateFile : tOnAddDuplicateFile read EventOnAddDuplicate write EventOnAddDuplicate;
                                    property onAddUnknownFileType : tOnAddUnknownFileType read EventOnAddUnknownFileType write EventOnAddUnknownFileType;
//                                  property OnAdvancePercentageProcessed : tOnAdvancePercentProcessed read EventOnAdvancePercentProcessed write EventOnAdvancePercentProcessed;

                                    property OnScanMessage : tOnScanMessage read EventOnScanMessage write EventOnScanMessage;
                                    property OnFullTraceScan : tOnFullTraceScan read EventOnFullScan write EventOnFullScan;
                                    property OnStartScanOfFile : tOnStartScanOfFile read EventOnStartScanOfFile write EventOnStartScanOfFile;
                                    property OnEndScanOfFile : tOnEndScanOfFile read EventOnEndScanOfFile write EventOnEndScanOfFile;
                                    property OnAdvancePercentageScanned : tOnAdvancePercentProcessed read EventOnAdvancePercentScanned write EventOnAdvancePercentScanned;
                                    property OnFileStatusChange : tOnFileStatusChange read EventOnFileStatusChange write EventOnFileStatusChange;
                                    property OnDiagnosticsMessages : tEventOnFileDiagnosticMessages read EventonDiagnosticsMessages write EventonDiagnosticsMessages;

                                    property FileList : tTracefileList read priv_filelist write priv_filelist;
                                    property FileCount : integer read getFileCount;
                                    property ENCFileCount : integer read getENCFileCount;
                                    property CAPFileCount : integer read getCAPFileCount;
                                    property PCAPFileCount : integer read getPCAPFileCount;
                                    property PCAPngFileCount : integer read getPCAPngFileCount;
                                    property MaxPCAPngSectionCount : integer read getMaximumSectionCount;
                                    property IgnoredFilesWhenAdding : integer read priv_ignoredfiles write priv_ignoredfiles;

                                    property TotalStatsIncomplete : boolean read priv_totalstatsincomplete;
                                    property TotalFrameCount : int64 read CalculateTotalFrameCount;
                                    property TotalFrameBytes : int64 read CalculateFrameBytes;

                                    property CurrentFileFrameReadCount : int64 read priv_currentfileframecount;
                                    property CurrentFileFrameByteCount : int64 read priv_currentfileframebytes;
                                    property TotalFrameReadCount : int64 read priv_totalframecount;
                                    property TotalFrameByteCount : int64 read priv_totalframebytes;
                                    property FilesToProcess : integer read priv_filesToProcess;
                                    property FilesProcessed : integer read priv_filesprocessed;

                                    property CurrentFileIndex : integer read priv_currentfileindex;
                                    property CurrentFilename : string read getCurrentFileName;
                                    property BatchSize : integer read priv_batchsize write priv_batchsize;
                                    property AutoScanThreshold : int64 read priv_autoscanthreshold write priv_autoscanthreshold;
                                    property FramesAvailable : integer read getFrameCount;
                                    property DiscardDuplicateFiles : boolean read priv_discardduplicates write priv_discardduplicates;
                                    property EOF : boolean read priv_eof;
                                    property IntelDBParams : tIntelDBParameters read priv_IntelDBParams write priv_IntelDBParams;
                                    property ConversationStore : tConversationStorage read priv_conversationstore;

                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Forces the loader to start from the
                                    ///   beginning for a new readBatch call.
                                    ///   As optional values the index of the
                                    ///   file to start and stop the reading
                                    ///   process at can be specified
                                    /// </summary>
                                    /// <param name="StartFileID">
                                    ///   Index of the file for the next batch
                                    ///   job to start reading from
                                    /// </param>
                                    /// <param name="StopAtFileID">
                                    ///   Index of the file to stop the batch
                                    ///   job at
                                    /// </param>
                                    {$ENDREGION}
                                    procedure reset (const StartAtFileID : integer = 0; const StopAtFileID : integer = 0);
                                    procedure readBatch;
                                    function getFirstFrame : tRawFrame;
                                    function getNextFrame : tRawFrame;
                                    function getFrameByIndex (index : integer) : tRawFrame;
                                    procedure clearFramelist;

                                    procedure SortFilelist (const SortBy : tTracefileSortType = stFirstFrameTime);
                                    procedure clearFilelist;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Aggregates all scan details from the
                                    ///   FileEntries in the list into one
                                    ///   global conversation store
                                    /// </summary>
                                    {$ENDREGION}
                                    procedure UpdateConversationStore;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Adds a new file to the trace file
                                    ///   list by filename. The new
                                    ///   tracefileentry object is returned in
                                    ///   case it is needed for display
                                    ///   purposes right away.
                                    /// </summary>
                                    /// <param name="Filename">
                                    ///   Filename of the new trace file to be
                                    ///   added
                                    /// </param>
                                    /// <returns>
                                    ///   Object created for the new file
                                    /// </returns>
                                    {$ENDREGION}
                                    function AddFile (Filename : string) : tTraceFileEntry;
                                    procedure RemoveFile (Filename : string);
                                    function getFile (Index : integer) : tTraceFileEntry;
                        end;

implementation

uses SysUtils,
     Math,
     toolkit,
     SiAuto,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,
     class_loader_pcapng;


(* ------------------------------------------- *)
{ tFrameCollector }
(* ------------------------------------------- *)

constructor tTracefileSetLoader.create (const DefaultBatchSize : integer = 100000);

begin
     inherited create;

     priv_batchsize := DefaultBatchSize;
     priv_filelist := tTracefilelist.create;
     priv_ownsfilelist := true;

     initialize;
end;

(* ------------------------------------------- *)

constructor tTracefileSetLoader.create (FileList: tTraceFileList;  const DefaultBatchSize: integer);

begin
     inherited create;

     priv_batchsize := DefaultBatchSize;
     priv_filelist := Filelist;
     priv_ownsfilelist := false;

     initialize;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.free;

begin
     // Free the current loader object if still active

     if not (priv_traceloader = nil) then
        begin
             inc (priv_filesprocessed);

             priv_traceloader.closeFile;
             priv_traceloader.free;
             priv_traceloader := nil;
        end;

     priv_namelist.Free;
     priv_interfacelist.free;

     if priv_ownsfilelist then
        priv_filelist.free;

     priv_conversationstore.free;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.initialize;

begin
     priv_framestorage := tObjectlist.Create (false);
     priv_namelist := tStringlist.Create;
     priv_conversationstore := tConversationStorage.create;

     priv_interfacelist := tCaptureInterfaceList.create (true);
     priv_autoscanthreshold := 102400000;                // Do not do file structure scans for files above 100 MB in size

     priv_currentfileindex := 0;
     priv_currentframe := 0;

     priv_discardduplicates := true;
     priv_eof := false;

     priv_ignoredfiles := 0;

     priv_IntelDBParams.initialize;

     priv_filestoprocess := 0;
     priv_filesprocessed := 0;

     priv_currentfileframecount := 0;
     priv_currentfileframebytes := 0;
     priv_totalframecount := 0;
     priv_totalframebytes := 0;

     priv_traceloader := nil;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.OnFileListEntryStartOfScan (Sender: tObject; const Status: tProcessingStatus);

begin
     if assigned (EventOnStartScanOfFile) then
        EventOnStartScanOfFile (Sender, Status);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.OnFileListEntryEndOfScan (Sender: tObject; const Status: tProcessingStatus);

begin
     if assigned (EventOnEndScanOfFile) then
        EventOnEndScanOfFile (Sender, Status);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.OnFileListEntryAdvancePercentageScanned (Sender: tObject; const PercentageScanned, FramesPerSecond : integer; var AbortScan : boolean);

begin
     if assigned (EventOnAdvancePercentScanned) then
        EventOnAdvancePercentScanned (Sender, PercentageScanned, FramesPerSecond, AbortScan);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.OnFileListEntryScanMessage (Sender: tObject; MessageText: string; out Continue: boolean);

begin
     if assigned (EventOnScanMessage) then
        EventOnScanMessage (Sender, MessageText, Continue);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.OnFileListEntryStatusChange (Sender: tObject; const Status: tProcessingStatus);

begin
     if assigned (EventOnFileStatusChange) then
        EventOnFileStatusChange (Sender, Status);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.onParserHasDiagnostics (Sender: tObject; Messagelist: tDiagnosticMsgList);

begin
     if assigned (EventonDiagnosticsMessages) then
        EventonDiagnosticsMessages (self, Sender as tTracefileEntry, Messagelist);
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.reset (const StartAtFileID : integer = 0; const StopAtFileID : integer = 0);

begin
     priv_eof := false;

     priv_filesprocessed := 0;

     priv_currentfileframecount := 0;
     priv_currentfileframebytes := 0;
     priv_totalframecount := 0;
     priv_totalframebytes := 0;

     if StartAtFileID > 0 then
        priv_currentfileindex := StartAtFileID - 1
     else
         priv_currentfileindex := 0;

     if StopAtFileID = 0 then
        priv_stopatfileindex := priv_filelist.Count
     else
         priv_stopatfileindex := min (StopAtFileID, priv_filelist.Count);

     priv_filesToProcess := priv_stopatfileindex - priv_currentfileindex;
end;

(* ------------------------------------------- *)
{
function tTracefileSetLoader.FrameIsNotFilteredOut (NewFrame: tRawFrame): boolean;

begin
     // TODO: Implement load filtering when we can

     result := true;
end;
}
(* ------------------------------------------- *)

procedure tTracefileSetLoader.openTraceFile (FileIndex: integer);

var newFile  : tTracefileEntry;
    filename : string;

begin
     newFile := priv_filelist.getFile (priv_currentfileindex);
     filename := newFile.FullName;

     SiMain.LogVerbose ('Multiloader is opening new file "' + FileName + '"');

     if fileexists (filename) then
        begin
             // Create the correct trace loader object depending on the trace file type
             // and make sure it does not destroy interface entries of its interface list,
             // because we handle them ourselves

             case newFile.TracefileType of
              cfENC    : priv_traceloader := tENCLoader.create (filename, FileIndex, false);
              cfCAP    : priv_traceloader := tCAPLoader.create (filename, FileIndex, false);
              cfPCAP   : priv_traceloader := tPCAPLoader.create (filename, FileIndex, false);
              cfPCAPng : priv_traceloader := tPCAPNGLoader.create (filename, FileIndex, false);
             end;

             // Set the traceloader limit to the current batchsize for better alignment

             priv_traceloader.MaxFramesInMem := priv_batchsize;

             priv_traceloader.openFile;
             priv_traceloader.readTraceFileheader;

             priv_currentfileframecount := 0;
             priv_currentfileframebytes := 0;

             // Todo: aggregate the interfaces found and determine how to handle duplicates
        end
     else
         begin
              // Todo: report problem with opening file
         end;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.readBatch;

var newFrame  : tRawFrame;
    i         : integer;
    maxframes : longword;
    SkipFile  : boolean;
    oldFile   : tTracefileEntry;
    newFile   : tTracefileEntry;

begin
     // Remove all old frames

     priv_framestorage.Clear;

     if priv_currentfileindex <= priv_filelist.count then
        begin
             // Load batchsize of frames - from multiple files if necessary

             repeat
                   // If no loader is active, open one for the next file. This is usually file index 1,
                   // but may be different if the reset function was used.

                   if (priv_traceloader = nil) then
                      begin
                           if priv_currentfileindex < priv_stopatfileindex then
                              begin
                                   repeat
                                         inc (priv_currentfileindex);

                                         newFile := priv_filelist.getFile (priv_currentfileindex);
                                         SkipFile := false;

                                         // Trigger nextFile event if assigned and allow skipping files if necessary

                                         if assigned (EventOnNextFile) then
                                            EventOnNextFile (self, priv_currentfileindex, newFile, SkipFile);

                                   until (not SkipFile) or (priv_currentfileindex >= priv_filelist.count);

                                   if (priv_currentfileindex > priv_filelist.count) or ((priv_currentfileindex = priv_filelist.count) and SkipFile) then
                                      priv_eof := true
                                   else
                                       openTraceFile (priv_currentfileindex);
                              end
                           else
                               priv_eof := true;
                      end
                   else
                       if priv_traceloader.EOF then
                          begin
                               // Close current file

                               if assigned (EventOnCloseFile) then
                                  begin
                                       OldFile := priv_filelist.getFile (priv_currentfileindex);

                                       EventOnCloseFile (self, priv_currentfileindex, OldFile);
                                  end;

                               SiMain.LogVerbose ('Multiloader is closing current file "' + OldFile.FullName + '"');

                               priv_traceloader.closeFile;
                               priv_traceloader.free;
                               priv_traceloader := nil;

                               inc (priv_filesprocessed);

                               if priv_currentfileindex < priv_stopatfileindex then
                                  begin
                                       repeat
                                             inc (priv_currentfileindex);

                                             newFile := priv_filelist.getFile (priv_currentfileindex);
                                             SkipFile := false;

                                             // Trigger nextFile event if assigned and allow skipping files if necessary

                                             if assigned (EventOnNextFile) then
                                                EventOnNextFile (self, priv_currentfileindex, newFile, SkipFile);

                                       until (not SkipFile) or (priv_currentfileindex >= priv_filelist.count);

                                       // Open the next file if there is one left

                                       if (priv_currentfileindex > priv_stopatfileindex) or ((priv_currentfileindex > priv_stopatfileindex) and SkipFile) then
                                          priv_eof := true
                                       else
                                           begin
                                                // Open the file and read the header

                                                openTraceFile (priv_currentfileindex);
                                           end;
                                  end
                               else
                                   priv_eof := true;
                          end;

                   if not priv_eof then
                      begin
                           // Read the next bunch - only as much as we need to have a full Batchsize

                           maxframes := priv_batchsize - priv_framestorage.Count;

                           priv_traceloader.MaxFramesInMem := maxframes;
                           priv_traceloader.readBatchFromDisk;

                           // Store if valid

                           for i := 1 to priv_traceloader.FramesAvailable do
                               begin
                                    newFrame := priv_traceloader.getRawFrame(i);

                                    priv_framestorage.Add (newFrame);

                                    // Statistics

                                    priv_currentfileframecount := priv_currentfileframecount + 1;
                                    priv_currentfileframebytes := priv_currentfileframecount + newFrame.FrameSize;
                                    priv_totalframecount := priv_totalframecount + 1;
                                    priv_totalframebytes := priv_totalframebytes + newFrame.FrameSize;
                               end;
                      end;

             until priv_eof or (priv_framestorage.Count = priv_batchsize);

             // Cleanup when all is done

             if priv_eof and not (priv_traceloader = nil) then
                begin
                     if assigned (EventOnCloseFile) then
                        begin
                             OldFile := priv_filelist.getFile (priv_currentfileindex);

                             EventOnCloseFile (self, priv_currentfileindex, OldFile);
                        end;

                     inc (priv_filesprocessed);

                     priv_traceloader.closeFile;
                     priv_traceloader.free;
                     priv_traceloader := nil;
                end;
        end
     else
         priv_eof := true;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.UpdateConversationStore;

var entry : tTracefileEntry;
    i     : integer;

begin
     SiMain.EnterMethod ('tTracefileSetLoader.UpdateConversationStore');

     priv_conversationstore.Clear;
     priv_totalstatsincomplete := false;

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist.getFile(i);

              // If the file has a conversation store, merge it to the global store and force
              // the current file id for the entries. This makes sure that no matter the order
              // of the file entries, the global store will have the correct values

              SiMain.LogDebug ('adding conversations for file "' + entry.FullName + '" to the conversation store');

              if entry.ConversationsScanned then
                 priv_conversationstore.AddConversationsFromStore (entry.ConversationStore, i);

              if not entry.PacketStatisticsRead then
                 priv_totalstatsincomplete := true;
         end;

     // Final calculations, e.g. RTT stuff

     priv_conversationstore.UpdateCalculatedValues;

     SiMain.LeaveMethod ('tTracefileSetLoader.UpdateConversationStore');
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.SortFilelist (const SortBy: tTracefileSortType = stFirstFrameTime);

begin
     priv_filelist.sort (SortBy);
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getFrameCount: integer;

begin
     result := priv_framestorage.Count;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getFirstFrame: tRawFrame;

begin
     priv_currentframe := 0;

     result := priv_framestorage[priv_currentframe] as tRawFrame;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getNextFrame: tRawFrame;

begin
     inc (priv_currentframe);

     if priv_currentframe <= priv_framestorage.Count then
        result := priv_framestorage[priv_currentframe] as tRawFrame
     else
         result := nil;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getFrameByIndex(index: integer): tRawFrame;

begin
     if index <= priv_framestorage.Count then
        result := priv_framestorage[index - 1] as tRawFrame
     else
         raise EInOutError.Create ('Error: frame index out of bounds (tFrameCollector.getFrameByIndex).');
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getFile(Index: integer): tTraceFileEntry;

begin
     result := priv_filelist.getFile(index);
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getFileCount: integer;

begin
     result := priv_filelist.Count;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getENCFileCount: integer;

begin
     result := priv_filelist.ENCFileCount;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getCAPFileCount: integer;

begin
     result := priv_filelist.CAPFileCount;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getPCAPFileCount: integer;

begin
     result := priv_filelist.PCAPFileCount;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getMaximumSectionCount: integer;

var i           : integer;
    FileEntry   : tTracefileEntry;
    maxsections : integer;

begin
     maxsections := 0;

     for i := 1 to priv_filelist.Count do
         begin
              FileEntry := priv_filelist.getFile(i);

              if FileEntry.TracefileType = cfPCAPng then
                 maxsections := max (maxsections, FileEntry.PCAPngStatistics.SectionCount);
         end;

     result := maxsections;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getPCAPngFileCount: integer;

begin
     result := priv_filelist.PCAPngFileCount;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.getCurrentFileName: string;

var FileEntry : tTraceFileEntry;
    filename  : string;

begin
     filename := '';

     if priv_currentfileindex > 0 then
        begin
             // In case we're in the last file already, priv_currentfile is one too high caused by the readBatch inc statement
             // so for this we assume we're in the last file

             if priv_currentfileindex = (priv_filelist.Count + 1) then
                FileEntry := priv_filelist.getFile(priv_currentfileindex - 1)
             else
                 FileEntry := priv_filelist.getFile(priv_currentfileindex);

             filename := FileEntry.FullName;
        end;

     result := filename;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.AddFile (Filename: string) : tTraceFileEntry;

var NewTracefile : tTracefileEntry;
    boAddOkay    : boolean;
    boDiscard    : boolean;

begin
     SiMain.LogVerbose('Adding file "' + Filename + '", current AutoScanThreshold is ' + PointText (priv_autoscanthreshold) + ' bytes');

     // Determine if the file is a duplicate

     if priv_namelist.IndexOf (Filename) > -1 then
        begin
             // it is a duplicate. See if there is an event that could tell us to still add the file

             if assigned (EventOnAddDuplicate) then
                begin
                     // Preset to whatever the current default is

                     boDiscard := priv_discardduplicates;

                     // Trigger the event and do what comes out of it

                     EventOnAddDuplicate (self, Filename, boDiscard);
                     boAddOkay := not boDiscard;
                end
             else
                 boAddOkay := not priv_discardduplicates;
        end
     else
         boAddOkay := true;

     if boAddOkay then
        begin
             NewTracefile := tTracefileEntry.create (Filename, priv_IntelDBParams, priv_autoscanthreshold);

             // Daisy chain events for GUI accessability

             NewTraceFile.OnStartScanOfFile := OnFileListEntryStartOfScan;
             NewTraceFile.OnEndScanOfFile := OnFileListEntryEndOfScan;
             NewTraceFile.OnAdvancePercentageScanned := OnFileListEntryAdvancePercentageScanned;
             NewTraceFile.OnScanMessage := OnFileListEntryScanMessage;
             NewTraceFile.OnFileStatusChange := OnFileListEntryStatusChange;
             NewTraceFile.OnDiagnosticsMessager := onParserHasDiagnostics;

             // Now we need to get some details about the file and store it with the file name

             if NewTracefile.TracefileType <> cfUnknown then
                priv_filelist.addFile (NewTraceFile)
             else
                 begin
                      // We don't know what format it is, so we ignore the file

                      if assigned (EventOnAddUnknownFileType) then
                         EventOnAddUnknownFileType (self, Filename);

                      inc (priv_ignoredfiles);
                 end;

             // Add the filename to the name list

             priv_namelist.Add (Filename);
        end
     else
         NewTraceFile := nil;

     result := NewTraceFile;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.RemoveFile (Filename : string);

var boFound : boolean;
    i       : integer;
    entry   : tTracefileEntry;

begin
     // Search for the entry with the given name. We do it the stupid way, but speed is not an issue here anyway

     boFound := false;
     i := 0;

     repeat
           inc (i);

           entry := priv_filelist.getFile(i);

           if uppercase (entry.FullName) = uppercase (Filename) then
              boFound := true;

     until boFound or (i = priv_filelist.Count);

     if boFound then
        begin
             priv_filelist.removeFile(i);

             // remove the name list entry, too

             i := priv_namelist.IndexOf(Filename);

             if i > -1 then
                priv_namelist.Delete(i);
        end;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.CalculateFrameBytes: int64;

var value : int64;
    entry : tTracefileEntry;
    i     : integer;

begin
     value := 0;
     priv_totalstatsincomplete := false;

     for i := 1 to  priv_filelist.Count do
         begin
              entry := priv_filelist.getFile (i);

              value := value + entry.GeneralStatistics.TotalFrameBytes;

              if not entry.PacketStatisticsRead then
                 priv_totalstatsincomplete := true;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tTracefileSetLoader.CalculateTotalFrameCount: int64;

var value : int64;
    entry : tTracefileEntry;
    i     : integer;

begin
     value := 0;
     priv_totalstatsincomplete := false;

     for i := 1 to  priv_filelist.Count do
         begin
              entry := priv_filelist.getFile (i);

              value := value + entry.GeneralStatistics.FrameCount;

              if not entry.PacketStatisticsRead then
                 priv_totalstatsincomplete := true;
         end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.clearFilelist;

begin
     priv_filelist.clear;
end;

(* ------------------------------------------- *)

procedure tTracefileSetLoader.clearFramelist;

begin
     priv_framestorage.Clear;
end;

(* ------------------------------------------- *)

end.
