{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

program ngconvert;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  System.SysUtils,
  Vcl.Graphics,
  System.Diagnostics,
  SiAuto,
  class_interfacelist in 'loader\class_interfacelist.pas',
  class_loader_enc in 'loader\class_loader_enc.pas',
  class_loader_naicap in 'loader\class_loader_naicap.pas',
  class_loader_pcap in 'loader\class_loader_pcap.pas',
  class_loader_pcapng in 'loader\class_loader_pcapng.pas',
  class_tracewriter in 'loader\class_tracewriter.pas',
  class_writer_pcapng in 'loader\class_writer_pcapng.pas',
  base_structures in 'base\base_structures.pas',
  toolkit in 'base\toolkit.pas',
  pcapng_structures in 'loader\pcapng_structures.pas',
  definition_ip_addresses in 'base\definition_ip_addresses.pas',
  definition_simple_nameresolution in 'base\definition_simple_nameresolution.pas',
  class_interfacestatistics in 'loader\class_interfacestatistics.pas',
  class_rawframe in 'base\class_rawframe.pas',
  class_traceloader in 'loader\class_traceloader.pas',
  class_dupechecker in 'loader\class_dupechecker.pas',
  class_timezonedetails in 'base\class_timezonedetails.pas',
  class_dnsserverlist in 'storage\class_dnsserverlist.pas',
  base_messages in 'base\base_messages.pas',
  tracefile_utils in 'loader\tracefile_utils.pas',
  class_dirparser in 'base\class_dirparser.pas',
  base_checksums in 'base\base_checksums.pas';

var cPercent  : integer;

(* ------------------------------------------- *)

procedure UpdatePercentage (const Position, Total : int64);

var dHelp : double;
    iHelp : integer;

begin
     if Position < Total then
        begin
             dHelp := 100 * Position / Total;
             iHelp := trunc (dHelp);

             if iHelp > cPercent then
                begin
                     write (#8#8#8#8 + Format ('%3d', [iHelp]) + '%');
                     cPercent := iHelp;
                end;
        end
     else
         write (#8#8#8#8'100%');
end;

(* ------------------------------------------- *)

function getOutfileName (Filename : string) : string;

var fileext : string;
    value   : string;

begin
     fileext := extractFileExt (Filename);

     if fileext <> '.pcapng' then
        value := changeFileExt (Filename, '.pcapng')
     else
         begin
              value := changeFileExt (Filename, '');
              value := value + '_converted.pcapng';
         end;

     result := value;
end;

(* ------------------------------------------- *)

procedure ConvertFile (const Input, Output : string);

var loader     : tTraceLoader;
    writer     : tPCAPngWriter;
    outfile    : string;
    i          : integer;
    frame      : tRawFrame;
    comment    : string;
    framecount : integer;
    stopwatch  : tStopWatch;

begin
     if (Output = '') or fileexists (Output) then
        outfile := getOutfileName (Input)
     else
         outfile := Output;

     loader := createTraceLoader (Input, 0, cfUnknown);

     if loader = nil then
        begin
             writeln ('Error: could not open file "' + Input + '": unknown file type');
             exit;
        end;

     if loader.FileFormat = cfPCAPng then
        begin
             writeln ('Skipping file "' + Input + '": already a PCAPng file');
             loader.free;
             exit;
        end;

     write ('Converting "' + input + '":   0%');

     loader.MaxFramesInMem := 100000;
     loader.openFile;
     loader.readTraceFileheader;

     framecount := 0;

     writer := tPCAPngWriter.create (outfile, true);
     writer.MaxFramesInMem := 100000;
     writer.Open;

     // Create new file header blocks

     comment := 'Converted by ngconvert from ' + CaptureFileTypeAsText (Loader.FileFormat);

     case loader.FileFormat of
      cfENC    : begin
                      writer.HeaderCommentList.Add (comment);
                      writer.WriteGenericHeader ((Loader as tENCLoader).LinkType);
                 end;
      cfCAP    : begin
                      writer.HeaderCommentList.Add (comment);
                      writer.WriteGenericHeader ((loader as tCAPLoader).CaptureInterfaces);
                 end;
      cfPCAP   : begin
                      writer.HeaderCommentList.Add (comment);
                      writer.WriteGenericHeader ((Loader as tPCAPLoader).LinkType);
                 end;
     end;

     stopwatch := tStopWatch.Create;
     stopwatch.Start;

     repeat
           loader.readBatchFromDisk;

           for i := 1 to loader.FramesAvailable do
               begin
                    frame := loader.getRawFrame(i);

                    // Add the frame to the writer. Since it owns the frame it will discard the
                    // frame object after flush operations

                    writer.addFrameToList (frame);

                    UpdatePercentage (loader.Position, loader.Filesize);
               end;

           inc (framecount, loader.FramesAvailable);

     until loader.EOF;

     writer.Flush;
     writer.Close;
     writer.free;

//   writeln (' - completed in ' + IntToStr (stopwatch.ElapsedMilliseconds) + ' milliseconds');

     loader.closeFile;
     loader.free;
end;

(* ------------------------------------------- *)

procedure ConvertBatch (const FileMask, Outfile : string; const IncludeSubDirs : boolean);

var dirparser : tDirParser;
    i         : integer;
    filename  : string;

begin
     dirParser := tDirParser.create;

     dirParser.BaseDirectory := extractfilepath (FileMask);
     dirParser.FileMask := extractfilename (FileMask);
     dirParser.IncludeSubDirs := IncludeSubDirs;

     dirparser.execute;

     for i := 1 to dirParser.FileCount do
         begin
              filename := dirParser.getFileName (i);

              if filename <> 'traceintel.db' then
                 ConvertFile (filename, '');
         end;
end;

(* ------------------------------------------- *)

procedure runMain;

var SourceFile : string;
    boShowHelp : boolean;
    boInclSub  : boolean;
    outfile    : string;

begin
     boShowHelp := false;
     boInclSub := FindCmdLineSwitch ('r');

     if ParamCount > 0 then
        begin
             if (ParamStr(2) <> '') and (pos ('/', ParamStr(2)) = 0) then
                outfile := ParamStr(2)
             else
                 outfile := '';
        
             if fileexists (ParamStr(1)) then
                ConvertFile (ParamStr(1), outfile)
             else
                 if pos ('*', ParamStr(1)) > 0 then
                    ConvertBatch (ParamStr(1), outfile, boInclSub)
                 else
                     writeln ('File not found.');
        end
     else
         boShowHelp := true;

     if boShowHelp then
        begin
             writeln ('This tool converts packet capture files (trace files) to the PCAPng format');
             writeln ('Usage: ngconvert <inputfile> [outputfile] [options]');
             writeln;
             writeln ('/r     include subdirectories (when using wildcards)');
             writeln;
             writeln ('Hints: - if outputfile is not specified the name will be autogenerated.');
             writeln ('       - you can use the asterisk as wildcard for the input file name.');
        end;
end;

(* ------------------------------------------- *)

begin
     try
        runMain;
     except
           on E: Exception do Writeln(E.ClassName, ': ', E.Message);
     end;
end.
