{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_arp;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;

type pARPStruct = ^tARPStruct;
     tARPStruct = packed record
                               HardwareType : word;
                               ProtocolType : word;
                               HardwareSize : byte;
                               ProtocolSize : byte;
                               Opcode       : word;
                               SenderMAC    : tMACBytes;
                               SenderIP     : longword;
                               TargetMAC    : tMACBytes;
                               TargetIP     : longword;
                  end;

     tARPParser  = class (tFrameParser)

                              private
                                     priv_HardwareType : word;
                                     priv_ProtocolType : word;
                                     priv_HardwareSize : byte;
                                     priv_ProtocolSize : byte;
                                     priv_OpCode       : word;
                                     priv_SenderMAC    : tMACAddress;
                                     priv_SenderIP     : tIPv4Address;
                                     priv_TargetMAC    : tMACAddress;
                                     priv_TargetIP     : tIPv4Address;
                              public
                                    property HardwareType : word read priv_HardwareType;
                                    property ProtocolType : word read priv_ProtocolType;
                                    property HardwareSize : byte read priv_HardwareSize;
                                    property ProtocolSize : byte read priv_ProtocolSize;
                                    property OpCode : word read priv_OpCode;
                                    property SenderMAC : tMACAddress read priv_SenderMAC;
                                    property SenderIP : tIPv4Address read priv_SenderIP;
                                    property TargetMAC : tMACAddress read priv_TargetMAC;
                                    property TargetIP : tIPv4Address read priv_TargetIP;

                                    function checkIsPlausible : boolean; override;
                                    function parse : tParseResult; override;
                        end;


implementation

uses toolkit;


(* ------------------------------------------- *)
{ tARPParser }
(* ------------------------------------------- *)

function tARPParser.checkIsPlausible: boolean;

begin
     result := true;                  // good enough for now
end;

(* ------------------------------------------- *)

function tARPParser.parse: tParseResult;

var pData    : pARPStruct;
    status   : tParseResult;
    HelpWord : word;
    HelpLong : longword;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := sizeOf (tARPStruct);

     status := prSuccess;

     if priv_framelength < (priv_frameoffset + sizeOf (tARPStruct)) then
        status := prFailUndersize
     else
         try
            pData := pARPStruct (priv_framedata + priv_frameoffset);

            HelpWord := pData^.HardwareType;
            priv_HardwareType := Swap16 (HelpWord);

            HelpWord := pData^.ProtocolType;
            priv_ProtocolType := Swap16 (HelpWord);

            priv_HardwareSize := pData^.HardwareSize;
            priv_ProtocolSize := pData^.ProtocolSize;

            HelpWord := pData^.Opcode;
            priv_OpCode := Swap16 (HelpWord);

            priv_SenderMAC.RangeBytes := pData^.SenderMAC;

            HelpLong := pData^.SenderIP;
            priv_SenderIP.ValueAsLong := Swap32 (HelpLong);
            priv_SenderIP.MaskBitCount := 32;

            priv_TargetMAC.RangeBytes := pData^.TargetMAC;

            HelpLong := pData^.TargetIP;
            priv_TargetIP.ValueAsLong := Swap32 (HelpLong);
            priv_TargetIP.MaskBitCount := 32;

            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tARPStruct);

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
