{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ethernet;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass,
     class_parser_llc;


type tEtherFrameType = (etEthernetII, et802dot3);

     tEthernetParser  = class (tFrameParser)

                              private
                                     // 802.3

                                     priv_ethsrc          : tMACAddress;
                                     priv_ethdst          : tMACAddress;
                                     priv_ethertypelength : word;

                                     function getEtherFrameType: tEtherFrameType;
                                     function getMACAddressPair : tMACAddressPair;

                              public
                                    property EtherFrameType : tEtherFrameType read getEtherFrameType;
                                    property SourceAddress : tMACAddress read priv_ethsrc;
                                    property DestinationAddress : tMACAddress read priv_ethdst;
                                    property EtherTypeLength : word read priv_ethertypelength;
                                    property EthernetAddressPair : tMACAddressPair read getMACAddressPair;

                                    function checkIsPlausible : boolean; override;
                                    function parse : tParseResult; override;
                        end;

     pStruct_Ethernet = ^tStruct_Ethernet;
     tStruct_Ethernet = packed record
                               dst           : tMACBytes;
                               src           : tMACBytes;
                               ethtypelength : word;
                        end;


const // Declaration of values (MAC level)

      VALUE_ETH_TYPE_IP            = $0800;
      VALUE_ETH_TYPE_ARP           = $0806;
      VALUE_ETH_WOL                = $0842;
      VALUE_ETH_DECNET_PHASE_IV    = $6003;
      VALUE_ETH_TYPE_RARP          = $8035;
      VALUE_ETH_APPLETALK          = $809b;
      VALUE_ETH_APPLETALK_ARP      = $80f3;
      VALUE_ETH_8021_Q             = $8100;  // 802.1q
      VALUE_ETH_NOVELL_IPX_OLD     = $8137;
      VALUE_ETH_NOVEL              = $8137;
      VALUE_ETH_TYPE_IPV6          = $86dd;
      VALUE_ETH_TYPE_MAC_CONTROL   = $8808;
      VALUE_ETH_TYPE_SLOW          = $8809;
      VALUE_ETH_MPLS_UNICAST       = $8847;
      VALUE_ETH_MPLS_MULTICAST     = $8848;
      VALUE_ETH_PPOE_DISCOVERY     = $8863;
      VALUE_ETH_PPOE_SESSION       = $8864;
      VALUE_ETH_MICROSOFT_NLB_HB   = $886f;
      VALUE_ETH_JUMBO_FRAMES       = $8870;
      VALUE_ETH_HOMEPLUG           = $887b;
      VALUE_ETH_EAP_OVER_LAN       = $888e;  // 802.1x
      VALUE_ETH_802_QINQ_Outer     = $88a8;
      VALUE_ETH_LLDP               = $88cc;
      VALUE_ETH_FCOE               = $8906;
      VALUE_ETH_FCOE_INIT          = $8914;
      VALUE_ETH_802_QINQ_Outer_OLD = $9100;
      VALUE_ETH_SNAP               = $aaaa;

      // SNAP Vendor Codes

      SNAP_VENDOR_ENCAPSULATEDETHERNET = 0;

implementation

uses Toolkit;

(* ------------------------------------------- *)
{ tEthernetParser }
(* ------------------------------------------- *)

function tEthernetParser.checkIsPlausible: boolean;

begin
     result := true;                       // obviously
end;

(* ------------------------------------------- *)

function tEthernetParser.parse: tParseResult;

var pData  : pStruct_Ethernet;
    status : tParseResult;
    help   : word;

begin
     priv_osilayer := 2;     // Set Layer this parser is for
     prot_layersize := 14;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tStruct_Ethernet) then
        status := prFailUndersize
     else
         try
            pData := pStruct_Ethernet (priv_framedata + priv_frameoffset);

            priv_ethsrc.RangeBytes := pData^.src;
            priv_ethdst.RangeBytes := pData^.dst;

            help := pData^.ethtypelength;

            priv_ethertypelength := swap16 (help);

            // Set the next header type - usually the same as the ether type

            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tStruct_Ethernet);

            if priv_ethertypelength > $0600 then
               priv_nextAnchor.HeaderType := priv_ethertypelength
            else
                priv_nextAnchor.HeaderType := NEXT_HEADER_LLC;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function tEthernetParser.getEtherFrameType: tEtherFrameType;

begin
     if priv_ethertypelength >= $0600 then
        result := etEthernetII
     else
         result := et802dot3;
end;

(* ------------------------------------------- *)

function tEthernetParser.getMACAddressPair: tMACAddressPair;

begin
     result.VLANID := 0;
     result.Source := priv_ethsrc;
     result.Destination := priv_ethdst;
end;

(* ------------------------------------------- *)

end.
