{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_gre;


interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;


// RFC 2890 / 1701 / 1702

type pGREStruct = ^tGREStruct;
     tGREStruct = packed record
                               FlagsVersion : word;
                               ProtocolType : word;
                  end;

     tGREParser  = class (tFrameParser)

                         private
                                priv_ChecksumPresent   : boolean;
                                priv_RoutingPresent    : boolean;
                                priv_KeyPresent        : boolean;
                                priv_SequencePresent   : boolean;
                                priv_strictSourceRoute : boolean;
                                priv_recursioncontrol  : byte;
                                priv_flags             : byte;
                                priv_versionnumber     : byte;
                                priv_checksum          : word;
                                priv_key               : longword;
                                priv_sequence          : longword;
                                priv_protocoltype      : word;

                         public
                               property CheckSumPresent : boolean read priv_ChecksumPresent;
                               property RoutingPresent : boolean read priv_RoutingPresent;
                               property KeyPresent : boolean read priv_KeyPresent;
                               property SequencePresent : boolean read priv_SequencePresent;
                               property StrictSourceRoute : boolean read priv_strictSourceRoute;
                               property RecursionControl : byte read priv_recursioncontrol;
                               property Flags : byte read priv_flags;
                               property VersionNumber : byte read priv_versionnumber;
                               property Checksum : word read priv_checksum;
                               property Key : longword read priv_key;
                               property Sequence : longword read priv_sequence;
                               property ProtocolType : word read priv_protocoltype;

                               function checkIsPlausible : boolean; override;
                               function parse : tParseResult; override;
                     end;


implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tGREParser }
(* ------------------------------------------- *)

function tGREParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tGREParser.parse: tParseResult;

var pData     : pGREStruct;
    modifier  : integer;
    status    : tParseResult;
    pHelp     : pWord;
    pLongHelp : pLongword;
    LongHelp  : Longword;
    help      : Word;
    help2     : word;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := sizeOf (tGREStruct);

     priv_checksum := 0;
     priv_key := 0;
     priv_sequence := 0;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tGREStruct) then
        status := prFailUndersize
     else
         try
            pData := pGREStruct (priv_framedata + priv_frameoffset);

            // Flags and Version

            pHelp := pWord (pData);
            help := pHelp^;
            help := Swap16 (help);

            priv_ChecksumPresent := (help and $8000) = $8000;
            priv_RoutingPresent := (help and $4000) = $4000;
            priv_KeyPresent := (help and $2000) = $2000;
            priv_SequencePresent := (help and $1000) = $1000;
            priv_strictSourceRoute := (help and $800) = $800;

            help2 := (help and $700) shr 8;
            priv_recursioncontrol := help2;

            help2 := (help and $f8) shr 3;
            priv_flags := help2;

            priv_versionnumber := help and $7;

            // Prepare next anchor

            modifier := 4;

            if priv_ChecksumPresent then
               begin
                    pLongHelp := pLongWord (priv_framedata + priv_frameoffset + modifier);
                    LongHelp := Swap32 (pLongHelp^);

                    priv_checksum := LongHelp;

                    inc (modifier, 4);
                    inc (prot_layersize, 4);
               end;


            if priv_KeyPresent then
               begin
                    pLongHelp := pLongWord (priv_framedata + priv_frameoffset + modifier);
                    priv_key := Swap32 (pLongHelp^);

                    inc (modifier, 4);
                    inc (prot_layersize, 4);
               end;

            if priv_SequencePResent then
               begin
                    pLongHelp := pLongWord (priv_framedata + priv_frameoffset + modifier);
                    priv_sequence := Swap32 (pLongHelp^);

                    inc (modifier, 4);
                    inc (prot_layersize, 4);
               end;

            priv_protocoltype := swap16 (pData^.ProtocolType);

            if not priv_RoutingPresent then
               begin
                    help := pData^.ProtocolType;

                    priv_nextAnchor.HeaderType := Swap16 (help);
                    priv_nextAnchor.Offset := priv_frameoffset + modifier;
               end
            else
                begin
                     // We can't parse routing information (yet) so we give up

                     priv_nextAnchor.HeaderType := NEXT_HEADER_UNKNOWN;
                     priv_nextAnchor.Offset := priv_frameoffset + modifier;
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)


end.
