{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ipv6_fragmentationheader;

interface

uses base_structures,
//   class_definition_ip,
     definition_ip_addresses,
     base_checksums,
     common_parse_helpers,
     class_parser_baseclass;

type pIPv6FragmentationStruct = ^tIPv6FragmentationStruct;
     tIPv6FragmentationStruct = packed record
                                             NextHeader           : byte;
                                             Reserved             : byte;
                                             OffsetAndFlags       : word;
                                             Identification       : longword;
                                       end;

     tIPv6FragmentationParser  = class (tFrameParser)

                                       private
                                              priv_flags          : tIPFlagSet;
                                              priv_fragmentoffset : word;
                                              priv_identification : longword;

                                       public
                                             property Offset : word read priv_fragmentoffset;
                                             property Flags : tIPFlagSet read priv_flags;
                                             property Identification : longword read priv_identification;

                                             function checkIsPlausible : boolean; override;
                                             function parse : tParseResult; override;
                                 end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ IPv6Parser }
(* ------------------------------------------- *)

function tIPv6FragmentationParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tIPv6FragmentationParser.parse: tParseResult;

var pData  : pIPv6FragmentationStruct;
    status : tParseResult;
    help16 : word;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := 8;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tIPv6FragmentationStruct) then
        status := prFailUndersize
     else
         try
            pData := pIPv6FragmentationStruct (priv_framedata + priv_frameoffset);

            help16 := pData^.OffsetAndFlags;
            help16 := Swap16 (help16);

            if (help16 and $1) = $1 then
               priv_flags := [ipfMoreFragments]
            else
                priv_flags := [ipfLastFragment];

            priv_fragmentoffset := (help16 and $FFF8) shr 3;

            priv_identification := pData^.Identification;
            priv_identification := Swap32 (priv_identification);

            // Anchor

            priv_nextAnchor.HeaderType := pData^.NextHeader;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tIPv6FragmentationStruct);

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
