{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ntp;

{
   NTP Header:

 0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |LI | VN  |Mode |    Stratum     |     Poll      |  Precision   |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Root Delay                            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                         Root Dispersion                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                          Reference ID                         |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                     Reference Timestamp (64)                  +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                      Origin Timestamp (64)                    +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                      Receive Timestamp (64)                   +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      +                      Transmit Timestamp (64)                  +
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      .                                                               .
      .                    Extension Field 1 (variable)               .
      .                                                               .
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      .                                                               .
      .                    Extension Field 2 (variable)               .
      .                                                               .
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                          Key Identifier                       |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                                                               |
      |                            dgst (128)                         |
      |                                                               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
}

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;

type tNTPMode = (ntmReserved, ntmSymActive, ntSymPassive, ntmClient, ntmServer, ntmBroadcast, ntmNTPControlMsg, ntmReservedPrivate);
     tNTPLeapIndicator = (nliNoWarning, nli61Seconds, nli59Seconds, nliUnknown);

     tNTPHdrStruct = packed record
                                  Flags              : byte;
                                  Stratum            : byte;
                                  Poll               : byte;
                                  Precision          : byte;
                                  RootDelay          : longword;
                                  RootDispersion     : longword;
                                  ReferenceID        : longword;
                                  ReferenceTimestamp : uint64;
                                  OriginTimestamp    : uint64;
                                  ReceiveTimestamp   : uint64;
                                  TransmitTimestamp  : uint64;
                       end;
     pNTPHdrStruct = ^tNTPHdrStruct;

     tNTPParser  = class (tFrameParser)

                           private
                                  priv_LeapIndicator      : tntpLeapIndicator;
                                  priv_VersionNumber      : byte;
                                  priv_NTPMode            : tNTPMode;
                                  priv_Stratum            : byte;
                                  priv_Poll               : byte;
                                  priv_Precision          : byte;
                                  priv_RootDelay          : longword;
                                  priv_RootDispersion     : longword;
                                  priv_ReferenceID        : longword;
                                  priv_ReferenceTimestamp : tFrametimeStamp;
                                  priv_OriginTimestamp    : tFrametimeStamp;
                                  priv_ReceiveTimestamp   : tFrametimeStamp;
                                  priv_TransmitTimestamp  : tFrametimeStamp;

                                  procedure ParseFlag (const flag : byte);

                           public
                                 property LeapIndicator : tntpLeapIndicator read priv_LeapIndicator;
                                 property VersionNumber : byte read priv_VersionNumber;
                                 property NTPMode : tNTPMode read priv_NTPMode;
                                 property Stratum : byte read priv_Stratum;
                                 property Poll : byte read priv_poll;
                                 property Precision : byte read priv_Precision;
                                 property RootDelay : longword read priv_RootDelay;
                                 property RootDispersion : longword read priv_RootDispersion;
                                 property ReferenceID : longword read priv_ReferenceID;
                                 property ReferenceTimestamp : tFrametimeStamp read priv_ReferenceTimestamp;
                                 property OriginTimestamp : tFrametimeStamp read priv_OriginTimestamp;
                                 property ReceiveTimestamp : tFrametimeStamp read priv_ReceiveTimestamp;
                                 property TransmitTimestamp : tFrametimeStamp read priv_TransmitTimestamp;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                       end;

implementation

uses Toolkit;

(* ------------------------------------------- *)
{ tNTPParser }
(* ------------------------------------------- *)

function tNTPParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

procedure tNTPParser.ParseFlag (const flag : byte);

var help : byte;

begin
     help := (flag and $c0) shr 6;

     case help of
      0 : priv_LeapIndicator := nliNoWarning;
      1 : priv_LeapIndicator := nli61Seconds;
      2 : priv_LeapIndicator := nli59Seconds;
      3 : priv_LeapIndicator := nliUnknown;
     end;

     priv_VersionNumber := (flag and $38) shr 3;

     help := (flag and $7);

     case help of
      0 : priv_NTPMode := ntmReserved;
      1 : priv_NTPMode := ntmSymActive;
      2 : priv_NTPMode := ntSymPassive;
      3 : priv_NTPMode := ntmClient;
      4 : priv_NTPMode := ntmServer;
      5 : priv_NTPMode := ntmBroadcast;
      6 : priv_NTPMode := ntmNTPControlMsg;
      7 : priv_NTPMode := ntmReservedPrivate;
     end;
end;

(* ------------------------------------------- *)

function tNTPParser.parse: tParseResult;

var status    : tParseResult;
    pData     : pNTPHdrStruct;
    bHelp     : byte;
    wHelp     : word;
    lHelp     : longword;

begin
     priv_osilayer := 5;     // Set Layer this parser is for
     prot_layersize := sizeOf (tNTPHdrStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tNTPHdrStruct) then
        status := prFailUndersize
     else
         try
            pData := pNTPHdrStruct (priv_framedata + priv_frameoffset);

            ParseFlag (pData^.Flags);

            priv_Stratum := pData^.Stratum;
            priv_Poll := pData^.Poll;
            priv_Precision := pData^.Precision;
            priv_RootDelay := pData^.RootDelay;
            priv_RootDispersion := swap32 (pData^.RootDispersion);
            priv_ReferenceID := swap32 (pData^.ReferenceID);
            priv_ReferenceTimestamp := NTPTimeStampToFrameTime (swap64 (pData^.ReferenceTimestamp));
            priv_OriginTimestamp := NTPTimeStampToFrameTime (swap64 (pData^.OriginTimestamp));
            priv_ReceiveTimestamp := NTPTimeStampToFrameTime (swap64 (pData^.ReceiveTimestamp));
            priv_TransmitTimestamp := NTPTimeStampToFrameTime (swap64 (pData^.TransmitTimestamp));

            // Determine next anchor

            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tNTPHdrStruct);

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.