{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_tokenring;

interface

uses base_structures,
     common_parse_helpers,
     definition_ip_addresses,
     class_parser_baseclass;


type tTokenRingParser  = class (tFrameParser)

                               private
                                      priv_AccessControl    : byte;
                                      priv_FrameControl     : byte;
                                      priv_Destination      : tMACBytes;
                                      priv_Source           : tMACBytes;
                                      priv_RoutingByteCount : byte;
                                      priv_RoutingControl   : word;
                                      priv_RouteDescriptors : array of word;


                               public
                                     property AccessControl : byte read priv_AccessControl;
                                     property FrameControl : byte read priv_FrameControl;
                                     property SourceAddress : tMACBytes read priv_Source;
                                     property DestinationAddress : tMACBytes read priv_Destination;
                                     property RoutingByteCount : byte read priv_RoutingByteCount;

                                     function checkIsPlausible : boolean;
                                     function parse : tParseResult; override;
                         end;

     pStruct_Tokenring = ^tStruct_TokenRing;
     tStruct_TokenRing = packed record
                                AccessControl        : byte;
                                FrameControl         : byte;
                                Destination          : tMACBytes;
                                Source               : tMACBytes;
                                RoutingByteCount     : byte;
                         end;

     pStruct_802dot3 = ^tStruct_802dot3;
     tStruct_802dot3 = packed record
                              DSAP           : byte;                  // Destination Service Access Point
                              SSAP           : byte;                  // Source Service Access Point, $AA for SNAP following the LLC
                              Control        : byte;                  // LLC Field Type
                       end;



implementation

uses Toolkit;

(* ------------------------------------------- *)
{ tTokenRingParser }
(* ------------------------------------------- *)

function tTokenRingParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tTokenRingParser.parse: tParseResult;

var pData  : pStruct_TokenRing;
    status : tParseResult;
    i      : integer;
    offset : integer;
    pHelp  : pWord;
    help   : word;

begin
     priv_osilayer := 2;     // Set Layer this parser is for

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tStruct_TokenRing) then
        status := prFailUndersize
     else
         try
            pData := pStruct_TokenRing (priv_framedata + priv_frameoffset);

            priv_AccessControl := pData^.AccessControl;
            priv_FrameControl := pData^.FrameControl;
            priv_Destination := pData^.Destination;
            priv_Source := pData^.Source;
            priv_RoutingByteCount := pData^.RoutingByteCount;

            offset := priv_frameoffset + sizeOf (tStruct_TokenRing);

            // first word is Routing Control, followed by n words of Route Descriptors

            if priv_RoutingByteCount > 2 then
               setLength (priv_RouteDescriptors, (priv_RoutingByteCount - 2) DIV 2);

            for i := 1 to (priv_RoutingByteCount DIV 2) do
                begin
                     pHelp := pWord (priv_framedata + offset);
                     help := pHelp^;
                     help := Swap16 (help);

                     // The first word is Routing Control, the rest are Route Descriptors

                     if i = 1 then
                        priv_RoutingControl := help
                     else
                         priv_RouteDescriptors[i] := help;

                     inc (offset);
                end;






{
            priv_nextAnchor.HeaderType := priv_ethertypelength;
            priv_nextAnchor.Offset := sizeOf (tStruct_Ethernet);
}
         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
