{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_dbfilemanager;

interface

uses SysUtils,
     base_structures,
     definition_ip_addresses,

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite,
     Data.DB, FireDAC.Comp.Client;

type tDBFileManager = class (tObject)

                            constructor create (DBFilename : TFileName);

                            private
                                   priv_filename           : TFileName;
                                   priv_db                 : TFDConnection;
                                   priv_sharedmode         : boolean;

                                   function DBisConnected : boolean;
                            public
                                  property Conntected : boolean read DBisConnected;
                                  property SharedMode : boolean read priv_sharedmode write priv_sharedmode;

                                  destructor Destroy; override;

                                  procedure Connect;
                                  procedure Disconnect;

                                  procedure UpdateDatabase;
                      end;

implementation

uses class_dirparser,
     class_inteldb_manager;

(* ------------------------------------------ *)
{ tDBFileManager }
(* ------------------------------------------ *)

constructor tDBFileManager.create (DBFilename: TFileName);

begin
     inherited create;

     priv_filename := DBFilename;
     priv_db := nil;
     priv_sharedmode := False;
end;

(* ------------------------------------------ *)

destructor tDBFileManager.Destroy;

begin
     if priv_db <> nil then
        begin
             priv_db.Free;
             priv_db := nil;
        end;

     inherited;
end;

(* ------------------------------------------ *)

procedure tDBFileManager.Connect;

begin
     if priv_db <> nil then
        begin
             priv_db.Free;
             priv_db := nil;
        end;

     if FileExists (priv_filename) then
        begin
             priv_db := TFDConnection.Create(nil);
             priv_db.DriverName := 'SQLite';
             priv_db.Params.Add ('SQLite');
             priv_db.Params.Add ('Database=' + priv_filename);

             if priv_sharedmode then
                begin
                     priv_db.Params.Add('LockingMode=Normal');
                     priv_db.Params.Add('Synchronous=Normal');
                end;

             priv_db.Open;
        end;
end;

(* ------------------------------------------ *)

procedure tDBFileManager.Disconnect;

begin
     if priv_db.Connected then
        priv_db.Close;
end;

(* ------------------------------------------ *)

function tDBFileManager.DBisConnected: boolean;

begin
     result := (priv_db <> nil) and (priv_db.Connected);
end;

(* ------------------------------------------ *)

procedure tDBFileManager.UpdateDatabase;

begin

end;

(* ------------------------------------------ *)

end.
