{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_diagmessagehandler;

interface

uses base_structures;

type tDiagnosticMsg = record
                            id             : integer;               // for eventual DB storage; usually set to 0 otherwise
                            parsertype     : integer;               // Used to tell what parser the message was created by
                            timestamp      : tDateTime;             // timestamp of the message being created
                            severity       : tMsgSeverity;
                            messagehandle  : string;                // Unique message handle for the given kind of message
                            frameindex     : int64;                 // 0 if not applicable
                            param1         : string;                // Additional parameter 1
                            param2         : string;                // Additional parameter 2
                            param3         : string;                // Additional parameter 3
                            param4         : string;                // Additional parameter 4

//                          procedure initialize (const MsgSeverity : tMsgSeverity; const MsgHandle : string);
                            procedure initialize; overload;
                            procedure initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer); overload;
                            procedure initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer; const MsgParam1 : string); overload;
                            procedure initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer; const MsgParam1, MsgParam2 : string); overload;
                            procedure initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer; const MsgParam1, MsgParam2, MsgParam3 : string); overload;
                            procedure initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer; const MsgParam1, MsgParam2, MsgParam3, MsgParam4 : string); overload;
                            function DisplayText : string;
                      end;

     tDiagnosticMsgList = array of tDiagnosticMsg;

{
     tDiagnosticMsgHandler = class (tObject)

                                   constructor create;
                                   procedure free;

                                   private
                                          priv_messagelist : tDiagnosticMsgList;

                                   public
                                         procedure clear;
                                         function Count : integer;

                                         procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer); overload;
                                         procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer; const Param1 : string); overload;
                                         procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer; const Param1, Param2 : string); overload;
                                         procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer; const Param1, Param2, Param3 : string); overload;
                                         procedure addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer; const Param1, Param2, Param3, Param4 : string); overload;

                                         function getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;
                             end;
}



implementation

uses SysUtils,
     base_messages;

(* ----------------------------------- *)
{tDiagnosticMsg}
(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize;

begin
     id             := 0;
     timestamp      := 0;
     severity       := msNone;
     messagehandle  := '';
     frameindex     := 0;
     param1         := '';
     param2         := '';
     param3         := '';
     param4         := '';
end;

(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize (const MsgSeverity : tMsgSeverity; const MsgID : integer);

begin
     id             := MsgID;
     timestamp      := now;
     severity       := MsgSeverity;
     messagehandle  := GetMessageHandle (MsgID);
     frameindex     := 0;
     param1         := '';
     param2         := '';
     param3         := '';
     param4         := '';
end;

(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize (const MsgSeverity: tMsgSeverity; const MsgID: integer; const MsgParam1: string);

begin
     initialize (MsgSeverity, MsgID);

     param1 := MsgParam1;
end;

(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize (const MsgSeverity: tMsgSeverity; const MsgID: integer; const MsgParam1, MsgParam2: string);

begin
     initialize (MsgSeverity, MsgID);

     param1 := MsgParam1;
     param2 := MsgParam2;
end;

(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize (const MsgSeverity: tMsgSeverity; const MsgID: integer; const MsgParam1, MsgParam2, MsgParam3: string);

begin
     initialize (MsgSeverity, MsgID);

     param1 := MsgParam1;
     param2 := MsgParam2;
     param3 := MsgParam3;
end;

(* ----------------------------------- *)

procedure tDiagnosticMsg.initialize (const MsgSeverity: tMsgSeverity; const MsgID: integer; const MsgParam1, MsgParam2, MsgParam3, MsgParam4: string);

begin
     initialize (MsgSeverity, MsgID);

     param1 := MsgParam1;
     param2 := MsgParam2;
     param3 := MsgParam3;
     param4 := MsgParam4;
end;

(* ----------------------------------- *)

function tDiagnosticMsg.DisplayText: string;

var help : string;

begin
     help := GetDiagMessageBaseText (id);

     result := Format (help, [param1, param2, param3, param4]);
end;

(* ----------------------------------- *)
{ tDiagnosticMsgHandler }
(* ----------------------------------- *)
{constructor tDiagnosticMsgHandler.create;

begin
     inherited;

     setLength (priv_messagelist, 0);
end;

(* ----------------------------------- *)

procedure tDiagnosticMsgHandler.free;

begin
     setLength (priv_messagelist, 0);

     inherited;
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.addDiagnosticMsg (const Severity : tMsgSeverity; const MsgID : integer);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgID);

     RecCount := length (priv_messagelist) + 1;
     setLength (priv_messagelist, RecCount);
     priv_messagelist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.addDiagnosticMsg(const Severity: tMsgSeverity; const MsgID : integer; const Param1: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgID);
     MsgRecord.param1 := Param1;

     RecCount := length (priv_messagelist) + 1;
     setLength (priv_messagelist, RecCount);
     priv_messagelist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.addDiagnosticMsg(const Severity: tMsgSeverity; const MsgID : integer; const Param1, Param2: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgID);
     MsgRecord.param1 := Param1;
     MsgRecord.param2 := Param2;

     RecCount := length (priv_messagelist) + 1;
     setLength (priv_messagelist, RecCount);
     priv_messagelist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.addDiagnosticMsg(const Severity: tMsgSeverity; const MsgID : integer; const Param1, Param2, Param3: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgID);
     MsgRecord.param1 := Param1;
     MsgRecord.param2 := Param2;
     MsgRecord.param3 := Param3;

     RecCount := length (priv_messagelist) + 1;
     setLength (priv_messagelist, RecCount);
     priv_messagelist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.addDiagnosticMsg(const Severity: tMsgSeverity; const MsgID : integer; const Param1, Param2, Param3, Param4: string);

var RecCount  : integer;
    MsgRecord : tDiagnosticMsg;

begin
     MsgRecord.initialize(Severity, MsgID);
     MsgRecord.param1 := Param1;
     MsgRecord.param2 := Param2;
     MsgRecord.param3 := Param3;
     MsgRecord.param4 := Param4;

     RecCount := length (priv_messagelist) + 1;
     setLength (priv_messagelist, RecCount);
     priv_messagelist[RecCount-1] := MsgRecord;
end;

(* ------------------------------------------- *)

function tDiagnosticMsgHandler.Count : integer;

begin
     result := length (priv_messagelist);
end;

(* ------------------------------------------- *)

function tDiagnosticMsgHandler.getDiagnosticMsg (const Index : integer) : tDiagnosticMsg;

begin
     result := priv_messagelist[Index - 1];
end;

(* ------------------------------------------- *)

procedure tDiagnosticMsgHandler.Clear;

begin
     setLength (priv_messagelist, 0);
end;
}
(* ------------------------------------------- *)

end.
