{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_formstorage_extracttask;

interface

{$ifndef _CONSOLE_TESTRUNNER}
{$i 'base\jedi.inc')}
{$else}
{$i '..\base\jedi.inc')}
{$endif}

uses Forms,
     Classes,

{$ifdef DELPHI20_UP}
     System.JSON,
{$endif}

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite, FireDAC.Phys.SQLiteWrapper,
     Data.DB, FireDAC.Comp.Client, FireDAC.Stan.Param,

     // Smartinspect

     VirtualTrees,
     SiAuto,
     class_formstorage,
     definition_filter_structures,
     class_parse_snortalertlog;

type tExtractTaskFormStorageManager = class (tFormStorageManager)

                                         constructor create (AssociatedForm : tForm);
                                         procedure free;

                                         private

                                         protected
                                                  function getDBStructure : string; override;

                                         public

                                               function storeFormToDB (DatabaseFile : string; const Comment : string = '') : integer; virtual;
                                               function retrieveFormFromDB (DatabaseFile : string) : integer; virtual;

                                               procedure StoreFiltersToDB (FilterList : tFilterSet);
                                               procedure LoadFiltersFromDB (var List : tFilterSet);
                                               procedure StoreMetaDataToDB (MetaData : tMetaDataSet);
                                               function LoadMetaDataFromDB : tMetaDataSet;

                                   end;

implementation

uses SysUtils,
{$ifdef VER270}
     System.JSON,
{$endif}
     DBXJSON,
     base_structures,
     definition_ip_addresses;

const CR = chr(13);
      LF = chr(10);
      CRLF = CR + LF;
      ExtractSettingsTables = 'CREATE TABLE "formFilters" ("id" INTEGER PRIMARY KEY NOT NULL UNIQUE, "expression" CHAR NOT NULL);' + CRLF +
                              'CREATE TABLE "formMetaData" ("id" INTEGER PRIMARY KEY NOT NULL UNIQUE, "expression" CHAR NOT NULL);';

(* -------------------------------------------- *)
{ tExtractTaskFormStorageManager }
(* -------------------------------------------- *)

constructor tExtractTaskFormStorageManager.create(AssociatedForm: TForm);

begin
     inherited create (AssociatedForm);
end;

(* -------------------------------------------- *)

procedure tExtractTaskFormStorageManager.free;

begin
     inherited free;
end;

(* -------------------------------------------- *)

function tExtractTaskFormStorageManager.getDBStructure: string;

var struct : string;

begin
     struct := ExtractSettingsTables + inherited;

     result := struct;
end;

(* -------------------------------------------- *)

function tExtractTaskFormStorageManager.retrieveFormFromDB (DatabaseFile: string): integer;

begin
     inherited;


end;

(* -------------------------------------------- *)

function tExtractTaskFormStorageManager.storeFormToDB (DatabaseFile: string; const Comment: string): integer;

begin
     inherited;
end;

(* -------------------------------------------- *)

procedure tExtractTaskFormStorageManager.StoreFiltersToDB (FilterList: tFilterSet);

var filter : tFrameFilter;
    query  : tFDQuery;
    i      : integer;
    JObj   : tJSONObject;

begin
     query := tFDQuery.Create(nil);
     query.Connection := priv_database;

     filter := FilterList.getFirst;
     i := 1;

//   for i := 1 to length (filters) do
     while not filter.isUnspecified do
           begin
//              filter := FilterList[i-1];
                JObj := filter.toJSON;

                query.SQL.Text := 'INSERT INTO formFilters (id,expression) VALUES (:id,:expression);';
                query.ParamByName ('id').AsInteger := i;
                query.ParamByName ('expression').AsString := JObj.ToString;

                query.ExecSQL;

                JObj.Free;

                filter := FilterList.getNext;
                inc (i);
           end;

     query.Free;
end;

(* -------------------------------------------- *)

procedure tExtractTaskFormStorageManager.LoadFiltersFromDB (var List : tFilterSet);

var query   : tFDQuery;
    filter  : tFrameFilter;
    jsonstr : string;
    i       : integer;

begin
     query := tFDQuery.Create(nil);
     query.Connection := priv_database;

     query.SQL.Text := 'SELECT * FROM formFilters ORDER BY id;';
     query.Open;
     query.FetchAll;

     list.clear;

//   setlength (value, Query.RecordCount);

     for i := 1 to query.RecordCount do
         begin
              jsonstr := Query.FieldByName ('expression').AsString;

//            list[i-1].initialize (jsonstr);
              filter.initialize (jsonstr);
              list.AddFilter (filter);

              query.Next;
         end;

     query.Free;
end;

(* -------------------------------------------- *)

procedure tExtractTaskFormStorageManager.StoreMetaDataToDB (MetaData : tMetaDataSet);

var entry  : tMetaDataEntry;
    query  : tFDQuery;
    i      : integer;
    JObj   : tJSONObject;

begin
     query := tFDQuery.Create(nil);
     query.Connection := priv_database;

     for i := 1 to length (MetaData) do
         begin
              entry := MetaData[i-1];
              JObj := entry.toJSON;

              query.SQL.Text := 'INSERT INTO formMetaData (id,expression) VALUES (:id,:expression);';
              query.ParamByName ('id').AsInteger := i;
              query.ParamByName ('expression').AsString := JObj.ToString;

              query.ExecSQL;

              JObj.Free;
         end;

     query.Free;
end;

(* -------------------------------------------- *)

function tExtractTaskFormStorageManager.LoadMetaDataFromDB: tMetaDataSet;

var value   : tMetaDataSet;
    query   : tFDQuery;
    jsonstr : string;
    i       : integer;

begin
     query := tFDQuery.Create(nil);
     query.Connection := priv_database;

     query.SQL.Text := 'SELECT * FROM formMetaData ORDER BY id;';
     query.Open;
     query.FetchAll;

     setlength (value, Query.RecordCount);

     for i := 1 to query.RecordCount do
         begin
              jsonstr := Query.FieldByName ('expression').AsString;

              value[i-1].initialize (jsonstr);

              query.Next;
         end;

     query.Free;

     result := value;
end;


(* -------------------------------------------- *)

end.
