{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_sanitize;

interface

uses System.Generics.Collections,
     System.Diagnostics,
     System.TimeSpan,
     base_structures,
     pcapng_structures,
     definition_anonymization_struct,
     definition_simple_nameresolution,
     definition_ip_addresses,
     definition_tcp_structures,
     base_checksums,
     class_tracefilelist,
     class_replacement_manager,
     class_taskprocessor_baseclass,
     class_rawframe,
     class_tracewriter,
     class_loader_pcapng,
     class_writer_pcapng,

     // Parser

     class_parser_baseclass,
     class_parser_802dot11_radiotap,
     class_parser_802dot11header,
     class_parser_802dot11management,
     class_parser_ethernet,
     class_parser_linux_sll,
     class_parser_tokenring,
     class_parser_802dot1q,
     class_parser_mpls,
     class_parser_llc,
     class_parser_arp,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_ipv6_fragmentationheader,
     class_parser_ayiya,
     class_parser_gre,
     class_parser_gtpv1,
     class_parser_vxlan,
     class_parser_icmpv4,
     class_parser_icmpv6,
     class_parser_tcp,
     class_parser_udp,
     class_parser_dhcpv4,
     class_parser_netflow,
     class_parser_hsrp,
     class_parser_rtps,
     class_parsemaster,
     class_ipfragment_manager;


type tQueryOnTargetFileExists = procedure (FileEntry : tTracefileEntry; var TargetFile : string; var AbortFile : boolean) of object;
     tQueryPCAPngSHBDetails = procedure (FileEntry : tTracefileEntry; OptionQueryList : tStringQueryList) of object;
     tQueryPCAPngNameResolution = procedure (FileEntry : tTracefileEntry; var IPv4List : tSimpleIPv4NameResolutionList; var IPv6List : tSimpleIPv6NameResolutionList) of object;

     tSanitizationProcessEntry = record
                                       IPAddress            : tIPAddress;
                                       IPIdentification     : longword;
                                       Frame                : tRawFrame;
                                       FragmentationDetails : tIPFragment;
                                       isFragment           : boolean;
                                 end;

     tSanitizationProcessChain = array of tSanitizationProcessEntry;

     tTaskProcessorAnonymize = class (tTaskProcessor)

                                     constructor create (InputFile : tTracefileEntry; Output : tOutputlocation; ReplacementManager : tReplacementManager; const Parameters: tAnonymizeParameters);
                                     procedure free;

                                     private
                                            EventOnTargetFileExists           : tQueryOnTargetFileExists;
                                            EventOnQueryPCAPngHeaderDetails   : tQueryPCAPngSHBDetails;
                                            EventOnQueryNameResolutionDetails : tQueryPCAPngNameResolution;

                                            priv_tracewriter                  : tPCAPngWriter;
                                            priv_AbortTask                    : boolean;
                                            priv_FragmentModeActive           : boolean;

                                            priv_params                       : tAnonymizeParameters;
                                            priv_replacementmgr               : tReplacementManager;

                                            priv_processchain                 : tSanitizationProcessChain;
                                            priv_ipfragmentmgr                : tIPFragmentManager;
                                            priv_reassemblystorage            : tObjectDictionary<string, tRawFrame>;

                                            priv_StopWatch                    : tStopWatch;
                                            priv_elapsed                      : tTimeSpan;
                                            priv_ProtocolsSeen                : tParsedProtocolSet;

                                            function SetFileTimestamp (const timestamp : tDateTime) : integer;
                                            /// <summary>
                                            ///   This routine checks if the
                                            ///   ICMP type and code was
                                            ///   configured to be completely
                                            ///   removed from the sanitized
                                            ///   trace via the anonymization
                                            ///   parameter dialog
                                            /// </summary>
                                            /// <param name="ParseMaster">
                                            ///   Parsed Frame to avoid parsing
                                            ///   it again
                                            /// </param>
                                            /// <returns>
                                            ///   True if the ICMP Type/Code is
                                            ///   configured to be removed
                                            /// </returns>
                                            function RemoveFrameFromTrace (ParseMaster : tParseMaster) : boolean;

                                            procedure SanitizeSHBHeaders (Loader : tPCAPNGLoader; Writer : tPCAPNGWriter);
                                            procedure SanitizeInterfaceBlock (Loader : tPCAPNGLoader; Writer : tPCAPNGWriter);
                                            procedure SanitizeNameResolutionBlock (Loader : tPCAPNGLoader; Writer : tPCAPNGWriter);

                                            procedure sanitizePCAPngHeaders (Loader : tPCAPNGLoader; Writer : tPCAPNGWriter);
                                            procedure SanitizeIPv4PseudoHeader (var IPv4PseudoHeader : tIPv4PseudoHeader);
                                            procedure SanitizeIPv6PseudoHeader (var IPv6PseudoHeader : tIPv6PseudoHeader);

                                            procedure SanitizeRTPSParameter (var Parameter : tRTPSSerializedDataEntry);

                                            function processFragmentChain : tTaskErrorCode;

                                            /// <summary>
                                            ///   This routine sanitizes a
                                            ///   single frame in stateless
                                            ///   mode (when there are not
                                            ///   fragments to deal with)
                                            /// </summary>
                                            function ProcessSingleFrame (var OriginalFrame : tRawFrame) : tTaskErrorCode;
                                            /// <summary>
                                            ///   This routine handles queueing
                                            ///   the frames when having to
                                            ///   deal with fragments
                                            /// </summary>
                                            function AddFrameToReassemblyProcessChain (const IPAddress : tIPAddress; const IPID : longword; const IPFragmentDetails : tIPFragment) : tTaskErrorCode;
                                            function ProcessFrame (var OriginalFrame : tRawFrame) : tTaskErrorCode;

                                            function getReplacementPayload (Payload : tFrameChunk) : tFrameChunk;
                                            function getSanitizedPayload (ParseMaster : tParseMaster) : tFrameChunk;

                                            function getIPv6PseudoHeader (ParseMaster : tParseMaster; StartLayer : integer) : tIPv6PseudoHeader;

                                            {$REGION 'Documentation'}
                                            ///	<param name="ForceEtherType">
                                            ///	  If not zero the assembly will
                                            ///	  use the given EtherType. This
                                            ///	  happens when a VLAN layer was
                                            ///	  removed earlier.
                                            ///	</param>
                                            {$ENDREGION}
                                            function GetEthernetAssembly (Parser : tEthernetParser; ForceEtherType : word) : tFrameChunk;
                                            function GetLinuxCookedAssembly (Parser : tLinuxSLLParser) : tFrameChunk;

                                            function Get802dot11RadioTapAssembly (Parser : t802dot11RadioTapParser) : tFrameChunk;
                                            function Get802dot11HdrAssembler (Parser : t802dot11HeaderParser) : tFrameChunk;
                                            function Get802dot11MgmtAssembler (Parser : t802dot11ManagementParser) : tFrameChunk;
                                            function GetLLCAssembly (Parser : tLLCParser) : tFrameChunk;
                                            {$REGION 'Documentation'}
                                            ///	<param name="ForceEtherType">
                                            ///	  If not zero the assembly will
                                            ///	  use the given EtherType. This
                                            ///	  happens when a VLAN layer was
                                            ///	  removed earlier.
                                            ///	</param>
                                            {$ENDREGION}
                                            function GetVLANAssembly (Parser : t802dot1qParser; ForceEtherType : word) : tFrameChunk;
                                            function GetARPAssembly (Parser : tARPParser) : tFrameChunk;
//                                          function GetICMPv4Assembly (Parser : tICMPv4Parser; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetICMPv4Assembly (Parser : tICMPv4Parser; QuotedPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetICMPv6Assembly (Parser : tICMPv6Parser; IPv6PseudoHeader : tIPv6PseudoHeader; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetICMPv6Assembly (Parser : tICMPv6Parser; IPv6PseudoHeader : tIPv6PseudoHeader; QuotedPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetMPLSAssembly (Parser : tMPLSParser) : tFrameChunk;
                                            function GetIPv4Assembly (Parser : tIPv4Parser; PseudoHeader : tIPv4PseudoHeader; PortPair : tPortPair; PayloadLen : integer; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetIPv4Assembly (Parser : tIPv4Parser; PortPair : tPortPair; PayloadLen : integer; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetIPv6Assembly (Parser : tIPv6Parser; PortPair : tPortPair) : tFrameChunk; overload;
                                            function GetIPv6Assembly (Parser : tIPv6Parser; PseudoHeader : tIPv6PseudoHeader; PortPair : tPortPair) : tFrameChunk; overload;
                                            function GetIPv6FragmentationAssembly (Parser : tIPv6FragmentationParser) : tFrameChunk;
                                            function GetTCPAssembly (Parser: tTCPParser; OriginalIPPair : tIPv4AddressPair; IPv4PseudoHeader : tIPv4PseudoHeader; TCPPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetTCPAssembly (Parser: tTCPParser; OriginalIPPair : tIPv6AddressPair; IPv6PseudoHeader : tIPv6PseudoHeader; TCPPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetUDPAssembly (Parser: tUDPParser; OriginalIPPair : tIPv4AddressPair; IPv4PseudoHeader : tIPv4PseudoHeader; UDPPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetUDPAssembly (Parser: tUDPParser; OriginalIPPair : tIPv6AddressPair; IPv6PseudoHeader : tIPv6PseudoHeader; UDPPayload : tFrameChunk; CRCWasOkay : boolean) : tFrameChunk; overload;
                                            function GetVXLANAssembly (Parser : tVXLANParser) : tFrameChunk;
                                            function GetAyiyaAssembly (Parser : tAYIYAParser) : tFrameChunk;
                                            function GetGREAssembly (Parser : tGREParser) : tFrameChunk;
                                            function GetGTPv1Assembly (Parser : tGTPv1Parser) : tFrameChunk;
                                            function GetDHCPv4Assembly (Parser : tDHCPv4Parser) : tFrameChunk;
                                            function GetNetFlowAssembly (Parser : tNetFlowParser) : tFrameChunk;
                                            function GetHSRPAssembly (Parser : tHSRPParser) : tFrameChunk;
                                            function GetRTPSAssembly (Parser : tRTPSParser) : tFrameChunk;

                                            function getNewRawFrameFromFrameChunk (Frame : tRawFrame; Payload : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;
                                            function getSanitizedFrame (Frame : tRawFrame) : tRawFrame;

                                     protected


                                     public
                                           property OnQueryTargetFileExists : tQueryOnTargetFileExists read EventOnTargetFileExists write EventOnTargetFileExists;
                                           property OnQueryPCAPngHeaderDetails : tQueryPCAPngSHBDetails read EventOnQueryPCAPngHeaderDetails write EventOnQueryPCAPngHeaderDetails;
                                           property OnQueryNameResolutionDetails : tQueryPCAPngNameResolution read EventOnQueryNameResolutionDetails write EventOnQueryNameResolutionDetails;

                                           function Execute: tTaskErrorCode; override;
                               end;

implementation

uses SysUtils,
     DateUtils,
     Math,
     SiAuto,
     toolkit,
     common_parse_helpers,

     // Loader & Writer

     class_interfacelist,
     class_traceloader,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,

     // Assemblies

     class_assembly_802dot11_radiotap,
     class_assembly_802dot11header,
     class_assembly_802dot11management,
     class_assembly_ethernet,
     class_assembly_linux_sll,
     class_assembly_802dot1q,
     class_assembly_llc,
     class_assembly_arp,
     class_assembly_mpls,
     class_assembly_ipv4,
     class_assembly_ipv6,
     class_assembly_ipv6_fragmentationheader,
     class_assembly_icmpv4,
     class_assembly_icmpv6,
     class_assembly_ayiya,
     class_assembly_gre,
     class_assembly_gtpv1,
     class_assembly_vxlan,
     class_assembly_tcp,
     class_assembly_udp,
     class_assembly_dhcpv4,
     class_assembly_netflow,
     class_assembly_hsrp,
     class_assembly_rtps;

var TraceWranglerHint : string;

(* ------------------------------------------- *)
{ tTaskProcessorAnonymize }
(* ------------------------------------------- *)

constructor tTaskProcessorAnonymize.create (InputFile: tTracefileEntry; Output: tOutputlocation; ReplacementManager : tReplacementManager; const Parameters: tAnonymizeParameters);

begin
     inherited create (InputFile, Output);

     priv_params := Parameters;
     priv_replacementmgr := ReplacementManager;
     priv_ipfragmentmgr := tIPFragmentManager.create;
     priv_reassemblystorage := tObjectDictionary<string, tRawFrame>.create ([doOwnsValues]);

     TraceWranglerHint := 'Sanitized by TraceWrangler v' + GetExeVersion (ParamStr(0))
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.free;

begin
//   priv_replacementmgr.free;         This is done in main.pas
     priv_ipfragmentmgr.Free;
     priv_reassemblystorage.Free;

     inherited free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.SetFileTimestamp (const timestamp: tDateTime) : integer;

var rcode    : integer;
    filename : string;

begin
     if timestamp <> 0 then
        begin
             filename := prot_OutputLocation.getOutputfilename;

             rcode := FileSetDate (filename, DateTimeToFileDate(timestamp));
        end;

     result := rcode;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.RemoveFrameFromTrace (ParseMaster: tParseMaster): boolean;

var ICMPv4Parser : tICMPv4Parser;
    i,j          : integer;
    value        : boolean;

begin
     value := false;

     // Check ICMPv4 types/codes

     if ppICMPv4 in ParseMaster.ParsedProtocols then
        begin
             i := 0;

             // Find the ICMP Parser(s)

             repeat
                   inc (i);

                   if (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_ICMP) then
                      begin
                           ICMPv4Parser := (Parsemaster.getParserByIndex (i) as tICMPv4Parser);

                           for j := 1 to length (priv_params.ICMPv4RemoveByTypeCode) do
                               if (priv_params.ICMPv4RemoveByTypeCode[j-1].ICMPv4Type = ICMPv4Parser.ICMPType) and
                                  (priv_params.ICMPv4RemoveByTypeCode[j-1].ICMPv4Code = ICMPv4Parser.ICMPCode) then
                                  value := true;
                      end;

             until (i >= Parsemaster.ParserCount);
        end;

     result := value;
end;

(* ------------------------------------------- *)
{ PCAPng Header Sanitization }
(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeSHBHeaders (Loader: tPCAPNGLoader; Writer: tPCAPNGWriter);

var HelpString      : string;
    QueryList       : tStringQueryList;
    i               : integer;

begin
     QueryList.initialize;

     if Loader.HeaderComments.Count = 0 then
        Writer.HeaderCommentList.Add (TraceWranglerHint)
     else
         for i := 1 to Loader.HeaderComments.Count do
             begin
                  HelpString := Loader.HeaderComments[i-1];

                  if (i = 1) and (pos ('Sanitized by TraceWrangler', helpstring) = 0) then
                     HelpString := HelpString + CRLF + TraceWranglerHint;

                  // Write the comment if we only remove packet comments, otherwise handle it (ask or just leave it away)

                  if priv_params.PCAPngSanitizeCommentsScope = scsPacket then
                     Writer.HeaderCommentList.Add(HelpString)
                  else
                      if priv_params.PCAPngSanitizeCommentsMode = scmAsk then
                         QueryList.add(qtSHBComment, Helpstring);
             end;

     for i := 1 to Loader.Hardware.Count do
         begin
              HelpString := Loader.Hardware[i-1];

              // Write the comment if we only remove packet comments, otherwise handle it (ask or just leave it away)

              if priv_params.PCAPngSanitizeCommentsScope = scsPacket then
                 Writer.HardwareInfoList.Add(HelpString)
              else
                  if priv_params.PCAPngSanitizeCommentsMode = scmAsk then
                     QueryList.add(qtSHBHardware, Helpstring);
         end;

     for i := 1 to Loader.OperatingSystem.Count do
         begin
              HelpString := Loader.OperatingSystem[i-1];

              // Write the comment if we only remove packet comments, otherwise handle it (ask or just leave it away)

              if priv_params.PCAPngSanitizeCommentsScope = scsPacket then
                 Writer.OperatingSystemList.Add(HelpString)
              else
                  if priv_params.PCAPngSanitizeCommentsMode = scmAsk then
                     QueryList.add(qtSHBOS, Helpstring);
         end;


     if priv_params.KeepOriginalApplicationString then
        begin
             for i := 1 to Loader.Userapplication.Count do
                 begin
                      HelpString := Loader.Userapplication[i-1];

                      // Write the comment if we only remove packet comments, otherwise handle it (ask or just leave it away)

                      if priv_params.PCAPngSanitizeCommentsScope = scsPacket then
                         Writer.UserApplicationList.Add(HelpString)
                      else
                          if priv_params.PCAPngSanitizeCommentsMode = scmAsk then
                             QueryList.add(qtSHBApplication, Helpstring);
                 end;
        end
     else
         Writer.UserApplicationList.Add (TraceWranglerHint);

     // TODO: handle GUID in PCAPngWriter

     // TODO: auto set replacement strings for those we already asked the user for in former traces

     // If we have to ask the user for changes, do that and apply them afterwards

     if ((priv_params.PCAPngSanitizeCommentsScope = scsFile) or (priv_params.PCAPngSanitizeCommentsScope = scsBoth)) and (priv_params.PCAPngSanitizeCommentsMode = scmAsk) then
        begin
             EventOnQueryPCAPngHeaderDetails (prot_Inputfile, QueryList);

             // TODO: Store changes to replacement manager - String storage routines are already there

             // Assign returned results

             for i := 1 to Querylist.Count do
                 begin
                      case QueryList.EntryList[i-1].StringOrigin of
                       qtSHBComment     : Writer.HeaderCommentList.Add(QueryList.EntryList[i-1].Replacement);
                       qtSHBHardware    : Writer.HardwareInfoList.Add(QueryList.EntryList[i-1].Replacement);
                       qtSHBOS          : Writer.OperatingSystemList.Add(QueryList.EntryList[i-1].Replacement);
                       qtSHBApplication : Writer.UserApplicationList.Add(QueryList.EntryList[i-1].Replacement);
                      end;
                 end;
        end;

     // Write to disk

     Writer.WriteSectionHeaderBlock;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeInterfaceBlock (Loader: tPCAPNGLoader; Writer: tPCAPNGWriter);

var InterfaceEntry : tCaptureInterface;
    InterfaceName  : string;
    FilterRec      : tPCAPNGCaptureFilter;
    CurrentGUID    : tGUID;
    NewGUID        : tGUID;
    i              : integer;

begin
     for i := 1 to Loader.CaptureInterfaces.Count do
         begin
              InterfaceEntry := Loader.CaptureInterfaces.getInterfaceByIndex(i-1);

              if priv_params.PCAPngIHBDoReplace then
                 begin
                      // Interface Names

                      case priv_params.PCAPngSanitizeIHBInterfaceName of
                       simReplaceGeneric : begin
                                                if pos ('\device\npf_{', lowercase (InterfaceEntry.Name)) > 0 then
                                                   begin
                                                        // Windows network card, based on GUID

                                                        InterfaceName := GetEntry (InterfaceEntry.Name, '_', 2);

                                                        CurrentGUID := StringToGUID (InterfaceName);

                                                        // Find GUID in Replacement DB

                                                        NewGUID := priv_replacementmgr.getReplacementForGUID(CurrentGUID);

                                                        // If we found nothing, create a new GUID and remember it

                                                        if NewGUID = ZERO_GUID then
                                                           begin
                                                                CreateGUID (NewGUID);

                                                                priv_replacementmgr.AddReplacementPair(CurrentGUID, NewGUID);
                                                           end;

                                                        InterfaceEntry.Name := GetEntry (InterfaceEntry.Name, '_', 1) + '_' + GUIDToString (NewGUID);
                                                   end
                                                else
                                                    begin
                                                         InterfaceEntry.Name := 'eth' + IntToStr (i-1);
                                                    end;
                                           end;

                       simSetToBlank     : begin
                                                InterfaceEntry.Name := '';
                                           end;
                      end;

                      // Capture Filters

                      case priv_params.PCAPngSanitizeIHBCaptureFilter of
                       cfmRemoveFilter : begin
                                              FilterRec.FilterType := cftUnknown;
                                              FilterRec.FilterString := '';

                                              InterfaceEntry.CaptureFilter := FilterRec;
                                         end;
                      end;
                 end;

              // TODO: handle IPs mentioned in capture filters, if any

              Writer.WriteInterfaceDescriptionBlock(InterfaceEntry);
         end;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeNameResolutionBlock (Loader: tPCAPNGLoader; Writer: tPCAPNGWriter);

var NREntryIPv4  : tSimpleIPv4NameResolutionEntry;
    NREntryIPv6  : tSimpleIPv6NameResolutionEntry;
    NewEntryIPv4 : tSimpleIPv4NameResolutionEntry;
    NewEntryIPv6 : tSimpleIPv6NameResolutionEntry;
    IPv4List     : tSimpleIPv4NameResolutionList;
    IPv6List     : tSimpleIPv6NameResolutionList;
    i            : integer;

begin
     IPv4List := Loader.NameResolutionManagerIPv4;
     IPv6List := Loader.NameResolutionManagerIPv6;

     case priv_params.PCAPngSanitizeNRB of
      snmAsk : begin
                    // Trigger opening the editor

                    EventOnQueryNameResolutionDetails (prot_Inputfile, IPv4List, IPv6List);

                    for i := 1 to IPv4List.count do
                        begin
                             // TODO : code NRB record replacement by popup values



                        end;

                    for i := 1 to IPv6List.count do
                        begin




                        end;
               end;
      snmRemoveBlock : begin
                            // Do nothing
                       end;
      snmReplaceIPs  : begin
                            // Do some automatic work on the entries

                           for i := 1 to IPv4List.count do
                               begin
                                    NREntryIPv4 := IPv4List.GetEntry(i);

                                    // Get replacement address (usually gets auto-stored for further references)

                                    NewEntryIPv4.IPv4Address :=  priv_replacementmgr.getReplacementForAddress(NREntryIPv4.IPv4Address);
                                    NewEntryIPv4.FQDN := 'sanitized.by.tracewrangler';                                                           // TODO: do real FQDN replacement

                                    // Store to PCAPng Writer

                                    Writer.IPv4NameResolutionMgr.AddEntry (NREntryIPv4);
                               end;

                           for i := 1 to IPv6List.count do
                               begin
                                    NREntryIPv6 := IPv6List.GetEntry(i);

                                    // Get replacement address (usually gets auto-stored for further references)

                                    NewEntryIPv6.IPv6Address :=  priv_replacementmgr.getReplacementForAddress(NREntryIPv6.IPv6Address);
                                    NewEntryIPv6.FQDN := 'sanitized.by.tracewrangler';                                                           // TODO: do real FQDN replacement

                                    Writer.IPv6NameResolutionMgr.AddEntry (NewEntryIPv6);
                               end;
                       end;

      snmNone        : begin
                            // Simply Transfer NRB values, if any

                            for i := 1 to Loader.NameResolutionManagerIPv4.count do
                                begin
                                     NREntryIPv4 := Loader.NameResolutionManagerIPv4.GetEntry(i);

                                     Writer.IPv4NameResolutionMgr.AddEntry(NREntryIPv4);
                                end;

                            for i := 1 to Loader.NameResolutionManagerIPv6.count do
                                begin
                                     NREntryIPv6 := Loader.NameResolutionManagerIPv6.GetEntry(i);

                                     Writer.IPv6NameResolutionMgr.AddEntry(NREntryIPv6);
                                end;
                       end;
     end;

     if (Writer.IPv4NameResolutionMgr.count > 0) or (Writer.IPv6NameResolutionMgr.count > 0) then
        Writer.WriteNameResolutionBlock;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.sanitizePCAPngHeaders (Loader: tPCAPNGLoader; Writer : tPCAPNGWriter);

begin
     if priv_params.PCAPngDoReplace then
        begin
             SanitizeSHBHeaders (Loader, Writer);
             SanitizeInterfaceBlock (Loader, Writer);
             SanitizeNameResolutionBlock (Loader, Writer);
        end
     else
         begin
              if Loader.HeaderComments.Count = 0 then
                 Writer.HeaderCommentList.Add (TraceWranglerHint);

              Writer.CopyPCAPngHeaderDetailsFromLoader (Loader);
         end;
end;

(* ------------------------------------------- *)
{ Frame Sanitization }
(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeIPv4PseudoHeader (var IPv4PseudoHeader : tIPv4PseudoHeader);

var Source : tIPv4Address;
    Dest   : tIPv4Address;

begin
     // The Pseudoheader is sanitized according to the IPv4 replacement settings. We cannot assume that every replacement is written
     // to the DB, so we'll use the same address values in the replacement of the actual layer.

     if priv_params.IPv4DoReplace then
        begin
             if priv_params.IPv4DoReplaceByIPList or priv_params.IPv4ReplaceBySubnetlist or priv_params.IPv4RandomizeIPAddress or priv_params.IPv4BlackMarkerIPAddress then
                begin
                     Source.ValueAsLong := IPv4PseudoHeader.Source;
                     Source.MaskBitCount := 32;
                     Dest.ValueAsLong := IPv4PseudoHeader.Destination;
                     Dest.MaskBitCount := 32;

                     Source := priv_replacementmgr.getReplacementForAddress (Source);
                     Dest := priv_replacementmgr.getReplacementForAddress (Dest);

                     IPv4PseudoHeader.Source := Source.ValueAsLong;
                     IPv4PseudoHeader.Destination := Dest.ValueAsLong;
                end;
        end;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeIPv6PseudoHeader (var IPv6PseudoHeader : tIPv6PseudoHeader);

var Source : tIPv6Address;
    Dest   : tIPv6Address;

begin
     // The Pseudoheader is sanitized according to the IPv4 replacement settings. We cannot assume that every replacement is written
     // to the DB, so we'll use the same address values in the replacement of the actual layer.

     if priv_params.IPv6DoReplace then
        begin
             Source.ValueAsWordStruct := IPv6PseudoHeader.Source;
             Source.MaskBitCount := 64;
             Dest.ValueAsWordStruct := IPv6PseudoHeader.Destination;
             Dest.MaskBitCount := 64;

             Source := priv_replacementmgr.getReplacementForAddress (Source);
             Dest := priv_replacementmgr.getReplacementForAddress (Dest);

             IPv6PseudoHeader.Source := Source.ValueAsWordStruct;
             IPv6PseudoHeader.Destination := Dest.ValueAsWordStruct;
        end;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetEthernetAssembly (Parser: tEthernetParser; ForceEtherType : word): tFrameChunk;

var assembly : tEthernetAssembler;
    help     : tMACAddress;

begin
     assembly := tEthernetAssembler.create;

     // Get values from parser and assign them

     assembly.CopyFromParser (Parser);

     // Modify what we know

     if priv_params.EthernetDoReplace then
        begin
             // The kind of replacement is decided inside the replacement manager

             assembly.SourceAddress := priv_replacementmgr.getReplacementForAddress (assembly.SourceAddress);
             assembly.DestinationAddress := priv_replacementmgr.getReplacementForAddress (assembly.DestinationAddress);
        end
     else
         if priv_params.EthernetDoOnlyDBBasedReplacement then
            begin
                 // This will only replace addresses that have a replacement entry stored to the DB
                 // which happens when IPv6 replacement is active

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.SourceAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.SourceAddress := help;

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.DestinationAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.DestinationAddress := help;
            end;

     // Set new ethertype in cases where we removed earlier VLAN layers

     if ForceEtherType > 0 then
        assembly.EtherTypeLength := ForceEtherType;

     // Create output

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.Get802dot11RadioTapAssembly (Parser: t802dot11RadioTapParser): tFrameChunk;

var assembly : t802dot11RadioTapAssembler;

begin
     assembly := t802dot11RadioTapAssembler.create;
     assembly.copyFromparser (Parser);
     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.Get802dot11HdrAssembler (Parser: t802dot11HeaderParser): tFrameChunk;

var assembly : t802dot11headerAssembler;
    help     : tMACAddress;

begin
     assembly := t802dot11headerAssembler.create (Parser.isOverCAPWAP);
     assembly.copyFromparser (Parser);

     if priv_params.EthernetDoReplace then
        begin
             // The kind of replacement is decided inside the replacement manager

             assembly.SourceAddress := priv_replacementmgr.getReplacementForAddress (assembly.SourceAddress);
             assembly.DestinationAddress := priv_replacementmgr.getReplacementForAddress (assembly.DestinationAddress);
             assembly.ReceiverAddress := priv_replacementmgr.getReplacementForAddress (assembly.ReceiverAddress);
             assembly.TransmitterAddress := priv_replacementmgr.getReplacementForAddress (assembly.TransmitterAddress);
        end
     else
         if priv_params.EthernetDoOnlyDBBasedReplacement then
            begin
                 // This will only replace addresses that have a replacement entry stored to the DB
                 // which happens when IPv6 replacement is active

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.SourceAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.SourceAddress := help;

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.DestinationAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.DestinationAddress := help;

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.ReceiverAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.ReceiverAddress := help;

                 help := priv_replacementmgr.getMACReplacementFromDB (assembly.TransmitterAddress);

                 if help <> VALUE_ZERO_MAC then
                    assembly.TransmitterAddress := help;
            end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.Get802dot11MgmtAssembler (Parser: t802dot11ManagementParser) : tFrameChunk;

var assembly    : t802dot11ManagementAssembler;
    i           : integer;


begin
     assembly := t802dot11ManagementAssembler.create (Parser.SubType);

     if priv_params.WiFiDoReplace then
        begin
             assembly.copyFromparser (Parser, priv_params.WiFiKeepUnknownTaggedParams);

             for i := 1 to length (assembly.TaggedParameters) do
                 begin
                      if assembly.TaggedParameters[i-1].ParamType = cSSIDParameterSet then
                         begin
                              assembly.TaggedParameters[i-1].SSID := priv_ReplacementMgr.getReplacementForString (assembly.TaggedParameters[i-1].SSID, qt802dot11);
                              assembly.TaggedParameters[i-1].ParamLen := length (assembly.TaggedParameters[i-1].SSID);
                         end;
                 end;
        end
     else
         assembly.copyFromparser (Parser, true);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetLLCAssembly (Parser: tLLCParser): tFrameChunk;

var assembly : tLLCAssembler;

begin
     assembly := tLLCAssembler.create;
     assembly.copyFromparser (Parser);
     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetLinuxCookedAssembly (Parser: tLinuxSLLParser): tFrameChunk;

var assembly : tLinuxSLLAssembler;

begin
     assembly := tLinuxSLLAssembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.EthernetDoReplace and (assembly.MACAddress <> VALUE_ZERO_MAC) then
        begin
             assembly.MACAddress := priv_replacementmgr.getReplacementForAddress (assembly.MACAddress);
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetVLANAssembly (Parser: t802dot1qParser; ForceEtherType : word): tFrameChunk;

var assembly : t802dot1qAssembler;

begin
     assembly := t802dot1qAssembler.create;

     // Get values from parser and assign them

     assembly.CopyFromParser (Parser);

     // Modify what we know

     if priv_params.VLANDoReplace then
        begin
             // The kind of replacement is decided inside the replacement manager

             assembly.VLANID := priv_replacementmgr.getReplacementForVLAN (assembly.VLANID);
        end;

     // Set new ethertype in cases where we removed earlier VLAN layers

     if ForceEtherType > 0 then
        assembly.EtherType := ForceEtherType;

     // Create output

     result := assembly.GetAssembly;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetARPAssembly (Parser: tARPParser): tFrameChunk;

var assembly : tARPAssembler;

begin
     assembly := tARPAssembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.EthernetDoReplace then
        begin
             assembly.SenderMAC := priv_replacementmgr.getReplacementForAddress (Parser.SenderMAC);
             assembly.TargetMAC := priv_replacementmgr.getReplacementForAddress (Parser.TargetMAC);
        end;

     if priv_params.IPv4DoReplace then
        begin
             assembly.SenderIP := priv_replacementmgr.getReplacementForAddress (Parser.SenderIP);
             assembly.TargetIP := priv_replacementmgr.getReplacementForAddress (Parser.TargetIP);
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetMPLSAssembly (Parser: tMPLSParser): tFrameChunk;

var assembly : tMPLSAssembler;
    help     : longword;

begin
     assembly := tMPLSAssembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.TunnelDoReplace then
        begin
             if priv_params.MPLSRandomizeLabel then
                assembly.MPLSLabel := priv_replacementmgr.getReplacementForMPLSLabel (assembly.MPLSLabel);
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetIPv4Assembly (Parser: tIPv4Parser; PseudoHeader : tIPv4PseudoHeader; PortPair : tPortPair; PayloadLen : integer; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tIPv4Assembler;
    Source     : tIPv4Address;
    SockPair   : tIPv4SocketPair;
    Dest       : tIPv4Address;
    Option     : tIPv4Option;
    Socket     : tIPv4Socket;
    SourceDone : boolean;
    DestDone   : boolean;
    complete   : boolean;
    i,j        : integer;

begin
     assembly := tIPv4Assembler.create (PayloadLen);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     // Check if we need to do a per socket pair replacement, and if so, flag the replacement as complete

     if priv_params.TCPReplaceSocketsByList and (length (priv_params.TCPSocketPortList) > 0) then
        begin
             Socket.IPAddress := Parser.Source;
             Socket.Port := PortPair.Value1;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Source := Socket.IPAddress;

                     sourcedone := true;
                end
             else
                 sourcedone := false;

             Socket.IPAddress := Parser.Destination;
             Socket.Port := PortPair.Value2;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Destination := Socket.IPAddress;

                     destdone := true;
                end
             else
                 destdone := false;
        end;

     if priv_params.IPv4DoReplace then
        begin
             // Addresses - they have been replaced in the pseudo header already, so we use those

             if priv_params.IPv4DoReplaceByIPList or priv_params.IPv4ReplaceBySubnetlist or priv_params.IPv4RandomizeIPAddress or priv_params.IPv4BlackMarkerIPAddress then
                begin
                     // Replace source and/or destination if not yet replaced by socket

                     if not sourcedone then
                        begin
                             Source.ValueAsLong := PseudoHeader.Source;
                             assembly.Source := Source;
                        end;

                     if not destdone then
                        begin
                             Dest.ValueAsLong := PseudoHeader.Destination;
                             assembly.Destination := Dest;
                        end;
                end;

             // IP Options

             if assembly.CurrentRoute.ValueAsLong <> 0 then
                assembly.CurrentRoute := priv_replacementmgr.getReplacementForAddress (assembly.CurrentRoute);

                for i := 1 to length (assembly.OptionList) do
                    begin
                         Option := assembly.OptionList[i-1];

                         for j := 1 to length (Option.RouteRecords) do
                             Option.RouteRecords[j-1] :=  priv_replacementmgr.getReplacementForAddress (Option.RouteRecords[j-1]);


                         for j := 1 to length (Option.TimestampRecords) do
                             Option.TimestampRecords[j-1].address :=  priv_replacementmgr.getReplacementForAddress (Option.TimestampRecords[j-1].address);
                    end;

             // IP Identification

             if priv_params.IPv4RandomizeIPID then
                begin
                     if Parser.IsFragment then
                        begin
                             // Get IP ID for the socket pair

                             SockPair.Source.IPAddress := Parser.Source;
                             SockPair.Destination.IPAddress := Parser.Destination;

                             assembly.Identification := priv_replacementmgr.getReplacementForIPID (Parser.Identification, SockPair);
                        end
                     else
                         assembly.Identification := random (65535);
                end;

             // CRC

             case priv_params.IPv4RecalculateCRCMode of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // If the checksum was 0, we force it to 0 again

                                           if Parser.Checksum = 0 then
                                              assembly.Checksum := 0
                                           else
                                               assembly.ForceBadCRC := true;
                                      end;
                              end;
              rcmFixZeroCRC : begin
                                   if not (Parser.Checksum = 0) then
                                      assembly.Checksum := random (65535);
                              end;
             end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetIPv4Assembly (Parser: tIPv4Parser; PortPair : tPortPair; PayloadLen : integer; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tIPv4Assembler;
    SockPair   : tIPv4SocketPair;
    Option     : tIPv4Option;
    Socket     : tIPv4Socket;
    SourceDone : boolean;
    DestDone   : boolean;
    i,j        : integer;

begin
     assembly := tIPv4Assembler.create (PayloadLen);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     // Check if we need to do a per socket pair replacement, and if so, flag the replacement as complete

     if priv_params.TCPReplaceSocketsByList and (length (priv_params.TCPSocketPortList) > 0) then
        begin
             Socket.IPAddress := Parser.Source;
             Socket.Port := PortPair.Value1;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Source := Socket.IPAddress;

                     sourcedone := true;
                end
             else
                 sourcedone := false;

             Socket.IPAddress := Parser.Destination;
             Socket.Port := PortPair.Value2;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Destination := Socket.IPAddress;

                     destdone := true;
                end
             else
                 destdone := false;
        end;

     if priv_params.IPv4DoReplace then
        begin
             // Addresses - replacement manager handles the different variants anyway

             if priv_params.IPv4DoReplaceByIPList or priv_params.IPv4ReplaceBySubnetlist or priv_params.IPv4RandomizeIPAddress or priv_params.IPv4BlackMarkerIPAddress then
                begin
                     if not sourcedone then
                        assembly.Source := priv_replacementmgr.getReplacementForAddress (Parser.Source);

                     if not destdone then
                        assembly.Destination := priv_replacementmgr.getReplacementForAddress (Parser.Destination);
                end;

             // IP Options

             if assembly.CurrentRoute.ValueAsLong <> 0 then
                assembly.CurrentRoute := priv_replacementmgr.getReplacementForAddress (assembly.CurrentRoute);

                for i := 1 to length (assembly.OptionList) do
                    begin
                         Option := assembly.OptionList[i-1];

                         for j := 1 to length (Option.RouteRecords) do
                             Option.RouteRecords[j-1] :=  priv_replacementmgr.getReplacementForAddress (Option.RouteRecords[j-1]);


                         for j := 1 to length (Option.TimestampRecords) do
                             Option.TimestampRecords[j-1].address :=  priv_replacementmgr.getReplacementForAddress (Option.TimestampRecords[j-1].address);
                    end;

             // IP Identification

             if priv_params.IPv4RandomizeIPID then
                begin
                     if Parser.IsFragment then
                        begin
                             SockPair.Source.IPAddress := Parser.Source;
                             SockPair.Destination.IPAddress := Parser.Destination;

                             assembly.Identification := priv_replacementmgr.getReplacementForIPID (Parser.Identification, SockPair);
                        end
                     else
                         assembly.Identification := random (65535);
                end;

              // CRC

              case priv_params.IPv4RecalculateCRCMode of
               rcmKeepCRC    : begin
                                    assembly.Checksum := Parser.Checksum;
                               end;
               rcmKeepBadCRC : begin
                                    if not CRCWasOkay then
                                       begin
                                            // If the checksum was 0, we force it to 0 again

                                            if Parser.Checksum = 0 then
                                               assembly.Checksum := 0
                                            else
                                                assembly.ForceBadCRC := true;
                                       end;
                               end;
               rcmFixZeroCRC : begin
                                    if not (Parser.Checksum = 0) then
                                       assembly.Checksum := random (65535);
                               end;
              end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetIPv6Assembly (Parser: tIPv6Parser; PseudoHeader : tIPv6PseudoHeader; PortPair : tPortPair): tFrameChunk;

var assembly   : tIPv6Assembler;
    Source     : tIPv6Address;
    Dest       : tIPv6Address;
    Socket     : tIPv6Socket;
    SourceDone : boolean;
    DestDone   : boolean;


begin
     assembly := tIPv6Assembler.create;
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     if priv_params.TCPReplaceSocketsByList and (length (priv_params.TCPSocketPortList) > 0) then
        begin
             Socket.IPAddress := Parser.Source;
             Socket.Port := PortPair.Value1;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Source := Socket.IPAddress;

                     sourcedone := true;
                end
             else
                 sourcedone := false;

             Socket.IPAddress := Parser.Destination;
             Socket.Port := PortPair.Value2;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Destination := Socket.IPAddress;

                     destdone := true;
                end
             else
                 destdone := false;
        end;

     if priv_params.IPv6DoReplace then
        begin
             if not sourcedone then
                begin
                     Source.ValueAsWordStruct := PseudoHeader.Source;
                     assembly.Source := Source;
                end;

             if not destdone then
                begin
                     Dest.ValueAsWordStruct := PseudoHeader.Destination;
                     assembly.Destination := Dest;
                end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetIPv6Assembly (Parser: tIPv6Parser; PortPair : tPortPair): tFrameChunk;

var assembly   : tIPv6Assembler;
    Socket     : tIPv6Socket;
    SourceDone : boolean;
    DestDone   : boolean;

begin
     assembly := tIPv6Assembler.create;
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     if priv_params.TCPReplaceSocketsByList and (length (priv_params.TCPSocketPortList) > 0) then
        begin
             Socket.IPAddress := Parser.Source;
             Socket.Port := PortPair.Value1;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Source := Socket.IPAddress;

                     sourcedone := true;
                end
             else
                 sourcedone := false;

             Socket.IPAddress := Parser.Destination;
             Socket.Port := PortPair.Value2;

             Socket := priv_replacementmgr.getReplacementForSocket (protTCP, Socket);

             // Got a replacement socket, so need to replace the IP accordingly

             if not Socket.isUndefined then
                begin
                     assembly.Destination := Socket.IPAddress;

                     destdone := true;
                end
             else
                 destdone := false;
        end;

     if priv_params.IPv6DoReplace then
        begin
             if not sourcedone then
                assembly.Source := priv_replacementmgr.getReplacementForAddress (Parser.Source);

             if not destdone then
                assembly.Destination := priv_replacementmgr.getReplacementForAddress (Parser.Destination);
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetIPv6FragmentationAssembly (Parser: tIPv6FragmentationParser): tFrameChunk;

var assembly : tIPv6FragHdrAssembler;

begin
     assembly := tIPv6FragHdrAssembler.create;
     assembly.CopyFromParser (Parser);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)
{
function tTaskProcessorAnonymize.GetICMPv4Assembly (Parser: tICMPv4Parser; CRCWasOkay : boolean): tFrameChunk;

var assembly : tICMPv4Assembler;
    i        : integer;

begin
     assembly := tICMPv4Assembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.IPv4DoReplace then
        begin
             case assembly.ICMPType of
              5 : begin
                       assembly.RedirectGwy := priv_replacementmgr.getReplacementForAddress(assembly.RedirectGwy);
                  end;
             end;
        end;

     if priv_params.ICMPv4DoReplace then
        begin
             case assembly.ICMPType of
              0, 8 : begin
                          if priv_params.ICMPv4RemovePingPayload then
                             begin
                                  assembly.clearPayload;         // TODO: Check if length is adjusted correctly
                             end;
                     end;
             end;

             // CRC

             case priv_params.ICMPv4RecalculateCRCMode of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // If the checksum was 0, we force it to 0 again

                                           if Parser.Checksum = 0 then
                                              assembly.Checksum := 0
                                           else
                                               assembly.ForceBadCRC := true;
                                      end;
                              end;
             end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;
}
(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetICMPv4Assembly (Parser: tICMPv4Parser; QuotedPayload: tFrameChunk; CRCWasOkay: boolean): tFrameChunk;

var assembly : tICMPv4Assembler;
    i        : integer;

begin
     assembly := tICMPv4Assembler.create (QuotedPayload);
     assembly.CopyFromParser (Parser);

     if priv_params.IPv4DoReplace then
        begin
             case assembly.ICMPType of
              5 : begin
                       assembly.RedirectGwy := priv_replacementmgr.getReplacementForAddress (assembly.RedirectGwy);
                  end;
             end;
        end;

     if priv_params.ICMPv4DoReplace then
        begin
             case assembly.ICMPType of
              0, 8 : begin
                          if priv_params.ICMPv4RemovePingPayload then
                             begin
                                  assembly.clearPayload;

                                  // TODO: need to find a way to set the frame length correctly.
                                  // Stripping the payload means the assembly.len is shorter, leading to a shortened wiresize instead of just shortening capturesize
                             end;
                     end;
             end;

             // CRC

             case priv_params.ICMPv4RecalculateCRCMode of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // If the checksum was 0, we force it to 0 again

                                           if Parser.Checksum = 0 then
                                              assembly.Checksum := 0
                                           else
                                               assembly.ForceBadCRC := true;
                                      end;
                              end;
             end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetICMPv6Assembly (Parser: tICMPv6Parser; IPv6PseudoHeader : tIPv6PseudoHeader; CRCWasOkay: boolean): tFrameChunk;

var assembly : tICMPv6Assembler;
    i,j      : integer;

begin
     assembly := tICMPv6Assembler.create (IPv6PseudoHeader);
     assembly.CopyFromParser (Parser);

     if priv_params.IPv6DoReplace then
        begin
             case assembly.ICMPType of
              135 : begin
                         // Neighbor solicitation

                         assembly.NDTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.NDTargetAddress);

                         assembly.SourceLinkLayerAddress := priv_replacementmgr.getReplacementForAddress (assembly.SourceLinkLayerAddress);
                    end;
              136 : begin
                         // Neighbor advertisement

                         assembly.NDTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.NDTargetAddress);

                         assembly.TargetLinkLayerAddress := priv_replacementmgr.getReplacementForAddress (assembly.TargetLinkLayerAddress);
                    end;
              137 : begin
                         // Redirect

                         assembly.RedirectTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.RedirectTargetAddress);
                         assembly.RedirectDestinationAddress := priv_replacementmgr.getReplacementForAddress (assembly.RedirectDestinationAddress);
                    end;
             end;
        end;

     if priv_params.ICMPv6DoReplace then
        begin
             case assembly.ICMPType of
              128,
              129 : begin
                         if priv_params.ICMPv6RemovePingPayload then
                            assembly.clearPayload;                      // TODO: Check if length is adjusted correctly
                    end;
             end;

             // CRC

             case priv_params.ICMPv6RecalculateCRCMode of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // If the checksum was 0, we force it to 0 again

                                           if Parser.Checksum = 0 then
                                              assembly.Checksum := 0
                                           else
                                               assembly.ForceBadCRC := true;
                                      end;
                              end;
             end;
        end;

     // Clear options

     for i := 1 to length (assembly.OptionList) do
         begin
              // MAC adresses

//            if priv_params.EthernetDoReplace and (assembly.OptionList[i-1].MACAddress <> VALUE_ZERO_MAC) then         disabled because Ethernet Replacement does not have a word here
              if (assembly.OptionList[i-1].MACAddress <> VALUE_ZERO_MAC) then
                 begin
                      // Only replace MAC addresses that were not already sanitized above
                      // Type 1 and 2 are Source/Target Link layer addresses, used by ICMP types 135 & 136

                      if assembly.OptionList[i-1].option_type > 2 then
                         assembly.OptionList[i-1].MACAddress := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].MACAddress);
                 end;

              // IPv6 addresses

              if priv_params.IPv6DoReplace then
                 begin
                      // Prefixes

                      if not (assembly.OptionList[i-1].PrefixAddress.IsUnspecifiedAddress) then
                         assembly.OptionList[i-1].PrefixAddress := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].PrefixAddress);

                      // DNS Server list

                      if assembly.OptionList[i-1].option_type = 25 then
                         for j := 1 to length (assembly.OptionList[i-1].DNSServerList) do
                             assembly.OptionList[i-1].DNSServerList[j-1] := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].DNSServerList[j-1]);
                 end;

              // FQDNs - this needs more attention when we know how to deal with DNS

              if priv_params.ICMPv6RemoveDNSFQDNsFromOptions then
                 begin
                      if assembly.OptionList[i-1].option_type = 31 then
                         begin
                              // For now, we clear all FQDNs

                              for j := 1 to length (assembly.OptionList[i-1].DNSDomainNameList) do
                                  assembly.OptionList[i-1].DNSDomainNameList[j-1] := priv_replacementmgr.getReplacementForFQDN (assembly.OptionList[i-1].DNSDomainNameList[j-1]);
                         end;
                 end;
         end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetICMPv6Assembly (Parser: tICMPv6Parser; IPv6PseudoHeader: tIPv6PseudoHeader; QuotedPayload: tFrameChunk; CRCWasOkay: boolean): tFrameChunk;

var assembly : tICMPv6Assembler;
    i,j      : integer;

begin
     assembly := tICMPv6Assembler.create (IPv6PseudoHeader, QuotedPayload);
     assembly.CopyFromParser (Parser);

     if priv_params.IPv6DoReplace then
        begin
             case assembly.ICMPType of
              135 : begin
                         // Neighbor solicitation

                         assembly.NDTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.NDTargetAddress);

                         assembly.SourceLinkLayerAddress := priv_replacementmgr.getReplacementForAddress (assembly.SourceLinkLayerAddress);
                    end;
              136 : begin
                         // Neighbor advertisement

                         assembly.NDTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.NDTargetAddress);

                         assembly.TargetLinkLayerAddress := priv_replacementmgr.getReplacementForAddress (assembly.TargetLinkLayerAddress);
                    end;
              137 : begin
                         // Redirect

                         assembly.RedirectTargetAddress := priv_replacementmgr.getReplacementForAddress (assembly.RedirectTargetAddress);
                         assembly.RedirectDestinationAddress := priv_replacementmgr.getReplacementForAddress (assembly.RedirectDestinationAddress);
                    end;
             end;
        end;

     if priv_params.ICMPv6DoReplace then
        begin
             case assembly.ICMPType of
              128,
              129 : begin
                         if priv_params.ICMPv6RemovePingPayload then
                            assembly.clearPayload;                      // TODO: Check if length is adjusted correctly
                    end;
             end;

             // CRC

             case priv_params.ICMPv6RecalculateCRCMode of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // If the checksum was 0, we force it to 0 again

                                           if Parser.Checksum = 0 then
                                              assembly.Checksum := 0
                                           else
                                               assembly.ForceBadCRC := true;
                                      end;
                              end;
             end;
        end;

     // Clear options

     for i := 1 to length (assembly.OptionList) do
         begin
              // MAC adresses

//            if priv_params.EthernetDoReplace and (assembly.OptionList[i-1].MACAddress <> VALUE_ZERO_MAC) then         disabled because Ethernet Replacement does not have a word here
              if (assembly.OptionList[i-1].MACAddress <> VALUE_ZERO_MAC) then
                 begin
                      // Only replace MAC addresses that were not already sanitized above
                      // Type 1 and 2 are Source/Target Link layer addresses, used by ICMP types 135 & 136

                      if assembly.OptionList[i-1].option_type > 2 then
                         assembly.OptionList[i-1].MACAddress := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].MACAddress);
                 end;

              // IPv6 addresses

              if priv_params.IPv6DoReplace then
                 begin
                      // Prefixes

                      if not (assembly.OptionList[i-1].PrefixAddress.IsUnspecifiedAddress) then
                         assembly.OptionList[i-1].PrefixAddress := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].PrefixAddress);

                      // DNS Server list

                      if assembly.OptionList[i-1].option_type = 25 then
                         for j := 1 to length (assembly.OptionList[i-1].DNSServerList) do
                             assembly.OptionList[i-1].DNSServerList[j-1] := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].DNSServerList[j-1]);
                 end;

              // FQDNs - this needs more attention when we know how to deal with DNS

              if priv_params.ICMPv6RemoveDNSFQDNsFromOptions then
                 begin
                      if assembly.OptionList[i-1].option_type = 31 then
                         begin
                              // For now, we clear all FQDNs

                              for j := 1 to length (assembly.OptionList[i-1].DNSDomainNameList) do
                                  assembly.OptionList[i-1].DNSDomainNameList[j-1] := priv_replacementmgr.getReplacementForFQDN (assembly.OptionList[i-1].DNSDomainNameList[j-1]);
                         end;
                 end;
         end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetGREAssembly (Parser: tGREParser): tFrameChunk;

var assembly : tGREAssembler;

begin
     assembly := tGREAssembler.create;
     assembly.CopyFromParser (Parser);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetGTPv1Assembly (Parser: tGTPv1Parser): tFrameChunk;

var assembly : tGTPv1Assembler;

begin
     assembly := tGTPv1Assembler.create;
     assembly.CopyFromParser (Parser);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetVXLANAssembly (Parser : tVXLANParser) : tFrameChunk;

var assembly : tVXLANAssembler;

begin
     assembly := tVXLANAssembler.create;
     assembly.CopyFromParser (Parser);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getAyiyaAssembly (Parser: tAYIYAParser): tFrameChunk;

var assembly : tAyiyaAssembler;
    help     : tBufferBytes;
    i        : integer;

begin
     assembly := tAyiyaAssembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.TunnelDoReplace then
        begin
             if priv_params.TunnelAYIYASanitizeIdentity = aisZero then
                begin
                     // Replace Identity with all zeroes

                     if Parser.IDLength > 0 then
                        begin
                             setlength (help, Parser.IDLength);

                             for i := 1 to Parser.IDLength do
                                 help[i-1] := 0;

                             assembly.Identity := help;
                        end;
                end;

             if priv_params.TunnelAYIYASanitizeSignature = assZero then
                begin
                     // Replace Signature with all zeroes

                     if Parser.SignatureLength > 0 then
                        begin
                             setlength (help, Parser.SignatureLength);

                             for i := 1 to Parser.SignatureLength do
                                 help[i-1] := 0;

                             assembly.Signature := help;
                        end;
                end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetTCPAssembly (Parser: tTCPParser; OriginalIPPair : tIPv4AddressPair; IPv4PseudoHeader : tIPv4PseudoHeader; TCPPayload : tFrameChunk; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tTCPAssembler;
    socket     : tIPv4Socket;
    sourcedone : boolean;
    destdone   : boolean;
    i          : integer;


begin
     assembly := tTCPAssembler.create (IPv4PseudoHeader, TCPPayload);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     if priv_params.IPv4DoReplace then
        begin
             // Handle MultiPath Options

             for i := 1 to length (assembly.MultiPathOptionList) do
                 begin
                      if assembly.MultiPathOptionList[i-1].IPv4Address.ValueAsLong <> 0 then
                         assembly.MultiPathOptionList[i-1].IPv4Address := priv_replacementmgr.getReplacementForAddress (assembly.MultiPathOptionList[i-1].IPv4Address);

                      if not assembly.MultiPathOptionList[i-1].IPv6Address.IsUnspecifiedAddress then
                         assembly.MultiPathOptionList[i-1].IPv6Address := priv_replacementmgr.getReplacementForAddress (assembly.MultiPathOptionList[i-1].IPv6Address);
                 end;
        end;

     if priv_params.TCPDoReplace then
        begin
             // Ports

             sourcedone := false;
             destdone := false;

             // Start with socket list, if available

             if priv_params.TCPReplaceSocketsByList then
                begin
                     socket.Port := Parser.SourcePort;
                     socket.IPAddress := OriginalIPPair.Source;
                     socket := priv_replacementmgr.getReplacementForSocket (protTCP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.SourcePort := socket.Port;
                             sourcedone := true;
                        end;

                     socket.Port := Parser.DestinationPort;
                     socket.IPAddress := OriginalIPPair.Destination;
                     socket := priv_replacementmgr.getReplacementForSocket (protTCP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.DestinationPort := socket.Port;
                             destdone := true;
                        end;
                end;

             // Now simple port replacements

             if not sourcedone then
                assembly.SourcePort := priv_replacementmgr.getReplacementForPort (protTCP, Parser.SourcePort);

             if not destdone then
                assembly.DestinationPort := priv_replacementmgr.getReplacementForPort (protTCP, Parser.DestinationPort);

             // CRC

             case priv_params.TCPRecalculateCRC of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           assembly.ForceBadCRC := true;
                                           assembly.DoNotRecalculateCRC := false;
                                      end
                                   else
                                       begin
                                            assembly.ForceBadCRC := false;
                                            assembly.DoNotRecalculateCRC := false;
                                       end;
                              end;
              rcmFixAll     : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
             end;
        end;
//   else
//       assembly.DoNotRecalculateCRC := true;             This is a problem if the assembler cannot reproduce the exact TCP header byte by byte, so lets recalculate instead

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetTCPAssembly (Parser: tTCPParser; OriginalIPPair : tIPv6AddressPair; IPv6PseudoHeader: tIPv6PseudoHeader; TCPPayload: tFrameChunk; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tTCPAssembler;
    socket     : tIPv6Socket;
    mpoption   : tTCPMultiPathOption;
    i          : integer;
    sourcedone : boolean;
    destdone   : boolean;

begin
     assembly := tTCPAssembler.create (IPv6PseudoHeader, TCPPayload);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     // Handle Multipath Options

     if priv_params.IPv4DoReplace then
        begin
             // Handle MultiPath Options

             for i := 1 to length (assembly.MultiPathOptionList) do
                 begin
                      mpoption := assembly.MultiPathOptionList[i-1];

                      if mpoption.IPv4Address.ValueAsLong <> 0 then
                         mpoption.IPv4Address := priv_replacementmgr.getReplacementForAddress (mpoption.IPv4Address);

                      if not mpoption.IPv6Address.IsUnspecifiedAddress then
                         mpoption.IPv6Address := priv_replacementmgr.getReplacementForAddress (mpoption.IPv6Address);
                 end;
        end;

     if priv_params.TCPDoReplace then
        begin
             // Ports

             sourcedone := false;
             destdone := false;

             // Start with socket list, if available

             if priv_params.TCPReplaceSocketsByList then
                begin
                     socket.Port := Parser.SourcePort;
                     socket.IPAddress := OriginalIPPair.Source;
                     socket := priv_replacementmgr.getReplacementForSocket (protTCP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.SourcePort := socket.Port;
                             sourcedone := true;
                        end;

                     socket.Port := Parser.DestinationPort;
                     socket.IPAddress := OriginalIPPair.Destination;
                     socket := priv_replacementmgr.getReplacementForSocket (protTCP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.DestinationPort := socket.Port;
                             destdone := true;
                        end;
                end;

             if not sourcedone then
                assembly.SourcePort := priv_replacementmgr.getReplacementForPort (protTCP, Parser.SourcePort);

             if not destdone then
                assembly.DestinationPort := priv_replacementmgr.getReplacementForPort (protTCP, Parser.DestinationPort);

             // CRC

             case priv_params.TCPRecalculateCRC of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           assembly.ForceBadCRC := true;
                                           assembly.DoNotRecalculateCRC := false;
                                      end
                                   else
                                       begin
                                            assembly.ForceBadCRC := false;
                                            assembly.DoNotRecalculateCRC := false;
                                       end;
                              end;
              rcmFixAll     : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
             end;
        end;
//   else
//       assembly.DoNotRecalculateCRC := true;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetUDPAssembly (Parser: tUDPParser; OriginalIPPair : tIPv4AddressPair; IPv4PseudoHeader: tIPv4PseudoHeader; UDPPayload: tFrameChunk; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tUDPAssembler;
    socket     : tIPv4Socket;
    sourcedone : boolean;
    destdone   : boolean;

begin
//   SiMain.EnterMethod ('tTaskProcessorAnonymize.GetUDPAssembly');
//   SiMain.LogVerbose (UDPPayload.getAsHEXString);

     assembly := tUDPAssembler.create (IPv4PseudoHeader, UDPPayload);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     if priv_params.UDPDoReplace then
        begin
             // Ports

             sourcedone := false;
             destdone := false;

             // Start with socket list, if available

             if priv_params.UDPReplaceSocketsByList then
                begin
                     socket.Port := Parser.SourcePort;
                     socket.IPAddress := OriginalIPPair.Source;
                     socket := priv_replacementmgr.getReplacementForSocket (protUDP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.SourcePort := socket.Port;
                             sourcedone := true;
                        end;

                     socket.Port := Parser.DestinationPort;
                     socket.IPAddress := OriginalIPPair.Destination;
                     socket := priv_replacementmgr.getReplacementForSocket (protUDP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.DestinationPort := socket.Port;
                             destdone := true;
                        end;
                end;

             if not sourcedone then
                assembly.SourcePort := priv_replacementmgr.getReplacementForPort (protUDP, Parser.SourcePort);

             if not destdone then
                assembly.DestinationPort := priv_replacementmgr.getReplacementForPort (protUDP, Parser.DestinationPort);

             // CRC

             case priv_params.UDPRecalculateCRC of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // Keep Zeros

                                           if Parser.Checksum = 0 then
                                              begin
                                                   assembly.Checksum := 0;
                                                   assembly.DoNotRecalculateCRC := true;
                                              end
                                           else
                                               begin
                                                    assembly.DoNotRecalculateCRC := false;
                                                    assembly.ForceBadCRC := true;
                                               end;
                                      end
                                   else
                                       assembly.DoNotRecalculateCRC := false;
                              end;
              rcmFixZeroCRC : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := true;
                              end;
              rcmFixAll     : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := true;
                              end;
             end;
        end;
//   else
//       assembly.DoNotRecalcuateCRC := true;

     result := assembly.GetAssembly;
     assembly.Free;

//   SiMain.LeaveMethod ('tTaskProcessorAnonymize.GetUDPAssembly');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetUDPAssembly (Parser: tUDPParser; OriginalIPPair : tIPv6AddressPair; IPv6PseudoHeader: tIPv6PseudoHeader; UDPPayload: tFrameChunk; CRCWasOkay : boolean): tFrameChunk;

var assembly   : tUDPAssembler;
    socket     : tIPv6Socket;
    sourcedone : boolean;
    destdone   : boolean;

begin
     assembly := tUDPAssembler.create (IPv6PseudoHeader, UDPPayload);
     assembly.CopyFromParser (Parser);

     SourceDone := false;
     DestDone := false;

     if priv_params.UDPDoReplace then
        begin
             // Ports

             sourcedone := false;
             destdone := false;

             // Start with socket list, if available

             if priv_params.UDPReplaceSocketsByList then
                begin
                     socket.Port := Parser.SourcePort;
                     socket.IPAddress := OriginalIPPair.Source;
                     socket := priv_replacementmgr.getReplacementForSocket (protUDP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.SourcePort := socket.Port;
                             sourcedone := true;
                        end;

                     socket.Port := Parser.DestinationPort;
                     socket.IPAddress := OriginalIPPair.Destination;
                     socket := priv_replacementmgr.getReplacementForSocket (protUDP, socket);

                     if not socket.isUndefined then
                        begin
                             assembly.DestinationPort := socket.Port;
                             destdone := true;
                        end;
                end;

             if not sourcedone then
                assembly.SourcePort := priv_replacementmgr.getReplacementForPort (protUDP, Parser.SourcePort);

             if not destdone then
                assembly.DestinationPort := priv_replacementmgr.getReplacementForPort (protUDP, Parser.DestinationPort);

             // CRC

             case priv_params.UDPRecalculateCRC of
              rcmKeepCRC    : begin
                                   assembly.Checksum := Parser.Checksum;
                                   assembly.DoNotRecalculateCRC := false;
                              end;
              rcmKeepBadCRC : begin
                                   if not CRCWasOkay then
                                      begin
                                           // Keep Zeros

                                           if Parser.Checksum = 0 then
                                              begin
                                                   assembly.Checksum := 0;
                                                   assembly.DoNotRecalculateCRC := true;
                                              end
                                           else
                                               begin
                                                    assembly.DoNotRecalculateCRC := false;
                                                    assembly.ForceBadCRC := true;
                                               end;
                                      end
                                   else
                                       assembly.DoNotRecalculateCRC := false;
                              end;
              rcmFixZeroCRC : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := true;
                              end;
              rcmFixAll     : begin
                                   assembly.ForceBadCRC := false;
                                   assembly.DoNotRecalculateCRC := true;
                              end;
             end;
        end;
//   else
//       assembly.DoNotRecalcuateCRC := true;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getDHCPv4Assembly (Parser: tDHCPv4Parser): tFrameChunk;

var assembly : tDHCPv4Assembler;
    i,j      : integer;

begin
     assembly := tDHCPv4Assembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.EthernetDoReplace then
        begin
             assembly.ClientHWAddress := priv_replacementmgr.getReplacementForAddress(assembly.ClientHWAddress);
        end;

     if priv_params.IPv4DoReplace then
        begin
             assembly.ClientIPAddress := priv_replacementmgr.getReplacementForAddress(assembly.ClientIPAddress);
             assembly.YourClientIPAddress := priv_replacementmgr.getReplacementForAddress(assembly.YourClientIPAddress);
             assembly.ServerIPAddress := priv_replacementmgr.getReplacementForAddress(assembly.ServerIPAddress);
             assembly.RelayIPAddress := priv_replacementmgr.getReplacementForAddress(assembly.RelayIPAddress);
        end;

     for i := 1 to length (assembly.OptionList) do
         begin
              // Replace MAC option values

              if priv_params.EthernetDoReplace and (assembly.OptionList[i-1].MACAddress <> VALUE_ZERO_MAC) then
                 assembly.OptionList[i-1].MACAddress := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].MACAddress);

              // Replace IPv4 option values

              if priv_params.IPv4DoReplace and (assembly.OptionList[i-1].addresscount > 0) then
                 begin
                      for j := 1 to assembly.OptionList[i-1].addresscount do
                          assembly.OptionList[i-1].addresslist[j] := priv_replacementmgr.getReplacementForAddress (assembly.OptionList[i-1].addresslist[j]);
                 end;

              // Replace DHCPv4 stuff

              if priv_params.DHCPv4DoReplace then
                 begin
                      // Handle FQDNs

                      if priv_params.DHCPv4ReplaceHostName then
                         begin
                              for j := 1 to assembly.OptionList[i-1].FQDNCount do
                                  assembly.OptionList[i-1].FQDNList[j] := priv_replacementmgr.getReplacementForFQDN(assembly.OptionList[i-1].FQDNList[j]);
                         end;
                 end;
         end;

     if priv_params.DHCPv4DoReplace then
        begin
             // Clear the Server and bootfile string areas

             assembly.ServerHostName := '';
             assembly.BootFileName := '';
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetNetFlowAssembly (Parser: tNetFlowParser): tFrameChunk;

var assembly : tNetFlowAssembler;
//  help     : tIPv4Socket;
    i        : integer;

begin
     assembly := tNetFlowAssembler.create;
     assembly.CopyFromParser (Parser);

     // Replace addresses

     for i := 1 to length (assembly.NetFlowV5Records) do
         begin
              assembly.NetFlowV5Records[i-1].SourceAddress := priv_replacementmgr.getReplacementForAddress (assembly.NetFlowV5Records[i-1].SourceAddress);
              assembly.NetFlowV5Records[i-1].DestinationAddress := priv_replacementmgr.getReplacementForAddress (assembly.NetFlowV5Records[i-1].DestinationAddress);
              assembly.NetFlowV5Records[i-1].NextHop := priv_replacementmgr.getReplacementForAddress (assembly.NetFlowV5Records[i-1].NextHop);
{
              help.IPAddress := assembly.NetFlowV5Records[i-1].SourceAddress;
              help.Port := assembly.NetFlowV5Records[i-1].SourcePort;

              assembly.NetFlowV5Records[i-1].SourcePort := priv_replacementmgr.getReplacementForPort (assembly.NetFlowV5Records[i-1].IPProtocolType, help);

              help.IPAddress := assembly.NetFlowV5Records[i-1].DestinationAddress;
              help.Port := assembly.NetFlowV5Records[i-1].DestinationPort;

              assembly.NetFlowV5Records[i-1].DestinationPort := priv_replacementmgr.getReplacementForPort (assembly.NetFlowV5Records[i-1].IPProtocolType, help);
}
              assembly.NetFlowV5Records[i-1].SourcePort := priv_replacementmgr.getReplacementForPort (assembly.NetFlowV5Records[i-1].IPProtocolType, assembly.NetFlowV5Records[i-1].SourcePort);
              assembly.NetFlowV5Records[i-1].DestinationPort := priv_replacementmgr.getReplacementForPort (assembly.NetFlowV5Records[i-1].IPProtocolType, assembly.NetFlowV5Records[i-1].DestinationPort);
         end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetHSRPAssembly (Parser: tHSRPParser): tFrameChunk;

var assembly : tHSRPAssembler;
    help     : tIPv4Socket;

begin
     assembly := tHSRPAssembler.create;
     assembly.CopyFromParser (Parser);

     // Replace vIP

     if not (assembly.OpCode = hoAdvertise) then
        assembly.VirtualIP := priv_replacementmgr.getReplacementForAddress (assembly.VirtualIP);

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

procedure tTaskProcessorAnonymize.SanitizeRTPSParameter (var Parameter: tRTPSSerializedDataEntry);

var help        : longword;
    AddrType    : longword;
    IPv4Address : tIPv4Address;
//  IPv4Socket  : tIPv4Socket;
    Port        : word;
    test        : string;

begin
     case Parameter.DataType of
      RTPS_PID_UNICAST_LOCATOR,
      RTPS_PID_MULTICAST_LOCATOR,
      RTPS_PID_DEFAULT_UNICAST_LOCATOR,
      RTPS_PID_METATRAFFIC_UNICAST_LOCATOR,
      RTPS_PID_METATRAFFIC_MULTICAST_LOCATOR,
      RTPS_PID_DEFAULT_MULTICAST_LOCATOR     : begin
                                                    // Check if it's IP (LOCATOR_KIND_UDPv4 0x00000001)

                                                    help := ExtractLongWordFromByteArray (Parameter.Data, 0);
                                                    AddrType := Swap32 (help);

                                                    if (AddrType = 1) and priv_params.IPv4DoReplace then
                                                       begin
                                                            help := ExtractLongWordFromByteArray (Parameter.Data, 4);
                                                            Port := swap32 (help);

                                                            help := ExtractLongWordFromByteArray (Parameter.Data, 20);

                                                            IPv4Address.ValueAsLong := help;
{
                                                            IPv4Socket.IPAddress := IPv4Address;
                                                            IPv4Socket.Port := Port;

                                                            // Do the replacements

                                                            Port := priv_replacementmgr.getReplacementForPort (protUDP, IPv4Socket);
}
                                                            Port := priv_replacementmgr.getReplacementForPort (protUDP, Port);

                                                            help := Port;
                                                            PatchLongwordToByteArray (Parameter.Data, 4, swap32 (help));

                                                            IPv4Address := priv_replacementmgr.getReplacementForAddress(IPv4address);
                                                            PatchLongwordToByteArray (Parameter.Data, 20, IPv4Address.ValueAsLong);
                                                       end
                                                    else
                                                        begin
                                                             if priv_params.RTPSReplacementBlackMarker then
                                                                sanizizeByteArray (Parameter.Data, asZero)
                                                             else
                                                                 sanizizeByteArray (Parameter.Data, asRandomize);
                                                        end;
                                               end;
      RTPS_PID_PROPERTY_LIST                 : begin
                                                    // Text list of hostname, username etc.

                                                    if priv_params.RTPSReplacementBlackMarker then
                                                       sanizizeByteArray (Parameter.Data, asZero)
                                                    else
                                                        sanizizeByteArray (Parameter.Data, asRandomize);
                                               end;
     end;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.GetRTPSAssembly (Parser: tRTPSParser): tFrameChunk;

var assembly : tRTPSAssembler;
    msg      : tRTPSSubMessage;
    len      : integer;
    i,j      : integer;

begin
     assembly := tRTPSAssembler.create;
     assembly.CopyFromParser (Parser);

     if priv_params.RTPSDoReplace then
        begin
             // Go through all submessages to see if they need sanitization

             for i := 1 to length (Assembly.SubMessages) do
                 begin
                      msg := Assembly.SubMessages[i-1];

                      case msg.MessageID of
                       RTPS_ID_DATA       : begin
                                                 // Remove PID_TYPECODE and PID_OBJECT for data(w) and data(r)

                                                 if (msg.DataMessage.WriterEntity = $3c2) or (msg.DataMessage.WriterEntity = $4c2) then
                                                    begin
                                                         len := length (msg.DataMessage.DataList);

                                                         for j := 1 to len do
                                                             begin
                                                                  if (msg.DataMessage.DataList[j-1].DataType = RTPS_PID_TYPECODE) or
                                                                     (msg.DataMessage.DataList[j-1].DataType = RTPS_PID_OBJECT) then
                                                                     begin
                                                                          if priv_params.RTPSReplacementBlackMarker then
                                                                             sanizizeByteArray (Assembly.SubMessages[i-1].DataMessage.DataList[j-1].Data, asZero)
                                                                          else
                                                                              sanizizeByteArray (Assembly.SubMessages[i-1].DataMessage.DataList[j-1].Data, asRandomize);
                                                                     end;
                                                             end;
                                                    end
                                                 else
                                                     // Only sanitize if not Data(p)

                                                     if (msg.DataMessage.WriterEntity = $100c2) then
                                                        begin
                                                             len := length (msg.DataMessage.DataList);

                                                             for j := 1 to len do
                                                                 SanitizeRTPSParameter (Assembly.SubMessages[i-1].DataMessage.DataList[j-1]);
                                                        end
                                                     else
                                                         begin
                                                              if priv_params.RTPSReplacementBlackMarker then
                                                                 sanizizeByteArray (Assembly.SubMessages[i-1].DataMessage.data, asZero)
                                                              else
                                                                  sanizizeByteArray (Assembly.SubMessages[i-1].DataMessage.data, asRandomize);
                                                         end
                                            end;
                       RTPS_ID_DATA_FRAG,
                       RTPS_ID_DATA_BATCH : begin
                                                 if priv_params.RTPSReplacementBlackMarker then
                                                    sanizizeByteArray (Assembly.SubMessages[i-1].data, asZero)
                                                 else
                                                     sanizizeByteArray (Assembly.SubMessages[i-1].data, asRandomize);
                                            end;
                      end;
                 end;
        end;

     result := assembly.GetAssembly;
     assembly.Free;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getReplacementPayload (Payload : tFrameChunk) : tFrameChunk;

var NewPayload  : tFrameChunk;
    ContentStr  : ANSIString;
    original    : ANSIString;
    replacement : ANSIString;
    help        : tHexRepresentation;
    found       : boolean;
    delta       : integer;
    i           : integer;

begin
     ContentStr := Payload.getAsANSIString;

     // Check if we have a replacement rule

     for i := 1 to length (priv_params.PayloadReplacementStrings) do
         begin
              found := false;

              original := priv_params.PayloadReplacementStrings[i-1].original;

              // Handle Snort formatted binary content

              if pos ('|', original) > 0 then
                 begin
                      help.ValueAsString := original;

                      original := help.ValueAsBinaryString;
                 end;

              if priv_params.PayLoadDoReplaceCaseInsensitve then
                 found := (pos (lowercase (original), lowercase (ContentStr)) > 0)
              else
                  found := (pos (original, ContentStr) > 0);

              if found then
                 begin
                      replacement := priv_params.PayloadReplacementStrings[i-1].replacement;

                      // Handle Snort formatted binary content

                      if pos ('|', replacement) > 0 then
                         begin
                              help.ValueAsString := replacement;

                              replacement := help.ValueAsBinaryString;
                         end;

                      if not priv_params.PayloadDoNotEnforceSameLength then
                         begin
                              // Make sure the replacement is the same size as the original to avoid trouble with oversizing
                              // the frame and TCP sequence numbers going nuts

                              if length (original) < length (replacement) then
                                 replacement := copy (replacement, 1, length (original))
                             else
                                  if length (original) > length (replacement) then
                                     replacement := replacement + StringOfChar (' ', length (original) - length (replacement));
                         end;

                      if priv_params.PayLoadDoReplaceCaseInsensitve then
                         ContentStr := stringreplace (ContentStr, original, replacement, [rfIgnoreCase, rfReplaceAll])
                      else
                          ContentStr := stringreplace (ContentStr, original, replacement, [rfReplaceAll]);
                 end;
         end;

     NewPayLoad.initialize;
     NewPayload.copyFrom (ContentStr);

     // Kill the existing Payload

     Payload.free;

     // Return the result

     result := NewPayload;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getSanitizedPayload (ParseMaster: tParseMaster): tFrameChunk;

var PayloadCopy             : tFrameChunk;
    SanitizedPayload        : tFrameChunk;

begin
     if ParseMaster.HasRemainingPayload and not priv_params.PayloadTruncateUnknownLayers  then
        begin
             // If we do not truncate the payload at all, keep it, otherwise mark it empty

             PayLoadCopy := ParseMaster.RemainingPayloadCopy;

             // Do payload string replacement if needed

             if priv_params.PayLoadDoReplaceStrings then
                begin
                     // Sanitize the payload

                     SanitizedPayLoad := getReplacementPayload (PayloadCopy);

                     // Discard the original payload copy

                     PayloadCopy.free;
                end
             else
                 begin
                      // Unknown payloads will be discarded or replaced

                      case priv_params.PayloadRemovalMode of
                       prmCut    : SanitizedPayload.initialize;
                       prmZero   : begin
                                        if ParseMaster.IsTruncated and priv_params.PayloadReAddPaddingIfSliced then
                                           SanitizedPayload.CreateZeroBuffer (ParseMaster.TruncatedByteCount)
                                        else
                                            SanitizedPayload.CreateZeroBuffer (ParseMaster.RemainingPayloadLen);
                                   end;
                       prmString : begin
                                        if ParseMaster.IsTruncated and priv_params.PayloadReAddPaddingIfSliced then
                                           SanitizedPayload.CreateFromPattern (priv_params.PayloadOverwritePattern, ParseMaster.TruncatedByteCount)
                                        else
                                            SanitizedPayload.CreateFromPattern (priv_params.PayloadOverwritePattern, ParseMaster.RemainingPayloadLen);
                                   end;
                      end;

                      // Discard the original payload copy

                      PayloadCopy.free;
                 end;
        end
     else
         SanitizedPayload.initialize;

     result := SanitizedPayload;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getNewRawFrameFromFrameChunk (Frame: tRawFrame; Payload: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
{
     if FrameLengthModifier <> 0 then
        SiMain.LogVerbose ('Frame length adjusted by ' + IntToStr (FrameLengthModifier) + ' for frame ' + IntToStr (Frame.FileFrameID) + '/' + IntToStr (Frame.TracefileID));
}
     WireDetails.WireSize := Frame.WireSize + FrameLengthModifier;

     if priv_params.PayloadDoTruncateAtOffset then
        WireDetails.CaptureSize := Min (Payload.Len, priv_params.PayloadTruncateAtOffset)
     else
         WireDetails.CaptureSize := Payload.Len;

     if WireDetails.CaptureSize > WireDetails.WireSize then
        begin
             SiMain.LogFatal('Invalid capture size detected in frame #' + IntToStr (Frame.FileFrameID) + '. Wiresize is ' + IntToStr (WireDetails.WireSize) + ' bytes, Capture Size is ' + IntToStr (WireDetails.CaptureSize) + ' bytes');


        end;

     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.LinkID := Frame.InterfaceLinkID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := 0;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, min (maxFrameSize, WireDetails.CaptureSize));

     NewFrame := tRawFrame.create (TimeDetails, WireDetails,FilePosition, FrameBuffer);

     // Copy frame comments if allowed

     if (Frame.Commentlist.Count > 0) then
        begin
             if not ((priv_params.PCAPngSanitizeCommentsScope = scsPacket) or (priv_params.PCAPngSanitizeCommentsScope = scsBoth)) then
                for i := 1 to Frame.Commentlist.Count do
                    NewFrame.Commentlist.Add (Frame.Commentlist[i-1]);
        end;

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getIPv6PseudoHeader (ParseMaster : tParseMaster; StartLayer : integer) : tIPv6PseudoHeader;

var IPv6PseudoHeader : tIPv6PseudoHeader;
    Parser           : tFrameParser;
    ParserType       : integer;
    found            : boolean;
    LayerIndex       : integer;

begin
     LayerIndex := StartLayer;
     found := false;

     repeat
           Parser := ParseMaster.getParserByIndex(LayerIndex);
           ParserType := ParseMaster.getParserTypeByIndex(LayerIndex);

           if (ParserType = NEXT_HEADER_IPV6) or (ParserType = NEXT_HEADER_TYPE_IPV6) then
              found := true
           else
               dec (LayerIndex);

     until found or (LayerIndex = 0);

     IPv6PseudoHeader := (Parser as tIPv6Parser).PseudoHeader;

     // TODO: when there are extension headers, we need to check how to handle those with regards to the TCP/UDP CRC process

     result := IPv6PseudoHeader;
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.getSanitizedFrame (Frame: tRawFrame): tRawFrame;

var ParseMaster             : tParseMaster;
    ParseMode               : tParseProcessMode;
    LayerIndex              : integer;
    IPPayloadLen            : word;
    SanitizedPayload        : tFrameChunk;
    MergePayload            : tFrameChunk;
    Parser                  : tFrameParser;
    PrevParser              : tFrameParser;
    IPv4PseudoHeader        : tIPv4PseudoHeader;
    IPv6PseudoHeader        : tIPv6PseudoHeader;
    L4IPv4OriginalPair      : tIPv4AddressPair;
    L4IPv6OriginalPair      : tIPv6AddressPair;
    IPv6Dummy               : tIPv6Address;
    PortPair                : tPortPair;
    Assembly                : tFrameChunk;
    PrevHeaderType          : tHeaderType;
    HeaderType              : tHeaderType;
    CRCOkay                 : boolean;
    SkipAssembly            : boolean;
    UseModifiedPseudoHeader : boolean;
    VLANID                  : word;
    NewEtherType            : word;
    HelpLen                 : integer;
    WireSizeModifier        : integer;
    PayloadLenModifier      : integer;
    EthernetFCS             : longword;
    iHelp                   : integer;
    i                       : integer;

begin
//   SiMain.EnterMethod ('tTaskProcessorAnonymize.getSanitizedFrame');

     EthernetFCS := 0;

     if priv_params.PayloadDoTruncateAfterLayer then
        begin
             case priv_params.PayloadTruncateAfterLayer of
              2 : ParseMode := ppmUpToLayer2;
              3 : ParseMode := ppmUpToLayer3;
              4 : ParseMode := ppmUpToLayer4;
             end;
        end
     else
         ParseMode := ppmSanitize;

     try
        ParseMaster := tParseMaster.create(Frame, ParseMode);
     except
           SiMain.LogFatal ('Creating ParseMaster failed for frame #' + Frame.FileFrameID.ToString + ' (tTaskProcessorAnonymize.getSanitizedFrame)');

           inc (prot_ParseErrorCount);
           Exit (nil);
     end;

     // Track protocols

     priv_ProtocolsSeen := priv_ProtocolsSeen + Parsemaster.ParsedProtocols;

     // Check if the frame should be completely removed from the trace, e.g.
     // when certain ICMP type/codes are found

     if RemoveFrameFromTrace (ParseMaster) then
        begin
             try
                ParseMaster.free;
             except
                   SiMain.LogFatal ('Freeing ParseMaster to remove the frame from trace failed for frame #' + Frame.FileFrameID.ToString + ' (tTaskProcessorAnonymize.getSanitizedFrame)');
             end;

             Exit (nil);
        end;

     MergePayload.initialize;

     // Init the modifier we need to keep track of frame size deviation

     WireSizeModifier := 0;
     PayloadLenModifier := 0;

     // Store the payload length for the IP layer in case we truncate layers. If the frame already is truncated we need to add that here, too

     IPPayloadLen := ParseMaster.RemainingPayloadLen;

     if ParseMaster.IsTruncated then
        IPPayloadLen := IPPayloadLen + ParseMaster.TruncatedByteCount;

     // Determine what to do with the remaining unparsed payload

{   Refactoring disabled     SanitizedPayload := getSanitizedPayload (ParseMaster);

     // If the payload still exists but is now shorter/longer we need to keep that in mind

     if priv_params.PayLoadDoReplaceStrings then
        begin
             PayloadLenModifier := ParseMaster.RemainingPayloadLen - SanitizedPayload.Len;

             IPPayloadLen := IPPayloadLen - PayloadLenModifier;
             WireSizeModifier := WireSizeModifier - PayloadLenModifier;
        end;
}
     // Determine what to do with the remaining unparsed payload

     if ParseMaster.HasRemainingPayload and not priv_params.PayloadTruncateUnknownLayers  then
        begin
             // If we do not truncate the payload at all, keep it, otherwise mark it empty

             SanitizedPayload := ParseMaster.RemainingPayloadCopy;

             // Do payload string replacement if needed

             if priv_params.PayLoadDoReplaceStrings then
                begin
                     PayloadLenModifier := SanitizedPayload.Len;

//                   SanitizedPayload := getSanitizedPayload (SanitizedPayload);     // refactored to getReplacementPayload
                     SanitizedPayload := getReplacementPayload (SanitizedPayload);

                     // If the payload is now shorter we need to keep that in mind

                     PayloadLenModifier := PayloadLenModifier - SanitizedPayload.Len;

                     IPPayloadLen := IPPayloadLen - PayloadLenModifier;
                     WireSizeModifier := WireSizeModifier - PayloadLenModifier;
                end;
        end
     else
         begin
              // Unknown payloads will be discarded or replaced

              case priv_params.PayloadRemovalMode of
               prmCut    : SanitizedPayload.initialize;
               prmZero   : begin
                                if ParseMaster.IsTruncated and priv_params.PayloadReAddPaddingIfSliced then
                                   SanitizedPayload.CreateZeroBuffer (ParseMaster.TruncatedByteCount)
                                else
                                    SanitizedPayload.CreateZeroBuffer (ParseMaster.RemainingPayloadLen);
                           end;
               prmString : begin
                                if ParseMaster.IsTruncated and priv_params.PayloadReAddPaddingIfSliced then
                                   SanitizedPayload.CreateFromPattern (priv_params.PayloadOverwritePattern, ParseMaster.TruncatedByteCount)
                                else
                                    SanitizedPayload.CreateFromPattern (priv_params.PayloadOverwritePattern, ParseMaster.RemainingPayloadLen);
                           end;
              end;
         end;

     // Now go through the layers from top to bottom and work on them
     // Determine where to start by looking at the OSI layers of the parsers

     if priv_params.PayloadDoTruncateAfterLayer then
        begin
             LayerIndex := ParseMaster.ParserCount + 1;

             repeat
                   dec (LayerIndex);

                   // Determine layer type and create sanitized assembly from it

                   HeaderType := ParseMaster.getParserTypeByIndex(LayerIndex);
                   Parser := ParseMaster.getParserByIndex(LayerIndex);

             until (Parser.OSILayer <= priv_params.PayloadTruncateAfterLayer) or (LayerIndex = 1);

             // If we didn't find the layer we keep them all

             if not (Parser.OSILayer <= priv_params.PayloadTruncateAfterLayer) then
                LayerIndex := ParseMaster.ParserCount;

             // Check if the remaining payload is directly after the start layer, if not, forget the payload

             if (LayerIndex <= ParseMaster.ParserCount) then
                SanitizedPayload.free;
        end
     else
         begin
              LayerIndex := ParseMaster.ParserCount;

              // Check if the remaining payload is directly after the start layer, if not, forget the payload

              if (LayerIndex < ParseMaster.ParserCount) then
                 SanitizedPayload.free;
         end;

     // Main Loop

//   Refactoring disabled   LayerIndex := ParseMaster.ParserCount;

     UseModifiedPseudoHeader := false;
     NewEtherType := 0;

     while LayerIndex > 0 do
           begin
                // Determine layer type and create sanitized assembly from it

                HeaderType := ParseMaster.getParserTypeByIndex(LayerIndex);
                Parser := ParseMaster.getParserByIndex(LayerIndex);
                CRCOkay := ParseMaster.getParserCRCOkayByIndex(LayerIndex);

                SkipAssembly := false;

                case HeaderType of
                 NEXT_HEADER_RTPS               : begin
                                                       Assembly := GetRTPSAssembly (Parser as tRTPSParser);
                                                  end;
                 NEXT_HEADER_HSRP               : begin
                                                       Assembly := GetHSRPAssembly (Parser as tHSRPParser);
                                                  end;
                 NEXT_HEADER_DHCPv4             : begin
                                                       Assembly := getDHCPv4Assembly (Parser as tDHCPv4Parser);
                                                  end;
                 NEXT_HEADER_NETFLOW            : begin
                                                       Assembly := GetNetFlowAssembly (Parser as tNetflowParser);
                                                  end;
                 NEXT_HEADER_TCP                : begin
                                                       PrevHeaderType := ParseMaster.getParserTypeByIndex(LayerIndex-1);
                                                       PrevParser := ParseMaster.getParserByIndex(LayerIndex-1);

                                                       // Store current port pair in case we need it in IP sanitation (because of socket pair replacement list)

                                                       PortPair := (Parser as tTCPParser).PortPair;

                                                       case PrevHeaderType of
                                                        NEXT_HEADER_IPv4,
                                                        NEXT_HEADER_IPv4_ENCAPSULATION : begin
                                                                                              IPv4PseudoHeader := (PrevParser as tIPv4Parser).PseudoHeader;

                                                                                              L4IPv4OriginalPair.Source.ValueAsLong := IPv4PseudoHeader.Source;
                                                                                              L4IPv4OriginalPair.Destination.ValueAsLong := IPv4PseudoHeader.Destination;

                                                                                              // Update Pseudoheaders, otherwise CRC calculation will fail because it is calculated against an outdated Pseudoheader

                                                                                              SanitizeIPv4PseudoHeader (IPv4PseudoHeader);

                                                                                              Assembly := GetTCPAssembly (Parser as tTCPParser, L4IPv4OriginalPair, IPv4PseudoHeader, SanitizedPayload, CRCOkay);
                                                                                         end;
                                                       else
                                                           begin
                                                                IPv6PseudoHeader := getIPv6PseudoHeader (ParseMaster, LayerIndex-1);

                                                                L4IPv6OriginalPair.Source.ValueAsWordStruct := IPv6PseudoHeader.Source;
                                                                L4IPv6OriginalPair.Destination.ValueAsWordStruct := IPv6PseudoHeader.Destination;

                                                                // Update Pseudoheaders, otherwise CRC calculation will fail because it is calculated against an outdated Pseudoheader

                                                                SanitizeIPv6PseudoHeader (IPv6PseudoHeader);

                                                                Assembly := GetTCPAssembly (Parser as tTCPParser, L4IPv6OriginalPair, IPv6PseudoHeader, SanitizedPayload, CRCOkay);
                                                           end;
                                                       end;

                                                       // TODO: store payload length modifier when not the same, to adjust sequences later






                                                       UseModifiedPseudoHeader := true;
                                                  end;
                 NEXT_HEADER_UDP                : begin
                                                       PrevHeaderType := ParseMaster.getParserTypeByIndex(LayerIndex-1);
                                                       PrevParser := ParseMaster.getParserByIndex(LayerIndex-1);

                                                       // Store current port pair in case we need it in IP sanitation (because of socket pair replacement list)

                                                       PortPair := (Parser as tUDPParser).PortPair;

                                                       case PrevHeaderType of
                                                        NEXT_HEADER_IPv4,
                                                        NEXT_HEADER_IPv4_ENCAPSULATION : begin
                                                                                              IPv4PseudoHeader := (PrevParser as tIPv4Parser).PseudoHeader;

                                                                                              L4IPv4OriginalPair.Source.ValueAsLong := IPv4PseudoHeader.Source;
                                                                                              L4IPv4OriginalPair.Destination.ValueAsLong := IPv4PseudoHeader.Destination;

                                                                                              // Update Pseudoheaders, otherwise CRC calculation will fail because it is calculated against an outdated Pseudoheader

                                                                                              SanitizeIPv4PseudoHeader (IPv4PseudoHeader);

                                                                                              Assembly := GetUDPAssembly (Parser as tUDPParser, L4IPv4OriginalPair, IPv4PseudoHeader, SanitizedPayload, CRCOkay);
                                                                                         end;
                                                       else
                                                           begin
                                                                IPv6PseudoHeader := getIPv6PseudoHeader (ParseMaster, LayerIndex-1);

                                                                L4IPv6OriginalPair.Source.ValueAsWordStruct := IPv6PseudoHeader.Source;
                                                                L4IPv6OriginalPair.Destination.ValueAsWordStruct := IPv6PseudoHeader.Destination;

                                                                // Update Pseudoheaders, otherwise CRC calculation will fail because it is calculated against an outdated Pseudoheader

                                                                SanitizeIPv6PseudoHeader (IPv6PseudoHeader);

                                                                Assembly := GetUDPAssembly (Parser as tUDPParser, L4IPv6OriginalPair, IPv6PseudoHeader, SanitizedPayload, CRCOkay);
                                                           end;
                                                       end;

                                                       UseModifiedPseudoHeader := true;
                                                  end;
                 NEXT_HEADER_MPLS_UNICAST,
                 NEXT_HEADER_MPLS_MULTICAST     : begin
                                                       Assembly := GetMPLSAssembly (Parser as tMPLSParser);
                                                  end;
                 NEXT_HEADER_IPV6,
                 NEXT_HEADER_TYPE_IPV6          : begin
                                                       IPv6PseudoHeader.L4Length := IPPayloadLen;

                                                       // TODO: implement delta adjustements for hard sliced frames like we do for IPv4

                                                       // If the pseudoheader was already modified in TCP/UDP/ICMP we use its values to be consistent, otherwise we don't care about it

                                                       if UseModifiedPseudoHeader then
                                                          Assembly := GetIPv6Assembly (Parser as tIPv6Parser, IPv6PseudoHeader, PortPair)
                                                       else
                                                           Assembly := GetIPv6Assembly (Parser as tIPv6Parser, PortPair);

                                                       // Prevent accidental use of the wrong pseudoheader in case of multiple IP layers following this one

                                                       UseModifiedPseudoHeader := false;
                                                  end;
                 NEXT_HEADER_IPV6_FRAGMENTATION : begin
                                                       Assembly := getIPv6FragmentationAssembly (Parser as tIPv6FragmentationParser);
                                                  end;
                 NEXT_HEADER_ICMP               : begin
{
                                                       case (Parser as tICMPv4Parser).ICMPType of
                                                        0,8 : begin
                                                                   // Echo Request / Echo Reply are the only types at the moment that can have an "ICMP payload"
                                                                   // all others (quoted layers) are handled by their protocol specific assemblers

                                                                   if SanitizedPayload.Len > 0 then
                                                                      Assembly := GetICMPv4Assembly (Parser as tICMPv4Parser, SanitizedPayload, CRCOkay)
                                                                   else
                                                                       Assembly := GetICMPv4Assembly (Parser as tICMPv4Parser, CRCOkay);
                                                              end;
                                                       else
}
                                                       Assembly := GetICMPv4Assembly (Parser as tICMPv4Parser, SanitizedPayload, CRCOkay);
{
                                                       end;
}

                                                  end;
                 NEXT_HEADER_IPV6_ICMP          : begin
                                                       // find the IPv6 layer to create the pseudo header. It should be the previous layer, but there may be extension headers in between...

                                                       i := LayerIndex;

                                                       repeat
                                                             dec (i);

                                                             PrevHeaderType := ParseMaster.getParserTypeByIndex(i);

                                                       until (PrevHeaderType = NEXT_HEADER_IPV6) or (PrevHeaderType = NEXT_HEADER_TYPE_IPV6);   // or kaboom (i < 1) :-)

                                                       PrevParser := ParseMaster.getParserByIndex (i);

                                                       IPv6PseudoHeader := (PrevParser as tIPv6Parser).PseudoHeader;
                                                       UseModifiedPseudoHeader := true;

                                                       // Check if target is solicited node multicast, because then we may to have to force its sanitization first
                                                       // We do this by requesting a replacement for the specific address the ND is looking for, which write a replacement
                                                       // to the database and creates a replacement for the Solicited Node address automatically. That way it already exists
                                                       // when we access it, and it matches the correct address.

                                                       IPv6Dummy := (PrevParser as tIPv6Parser).Destination;

                                                       if IPv6Dummy.isSolicitedNodeMCast then
                                                          IPv6Dummy := priv_replacementmgr.getReplacementForAddress ((Parser as tICMPv6Parser).NDTargetAddress);

                                                       // Update Pseudoheaders, otherwise CRC calculation will fail because it is calculated against an outdated Pseudoheader

                                                       SanitizeIPv6PseudoHeader (IPv6PseudoHeader);

                                                       if SanitizedPayload.Len > 0 then
                                                          Assembly := GetICMPv6Assembly (Parser as tICMPv6Parser, IPv6PseudoHeader, SanitizedPayload, CRCOkay)
                                                       else
                                                           Assembly := GetICMPv6Assembly (Parser as tICMPv6Parser, IPv6PseudoHeader, CRCOkay);
                                                  end;
                 NEXT_HEADER_IPv4,
                 NEXT_HEADER_IPv4_ENCAPSULATION : begin
                                                       iHelp := (Parser as tIPv4Parser).FrameTotalLengthDelta;

                                                       // Determine if the IP header has a different length than the wiresize of the frame, e.g. when
                                                       // hard sliced. Example: IP Total Length 1500, but frame wire size is only 1510, missing 4 bytes -> ihelp := -4
                                                       // if iHelp > 0 then we may have an Ethernet trailer, which we ignore here, but this can also happen if padding was sliced, so we need
                                                       // to make sure not to exceed the original payload length

                                                       if iHelp < 0 then
                                                          IPPayloadLen := IPPayloadLen + (iHelp * -1)
                                                       else
                                                           if iHelp > 0 then
                                                              begin
                                                                   // check the original payload length. If it's less than IPPayloadLen we may have a case of sliced padding and need to adjust IPPayloadLen.
                                                                   // If we don't do this TCP will get into trouble with Sequence number tracking, caused by misleading IP Total Len values.

                                                                   if IPPayloadLen > SanitizedPayload.len then
                                                                      IPPayloadLen := SanitizedPayload.len;
                                                              end;

                                                       IPv4PseudoHeader.L4Length := IPPayloadLen;

                                                       // If the pseudoheader was already modified in TCP/UDP/ICMP we use its values to be consistent, otherwise we don't care about it

                                                       if UseModifiedPseudoHeader then
                                                          Assembly := GetIPv4Assembly (Parser as tIPv4Parser, IPv4PseudoHeader, PortPair, IPPayLoadLen, CRCOkay)
                                                       else
                                                           Assembly := GetIPv4Assembly (Parser as tIPv4Parser, PortPair, IPPayLoadLen, CRCOkay);

                                                       // Prevent accidental use of the wrong pseudoheader in case of multiple IP layers following this one

                                                       UseModifiedPseudoHeader := false;
                                                  end;
                 NEXT_HEADER_VXLAN              : begin
                                                       Assembly := getVXLANAssembly (Parser as tVXLANParser);
                                                  end;
                 NEXT_HEADER_AYIYA              : begin
                                                       Assembly := getAyiyaAssembly (Parser as tAYIYAParser);
                                                  end;
                 NEXT_HEADER_GRE                : begin
                                                       Assembly := getGREAssembly (Parser as tGREParser);
                                                  end;
                 NEXT_HEADER_GTPv1              : begin
                                                       Assembly := GetGTPv1Assembly (Parser as tGTPv1Parser);
                                                  end;
                 NEXT_HEADER_8021_Q,
                 NEXT_HEADER_802_QINQ_Outer,
                 NEXT_HEADER_802_QINQ_Outer_OLD1,
                 NEXT_HEADER_802_QINQ_Outer_OLD2,
                 NEXT_HEADER_802_QINQ_Outer_OLD3: begin
                                                       VLANID := (Parser as t802dot1qParser).VLANID;

                                                       if priv_params.VLANDoReplace then
                                                          begin
                                                               case priv_params.VLANRemoveMode of
                                                                vrmRemoveAll         : SkipAssembly := true;
                                                                vrmRemoveAllListed   : begin
                                                                                            if priv_params.VLANRemoveList[VLANID] then
                                                                                               SkipAssembly := true;
                                                                                       end;
                                                                vrmRemoveAllUnlisted : begin
                                                                                            if not priv_params.VLANRemoveList[VLANID] then
                                                                                               SkipAssembly := true;
                                                                                       end;
                                                               end;
                                                          end;

                                                          if SkipAssembly then
                                                             begin
                                                                  NewEtherType := (Parser as t802dot1qParser).EtherType;
                                                                  WireSizeModifier := WireSizeModifier - 4;
                                                             end
                                                          else
                                                              begin
                                                                   Assembly := GetVLANAssembly (Parser as t802dot1qParser, NewEtherType);
                                                                   NewEtherType := 0;
                                                              end;
                                                  end;
                 NEXT_HEADER_ARP,
                 NEXT_HEADER_RARP               : Assembly := GetARPAssembly (Parser as tARPParser);
                 NEXT_HEADER_LINUX_COOKED       : Assembly := GetLinuxCookedAssembly (Parser as tLinuxSLLParser);
                 NEXT_HEADER_ETHERNET,
                 NEXT_HEADER_TRANS_ETHER_BRIDGE : begin
                                                       // If this is a transparent Ethernet bridge AND we have no forced Ethertype, set it to transparent Ethernet bridge

                                                       if (NewEtherType = 0) and (HeaderType = NEXT_HEADER_TRANS_ETHER_BRIDGE) then
                                                          NewEtherType := NEXT_HEADER_TRANS_ETHER_BRIDGE;

                                                       Assembly := GetEthernetAssembly (Parser as tEthernetParser, NewEtherType);

                                                       // If the original frame had the FCS it will be lost, so reduce the frame size by 4, but only if we're not
                                                       // going to add it again later (which only happens if the frame isn't sliced)

                                                       if priv_params.ShortenPayload and ParseMaster.HasRemainingPayload and ParseMaster.HasEthernetFCS then
                                                          WireSizeModifier := WireSizeModifier - 4;
                                                  end;

                 NEXT_HEADER_LLC                : begin
                                                       Assembly := GetLLCAssembly (Parser as tLLCParser);
                                                  end;
                 NEXT_HEADER_802dot11Mgmt       : begin
                                                       Assembly := Get802dot11MgmtAssembler (Parser as t802dot11ManagementParser);
                                                  end;
                 NEXT_HEADER_802dot11Hdr        : begin
                                                       Assembly := Get802dot11HdrAssembler (Parser as t802dot11HeaderParser);
                                                  end;
                 NEXT_HEADER_80211RADIO         : begin
                                                       EthernetFCS := getCRC32FromPointer (SanitizedPayload.Data, SanitizedPayload.Len);

                                                       Assembly := Get802dot11RadioTapAssembly (Parser as t802dot11RadioTapParser);
                                                  end;
                else
                    begin
                         Assembly.initialize;

                         SkipAssembly := true;
                    end;
                end;

                // Adjust IPPayload Length (this is also counted even when we have already created the IP layer, but doesn't hurt because it is only used when creating one)

                inc (IPPayloadLen, Assembly.Len);

                // If assembly was done, merge it

                if not SkipAssembly then
                   begin
                        // Sanitity check

                        if Assembly.Len = 0 then
                           begin
                                SiMain.LogFatal('Error: assembly for frame #' + IntToStr (Frame.FileFrameID) + ' failed, returned assembly length of 0');

                                SkipAssembly := true;
                           end;

                        // Check length deviation

                        if Parser.LayerSize <> Assembly.Len then
                           begin
                                HelpLen := Assembly.Len - Parser.LayerSize;

                                if HelpLen <> 0 then
                                   WireSizeModifier := WireSizeModifier + HelpLen;
                           end;

                        // Merge with the new layer coming in front of what we already have

                        if SanitizedPayload.Len = 0 then
                           begin
                                SanitizedPayload.Len := Assembly.Len;
                                SanitizedPayload.Data := Assembly.Data;
                           end
                        else
                            begin
                                 MergePayload := mergeFrameChunks (Assembly, SanitizedPayload);

                                 // Discard the single chunks

                                 Assembly.free;
                                 SanitizedPayload.free;

                                 // Keep the merge result

                                 SanitizedPayload := MergePayload;
                            end;
                   end;

                // Go down to the next layer

                dec (LayerIndex);
           end;

     // Add Ethernet padding/trailer/FCS if it makes sense (meaning: when the payload isn't truncated anyway, or there is nothing to truncate)

     if (not priv_params.ShortenPayload) or (not ParseMaster.HasRemainingPayload) then
        begin
             // Handle Trailer

             if ParseMaster.PaddingByteCount > 0 then
                begin
                     // Check if we need to copy padding bytes or zero them

                     if priv_params.EthernetDoReplace and priv_params.EthernetPreserveTrailingBytes then
                        Assembly := ParseMaster.PaddingByteCopy
                     else
                         Assembly.CreateZeroBuffer (ParseMaster.PaddingByteCount);

                     MergePayload := mergeFrameChunks (SanitizedPayload, Assembly);
                     Assembly.free;
                     SanitizedPayload.free;

                     SanitizedPayload := MergePayload;
                end;

             // Handle FCS, if the original frame had one

             if ParseMaster.HasEthernetFCS then
                begin
                     case Frame.LinkType of
                      LINKTYPE_ETHERNET         : begin
                                                       case priv_params.EthernetRecalculateFCSMode of
                                                        rcmKeepCRC    : SanitizedPayload.AttachCRC32 (ParseMaster.EthernetFCS);
                                                        rcmKeepBadCRC : SanitizedPayload.AttachCRC32 (not Parsemaster.EthernetFSCCorrect);
                                                        rcmFixZeroCRC : begin
                                                                             if ParseMaster.EthernetFCS = 0 then
                                                                                SanitizedPayload.AttachCRC32 (false)
                                                                             else
                                                                                 SanitizedPayload.AttachCRC32 (not Parsemaster.EthernetFSCCorrect);
                                                                        end;
                                                        rcmFixAll     : SanitizedPayload.AttachCRC32 (false);
                                                       end;
                                                  end;
                      LINKTYPE_IEEE802_11_RADIO : begin
                                                       // for 802.11, we always set the FCS bad or correct depending on the original. No fixing FCS here,
                                                       // because bad CRCs are important to keep.

                                                       SanitizedPayload.AttachCRC32 (not Parsemaster.EthernetFSCCorrect);
                                                  end;
                     end;
                end;
        end;

     try
        ParseMaster.free;
        ParseMaster := nil;
     except
           SiMain.LogFatal ('Freeing ParseMaster failed (tTaskProcessorAnonymize.getSanitizedFrame)');
     end;

     // Turn Payload into a tRawFrame

     result := getNewRawFrameFromFrameChunk (Frame, SanitizedPayload, WireSizeModifier);

     SanitizedPayload.free;

//   SiMain.LeaveMethod ('tTaskProcessorAnonymize.getSanitizedFrame');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.ProcessSingleFrame (var OriginalFrame : tRawFrame) : tTaskErrorCode;

var SanitizedFrame    : tRawFrame;
    InterfaceEntry    : tCaptureInterface;
    ErrorCode         : tTaskErrorCode;

begin
//   SiMain.EnterMethod ('tTaskProcessorAnonymize.ProcessSingleFrame');

     ErrorCode := ecSuccess;

     try
        SanitizedFrame := getSanitizedFrame (OriginalFrame);

//      SanitizedFrame := tRawFrame.create (OriginalFrame);     debug only, for excluding the sanitization process

     except
           SiMain.LogFatal ('Could not get sanitized frame for frame #' + OriginalFrame.FileFrameID.ToString + ', exiting (tTaskProcessorAnonymize.ProcessSingleFrame)');

           ErrorCode := ecFailure;

           SiMain.LeaveMethod ('tTaskProcessorAnonymize.ProcessSingleFrame');

           Exit (ErrorCode);
     end;

     // If we get a sanitized frame back, write it

     if SanitizedFrame <> nil then
        begin
             // Assign the new interface by looking up the writer interface of the same index
             // since the writing process created the new interfaces in the same order

             InterfaceEntry := priv_tracewriter.InterfaceList.getInterfaceByIndex(OriginalFrame.InterfaceID);
             SanitizedFrame.InterfaceLinkID := InterfaceEntry.PacketLinkID;

             // Add sanitized frame to writer

             try
                priv_tracewriter.addFrameToList (SanitizedFrame);
             except
                   SiMain.LogFatal ('Could add sanitized frame for original frame #' + OriginalFrame.FileFrameID.ToString + 'to TraceWriter, exiting (tTaskProcessorAnonymize.ProcessSingleFrame)');

                   ErrorCode := ecFailure;

                   SiMain.LeaveMethod ('tTaskProcessorAnonymize.ProcessSingleFrame');

                   Exit (ErrorCode);
             end;
        end;

     result := ErrorCode;

//   SiMain.LeaveMethod ('tTaskProcessorAnonymize.ProcessSingleFrame');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.processFragmentChain : tTaskErrorCode;

var OriginalFrame      : tRawFrame;
    SanitizedFullFrame : tRawFrame;
    SanitizedFrame     : tRawFrame;
    InterfaceEntry     : tCaptureInterface;
    ErrorCode          : tTaskErrorCode;
    Entry              : tSanitizationProcessEntry;
    FragmentedPacket   : tFragmentedPacket;
    MaxIPPayloadLen    : integer;
    ProcessSingleFrame : boolean;
    len                : integer;
    key                : string;
    i                  : integer;

begin
     SiMain.EnterMethod ('tTaskProcessorAnonymize.processFragmentChain');

     ErrorCode := ecSuccess;

     len := Length (priv_processchain);

     SiMain.LogDebug ('Processing fragment processchain with length ' + len.ToString);

     // Get sanitized frames for all frames in the chain

     for i := 1 to len do
         begin
              SiMain.LogDebug ('Processing frame #' + IntToStr (priv_processchain[i-1].Frame.FileFrameID) + ' (chain location ' + IntToStr (i) + ')');

              Entry := priv_processchain[i-1];

              // Check if the current frame is handled solo

              processSingleFrame := not priv_processchain[i-1].isFragment;

              if not processSingleFrame then
                 begin
                      // Check if this fragment is part of a complete set

                      FragmentedPacket := priv_ipfragmentmgr.GetFragmentedPacketEntry (Entry.IPAddress, Entry.IPIdentification);

                      // if not, process solo

                      processSingleFrame := FragmentedPacket.isIncomplete;

                      SiMain.LogVerbose ('Sanitizing incomplete fragment frame #' + IntToStr (priv_processchain[i-1].Frame.FileFrameID) + ' (chain location ' + IntToStr (i) + ')');
                 end
              else
                  SiMain.LogVerbose ('Sanitizing non-fragment frame #' + IntToStr (priv_processchain[i-1].Frame.FileFrameID) + ' (chain location ' + IntToStr (i) + ')');

              // If this is just an innocent non-fragmented packet that got caught between fragments,
              // or a fragment of an incomplete set, simply process it as usual

              if processSingleFrame then
                 begin
                      try
                         SanitizedFrame := getSanitizedFrame (priv_processchain[i-1].Frame);

                      except
                            SiMain.LogFatal ('Could not get sanitized non-fragment frame, exiting (tTaskProcessorAnonymize.processFragmentChain)');

                            ErrorCode := ecFailure;

                            SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');

                            Exit (ErrorCode);
                      end;
                 end
              else
                  begin
                       // this is a fragment, so processing it is much more complex

                       key := getFragmentedPacketDictionaryKey (Entry.IPAddress, Entry.IPIdentification);

                       SiMain.LogDebug ('Searching for sanitized reassembled frame with key "' + key + '"');

                       if priv_reassemblystorage.ContainsKey (key) then
                          begin
                               SiMain.LogDebug ('Sanitized Full frame found, retrieving from dictionary');

                               // We have already reassembled and sanitized the frame from which we need a fragment

                               SanitizedFullFrame := priv_reassemblystorage.Items[key];
                          end
                       else
                           begin
//                              SiMain.LogDebug ('No sanitized full frame available yet, constructing new reassembled frame from fragments for sanitization');

                                // We haven't yet reassembled the frame we will get the fragments from
                                // so we need to reassemble it first from all original fragments

                                OriginalFrame := priv_ipfragmentmgr.GetReassembledFrame (Entry.IPAddress, Entry.IPIdentification);

//                              SiMain.LogDebug ('Length of reassembled frame: ' + OriginalFrame.FrameSize.ToString);

                                // Make sure we got a frame

                                assert (assigned (OriginalFrame), 'Failed to get reassembled frame ' + Entry.Frame.FileFrameID.ToString + ' (tTaskProcessorAnonymize.ProcessFragmentChain)');

                                // Continue if the reassembly was successfull

                                if assigned (OriginalFrame) then
                                   begin
                                        // Also check what the maximum size for a fragment is

                                        MaxIPPayloadLen := priv_ipfragmentmgr.GetMaximumIPPayloadLength (Entry.IPAddress, Entry.IPIdentification);

//                                      SiMain.LogDebug ('Maximum IP payload length per fragment is ' + MaxIPPayloadLen.ToString);

                                        // Get a sanitized version of the reassembled frame

                                        try
//                                         SiMain.LogVerbose ('Sanitizing reassembled frame');

                                           SanitizedFullFrame := getSanitizedFrame (OriginalFrame);

                                        except
                                              SiMain.LogFatal('Could not get sanitized frame for reassembled full frame #' + OriginalFrame.FileFrameID.ToString + ', exiting (tTaskProcessorAnonymize.processFragmentChain)');

                                              ErrorCode := ecFailure;

                                              SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');

                                              Exit (ErrorCode);
                                        end;

//                                      SiMain.LogDebug ('Length of sanitized reassembled frame: ' + SanitizedFullFrame.FrameSize.ToString);

                                        // Add to reassembly storage so we don't have to do it every time we see a fragment of this frame

//                                      SiMain.LogDebug ('Storing new sanitized frame with key "' + key + '"');

                                        priv_reassemblystorage.Add (key, SanitizedFullFrame);
                                   end
                                else
                                    begin
                                         // Looks like we have at least one missing fragment, or they are sliced

                                         try
//                                          SiMain.LogVerbose ('Sanitizing fragmented frame #' + IntToStr (priv_processchain[i-1].Frame.FileFrameID) + 'that could not be reassembled (chain location ' + IntToStr (i) + ')');

                                            SanitizedFrame := getSanitizedFrame (priv_processchain[i-1].Frame);

                                         except
                                               SiMain.LogFatal('Could not get sanitized non-fragment frame #' + priv_processchain[i-1].Frame.FileFrameID.ToString + ', exiting (tTaskProcessorAnonymize.processFragmentChain)');

                                               ErrorCode := ecFailure;

                                               SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');

                                               Exit (ErrorCode);
                                         end;
                                    end;
                           end;

                       // Now get the fragment we need

                       SanitizedFrame := GetFragmentFromPacket (SanitizedFullFrame, Entry.FragmentationDetails.Offset, MaxIPPayloadLen, Entry.Frame.Ticks, Entry.Frame.Timestamp);

                       assert (assigned (SanitizedFrame), 'Failed to get fragment from sanitized reassembled packet #' + SanitizedFullFrame.FileFrameID.ToString + ' (tTaskProcessorAnonymize.ProcessFragmentChain)');
                  end;

              // If we get a sanitized frame back, write it

              if SanitizedFrame <> nil then
                 begin
//                    SiMain.LogVerbose ('Adding sanitized frame #' + IntToStr (priv_processchain[i-1].Frame.FileFrameID) + ' to PCAPng writer (chain location ' + IntToStr (i) + ')');

                      // Assign the new interface by looking up the writer interface of the same index
                      // since the writing process created the new interfaces in the same order

                      InterfaceEntry := priv_tracewriter.InterfaceList.getInterfaceByIndex (Entry.Frame.InterfaceID);
                      SanitizedFrame.InterfaceLinkID := InterfaceEntry.PacketLinkID;

                      // Add sanitized frame to writer

                      try
                         priv_tracewriter.addFrameToList (SanitizedFrame);
                      except
                            SiMain.LogFatal ('Could add frame #' + SanitizedFullFrame.FileFrameID.ToString + ' to TraceWriter, exiting (tTaskProcessorAnonymize.ProcessSingleFrame)');

                            ErrorCode := ecFailure;

                            SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');

                            Exit (ErrorCode);
                      end;
                 end;

              // Drop the original frame, we're done with it

//            SiMain.LogDebug ('Discarding Original Frame');

              priv_processchain[i-1].Frame.free;
              priv_processchain[i-1].Frame := nil;

              // TODO: discard fragment entries in the fragmentation manager to avoid reusage problems with the same IP:IPID pair (after some time)

              if ErrorCode <> ecSuccess then
                 begin
                      SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');

                      Exit (ErrorCode);
                 end;
         end;

     // Disable fragment chaining mode

     priv_FragmentModeActive := False;

     SiMain.LogDebug ('FragmentMode not active');

     // Empty chain

     setLength (priv_processchain, 0);

     result := ErrorCode;

     SiMain.LeaveMethod ('tTaskProcessorAnonymize.processFragmentChain');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.AddFrameToReassemblyProcessChain (const IPAddress : tIPAddress; const IPID : longword; const IPFragmentDetails : tIPFragment): tTaskErrorCode;

var ErrorCode  : tTaskErrorCode;
    timestamp  : tFrametimestamp;
    isFragment : boolean;
    len        : integer;

begin
     SiMain.EnterMethod ('AddFrameToReassemblyProcessChain');

     ErrorCode := ecSuccess;

     isFragment := (IPFragmentDetails.Offset > 0) or (ipfMoreFragments in IPFragmentDetails.Flags) or (ipfLastFragment in IPFragmentDetails.Flags);

     len := Length (priv_processchain);
     setLength (priv_processchain, len + 1);

     priv_processchain[len].IPAddress := IPAddress;
     priv_processchain[len].IPIdentification := IPID;
     priv_processchain[len].Frame := IPFragmentDetails.Frame;                    // for easier access
     priv_processchain[len].FragmentationDetails := IPFragmentDetails;
     priv_processchain[len].isFragment := isFragment;

     SiMain.LogDebug ('New Length of process chain is ' + IntToStr (len + 1));

     // Add fragment to Fragmentation Manager if the frame is a fragment

     if isFragment then
        begin
             SiMain.LogDebug ('Frame #' + IntToStr (IPFragmentDetails.Frame.FileFrameID) + ' is a fragment, adding to Fragmentation manager');

             priv_ipfragmentmgr.AddPacketFragmentAndCheckComplete (IPAddress, IPID, IPFragmentDetails.Frame, IPFragmentDetails.Offset, IPFragmentDetails.Len, IPFragmentDetails.Flags);
        end
     else
         SiMain.LogDebug ('Frame #' + IntToStr (IPFragmentDetails.Frame.FileFrameID) + ' is not a fragment');

     // Check if we have a complete set of fragments (may involve multiple interleaved fragmeted packets)
     // if so, we can process the chain and clear it

     if priv_ipfragmentmgr.AllPacketFragmentsAvailable then
        begin
             SiMain.LogDebug ('All fragments available, processing Fragment Chain (containing '+ IntToStr (len) + ' frames)');

             ErrorCode := ProcessFragmentChain;
        end
     else
         begin
              SiMain.LogDebug ('Some fragments still missing, cannot yet process Fragment Chain (containing '+ IntToStr (len + 1) + ' frames)');

              // Failsafe check, in case we have missing fragments in the trace

              timestamp := priv_processchain[0].Frame.Timestamp;

              // Check if the oldest frame is more than 1 second older than the current frame

              if MilliSecondsBetweenFrames (IPFragmentDetails.Frame.Timestamp, timestamp) > priv_params.AdvReassemblyChainTimeoutMS then
                 begin
                      SiMain.LogDebug ('Reassembly wait timer of ' + priv_params.AdvReassemblyChainTimeoutMS.ToString +  ' milliseconds exceeded, processing Fragment Chain (containing '+ IntToStr (len) + ' frames)');

                      ErrorCode := ProcessFragmentChain;

                      // TODO: handle forced fragment chain processing with better edge case (last fragment being incomplete but would have been if waited longer)
                 end;
         end;

     result := ErrorCode;

     SiMain.LeaveMethod ('AddFrameToReassemblyProcessChain');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.ProcessFrame (var OriginalFrame : tRawFrame) : tTaskErrorCode;

var FragTestParser  : tParseMaster;
    IPv4Parser      : tIPv4Parser;
    IPv6Parser      : tIPv6Parser;
    IPv6FragParser  : tIPv6FragmentationParser;
    IPAddress       : tIPAddress;
    IPFragment      : tIPFragment;
    IPID            : longword;
    ErrorCode       : tTaskErrorCode;
    HeaderType      : tHeaderType;
    found           : boolean;
    DiscardOriginal : boolean;
    i               : integer;

begin
//   SiMain.EnterMethod ('tTaskProcessorAnonymize.ProcessFrame');

//   SiMain.LogVerbose ('Frame Index: ' + IntToStr (OriginalFrame.FileFrameID));

     ErrorCode := ecSuccess;
     DiscardOriginal := False;

     // Check if this is a fragment and not sliced (which would mean we can't reassemble)

     try
        FragTestParser := tParseMaster.create (OriginalFrame, ppmUpToLayer3);
     except
           SiMain.LogFatal ('Creating ParseMaster failed (tTaskProcessorAnonymize.ProcessFrame)');

           ErrorCode := ecFailure;
//         SiMain.LeaveMethod ('tTaskProcessorAnonymize.ProcessFrame');

           Exit (ErrorCode);
     end;

     // If yes, enable FragmentMode, which will be turned off as soon as all fragments were handled.
     // The problem here is that there can be multiple interleaved fragmented original packets, also mixed
     // with non-fragments, so we need to keep the order of frames while gathering all parts

     if not priv_FragmentModeActive then
        begin
             // Check if it's a non-truncated fragment and that we are allowed to reassemble

             if (FragTestParser.IsFragment and not FragTestParser.IsTruncated) and not priv_params.AdvDoNotReassembleIPFragments then
                priv_FragmentModeActive := True
             else
                 begin
                      // We're not working on fragments, so we can just process the frame

                      ErrorCode := ProcessSingleFrame (OriginalFrame);

                      // Mark the original frame for removal, we're done with it

                      DiscardOriginal := true;
                 end;
        end;

     if priv_FragmentModeActive then
        begin
             SiMain.LogDebug ('FragmentMode active');

             IPFragment.initialize;

             // Set general values

             IPFragment.IsSliced := FragTestParser.IsTruncated;
             IPFragment.Fileposition := OriginalFrame.FilePosition;
             IPFragment.Frame := OriginalFrame;

             // Gather fragment details

             i := 0;
             found := false;

             repeat
                   inc (i);
                   HeaderType := FragTestParser.getParserTypeByIndex(i);

                   // IPv4 is easy, just one header to look at

                   if (HeaderType = NEXT_HEADER_IPv4) or (HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) then
                      begin
                           IPv4Parser := FragTestParser.getParserByIndex(i) as tIPv4Parser;

                           IPAddress.initialize (IPv4Parser.Source);

                           IPID := IPv4Parser.Identification;
                           IPFragment.Offset := IPv4Parser.Offset;
                           IPFragment.Len := IPv4Parser.IPPayloadLength;
                           IPFragment.Flags := IPv4Parser.Flags;

                           found := True;
                      end;

                   // IPv6 is a two step process, requiring the IPv6 header and the fragmentation extension header

                   if (HeaderType = NEXT_HEADER_IPV6) or (HeaderType = NEXT_HEADER_TYPE_IPV6) then
                      begin
                           IPv6Parser := FragTestParser.getParserByIndex(i) as tIPv6Parser;

                           IPAddress.initialize (IPv6Parser.Source);

                           IPFragment.Len := IPv6Parser.PayloadLength;
                      end;

                   if (HeaderType = NEXT_HEADER_IPV6_FRAGMENTATION) then
                      begin
                           IPv6FragParser := FragTestParser.getParserByIndex(i) as tIPv6FragmentationParser;

                           IPID := IPv6FragParser.Identification;
                           IPFragment.Offset := IPv6FragParser.Offset;
                           IPFragment.Flags := IPv6FragParser.Flags;

                           IPID := IPv6FragParser.Identification;

                           found := True;
                      end;

             until found or (i = FragTestParser.ParserCount);

             // Handle IP Fragments, which means gathering all parts, assemble them to a big frame, sanitize it, and cut/write it
             // plus deal with non-related frames that may show up in between

             SiMain.LogDebug ('Adding frame #' + IntToStr (OriginalFrame.FileFrameID) + ' to Reassembly Process Chain');

             ErrorCode := AddFrameToReassemblyProcessChain (IPAddress, IPID, IPFragment);
        end;

//   SiMain.LogDebug ('Discarding ParseMaster');

     try
        FragTestParser.free;
        FragTestParser := nil;
     except
           SiMain.LogFatal ('Freeing ParseMaster failed (tTaskProcessorAnonymize.ProcessFrame)');
     end;

     if DiscardOriginal then
        begin
             // Drop the original frame, we're done with it

//           SiMain.LogDebug ('Discarding Original Frame');

             DiscardOriginal := False;

             OriginalFrame.free;
             OriginalFrame := nil;
        end;

     result := ErrorCode;

//   SiMain.LeaveMethod ('tTaskProcessorAnonymize.ProcessFrame');
end;

(* ------------------------------------------- *)

function tTaskProcessorAnonymize.Execute : tTaskErrorCode;

var outputfile        : string;
    AbortFile         : boolean;
    fileframesscanned : integer;
    percentagescanned : integer;
    traceloader       : tTraceLoader;
//  tracewriter       : tPCAPNGWriter;
    OriginalFrame     : tRawFrame;
    FrameOffset       : integer;
//  SanitizedFrame    : tRawFrame;
    dHelp             : double;
    iHelp             : integer;
    i                 : integer;
    lLinkType         : longword;
    boAbort           : boolean;
    FirstFrameTime    : tFrametimestamp;
    LastFrameTime     : tFrametimestamp;
    ErrorCode         : tTaskErrorCode;
    chainlen          : integer;
//  InterfaceEntry    : tCaptureInterface;

begin
     SiMain.EnterMethod ('tTaskProcessorAnonymize.Execute');

     priv_ProtocolsSeen := [];                              // tracks what protocols were found in total in all frames
     ErrorCode := ecSuccess;
     percentagescanned := 0;
     fileframesscanned := 0;
     prot_ParseErrorCount := 0;
     priv_FragmentModeActive := False;

     outputfile := prot_OutputLocation.getOutputfilename;

     // Some file security measures

     if LowerCase (outputfile) = LowerCase (prot_Inputfile.FullName) then
        begin
             prot_Inputfile.ProcessingStatus := psTaskAborted;
             ErrorCode := ecAbort;

             Exit (ErrorCode);
        end;

     if FileExists (outputfile) then
        begin
             // default is to overwrite a file

             AbortFile := priv_params.OverwriteExistingTarget;

             if assigned (EventOnTargetFileExists) then
                begin
                     EventOnTargetFileExists (prot_Inputfile, outputfile, AbortFile);

                     if AbortFile then
                        begin
                             prot_Inputfile.ProcessingStatus := psTaskAborted;
                             ErrorCode := ecAbort;

                             exit (ErrorCode);
                        end;
                end;
        end;

     // Get the right trace loader

     case prot_Inputfile.TracefileType of
      cfENC    : traceloader := tENCLoader.create (prot_Inputfile.FullName, priv_params.TracefileListIndex, true);
      cfCAP    : traceloader := tCAPLoader.create (prot_Inputfile.FullName, priv_params.TracefileListIndex, true);
      cfPCAP   : traceloader := tPCAPLoader.create (prot_Inputfile.FullName, priv_params.TracefileListIndex, true);
      cfPCAPng : traceloader := tPCAPNGLoader.create (prot_Inputfile.FullName, priv_params.TracefileListIndex, true);
     end;

     // Update file status and report it if possible

     prot_Inputfile.ProcessingStatus := psTaskProcessing;

     if assigned (EventOnStartProcessingOfFile) then
        EventOnStartProcessingOfFile (prot_Inputfile);

     SiMain.LogVerbose ('Starting sanitization of file "' + prot_Inputfile.FullName + '"');

     // Open the file

     traceloader.openFile;
     traceloader.readTraceFileheader;

     // Create the trace writer

     priv_tracewriter := tPCAPngWriter.create (outputfile, true);
     priv_tracewriter.TimeStampMode := prot_OutputLocation.TimestampMode;
     priv_tracewriter.AbsoluteTimestamp := prot_OutputLocation.OriginalFileTime;

     priv_tracewriter.Open;

     // Handle PCAPng and CAP specialties

     case prot_Inputfile.TracefileType of
      cfENC    : begin
                      priv_tracewriter.HeaderCommentList.Add (TraceWranglerHint);

                      lLinkType := (traceloader as tENCLoader).LinkType;

                      priv_tracewriter.WriteGenericHeader (lLinkType);
                 end;
      cfCAP    : begin
                      priv_tracewriter.HeaderCommentList.Add (TraceWranglerHint);

                      lLinkType := (traceloader as tCAPLoader).LinkType;

                      priv_tracewriter.WriteGenericHeader ((traceloader as tCAPLoader).CaptureInterfaces);
                 end;
      cfPCAP   : begin
                      priv_tracewriter.HeaderCommentList.Add (TraceWranglerHint);

                      lLinkType := (traceloader as tPCAPLoader).LinkType;

                      priv_tracewriter.WriteGenericHeader (lLinkType);
                 end;
      cfPCAPng : begin
                      sanitizePCAPngHeaders (traceloader as tPCAPNGLoader, priv_tracewriter);
                 end;
     end;

     // Read frames

     repeat
           try
              traceloader.readBatchFromDisk;
           except
                 ErrorCode := ecFailure;

                 SiMain.LogFatal ('Error while reading frames from disk');
           end;

           if ErrorCode = ecFailure then
              begin
                   if assigned (onErrorProcessingFrame) then
                      onErrorProcessingFrame (prot_Inputfile, fileframesscanned, ateLoader, priv_AbortTask);

                   if priv_AbortTask then
                      Exit (ecFailure);
              end;

           for i := 1 to traceloader.FramesAvailable do
               begin
//                  priv_StopWatch := TStopWatch.StartNew;

                    priv_AbortTask := false;
                    OriginalFrame := traceloader.getRawFrame(i);
                    FrameOffset := OriginalFrame.FileOffset;

                    // Record first and last frame time in case we want to use them for setting the file date later

                    if OriginalFrame.FileFrameID = 1 then
                       FirstFrameTime := OriginalFrame.Timestamp;

                    LastFrameTime := OriginalFrame.Timestamp;

                    // Sanitize the frame

                    try
                       ErrorCode := ProcessFrame (OriginalFrame);
                    except
                          SiMain.LogFatal ('Failed to sanitize frame #' + OriginalFrame.FileFrameID.ToString);
                    end;

                    // Check if there was an error

                    if ErrorCode = ecFailure then
                       begin
                            SiMain.LogError ('Failed to process frame #' + OriginalFrame.FileFrameID.ToString);

                            inc (prot_ParseErrorCount);

                            if assigned (onErrorProcessingFrame) then
                               onErrorProcessingFrame (prot_Inputfile, OriginalFrame.FileFrameID, ateWriter, boAbort);

                            if boAbort then
                               Exit (ErrorCode);
                       end;

                    inc (fileframesscanned);

                    // Trigger report if possible

                    ReportProgress (FrameOffset);
{
                    priv_elapsed := priv_stopWatch.Elapsed;

                    SiMain.LogVerbose ('Elapsed processing time: ' + IntToStr (priv_elapsed.Milliseconds));
}
               end;

     until traceloader.EOF or priv_AbortTask;

     SiMain.LogVerbose ('All frames read and sanitized, with protocols [' + ParsedProtocolsToString (priv_ProtocolsSeen) + ']');

     // Flush fragment chain if required

     chainlen := Length (priv_processchain);

     if chainlen > 0 then
        try
           processFragmentChain;
        except
              SiMain.LogFatal ('Processing fragment processchain failed');
        end;

     // Cleanup loader and writer

     SiMain.LogDebug ('Cleaning up');

     priv_tracewriter.Flush;
     priv_tracewriter.Close;
     priv_tracewriter.free;

     traceloader.closeFile;
     traceloader.free;

     SiMain.LeaveMethod ('tTaskProcessorAnonymize.Execute');
end;

(* ------------------------------------------- *)

end.
