{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit form_task_edit;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls,
  Vcl.ComCtrls,

  class_taskprocessor_edit;

type
  TFormTaskEdit = class(TForm)
    TreeViewNavigation: TTreeView;
    ButtonOkay: TButton;
    ButtonCancel: TButton;
    PageControlEditSettings: TPageControl;
    TabSheetPacketList: TTabSheet;
    TabSheetRemoveLayers: TTabSheet;
    TabSheetOutput: TTabSheet;
    GroupBoxOutputLocation: TGroupBox;
    RadioButtonFileOutputSameDirectory: TRadioButton;
    RadioButtonFileOutputToSubdirectory: TRadioButton;
    LabeledEditOutputToSubdirectory: TLabeledEdit;
    RadioButtonFileOutputToAbsoluteDirectory: TRadioButton;
    LabeledEditOutputToAbsoluteDirectory: TLabeledEdit;
    ButtonSelectOutputDirectory: TButton;
    GroupBoxFileOutputNamingOptions: TGroupBox;
    LabelOutputFilename: TLabel;
    EditOutputFilename: TEdit;
    MemoOutputHint: TMemo;
    GroupBoxPacketAction: TGroupBox;
    RadioButtonPacketActionNone: TRadioButton;
    RadioButtonEditPacketListActive: TRadioButton;
    MemoPacketActionHint: TMemo;
    GroupBoxActionPackets: TGroupBox;
    CheckBoxReorderPacketsByAbsoluteTime: TCheckBox;
    CheckBoxDeduplicatePackets: TCheckBox;
    LabelFrameDeduplicationBuffer: TLabel;
    EditPacketListDeduplicateFrames: TEdit;
    LabelFrameDeduplicationBufferUnit: TLabel;
    LabelFrameDeduplicationBufferTimeLimit: TLabel;
    EditFrameHistoryMsec: TEdit;
    LabelFrameDeduplicationTimeLimitUnit: TLabel;
    CheckBoxThoroughSearch: TCheckBox;
    GroupBoxRemoveLayers: TGroupBox;
    CheckBoxRemoveJuniperHeader: TCheckBox;
    CheckBoxRemoveGTPUHeader: TCheckBox;
    CheckBoxRemoveGREHeader: TCheckBox;
    CheckBoxAddSHBComment: TCheckBox;
    TabSheetEditLayers: TTabSheet;
    GroupBoxEditLayers: TGroupBox;
    CheckBoxReplaceLinuxCookedHeaderWithEthernet: TCheckBox;
    CheckBoxAddJuniperL2Header: TCheckBox;
    CheckBoxSetFileDateTo: TCheckBox;
    ComboBoxSetFileTimestampScope: TComboBox;
    CheckBoxAddJuniperL2Comment: TCheckBox;
    CheckBoxSLLReplaceEtherType: TCheckBox;
    LabelOriginalEthertype: TLabel;
    LabelReplacementEtherType: TLabel;
    EditOriginalEtherType: TEdit;
    EditReplacementEtherType: TEdit;
    CheckBoxRemoveMPLSShims: TCheckBox;
    CheckBoxFixFrameSizeMeta: TCheckBox;
    CheckBoxFixFrameSizeMetaNonHardSliced: TCheckBox;
    CheckBoxRemoveCAPWAP: TCheckBox;
    LabelSLLMACAddress: TLabel;
    EditSLLMACAddress: TEdit;
    LabelSLLReplaceEtherTypeHint: TLabel;
    CheckboxForceSLLMACReplacement: TCheckBox;
    CheckBoxRemoveVLANs: TCheckBox;
    CheckBoxFixBrokenEthertypes: TCheckBox;
    TabSheetPCAPng: TTabSheet;
    GroupBoxPCAPngIDB: TGroupBox;
    CheckBoxModifyPCAPngIDB: TCheckBox;
    EditPCAPngIDBInterfaceName: TEdit;
    procedure onChangeTreeViewNavigation(Sender: TObject; Node: TTreeNode);
    procedure onShow(Sender: TObject);
    procedure onKeyPressAllowNumbersOnly(Sender: TObject; var Key: Char);
    procedure onChangePacketListSettings(Sender: TObject);
    procedure ButtonSelectOutputDirectoryClick(Sender: TObject);
    procedure onChangeOutputSettings(Sender: TObject);
    procedure onChangeRemoveLayersSettings(Sender: TObject);
    procedure onChangeEditLayerSettings(Sender: TObject);
    procedure onKeyPressEthertypes(Sender: TObject; var Key: Char);
    procedure onExitEditSSLMACAddress(Sender: TObject);
    procedure onKeyPressEditSLLMACAddress(Sender: TObject; var Key: Char);
    procedure onCreate(Sender: TObject);
  private
    { Private declarations }

         function BrowseDialog (const Title: string; const Flag: integer): string;

  public
    { Public declarations }

        procedure LoadSettings(const Filename: string);
        procedure SaveSettings(const Filename: string);

        {$REGION 'Documentation'}
        ///	<summary>
        ///	  This function returns the settings made on the edit task form as
        ///	  a parameter record which is then used by the task processor
        ///	</summary>
        {$ENDREGION}
        function getParameters : tEditParameters;
  end;

var
  FormTaskEdit: TFormTaskEdit;

const TaskTypeEdit = 'TaskEdit';

implementation

{$R *.dfm}

uses ShlObj,
     Toolkit,
     base_structures,
     class_formstorage;

(* ------------------------------------------ *)

function TFormTaskEdit.BrowseDialog (const Title: string; const Flag: integer): string;

var lpItemID    : PItemIDList;
    BrowseInfo  : TBrowseInfo;
    DisplayName : array[0..MAX_PATH] of char;
    TempPath    : array[0..MAX_PATH] of char;

begin
     Result:='';
     FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);

     with BrowseInfo do
          begin
               hwndOwner := Application.Handle;
               pszDisplayName := @DisplayName;
               lpszTitle := PChar(Title);
               ulFlags := Flag;
          end;

  lpItemID := SHBrowseForFolder(BrowseInfo);

  if lpItemId <> nil then
     begin
          SHGetPathFromIDList(lpItemID, TempPath);
          Result := TempPath;
          GlobalFreePtr(lpItemID);
     end;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onShow(Sender: TObject);

var i : integer;

begin
     TreeViewNavigation.FullExpand;
     PageControlEditSettings.ActivePageIndex := 0;

     for i := 1 to PageControlEditSettings.PageCount do
         PageControlEditSettings.Pages[i-1].TabVisible := false;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onChangeEditLayerSettings(Sender: TObject);

begin
     CheckBoxSLLReplaceEtherType.Enabled := CheckBoxReplaceLinuxCookedHeaderWithEthernet.Checked;

     LabelOriginalEthertype.Enabled := CheckBoxSLLReplaceEtherType.Enabled and CheckBoxSLLReplaceEtherType.Checked;
     EditOriginalEthertype.Enabled := LabelOriginalEthertype.Enabled;
     LabelReplacementEtherType.Enabled := LabelOriginalEthertype.Enabled;
     EditReplacementEtherType.Enabled := LabelOriginalEthertype.Enabled;
     LabelSLLReplaceEtherTypeHint.Enabled := EditOriginalEtherType.Enabled;

     LabelSLLMACAddress.Enabled := CheckBoxReplaceLinuxCookedHeaderWithEthernet.Checked;
     EditSLLMACAddress.Enabled := LabelSLLMACAddress.Enabled;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onChangeOutputSettings(Sender: TObject);

begin
     LabeledEditOutputToSubdirectory.Enabled := RadioButtonFileOutputToSubdirectory.Enabled and RadioButtonFileOutputToSubdirectory.Checked;
     LabeledEditOutputToAbsoluteDirectory.Enabled := RadioButtonFileOutputToAbsoluteDirectory.Enabled and RadioButtonFileOutputToAbsoluteDirectory.Checked;
     ButtonSelectOutputDirectory.Enabled := LabeledEditOutputToAbsoluteDirectory.Enabled;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onChangePacketListSettings(Sender: TObject);

begin
     GroupBoxActionPackets.Enabled := RadioButtonEditPacketListActive.Checked;

     CheckBoxReorderPacketsByAbsoluteTime.Enabled := GroupBoxActionPackets.Enabled;
//   CheckBoxDeduplicatePackets.Enabled := GroupBoxActionPackets.Enabled;

     CheckBoxFixFrameSizeMeta.Enabled := GroupBoxActionPackets.Enabled;
     CheckBoxFixFrameSizeMetaNonHardSliced.Enabled := CheckBoxFixFrameSizeMeta.Enabled and CheckBoxFixFrameSizeMeta.Checked;

     LabelFrameDeduplicationBuffer.Enabled := CheckBoxDeduplicatePackets.Enabled and CheckBoxDeduplicatePackets.Checked;
     EditPacketListDeduplicateFrames.Enabled := LabelFrameDeduplicationBuffer.Enabled;
     LabelFrameDeduplicationBufferUnit.Enabled := EditPacketListDeduplicateFrames.Enabled;

     LabelFrameDeduplicationBufferTimeLimit.Enabled := CheckBoxDeduplicatePackets.Enabled and CheckBoxDeduplicatePackets.Checked;
     EditFrameHistoryMsec.Enabled := LabelFrameDeduplicationBufferTimeLimit.Enabled;
     LabelFrameDeduplicationTimeLimitUnit.Enabled := EditFrameHistoryMsec.Enabled;

     CheckBoxThoroughSearch.Enabled := CheckBoxDeduplicatePackets.Enabled and CheckBoxDeduplicatePackets.Checked;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onChangeRemoveLayersSettings(Sender: TObject);

begin
     CheckBoxAddJuniperL2Header.Enabled := CheckBoxRemoveJuniperHeader.Checked;
     CheckBoxAddJuniperL2Comment.Enabled := CheckBoxAddJuniperL2Header.Enabled and CheckBoxAddJuniperL2Header.Checked
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onChangeTreeViewNavigation(Sender: TObject; Node: TTreeNode);

var PageIndex : integer;

begin
     PageIndex := -1;

     case TreeViewNavigation.Selected.AbsoluteIndex of
      0,1  : PageIndex := TreeViewNavigation.Selected.AbsoluteIndex;
      2    : PageIndex := 1;
      3    : PageIndex := 2;
      4    : PageIndex := 3;
      5    : PageIndex := 4;
     end;

     if PageIndex > -1 then
        PageControlEditSettings.ActivePageIndex := PageIndex;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onCreate(Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onExitEditSSLMACAddress(Sender: TObject);

begin
     if isValidMAC (EditSLLMACAddress.Text) then
        EditSLLMACAddress.Font.Color := clWindowText
     else
         EditSLLMACAddress.Font.Color := clRed;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onKeyPressAllowNumbersOnly(Sender: TObject;  var Key: Char);

begin
     if not (key in [#8, #9, '0' .. '9']) then
        key:=#0
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onKeyPressEditSLLMACAddress(Sender: TObject; var Key: Char);

begin
     if not (key in [#8, #9, '0'..'9','a'..'f','A'..'F', ':']) then
        key:=#0;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.onKeyPressEthertypes(Sender: TObject; var Key: Char);

begin
     // allow hex (x as indicator)

     if not (key in [#8, #9, '0' .. '9', 'a' .. 'f', 'x']) then
        key:=#0
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.LoadSettings (const Filename: string);

var StorMan : tFormStorageManager;

begin
     StorMan := tFormStorageManager.create(self);

     StorMan.retrieveFormFromDB(Filename);
     StorMan.free;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.SaveSettings (const Filename: string);

var StorMan : tFormStorageManager;

begin
     StorMan := tFormStorageManager.create (self);

     StorMan.storeFormToDB (Filename, TaskTypeEdit);
     StorMan.free;
end;

(* ------------------------------------------- *)

procedure TFormTaskEdit.ButtonSelectOutputDirectoryClick(Sender: TObject);

var sPath : string;

begin
     sPath := BrowseDialog ('Select Path', BIF_RETURNONLYFSDIRS);

     if DirectoryExists (sPath) then
        LabeledEditOutputToAbsoluteDirectory.Text := sPath;
end;

(* ------------------------------------------- *)

function TFormTaskEdit.getParameters: tEditParameters;

var EditParams : tEditParameters;
    help       : string;

begin
     EditParams.ReorderFrames := RadioButtonEditPacketListActive.Checked and CheckBoxReorderPacketsByAbsoluteTime.Checked;

     EditParams.ReplaceCookedHdrWithEthernet := CheckBoxReplaceLinuxCookedHeaderWithEthernet.Checked;
     EditParams.ReplaceEtherType := CheckBoxSLLReplaceEtherType.Checked;

     EditParams.ReplacementSLLMACAddress.setfromString (EditSLLMACAddress.Text, ':');
     EditParams.ForceSLLMACReplacement := CheckboxForceSLLMACReplacement.checked;

     if EditParams.ReplaceEtherType then
        begin
             help := EditOriginalEtherType.Text;

             if pos ('x', help) > 0 then
                help := '$' + rightstring (help, 'x');

             tryStrToInt (help, EditParams.OriginalEtherType);

             help := EditReplacementEtherType.Text;

             if pos ('x', help) > 0 then
                help := '$' + rightstring (help, 'x');

             tryStrToInt (help, EditParams.ReplacementEtherType);
        end
     else
         begin
              EditParams.OriginalEtherType := 0;
              EditParams.ReplacementEtherType := 0;
         end;

     EditParams.RemoveJuniperHeader := CheckBoxRemoveJuniperHeader.Checked;

     if CheckBoxAddJuniperL2Header.Checked and CheckBoxAddJuniperL2Header.Enabled then
        begin
             if CheckBoxAddJuniperL2Comment.Checked then
                EditParams.AddPseudoL2HeaderForJuniper := plmWithFrameComment
             else
                 EditParams.AddPseudoL2HeaderForJuniper := plmSilent;
        end
     else
         EditParams.AddPseudoL2HeaderForJuniper := plmNone;

     EditParams.RemoveVLANs := CheckboxRemoveVLANs.Checked;
     EditParams.RemoveGREHeader := CheckBoxRemoveGREHeader.Checked;
     EditParams.RemoveGTPUHeader := CheckBoxRemoveGTPUHeader.Checked;
     EditParams.RemoveMPLSShims := CheckBoxRemoveMPLSShims.Checked;
     EditParams.RemoveCAPWAP := CheckboxRemoveCAPWAP.Checked;
     EditParams.FixFrameSizeMetaData := CheckBoxFixFrameSizeMeta.Checked;
     EditParams.FixFrameSizeMetaNonHardSliced := CheckBoxFixFrameSizeMetaNonHardSliced.Checked;
     EditParams.FixEtherTypes := CheckBoxFixBrokenEthertypes.Checked;
     EditParams.AddSHBComment := CheckBoxAddSHBComment.Checked;

     EditParams.ChangeIDBInterfaceName := CheckBoxModifyPCAPngIDB.Checked;
     EditParams.IDBInterfaceNameValue := EditPCAPngIDBInterfaceName.Text;

     result := EditParams;
end;

(* ------------------------------------------- *)

end.

