{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit form_task_extract;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls,
  Vcl.ComCtrls, Vcl.Menus, VirtualTrees, Generics.Collections,

  base_structures,
  definition_filter_structures,
  class_interfacemapper,
  class_tracefilelist,
  class_interfacelist,
  class_parse_snortalertlog,
  class_parse_csvfilterfile,
  class_conversationstorage,
  class_formstorage_extracttask,
  class_taskprocessor_extract,
  dialog_import_filter_from_file;

type
  TFormTaskExtract = class(TForm)
    TreeViewNavigation: TTreeView;
    PageControlExtractSettings: TPageControl;
    TabSheetOutput: TTabSheet;
    GroupBoxOutputLocation: TGroupBox;
    RadioButtonFileOutputSameDirectory: TRadioButton;
    RadioButtonFileOutputToSubdirectory: TRadioButton;
    LabeledEditOutputToSubdirectory: TLabeledEdit;
    RadioButtonFileOutputToAbsoluteDirectory: TRadioButton;
    LabeledEditOutputToAbsoluteDirectory: TLabeledEdit;
    ButtonSelectOutputDirectory: TButton;
    MemoOutputHint: TMemo;
    GroupBoxFileOutputOptions: TGroupBox;
    LabeledEditFileNamePattern: TEdit;
    CheckBoxSetFileDateTo: TCheckBox;
    ComboBoxSetFileTimestampScope: TComboBox;
    ButtonOkay: TButton;
    ButtonCancel: TButton;
    TabSheetFilterSettings: TTabSheet;
    GroupBoxFilterSettings: TGroupBox;
    VSTFilterList: TVirtualStringTree;
    ButtonAddFilter: TButton;
    ButtonEditFilter: TButton;
    ButtonRemoveFilter: TButton;
    PopupMenuFilterList: TPopupMenu;
    MenuFilterLookup: TMenuItem;
    MenuFilterBeaconFile: TMenuItem;
    DateTimePickerOutputFileTimestamp: TDateTimePicker;
    TabSheetMetadataEnrichment: TTabSheet;
    GroupBoxMetadataEnrichment: TGroupBox;
    VSTMetaData: TVirtualStringTree;
    ButtonAddMetaData: TButton;
    ButtonEditMetaData: TButton;
    ButtonRemoveMetaData: TButton;
    PopupMenuMetaData: TPopupMenu;
    MenuImportFromSnortAlertFile: TMenuItem;
    OpenDialogAlertLog: TOpenDialog;
    MenuFilterSnortAlertLog: TMenuItem;
    MenuFilterFromFile: TMenuItem;
    LabelFilterHint: TLabel;
    ComboBoxFileNamePattern: TComboBox;
    LabelFilename: TLabel;
    PanelFilterMessage: TPanel;
    procedure ButtonSelectOutputDirectoryClick(Sender: TObject);
    procedure onChangeFilenamePattern(Sender: TObject);
    procedure MenuFilterLookupClick(Sender: TObject);
    procedure onCreate(Sender: TObject);
    procedure onShow(Sender: TObject);
    procedure onVSTFilterListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure onVSTFilterListDblClick(Sender: TObject);
    procedure onVSTFilterListChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure onChangeNavigation(Sender: TObject; Node: TTreeNode);
    procedure ButtonAddFilterClick(Sender: TObject);
    procedure ButtonEditFilterClick(Sender: TObject);
    procedure ButtonRemoveFilterClick(Sender: TObject);
    procedure onChangeTimestampScope(Sender: TObject);
    procedure onChangeOutputSettings(Sender: TObject);
    procedure onVSTMetaDataListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure onVSTMetaDataChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure onVSTMetaDataDblClick(Sender: TObject);
    procedure MenuImportFromSnortAlertFileClick(Sender: TObject);
    procedure ButtonEditMetaDataClick(Sender: TObject);
    procedure ButtonAddMetaDataClick(Sender: TObject);
    procedure ButtonRemoveMetaDataClick(Sender: TObject);
    procedure MenuFilterSnortAlertLogClick(Sender: TObject);
    procedure onKeyPressDirectory(Sender: TObject; var Key: Char);
    procedure onKeyPressFilename(Sender: TObject; var Key: Char);
    procedure MenuFilterFromFileClick(Sender: TObject);
    procedure onVSTFilterListCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
    procedure onVSTMetaDataCompareNodes(Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);
    procedure onVSTFilterListHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);
    procedure onVSTMetaDataHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);
    procedure PageControlExtractSettingOnChange(Sender: TObject);
    procedure onDestroy(Sender: TObject);
    procedure onVSTFilterListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onVSTMetaDataKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onKeyUpEdit(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onKeyUpLabeledEdit(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure onVSTFilterListDrawText(Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; const Text: string; const CellRect: TRect; var DefaultDraw: Boolean);
  private
         priv_filelist           : tTracefileList;
         priv_conversationstore  : tConversationStorage;
         priv_filterset          : tFilterSet;
         priv_metadatalist       : tDictionary<string, tMetaDataEntry>;

         priv_importoptions      : tImportOptionSet;

         priv_InvalidFilterCount : integer;

         // Windows API stuff

         procedure WMDROPFILES (var msg : TWMDropFiles) ; message WM_DROPFILES;

         procedure OnBeaconsPercentProcessed (Sender : tObject; const FileSize : uint64; const FiltersAdded : integer; const PercentageParsed : integer);
         procedure OnSnortAlertsPercentProcessed (Sender : tObject; const FileSize : uint64; const AlertsAdded : integer; const PercentageParsed : integer);
         procedure OnAddFilterToSet (Sender : tObject; var Filter : tFrameFilter);

         function IndicatorToStr (const L4Protocol : tLayer4Protocol; const Sequence : longword; const IPID : word) : string;

         procedure AddFiltersFromBeaconFile (Filename : string);
         procedure AddFiltersFromCSVFile (Filename : string);
         procedure AddDataFromSnortAlertFile (Filename : string);
         procedure AddMetaDataFromSnortAlertFile (Filename : string);

//       procedure MergeFilterListToSet (List : tFilterSet);

         procedure OnAddFilterFromList (const Filter : tFrameFilter);

         procedure BoldTreeNode (TreeNode: TTreeNode; SetBold: boolean);
         procedure UpdateFormElements;
         procedure UpdateFilterList;
         procedure UpdateMetaDataList;
         procedure UpdateNavigationTree;

         function MetaDataDictionaryToList : tMetaDataSet;
         /// <summary>
         ///   Creates a copy of the internal list and returns it. Freeing that
         ///   copy needs to be done by the caller of this method.
         /// </summary>
         function getFilterListCopy : tFilterSet;

  public
        property FileList : tTracefileList read priv_filelist write priv_filelist;
        property ConversationStore : tConversationStorage read priv_conversationstore write priv_conversationstore;

        procedure LoadSettings (const Filename: string);
        procedure SaveSettings (const Filename: string);

        function getParameters : tExtractionParameters;
  end;

var
  FormTaskExtract: TFormTaskExtract;

implementation

{$R *.dfm}

uses Shlobj,
     Clipbrd,
     CommCtrl,
     Math,
     ShellAPI,
     System.Types,
     SiAuto,
     tracefile_utils,
     class_beaconfile,
     definition_ip_addresses,
     definition_tcp_structures,
     dialog_conversation_summary,
     dialog_editfilter,
     dialog_editmetadata;

type tFilterListEntry = record
                              Index                     : integer;
                              FilterAsDisplayFilter     : string;
                              FilterLabel               : string;
                              FilterWillNotMatch        : boolean;
                              key                       : string;
                        end;
     pFilterListEntry = ^tFilterListEntry;

     tMetaDataListEntry = record
                                index         : integer;
                                IPVersion     : byte;
                                Address4      : tIPv4AddressPair;
                                Address6      : tIPv6AddressPair;
                                PortA         : word;
                                PortB         : word;
                                Layer4Prot    : tLayer4Protocol;
                                TCPFlags      : tTCPFlagSet;
                                Indicator     : string;
                                Priority      : integer;
                                Comment       : string;
                                metakey       : string;
                          end;
     pMetaDataListEntry = ^tMetaDataListEntry;

const TaskTypeExtract = 'TaskExtract';
      FilterHintMatchAll = '** With no filter defined ALL frames will be extracted **';
      FilterHintNoMatch = '** Grayed out filters will not match any frames **';
      BeaconFileStatusDisplayThreshold = 10000000;  // 10MB
      SnortFileStatusDisplayThreshold = 10000;      // 10k


(* ------------------------------------------ *)

procedure TFormTaskExtract.WMDROPFILES(var msg: TWMDropFiles);

const MAXFILENAME = 255;

var i                : integer;
    FileCount        : integer;
    OSFileName       : array [0..MAXFILENAME] of char;
    Filename         : string;
    boOkay           : boolean;
    captype          : tCaptureFileType;
    sFile            : tSnortAlertParser;
    isValidSnortFile : boolean;

begin
     // how many files dropped?

     fileCount := DragQueryFile(msg.Drop, $FFFFFFFF, OSFileName, MAXFILENAME) ;

     // query for file names

     for i := 0 to fileCount - 1 do
         begin
              DragQueryFile(msg.Drop, i, OSFileName, MAXFILENAME) ;
              Filename := string (OSFileName);

              if FileExists (Filename) and (lowercase (ExtractFileName (filename)) <> LocalTraceIntelFilename) then
                 begin
                      // Set default options for this kind of thing since we don't show the options dialog

                      priv_importoptions.initialize;

                      // Import depending on where we are and what we get

                      case PageControlExtractSettings.ActivePageIndex of
                       0 : begin
                                captype := getTracefileType (Filename);

                                if captype = cfUnknown then
                                   begin
                                        // Check if it is a snort alert file

                                        sFile := tSnortAlertParser.create (Filename);

                                        try
                                           isValidSnortFile := sFile.isValidFile;
                                        finally
                                               sFile.Free;
                                        end;

                                        if isValidSnortFile then
                                           AddDataFromSnortAlertFile (Filename);
                                   end
                                else
                                    begin
                                         if extractFileExt (Filename) = '.csv' then
                                            begin
                                                 // Assume this is a csv file, so lets load it

                                                 priv_importoptions.CSVParseMode := cpmAuto;

                                                 AddFiltersFromCSVFile (Filename);
                                            end
                                         else
                                             begin
                                                  if getTracefileType (Filename) <> cfUnknown then
                                                     AddFiltersFromBeaconFile (Filename);
                                             end;
                                    end;
                           end;
                       1 : begin
                                // Check if it is a snort alert file

                                sFile := tSnortAlertParser.create (Filename);

                                try
                                   isValidSnortFile := sFile.isValidFile;
                                finally
                                       sFile.Free;
                                end;

                                if isValidSnortFile then
                                   AddMetaDataFromSnortAlertFile (Filename);
                           end;
                      end;
                 end;
         end;

     // Release memory

     DragFinish(msg.Drop) ;
end;

(* ------------------------------------------ *)

function BrowseDialog (const Title: string; const Flag: integer): string;

var lpItemID    : PItemIDList;
    BrowseInfo  : TBrowseInfo;
    DisplayName : array[0..MAX_PATH] of char;
    TempPath    : array[0..MAX_PATH] of char;

begin
     Result:='';
     FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);

     with BrowseInfo do
          begin
               hwndOwner := Application.Handle;
               pszDisplayName := @DisplayName;
               lpszTitle := PChar(Title);
               ulFlags := Flag;
          end;

  lpItemID := SHBrowseForFolder(BrowseInfo);

  if lpItemId <> nil then
     begin
          SHGetPathFromIDList(lpItemID, TempPath);
          Result := TempPath;
          GlobalFreePtr(lpItemID);
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onCreate (Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     VSTFilterList.NodeDataSize := sizeOf (tFilterListEntry);
     VSTMetaData.NodeDataSize := sizeOf (tMetaDataListEntry );

     priv_filterset := tFilterSet.create;
     priv_filterset.OnAddFilterToSet := OnAddFilterToSet;

     priv_metadatalist := tDictionary<string, tMetaDataEntry>.create;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onDestroy(Sender: TObject);

begin
     priv_filterset.Free;
     priv_metadatalist.Free;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onShow (Sender: TObject);

var i     : integer;

begin
     UpdateFilterList;
     UpdateMetaDataList;

     for i := 1 to PageControlExtractSettings.PageCount do
         PageControlExtractSettings.Pages[i-1].TabVisible := false;

     PageControlExtractSettings.ActivePageIndex := 0;

     UpdateFormElements;

     // Make sure the first node is active

     TreeViewNavigation.Select(TreeViewNavigation.Items[0]);

     // Declare that we do accept dropped files

     DragAcceptFiles (Handle, true);
end;

(* ------------------------------------------ *)

function TFormTaskExtract.IndicatorToStr (const L4Protocol: tLayer4Protocol; const Sequence: longword; const IPID: word): string;

begin
     case L4Protocol of
      l4TCP  : Result := 'Seq: ' + IntToStr (Sequence);
      l4UDP,
      l4ICMP : Result := 'IPID: ' + IntToStr (IPID);
     else
         Result := 'undefined';
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onKeyPressDirectory (Sender: TObject; var Key: Char);

begin
     if not (key in [#1, #3, #8, #9, #$16, #$1a, ' ', 'a'..'z', 'A'..'Z', '0'..'9', '\', '-', '_', '#', '[', ']', '(', ')', '<', '>']) then
        key:=#0;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onKeyPressFilename (Sender: TObject; var Key: Char);

begin
     // #1 is CTRL-A, #3 is CTRL-C, #8 is tab, #9 is backspace, #$16 is CTRL-V, #$1a is CTRL-Z

     if not (key in [#1, #3, #8, #9, #$16, #$1a, ' ', 'a'..'z', 'A'..'Z', '0'..'9', '-', '_', '#', '[', ']', '(', ')', '<', '>']) then
        Key := #0;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onKeyUpEdit (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if ssCtrl in Shift then
        begin
             case key of
              65 : (Sender as tEdit).SelectAll;
              67 : Clipboard.AsText := (Sender as tEdit).Text;
              86 : (Sender as tEdit).Text := Clipboard.AsText;
             end;
        end
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onKeyUpLabeledEdit (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if ssCtrl in Shift then
        begin
             case key of
              65 : (Sender as tLabeledEdit).SelectAll;
              67 : Clipboard.AsText := (Sender as tLabeledEdit).Text;
              86 : (Sender as tLabeledEdit).Text := Clipboard.AsText;
             end;
        end
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListChange (Sender: TBaseVirtualTree; Node: PVirtualNode);

begin
     if VSTFilterList.SelectedCount > 0 then
        begin
             ButtonEditFilter.Enabled := (VSTFilterList.SelectedCount = 1);

             ButtonRemoveFilter.Enabled := true;
        end
     else
         begin
             ButtonEditFilter.Enabled := false;
             ButtonRemoveFilter.Enabled := false;
         end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1 : pFilterListEntry;
    data2 : pFilterListEntry;

begin
     data1 := VSTFilterList.GetNodeData (node1);
     data2 := VSTFilterList.GetNodeData (node2);

     case column of
      0 : Result := compareValue (data1^.Index, data2^.Index);
      1 : Result := compareText (data1^.FilterAsDisplayFilter, data2^.FilterAsDisplayFilter);
      2 : Result := compareText (data1^.FilterLabel, data2^.FilterLabel);
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataChange(Sender: TBaseVirtualTree; Node: PVirtualNode);

begin
     if VSTMetaData.SelectedCount > 0 then
        begin
             ButtonEditMetaData.Enabled := (VSTMetaData.SelectedCount = 1);

             ButtonRemoveMetaData.Enabled := true;
        end
     else
         begin
             ButtonEditMetaData.Enabled := false;
             ButtonRemoveMetaData.Enabled := false;
         end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataCompareNodes (Sender: TBaseVirtualTree; Node1, Node2: PVirtualNode; Column: TColumnIndex; var Result: Integer);

var data1  : pMetaDataListEntry;
    data2  : pMetaDataListEntry;

begin
     data1 := VSTMetaData.GetNodeData (node1);
     data2 := VSTMetaData.GetNodeData (node2);

     case column of
      0 : Result := compareValue (data1^.index, data2^.index);
      1 : begin
               if data1^.IPVersion = 4 then
                  begin
                       if data2^.IPVersion = 4 then
                          Result := CompareValue (data1^.Address4.Source, data2^.Address4.Source)
                       else
                           Result := GreaterThanValue;
                  end
               else
                   if data2^.IPVersion = 6 then
                      Result := CompareValue (data1^.Address6.Source, data2^.Address6.Source)
                   else
                       Result := LessThanValue;
          end;
      2 : Result := CompareValue (data1^.PortA, data2^.PortA);
      3 : begin
               if data1^.IPVersion = 4 then
                  begin
                       if data2^.IPVersion = 4 then
                          Result := CompareValue (data1^.Address4.Destination, data2^.Address4.Destination)
                       else
                           Result := GreaterThanValue;
                  end
               else
                   if data2^.IPVersion = 6 then
                      Result := CompareValue (data1^.Address6.Destination, data2^.Address6.Destination)
                   else
                       Result := LessThanValue;
          end;
      4 : Result := CompareValue (data1^.PortB, data2^.PortB);
      5 : Result := compareText (Layer4ProtocolToStr (data1^.Layer4Prot), Layer4ProtocolToStr (data2^.Layer4Prot));
      6 : begin
               // TCP flags - only compare them if both nodes are TCP, otherwise compare L4 protocol types

               if (data1^.Layer4Prot = l4TCP) and (data2^.Layer4Prot = l4TCP) then
                  // Result := CompareFlags (data1^.TCPFlags, data2^.TCPFlags)  this doesn't seem to work; results are split for some reason
                  Result := CompareText (TCPFlagSetToTCPDump (data1^.TCPFlags), TCPFlagSetToTCPDump (data2^.TCPFlags))
               else
                   compareText (Layer4ProtocolToStr (data1^.Layer4Prot), Layer4ProtocolToStr (data2^.Layer4Prot));
          end;
      7 : Result := compareText (data1^.Indicator, data2^.Indicator);
      8 : begin
               if data1^.Priority = data2^.Priority then
                  Result := CompareText (data1^.Comment, data2^.Comment)
               else
                   Result := CompareValue (data1^.Priority, data2^.Priority);
          end;
      9 : Result := compareText (data1^.Comment, data2^.Comment);
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataDblClick(Sender: TObject);

begin
     if VSTMetaData.SelectedCount > 0 then
        ButtonEditMetaData.Click;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);

var node : pVirtualNode;

begin
     node := Sender.Treeview.GetFirstSelected(false);

     Sender.Treeview.SortTree (HitInfo.Column, Sender.SortDirection, true);

     // Since we alternate between ascending and descending we need to reverse the setting for next time

     if Sender.sortDirection = sdAscending then
        Sender.SortDirection := sdDescending
     else
         Sender.SortDirection := sdAscending;

     if node <> nil then
        Sender.Treeview.ScrollIntoView (node, true);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataKeyUp (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if Key = 46 then
        ButtonRemoveMetaDataClick (Sender)
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListDblClick (Sender: TObject);

begin
     if VSTFilterList.SelectedCount > 0 then
        ButtonEditFilter.Click;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var nodedata : pFilterListEntry;

begin
     nodedata := VSTFilterList.GetNodeData (Node);

     case column of
      0 : CellText := IntToStr (nodedata^.Index);
      1 : CellText := nodedata^.FilterAsDisplayFilter;
      2 : CellText := nodedata^.FilterLabel;
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListDrawText (Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; const Text: string; const CellRect: TRect; var DefaultDraw: Boolean);

var nodedata : pFilterListEntry;

begin
     nodedata := VSTFilterList.GetNodeData (Node);

     if nodedata^.FilterWillNotMatch then
        TargetCanvas.Font.Color := clGray;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListHeaderClick(Sender: TVTHeader; HitInfo: TVTHeaderHitInfo);

var node : pVirtualNode;

begin
     node := Sender.Treeview.GetFirstSelected(false);

     Sender.Treeview.SortTree (HitInfo.Column, Sender.SortDirection, true);

     // Since we alternate between ascending and descending we need to reverse the setting for next time

     if Sender.sortDirection = sdAscending then
        Sender.SortDirection := sdDescending
     else
         Sender.SortDirection := sdAscending;

     if node <> nil then
        Sender.Treeview.ScrollIntoView (node, true);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTFilterListKeyUp (Sender: TObject; var Key: Word; Shift: TShiftState);

begin
     if Key = 46 then
        ButtonRemoveFilterClick (Sender);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onVSTMetaDataListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var nodedata : pMetaDataListEntry;
    entry    : tMetaDataEntry;
    key      : string;

begin
     nodedata := Sender.GetNodeData (Node);

     case column of
      0 : CellText := IntToStr (nodedata^.index);
      1 : begin
               if nodedata^.IPVersion = 4 then
                  CellText := nodedata^.Address4.Source.ValueAsString
               else
                   CellText := nodedata^.Address6.Source.ValueAsString;
          end;
      2 : CellText := IntToStr (nodedata^.PortA);
      3 : begin
               if nodedata^.IPVersion = 4 then
                  CellText := nodedata^.Address4.Destination.ValueAsString
               else
                   CellText := nodedata^.Address6.Destination.ValueAsString;
          end;
      4 : CellText := IntToStr (nodedata^.PortB);
      5 : CellText := Layer4ProtocolToStr (nodedata^.Layer4Prot);
      6 : begin
               if nodedata^.Layer4Prot = l4TCP then
                  CellText := TCPFlagSetToTCPDump (nodedata^.TCPFlags)
               else
                   CellText := '';
          end;
      7 : CellText := nodedata^.Indicator;
      8 : CellText := IntToStr (nodedata^.Priority);
      9 : CellText := nodedata^.Comment;
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.PageControlExtractSettingOnChange (Sender: TObject);

var boAccept : boolean;

begin
     boAccept := (PageControlExtractSettings.ActivePageIndex < 2);

     // Declare that we do accept dropped files

     DragAcceptFiles (Handle, boAccept);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.BoldTreeNode (TreeNode: TTreeNode; SetBold: boolean);

var treeItem: TTVItem;

begin
     if not Assigned (treeNode) then
        Exit;

     with treeItem do
          begin
               hItem := treeNode.ItemId;
               stateMask := TVIS_BOLD;
               mask := TVIF_HANDLE or TVIF_STATE;

               if SetBold then
                  state := TVIS_BOLD
               else
                   state := 0;
          end;

     TreeView_SetItem(treeNode.Handle, treeItem) ;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.UpdateNavigationTree;

begin
     // Update navigation tree

     if priv_filterset.Count > 0 then
        begin
             TreeViewNavigation.Items[0].Text := 'Flow Filters (' + IntToStr (priv_filterset.Count) + ')';
             BoldTreeNode (TreeViewNavigation.Items[0], true);

             if priv_InvalidFilterCount > 0 then
                begin
                     LabelFilterHint.Caption := FilterHintNoMatch;
                     LabelFilterHint.Visible := true;
                end
             else
                 LabelFilterHint.Visible := false;
        end
     else
         begin
              TreeViewNavigation.Items[0].Text := 'Flow Filters';
              BoldTreeNode (TreeViewNavigation.Items[0], false);

              LabelFilterHint.Caption := FilterHintMatchAll;
              LabelFilterHint.Visible := True;
         end;

     if priv_metadatalist.Count > 0 then
        begin
             TreeViewNavigation.Items[1].Text := 'Metadata (' + IntToStr (priv_metadatalist.Count) + ')';
             BoldTreeNode (TreeViewNavigation.Items[1], true);
        end
     else
         begin
              TreeViewNavigation.Items[1].Text := 'Metadata';
              BoldTreeNode (TreeViewNavigation.Items[1], false);
         end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.UpdateFilterList;

var Node   : pVirtualNode;
    Data   : pFilterListEntry;
    Filter : tFrameFilter;
    len    : integer;
    i      : Integer;

begin
     VSTFilterList.BeginUpdate;
     VSTFilterList.Clear;

     len := priv_metadatalist.Count;
     i := 0;

     priv_InvalidFilterCount := 0;

     filter := priv_filterset.getFirst;

     while not filter.isUnspecified do
         begin
              inc (i);

              Node := VSTFilterList.AddChild(nil);
              Data := VSTFilterList.GetNodeData (Node);

              Data^.Index := i;
              Data^.FilterAsDisplayFilter := Filter.asDisplayFilter;
              Data^.FilterLabel := Filter.FilterLabel;

              if Filter.HelpFirstSeen.isZero then
                 begin
                      Data^.FilterWillNotMatch := true;
                      inc (priv_InvalidFilterCount);
                 end
              else
                  Data^.FilterWillNotMatch := false;

              Data^.key := Filter.DictionaryKey;

              filter := priv_filterset.getNext;
         end;

     VSTFilterList.EndUpdate;

     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.UpdateMetaDataList;

var Node     : pVirtualNode;
    Data     : pMetaDataListEntry;
    entry    : tMetaDataEntry;
    len      : integer;
    key      : string;
    i        : integer;

begin
     VSTMetaData.BeginUpdate;
     VSTMetaData.Clear;

     len := priv_metadatalist.Count;
     i := 0;

     for key in priv_metadatalist.Keys do
         begin
              inc (i);
              entry := priv_metadatalist.Items[key];

              Node := VSTMetaData.AddChild (nil);
              Data := VSTMetaData.GetNodeData (Node);

              case entry.IPVersion of
               4 : begin
                        Data^.Address4.Source := entry.IPv4SocketPair.Source.IPAddress;
                        Data^.PortA := entry.IPv4SocketPair.Source.Port;
                        Data^.Address4.Destination := entry.IPv4SocketPair.Destination.IPAddress;
                        Data^.PortB := entry.IPv4SocketPair.Destination.Port;
                   end;
               6 : begin
                        Data^.Address6.Source := entry.IPv6SocketPair.Source.IPAddress;
                        Data^.PortA := entry.IPv6SocketPair.Source.Port;
                        Data^.Address6.Destination := entry.IPv6SocketPair.Destination.IPAddress;
                        Data^.PortB := entry.IPv6SocketPair.Destination.Port;
                   end;
              end;

              Data^.index := i;
              Data^.IPVersion := entry.IPVersion;
              Data^.Layer4Prot := entry.Layer4Protocol;
              Data^.TCPFlags := entry.TCPFlags;
              Data^.Indicator := IndicatorToStr (entry.Layer4Protocol, entry.TCPSequenceNumber, entry.IPID);
              Data^.Priority := entry.RuleDetails.Priority;
              Data^.Comment := entry.RuleDetails.Comment;

              Data^.metakey := key;
         end;

     VSTMetaData.EndUpdate;

     // Update navigation tree

     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.UpdateFormElements;

begin
     // Item Settings

     DateTimePickerOutputFileTimestamp.Visible := ComboBoxSetFileTimestampScope.ItemIndex = 3;
     DateTimePickerOutputFileTimestamp.Enabled := ComboBoxSetFileTimestampScope.Enabled;

     ButtonEditFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);
     ButtonRemoveFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);

     // Output settings

     LabeledEditOutputToSubdirectory.Enabled := RadioButtonFileOutputToSubdirectory.Enabled and RadioButtonFileOutputToSubdirectory.Checked;
     LabeledEditOutputToAbsoluteDirectory.Enabled := RadioButtonFileOutputToAbsoluteDirectory.Enabled and RadioButtonFileOutputToAbsoluteDirectory.Checked;
     ButtonSelectOutputDirectory.Enabled := LabeledEditOutputToAbsoluteDirectory.Enabled;

     ComboBoxSetFileTimestampScope.Enabled := CheckBoxSetFileDateTo.Checked;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonAddFilterClick (Sender: TObject);

var Filter : tFrameFilter;
    len    : integer;
    help   : string;
    key    : string;

begin
     FormEditFilter.clear;
     FormEditFilter.PresetFilterSettings (fsConversation);
     FormEditFilter.showmodal;

     if FormEditFilter.ModalResult = mrOK then
        begin
             Filter := FormEditFilter.getFrameFilter;

             if priv_filterset.AddFilter (Filter) then
                begin
                     // Update filter list and navigation

                     UpdateFilterList;
                     UpdateNavigationTree;
                end;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonEditFilterClick (Sender: TObject);

var filter    : tFrameFilter;
    newfilter : tFrameFilter;
    node      : pVirtualNode;
    data      : pFilterListEntry;

begin
     node := VSTFilterList.GetFirstSelected (false);
     data := VSTFilterList.GetNodeData (node);

     filter := priv_filterset.getFilterByKey (data^.key);

     assert (not filter.isUnspecified, 'Filter not found (TFormTaskExtract.ButtonEditFilterClick)');

     FormEditFilter.PrepareForm (filter);
     FormEditFilter.showmodal;

     if FormEditFilter.ModalResult = mrOK then
        begin
             // Check if something has changed which would mean the key changed

             newfilter := FormEditFilter.getFrameFilter;

             if (filter.DictionaryKey <> newfilter.DictionaryKey) or (filter.FilterLabel <> newfilter.FilterLabel) then
                begin
                     // remove old entry, add new

                     priv_filterset.RemoveFilter (filter.DictionaryKey);
                     priv_filterset.AddFilter (newfilter);

                     // Update the new filter too (the newly added filter will not transfer it's setting back here)

                     newfilter.updateRecType;
                     priv_conversationstore.UpdateFilterHelpers (newfilter);

                     // Update the key of the node

                     Data^.FilterAsDisplayFilter := newfilter.asDisplayFilter;
                     Data^.FilterLabel := newfilter.FilterLabel;
                     Data^.FilterWillNotMatch := newfilter.HelpFirstSeen.isZero;
                     data^.key := newfilter.DictionaryKey;
                end;

             VSTFilterList.RepaintNode (Node);

             UpdateNavigationTree;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonRemoveFilterClick (Sender: TObject);

var node : pVirtualNode;
    last : pVirtualNode;
    data : pFilterListEntry;

begin
     node := VSTFilterList.GetFirstSelected (false);

     while node <> nil do
           begin
                data := VSTFilterList.GetNodeData (node);
//              priv_filterset.Remove (data^.key);
                priv_filterset.RemoveFilter (data^.key);

                last := node;
                node := VSTFilterList.GetNextSelected(node);

                VSTFilterList.DeleteNode(last);
           end;

     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonAddMetaDataClick(Sender: TObject);

var node  : pVirtualNode;
    data  : pMetaDataListEntry;
    entry : tMetaDataEntry;
    key   : string;
    len   : integer;

begin
     FormEditMetaData.Clear;
     FormEditMetaData.ShowModal;

     if FormEditMetaData.ModalResult = mrOK then
        begin
             entry := FormEditMetaData.DataEntry;
             key := entry.DictionaryKey;

             if not priv_metadatalist.ContainsKey (key) then
                begin
                     priv_metadatalist.Add (key, entry);

                     UpdateMetaDataList;
                end;

             UpdateNavigationTree;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonEditMetaDataClick(Sender: TObject);

var node  : pVirtualNode;
    data  : pMetaDataListEntry;
    entry : tMetaDataEntry;
    index : integer;

begin
     node := VSTMetaData.GetFirstSelected (false);
     data := VSTMetaData.GetNodeData (Node);

     entry := priv_metadatalist.Items[data^.metakey];
     FormEditMetaData.DataEntry := entry;

     FormEditMetaData.showmodal;

     if FormEditMetaData.ModalResult = mrOK then
        begin
             // Check if something has changed which would mean the key changed

             if entry.DictionaryKey <> FormEditMetaData.DataEntry.DictionaryKey then
                begin
                     // remove old entry, add new

                     priv_metadatalist.Remove(entry.DictionaryKey);

                     entry := FormEditMetaData.DataEntry;
                     priv_metadatalist.Add (entry.DictionaryKey, entry);

                     // Update the key of the node

                     data^.metakey := entry.DictionaryKey;
                end;

             VSTMetaData.RepaintNode (Node);

             UpdateNavigationTree;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonRemoveMetaDataClick(Sender: TObject);

var node : pVirtualNode;
    last : pVirtualNode;
    data : pMetaDataListEntry;

begin
     node := VSTMetaData.GetFirstSelected (false);

     while node <> nil do
           begin
                data := VSTMetaData.GetNodeData (node);
                priv_metadatalist.Remove (data^.metakey);

                last := node;
                node := VSTMetaData.GetNextSelected(node);

                VSTMetaData.DeleteNode(last);
           end;

     // Update navigation tree

     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.ButtonSelectOutputDirectoryClick (Sender: TObject);

var sPath : string;

begin
     sPath := BrowseDialog ('Select Path', BIF_RETURNONLYFSDIRS);

     LabeledEditOutputToAbsoluteDirectory.Text := includetrailingpathdelimiter (sPath);
end;

(* ------------------------------------------ *)
{
procedure TFormTaskExtract.MergeFilterListToSet (List : tFilterSet);

var filter      : tFrameFilter;
    key         : string;
    i           : integer;

begin
     for i := 1 to length (List) do
         begin
              filter := List[i-1];
              key := filter.DictionaryKey;

              // This will merge the new filter into the actual filter list
              // Note that filters that are just having different labels will
              // be considered identical - so only the first filter will be in
              // the final list

              if not priv_filterset.ContainsKey (key) then
                 priv_filterset.Add (key, filter);
         end;
end;
}
(* ------------------------------------------ *)

procedure TFormTaskExtract.OnBeaconsPercentProcessed (Sender: tObject; const FileSize : uint64; const FiltersAdded : integer; const PercentageParsed: integer);

begin
     if Filesize > BeaconFileStatusDisplayThreshold then
        begin
             PanelFilterMessage.Left := VSTFilterList.Left + 100;
             PanelFilterMessage.Width := VSTFilterList.Width - 200;
             PanelFilterMessage.Top := VSTFilterList.Top + ((VSTFilterList.Height - PanelFilterMessage.Height) DIV 2);
             PanelFilterMessage.Visible := true;

             PanelFilterMessage.Caption := 'Processed: ' + PercentageParsed.ToString + '%, ' + FiltersAdded.ToString + ' filters created';
             Application.ProcessMessages;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.OnSnortAlertsPercentProcessed (Sender : tObject; const FileSize : uint64; const AlertsAdded : integer; const PercentageParsed : integer);

begin
     if Filesize > SnortFileStatusDisplayThreshold then
        begin
             PanelFilterMessage.Left := VSTFilterList.Left + 100;
             PanelFilterMessage.Width := VSTFilterList.Width - 200;
             PanelFilterMessage.Top := VSTFilterList.Top + ((VSTFilterList.Height - PanelFilterMessage.Height) DIV 2);

             PanelFilterMessage.Caption := 'Processed: ' + PercentageParsed.ToString + '%, ' + AlertsAdded.ToString + ' alerts added';
             Application.ProcessMessages;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.AddFiltersFromBeaconFile (Filename: string);

var beaconfile  : tBeaconFile;
    filterset   : tFilterSet;

begin
     try
        beaconfile := tBeaconFile.create (FileName);
        beaconfile.OnBeaconsPercentProcessed := OnBeaconsPercentProcessed;

        beaconfile.IgnoreSYN := priv_importoptions.IgnoreSYN;
        beaconfile.IgnoreICMP := priv_importoptions.IgnoreICMP;

        beaconfile.DetermineFilterList;

        filterset := beaconfile.FilterList;

        if PanelFilterMessage.Visible then
           begin
                PanelFilterMessage.Caption := 'Processed: 100%, ' + filterset.ToString + ' filters created, updating display...';
                Application.ProcessMessages;
           end;

        // add the new filters to the current set (the beacon filterlist will be freed when the beaconfile is freed) and display them if any new filters were added

        if priv_filterset.AddFilterSet (filterset) > 0 then
           UpdateFilterList;

     finally
            beaconfile.Free;
     end;

     PanelFilterMessage.Visible := false;
     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.AddFiltersFromCSVFile (Filename: string);

var CSVFilterFile  : tCSVFilterFile;
    filterset      : tFilterSet;

begin
     try
        CSVFilterFile := tCSVFilterFile.create (Filename, priv_importoptions.CSVParseMode, priv_importoptions.CSVSeparator);
        CSVFilterFile.execute;

        filterset := CSVFilterFile.FilterList;

        // add the new filters to the current set (the CSVFile filterlist will be freed when the beaconfile is freed), and display them if any new filters were added

//      MergeFilterListToSet (filterset);

        if priv_filterset.AddFilterSet (filterset) > 0 then
           UpdateFilterList;

     finally
            CSVFilterFile.Free;
     end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.AddDataFromSnortAlertFile (Filename: string);

var SnortParser : tSnortAlertParser;
    filterset   : tFilterSet;
    entry       : tMetaDataEntry;
    filelist    : tStringlist;
    key         : string;
    rulefile    : string;
    existlen    : integer;
    len         : integer;
    i           : Integer;

begin
     try
        SnortParser := tSnortAlertParser.create (FileName);
        SnortParser.IgnoreSYN := priv_importoptions.IgnoreSYN;
        SnortParser.IgnoreICMP := priv_importoptions.IgnoreICMP;
        SnortParser.LabelTemplate := priv_importoptions.FilterLabel;
        SnortParser.OnAdvanceAlertPercentProcessed := OnSnortAlertsPercentProcessed;

        PanelFilterMessage.Visible := true;

        if priv_ImportOptions.SnortRuleFiles <> '' then
           begin
                filelist := tStringlist.Create;
                filelist.CommaText := priv_ImportOptions.SnortRuleFiles;

                for i := 1 to filelist.Count do
                    begin
                         rulefile := filelist[i-1];

                         if fileexists (rulefile) then
                            SnortParser.LoadRuleDefinitions (rulefile);
                    end;

                filelist.Free;
           end;

        SnortParser.execute;

        if PanelFilterMessage.Visible then
           begin
                PanelFilterMessage.Caption := 'Processed: 100%, ' + length (SnortParser.AlertList).ToString + ' alerts created, updating display...';
                Application.ProcessMessages;
           end;

        // Get list of filters from the snort parser

        filterset := SnortParser.GetFilterList;

        // add the new filters to the current set and display them if any new were added

        if priv_filterset.AddFilterSet (filterset) > 0 then
           UpdateFilterList;

        // Discard the list from the snortparser

        filterset.Free;

        // Add metadata if option was selected

        if priv_importoptions.AutoImportMetaData then
           begin
//              mergeMetaSetDataToSet (SnortParser.MetaDataList, priv_metadatalist);

                existlen := priv_metadatalist.Count;
                len := length (SnortParser.MetaDataList);

                for i := 1 to len do
                    begin
                         entry := SnortParser.MetaDataList[i-1];
                         key := entry.DictionaryKey;

                         if not priv_metadatalist.ContainsKey (key) then
                            priv_metadatalist.Add (key, entry)
                    end;

                // Display metadata

                UpdateMetaDataList;
           end;
     finally
            SnortParser.Free;
     end;

     PanelFilterMessage.Visible := false;
     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.AddMetaDataFromSnortAlertFile (Filename: string);

var SnortParser : tSnortAlertParser;
    key         : string;
    existlen    : integer;
    len         : integer;
    entry       : tMetaDataEntry;
    i           : integer;

begin
     try
        SnortParser := tSnortAlertParser.create (FileName);
        SnortParser.execute;

//      mergeMetaSetDataToSet (SnortParser.MetaDataList, priv_metadatalist);

        existlen := priv_metadatalist.Count;
        len := length (SnortParser.MetaDataList);

        for i := 1 to len do
            begin
                 entry := SnortParser.MetaDataList[i-1];
                 key := entry.DictionaryKey;

                 if not priv_metadatalist.ContainsKey (key) then
                    priv_metadatalist.Add (key, entry)
            end;

        UpdateMetaDataList;

     finally
            SnortParser.Free;
     end;

     UpdateNavigationTree;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.MenuFilterFromFileClick(Sender: TObject);

var Filename    : string;

begin
     FormImportFilterFromFile.showModal;

     if FormImportFilterFromFile.ModalResult = mrOK then
        begin
             Filename := FormImportFilterFromFile.EditFilename.Text;

             priv_importoptions := FormImportFilterFromFile.OptionSet;

             // Check which kind of import we have, and do it

             case FormImportFilterFromFile.ImportFileType of
              iftBeacon     : AddFiltersFromBeaconFile (Filename);
              iftSnortAlert : AddDataFromSnortAlertFile (Filename);
              iftCSV        : AddFiltersFromCSVFile (Filename);
             end;
        end;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.MenuFilterLookupClick (Sender: TObject);

begin
     FormConversationSummary.ConversationStore := priv_conversationstore;
     FormConversationSummary.TracefileList := priv_filelist;
     FormConversationSummary.SetMode (cmmExtractForm);
     FormConversationSummary.onAddFilterToList := OnAddFilterFromList;
     FormConversationSummary.ShowModal;

     FormConversationSummary.onAddFilterToList := nil;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.MenuFilterSnortAlertLogClick(Sender: TObject);

var SnortParser : tSnortAlertParser;
    filterset   : tFilterSet;
    len         : integer;
    i           : integer;

begin
{
     if OpenDialogAlertLog.Execute then
        begin
             SnortParser := tSnortAlertParser.create (OpenDialogAlertLog.FileName);
             SnortParser.execute;
             SnortParser.DetermineFilterList;

             filterset := SnortParser.FilterList;

             // add the new filters to the current set

             mergeFilterSetToSet (filterset, priv_filterset);

             // Display the new filters

             UpdateFilterList;

             SnortParser.Free;
        end;
}
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.MenuImportFromSnortAlertFileClick (Sender: TObject);

begin
     if OpenDialogAlertLog.Execute then
        AddMetaDataFromSnortAlertFile (OpenDialogAlertLog.FileName);
end;

(* ------------------------------------------ *)

function TFormTaskExtract.getFilterListCopy : tFilterSet;

var list : tFilterSet;
    key  : string;
    i    : integer;

begin
{
     setLength (list, priv_filterset.count);

     i := 0;

     for key in priv_filterset.Keys do
         begin
              inc (i);

              list[i-1] := priv_filterset.Items[key];
         end;
}
     list := tFilterset.create;
     list.AddFilterSet (priv_filterset);

     result := list;
end;

(* ------------------------------------------ *)

function TFormTaskExtract.MetaDataDictionaryToList: tMetaDataSet;

var list : tMetaDataSet;
    key  : string;
    i    : integer;

begin
     setLength (list, priv_metadatalist.Count);
     i := 0;

     for key in priv_metadatalist.Keys do
         begin
              inc (i);

              list[i-1] := priv_metadatalist.Items[key];
         end;

     result := list;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.OnAddFilterFromList (const Filter: tFrameFilter);

// var key : string;

begin
     if priv_filterset.AddFilter (Filter) then
        UpdateFilterList;
{
     key := Filter.DictionaryKey;

     if priv_filterset.ContainsKey (key) then
        begin
             priv_filterset.Add (key, filter);

             // Update tree view

             UpdateFilterList;
        end;
}
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.OnAddFilterToSet (Sender: tObject; var Filter: tFrameFilter);

begin
     // We use this to update the helper variables of the filter, so first we make sure
     // we know what kind of filter this is for;

     filter.updateRecType;

     priv_conversationstore.UpdateFilterHelpers (filter);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onChangeNavigation (Sender: TObject; Node: TTreeNode);

var PageIndex : integer;
    boAccept  : boolean;

begin
     PageIndex := -1;

     case TreeViewNavigation.Selected.AbsoluteIndex of
      0    : PageIndex := 0;
      1    : PageIndex := 1;
      2    : PageIndex := 2;
     end;

     if PageIndex > -1 then
        PageControlExtractSettings.ActivePageIndex := PageIndex;

     boAccept := (PageControlExtractSettings.ActivePageIndex < 2);

     // Declare that we do accept dropped files

     DragAcceptFiles (Handle, boAccept);
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onChangeOutputSettings(Sender: TObject);

begin
     UpdateFormElements;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onChangeFilenamePattern(Sender: TObject);

var valid : boolean;
    help  : char;
    test  : string;
    i     : integer;

begin
     valid := true;

     // remove known placeholders

//   test := LabeledEditFileNamePattern.Text;
     test := ComboBoxFileNamePattern.Text;

     test := stringreplace (test, '<filename>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<count>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<timestamp>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<vlanid>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<sourcemac>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<destinationmac>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<sourceip>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<destinationip>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<sourceport>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<destinationport>', '', [rfReplaceAll, rfIgnoreCase]);
     test := stringreplace (test, '<filterlabel>', '', [rfReplaceAll, rfIgnoreCase]);

     // Check for invalid characters

     for i := 1 to length (test) do
         begin
              help := test[i];

              if help in ['<', '>', ':', '"', '/', '|', '?', '*'] then
                 valid := false;
         end;
{
     if valid then
        LabeledEditFileNamePattern.Font.Color := clWindowText
     else
         LabeledEditFileNamePattern.Font.Color := clRed;
}
     if valid then
        ComboBoxFileNamePattern.Font.Color := clWindowText
     else
         ComboBoxFileNamePattern.Font.Color := clRed;

     ButtonOkay.Enabled := valid;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.onChangeTimestampScope(Sender: TObject);

begin
     DateTimePickerOutputFileTimestamp.Visible := ComboBoxSetFileTimestampScope.ItemIndex = 3;
     DateTimePickerOutputFileTimestamp.Enabled := ComboBoxSetFileTimestampScope.Enabled;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.LoadSettings (const Filename: string);

var StorMan    : tExtractTaskFormStorageManager;
    metalist   : tMetaDataSet;
    filterlist : tFilterSet;
    entry      : tMetaDataEntry;
    len        : integer;
    key        : string;
    i          : integer;

begin
     StorMan := tExtractTaskFormStorageManager.create (self);

     StorMan.retrieveFormFromDB(Filename);
     StorMan.LoadFiltersFromDB (priv_filterset);

     // Add MetaData to list

     metalist := StorMan.LoadMetaDataFromDB;

     priv_metadatalist.Clear;
     len := length (metalist);

     for i := 1 to len do
         begin
              entry := metalist[i-1];
              key := entry.DictionaryKey;

              if not priv_metadatalist.ContainsKey (key) then
                 priv_metadatalist.Add (key, entry)
              else
                  MessageDlg('Warning: duplicate metadata entry found, ignoring! (tFormTaskExtract.LoadSettings)', mtWarning, [mbOK], 0);
         end;

     StorMan.free;
end;

(* ------------------------------------------ *)

procedure TFormTaskExtract.SaveSettings (const Filename: string);

var StorMan    : tExtractTaskFormStorageManager;
    filterlist : tFilterSet;
    metalist   : tMetaDataSet;

begin
     StorMan := tExtractTaskFormStorageManager.create (self);

     StorMan.storeFormToDB (Filename, TaskTypeExtract);

     StorMan.StoreFiltersToDB (priv_filterset);

//   StorMan.StoreMetaDataToDB (priv_metadatalist);

     metalist := MetaDataDictionaryToList;
     StorMan.StoreMetaDataToDB (metalist);

     StorMan.free;
end;

(* ------------------------------------------ *)

function TFormTaskExtract.getParameters: tExtractionParameters;

var fileentry  : tTraceFileEntry;
    params     : tExtractionParameters;

begin
     params.initialize;

     params.FilterList := getFilterListCopy;
//   params.MatchAllOnEmptyList := length (params.FilterList) = 0;
     params.MatchAllOnEmptyList := params.FilterList.count = 0;

     params.MetaDataSet := MetaDataDictionaryToList;

     params.ActiveInterfaceList := priv_conversationstore.GetInterfaceMapListForFilterSet (params.FilterList);

     params.ExtractionMode := xmFilePerConversation;
//   params.WriteMode := mwmAutomatic;

     if RadioButtonFileOutputSameDirectory.Checked then
        begin
             fileentry := priv_filelist.getFile (1);

             params.OutputSchema := includeTrailingPathDelimiter (fileentry.FilePath) + ComboBoxFileNamePattern.Text;
        end
     else
     if RadioButtonFileOutputToSubdirectory.Checked then
        begin
             fileentry := priv_filelist.getFile (1);

             params.OutputSchema := includeTrailingPathDelimiter (fileentry.FilePath) + includeTrailingPathDelimiter (LabeledEditOutputToSubdirectory.Text) + ComboBoxFileNamePattern.Text;
        end
     else
     if RadioButtonFileOutputToAbsoluteDirectory.Checked then
        begin
             params.OutputSchema := includeTrailingPathDelimiter (LabeledEditOutputToAbsoluteDirectory.Text) + ComboBoxFileNamePattern.Text;
        end;

     params.AdjustFileTimestamp := CheckBoxSetFileDateTo.Checked;
     params.ManualTimestamp := 0;

     case ComboBoxSetFileTimestampScope.ItemIndex of
      0 : params.AdjustTimestampMode := tsmOriginalFile;
      1 : params.AdjustTimestampMode := tsmFirstFrame;
      2 : params.AdjustTimestampMode := tsmLastFrame;
      3 : begin
               params.AdjustTimestampMode := tsmSpecificDate;
               params.ManualTimestamp := DateTimePickerOutputFileTimestamp.DateTime;
          end;
     end;

     result := params;
end;

(* ------------------------------------------ *)

end.
