{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit form_task_merge;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls,
  Vcl.ComCtrls, Vcl.Menus, VirtualTrees,

  definition_filter_structures,
  class_interfacemapper,
  class_tracefilelist,
  class_interfacelist,
  class_conversationstorage,
  class_taskprocessor_merge;

type
  TFormTaskMerge = class(TForm)
    TreeViewNavigation: TTreeView;
    ButtonOkay: TButton;
    ButtonCancel: TButton;
    PageControlMergeSettings: TPageControl;
    TabSheetGeneralSettings: TTabSheet;
    GroupBoxActionPackets: TGroupBox;
    TabSheetOutput: TTabSheet;
    GroupBoxFileOutputNamingOptions: TGroupBox;
    LabelOutputFilename: TLabel;
    EditOutputFilename: TEdit;
    TabSheetFilterSettings: TTabSheet;
    RadioButtonMergeModeConcatenate: TRadioButton;
    RadioButtonMergeModeMergeByTimestamp: TRadioButton;
    TabSheetInterfaceMappings: TTabSheet;
    CheckBoxSetFileDateTo: TCheckBox;
    ComboBoxSetFileTimestampScope: TComboBox;
    GroupBoxInterfaceSettings: TGroupBox;
    RadioButtonInterfaceMergeAll: TRadioButton;
    RadioButtonUniqueInterfaces: TRadioButton;
    MemoMergeHint: TMemo;
    PopupMenuInterfaceMappingList: TPopupMenu;
    MenuPopupMappingListExpandAll: TMenuItem;
    MenuPopupMappingListCollapseAll: TMenuItem;
    PopupMenuTargetInterfaceList: TPopupMenu;
    MenuTargetInterfaceAdd: TMenuItem;
    MenuTargetInterfaceEdit: TMenuItem;
    MenuTargetInterfaceDelete: TMenuItem;
    LabelMergeWarning: TLabel;
    PanelFileMappings: TPanel;
    PanelInterfaceTargetList: TPanel;
    Splitter1: TSplitter;
    VSTInterfaceMappingList: TVirtualStringTree;
    GroupBoxMappingDetails: TGroupBox;
    VSTTargetInterfaceList: TVirtualStringTree;
    ButtonSelectOutputFilename: TButton;
    GroupBoxFilterAction: TGroupBox;
    RadioButtonFilterActionPassthrough: TRadioButton;
    RadioButtonFilterActionEnable: TRadioButton;
    MemoFilterActionHint: TMemo;
    GroupBoxFilterSettings: TGroupBox;
    SaveDialogTargetFile: TSaveDialog;
    RadioButtonMapToSingleInterface: TRadioButton;
    VSTFilterList: TVirtualStringTree;
    ButtonAddFilter: TButton;
    ButtonEditFilter: TButton;
    ButtonRemoveFilter: TButton;
    LabelWarningFileExists: TLabel;
    DateTimePickerOutputFileTimestamp: TDateTimePicker;
    PopupMenuFilterList: TPopupMenu;
    MenuFilterLookup: TMenuItem;
    CheckBoxSetFileComment: TCheckBox;
    ComboBoxFileCommentSetting: TComboBox;
    CheckBoxAnnotateSourceFilename: TCheckBox;
    procedure onShow(Sender: TObject);
    procedure onCreate(Sender: TObject);
    procedure onVSTInterfaceMappingListDrawText(Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; const Text: string; const CellRect: TRect; var DefaultDraw: Boolean);
    procedure onVSTInterfaceMappingListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure onVSTTargetInterfaceListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure onVSTInterfaceMappingListChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure MenuPopupMappingListExpandAllClick(Sender: TObject);
    procedure MenuPopupMappingListCollapseAllClick(Sender: TObject);
    procedure MenuTargetInterfaceAddClick(Sender: TObject);
    procedure MenuTargetInterfaceEditClick(Sender: TObject);
    procedure MenuTargetInterfaceDeleteClick(Sender: TObject);
    procedure onTargetInterfaceListPopUp(Sender: TObject);
    procedure onVSTTargetInterfaceListClick(Sender: TObject);
    procedure onClickGeneralSettings(Sender: TObject);
    procedure ButtonSelectOutputFilenameClick(Sender: TObject);
    procedure onChangeNavigation(Sender: TObject; Node: TTreeNode);
    procedure onClickInterfaceMergeSettings(Sender: TObject);
    procedure ButtonAddFilterClick(Sender: TObject);
    procedure ButtonEditFilterClick(Sender: TObject);
    procedure ButtonRemoveFilterClick(Sender: TObject);
    procedure onVSTFilterListGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure onVSTFilterListChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure onVSTFilterListDblClick(Sender: TObject);
    procedure onClickOutputSettings(Sender: TObject);
    procedure onClickFilterSettings(Sender: TObject);
    procedure onChangeOutputFilename(Sender: TObject);
    procedure onChangeTimestampScope(Sender: TObject);
    procedure MenuFilterLookupClick(Sender: TObject);
    procedure onPopupMenuFilterList(Sender: TObject);
    procedure onResizeVSTFilterList(Sender: TObject);
    procedure onDestroy(Sender: TObject);
  private
         priv_filelist           : tTracefileList;
         priv_conversationstore  : tConversationStorage;

         priv_interfacemapper    : tInterfaceMapper;

         priv_current_targetnode : pVirtualNode;     // Indicates which node in the target interface list is currently selected

         priv_filterset : tFilterSet;

         procedure BoldTreeNode (TreeNode: TTreeNode; SetBold: boolean);
         procedure DetermineWarnLabelStatus;
         procedure UpdateInterfaceMappingList;
         procedure UpdateAvailableInterfaceList;
         procedure UpdateFormElements;
         procedure UpdateFilterList;

         procedure OnAddFilterFromList (const Filter : tFrameFilter);

  public
        property FileList : tTracefileList read priv_filelist;
        property ConversationStore : tConversationStorage read priv_conversationstore write priv_conversationstore;

        {$REGION 'Documentation'}
        /// <summary>
        ///   Uses the already given filelist to create initial interface
        ///   mapping
        /// </summary>
        {$ENDREGION}
        procedure PrepareMapper (FileList : tTracefilelist);
        procedure DiscardMapper;

        procedure LoadSettings (const Filename: string);
        procedure SaveSettings (const Filename: string);

        function getParameters : tMergeParameters;
  end;

var
  FormTaskMerge: TFormTaskMerge;

const TaskTypeMerge = 'TaskMerge';

implementation

{$R *.dfm}

uses ShlObj,
     DBXJSON,
     Toolkit,
     CommCtrl,
     base_structures,
     class_formstorage_mergetask,
     dialog_conversation_summary,
     dialog_edit_interface,
     dialog_editfilter;

type tMappingListEntry = record
                               filelistindex            : integer;
                               interfaceindex           : integer;
                               isFileNode               : boolean;
                               OriginalCaptureInterface : tCaptureInterface;
                               MappedCaptureInterface   : tCaptureInterface;
                         end;
     pMappingListEntry = ^tMappingListEntry;

     tAvailableInterfaceEntry = record
                                      Index                  : integer;
                                      MappedCaptureInterface : tCaptureInterface;
                                      UseCount               : integer;
                                end;
     pAvailableInterfaceEntry = ^tAvailableInterfaceEntry;

     tFilterListEntry = record
                              Index                     : integer;
                              FilterAsDisplayFilter     : string;
                              FilterKey                 : string;
                        end;
     pFilterListEntry = ^tFilterListEntry;

(* ------------------------------------------- *)

procedure TFormTaskMerge.BoldTreeNode (TreeNode: TTreeNode; SetBold: boolean);

var treeItem: TTVItem;

begin
     if not Assigned (treeNode) then
        Exit;

     with treeItem do
          begin
               hItem := treeNode.ItemId;
               stateMask := TVIS_BOLD;
               mask := TVIF_HANDLE or TVIF_STATE;

               if SetBold then
                  state := TVIS_BOLD
               else
                   state := 0;
          end;

     TreeView_SetItem(treeNode.Handle, treeItem) ;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.LoadSettings (const Filename: string);

var StorMan : tMergeTaskFormStorageManager;

begin
     StorMan := tMergeTaskFormStorageManager.create(self);

     StorMan.retrieveFormFromDB(Filename);

     // Make sure we have an interface mapper object, because otherwise
     // the load will fail. This can happen at program start when preparing
     // the factory defaults file

     if not (priv_interfacemapper = nil) then
        priv_interfacemapper.LoadFromDatabase (StorMan.Database);

     StorMan.LoadFiltersFromDB (priv_filterset);

     StorMan.free;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.SaveSettings (const Filename: string);

var StorMan : tMergeTaskFormStorageManager;
    filter  : tFilterSet;
    i       : integer;

begin
     StorMan := tMergeTaskFormStorageManager.create (self);

     StorMan.storeFormToDB (Filename, TaskTypeMerge);

     // Make sure we have an interface mapper object, because otherwise
     // the save will fail. This can happen when writing the factory defaults

     if not (priv_interfacemapper = nil) then
        priv_interfacemapper.StoreToDatabase (StorMan.Database);

     StorMan.StoreFiltersToDB (priv_filterset);

     StorMan.free;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.DetermineWarnLabelStatus;

var entry   : tTracefileEntry;
    orderok : boolean;
    i       : integer;

begin
     if RadioButtonMergeModeMergeByTimestamp.Checked then
        begin
             orderok := true;

             for i := 1 to priv_filelist.Count do
                 begin
                      entry := priv_filelist.getFile(i);

                      if not entry.GeneralStatistics.TimeOrderCorrect then
                         orderok := false;
                 end;

             if orderok then
                LabelMergeWarning.Visible := false
             else
                 LabelMergeWarning.Visible := true;
        end
     else
         LabelMergeWarning.Visible := false;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onCreate (Sender: TObject);

begin
     // Workaround for scaling problems: look for /noscale command line option

     self.Scaled := FindCmdLineSwitch ('noscale');

     VSTInterfaceMappingList.NodeDataSize := sizeOf (tMappingListEntry);
     VSTTargetInterfaceList.NodeDataSize := sizeOf (tAvailableInterfaceEntry);
     VSTFilterList.NodeDataSize := sizeOf (tFilterListEntry);

     // Set the interface mapper to nil to prevent problems with storing the form

     priv_interfacemapper := nil;

//   setLength (priv_filterset, 0);
     priv_filterset := tFilterSet.create;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onDestroy(Sender: TObject);

begin
     priv_filterset.Free;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onPopupMenuFilterList(Sender: TObject);

begin
     MenuFilterLookup.Enabled := priv_conversationstore.HasRecords;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onResizeVSTFilterList(Sender: TObject);

begin
     VSTFilterList.Header.Columns[1].Width := VSTFilterList.Width - 83;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onShow (Sender: TObject);

var entry : tTracefileEntry;
    i     : integer;

begin
     // Determine if we need to show a warning

     DetermineWarnLabelStatus;

     UpdateInterfaceMappingList;
     UpdateAvailableInterfaceList;
     UpdateFilterList;

     if EditOutputFilename.Text = '' then
        begin
             // Determine default output name

             entry := priv_filelist.getFile(1);

             EditOutputFilename.Text := entry.FilePath + 'MergedFile.pcapng';
        end;

     for i := 1 to PageControlMergeSettings.PageCount do
         PageControlMergeSettings.Pages[i-1].TabVisible := false;

     PageControlMergeSettings.ActivePageIndex := 0;

     UpdateFormElements;

     // Make sure the first node is active

     TreeViewNavigation.Select(TreeViewNavigation.Items[0]);
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.PrepareMapper (FileList : tTracefilelist);

var Linktype       : longword;
    boSameLinktype : boolean;
    entry          : tTraceFileEntry;
    ifentry        : tCaptureInterface;
    i,j            : integer;

begin
     priv_filelist := Filelist;

     priv_interfacemapper := tInterfaceMapper.create (priv_filelist);

     if RadioButtonInterfaceMergeAll.Checked then
        begin
             priv_interfacemapper.MappingMode := immMerge;
             priv_interfacemapper.doAutoMapping;
        end;

     if RadioButtonUniqueInterfaces.Checked then
        begin
             priv_interfacemapper.MappingMode := immDistinctIFace;
             priv_interfacemapper.doAutoMapping;
        end;

     // Allow certain options if everything is alright

     boSameLinktype := true;
     LinkType := 0;

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist.getFile(i);

              for j := 1 to entry.CaptureInterfaces.Count do
                  begin
                       ifEntry := entry.CaptureInterfaces.getInterfaceByIndex(j-1);

                       if Linktype = 0 then
                          LinkType := ifEntry.LinkType
                       else
                           if LinkType <> ifEntry.LinkType then
                              boSameLinkType := false;
                  end;
         end;

     RadioButtonMapToSingleInterface.Enabled := boSameLinkType;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.DiscardMapper;

begin
     priv_interfacemapper.free;
     priv_interfacemapper := nil;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onChangeNavigation (Sender: TObject; Node: TTreeNode);

var PageIndex : integer;

begin
     PageIndex := -1;

     case TreeViewNavigation.Selected.AbsoluteIndex of
      0    : PageIndex := 0;
      1    : PageIndex := 1;
      2    : PageIndex := 2;
      3    : PageIndex := 3;
     end;

     if PageIndex > -1 then
        PageControlMergeSettings.ActivePageIndex := PageIndex;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onChangeOutputFilename(Sender: TObject);

var path : string;

begin
     path := extractfilepath (EditOutputFilename.Text);

     if DirectoryExists (path) then
        begin
             EditOutputFilename.Font.Color := clWindowText;
             ButtonOkay.Enabled := true;
        end
     else
         begin
              EditOutputFilename.Font.Color := clRed;
              ButtonOkay.Enabled := false;
         end;

     LabelWarningFileExists.Visible := fileexists (EditOutputFilename.Text);
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onChangeTimestampScope(Sender: TObject);

begin
     DateTimePickerOutputFileTimestamp.Visible := ComboBoxSetFileTimestampScope.ItemIndex = 3;
     DateTimePickerOutputFileTimestamp.Enabled := ComboBoxSetFileTimestampScope.Enabled;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.UpdateFormElements;

begin
     // Item Settings

     CheckBoxAnnotateSourceFilename.Enabled := RadioButtonMergeModeConcatenate.Checked;

     DateTimePickerOutputFileTimestamp.Visible := ComboBoxSetFileTimestampScope.ItemIndex = 3;
     DateTimePickerOutputFileTimestamp.Enabled := ComboBoxSetFileTimestampScope.Enabled;

     GroupBoxFilterSettings.Enabled := RadioButtonFilterActionEnable.Checked;

     VSTFilterList.Enabled := GroupBoxFilterSettings.Enabled;
     ButtonAddFilter.Enabled := GroupBoxFilterSettings.Enabled;
     ButtonEditFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);
     ButtonRemoveFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);

     BoldTreeNode (TreeViewNavigation.Items[2], RadioButtonFilterActionEnable.Checked);

     // Hint texts

     // Merge mode

     if RadioButtonMergeModeConcatenate.Checked then
        begin
             MemoMergeHint.Text := 'In concatenation mode, files will be concatenated in the order as found in the file list, which is based on the timestamp of the first frame of each file.' + CRLF +
                                   CRLF;
        end;

     if RadioButtonMergeModeMergeByTimestamp.Checked then
        begin
             MemoMergeHint.Text := 'In merge mode, files will be merge on per-packet timestamp base.' + CRLF +
                                   CRLF;
        end;

     // Mapping mode

     if RadioButtonInterfaceMergeAll.Checked then
        MemoMergeHint.Text := MemoMergeHint.Text + 'All interfaces found in all files are compared to each other by interface name, linklayer type and timestamp resolution, and merged into a deduplicated set of interfaces';

     if RadioButtonUniqueInterfaces.Checked then
        MemoMergeHint.Text := MemoMergeHint.Text + 'Each source interface will be mapped to a single new interface of it''s own, resulting in a file that has as many different interfaces as the sum of all interfaces in all files.';

     if RadioButtonMapToSingleInterface.Checked then
        MemoMergeHint.Text := MemoMergeHint.Text + 'Each source interface will be mapped to the same single new interface, basically creating a file with only one interface that all packets are linked to.';

     // Filter Settings

     if RadioButtonFilterActionEnable.Enabled then
        MemoFilterActionHint.Lines.Text := 'Frames will be filtered before writing merged frames to disk. So if any filters are defined in the list below only frames matching at least one filter will be written to the output file'
     else
         MemoFilterActionHint.Lines.Text := 'No filtering will be performed when writing merged frames to disk. All frames found in all files in the merge set will be written to the output file.';

     DetermineWarnLabelStatus;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onClickGeneralSettings(Sender: TObject);

begin
     UpdateFormElements;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onClickInterfaceMergeSettings(Sender: TObject);

begin
     if RadioButtonInterfaceMergeAll.Checked then
        priv_interfacemapper.MappingMode := immMerge;

     if RadioButtonUniqueInterfaces.Checked then
        priv_interfacemapper.MappingMode := immDistinctIFace;

     if RadioButtonMapToSingleInterface.Checked then
        priv_interfacemapper.MappingMode := immSingleIFace;

     priv_interfacemapper.doAutoMapping;

     UpdateFormElements;

     UpdateInterfaceMappingList;
     UpdateAvailableInterfaceList;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onClickFilterSettings(Sender: TObject);

begin
     GroupBoxFilterSettings.Enabled := RadioButtonFilterActionEnable.Checked;
     VSTFilterList.Enabled := GroupBoxFilterSettings.Enabled;
     ButtonAddFilter.Enabled := GroupBoxFilterSettings.Enabled;
     ButtonEditFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);
     ButtonRemoveFilter.Enabled := GroupBoxFilterSettings.Enabled and (VSTFilterList.SelectedCount > 0);

     if RadioButtonFilterActionEnable.Enabled then
        begin
             MemoFilterActionHint.Lines.Text := 'Frames will be filtered before writing merged frames to disk.';
        end
     else
         begin
              MemoFilterActionHint.Lines.Text := 'No filtering will be performed when writing merged frames to disk.';
         end;

     BoldTreeNode (TreeViewNavigation.Items[2], RadioButtonFilterActionEnable.Checked);
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onClickOutputSettings(Sender: TObject);

begin
     ComboBoxSetFileTimestampScope.Enabled := CheckBoxSetFileDateTo.Checked;

     DateTimePickerOutputFileTimestamp.Visible := ComboBoxSetFileTimestampScope.ItemIndex = 3;
     DateTimePickerOutputFileTimestamp.Enabled := DateTimePickerOutputFileTimestamp.Visible and CheckBoxSetFileDateTo.Checked;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onTargetInterfaceListPopUp(Sender: TObject);

var node : pVirtualNode;
    data : pAvailableInterfaceEntry;

begin
     if (VSTTargetInterfaceList.SelectedCount > 0) then
        begin
             MenuTargetInterfaceEdit.Enabled := true;

             // Delete option is not available for interfaces found in any of the files

             Node := VSTTargetInterfaceList.GetFirstSelected;
             Data := VSTTargetInterfaceList.GetNodeData(Node);

             MenuTargetInterfaceDelete.Enabled := (Data^.MappedCaptureInterface.Origin <> ioFromFile);
        end
     else
         begin
              MenuTargetInterfaceEdit.Enabled := false;
              MenuTargetInterfaceDelete.Enabled := false;
         end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuPopupMappingListCollapseAllClick (Sender: TObject);

begin
     VSTInterfaceMappingList.FullCollapse;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuPopupMappingListExpandAllClick (Sender: TObject);

begin
     VSTInterfaceMappingList.FullExpand;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuFilterLookupClick(Sender: TObject);
{
var count  : integer;
    i      : integer;
    len    : integer;
    filter : tFrameFilter;
}
begin
     FormConversationSummary.ConversationStore := priv_conversationstore;
     FormConversationSummary.TracefileList := priv_filelist;
     FormConversationSummary.SetMode (cmmMergeForm);
     FormConversationSummary.onAddFilterToList := OnAddFilterFromList;
     FormConversationSummary.ShowModal;

     FormConversationSummary.onAddFilterToList := nil;
{
     Deprecated, now done via OnAddFilterFromList event

     if FormConversationSummary.ModalResult = mrOk then
        begin
             count := length (FormConversationSummary.FilterList);
             len := length (priv_filterset);

             for i := 1 to count do
                 begin
                      inc (len);
                      setlength (priv_filterset, len);

                      filter := FormConversationSummary.FilterList[i-1];
                      priv_filterset[len-1] := filter;
                 end;

             // Update tree view

             UpdateFilterList;
        end;
}
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuTargetInterfaceAddClick(Sender: TObject);

var Entry    : tCaptureInterface;
    LinkType : integer;
    IntIndex : integer;

begin
     FormEditInterface.Initialize;
     FormEditInterface.ShowModal;

     if FormEditInterface.ModalResult = mrOK then
        begin
             case FormEditInterface.ComboBoxLinkLayerType.ItemIndex of
              0 : LinkType := LINKTYPE_ETHERNET;
              1 : LinkType := LINKTYPE_LINUX_SLL;
              2 : LinkType := LINKTYPE_JUNIPER;
             end;

             // Determine the new interface index, which starts at 0 so we just set it to "count"

             IntIndex := priv_interfacemapper.TargetInterfaceList.Count;

             Entry := tCaptureInterface.create (IntIndex, LinkType, ioUserCreated, true);
             Entry.Name := FormEditInterface.EditInterfaceName.Text;
             Entry.Description := FormEditInterface.EditInterfaceDescription.Text;
             Entry.SnapLength := FormEditInterface.UpDownSnapLength.Position;
             Entry.TimeStampResolution := StrToInt (FormEditInterface.EditCustomTSResolution.Text);

             priv_interfacemapper.TargetInterfaceList.Add (Entry);

             UpdateAvailableInterfaceList;
        end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuTargetInterfaceDeleteClick(Sender: TObject);

var Node     : pVirtualNode;
    Data     : pAvailableInterfaceEntry;
    Entry    : tCaptureInterface;
    i        : integer;

begin
     Node := VSTTargetInterfaceList.GetFirstSelected;
     Data := VSTTargetInterfaceList.GetNodeData(Node);

     // Find the interface in the list by simple search and remove it

     for i := 1 to priv_interfacemapper.TargetInterfaceList.Count do
         begin
              Entry := priv_interfacemapper.TargetInterfaceList.getInterfaceByIndex(i-1);

              if Entry = Data^.MappedCaptureInterface then
                 begin
                      priv_interfacemapper.TargetInterfaceList.Delete(i-1);

                      // Remove the display node

                      VSTTargetInterfaceList.DeleteNode(Node);
                      VSTTargetInterfaceList.Repaint;
                 end;
         end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.MenuTargetInterfaceEditClick(Sender: TObject);

var Entry    : tCaptureInterface;
    LinkType : integer;
    IntIndex : integer;
    Node     : pVirtualNode;
    Data     : pAvailableInterfaceEntry;

begin
     Node := VSTTargetInterfaceList.GetFirstSelected;
     Data := VSTTargetInterfaceList.GetNodeData(Node);
     Entry := Data^.MappedCaptureInterface;

     FormEditInterface.EditInterfaceName.Text := Entry.Name;
     FormEditInterface.EditInterfaceDescription.Text := Entry.Description;
     FormEditInterface.UpDownSnapLength.Position := Entry.SnapLength;
     FormEditInterface.EditCustomTSResolution.Text := IntToStr (Entry.TimeStampResolution);

     case Entry.LinkType of
      LINKTYPE_ETHERNET  : FormEditInterface.ComboBoxLinkLayerType.ItemIndex := 0;
      LINKTYPE_LINUX_SLL : FormEditInterface.ComboBoxLinkLayerType.ItemIndex := 1;
      LINKTYPE_JUNIPER   : FormEditInterface.ComboBoxLinkLayerType.ItemIndex := 2;
     end;

     case Entry.TimeStampResolution of
      3 : FormEditInterface.ComboBoxTimestampResolution.ItemIndex := 0;
      6 : FormEditInterface.ComboBoxTimestampResolution.ItemIndex := 1;
      9 : FormEditInterface.ComboBoxTimestampResolution.ItemIndex := 2;
     else
         FormEditInterface.ComboBoxTimestampResolution.ItemIndex := 3;
     end;

     FormEditInterface.ShowModal;

     if FormEditInterface.ModalResult = mrOK then
        begin
             case FormEditInterface.ComboBoxLinkLayerType.ItemIndex of
              0 : LinkType := LINKTYPE_ETHERNET;
              1 : LinkType := LINKTYPE_LINUX_SLL;
              2 : LinkType := LINKTYPE_JUNIPER;
             end;

             // Determine the new interface index, which starts at 0 so we just set it to "count"

             IntIndex := priv_interfacemapper.TargetInterfaceList.Count;

             Entry := tCaptureInterface.create (IntIndex, LinkType, ioUserCreated, true);
             Entry.Name := FormEditInterface.EditInterfaceName.Text;
             Entry.Description := FormEditInterface.EditInterfaceDescription.Text;
             Entry.SnapLength := FormEditInterface.UpDownSnapLength.Position;
             Entry.TimeStampResolution := StrToInt (FormEditInterface.EditCustomTSResolution.Text);

             VSTTargetInterfacelist.RepaintNode(Node);
        end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTInterfaceMappingListChange (Sender: TBaseVirtualTree; Node: PVirtualNode);

var FileData   : pMappingListEntry;
    IfData     : pAvailableInterfaceEntry;
    TargetNode : pVirtualNode;
    boFound    : boolean;
    i          : integer;

begin
     if VSTInterfaceMappingList.SelectedCount > 0 then
        begin
             FileData := VSTInterfaceMappingList.GetNodeData(VSTInterfaceMappingList.GetFirstSelected);

             if not FileData^.isFileNode then
                begin
                     // Find the mapped interface in the interface target list and select it

                     boFound := false;
                     IfData := nil;

                     repeat
                           if IfData = nil then
                              TargetNode := VSTTargetInterfaceList.GetFirst
                           else
                               TargetNode := VSTTargetInterfaceList.GetNext(TargetNode);

                           IfData := VSTTargetInterfaceList.GetNodeData(TargetNode);

                           boFound := FileData^.MappedCaptureInterface = IfData^.MappedCaptureInterface;

                     until boFound or (VSTTargetInterfaceList.GetNext(Node) = nil);

                     if boFound and (TargetNode <> priv_current_targetnode) then
                        begin
                             if priv_current_targetnode <> nil then
                                begin
                                     priv_current_targetnode.States := priv_current_targetnode.States - [vsSelected];
                                     VSTTargetInterfaceList.RepaintNode (priv_current_targetnode);
                                end;

                             TargetNode.States := TargetNode.States + [vsSelected];

                             VSTTargetInterfaceList.ScrollIntoView(TargetNode, true);
                             VSTTargetInterfaceList.FocusedNode := TargetNode;
                             VSTTargetInterfaceList.RepaintNode(TargetNode);

                             priv_current_targetnode := TargetNode;
                        end;
                end
             else
                 begin
                      VSTTargetInterfaceList.ClearSelection;

                      if priv_current_targetnode <> nil then
                         begin
                              priv_current_targetnode.States := priv_current_targetnode.States - [vsSelected];
                              VSTTargetInterfaceList.RepaintNode (priv_current_targetnode);
                         end;

                      priv_current_targetnode := nil;
                 end;
        end
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTFilterListChange(Sender: TBaseVirtualTree; Node: PVirtualNode);

begin
     if VSTFilterList.SelectedCount > 0 then
        begin
             ButtonEditFilter.Enabled := (VSTFilterList.SelectedCount = 1);
             ButtonRemoveFilter.Enabled := true;
        end
     else
         begin
             ButtonEditFilter.Enabled := false;
             ButtonRemoveFilter.Enabled := false;
         end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTFilterListDblClick(Sender: TObject);

begin
     if VSTFilterList.SelectedCount > 0 then
        ButtonEditFilter.Click;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTInterfaceMappingListDrawText (Sender: TBaseVirtualTree; TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; const Text: string; const CellRect: TRect; var DefaultDraw: Boolean);

begin
     // This doesn't do much yet, but it may be extended later

     TargetCanvas.TextOut(CellRect.Left, CellRect.Top + 2, text);

     DefaultDraw := false;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTInterfaceMappingListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var nodedata  : pMappingListEntry;
    fileentry : tTracefileEntry;

begin
     nodedata := VSTInterfaceMappingList.GetNodeData(Node);

     case column of
      0 : begin
               if nodedata^.isFileNode then
                  begin
                       fileentry := priv_filelist.getFile(nodedata^.filelistindex);
                       celltext := fileentry.FileName;
                  end
               else
                   begin
                        if nodedata^.OriginalCaptureInterface.Name <> '' then
                           celltext := nodedata^.OriginalCaptureInterface.Name
                        else
                            celltext := '<no name available>';
                   end;
          end;
      1 : begin
               if nodedata^.isFileNode then
                  celltext := ''
               else
                   begin
                        if nodedata^.MappedCaptureInterface.Name <> '' then
                           celltext := nodedata^.MappedCaptureInterface.Name
                        else
                            celltext := '<no description available>';
                   end;
          end;
      2 : begin
               if nodedata^.isFileNode then
                  celltext := ''
               else
                   celltext := IntToStr (nodedata^.MappedCaptureInterface.OriginalID);
          end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTFilterListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var nodedata : pFilterListEntry;

begin
     nodedata := VSTFilterList.GetNodeData (Node);

     case column of
      0 : CellText := IntToStr (nodedata^.Index);
      1 : CellText := nodedata^.FilterAsDisplayFilter;
     end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTTargetInterfaceListClick(Sender: TObject);

var node     : pVirtualNode;
    FileData : pMappingListEntry;
    IfData   : pAvailableInterfaceEntry;
    loc      : tPoint;
    entry    : tTraceFileEntry;
    IfEntry  : tCaptureInterface;

begin
     loc := VSTTargetInterfaceList.ScreenToClient(Mouse.CursorPos);

     // Get the node under the cursor

     node := VSTTargetInterfaceList.GetNodeAt(loc);

     if assigned (node) then
        begin
             // If it is a different one than the current select node, unselect the old one

             if not (priv_current_targetnode = nil) and (node <> priv_current_targetnode) then
                begin
                     priv_current_targetnode.States := priv_current_targetnode.States - [vsSelected];
                     VSTTargetInterfaceList.RepaintNode(priv_current_targetnode);
                end;

             priv_current_targetnode := node;

             // Check if an interface in the file list is selected

             if VSTInterfaceMappingList.SelectedCount > 0 then
                begin
                     // Get target data for later assignment

                     IfData := VSTTargetInterfaceList.GetNodeData(node);

                     node := VSTInterfaceMappingList.GetFirstSelected;
                     FileData := VSTInterfaceMappingList.GetNodeData(node);

                     // Change interface assignment

                     entry := priv_filelist.getFile(FileData^.filelistindex);
                     IfEntry := entry.CaptureInterfaces.getInterfaceByIndex (FileData^.interfaceindex - 1);
                     IfEntry.MappedToInterface := IfData^.MappedCaptureInterface;

                     // Update Node

                     FileData^.MappedCaptureInterface := IfData^.MappedCaptureInterface;
                     VSTInterfaceMappingList.RepaintNode(node);
                end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.onVSTTargetInterfaceListGetText (Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);

var nodedata : pAvailableInterfaceEntry;

begin
     nodedata := VSTTargetInterfaceList.GetNodeData(Node);

     case column of
      0 : begin
               CellText := IntTostr (nodedata^.Index);
          end;
      1 : begin
               if nodedata^.MappedCaptureInterface.Name <> '' then
                  CellText := nodedata^.MappedCaptureInterface.Name
               else
                   CellText := '<no name available>';
          end;
      2 : begin
               if nodedata^.MappedCaptureInterface.Description <> '' then
                  CellText := nodedata^.MappedCaptureInterface.Description
               else
                   CellText := '<no description available>';
          end;
      3 : begin
               CellText := getLinkTypeName (nodedata^.MappedCaptureInterface.LinkType);
          end;
      4 : begin
               CellText := GetInterfaceOriginAsString (nodedata^.MappedCaptureInterface.Origin);
          end;
     end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.UpdateInterfaceMappingList;

var FileNode : pVirtualNode;
    SubNode  : pVirtualNode;
    Data     : pMappingListEntry;
    entry    : tTraceFileEntry;
    iface    : tCaptureInterface;
    i,j      : integer;

begin
     VSTInterfaceMappingList.BeginUpdate;
     VSTInterfaceMappingList.Clear;

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist.getFile(i);

              FileNode := VSTInterfaceMappingList.AddChild (nil);
              Data := VSTInterfaceMappingList.GetNodeData (FileNode);

              Data^.filelistindex := i;
              Data^.OriginalCaptureInterface := nil;
              Data^.MappedCaptureInterface := nil;
              Data^.isFileNode := true;

              // Add subnodes, one per interface

              for j := 1 to entry.CaptureInterfaces.Count do
                  begin
                       iface := entry.CaptureInterfaces.getInterfaceByIndex(j-1);

                       SubNode := VSTInterfaceMappingList.AddChild(FileNode);
                       Data := VSTInterfaceMappingList.GetNodeData(SubNode);

                       Data^.filelistindex := i;
                       Data^.interfaceindex := j;
                       Data^.OriginalCaptureInterface := iface;
                       Data^.MappedCaptureInterface := iface.MappedToInterface;
                       Data^.isFileNode := false;
                  end;
         end;

     VSTInterfaceMappingList.EndUpdate;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.UpdateAvailableInterfaceList;

var Node   : pVirtualNode;
    Data   : pAvailableInterfaceEntry;
    iface  : tCaptureInterface;
    i      : integer;

begin
     priv_current_targetnode := nil;

     VSTTargetInterfaceList.BeginUpdate;
     VSTTargetInterfaceList.Clear;

     for i := 1 to priv_interfacemapper.TargetInterfaceList.Count do
         begin
              iface := priv_interfacemapper.TargetInterfaceList.getInterfaceByIndex(i-1);

              Node := VSTTargetInterfaceList.AddChild (nil);
              Data := VSTTargetInterfaceList.GetNodeData (Node);

              Data^.Index := i;
              Data^.MappedCaptureInterface := iFace;
              Data^.UseCount := 0;
         end;

     VSTTargetInterfaceList.EndUpdate;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.UpdateFilterList;

var Node   : pVirtualNode;
    Data   : pFilterListEntry;
    Filter : tFrameFilter;
    i      : Integer;

begin
     VSTFilterList.BeginUpdate;
     VSTFilterList.Clear;

     Filter := priv_filterset.getFirst;
     i := 0;

//   for i := 1 to length (priv_filterset) do
     while not filter.isUnspecified do
         begin
//            filter := priv_filterset[i-1];
              inc (i);

              Node := VSTFilterList.AddChild(nil);
              Data := VSTFilterList.GetNodeData (Node);

              Data^.Index := i;
              Data^.FilterAsDisplayFilter := filter.asDisplayFilter;
              Data^.FilterKey := filter.DictionaryKey;

              filter := priv_filterset.getNext;
         end;

     VSTFilterList.EndUpdate;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.ButtonAddFilterClick(Sender: TObject);

var Filter : tFrameFilter;
    len    : integer;
    help   : string;

begin
     FormEditFilter.clear;
     FormEditFilter.showmodal;

     if FormEditFilter.ModalResult = mrOK then
        begin
             Filter := FormEditFilter.getFrameFilter;

//           len := length (priv_filterset);
//           setLength (priv_filterset, len + 1);

//           priv_filterset[len] := Filter;
             priv_filterset.AddFilter (Filter);

             UpdateFilterList;
        end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.ButtonEditFilterClick (Sender: TObject);

var filter    : tFrameFilter;
    newfilter : tFrameFilter;
    node      : pVirtualNode;
    data      : pFilterListEntry;

begin
     node := VSTFilterList.GetFirstSelected (false);
     data := VSTFilterList.GetNodeData (node);

//   filter := priv_filterset[data^.Index-1];
     Filter := priv_filterset.getFilterByKey (data^.FilterKey);

     FormEditFilter.PrepareForm (filter);

     FormEditFilter.showmodal;

     if FormEditFilter.ModalResult = mrOK then
        begin
             NewFilter := FormEditFilter.getFrameFilter;

             // Remove the old, add the new

             if Filter.DictionaryKey <> NewFilter.DictionaryKey then
                begin
                     priv_filterset.RemoveFilter (data^.FilterKey);
                     priv_filterset.AddFilter (Filter);
//                   priv_filterset[data^.Index-1] := Filter;

                     data^.FilterAsDisplayFilter := NewFilter.asDisplayFilter;
                     data^.FilterKey := NewFilter.DictionaryKey;

                     VSTFilterList.RepaintNode (node);
                end;
        end;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.ButtonRemoveFilterClick(Sender: TObject);

var node : pVirtualNode;
    last : pVirtualNode;
    data : pFilterListEntry;

begin
     node := VSTFilterList.GetFirstSelected (false);

     // find the last node, because we need to go backwards

     repeat
           last := node;

           node := VSTFilterList.GetNextSelected (node);

     until node = nil;

     // start at the last node

     node := last;

     repeat
           data := VSTFilterList.GetNodeData (node);

//         deleteFilterFromSet (data^.Index - 1, priv_filterset);
           priv_filterset.RemoveFilter (data^.FilterKey);

           node := VSTFilterList.GetPreviousSelected (node);

     until node = nil;

     UpdateFilterList;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.ButtonSelectOutputFilenameClick(Sender: TObject);

var entry   : tTracefileEntry;

begin
     if SaveDialogTargetFile.InitialDir = '' then
        begin
             // Determine default output name

             entry := priv_filelist.getFile(1);

             SaveDialogTargetFile.InitialDir := entry.FilePath;
        end;

     if SaveDialogTargetFile.Execute then
        EditOutputFilename.Text := SaveDialogTargetFile.FileName;
end;

(* ------------------------------------------- *)

function TFormTaskMerge.getParameters: tMergeParameters;

var params : tMergeParameters;

begin
     if RadioButtonMergeModeConcatenate.Checked then
        begin
             params.MergeMode := mmConcatenate;
             params.AddSourceFilePacketComment := CheckBoxAnnotateSourceFilename.Checked;
        end
     else
         params.MergeMode := mmMergeByTimestamp;

     params.MergeTargetFilename := EditOutputFilename.Text;

     params.InterfaceMapper := priv_interfacemapper;
     params.FilterFrames := RadioButtonFilterActionEnable.Checked;
     params.FilterSet := priv_FilterSet;

     params.AdjustFileTimestamp := CheckBoxSetFileDateTo.checked;
     params.ManualTimestamp := 0;

     case ComboBoxSetFileTimestampScope.ItemIndex of
      0 : params.AdjustTimestampMode := tsmOriginalFile;
      1 : params.AdjustTimestampMode := tsmFirstFrame;
      2 : params.AdjustTimestampMode := tsmLastFrame;
      3 : begin
               params.AdjustTimestampMode := tsmSpecificDate;
               params.ManualTimestamp := DateTimePickerOutputFileTimestamp.DateTime;
          end;
     end;

     if CheckBoxSetFileComment.Checked then
        begin
             case ComboBoxFileCommentSetting.ItemIndex of
              0 : params.AddSHBComment := mscListFiles;
              1 : params.AddSHBComment := mscFileCount;
             end;
        end
     else
         params.AddSHBComment := mscNone;

     result := params;
end;

(* ------------------------------------------- *)

procedure TFormTaskMerge.OnAddFilterFromList (const Filter: tFrameFilter);

var len : integer;

begin
//   len := length (priv_filterset);

//   setlength (priv_filterset, len + 1);
//   priv_filterset[len] := Filter;
     priv_filterset.AddFilter (Filter);

     // Update tree view

     UpdateFilterList;
end;

(* ------------------------------------------- *)

end.
