{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_base_structures;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit 
  being tested.

}

interface

uses
  TestFramework, Types, base_structures;

type
  // Test methods for class

  TestBaseStructs = class (TTestCase)

	                        strict private
	
	                        public
	
	                        published
	                        				 procedure TestPatching;
                                   procedure TestCutMarking;
                                   procedure TestNTPTimestamp;
	                  end;

implementation

uses Classes,
     SysUtils,
     Toolkit;

(* -------------------------------------- *)

procedure TestBaseStructs.TestPatching;

var help : tFrameChunk;

begin
     help.CreateZeroBuffer (10);
     help.Patch (0, $54);
     help.Patch (1, $6573);
     help.Patch (3, $74737472);
     help.Patch (7, 'ing');

     CheckEquals ('Teststring', help.getAsANSIString, 'Patching failed');
end;

(* -------------------------------------- *)

procedure TestBaseStructs.TestCutMarking;

var help1   : tFrameChunk;
    help2   : tFrameChunk;
    help3   : tFrameChunk;
    help4   : tFrameChunk;
    merged1 : tFrameChunk;
    merged2 : tFrameChunk;
    merged3 : tFrameChunk;
    extract : tFrameChunk;

begin
     help1.CreateFromPattern ('1', 10);
     help2.CreateFromPattern ('2', 10);
     help3.CreateFromPattern ('3', 10);
     help4.CreateFromPattern ('4', 10);

     merged1 := mergeFrameChunks (help1, help2);
     merged2 := mergeFrameChunks (help3, help4);
     merged3 := mergeFrameChunks (merged1, merged2);

     CheckEquals (1, length (merged1.cutmarkers), 'Number of Markers = 1');
     CheckEquals (2, merged1.SectionCount, 'Section Count = 2');
     CheckEquals (1, length (merged2.cutmarkers), 'Number of Markers = 1');
     CheckEquals (2, merged2.SectionCount, 'Section Count = 2');
     CheckEquals (3, length (merged3.cutmarkers), 'Number of Markers = 3');

     extract := GetSectionFromFrameChunk (merged3, 1);
     CheckEquals ('1111111111', extract.getAsANSIString);

     extract.free;

     extract := GetSectionFromFrameChunk (merged3, 4);
     CheckEquals ('4444444444', extract.getAsANSIString);

     extract.free;

     help1.free;
     help2.free;
     help3.free;
     help4.free;

     merged1.free;
     merged2.free;
     merged3.free;
end;

(* -------------------------------------- *)

procedure TestBaseStructs.TestNTPTimestamp;

const value = 5741542067754219481;

var timestamp : tFrameTimeStamp;
    test      : string;

begin
     timestamp := NTPTimeStampToFrameTime (swap64 (value));
     test := timestamp.asFullString;

     CheckEquals ('', test);
end;

(* -------------------------------------- *)

initialization
              // Register any test cases with the test runner

              RegisterTest(TestBaseStructs.Suite);
end.
