{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_definition_filter_structures;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit
  being tested.

}

interface

uses
  TestFramework, Types;

type
  // Test methods for class

  TestFilters = class (TTestCase)

                        strict private
  private

                        public

                        published
                                 procedure TestEmptyFilter;
                                 procedure TestMACString;
                                 procedure TestIPv4String;
                                 procedure TestIPv6String;
                                 procedure TestPortString;
                                 procedure TestIPv4AndPortString;
                  end;

implementation

uses SysUtils,
     DBXJSON,
     System.JSON,
     base_structures,
     definition_filter_structures,
     definition_ip_addresses;

(* -------------------------------------- *)
{ TestFilters }
(* -------------------------------------- *)

procedure TestFilters.TestEmptyFilter;

var filter : tFrameFilter;
    jobj   : tJSONObject;
    jsval  : string;

begin
     filter.initialize;

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{}', jsval, 'Empty JSON object');

end;

(* -------------------------------------- *)

procedure TestFilters.TestMACString;

var filter : tFrameFilter;
    jobj   : tJSONObject;
    jsval  : string;

begin
     filter.initialize;

     filter.MACFilter.AddressFilterType := ftSingle;
     filter.MACFilter.AddressScope := fsAny;
     filter.MACFilter.SingleAddress.setfromString ('01:02:03:04:05:06', ':');

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{"MACFilter":{"FilterType":"Single","Scope":"Any","SingleAddress":"01:02:03:04:05:06","SourceAddress":"00:00:00:00:00:00","DestinationAddress":"00:00:00:00:00:00"}}', jsval, 'MAC Simple Test');
end;

(* -------------------------------------- *)


procedure TestFilters.TestIPv4String;

var filter : tFrameFilter;
    jobj   : tJSONObject;
    jsval  : string;

begin
     filter.initialize;

     filter.IPv4Filter.AddressFilterType := ftSingle;
     filter.IPv4Filter.AddressScope := fsAny;
     filter.IPv4Filter.SingleAddress.setValueFromString('192.168.124.100');

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{"IPv4Filter":{"FilterType":"Single","Scope":"Any","SingleAddress":"192.168.124.100/32","SourceAddress":"0.0.0.0/32","DestinationAddress":"0.0.0.0/32","FirstAddress":"0.0.0.0/32","LastAddress":"0.0.0.0/32"}}', jsval, 'IPv4 Simple Test');
end;

(* -------------------------------------- *)

procedure TestFilters.TestIPv6String;

var filter : tFrameFilter;
    jobj   : tJSONObject;
    jsval  : string;

begin
     filter.initialize;

     filter.IPv6Filter.AddressFilterType := ftSingle;
     filter.IPv6Filter.AddressScope := fsAny;
     filter.IPv6Filter.SingleAddress.setValueFromString('2001:db8:1234::1', 128);

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{"IPv6Filter":{"FilterType":"Single","Scope":"Any","SingleAddress":"2001:db8:1234::1/128","SourceAddress":"::/128","DestinationAddress":"::/128","FirstAddress":"::/128","LastAddress":"::/128"}}', jsval, 'IPv6 Simple Test');
end;

(* -------------------------------------- *)

procedure TestFilters.TestPortString;

var filter  : tFrameFilter;
    jobj    : tJSONObject;
    portobj : tJSONObject;
    jval    : tJSONValue;
    jsval   : string;

begin
     filter.initialize;

     filter.PortFilter.Layer4Protocol := l4Any;
     filter.PortFilter.PortFilterType := ftSingle;
     filter.PortFilter.PortScope := fsAny;
     filter.PortFilter.SinglePort := 80;

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{"PortFilter":{"Layer4Protocol":"Any","FilterType":"Single","Scope":"Any","SinglePort":80,"SourcePort":0,"DestinationPort":0,"FirstPort":0,"LastPort":0}}', jsval, 'Port Simple Test');

     portobj := jobj.GetValue('PortFilter') as tJSONObject;
     jsVal := portobj.ToString;

     CheckEquals ('{"Layer4Protocol":"Any","FilterType":"Single","Scope":"Any","SinglePort":80,"SourcePort":0,"DestinationPort":0,"FirstPort":0,"LastPort":0}', jsval, 'PortFilter Subobject');

     jval := portobj.GetValue ('SinglePort');
     jsVal := jval.ToString;

     CheckEquals ('80', jsval, 'SinglePort');
end;

(* -------------------------------------- *)

procedure TestFilters.TestIPv4AndPortString;

var filter : tFrameFilter;
    jobj   : tJSONObject;
    jsval  : string;

begin
     filter.initialize;

     filter.IPv4Filter.AddressFilterType := ftSingle;
     filter.IPv4Filter.AddressScope := fsAny;
     filter.IPv4Filter.SingleAddress.setValueFromString('192.168.124.100');

     filter.PortFilter.Layer4Protocol := l4Any;
     filter.PortFilter.PortFilterType := ftSingle;
     filter.PortFilter.PortScope := fsAny;
     filter.PortFilter.SinglePort := 80;

     jobj := filter.toJSON;
     jsval := jobj.ToString;

     CheckEquals ('{"IPv4Filter":{"FilterType":"Single","Scope":"Any","SingleAddress":"192.168.124.100/32","SourceAddress":"0.0.0.0/32","DestinationAddress":"0.0.0.0/32","FirstAddress":"0.0.0.0/32","LastAddress":"0.0.0.0/32"},' +
                  '"PortFilter":{"Layer4Protocol":"Any","FilterType":"Single","Scope":"Any","SinglePort":80,"SourcePort":0,"DestinationPort":0,"FirstPort":0,"LastPort":0}}', jsval, 'Port Simple Test');

     jObj.Free;

     // Create JSON object from String

     jObj := TJSONObject.ParseJSONValue (TEncoding.ASCII.GetBytes (jsval),0) as tJSONObject;
     jsval := jobj.ToString;

     CheckEquals ('{"IPv4Filter":{"FilterType":"Single","Scope":"Any","SingleAddress":"192.168.124.100/32","SourceAddress":"0.0.0.0/32","DestinationAddress":"0.0.0.0/32","FirstAddress":"0.0.0.0/32","LastAddress":"0.0.0.0/32"},' +
                  '"PortFilter":{"Layer4Protocol":"Any","FilterType":"Single","Scope":"Any","SinglePort":80,"SourcePort":0,"DestinationPort":0,"FirstPort":0,"LastPort":0}}', jsval, 'Port Simple Test');

     // Create Filter from JSON

     filter.initialize (jObj);
end;

(* -------------------------------------- *)

initialization
              // Register any test cases with the test runner

              RegisterTest(TestFilters.Suite);
end.

