{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_ipv4subnethandler;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit 
  being tested.

}

interface

uses
  TestFramework, class_ipv4subnethandler, contnrs, definition_ip_addresses;

type
  // Test methods for class tSubnetHandler

  TesttSubnetHandler = class(TTestCase)
  strict private
    FtSubnetHandler: tIPv4SubnetHandler;
  public
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestGetSubnetIndexForIPAddress;
  end;

implementation

uses SysUtils;

(* -------------------------------------- *)

procedure TesttSubnetHandler.SetUp;

begin
     FtSubnetHandler := tIPv4SubnetHandler.Create;
end;

(* -------------------------------------- *)

procedure TesttSubnetHandler.TearDown;

begin
     FtSubnetHandler.Free;
     FtSubnetHandler := nil;
end;

(* -------------------------------------- *)

procedure TesttSubnetHandler.TestGetSubnetIndexForIPAddress;

var NetIndex    : Integer;
    IPAddress   : tIPv4Address;
    subnet      : tIPv4SubnetDefinition;
    i           : integer;

begin
     subnet := tIPv4SubnetDefinition.create('192.168.0.0', '255.255.0.0', 'Private Networks');
     FtSubnetHandler.AddSubnet(subnet);

     subnet := tIPv4SubnetDefinition.create('192.168.122.0', '255.255.255.0', 'Mgmt');
     FtSubnetHandler.AddSubnet(subnet);

     subnet := tIPv4SubnetDefinition.create('192.168.124.0', '255.255.255.0', 'JayNet');
     FtSubnetHandler.AddSubnet(subnet);

     subnet := tIPv4SubnetDefinition.create('81.209.179.0', '255.255.255.0', 'Public179');
     FtSubnetHandler.AddSubnet(subnet);

     subnet := tIPv4SubnetDefinition.create('81.209.147.0', '255.255.255.0', 'TooBig147');
     FtSubnetHandler.AddSubnet(subnet);

     subnet := tIPv4SubnetDefinition.create('81.209.147.0', '255.255.255.240', 'Public147');
     FtSubnetHandler.AddSubnet(subnet);

     // Test for OldNet

     for i := 1 to 255 do
         begin
              IPAddress.setValueFromString ('81.209.147.' + IntToStr (i));
              NetIndex := FtSubnetHandler.GetSubnetIndexForIPAddress(IPAddress);
              subnet := FtSubnetHandler.GetSubnetByIndex(NetIndex);
              CheckEquals (subnet.SubnetAddress.ValueAsString, '81.209.147.0', 'Public147');

              if i < 16 then
                 CheckEquals (28, subnet.SubnetAddress.MaskBitCount, 'Public 147 Mask wrong for IP ' + '81.209.147.' + IntToStr (i))
              else
                  CheckEquals (24, subnet.SubnetAddress.MaskBitCount, 'Public 147 Mask wrong for IP ' + '81.209.147.' + IntToStr (i));

              // Test for NewNet

              IPAddress.setValueFromString ('81.209.179.' + IntToStr (i));
              NetIndex := FtSubnetHandler.GetSubnetIndexForIPAddress(IPAddress);
              subnet := FtSubnetHandler.GetSubnetByIndex(NetIndex);
              CheckEquals (subnet.SubnetAddress.ValueAsString, '81.209.179.0', 'NewNet');

              // Test for Mgmt

              IPAddress.setValueFromString ('192.168.122.' + IntToStr (i));
              NetIndex := FtSubnetHandler.GetSubnetIndexForIPAddress(IPAddress);
              subnet := FtSubnetHandler.GetSubnetByIndex(NetIndex);
              CheckEquals (subnet.SubnetAddress.ValueAsString, '192.168.122.0', 'Mgmt');

              // Test for None

              IPAddress.setValueFromString ('137.226.181.' + IntToStr (i));
              NetIndex := FtSubnetHandler.GetSubnetIndexForIPAddress(IPAddress);

              CheckEquals (NetIndex, 0, 'No Match should have been found');
         end;
end;

(* -------------------------------------- *)

initialization
              // Register any test cases with the test runner

              RegisterTest(TesttSubnetHandler.Suite);
end.

