{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_parser_dns;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit
  being tested.

}

interface

uses
  TestFramework, classes, class_parser_baseclass, base_structures,
  definition_ip_addresses, class_parsemaster, class_parser_dns;

type
  // Test methods for class tDNSParser

  TesttDNSParser = class(TTestCase)

                        strict private
                                      Parsemaster : tParseMaster;

                                      function PrepareTest (Tracefile : string) : tDNSParser;
                        public

                        published
                                 procedure TestparseICMPQuoteDNS;
                                 procedure TestparseICMPQuoteDNS2;
                   end;

implementation

uses class_rawframe,
     common_parse_helpers,
     class_loader_pcapng;

(* ------------------------------------------- *)

function TesttDNSParser.PrepareTest(Tracefile: string): tDNSParser;

var Loader      : tPCAPNGLoader;
    Frame       : tRawFrame;
    DNSParser   : tDNSParser;
    OpenResult  : boolean;
    i           : integer;

begin
     Loader := tPCAPNGLoader.create (Tracefile);
     OpenResult := Loader.openFile;

     Loader.readTraceFileheader;
     Loader.readBatchFromDisk;

     Frame := Loader.getRawFrame(1);

     ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     Loader.Free;
     Loader := nil;

     CheckEquals (true, Openresult, 'PCAPNGLoader.OpenFile Failed');

     i := 0;

     // Find the ICMP Parser

     repeat
           inc (i)
     until (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_DNS) or (i >= Parsemaster.ParserCount);

     if (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_DNS) then
        DNSParser := (Parsemaster.getParserByIndex (i) as tDNSParser)
     else
         DNSParser := nil;

     result := DNSParser;
end;

(* ------------------------------------------- *)

procedure TesttDNSParser.TestparseICMPQuoteDNS;

var DNSParser : tDNSParser;

begin
     DNSParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\DNS-Destination-unreachable-Quote.pcapng');

     if DNSParser <> nil then
        begin
             CheckEquals ($2946, DNSParser.TransactionID, 'Transaction ID');
             CheckTrue (DNSParser.isResponse, 'Is DNS Response');
             CheckTrue (DNSParser.RecursionDesired, 'Recursion Desired');
             CheckTrue (DNSParser.RecursionAvailable, 'Recursion Available');
             CheckEquals (1, DNSParser.QueryCount, 'Query Count');
             CheckEquals (1, DNSParser.AnswerCount, 'Answer Count');
             CheckEquals (21, DNSParser.AuthorityCount, 'Authority Ressource Records');

             CheckEquals ('122.63.189.61.zen.spamhaus.org', DNSParser.QueryRecordSet[0].Name.FullFQDN, 'Query Name');
        end;
end;

(* ------------------------------------------- *)

procedure TesttDNSParser.TestparseICMPQuoteDNS2;

var DNSParser : tDNSParser;

begin
     DNSParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\DNS-Destination-unreachable-Quote2.pcapng');

     if DNSParser <> nil then
        begin
             CheckEquals ($b5c6, DNSParser.TransactionID, 'Transaction ID');
             CheckTrue (DNSParser.isResponse, 'Is DNS Response');
             CheckTrue (DNSParser.IsAuthoritative, 'Is Authoritative');
             CheckFalse (DNSParser.RecursionDesired, 'Recursion Desired');
             CheckFalse (DNSParser.RecursionAvailable, 'Recursion Available');
             CheckEquals (1, DNSParser.QueryCount, 'Query Count');
             CheckEquals (1, DNSParser.AnswerCount, 'Answer Count');
             CheckEquals (2, DNSParser.AuthorityCount, 'Authority Ressource Records');
             CheckEquals (4, DNSParser.AdditionalInfoCount, 'Authority Ressource Records');

             CheckEquals ('wiki.dynamics4.net', DNSParser.QueryRecordSet[0].Name.FullFQDN, 'Query Name');
             CheckEquals ('wiki.dynamics4.net', DNSParser.AnswerRecordSet[0].Name.FullFQDN, 'Answer Name');
             CheckEquals ('81.209.179.101', DNSParser.AnswerRecordSet[0].Answer.AsString, 'Answer IP Address');
        end;
end;

(* ------------------------------------------- *)

initialization
              // Register any test cases with the test runner

              RegisterTest(TesttDNSParser.Suite);
end.

