{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_parser_icmpv4;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit 
  being tested.

}

interface

uses TestFramework, class_parser_baseclass, class_parser_icmpv4,
     definition_ip_addresses, base_structures, class_parsemaster;

type
     // Test methods for class tICMPv4Parser

     TesttICMPv4Parser = class (TTestCase)

                               strict private
                                             FtICMPv4Parser: tICMPv4Parser;

                               private
                                      Parsemaster : tParseMaster;

                                      function PrepareTest (Tracefile : string) : tICMPv4Parser;

                               public
                                     procedure SetUp; override;
                                     procedure TearDown; override;

                               published
                                        procedure TestParsePingRequest;
                                        procedure TestParseRedirectForNetwork;
                                        procedure TestParseFragmentationNeeded;
                                        procedure TestParseHostUnreachable;
                                        procedure TestParseAdministrativelyFiltered;
                                        procedure TestParseTTLExceeded;
                                        procedure TestParsePortUnreachable;
                         end;

implementation

uses class_rawframe,
     class_loader_pcapng;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.SetUp;

begin

end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TearDown;

begin

end;

(* ------------------------------------------- *)

function TesttICMPv4Parser.PrepareTest(Tracefile: string) : tICMPv4Parser;

var Loader      : tPCAPNGLoader;
    Frame       : tRawFrame;
    ICMPParser  : tICMPv4Parser;
    OpenResult  : boolean;
    i           : integer;

begin
     Loader := tPCAPNGLoader.create (Tracefile);
     OpenResult := Loader.openFile;

     Loader.readTraceFileheader;
     Loader.readBatchFromDisk;

     Frame := Loader.getRawFrame(1);

     ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     Loader.Free;
     Loader := nil;

     CheckEquals (true, Openresult, 'PCAPNGLoader.OpenFile Failed');

     i := 0;

     // Find the ICMP Parser

     repeat
           inc (i)
     until (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_ICMP) or (i >= Parsemaster.ParserCount);

     if (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_ICMP) then
        ICMPParser := (Parsemaster.getParserByIndex (i) as tICMPv4Parser)
     else
         ICMPParser := nil;

     result := ICMPParser;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParsePingRequest;

var ICMPParser  : tICMPv4Parser;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-PingRequest.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (8, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($765b, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (1280, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (53760, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals ('abcdefghijklmnopqrstuvwabcdefghi', ICMPParser.PayloadAsString, 'Payload');
             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParseRedirectForNetwork;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-RedirectForNetwork.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (5, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($e2f9, ICMPParser.Checksum, 'Checksum');
             CheckEquals ('192.168.2.1/32', ICMPParser.RedirectGwy.ValueAsCIDRString, 'Next Hop Gateway');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($00, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParseFragmentationNeeded;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-FragmentationNeeded.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (3, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (4, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($9f7b, ICMPParser.Checksum, 'Checksum');
             CheckEquals (800, ICMPParser.NextHopMTU, 'Next Hop MTU');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($00, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParsePortUnreachable;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-DestinationPortUnreachable.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (3, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (3, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($128a, ICMPParser.Checksum, 'Checksum');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($01, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParseHostUnreachable;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-DestinationHostUnreachable.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (3, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (1, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($a7a2, ICMPParser.Checksum, 'Checksum');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($00, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParseAdministrativelyFiltered;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-DestinationAdministrativelyFiltered.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (3, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (13, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($33cb, ICMPParser.Checksum, 'Checksum');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($c3, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv4Parser.TestParseTTLExceeded;

var ICMPParser  : tICMPv4Parser;
    PayLoad     : tRawBuffer;

begin
     ICMPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMP-TTLExceeded.pcapng');

     if ICMPParser <> nil then
        begin
             PayLoad := ICMPParser.Payload;

             CheckEquals (11, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($cfbf, ICMPParser.Checksum, 'Checksum');
             CheckEquals (964, ICMPParser.OriginalDatagramLength, 'Original Datagram Length');

             CheckEquals ($45, PayLoad[1], 'First Payload Byte');
             CheckEquals ($00, PayLoad[ICMPParser.PayloadLength], 'Last Payload Byte');
             CheckEquals (NEXT_HEADER_IPv4, ICMPParser.NextAnchor.HeaderType, 'Next Header');
        end;
end;

(* ------------------------------------------- *)

initialization
              // Register any test cases with the test runner
              RegisterTest(TesttICMPv4Parser.Suite);
end.

