{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_parser_icmpv6;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit
  being tested.

}

interface

uses TestFramework, class_parser_baseclass, class_parser_icmpv6,
     definition_ip_addresses, base_structures, class_parsemaster;

type
     // Test methods for class tICMPv4Parser

     TesttICMPv6Parser = class (TTestCase)

                               strict private
                                             FtICMPv4Parser: tICMPv6Parser;

                               private
                                      Parsemaster : tParseMaster;

                                      function PrepareTest (Tracefile : string) : tICMPv6Parser;

                               public
                                     procedure SetUp; override;
                                     procedure TearDown; override;

                               published
                                        procedure TestParsePingRequest;
                                        procedure TestParseDestinationUnreachable;
                                        procedure TestParsePacketTooBig;
                                        procedure TestParseRouterSolicitation;
                                        procedure TestParseRouterAdvertisement;
                                        procedure TestParseRouterAdvertisement2;
                                        procedure TestParseRouterAdvertisement3;
                                        procedure TestParseNeighborSolicitation;
                                        procedure TestParseNeighborAdvertisement;
                         end;

implementation

uses class_rawframe,
     class_loader_pcapng;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.SetUp;

begin

end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TearDown;

begin

end;

(* ------------------------------------------- *)

function TesttICMPv6Parser.PrepareTest(Tracefile: string) : tICMPv6Parser;

var Loader      : tPCAPNGLoader;
    Frame       : tRawFrame;
    ICMPParser  : tICMPv6Parser;
    OpenResult  : boolean;
    i           : integer;

begin
     Loader := tPCAPNGLoader.create (Tracefile);
     OpenResult := Loader.openFile;

     Loader.readTraceFileheader;
     Loader.readBatchFromDisk;

     Frame := Loader.getRawFrame(1);

     ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     Loader.Free;
     Loader := nil;

     CheckEquals (true, Openresult, 'PCAPNGLoader.OpenFile Failed');

     i := 0;

     // Find the ICMP Parser

     repeat
           inc (i)
     until (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_IPV6_ICMP) or (i >= Parsemaster.ParserCount);

     if (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_IPV6_ICMP) then
        ICMPParser := (Parsemaster.getParserByIndex (i) as tICMPv6Parser)
     else
         ICMPParser := nil;

     result := ICMPParser;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParsePingRequest;

var ICMPParser  : tICMPv6Parser;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-EchoRequest.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (40, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (128, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($5492, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (1, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals ('abcdefghijklmnopqrstuvwabcdefghi', ICMPParser.PayloadAsString, 'Payload');
             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($5e, ICMPParser.NextAnchor.Offset, 'Next Offset');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseDestinationUnreachable;

var ICMPParser  : tICMPv6Parser;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-DestinationUnreachable.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (8, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (1, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($18b8, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (NEXT_HEADER_IPv6, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($3e, ICMPParser.NextAnchor.Offset, 'Next Offset');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParsePacketTooBig;

var ICMPParser  : tICMPv6Parser;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-PacketTooBig.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (8, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (2, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($ad3d, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (1280, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals (NEXT_HEADER_IPv6, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($3e, ICMPParser.NextAnchor.Offset, 'Next Offset');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseRouterSolicitation;

var ICMPParser  : tICMPv6Parser;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-RouterSolicitation.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (16, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (133, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($4b20, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($46, ICMPParser.NextAnchor.Offset, 'Next Offset');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseRouterAdvertisement;

var ICMPParser  : tICMPv6Parser;
    Option      : tICMPv6NDOption;
    i           : integer;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-RouterAdvertisement.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (96, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (134, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($1f9f, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals (0, ICMPParser.RAHopLimit, 'RA Hop Limit');
             CheckEquals (0, ICMPParser.RAReachableTime, 'RA Reachable Time');
             CheckEquals (0, ICMPParser.RARetransTimer, 'RA RetransTime');
             CheckEquals (0, ICMPParser.RALifeTime, 'Router Lifetime');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($96, ICMPParser.NextAnchor.Offset, 'Next Offset');

             for i := 1 to length (ICMPParser.OptionList) do
                 begin
                      Option := ICMPParser.OptionList[i-1];

                      case i of
                       1 : begin
                                CheckEquals (1, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals ('00:b0:d0:23:47:33', Option.MACAddress.asString (':', true), 'Source Link Layer Address');
                           end;
                       2 : begin
                                CheckEquals (5, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals (1500, Option.MTU, 'MTU');

                           end;
                       3 : begin
                                CheckEquals (3, Option.option_type, 'Option Type');
                                CheckEquals (32, Option.option_len, 'Option Length');
                                CheckEquals (64, Option.PrefixLength, 'Prefix Length');
                                CheckEquals ($ffffffff, Option.PrefixValidLifetime, 'Valid Lifetime');
                                CheckEquals ($ffffffff, Option.PrefixPrefLifetime, 'Preferred Lifetime');
                                CheckEquals ('2001:db8:0:2::', Option.PrefixAddress.ValueAsString, 'Prefix Address');
                           end;
                       4 : begin
                                CheckEquals (3, Option.option_type, 'Option Type');
                                CheckEquals (32, Option.option_len, 'Option Length');
                                CheckEquals (64, Option.PrefixLength, 'Prefix Length');
                                CheckEquals ($ffffffff, Option.PrefixValidLifetime, 'Valid Lifetime');
                                CheckEquals ($ffffffff, Option.PrefixPrefLifetime, 'Preferred Lifetime');
                                CheckEquals ('fd43:2da1:3fe9:2::', Option.PrefixAddress.ValueAsString, 'Prefix Address');
                           end;
                      end;
                 end;
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseRouterAdvertisement2;

var ICMPParser  : tICMPv6Parser;
    Option      : tICMPv6NDOption;
    i           : integer;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-RouterAdvertisement2.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (64, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (134, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($c160, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals (64, ICMPParser.RAHopLimit, 'RA Hop Limit');
             CheckEquals (0, ICMPParser.RAReachableTime, 'RA Reachable Time');
             CheckEquals (0, ICMPParser.RARetransTimer, 'RA RetransTime');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($76, ICMPParser.NextAnchor.Offset, 'Next Offset');

             for i := 1 to length (ICMPParser.OptionList) do
                 begin
                      Option := ICMPParser.OptionList[i-1];

                      case i of
                       1 : begin
                                CheckEquals (1, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals ('00:12:80:31:d6:80', Option.MACAddress.asString (':', true), 'Source Link Layer Address');
                           end;
                       2 : begin
                                CheckEquals (5, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals (1500, Option.MTU, 'MTU');

                           end;
                       3 : begin
                                CheckEquals (3, Option.option_type, 'Option Type');
                                CheckEquals (32, Option.option_len, 'Option Length');
                                CheckEquals (64, Option.PrefixLength, 'Prefix Length');
                                CheckEquals (900, Option.PrefixValidLifetime, 'Valid Lifetime');
                                CheckEquals (900, Option.PrefixPrefLifetime, 'Preferred Lifetime');
                                CheckEquals ('2001:4da0:0:3::', Option.PrefixAddress.ValueAsString, 'Prefix Address');
                           end;
                      end;
                 end;
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseRouterAdvertisement3;

var ICMPParser  : tICMPv6Parser;
    Option      : tICMPv6NDOption;
    i           : integer;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-RouterAdvertisementWithDNS.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (128, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (134, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($29b6, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals (64, ICMPParser.RAHopLimit, 'RA Hop Limit');
             CheckEquals (30, ICMPParser.RALifeTime, 'RA Reachable Time');
             CheckEquals (0, ICMPParser.RAReachableTime, 'RA Reachable Time');
             CheckEquals (0, ICMPParser.RARetransTimer, 'RA RetransTime');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($b6, ICMPParser.NextAnchor.Offset, 'Next Offset');

             for i := 1 to length (ICMPParser.OptionList) do
                 begin
                      Option := ICMPParser.OptionList[i-1];

                      case i of
                       1 : begin
                                CheckEquals (3, Option.option_type, 'Option Type');
                                CheckEquals (32, Option.option_len, 'Option Length');
                                CheckEquals (64, Option.PrefixLength, 'Prefix Length');

                                CheckTrue (poOnLink in Option.PrefixFlags);
                                CheckTrue (poAutonomous in Option.PrefixFlags);
                                CheckTrue (poRouterAddress in Option.PrefixFlags);

                                CheckEquals (86400, Option.PrefixValidLifetime, 'Valid Lifetime');
                                CheckEquals (14400, Option.PrefixPrefLifetime, 'Preferred Lifetime');
                                CheckEquals ('2001:4dd0:fb48:2::', Option.PrefixAddress.ValueAsString, 'Prefix Address');
                           end;
                       2 : begin
                                CheckEquals (25, Option.option_type, 'Option Type');
                                CheckEquals (40, Option.option_len, 'Option Length');

                                CheckEquals (10, Option.DNSLifeTime, 'DNS Lifetime');
                                CheckEquals ('2001:4860:4860::8888', Option.DNSServerList[0].ValueAsString, 'DNS Server 1');
                                CheckEquals ('2001:4860:4860::8844', Option.DNSServerList[1].ValueAsString, 'DNS Server 1');
                           end;
                       3 : begin
                                CheckEquals (31, Option.option_type, 'Option Type');
                                CheckEquals (24, Option.option_len, 'Option Length');

                                CheckEquals (10, Option.DNSLifeTime, 'DNS Lifetime');
                                CheckEquals ('ipv6net.local', Option.DNSDomainNameList[0].FullFQDN, 'DNS Name Server');
                           end;
                       4 : begin
                                CheckEquals (5, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals (1280, Option.MTU, 'MTU');
                           end;
                       5 : begin
                                CheckEquals (1, Option.option_type, 'Option Type');
                                CheckEquals (8, Option.option_len, 'Option Length');
                                CheckEquals ('00:0c:29:e7:73:6b', Option.MACAddress.asString (':', true), 'Source Link Layer Address');
                           end;
                      end;
                 end;
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseNeighborSolicitation;

var ICMPParser  : tICMPv6Parser;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-NeighborSolicitation.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (32, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (135, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($0f35, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals ('fe80::260:97ff:fe02:6ea5', ICMPParser.NDTargetAddress.ValueAsString, 'ND Target Address');

             CheckEquals ('00:10:5a:aa:20:a2', ICMPParser.OptionList[0].MACAddress.asString (':', true), 'Link Layer Address option');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($56, ICMPParser.NextAnchor.Offset, 'Next Offset');
        end;
end;

(* ------------------------------------------- *)

procedure TesttICMPv6Parser.TestParseNeighborAdvertisement;

var ICMPParser  : tICMPv6Parser;
    Option      : tICMPv6NDOption;

begin
     ICMPParser := PrepareTest ('D:\Software Development\Jays Utilities\Trace Wrangler\test\traces\ICMPv6-NeighborAdvertisement.pcapng');

     if ICMPParser <> nil then
        begin
             CheckEquals (32, ICMPParser.LayerSize, 'LayerSize');
             CheckEquals (136, ICMPParser.ICMPType, 'ICMP Type');
             CheckEquals (0, ICMPParser.ICMPCode, 'ICMP code');
             CheckEquals ($89ac, ICMPParser.Checksum, 'ICMP Checksum');
             CheckEquals (0, ICMPParser.EchoIdentifier, 'ECHO Identifier');
             CheckEQuals (0, ICMPParser.EchoSequence, 'ECHO Sequence');
             CheckEquals (0, ICMPParser.NextHopMTU, 'MTU');
             CheckEquals ('fe80::260:97ff:fe02:6ea5', ICMPParser.NDTargetAddress.ValueAsString, 'ND Target Address');

             CheckEquals ('00:60:97:02:6e:a5', ICMPParser.OptionList[0].MACAddress.asString (':', true), 'Link Layer Address option');

             CheckEquals (NEXT_HEADER_NONE, ICMPParser.NextAnchor.HeaderType, 'Next Header');
             CheckEquals ($56, ICMPParser.NextAnchor.Offset, 'Next Offset');

             Option := ICMPParser.OptionList[0];

             CheckEquals (2, Option.option_type, 'Option Type');
             CheckEquals (8, Option.option_len, 'Option Length');
             CheckEquals ('00:60:97:02:6e:a5', Option.MACAddress.asString(':', true), 'Link Layer Address');
        end;
end;

(* ------------------------------------------- *)

initialization
              // Register any test cases with the test runner
              RegisterTest(TesttICMPv6Parser.Suite);
end.

