{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}
 
unit Testclass_parser_tcp;
{

  Delphi DUnit Test Case
  ----------------------
  This unit contains a skeleton test case class generated by the Test Case Wizard.
  Modify the generated code to correctly setup and call the methods from the unit
  being tested.

}

interface

uses
  TestFramework, class_parser_baseclass, class_parser_tcp, base_structures,
  base_checksums, class_parsemaster;

type
  // Test methods for class

  TesttTCPParser = class(TTestCase)
  strict private
    Parsemaster : tParseMaster;
  public
    procedure SetUp; override;
    procedure TearDown; override;
    function PrepareTest (Tracefile: string) : tTCPParser;
  published
    procedure ParseFTPPortCommand;
    procedure ParseTCPACKWithPaddingAndFCS;
    procedure ParseJumboframe;
    procedure ParseCRCProblemFrame;
    procedure ParseMultiPathCapable;
    procedure ParseMultiPathJoin;
  end;

implementation

uses definition_tcp_structures,
     class_rawframe,
     class_loader_pcapng;

(* -------------------------------------- *)

procedure TesttTCPParser.SetUp;

begin
//   FtTCPParser := tTCPParser.Create;
end;

(* -------------------------------------- *)

procedure TesttTCPParser.TearDown;

begin
//   FtTCPParser.Free;
//   FtTCPParser := nil;
end;

(* ------------------------------------------- *)

function TesttTCPParser.PrepareTest(Tracefile: string) : tTCPParser;

var Loader      : tPCAPNGLoader;
    Frame       : tRawFrame;
    Parser      : tTCPParser;
    OpenResult  : boolean;
    i           : integer;

begin
     Loader := tPCAPNGLoader.create (Tracefile);
     OpenResult := Loader.openFile;

     Loader.readTraceFileheader;
     Loader.readBatchFromDisk;

     Frame := Loader.getRawFrame(1);

     ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     Loader.Free;
     Loader := nil;

     CheckEquals (true, Openresult, 'PCAPNGLoader.OpenFile Failed');

     i := 0;

     // Find the ICMP Parser

     repeat
           inc (i)
     until (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_TCP) or (i >= Parsemaster.ParserCount);

     if (Parsemaster.getParserTypeByIndex (i) = NEXT_HEADER_TCP) then
        Parser := (Parsemaster.getParserByIndex (i) as tTCPParser)
     else
         Parser := nil;

     result := Parser;
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseFTPPortCommand;

var TCPParser  : tTCPParser;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\FTP-PortCommand.pcapng');

     if TCPParser <> nil then
        begin
             CheckEquals (27, Parsemaster.RemainingPayloadCopy.Len, 'Remaining Payload Length');
             CheckEquals (true, Parsemaster.HasEthernetFCS, 'Has Ethernet FCS');
             CheckEquals ($91d234c3, Parsemaster.EthernetFCS, 'Ethernet FCS');
             CheckEquals (32, TCPParser.HeaderLength, 'TCP Header Length');
             CheckEquals (27, TCPParser.PayloadLength, 'TCP Payload Length');
             CheckEquals ($8626, TCPParser.Checksum, 'TCP Checksum');
             CheckEquals (1371, TCPParser.SourcePort, 'Source Port');
             CheckEquals (21, TCPParser.DestinationPort, 'Destination Port');
             CheckEquals (2139489536, TCPParser.SequenceNumber.Value, 'Sequence Number');
             CheckEquals (1869092573, TCPParser.AcknowledgeNumber.Value, 'Acknowledge Number');

             if not (TCPParser.TCPOptionsSet = [tcpNOP,tcpTimestamp]) then
                fail ('NOP and Timestamp Option');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseTCPACKWithPaddingAndFCS;

var TCPParser  : tTCPParser;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\TCP-ACKwithPaddingAndFCS.pcapng');

     if TCPParser <> nil then
        begin
             CheckEquals (0, Parsemaster.RemainingPayloadCopy.Len, 'Remaining Payload Length');
             CheckEquals (true, Parsemaster.HasEthernetFCS, 'Has Ethernet FCS');
             CheckEquals ($fc987b0a, Parsemaster.EthernetFCS, 'Ethernet FCS');
             CheckEquals (6, Parsemaster.PaddingByteCount, 'Padding Byte Count');
             CheckEquals (20, TCPParser.HeaderLength, 'TCP Header Length');
             CheckEquals (0, TCPParser.PayloadLength, 'TCP Payload Length');
             CheckEquals ($5276, TCPParser.Checksum, 'TCP Checksum');
             CheckEquals (1375, TCPParser.SourcePort, 'Source Port');
             CheckEquals (20, TCPParser.DestinationPort, 'Destination Port');
             CheckEquals (1789961405, TCPParser.SequenceNumber.Value, 'Sequence Number');
             CheckEquals (1333028994, TCPParser.AcknowledgeNumber.Value, 'Acknowledge Number');

             if not (TCPParser.TCPOptionsSet = []) then
                fail ('No TCP Options');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseJumboframe;

var TCPParser  : tTCPParser;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\TCP-Jumboframe.pcapng');

     if TCPParser <> nil then
        begin
             CheckEquals (8948, Parsemaster.RemainingPayloadCopy.Len, 'Remaining Payload Length');
             CheckEquals (false, Parsemaster.HasEthernetFCS, 'Has Ethernet FCS');
             CheckEquals ($0, Parsemaster.EthernetFCS, 'Ethernet FCS');
             CheckEquals (0, Parsemaster.PaddingByteCount, 'Padding Byte Count');
             CheckEquals (32, TCPParser.HeaderLength, 'TCP Header Length');
             CheckEquals (8948, TCPParser.PayloadLength, 'TCP Payload Length');
             CheckEquals ($f542, TCPParser.Checksum, 'TCP Checksum');
             CheckEquals (1399, TCPParser.SourcePort, 'Source Port');
             CheckEquals (18767, TCPParser.DestinationPort, 'Destination Port');
             CheckEquals (2927326338, TCPParser.SequenceNumber.Value, 'Sequence Number');
             CheckEquals (2205466041, TCPParser.AcknowledgeNumber.Value, 'Acknowledge Number');

             if not (TCPParser.TCPOptionsSet = [tcpNOP,tcpTimestamp]) then
                fail ('NOP and Timestamp Option');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseCRCProblemFrame;

var TCPParser  : tTCPParser;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\TCP-CRCTest1.pcapng');

     if TCPParser <> nil then
        begin
             CheckEquals (268, Parsemaster.RemainingPayloadCopy.Len, 'Remaining Payload Length');
             CheckEquals (false, Parsemaster.HasEthernetFCS, 'Has Ethernet FCS');
             CheckEquals ($0, Parsemaster.EthernetFCS, 'Ethernet FCS');
             CheckEquals (0, Parsemaster.PaddingByteCount, 'Padding Byte Count');
             CheckEquals (32, TCPParser.HeaderLength, 'TCP Header Length');
             CheckEquals (268, TCPParser.PayloadLength, 'TCP Payload Length');
             CheckEquals ($ffd2, TCPParser.Checksum, 'TCP Checksum');
             CheckEquals (1399, TCPParser.SourcePort, 'Source Port');
             CheckEquals (18767, TCPParser.DestinationPort, 'Destination Port');
             CheckEquals (2927335286, TCPParser.SequenceNumber.Value, 'Sequence Number');
             CheckEquals (2205466041, TCPParser.AcknowledgeNumber.Value, 'Acknowledge Number');

             if not (TCPParser.TCPOptionsSet = [tcpNOP,tcpTimestamp]) then
                fail ('NOP and Timestamp Option');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseMultiPathCapable;

var TCPParser  : tTCPParser;
    MPathOpt   : tTCPMultiPathOption;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\TCP-MPathCapable.pcapng');

     if TCPParser <> nil then
        begin
             CheckEquals (0, Parsemaster.RemainingPayloadCopy.Len, 'Remaining Payload Length');
             CheckEquals (false, Parsemaster.HasEthernetFCS, 'Has Ethernet FCS');
             CheckEquals ($0, Parsemaster.EthernetFCS, 'Ethernet FCS');
             CheckEquals (0, Parsemaster.PaddingByteCount, 'Padding Byte Count');
             CheckEquals (44, TCPParser.HeaderLength, 'TCP Header Length');
             CheckEquals (0, TCPParser.PayloadLength, 'TCP Payload Length');
             CheckEquals ($1dab, TCPParser.Checksum, 'TCP Checksum');
             CheckEquals (34522, TCPParser.SourcePort, 'Source Port');
             CheckEquals (80, TCPParser.DestinationPort, 'Destination Port');
             CheckEquals (4149196228, TCPParser.SequenceNumber.Value, 'Sequence Number');
             CheckEquals (0, TCPParser.AcknowledgeNumber.Value, 'Acknowledge Number');

             if not (TCPParser.TCPOptionsSet = [tcpNOP, tcpMSS, tcpSACKPermitted, tcpScalefactor, tcpMPath]) then
                fail ('Wrong options');

             CheckEquals (1, length (TCPParser.MultiPathOptionList), 'Multipath List Length');

             MPathOpt := TCPParser.MultiPathOptionList[0];

             CheckTrue (MPathOpt.subType = mtcpCapable, 'Multipath Option Type is CAPABLE');
             CheckEquals (MPathOpt.TCPVersion, 0, 'TCP Version');
             CheckTrue (MPathOpt.MPathFlags = [tmpChecksumReq, tmpCryptoUseHMAC], 'MultiPath Flags');
             CheckEquals (MPathOpt.SendersKey, 11353787690014689895, 'Senders Key');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

procedure TesttTCPParser.ParseMultiPathJoin;

var TCPParser  : tTCPParser;
    MPathOpt   : tTCPMultiPathOption;

begin
     TCPParser := PrepareTest ('d:\Software Development\Jays Utilities\Trace Wrangler\test\traces\TCP-MPathJoin.pcapng');

     if TCPParser <> nil then
        begin
             if not (TCPParser.TCPOptionsSet = [tcpNOP, tcpMSS, tcpSACKPermitted, tcpScalefactor, tcpMPath]) then
                fail ('Wrong options');

             CheckEquals (1, length (TCPParser.MultiPathOptionList), 'Multipath List Length');

             MPathOpt := TCPParser.MultiPathOptionList[0];

             CheckTrue (MPathOpt.subType = mtcpJoinConnection, 'Multipath Option Type is JoinConnection');
             CheckEquals (MPathOpt.AddressID, 1, 'Multipath AddressID');
             CheckTrue (MPathOpt.MPathFlags = [], 'MultiPath Flags');
             CheckEquals (MPathOpt.ReceiversToken, 1539751264, 'Receivers Token');
             CheckEquals (MPathOpt.SendersRndNr, 1162397042, 'Senders Random Number');
        end
     else
         Fail ('Parser not found');
end;

(* -------------------------------------- *)

initialization
              // Register any test cases with the test runner

              RegisterTest(TesttTCPParser.Suite);
end.

