{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

program tracedbtool;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  System.SysUtils,
  DateUtils,
  {$i}
  class_dbfilemanager in 'storage\class_dbfilemanager.pas',
  class_dirparser in 'base\class_dirparser.pas',
  class_conversationstorage in 'storage\class_conversationstorage.pas';

type tToolCommand = (cNone, cUpdate, cList);

     tParameters = record
                         command        : tToolCommand;
                         discardOrphans : boolean;

                         procedure initialize;
                   end;

(* ------------------------------------------ *)
{ tParameters }
(* ------------------------------------------ *)

procedure tParameters.initialize;

begin
     command := cNone;
     discardOrphans := False;
end;

(* ------------------------------------------ *)
{ Main }
(* ------------------------------------------ *)

procedure OnStartScanOfFile (Sender : tObject; const Status : tProcessingStatus);

var TraceFile  : tTraceFileEntry;

begin
     TraceFile := (Sender as tTraceFileEntry);

     write ('Scanning ' + Tracefile.FileName + '   0%');
end;

(* ------------------------------------------ *)

procedure OnAdvancePercentProcessed (Sender : tObject; const PercentageParsed : integer; var AbortScan : boolean);

begin
     write (#8#8#8#8 + Format ('%3d', [PercentageParsed]) + '%');
end;

(* ------------------------------------------ *)

procedure OnEndScanOfFile (Sender : tObject; const Status : tProcessingStatus);

begin
     writeln;
end;

(* ------------------------------------------ *)

procedure UpdateDatabase (DBFilename : TFileName);

var DirParser     : tDirParser;
    TraceFilename : string;
    TraceFile     : tTraceFileEntry;
    DBParams      : tIntelDBParameters;
    i             : integer;

begin
     DBParams.initialize;

     DirParser := tDirParser.create;
     DirParser.BaseDirectory := ExtractFilePath (DBFilename);
     DirParser.FileMask := '*.*';
     DirParser.execute;

     for i  := 1 to DirParser.FileCount do
         begin
              TraceFilename := DirParser.getFileName(i);

              TraceFile := tTraceFileEntry.create (Tracefilename, 0);

              TraceFile.ReadFileDetails;
         end;
end;

(* ------------------------------------------ *)

procedure ShowFileList (Manager : tDBFileManager);

begin


end;

(* ------------------------------------------ *)

procedure ProcessDatabase (const Filename : TFileName; const Parameters : tParameters);

var Manager : tDBFileManager;

begin
     if FileExists (Filename) then
        begin
             Manager := tDBFileManager.create (Filename);
             Manager.Connect;

             case Parameters.command of
              cUpdate : UpdateDatabase (Filename);
              cList   : ShowFileList (Manager);
             end;

             Manager.Disconnect;
             Manager.Free;
        end
     else
         begin
              writeln ('Error: file not found');
         end;
end;

(* ------------------------------------------ *)

procedure doShowhelp;

begin
     writeln ('This tool helps maintaining TraceIntelDB database files');
     writeln ('Usage: tracedbtool <command> <databasefile> [options]');
     writeln;
     writeln ('Commands:          update');
     writeln ('                   list');
     writeln;
     writeln ('Options:           /discardorphans');
     writeln ('                   /h');
end;

(* ------------------------------------------ *)

procedure main;

var ShowHelp : boolean;
    filename : tFilename;
    params   : tParameters;

begin
     ShowHelp := findCmdLineSwitch ('h') or (ParamCount < 2);

     if not ShowHelp then
        begin
             filename := ParamStr(2);

             if (lowercase (trim (ParamStr (1))) = 'update') then
                params.command := cUpdate
             else
                 params.command := cNone;

             if (lowercase (trim (ParamStr (1))) = 'list') then
                params.command := cList;

             params.initialize;
             params.discardOrphans := findCmdLineSwitch ('discardorphans');

             if not ShowHelp then
                ShowHelp := fileexists (filename);

             ShowHelp := (params.command > cNone);
        end;

     if ShowHelp then
        doshowHelp
     else
         ProcessDatabase (filename, params);
end;

(* ------------------------------------------ *)

begin
     try
        main;
     except
           on E: Exception do
              Writeln(E.ClassName, ': ', E.Message);
     end;
end.

